/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

GtkWidget* create_main_window (void);
GtkWidget* create_open_file_selection (void);
GtkWidget* create_font_selection (void);
GtkWidget* create_about_dialog (void);
GtkWidget* create_save_file_selection (void);
