// Track properties dialog
_("Artist")
_("Track Title")
_("Performer")
_("Band / Album Artist")
_("Album")
_("Date")
_("Track Number")
_("Total Tracks")
_("Genre")
_("Composer")
_("Disc Number")
_("Total Discs")
_("Comment")
_("Encoder / Vendor")
_("Copyright")
_("Location")
_("Subtrack Index")
_("Tag Type(s)")
_("Embedded Cuesheet")
_("Codec")
