#ifndef IO68_DEFAULT_H
#define IO68_DEFAULT_H

#include <sc68/file68_features.h>

enum {
  SPR_DEF = FILE68_SPR_DEF,             /**< Default sampling rate. */
  SPR_MIN = FILE68_SPR_MIN,             /**< Minimal sampling rate. */
  SPR_MAX = FILE68_SPR_MAX              /**< Maximal sampling rate. */
};

#endif
