//
//  SplitterWidget.h
//  deadbeef
//
//  Created by Oleksiy Yakovenko on 20/02/2021.
//  Copyright © 2021 Oleksiy Yakovenko. All rights reserved.
//

#import <Foundation/Foundation.h>
#import "WidgetBase.h"

NS_ASSUME_NONNULL_BEGIN

@interface HSplitterWidget : WidgetBase
@end

@interface VSplitterWidget : HSplitterWidget
@end


NS_ASSUME_NONNULL_END
