/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

GtkWidget* create_mainwin (void);
GtkWidget* create_searchwin (void);
GtkWidget* create_traymenu (void);
GtkWidget* create_progressdlg (void);
GtkWidget* create_helpwindow (void);
GtkWidget* create_trackproperties (void);
GtkWidget* create_editcolumndlg (void);
GtkWidget* create_prefwin (void);
GtkWidget* create_entrydialog (void);
GtkWidget* create_addlocationdlg (void);
GtkWidget* create_groupbydlg (void);
GtkWidget* create_sortbydlg (void);
GtkWidget* create_select_dsp_plugin (void);
GtkWidget* create_tagwritersettings (void);
GtkWidget* create_ctmappingdlg (void);
GtkWidget* create_ctmappingeditdlg (void);
GtkWidget* create_button_properties (void);
GtkWidget* create_select_action (void);
GtkWidget* create_setcustomtitledlg (void);
GtkWidget* create_trkproperties_popup_menu (void);
GtkWidget* create_edit_tag_value_dlg (void);
GtkWidget* create_log_window (void);
GtkWidget* create_rg_scan_progress (void);
GtkWidget* create_rg_scan_results (void);
GtkWidget* create_plugin_list_popup_menu (void);
GtkWidget* create_edit_multiple_tracks_dialog (void);
