//
//  DspPreferencesViewController.h
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 2/26/20.
//  Copyright © 2020 Oleksiy Yakovenko. All rights reserved.
//

#import <Cocoa/Cocoa.h>
#import "ScriptableTableDataSource.h"

NS_ASSUME_NONNULL_BEGIN

@interface DspPreferencesViewController : NSViewController<ScriptableItemDelegate>

@end

NS_ASSUME_NONNULL_END
