//
//  PlaybackPreferencesViewController.h
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 2/25/20.
//  Copyright © 2020 Oleksiy Yakovenko. All rights reserved.
//

#import <Cocoa/Cocoa.h>

NS_ASSUME_NONNULL_BEGIN

@interface SoundPreferencesViewController : NSViewController

- (void)outputDeviceChanged;

@end

NS_ASSUME_NONNULL_END
