/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

GtkWidget* create_converterdlg (void);
GtkWidget* create_convpreset_editor (void);
GtkWidget* create_dsppreset_editor (void);
GtkWidget* create_select_dsp_plugin (void);
GtkWidget* create_preset_list (void);
