#ifndef IDC_STATIC
#define IDC_STATIC (-1)
#endif

#define IDD_INF                                 103
#define IDD_SEL                                 104
#define IDD_CFG                                 105
#define IDC_INF_URI                             1003
#define IDC_INF_FORMAT                          1005
#define IDC_INF_GENRE                           1009
#define IDC_INF_RIPPER                          1022
#define IDC_INF_CONVERTER                       1023
#define IDC_INF_TITLE                           1030
#define IDC_INF_TIME                            1034
#define IDC_INF_ARTIST                          1042
#define IDC_INF_ALBUM                           1045
#define IDC_INF_ASID                            1054
#define IDC_INF_STE                             1055
#define IDC_INF_AMIGA                           1056
#define IDC_INF_YM                              1057
#define IDC_INF_TRACK                           1058
#define IDC_SEL_ALBUM                           1085
#define IDC_SEL_REMEMBER                        1090
#define IDC_CFG_DEFTIME                         40000
#define IDC_INF_YEAR                            40000
#define IDC_SEL_ASID                            40000
#define IDC_CFG_SPRSPIN                         40001
#define IDC_SEL_TRACK                           40001
#define IDC_CFG_USRSPR                          40002
#define IDC_INF_TAGNAME                         40002
#define IDC_CFG_SPR                             40003
#define IDC_INF_TAGVALUE                        40003
#define IDC_CFG_TIMESPIN                        40004
#define IDC_CFG_ASID                            40005
#define IDC_CFG_YMENGINE                        40006
#define IDC_CFG_YMFILTER                        40007
#define IDC_CFG_YMVOLUME                        40008
#define IDC_CFG_AGABLEND                        40009
#define IDC_CFG_AGAFILTER                       40010
#define IDC_CFG_UFI                             40011
#define IDC_CFG_HOOK                            40012
