// Generated by ResEdit 1.6.6
// Copyright (C) 2006-2015
// http://www.resedit.net

#include <windows.h>
#include <commctrl.h>
#include <richedit.h>
#include "resource.h"




//
// Dialog resources
//
LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDD_CFG DIALOG 0, 0, 185, 247
STYLE DS_3DLOOK | DS_CENTER | DS_MODALFRAME | DS_SHELLFONT | WS_CAPTION | WS_VISIBLE | WS_POPUP | WS_SYSMENU
CAPTION "sc68 configuration"
FONT 8, "Ms Shell Dlg"
{
    GROUPBOX        "General", IDC_STATIC, 5, 7, 175, 50, WS_GROUP, WS_EX_LEFT
    CONTROL         "Sampling", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 22, 30, 8, WS_EX_LEFT
    COMBOBOX        IDC_CFG_SPR, 47, 16, 51, 60, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    EDITTEXT        IDC_CFG_USRSPR, 123, 17, 39, 12, ES_AUTOHSCROLL | ES_NUMBER | ES_WANTRETURN, WS_EX_LEFT
    CONTROL         "", IDC_CFG_SPRSPIN, UPDOWN_CLASS, WS_TABSTOP | UDS_ARROWKEYS | UDS_AUTOBUDDY | UDS_SETBUDDYINT, 163, 17, 11, 12, WS_EX_LEFT
    CONTROL         "Default time", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 41, 38, 8, WS_EX_LEFT
    EDITTEXT        IDC_CFG_DEFTIME, 57, 37, 30, 12, ES_AUTOHSCROLL | ES_NUMBER | ES_WANTRETURN, WS_EX_LEFT
    CONTROL         "", IDC_CFG_TIMESPIN, UPDOWN_CLASS, WS_TABSTOP | UDS_ARROWKEYS | UDS_AUTOBUDDY | UDS_SETBUDDYINT, 87, 37, 11, 12, WS_EX_LEFT
    CONTROL         "aSIDifier", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 107, 41, 28, 8, WS_EX_LEFT
    COMBOBOX        IDC_CFG_ASID, 140, 37, 34, 60, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    GROUPBOX        "YM-2149", IDC_STATIC, 5, 67, 175, 70, WS_GROUP, WS_EX_LEFT
    CONTROL         "Simulation model", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 82, 54, 8, WS_EX_LEFT
    COMBOBOX        IDC_CFG_YMENGINE, 67, 77, 105, 60, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    CONTROL         "Filter", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 102, 16, 8, WS_EX_LEFT
    COMBOBOX        IDC_CFG_YMFILTER, 67, 97, 105, 60, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    CONTROL         "Volume model", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 122, 45, 8, WS_EX_LEFT
    COMBOBOX        IDC_CFG_YMVOLUME, 67, 117, 105, 60, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    GROUPBOX        "Amiga", IDC_STATIC, 5, 145, 175, 30, WS_GROUP, WS_EX_LEFT
    CONTROL         "L/R Blend", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 12, 162, 34, 8, WS_EX_LEFT
    CONTROL         "", IDC_CFG_AGABLEND, TRACKBAR_CLASS, WS_TABSTOP | TBS_BOTH, 45, 152, 92, 20, WS_EX_LEFT
    AUTOCHECKBOX    "Filter", IDC_CFG_AGAFILTER, 142, 162, 31, 8, 0, WS_EX_LEFT
    GROUPBOX        "Winamp", IDC_STATIC, 5, 185, 175, 30, WS_GROUP, WS_EX_LEFT
    AUTOCHECKBOX    "Unified file info", IDC_CFG_UFI, 12, 200, 62, 8, 0, WS_EX_LEFT
    AUTOCHECKBOX    "Hook prev/next track", IDC_CFG_HOOK, 90, 200, 84, 8, 0, WS_EX_LEFT
    DEFPUSHBUTTON   "OK", IDOK, 35, 225, 50, 14, 0, WS_EX_LEFT
    PUSHBUTTON      "Cancel", IDCANCEL, 100, 225, 50, 14, 0, WS_EX_LEFT
}



LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDD_INF DIALOG 0, 0, 224, 200
STYLE DS_3DLOOK | DS_CENTER | DS_MODALFRAME | DS_SHELLFONT | WS_CAPTION | WS_VISIBLE | WS_POPUP | WS_SYSMENU
CAPTION "sc68 file info"
FONT 8, "Ms Shell Dlg"
{
    CONTROL         "", IDC_INF_URI, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE | SS_SUNKEN, 7, 8, 210, 12, WS_EX_LEFT
    CONTROL         "Format", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 8, 27, 22, 8, WS_EX_LEFT
    CONTROL         "Genre", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 37, 27, 20, 8, WS_EX_LEFT
    CONTROL         "Hardware", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 107, 27, 32, 8, WS_EX_LEFT
    CONTROL         "", IDC_INF_FORMAT, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 12, 37, 20, 12, WS_EX_STATICEDGE
    CONTROL         "", IDC_INF_GENRE, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 42, 37, 65, 12, WS_EX_STATICEDGE
    AUTORADIOBUTTON "Amiga", IDC_INF_AMIGA, 112, 37, 32, 8, NOT WS_VISIBLE, WS_EX_LEFT
    AUTORADIOBUTTON "YM", IDC_INF_YM, 112, 37, 32, 8, 0, WS_EX_LEFT
    AUTORADIOBUTTON "STE", IDC_INF_STE, 144, 37, 32, 8, 0, WS_EX_LEFT
    AUTORADIOBUTTON "aSID", IDC_INF_ASID, 177, 37, 32, 8, 0, WS_EX_LEFT
    CONTROL         "Artist", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 7, 57, 16, 8, WS_EX_LEFT
    CONTROL         "Album", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 112, 57, 20, 8, WS_EX_LEFT
    CONTROL         "", IDC_INF_ARTIST, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 12, 67, 95, 12, WS_EX_STATICEDGE
    CONTROL         "", IDC_INF_ALBUM, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 117, 67, 100, 12, WS_EX_STATICEDGE
    CONTROL         "#", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 9, 87, 8, 8, WS_EX_LEFT
    CONTROL         "Time", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 40, 87, 16, 8, WS_EX_LEFT
    CONTROL         "Title", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 67, 87, 14, 8, WS_EX_LEFT
    COMBOBOX        IDC_INF_TRACK, 12, 97, 27, 30, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    CONTROL         "", IDC_INF_TIME, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 45, 97, 25, 12, WS_EX_STATICEDGE
    CONTROL         "", IDC_INF_TITLE, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 75, 97, 142, 12, WS_EX_STATICEDGE
    CONTROL         "Ripper", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 7, 117, 22, 8, WS_EX_LEFT
    CONTROL         "Converter", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 97, 117, 32, 8, WS_EX_LEFT
    CONTROL         "Year", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 187, 117, 16, 8, WS_EX_LEFT
    CONTROL         "", IDC_INF_RIPPER, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 14, 127, 80, 12, WS_EX_STATICEDGE
    CONTROL         "", IDC_INF_CONVERTER, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 102, 127, 80, 12, WS_EX_STATICEDGE
    CONTROL         "", IDC_INF_YEAR, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 190, 127, 26, 12, WS_EX_STATICEDGE
    CONTROL         "All metatags", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 7, 147, 39, 8, WS_EX_LEFT
    COMBOBOX        IDC_INF_TAGNAME, 10, 158, 65, 34, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    CONTROL         "", IDC_INF_TAGVALUE, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_CENTERIMAGE, 80, 158, 136, 12, WS_EX_STATICEDGE
    DEFPUSHBUTTON   "OK", IDOK, 87, 179, 50, 14, 0, WS_EX_LEFT
}



LANGUAGE LANG_NEUTRAL, SUBLANG_NEUTRAL
IDD_SEL DIALOG 0, 0, 216, 88
STYLE DS_3DLOOK | DS_CENTER | DS_MODALFRAME | DS_SHELLFONT | WS_CAPTION | WS_VISIBLE | WS_POPUP | WS_SYSMENU
CAPTION "sc68 track selection"
FONT 8, "Ms Shell Dlg"
{
    CONTROL         "", IDC_SEL_ALBUM, WC_STATIC, NOT WS_GROUP | SS_SIMPLE | SS_NOPREFIX | SS_SUNKEN, 7, 12, 156, 15, WS_EX_LEFT
    COMBOBOX        IDC_SEL_TRACK, 7, 35, 156, 30, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    CONTROL         "aSIDifier", IDC_STATIC, WC_STATIC, NOT WS_GROUP | SS_SIMPLE, 178, 24, 28, 8, WS_EX_LEFT
    COMBOBOX        IDC_SEL_ASID, 172, 35, 34, 30, WS_TABSTOP | CBS_DROPDOWNLIST | CBS_HASSTRINGS, WS_EX_LEFT
    AUTOCHECKBOX    "Remember", IDC_SEL_REMEMBER, 156, 65, 50, 14, 0, WS_EX_LEFT
    PUSHBUTTON      "OK", IDOK, 28, 65, 50, 14, 0, WS_EX_LEFT
    PUSHBUTTON      "Cancel", IDCANCEL, 83, 65, 50, 14, 0, WS_EX_LEFT
}
