//
//  MediaLibraryPreferencesViewController.h
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 8/24/20.
//  Copyright © 2020 Oleksiy Yakovenko. All rights reserved.
//

#import <Cocoa/Cocoa.h>

NS_ASSUME_NONNULL_BEGIN

@interface MediaLibraryPreferencesViewController : NSViewController

@property (nonatomic,readonly) BOOL isAvailable;

@end

NS_ASSUME_NONNULL_END
