//
//  AlbumArtImageView.h
//  ddbcore
//
//  Created by Oleksiy Yakovenko on 21/03/2021.
//  Copyright © 2021 Oleksiy Yakovenko. All rights reserved.
//

#import <Cocoa/Cocoa.h>

NS_ASSUME_NONNULL_BEGIN

/// A wrapper to allow the NSImageView to shrink to smaller size than the image size
@interface AlbumArtImageView : NSImageView

@end

NS_ASSUME_NONNULL_END
