#ifndef _UNIX_SIDINT_H
#define _UNIX_SIDINT_H 1
#ifndef _GENERATED_STDINT_H
#define _GENERATED_STDINT_H "libsidplay 2.1.0"
/* generated using a gnu compiler version 2.96 */

#include <stdint.h>


/* system headers have good uint64_t */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
#endif

  /* once */
#endif
#endif
