//
//  PluginActionMenuItem.m
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 2/23/20.
//  Copyright © 2020 Oleksiy Yakovenko. All rights reserved.
//

#import "PluginActionMenuItem.h"

@implementation PluginActionMenuItem

@end
