//
//  medialibwidget.h
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 11/08/2021.
//  Copyright © 2021 Oleksiy Yakovenko. All rights reserved.
//

#ifndef medialibwidget_h
#define medialibwidget_h

#include "../widgets.h"

ddb_gtkui_widget_t *
w_medialib_viewer_create (void);

#endif /* medialibwidget_h */
