//
//  ScopeWidget.h
//  deadbeef
//
//  Created by Oleksiy Yakovenko on 18/10/2021.
//  Copyright © 2021 Oleksiy Yakovenko. All rights reserved.
//

#import "WidgetBase.h"

NS_ASSUME_NONNULL_BEGIN

@interface ScopeWidget : WidgetBase

@end

NS_ASSUME_NONNULL_END
