//
//  SeekbarOverlay.h
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 3/13/20.
//  Copyright © 2020 Oleksiy Yakovenko. All rights reserved.
//

#import <Cocoa/Cocoa.h>

NS_ASSUME_NONNULL_BEGIN

IB_DESIGNABLE
@interface SeekbarOverlay : NSView

@property (nonatomic) NSString *text;

@end

NS_ASSUME_NONNULL_END
