// includes all chip-headers

#include "sn764intf.h"
#include "2413intf.h"
#include "2612intf.h"
#include "2151intf.h"
#include "segapcm.h"
#include "rf5c68.h"
#include "2203intf.h"
#include "2608intf.h"
#include "2610intf.h"
#include "3812intf.h"
#include "3526intf.h"
#include "8950intf.h"
#include "262intf.h"
#include "ymz280b.h"
#include "ymf271.h"
#include "ymf278b.h"
#include "scd_pcm.h"
#include "pwm.h"
#include "ay_intf.h"
#include "dac_control.h"
#include "gb.h"
#include "nes_intf.h"
#include "multipcm.h"
#include "upd7759.h"
#include "okim6258.h"
#include "okim6295.h"
#include "k051649.h"
#include "k054539.h"
#include "c6280intf.h"
#include "c140.h"
#include "k053260.h"
#include "pokey.h"
#include "qsound.h"
#include "scsp.h"
#include "ws_audio.h"
#include "vsu.h"
#include "saa1099.h"
#include "es5503.h"
#include "es5506.h"
#include "x1_010.h"
#include "c352.h"
#include "iremga20.h"
