/* line7.c - EMU68 generated code by
 * gen68 2013-05-25 00:26:49
 * Copyright (c) 1998-2013 Benjamin Gerard
 *
 * $Id: line7.c 458 2014-07-03 05:29:47Z benjihan $
 */

/* Line 7: MOVEQ */

DECL_LINE68(line700)
{
  /* MOVEQ #N,Dn [0..7] */
  int68_t a = reg0   +0;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line701)
{
  /* MOVEQ #N,Dn [8..15] */
  int68_t a = reg0   +8;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line702)
{
  /* MOVEQ #N,Dn [16..23] */
  int68_t a = reg0  +16;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line703)
{
  /* MOVEQ #N,Dn [24..31] */
  int68_t a = reg0  +24;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line704)
{
  /* MOVEQ #N,Dn [32..39] */
  int68_t a = reg0  +32;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line705)
{
  /* MOVEQ #N,Dn [40..47] */
  int68_t a = reg0  +40;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line706)
{
  /* MOVEQ #N,Dn [48..55] */
  int68_t a = reg0  +48;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line707)
{
  /* MOVEQ #N,Dn [56..63] */
  int68_t a = reg0  +56;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line708)
{
  /* MOVEQ #N,Dn [64..71] */
  int68_t a = reg0  +64;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line709)
{
  /* MOVEQ #N,Dn [72..79] */
  int68_t a = reg0  +72;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70A)
{
  /* MOVEQ #N,Dn [80..87] */
  int68_t a = reg0  +80;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70B)
{
  /* MOVEQ #N,Dn [88..95] */
  int68_t a = reg0  +88;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70C)
{
  /* MOVEQ #N,Dn [96..103] */
  int68_t a = reg0  +96;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70D)
{
  /* MOVEQ #N,Dn [104..111] */
  int68_t a = reg0 +104;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70E)
{
  /* MOVEQ #N,Dn [112..119] */
  int68_t a = reg0 +112;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line70F)
{
  /* MOVEQ #N,Dn [120..127] */
  int68_t a = reg0 +120;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line710)
{
  /* MOVEQ #N,Dn [-128..-121] */
  int68_t a = reg0 -128;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line711)
{
  /* MOVEQ #N,Dn [-120..-113] */
  int68_t a = reg0 -120;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line712)
{
  /* MOVEQ #N,Dn [-112..-105] */
  int68_t a = reg0 -112;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line713)
{
  /* MOVEQ #N,Dn [-104..-97] */
  int68_t a = reg0 -104;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line714)
{
  /* MOVEQ #N,Dn [-96..-89] */
  int68_t a = reg0  -96;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line715)
{
  /* MOVEQ #N,Dn [-88..-81] */
  int68_t a = reg0  -88;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line716)
{
  /* MOVEQ #N,Dn [-80..-73] */
  int68_t a = reg0  -80;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line717)
{
  /* MOVEQ #N,Dn [-72..-65] */
  int68_t a = reg0  -72;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line718)
{
  /* MOVEQ #N,Dn [-64..-57] */
  int68_t a = reg0  -64;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line719)
{
  /* MOVEQ #N,Dn [-56..-49] */
  int68_t a = reg0  -56;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71A)
{
  /* MOVEQ #N,Dn [-48..-41] */
  int68_t a = reg0  -48;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71B)
{
  /* MOVEQ #N,Dn [-40..-33] */
  int68_t a = reg0  -40;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71C)
{
  /* MOVEQ #N,Dn [-32..-25] */
  int68_t a = reg0  -32;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71D)
{
  /* MOVEQ #N,Dn [-24..-17] */
  int68_t a = reg0  -24;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71E)
{
  /* MOVEQ #N,Dn [-16..-9] */
  int68_t a = reg0  -16;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line71F)
{
  /* MOVEQ #N,Dn [-8..-1] */
  int68_t a = reg0   -8;
  REG68.d[reg9] = (u32)a;
  MOVEL(a);
}

DECL_LINE68(line720)
{
  ILLEGAL; /* op:040 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line721)
{
  ILLEGAL; /* op:041 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line722)
{
  ILLEGAL; /* op:042 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line723)
{
  ILLEGAL; /* op:043 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line724)
{
  ILLEGAL; /* op:044 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line725)
{
  ILLEGAL; /* op:045 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line726)
{
  ILLEGAL; /* op:046 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line727)
{
  ILLEGAL; /* op:047 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line728)
{
  ILLEGAL; /* op:050 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line729)
{
  ILLEGAL; /* op:051 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72A)
{
  ILLEGAL; /* op:052 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72B)
{
  ILLEGAL; /* op:053 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72C)
{
  ILLEGAL; /* op:054 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72D)
{
  ILLEGAL; /* op:055 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72E)
{
  ILLEGAL; /* op:056 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line72F)
{
  ILLEGAL; /* op:057 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line730)
{
  ILLEGAL; /* op:060 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line731)
{
  ILLEGAL; /* op:061 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line732)
{
  ILLEGAL; /* op:062 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line733)
{
  ILLEGAL; /* op:063 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line734)
{
  ILLEGAL; /* op:064 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line735)
{
  ILLEGAL; /* op:065 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line736)
{
  ILLEGAL; /* op:066 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line737)
{
  ILLEGAL; /* op:067 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line738)
{
  ILLEGAL; /* op:070 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line739)
{
  ILLEGAL; /* op:071 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73A)
{
  ILLEGAL; /* op:072 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73B)
{
  ILLEGAL; /* op:073 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73C)
{
  ILLEGAL; /* op:074 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73D)
{
  ILLEGAL; /* op:075 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73E)
{
  ILLEGAL; /* op:076 */
  assert(EMU68_BREAK);
}

DECL_LINE68(line73F)
{
  ILLEGAL; /* op:077 */
  assert(EMU68_BREAK);
}

