//
//  SpectrumAnalyzerSettings.m
//  DeaDBeeF
//
//  Created by Oleksiy Yakovenko on 28/07/2021.
//  Copyright © 2021 Oleksiy Yakovenko. All rights reserved.
//

#import "SpectrumAnalyzerSettings.h"

@implementation SpectrumAnalyzerSettings
@end
