TIMEDB_ENTRY( 000f819c,  0,  73089,       YM+TA), /* 03713 - Chris of TGE - Combo Racer Cracktro.sndh */
TIMEDB_ENTRY( 00411850,  0,   8277,          YM), /* 02509 - Crazy Q - Crators-DHS-YMR - Monolouge(1ch Minimalism).sndh */
TIMEDB_ENTRY( 005a16ee,  0,   8705,          YM), /* 01049 - Dr Beat - NTM - Menu.sndh */
TIMEDB_ENTRY( 00661cb2,  0,   9600,          YM), /* 01537 - Ralph - Ralph 1.sndh */
TIMEDB_ENTRY( 007288df,  0, 170122,       YM+TA), /* 03940 - Unknown - Bangkok Knights.sndh */
TIMEDB_ENTRY( 00768743,  0,   9849,    YM+TA+TB), /* 01138 - Tyan - Julia 2.sndh */
TIMEDB_ENTRY( 007b01fb,  0,   5374,          YM), /* 03197 - Allister Brimble - Arcade Classix - Missile Command.sndh */
TIMEDB_ENTRY( 008734d8,  0,  19968,          YM), /* 02166 - Mad Max - Cybernoid.sndh */
TIMEDB_ENTRY( 0096ba8c,  0, 154881,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  1,  51757,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  2,  27873,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  3,  93513,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  4,  20065,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  5,  80060,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 0096ba8c,  6, 165812,       YM+TA), /* 03789 - Roggie - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 00be4e4f,  0,   2555, YM+TA+TB+TD), /* 01122 - Toodeloo - Test 3.sndh */
TIMEDB_ENTRY( 00c16593,  0,   3054,          YM), /* 00742 - Sean Conran - Duckula.sndh */
TIMEDB_ENTRY( 00c5a987,  0,  50526,       YM+TA), /* 04823 - Twitch - Straight Up.sndh */
TIMEDB_ENTRY( 00d3d382,  0,   2688,          YM), /* 01170 - Sally - Mindbend.sndh */
TIMEDB_ENTRY( 00db87ff,  0,   8058,          YM), /* 02420 - 505 - Checkpoint - 8x8 Hero.sndh */
TIMEDB_ENTRY( 00e0b4f8,  0,  24580,       YM+TD), /* 03991 - XyNo - Hip STfucck Hop !.sndh */
TIMEDB_ENTRY( 00e2f6c3,  0,   3842,          YM), /* 02950 - Justin Scharvona - Robozone.sndh */
TIMEDB_ENTRY( 00e2f6c3,  1,   5764,          YM), /* 02950 - Justin Scharvona - Robozone.sndh */
TIMEDB_ENTRY( 00e2f6c3,  2,   7301,          YM), /* 02950 - Justin Scharvona - Robozone.sndh */
TIMEDB_ENTRY( 00e2f6c3,  3,   6533,          YM), /* 02950 - Justin Scharvona - Robozone.sndh */
TIMEDB_ENTRY( 00f4ae42,  0,   2877,          YM), /* 02967 - Unknown - Sleeping Gods Lie.sndh */
TIMEDB_ENTRY( 00f64fd1,  0,   2877,       YM+TD), /* 04228 - 505 - Checkpoint - Siebenschlaefer (unreleased).sndh */
TIMEDB_ENTRY( 011166dd,  0,  35840,    YM+TB+TD), /* 03378 - STU - STEREOID REMIX.sndh */
TIMEDB_ENTRY( 0124a56a,  0,   3069,    YM+TA+TD), /* 01858 - Lotek Style - Loadatune.sndh */
TIMEDB_ENTRY( 0145aae3,  0,   8225,          YM), /* 03892 - Unknown - Rumores.sndh */
TIMEDB_ENTRY( 014e6bf0,  0,   4608,       YM+TD), /* 03311 - Tomchi - 1st Intro.sndh */
TIMEDB_ENTRY( 01610ab4,  0,   8057, YM+TA+TB+TD), /* 01943 - MSG - Chu Chu Rocket - Challenge.sndh */
TIMEDB_ENTRY( 016f4cd6,  0,   6144,    YM+TA+TD), /* 02266 - Tao - Cream - Stay away (Steps).sndh */
TIMEDB_ENTRY( 018a6862,  0,  35038,       YM+TA), /* 03007 - Mad Max - Great Giana Sisters Title Music.sndh */
TIMEDB_ENTRY( 0192b02e,  0,  10240,       YM+TA), /* 04116 - Tao - The Arrival - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 01972878,  0,   8192,          YM), /* 01221 - ENS Music - ENS 4.sndh */
TIMEDB_ENTRY( 01b879d6,  0,    315,       YM+TA), /* 02642 - Miggy - Roar(Unfinished).sndh */
TIMEDB_ENTRY( 01d08df0,  0,   4176,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  1,   8160,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  2,   4407,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  3,   4320,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  4,   3276,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  5,    673,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  6,   1638,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  7,   6720,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  8,   4704,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01d08df0,  9,   2880,          YM), /* 00174 - Martin Walker - Atomic Robokid.sndh */
TIMEDB_ENTRY( 01f8f345,  0,   4000,          YM), /* 01548 - Rhino - Torment Christmas 2000 Intro.sndh */
TIMEDB_ENTRY( 01fa9141,  0,  37959,          YM), /* 00328 - Gary Wheaton - Solar Gammon.sndh */
TIMEDB_ENTRY( 01fbd6e2,  0,   4221,       YM+TA), /* 02436 - Beast - Oldie.sndh */
TIMEDB_ENTRY( 01fd87a5,  0,   6144,          YM), /* 00598 - Chris Mad - Overdrive Fullscreen.sndh */
TIMEDB_ENTRY( 02025205,  0,  10086,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  1,    101,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  2,    101,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  3,      7,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  4,      6,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  5,     16,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  6,     26,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 02025205,  7,     13,          YM), /* 02725 - Jeremy Das - Burger Man.sndh */
TIMEDB_ENTRY( 020b6a9e,  0,  13727, YM+TA+TB+TD), /* 03016 - Dubmood - RaZoR 1911 - Purple Motions 2nd Reality.sndh */
TIMEDB_ENTRY( 022cea81,  0,   9849, YM+TA+TB+TD), /* 01941 - MSG - Chu Chu Rocket - Battle.sndh */
TIMEDB_ENTRY( 022de5dc,  0,   9986,          YM), /* 01776 - Lap - Lap 16.sndh */
TIMEDB_ENTRY( 024cf349,  0,   3199,          YM), /* 04580 - blind io - apologies all around.sndh */
TIMEDB_ENTRY( 024f7804,  0,  16270,          YM), /* 00735 - Sean Connolly - Guy Spy.sndh */
TIMEDB_ENTRY( 024f7804,  1,  12528,          YM), /* 00735 - Sean Connolly - Guy Spy.sndh */
TIMEDB_ENTRY( 024f7804,  2,  13056,          YM), /* 00735 - Sean Connolly - Guy Spy.sndh */
TIMEDB_ENTRY( 024f7804,  3,  13538,          YM), /* 00735 - Sean Connolly - Guy Spy.sndh */
TIMEDB_ENTRY( 024f7804,  4,  10608,          YM), /* 00735 - Sean Connolly - Guy Spy.sndh */
TIMEDB_ENTRY( 024ff46a,  0,   9978, YM+TA+TB+TD), /* 01835 - Lotek Style - Burning Chrome.sndh */
TIMEDB_ENTRY( 027447f2,  0,  11983, YM+TA+TB+TD), /* 01856 - Lotek Style - Lotusflower.sndh */
TIMEDB_ENTRY( 02b1f999,  0,   6876,          YM), /* 00302 - Gary Wheaton - Colours.sndh */
TIMEDB_ENTRY( 02efdd55,  0,  18048,          YM), /* 01650 - Jedi - Megatizer Intro.sndh */
TIMEDB_ENTRY( 030ad120,  0,   8448,       YM+TD), /* 03222 - Tomchi - Sinus Screen.sndh */
TIMEDB_ENTRY( 0310839a,  0,   5376,          YM), /* 04639 - Crawdaddy - Dropzone.sndh */
TIMEDB_ENTRY( 032b08a1,  0,   4378,          YM), /* 03157 - n-a - SAILORS' BOOGIE.sndh */
TIMEDB_ENTRY( 0343ebe8,  0,   9971, YM+TA+TB+TD), /* 03618 - gwEm - Ex in my Bed.sndh */
TIMEDB_ENTRY( 03527bd9,  0,   4095,          YM), /* 00029 - 505 - Rectangular Waves.sndh */
TIMEDB_ENTRY( 037aecb0,  0,  14976,    YM+TA+TD), /* 03314 - SH3-RG - Saboteur 2.sndh */
TIMEDB_ENTRY( 03834eef,  0,  11524,    YM+TA+TD), /* 02298 - Tao - Cream - Pop 2149 (Alive 6).sndh */
TIMEDB_ENTRY( 03840236,  0,   4031,    YM+TA+TD), /* 03339 - marcer - jambala 8 mix.sndh */
TIMEDB_ENTRY( 038c2a30,  0,  16030,          YM), /* 02044 - Mad Max - Masters Of Magic.sndh */
TIMEDB_ENTRY( 038c2a30,  1,    574,          YM), /* 02044 - Mad Max - Masters Of Magic.sndh */
TIMEDB_ENTRY( 038c2a30,  2,   1153,          YM), /* 02044 - Mad Max - Masters Of Magic.sndh */
TIMEDB_ENTRY( 03b2ad1f,  0,   1530, YM+TA+TB+TD), /* 01945 - MSG - Chu Chu Rocket - Hurry Up.sndh */
TIMEDB_ENTRY( 03b4dc3b,  0,  28160,          YM), /* 01419 - Sonic - Aura.sndh */
TIMEDB_ENTRY( 03c1b536,  0,   3200,       YM+TA), /* 04589 - Illusion - Pure Energy 5.sndh */
TIMEDB_ENTRY( 03d0fbe6,  0,   1293,          YM), /* 02966 - Unknown - Quiz Construction Kit.sndh */
TIMEDB_ENTRY( 03d0fbe6,  1,   4042,          YM), /* 02966 - Unknown - Quiz Construction Kit.sndh */
TIMEDB_ENTRY( 03d0fbe6,  2,   2730,          YM), /* 02966 - Unknown - Quiz Construction Kit.sndh */
TIMEDB_ENTRY( 03e4f84a,  0,   2319,          YM), /* 03065 - Chris French - Deck the Halls.sndh */
TIMEDB_ENTRY( 0431a07d,  0,   2239,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 0431a07d,  1,    767,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 0431a07d,  2,   1455,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 0431a07d,  3,    181,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 0431a07d,  4,    191,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 0431a07d,  5,    950,          YM), /* 02447 - Tony Williams - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 043a3a3f,  0,   4231,          YM), /* 02408 - Epic - Through The Never.sndh */
TIMEDB_ENTRY( 04481e41,  0,   1788,       YM+TA), /* 04237 - 505 - 21st.sndh */
TIMEDB_ENTRY( 04497496,  0,   1785, YM+TA+TB+TD), /* 01955 - MSG - Godpey - Awards.sndh */
TIMEDB_ENTRY( 04563641,  0,  13312,          YM), /* 03216 - Tomchi - Endck.sndh */
TIMEDB_ENTRY( 0467f523,  0,  69588,       YM+TA), /* 03944 - Unknown - Dyer Maker.sndh */
TIMEDB_ENTRY( 0473cf3e,  0,  11904,          YM), /* 03258 - T1000 - The Beast Club - TS 6.sndh */
TIMEDB_ENTRY( 047b4e0f,  0,   5632,    YM+TA+TD), /* 02261 - Tao - Cream - Prelude (Steps).sndh */
TIMEDB_ENTRY( 047bfb94,  0,   7040,          YM), /* 00130 - David Whittaker - Renegade.sndh */
TIMEDB_ENTRY( 048e952e,  0,   7120,       YM+TD), /* 02373 - Nexus 6 - Disharmonica.sndh */
TIMEDB_ENTRY( 04ab02cc,  0,   6143,       YM+TD), /* 03361 - marcer - heroes intro.sndh */
TIMEDB_ENTRY( 04b12049,  0,  13881, YM+TA+TB+TD), /* 04327 - 505-cpt^pdx - Z.sndh */
TIMEDB_ENTRY( 04bc31b2,  0,  12096,       YM+TA), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  1,  20931,       YM+TA), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  2,  12096,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  3,  11904,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  4,   4480,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  5,   6400,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  6,   5396,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  7,   5963,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  8,   6784,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2,  9,   5888,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2, 10,   3360,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2, 11,   2886,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2, 12,   2256,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2, 13,   3077,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04bc31b2, 14,   6664,          YM), /* 02201 - Mad Max (Jochen Hippel) - The Seven Gates of Jambala.sndh */
TIMEDB_ENTRY( 04c86201,  0,   7876,          YM), /* 02514 - Crazy Q - DHS^Creators - Zipp Zapp Space Attack.sndh */
TIMEDB_ENTRY( 04c9de9e,  0,   1789,       YM+TA), /* 04106 - 505-cpt^dox - arping.sndh */
TIMEDB_ENTRY( 04cf71da,  0,    445,          YM), /* 02461 - Iso - Art, I presume?.sndh */
TIMEDB_ENTRY( 04d2a4e4,  0,   6047,          YM), /* 04753 - 505 - Sommarhack 2015 invite.sndh */
TIMEDB_ENTRY( 04f8965b,  0,   3072,          YM), /* 02053 - Mad Max - Small Title.sndh */
TIMEDB_ENTRY( 04f9af8d,  0,   5840,       YM+TD), /* 02393 - Nexus 6 - Praximate Digi.sndh */
TIMEDB_ENTRY( 05259830,  0,   9830,          YM), /* 00579 - Jas C Brooke - Vixen.sndh */
TIMEDB_ENTRY( 05259830,  1,   2048,          YM), /* 00579 - Jas C Brooke - Vixen.sndh */
TIMEDB_ENTRY( 05259830,  2,     81,          YM), /* 00579 - Jas C Brooke - Vixen.sndh */
TIMEDB_ENTRY( 053b080a,  0,  17921,    YM+TA+TD), /* 02195 - Tao - Cream - ..do things (intro).sndh */
TIMEDB_ENTRY( 05462254,  0,   7584,    YM+TB+TD), /* 04765 - 505 - Pulse (compo edit).sndh */
TIMEDB_ENTRY( 05532f21,  0,   6655,          YM), /* 03875 - Gary Wheaton - Juandeho.sndh */
TIMEDB_ENTRY( 055f5d72,  0,  36096, YM+TA+TB+TD), /* 02541 - SCHMX & STU - Q DEDICATION.sndh */
TIMEDB_ENTRY( 056a61e5,  0,  17664,          YM), /* 02903 - Techno - Puppet Show Intro.sndh */
TIMEDB_ENTRY( 056f21c8,  0,   5378,          YM), /* 00368 - Steve Barrett - Treasure Island Dizzy.sndh */
TIMEDB_ENTRY( 056f21c8,  1,   3366,          YM), /* 00368 - Steve Barrett - Treasure Island Dizzy.sndh */
TIMEDB_ENTRY( 056f21c8,  2,    242,          YM), /* 00368 - Steve Barrett - Treasure Island Dizzy.sndh */
TIMEDB_ENTRY( 056f21c8,  3,    242,          YM), /* 00368 - Steve Barrett - Treasure Island Dizzy.sndh */
TIMEDB_ENTRY( 0574488b,  0,  18434,          YM), /* 03489 - Donald Campbell - Rolling Thunder.sndh */
TIMEDB_ENTRY( 0574488b,  1,   2114,          YM), /* 03489 - Donald Campbell - Rolling Thunder.sndh */
TIMEDB_ENTRY( 057595ad,  0,   6142,          YM), /* 01307 - Matt Furniss - Manchester United Europe.sndh */
TIMEDB_ENTRY( 0589d640,  0,   5760,       YM+TA), /* 00918 - Dark Angel - Blood Title.sndh */
TIMEDB_ENTRY( 0597da5f,  0,      8,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  1,     14,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  2,    507,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  3,     14,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  4,     30,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  5,     36,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  6,     35,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 0597da5f,  8,     26,          YM), /* 04494 - Daniel Decourdemanche - Avalon.sndh */
TIMEDB_ENTRY( 05a9b513,  0,    453,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  1,     91,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  2,    213,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  3,    112,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  4,    400,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  5,    684,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05a9b513,  6,    438,          YM), /* 04537 - Darren Ithell - Fun School 4 (Under 5).sndh */
TIMEDB_ENTRY( 05cc2736,  0,   6145,          YM), /* 03307 - Orall Cornelius - Pyramax.sndh */
TIMEDB_ENTRY( 05cc2736,  1,   2305,          YM), /* 03307 - Orall Cornelius - Pyramax.sndh */
TIMEDB_ENTRY( 05cc2736,  2,    769,          YM), /* 03307 - Orall Cornelius - Pyramax.sndh */
TIMEDB_ENTRY( 05cc2736,  3,    897,          YM), /* 03307 - Orall Cornelius - Pyramax.sndh */
TIMEDB_ENTRY( 05d290a5,  0,  16696, YM+TA+TB+TD), /* 02095 - Mad Max - Sanxion Loader - Sid.sndh */
TIMEDB_ENTRY( 05d9ec36,  0,   8928,       YM+TA), /* 01752 - Kanewood - Phantom Demo - End.sndh */
TIMEDB_ENTRY( 05e13969,  0,    638,          YM), /* 04528 - Bobby Kirchoff - Fang Die Ratte.sndh */
TIMEDB_ENTRY( 05e18cd3,  0,  13740,          YM), /* 00737 - Sean Connolly - No Mercy.sndh */
TIMEDB_ENTRY( 05efff3f,  0,  80585,       YM+TA), /* 04024 - Roggie - Variations.sndh */
TIMEDB_ENTRY( 05f8cadb,  0,  44513,    YM+TA+TD), /* 02227 - Tao - Cream - Credits (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 061beb1a,  0,   8651, YM+TA+TB+TD), /* 03625 - 505-checkpoint - Slowslip.sndh */
TIMEDB_ENTRY( 062aa125,  0,   3068,    YM+TA+TD), /* 04259 - 505-cpt^pdx - foolish.sndh */
TIMEDB_ENTRY( 062aa850,  0,   2688,          YM), /* 03498 - Shamen - Certified Dope - Menu.sndh */
TIMEDB_ENTRY( 0660bff2,  0,   6155,       YM+TD), /* 02687 - Donald Fakk - Syntax Terror - Loader (Chip).sndh */
TIMEDB_ENTRY( 066db576,  0,    105,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  1,     93,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  2,    139,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  3,    130,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  4,     79,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  5,    133,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  6,    133,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  7,    176,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  8,    177,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576,  9,    170,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 10,     12,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 11,    145,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 12,     93,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 13,    131,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 14,     91,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 15,    467,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 16,   2303,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 17,    124,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 18,    145,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 19,    181,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 20,     58,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 21,   4030,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 066db576, 22,    105,          YM), /* 00018 - 4-Mat - Spot.sndh */
TIMEDB_ENTRY( 0680ba94,  0,  11839,          YM), /* 03444 - Per Almered (XiA) 2009 - (S)ave (T)he (E)arth.sndh */
TIMEDB_ENTRY( 0683027e,  0,  77491,       YM+TA), /* 04665 - Robert Wolf - Bilinski.sndh */
TIMEDB_ENTRY( 06846733,  0,   2685,          YM), /* 01523 - Iso - Faster and Faster.sndh */
TIMEDB_ENTRY( 068751ac,  0,   3626,          YM), /* 03127 - n-a - Joy to the World.sndh */
TIMEDB_ENTRY( 06899b0e,  0,  13824,          YM), /* 01815 - Loony - Loony.sndh */
TIMEDB_ENTRY( 068c5209,  0,   4032,          YM), /* 00605 - Chris Mad - Phalanx Crazy.sndh */
TIMEDB_ENTRY( 0695fcf3,  0,   4607, YM+TA+TB+TD), /* 04305 - 505-Checkpoint - sherlock.sndh */
TIMEDB_ENTRY( 069b3159,  0,   3840, YM+TA+TB+TD), /* 00187 - 505 - Posh - Part 4.sndh */
TIMEDB_ENTRY( 06a00646,  0,   2076,          YM), /* 04721 - Richard Jacques - Tron.sndh */
TIMEDB_ENTRY( 06a6d743,  0,   3996,          YM), /* 01104 - Thomas - Flu 4kb Intro.sndh */
TIMEDB_ENTRY( 06a74be7,  0,  13824,          YM), /* 02026 - Mad Max - Crazy Comets.sndh */
TIMEDB_ENTRY( 06a74be7,  1,   3264,          YM), /* 02026 - Mad Max - Crazy Comets.sndh */
TIMEDB_ENTRY( 06a95d81,  0,  42851,    YM+TA+TD), /* 02337 - Tao - Cream - One On One (UCM 11).sndh */
TIMEDB_ENTRY( 06b06a33,  0,   3072,    YM+TB+TD), /* 04317 - 505 - Checkpoint - Tambumbi (Unrel.).sndh */
TIMEDB_ENTRY( 06b5c941,  0,  18944,       YM+TA), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  1,   4032,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  2,   4992,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  3,    400,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  4,    215,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  5,   1920,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  6,   7872,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  7,    219,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  8,    129,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941,  9,    181,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941, 10,     87,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941, 11,   8450,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941, 12,   5888,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941, 13,   7732,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06b5c941, 14,   3888,          YM), /* 02174 - Mad Max - A Prehistoric Tale.sndh */
TIMEDB_ENTRY( 06d77e3e,  0,  20730,       YM+TA), /* 01719 - Jess - The Twin Paradox.sndh */
TIMEDB_ENTRY( 06ea5b50,  0,   8561,          YM), /* 03130 - n-a - Life in a Northern Town.sndh */
TIMEDB_ENTRY( 06edd58b,  0,   1343,          YM), /* 04776 - V!X - Remote Entry #2.sndh */
TIMEDB_ENTRY( 06f25069,  0,  11264,       YM+TA), /* 03557 - Tomchi - No Extra.sndh */
TIMEDB_ENTRY( 06fbc75f,  0,   1937,          YM), /* 02231 - Tao - Ikari (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 06fc4b30,  0,   8703,          YM), /* 01274 - Frazer - TSC 28.sndh */
TIMEDB_ENTRY( 06fcc8e1,  0,   5520,          YM), /* 03539 - David Whittaker - European Champions.sndh */
TIMEDB_ENTRY( 06fcc8e1,  1,    769,          YM), /* 03539 - David Whittaker - European Champions.sndh */
TIMEDB_ENTRY( 06fcc8e1,  2,    491,          YM), /* 03539 - David Whittaker - European Champions.sndh */
TIMEDB_ENTRY( 06fcc8e1,  3,    242,          YM), /* 03539 - David Whittaker - European Champions.sndh */
TIMEDB_ENTRY( 070471b8,  0,   5266,          YM), /* 00108 - David Whittaker - Hunter Killer.sndh */
TIMEDB_ENTRY( 070e4547,  0,    190,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  1,    292,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  2,    261,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  3,    450,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  4,     16,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  5,     14,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 070e4547,  6,    608,          YM), /* 04501 - Paul Dowers - Boxing champ.sndh */
TIMEDB_ENTRY( 07150380,  0,   3328,          YM), /* 01191 - Excellence In Art - Musolini.sndh */
TIMEDB_ENTRY( 075c8520,  0, 127496,       YM+TA), /* 04058 - Roggie - Powerbass.sndh */
TIMEDB_ENTRY( 077fce28,  0,  15753,          YM), /* 02241 - Tao - Cream (orig. Rob Hubbard) - Master Of Magic (Stmygm).sndh */
TIMEDB_ENTRY( 0789a8fa,  0,   5840,          YM), /* 02394 - Nexus 6 - The Surgery.sndh */
TIMEDB_ENTRY( 0791da6b,  0,   6912,          YM), /* 01114 - Keith Tinman - Red Heat.sndh */
TIMEDB_ENTRY( 0791da6b,  1,    193,          YM), /* 01114 - Keith Tinman - Red Heat.sndh */
TIMEDB_ENTRY( 0791da6b,  2,    399,          YM), /* 01114 - Keith Tinman - Red Heat.sndh */
TIMEDB_ENTRY( 079eb096,  0,   4224,          YM), /* 01158 - Tyan - Revolution - Snmpfh.sndh */
TIMEDB_ENTRY( 079f22cc,  0,   2257,          YM), /* 03038 - Ed Bogas - Bebop.sndh */
TIMEDB_ENTRY( 07b5f83c,  0,  55030,       YM+TA), /* 03768 - Warren Cann - Wreckers.sndh */
TIMEDB_ENTRY( 07dc77b6,  0,  20886,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  1,   3083,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  2,   4032,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  3,   4010,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  4,  15552,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  5,   2952,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  6,   4608,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  7,   4608,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  8,   3489,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6,  9,   3840,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 10,   2688,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 11,  12288,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 12,   7168,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 13,   8961,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 14,   8338,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 15,   8256,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 16,  10770,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 17,   7168,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 18,   7168,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 19,   4726,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 20,   4634,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 21,   3840,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07dc77b6, 22,   4472,          YM), /* 02189 - Mad Max - Dragon Flight.sndh */
TIMEDB_ENTRY( 07f97429,  0,   4682,          YM), /* 00878 - Ben Daglish - Ultimate Darts.sndh */
TIMEDB_ENTRY( 07f97429,  1,    330,          YM), /* 00878 - Ben Daglish - Ultimate Darts.sndh */
TIMEDB_ENTRY( 0803a8f3,  0,  10752, YM+TA+TB+TD), /* 02844 - Marcer - POKER.sndh */
TIMEDB_ENTRY( 08054e98,  0,   2303,          YM), /* 03438 - Proteque - Ratu.sndh */
TIMEDB_ENTRY( 0809146d,  0,   3350,          YM), /* 03107 - n-a - Eastenders - May-Osborne.sndh */
TIMEDB_ENTRY( 08104278,  0,   5039,          YM), /* 02376 - Nexus 6 - Happy Muff.sndh */
TIMEDB_ENTRY( 0812636d,  0,  26894,       YM+TA), /* 03880 - Mad Max - Labello.sndh */
TIMEDB_ENTRY( 081efa6f,  0,   3072,       YM+TD), /* 02635 - Miggy - Hyper dog poo ( Unfinished ).sndh */
TIMEDB_ENTRY( 0827bf66,  0,  23802,          YM), /* 01674 - Jess - The Battle of Arpeggios.sndh */
TIMEDB_ENTRY( 0857c9e8,  0,   5710,         STE), /* 03450 - 505 - Paradox - Megademo Intro.sndh */
TIMEDB_ENTRY( 0858e24f,  0,   5772,       YM+TA), /* 00821 - !Cube - Admirable's Mod Compil #31.sndh */
TIMEDB_ENTRY( 0865e3bb,  0,  20736,          YM), /* 01447 - Stax - Revotus - Chip.sndh */
TIMEDB_ENTRY( 08687258,  0,   4473,    YM+TA+TB), /* 01597 - Nemo - Loading - Warm Up.sndh */
TIMEDB_ENTRY( 087c4b4d,  0, 156946,       YM+TA), /* 03821 - Tufty - Roger Rabbit.sndh */
TIMEDB_ENTRY( 088687c4,  0,   7296,    YM+TA+TD), /* 04577 - Crazy Q - DHS - From Da Ghetto With Love.sndh */
TIMEDB_ENTRY( 0886969b,  0,   2685,          YM), /* 01530 - Iso - Panache Theme.sndh */
TIMEDB_ENTRY( 0889f290,  0,  39936,          YM), /* 01437 - Stax - Happy.sndh */
TIMEDB_ENTRY( 08ec970f,  0,  25080,       YM+TA), /* 01216 - Energizer - Floating - Digi.sndh */
TIMEDB_ENTRY( 08f09c66,  0,   2370,          YM), /* 04697 - Unknown - MGD2.sndh */
TIMEDB_ENTRY( 09052c4b,  0,   4229,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  1,   9654,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  2,   9654,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  3,   9362,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  4,  11702,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  5,   8936,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  6,  10240,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  7,   8192,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09052c4b,  8,   3606,          YM), /* 00081 - David Whittaker - Buffalo Bill.sndh */
TIMEDB_ENTRY( 09183d32,  0,   2701,       YM+TA), /* 02236 - Tao - Cream (orig. Rob Hubbard) - Commando (Stmygm).sndh */
TIMEDB_ENTRY( 0942df90,  0,  13568,       YM+TA), /* 00928 - Dark Angel - Kaskades - Digi.sndh */
TIMEDB_ENTRY( 0947c528,  0,   1600,          YM), /* 02146 - Mad Max - Mad Max Tune 1 - LSD.sndh */
TIMEDB_ENTRY( 094adf24,  0,  29328,          YM), /* 01457 - Stax - X Large 2.sndh */
TIMEDB_ENTRY( 094d244a,  0,   4992,          YM), /* 02933 - Chris Denman - Cyber Cop.sndh */
TIMEDB_ENTRY( 096862bb,  0,  35328, YM+TA+TB+TD), /* 02603 - STU - MACHINE.sndh */
TIMEDB_ENTRY( 096ae5c5,  0,   8640,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 096ae5c5,  1,   8704,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 096ae5c5,  2,   4224,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 096ae5c5,  3,   3840,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 096ae5c5,  4,   3061,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 096ae5c5,  5,   4608,          YM), /* 01115 - Keith Tinman - Robocop 2.sndh */
TIMEDB_ENTRY( 09770bbd,  0,   6522,       YM+TA), /* 03301 - 505-checkpoint - GLENZ.sndh */
TIMEDB_ENTRY( 099d239f,  0,    775,          YM), /* 03936 - Unknown - Ungarn.sndh */
TIMEDB_ENTRY( 09b30324,  0,    762,       YM+TB), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  1,   4921,    YM+TB+TD), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  2,    441,    YM+TB+TD), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  3,   8058, YM+TA+TB+TD), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  4,   5754, YM+TA+TB+TD), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  5,   8253, YM+TA+TB+TD), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09b30324,  6,   1926,          YM), /* 03607 - DMA-SC - Kolmik.sndh */
TIMEDB_ENTRY( 09cc4ad7,  0,  10068, YM+TA+TB+TD), /* 00044 - 505 - Clouds.sndh */
TIMEDB_ENTRY( 09d2f42e,  0,   9855,    YM+TB+TD), /* 02694 - Dubmood - Alla man borde ringt.sndh */
TIMEDB_ENTRY( 09e4040c,  0,   4480,          YM), /* 02196 - Mad Max - Grand Monster Slam.sndh */
TIMEDB_ENTRY( 09e4040c,  1,    256,          YM), /* 02196 - Mad Max - Grand Monster Slam.sndh */
TIMEDB_ENTRY( 09e4040c,  2,    303,          YM), /* 02196 - Mad Max - Grand Monster Slam.sndh */
TIMEDB_ENTRY( 09e4040c,  3,    241,          YM), /* 02196 - Mad Max - Grand Monster Slam.sndh */
TIMEDB_ENTRY( 09efef6f,  0,   2004,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09efef6f,  1,    175,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09efef6f,  2,    624,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09efef6f,  3,   1344,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09efef6f,  4,   9600,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09efef6f,  5,   2592,          YM), /* 04610 - Richard Joseph - Moonshine Racers.sndh */
TIMEDB_ENTRY( 09f8ce12,  0,   4600,          YM), /* 03679 - Dma-Sc - Antiques - Hidden.sndh */
TIMEDB_ENTRY( 09fcded8,  0,  19793,       YM+TA), /* 03734 - David Whittaker - Gazza 2.sndh */
TIMEDB_ENTRY( 0a1a5a71,  0,   2707,          YM), /* 01754 - Dave Kelly - Dragon Spirit.sndh */
TIMEDB_ENTRY( 0a1b1734,  0,   7236,          YM), /* 03162 - n-a - Sidestep by Davidow.sndh */
TIMEDB_ENTRY( 0a28707f,  0,  11142,          YM), /* 04098 - Gary Wheaton - Out on the Weekend.sndh */
TIMEDB_ENTRY( 0a2c13b6,  0,  11448, YM+TA+TB+TD), /* 03471 - gwEm (Jester conversion) - Stardust Memories.sndh */
TIMEDB_ENTRY( 0a62127c,  0,  16513,          YM), /* 00615 - Chrispy Noodle - Sapiens 1.sndh */
TIMEDB_ENTRY( 0a66d43f,  0,   1177,          YM), /* 00565 - Jas C Brooke - Flying Shark.sndh */
TIMEDB_ENTRY( 0a66d43f,  1,   4096,          YM), /* 00565 - Jas C Brooke - Flying Shark.sndh */
TIMEDB_ENTRY( 0a66d43f,  2,   1224,          YM), /* 00565 - Jas C Brooke - Flying Shark.sndh */
TIMEDB_ENTRY( 0a66d43f,  3,   4320,          YM), /* 00565 - Jas C Brooke - Flying Shark.sndh */
TIMEDB_ENTRY( 0a66d43f,  4,   1141,          YM), /* 00565 - Jas C Brooke - Flying Shark.sndh */
TIMEDB_ENTRY( 0a79bcc4,  0,   8657,          YM), /* 00225 - ACC - Ala Turk.sndh */
TIMEDB_ENTRY( 0a8138e8,  0,   7162,          YM), /* 00457 - Augustin Vidovic - Punch.sndh */
TIMEDB_ENTRY( 0a8acf9e,  0,   5120,          YM), /* 00805 - Crawdaddy - Party Muz.sndh */
TIMEDB_ENTRY( 0a8de2f4,  0,   3648,          YM), /* 01100 - Techno - Techno 5.sndh */
TIMEDB_ENTRY( 0a9036d3,  0,  17280, YM+TA+TB+TD), /* 02101 - Mad Max - Thrust - Sid.sndh */
TIMEDB_ENTRY( 0aa9116a,  0,   6272, YM+TA+TB+TD), /* 00188 - 505 - Posh - Part 5.sndh */
TIMEDB_ENTRY( 0ac1c32b,  0,   9654,          YM), /* 00155 - David Whittaker - Crossbow - William Tell.sndh */
TIMEDB_ENTRY( 0ac1c32b,  1,   1365,          YM), /* 00155 - David Whittaker - Crossbow - William Tell.sndh */
TIMEDB_ENTRY( 0ac1c32b,  2,     88,          YM), /* 00155 - David Whittaker - Crossbow - William Tell.sndh */
TIMEDB_ENTRY( 0ad1a60e,  0,     64,          YM), /* 04515 - J.P. Peach - Color Quest.sndh */
TIMEDB_ENTRY( 0ad73d30,  0,   8000,          YM), /* 03244 - T1000 - The Beast Club - TBC 16.sndh */
TIMEDB_ENTRY( 0ad9b6c2,  0,   2024,          YM), /* 04431 - Lotus - Synth Dream 2 - Max intro.sndh */
TIMEDB_ENTRY( 0af49608,  0,   1926,          YM), /* 03920 - Unknown - Klassik1.sndh */
TIMEDB_ENTRY( 0af93806,  0,  17408,          YM), /* 02139 - Mad Max - One Man And His Droid - LCD.sndh */
TIMEDB_ENTRY( 0b06d0dd,  0,  43995, YM+TA+TB+TD), /* 02442 - cerror - Kicking it oldskool.sndh */
TIMEDB_ENTRY( 0b19f70f,  0,   2240,          YM), /* 03189 - n-a - Xeron Overture.sndh */
TIMEDB_ENTRY( 0b1d5135,  0,  59618,       YM+TA), /* 03751 - Tufty - Killer - Adamski.sndh */
TIMEDB_ENTRY( 0b1e485c,  0,   6914,          YM), /* 02117 - Mad Max - Starball.sndh */
TIMEDB_ENTRY( 0b1e485c,  1,    770,          YM), /* 02117 - Mad Max - Starball.sndh */
TIMEDB_ENTRY( 0b2b30f2,  0,   8953,          YM), /* 03565 - DMA SC - Coast II Coast.sndh */
TIMEDB_ENTRY( 0b33f385,  0,  10301, YM+TA+TB+TD), /* 02521 - ZABUTOM UNT DUBMOOD - DRUMNBICE.sndh */
TIMEDB_ENTRY( 0b4b79f3,  0,  15231, YM+TA+TB+TD), /* 02707 - DUBMOOD <DEAD KENNEDYS> - TOO DRUNK TO FUCK!.sndh */
TIMEDB_ENTRY( 0b587f22,  0,   3232,          YM), /* 03185 - Unknown - The Windmills of Your Mind.sndh */
TIMEDB_ENTRY( 0b5cd2b1,  0,   2309,          YM), /* 02291 - Tao - Cream (orig. Karsten Obarski) - End Theme.sndh */
TIMEDB_ENTRY( 0b5d1a03,  0,   7040,          YM), /* 01501 - Dave Hasler - Drivin Force.sndh */
TIMEDB_ENTRY( 0b5d1a03,  1,    180,          YM), /* 01501 - Dave Hasler - Drivin Force.sndh */
TIMEDB_ENTRY( 0b5fdd60,  0,   5661,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  1,   4804,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  2,   9888,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  3,   9024,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  4,   4608,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  5,   5396,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b5fdd60,  6,   6912,          YM), /* 03501 - Christian Wirsig - Blaxx.sndh */
TIMEDB_ENTRY( 0b6287d6,  0,   7616,          YM), /* 02010 - Mad Max - LM.sndh */
TIMEDB_ENTRY( 0b6287d6,  1,    206,          YM), /* 02010 - Mad Max - LM.sndh */
TIMEDB_ENTRY( 0b6d77af,  0,   6590,          YM), /* 00512 - Tangens - Synthetic Visions - Parallax.sndh */
TIMEDB_ENTRY( 0b753dff,  0,  16896,    YM+TB+TD), /* 02911 - helene - temptrance.sndh */
TIMEDB_ENTRY( 0b7c380e,  0,   6140,          YM), /* 00061 - 505 - I Want Sid.sndh */
TIMEDB_ENTRY( 0b8092e4,  0,   1921,          YM), /* 00392 - Wally Beben - Roadwars.sndh */
TIMEDB_ENTRY( 0b8cc2cd,  0,   3839,          YM), /* 03685 - Dma-Sc - Saboteur 3 - Ending.sndh */
TIMEDB_ENTRY( 0b934018,  0, 268756,       YM+TA), /* 04795 - Unknown - Acid.sndh */
TIMEDB_ENTRY( 0ba741ba,  0,  38785,       YM+TA), /* 03760 - Tufty - Robo 16.sndh */
TIMEDB_ENTRY( 0bb8ecc4,  0,  11904,    YM+TA+TD), /* 02287 - Tao - Cream - Seagulls (Steps).sndh */
TIMEDB_ENTRY( 0bbcdf93,  0,   8791,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bbcdf93,  1,    387,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bbcdf93,  2,   2108,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bbcdf93,  3,   1888,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bbcdf93,  4,    458,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bbcdf93,  5,    291,          YM), /* 03976 - Christopher Barker - The Third Courier.sndh */
TIMEDB_ENTRY( 0bc313eb,  0,   5376,          YM), /* 00494 - Ultimatum - Kidney Bean 10.sndh */
TIMEDB_ENTRY( 0bd1e4bb,  0,  12096,          YM), /* 03235 - T1000 - The Beast Club - TBC 5.sndh */
TIMEDB_ENTRY( 0bddfee5,  0,   4267,       YM+TA), /* 03481 - Peter Jorgensen - Russer.sndh */
TIMEDB_ENTRY( 0be6ee8f,  0,   7179,          YM), /* 00141 - David Whittaker - Starray.sndh */
TIMEDB_ENTRY( 0be6ee8f,  1,    124,          YM), /* 00141 - David Whittaker - Starray.sndh */
TIMEDB_ENTRY( 0be6ee8f,  2,   5671,          YM), /* 00141 - David Whittaker - Starray.sndh */
TIMEDB_ENTRY( 0be6ee8f,  3,   4468,          YM), /* 00141 - David Whittaker - Starray.sndh */
TIMEDB_ENTRY( 0bf6b1df,  0,   4160,          YM), /* 01516 - Illusion - Legal Tools 5.sndh */
TIMEDB_ENTRY( 0bf736e9,  0,   1506,          YM), /* 03918 - Unknown - Finnland.sndh */
TIMEDB_ENTRY( 0c0ab5c8,  0,   4146,          YM), /* 04067 - Gary Wheaton - Dear Boy.sndh */
TIMEDB_ENTRY( 0c2005e9,  0,   9232,          YM), /* 03466 - Marcer - Stax.sndh */
TIMEDB_ENTRY( 0c2905ad,  0,   2055,          YM), /* 04651 - Unknown - Japs the Sequence.sndh */
TIMEDB_ENTRY( 0c2b5778,  0,   8448,          YM), /* 04847 - t1000 - Bits 56.sndh */
TIMEDB_ENTRY( 0c31c5a5,  0,   7296,          YM), /* 00236 - AJT - Colorz.sndh */
TIMEDB_ENTRY( 0c3fde19,  0,    448,          YM), /* 01084 - Techno - Dentro.sndh */
TIMEDB_ENTRY( 0c4006c4,  0,   4605,          YM), /* 01226 - Epic - Doc Dots.sndh */
TIMEDB_ENTRY( 0c59f3e8,  0,  18152,          YM), /* 02457 - Starbuck^Creative Mind - Hurry-Scurry.sndh */
TIMEDB_ENTRY( 0c6133b1,  0,   8880,          YM), /* 02957 - Rudi Stember - The Patrician.sndh */
TIMEDB_ENTRY( 0c6133b1,  1,   4800,          YM), /* 02957 - Rudi Stember - The Patrician.sndh */
TIMEDB_ENTRY( 0c6133b1,  2,   3856,          YM), /* 02957 - Rudi Stember - The Patrician.sndh */
TIMEDB_ENTRY( 0c63ff61,  0,   5088,          YM), /* 01181 - Excellence In Art - Gritty.sndh */
TIMEDB_ENTRY( 0c734b9b,  0,   9711,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  1,   3868,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  2,  23052,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  3,  15582,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  4,  12683,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  5,  13210,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  6,   8317,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  7,   1262,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c734b9b,  8,  10447,          YM), /* 03011 - Paul Summers - S.D.I. (Activision).sndh */
TIMEDB_ENTRY( 0c8ae262,  0,   9983,          YM), /* 04572 - Tomchi  Noextra - ltc 01 extra volume 2.sndh */
TIMEDB_ENTRY( 0c9094a7,  0,   8064,          YM), /* 02369 - Nexus 6 - Cried Out Eyes.sndh */
TIMEDB_ENTRY( 0c90f9d3,  0,   4992, YM+TA+TB+TD), /* 01827 - Lotek Style - 28 Style.sndh */
TIMEDB_ENTRY( 0c951326,  0,  62725,       YM+TA), /* 03951 - Scott Lemmon - Geosong.sndh */
TIMEDB_ENTRY( 0ca0ebd7,  0,  11441,          YM), /* 01389 - Andy Severn - Twin Turbos.sndh */
TIMEDB_ENTRY( 0ca0ebd7,  1,    109,          YM), /* 01389 - Andy Severn - Twin Turbos.sndh */
TIMEDB_ENTRY( 0ca0ebd7,  2,    865,          YM), /* 01389 - Andy Severn - Twin Turbos.sndh */
TIMEDB_ENTRY( 0caa6961,  0,   1024,       YM+TD), /* 02763 - Bobby Prince [ST by Dma-Sc] - Ending.sndh */
TIMEDB_ENTRY( 0cbf3abc,  0,   2144,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  1,   1066,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  2,    842,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  3,    550,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  4,   1969,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  5,    880,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  6,    609,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  7,   1066,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  8,   1505,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc,  9,   1138,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 10,   1066,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 11,    345,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 12,     64,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 13,   1066,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 14,   1600,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 15,   4474,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 16,   5490,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 17,   2260,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 18,   4224,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cbf3abc, 19,   1706,          YM), /* 04190 - Jim Cuomo - Astate.sndh */
TIMEDB_ENTRY( 0cc1715e,  0,   8442,       YM+TA), /* 01370 - Scavenger - Working Man.sndh */
TIMEDB_ENTRY( 0cc657cd,  0,  32768, YM+TA+TB+TD), /* 02553 - Marcer - BONITA - MARCER.sndh */
TIMEDB_ENTRY( 0cc9cee8,  0,   8256,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  1,    320,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  2,    367,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  3,   2240,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  4,   2240,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  5,   8064,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0cc9cee8,  6,   4480,          YM), /* 04186 - Dave Hasler - Elvira - Mistress of the Dark.sndh */
TIMEDB_ENTRY( 0ccef3de,  0,   2048, YM+TA+TB+TD), /* 01025 - Doclands - Doclands Sid.sndh */
TIMEDB_ENTRY( 0cd1d6a2,  0, 124133,       YM+TA), /* 04678 - Unknown - CH 11.sndh */
TIMEDB_ENTRY( 0ce6d43f,  0,   5907,          YM), /* 03083 - Tom Walden - The Battle of New Orleans.sndh */
TIMEDB_ENTRY( 0cf0f911,  0,  52821,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  1,  84510,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  2,  65692,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  3,  78390,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  4,  63146,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  5,  56812,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  6,  94554,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  7,  96567,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  8,  36811,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911,  9,  68523,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf0f911, 10, 128134,       YM+TA), /* 03774 - Exarch - Son Et Lumiere.sndh */
TIMEDB_ENTRY( 0cf4ed4b,  0,   9984,       YM+TA), /* 04435 - Lotus - Synth Dream 2 - Running Status.sndh */
TIMEDB_ENTRY( 0cfbd1c3,  0,   8587, YM+TA+TB+TD), /* 03391 - Ultrasyd - Cats love ST.sndh */
TIMEDB_ENTRY( 0d07145f,  0,   7193,          YM), /* 01579 - Donovan Prince - Hellbent.sndh */
TIMEDB_ENTRY( 0d13205e,  0,   7295, YM+TA+TB+TD), /* 02807 - Starbuck^Creative Mind - One Step foreward (SNDH Version).sndh */
TIMEDB_ENTRY( 0d1f0879,  0,   8192,          YM), /* 01168 - Tyan - Scoopex.sndh */
TIMEDB_ENTRY( 0d2c38b9,  0,     86,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2c38b9,  1,   1861,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2c38b9,  2,    209,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2c38b9,  3,     66,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2c38b9,  4,     57,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2c38b9,  5,    365,          YM), /* 04490 - Mikael Lundberg - Amazeing.sndh */
TIMEDB_ENTRY( 0d2f26a5,  0,   4160,          YM), /* 00573 - Jas C Brooke - Rockford.sndh */
TIMEDB_ENTRY( 0d2f26a5,  1,    321,          YM), /* 00573 - Jas C Brooke - Rockford.sndh */
TIMEDB_ENTRY( 0d2f26a5,  2,     13,          YM), /* 00573 - Jas C Brooke - Rockford.sndh */
TIMEDB_ENTRY( 0d2f26a5,  3,     51,          YM), /* 00573 - Jas C Brooke - Rockford.sndh */
TIMEDB_ENTRY( 0d3f87e8,  0, 200022,       YM+TA), /* 03731 - Martin Walker - Shadow Sorcerer.sndh */
TIMEDB_ENTRY( 0d4dbf05,  0,  11271,    YM+TA+TD), /* 02272 - Tao - Cream - The Delegate (Steps).sndh */
TIMEDB_ENTRY( 0d58eb4b,  0,   5003,       YM+TA), /* 04416 - EVE - Noisy Pillars.sndh */
TIMEDB_ENTRY( 0d5cce44,  0,   2506,          YM), /* 03044 - Ed Bogas - Country Too.sndh */
TIMEDB_ENTRY( 0d61dd75,  0,   5039, YM+TA+TB+TD), /* 02980 - 505-paradox - coffee break.sndh */
TIMEDB_ENTRY( 0d63d768,  0,   6720,          YM), /* 01257 - Tim & Geoff Follin - Gauntlet 3.sndh */
TIMEDB_ENTRY( 0d63d768,  1,    354,          YM), /* 01257 - Tim & Geoff Follin - Gauntlet 3.sndh */
TIMEDB_ENTRY( 0d63d768,  2,    343,          YM), /* 01257 - Tim & Geoff Follin - Gauntlet 3.sndh */
TIMEDB_ENTRY( 0d63d768,  3,   3073,          YM), /* 01257 - Tim & Geoff Follin - Gauntlet 3.sndh */
TIMEDB_ENTRY( 0d684b94,  0,   6754,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  1,   7125,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  2,    342,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  3,    365,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  4,    497,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  5,    371,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  6,    160,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d684b94,  7,    224,          YM), /* 00416 - Big Alec - No Second Prize.sndh */
TIMEDB_ENTRY( 0d71ea78,  0,   6151,          YM), /* 00959 - Jeremy Das - Radio Controlled Racer.sndh */
TIMEDB_ENTRY( 0d762eb1,  0,   3013,          YM), /* 00727 - Tebirod - No Cooper Rasters.sndh */
TIMEDB_ENTRY( 0d762eb1,  1,   3013,          YM), /* 00727 - Tebirod - No Cooper Rasters.sndh */
TIMEDB_ENTRY( 0d762eb1,  2,   4997,          YM), /* 00727 - Tebirod - No Cooper Rasters.sndh */
TIMEDB_ENTRY( 0d762eb1,  3,   3013,          YM), /* 00727 - Tebirod - No Cooper Rasters.sndh */
TIMEDB_ENTRY( 0d762eb1,  4,   3587,          YM), /* 00727 - Tebirod - No Cooper Rasters.sndh */
TIMEDB_ENTRY( 0d7c187d,  0,   3072,          YM), /* 01883 - Dave Lowe - F-15 Strike Eagle 2.sndh */
TIMEDB_ENTRY( 0d7c187d,  1,   2240,          YM), /* 01883 - Dave Lowe - F-15 Strike Eagle 2.sndh */
TIMEDB_ENTRY( 0d7c187d,  2,   8192,          YM), /* 01883 - Dave Lowe - F-15 Strike Eagle 2.sndh */
TIMEDB_ENTRY( 0d7c187d,  3,   1152,          YM), /* 01883 - Dave Lowe - F-15 Strike Eagle 2.sndh */
TIMEDB_ENTRY( 0d7c187d,  4,   1920,          YM), /* 01883 - Dave Lowe - F-15 Strike Eagle 2.sndh */
TIMEDB_ENTRY( 0dc79d68,  0,   4845,          YM), /* 03183 - n-a - WhatChildIsThis.sndh */
TIMEDB_ENTRY( 0dc7e44f,  0,   1595,       YM+TA), /* 02638 - Miggy - Meow (Unfinished).sndh */
TIMEDB_ENTRY( 0dd0f1e9,  0,   2925, YM+TA+TB+TD), /* 04308 - 505-checkpoint - ST_M.sndh */
TIMEDB_ENTRY( 0ded0765,  0,   4218, YM+TA+TB+TD), /* 04255 - 505-checkpoint - eighties.sndh */
TIMEDB_ENTRY( 0df05e40,  0,   8448, YM+TA+TB+TD), /* 02067 - Mad Max - Bump Set Spike - Sid.sndh */
TIMEDB_ENTRY( 0df05e40,  1,   9504, YM+TA+TB+TD), /* 02067 - Mad Max - Bump Set Spike - Sid.sndh */
TIMEDB_ENTRY( 0df45c86,  0,    575,          YM), /* 04525 - Philip Rankin - Easy Learning Maths.sndh */
TIMEDB_ENTRY( 0df45c86,  1,   2944,          YM), /* 04525 - Philip Rankin - Easy Learning Maths.sndh */
TIMEDB_ENTRY( 0df45c86,  2,    280,          YM), /* 04525 - Philip Rankin - Easy Learning Maths.sndh */
TIMEDB_ENTRY( 0e01f55f,  0,  17792,          YM), /* 02187 - Mad Max - Crimetime.sndh */
TIMEDB_ENTRY( 0e106f8d,  0,   4473,       YM+TA), /* 04234 - 505 - 12th.sndh */
TIMEDB_ENTRY( 0e128fd6,  0,  23264, YM+TA+TB+TD), /* 04241 - 505-cpt^pdx - Arcarde.sndh */
TIMEDB_ENTRY( 0e15e766,  0,  10800, YM+TA+TB+TD), /* 02066 - Mad Max - Battle of Britain - Sid.sndh */
TIMEDB_ENTRY( 0e16a0fa,  0,  18442,          YM), /* 00463 - Augustin Vidovic - So Watt - TCB (SOS).sndh */
TIMEDB_ENTRY( 0e1a0d56,  0, 179792,       YM+TA), /* 04794 - Unknown - 808 State.sndh */
TIMEDB_ENTRY( 0e2b2de0,  0,   9663, YM+TA+TB+TD), /* 03159 - Frequent-Ephidrena - Steinras DMA.sndh */
TIMEDB_ENTRY( 0e38278e,  0,  11136,          YM), /* 04011 - Danny Oneway - The Year After - Risk #2.sndh */
TIMEDB_ENTRY( 0e45e998,  0,   5183,          YM), /* 04095 - Gary Wheaton - Strawberry Fields.sndh */
TIMEDB_ENTRY( 0e636894,  0,   7681,          YM), /* 00363 - Steve Barrett - Italy 1990.sndh */
TIMEDB_ENTRY( 0e68ab31,  0,   1536,          YM), /* 04844 - t1000 - Bits 40.sndh */
TIMEDB_ENTRY( 0e9487aa,  0,   8112,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  1,   1561,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  2,   4032,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  3,   3852,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  4,   6240,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  5,   5664,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0e9487aa,  6,    241,          YM), /* 01259 - Tim Follin - Led Storm.sndh */
TIMEDB_ENTRY( 0ea0d826,  0,   5088,          YM), /* 02414 - Nexus 6 - The Vulcan.sndh */
TIMEDB_ENTRY( 0ea105d8,  0,  32060,          YM), /* 00001 - gwEm - Full On! (re-version).sndh */
TIMEDB_ENTRY( 0ea4f751,  0,   7290,          YM), /* 02888 - Epic - World of Wonder - Legend.sndh */
TIMEDB_ENTRY( 0eb0298d,  0,   6969,       YM+TA), /* 01899 - Luebke - Compo.sndh */
TIMEDB_ENTRY( 0ec74db4,  0,   6166,          YM), /* 00409 - Big Alec - Killer.sndh */
TIMEDB_ENTRY( 0eeb8343,  0,  15360,       YM+TA), /* 04114 - Tao - 150 MPH - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 0eec7c72,  0,  10355,          YM), /* 00839 - Ben Daglish - Artura.sndh */
TIMEDB_ENTRY( 0eec7c72,  1,   7160,          YM), /* 00839 - Ben Daglish - Artura.sndh */
TIMEDB_ENTRY( 0ef563a2,  0,  11641, YM+TA+TB+TD), /* 03319 - Msg - Meikever.sndh */
TIMEDB_ENTRY( 0efeb974,  0,  15676, YM+TA+TB+TD), /* 00205 - 505 - UCM 12 1.sndh */
TIMEDB_ENTRY( 0f191d4d,  0,   3583, YM+TA+TB+TD), /* 02838 - maktone-fairlight - nelson.sndh */
TIMEDB_ENTRY( 0f259465,  0,   1920,          YM), /* 04605 - White Ninja - Love Desaster.sndh */
TIMEDB_ENTRY( 0f2e0c09,  0,  11945,          YM), /* 04379 - 505 - RoboST (STE).sndh */
TIMEDB_ENTRY( 0f540c0c,  0,   4992,          YM), /* 00497 - Ultimatum - Kidney Bean - Reset.sndh */
TIMEDB_ENTRY( 0f546b2b,  0,   5361,          YM), /* 02506 - Crazy Q - Crushed Insect (RG mix).sndh */
TIMEDB_ENTRY( 0f55518f,  0,   2690,       YM+TA), /* 00930 - Dark Angel - Logic Force.sndh */
TIMEDB_ENTRY( 0f6c0b66,  0,   6528,       YM+TD), /* 01385 - Scy - TSCC Mod Compil 13.sndh */
TIMEDB_ENTRY( 0f797ef0,  0,   7680,       YM+TD), /* 00586 - Celtic - Lethal Suck.sndh */
TIMEDB_ENTRY( 0f7ac162,  0,     90,          YM), /* 04710 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - screen 5.sndh */
TIMEDB_ENTRY( 0f7cd0f4,  0,  28561,       YM+TA), /* 03850 - Mr Saigon - MSI Sound Demo.sndh */
TIMEDB_ENTRY( 0f7cd0f4,  1,  27272,       YM+TA), /* 03850 - Mr Saigon - MSI Sound Demo.sndh */
TIMEDB_ENTRY( 0f7cd0f4,  2,  43265,       YM+TA), /* 03850 - Mr Saigon - MSI Sound Demo.sndh */
TIMEDB_ENTRY( 0f7cd0f4,  3,  10001,       YM+TA), /* 03850 - Mr Saigon - MSI Sound Demo.sndh */
TIMEDB_ENTRY( 0f8ea27d,  0,  12042,       YM+TA), /* 02673 - Donald Fakk - Jag Springer Fort.sndh */
TIMEDB_ENTRY( 0f9490b8,  0,   6724,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  1,   5124,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  2,    964,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  3,   1924,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  4,    644,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  5,   5764,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9490b8,  6,   5764,          YM), /* 01374 - Justin Scharvona - Armalyte.sndh */
TIMEDB_ENTRY( 0f9803b3,  0,  13760, YM+TA+TB+TD), /* 00826 - Crazy Q - TiMeBoMb - Evil Chip Of Dr.Zool.sndh */
TIMEDB_ENTRY( 0fdf6f5e,  0,   1536,       YM+TA), /* 02885 - Deneb - The Puppy Show - Loader.sndh */
TIMEDB_ENTRY( 1001b206,  0,   7488,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  1,   1717,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  2,   1703,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  3,   1550,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  4,   1594,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  5,    282,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  6,    275,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 1001b206,  7,    102,          YM), /* 00137 - David Whittaker - Speedball.sndh */
TIMEDB_ENTRY( 10101bfe,  0,   8273, YM+TA+TB+TD), /* 02984 - 505-paradox - everything.sndh */
TIMEDB_ENTRY( 10154b5a,  0,  14777, YM+TA+TB+TD), /* 00065 - 505-Checkpoint - UCM #20 music.sndh */
TIMEDB_ENTRY( 1020c8c4,  0,   6532,          YM), /* 01773 - Lap - Lap 13.sndh */
TIMEDB_ENTRY( 10250c70,  0,   8727,          YM), /* 04075 - Gary Wheaton - Fell in the horse shit in his new suit.sndh */
TIMEDB_ENTRY( 1042340d,  0,  10368,          YM), /* 01991 - Mad Max - Last Ninja.sndh */
TIMEDB_ENTRY( 10443af2,  0,   8384,          YM), /* 02052 - Mad Max - Harvey Smiths Show Jumping.sndh */
TIMEDB_ENTRY( 10477923,  0,  10861,          YM), /* 00170 - Martin Walker - Indiana Jones and the Fate of Atlantis.sndh */
TIMEDB_ENTRY( 106bb9c8,  0,  49675,       YM+TA), /* 04667 - Zbyhoo & Kijop - Bob Dylan.sndh */
TIMEDB_ENTRY( 1075f97c,  0,  15354,          YM), /* 01693 - Jess - Jochen Spirit.sndh */
TIMEDB_ENTRY( 107f3197,  0,   1680,       YM+TA), /* 02886 - Deneb - The Puppy Show - Menu.sndh */
TIMEDB_ENTRY( 1099deae,  0,  10171,          YM), /* 03145 - n-a - Orange Blossom Special.sndh */
TIMEDB_ENTRY( 10a1171f,  0,   4600,       YM+TA), /* 04227 - 505 - 6th.sndh */
TIMEDB_ENTRY( 10a18d1f,  0,   9412,          YM), /* 02234 - Tao - Main menu (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 10a51096,  0,  33792, YM+TA+TB+TD), /* 02540 - SCHMX & STU - HIGHWAY UP NORTH.sndh */
TIMEDB_ENTRY( 10ab8c09,  0,   3328,          YM), /* 01222 - ENS Music - ENS 5.sndh */
TIMEDB_ENTRY( 10c356e4,  0,  13483,       YM+TA), /* 01328 - Scavenger - Cuddly This Year.sndh */
TIMEDB_ENTRY( 10cbaa59,  0,   3200,          YM), /* 01428 - Starbuck - Living on Video remix.sndh */
TIMEDB_ENTRY( 10d28bc3,  0,   6528,          YM), /* 01767 - Lap - Lap 7 - Ghosts and Goblins.sndh */
TIMEDB_ENTRY( 10eb444e,  0,   9409, YM+TA+TB+TD), /* 02069 - Mad Max - Clumsy Colin Action Biker - Sid.sndh */
TIMEDB_ENTRY( 10eb444e,  1,   3073, YM+TA+TB+TD), /* 02069 - Mad Max - Clumsy Colin Action Biker - Sid.sndh */
TIMEDB_ENTRY( 10eb444e,  2,    541, YM+TA+TB+TD), /* 02069 - Mad Max - Clumsy Colin Action Biker - Sid.sndh */
TIMEDB_ENTRY( 10ff790b,  0,   8955, YM+TA+TB+TD), /* 01586 - Nemo - Alive Intro.sndh */
TIMEDB_ENTRY( 112e3abc,  0,   4367,    YM+TB+TD), /* 04321 - 505 - Checkpoint - Terror (Unrel.).sndh */
TIMEDB_ENTRY( 114ba133,  0,   7207,       YM+TA), /* 03844 - Dave Lowe - Starglider - Title.sndh */
TIMEDB_ENTRY( 114ef6a5,  0,  16751, YM+TA+TB+TD), /* 02986 - 505-paradox - faster.sndh */
TIMEDB_ENTRY( 115e0a13,  0,   1920,          YM), /* 00940 - Dark Angel - Oh Je Mi Neeeh!.sndh */
TIMEDB_ENTRY( 1165187b,  0,   8898,          YM), /* 00388 - Wally Beben - Lancaster.sndh */
TIMEDB_ENTRY( 116ae9d7,  0,   9216,          YM), /* 02451 - TPM - Journey to the Centre of the Earth.sndh */
TIMEDB_ENTRY( 116e17b0,  0,   2560,          YM), /* 00608 - Chris Mad - So Watt - Ghost Reset.sndh */
TIMEDB_ENTRY( 117282eb,  0,   7680,          YM), /* 00240 - Alx 33 - Wow.sndh */
TIMEDB_ENTRY( 11786b23,  0,  23808,    YM+TB+TD), /* 04751 - Xrwwr - SOS (2015 Remix).sndh */
TIMEDB_ENTRY( 11906bd8,  0,  11641,       YM+TA), /* 01936 - Miq - Rolling.sndh */
TIMEDB_ENTRY( 11ae621c,  0,   1339,          YM), /* 03488 - Kevin Toms - Football Manager.sndh */
TIMEDB_ENTRY( 11dc782e,  0,   5193,       YM+TA), /* 01610 - No More - Blast.sndh */
TIMEDB_ENTRY( 11e19312,  0,   5005,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  1,    385,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  2,    768,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  3,   3584,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  4,   1808,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  5,    768,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  6,   2688,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  7,   3462,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  8,   1536,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312,  9,    672,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 10,   2640,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 11,    576,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 12,   1278,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 13,   1632,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 14,   3072,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 15,    324,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e19312, 16,   2064,          YM), /* 01119 - Paul Tonge - Civilisation.sndh */
TIMEDB_ENTRY( 11e76d9f,  0,   3840,       YM+TA), /* 04427 - Lotus - Synth Dream 2 - Eternal Intro Tune.sndh */
TIMEDB_ENTRY( 1202b75e,  0,   4480,          YM), /* 01209 - Excellence In Art - Top Gun.sndh */
TIMEDB_ENTRY( 1206cc06,  0,    511,          YM), /* 03869 - Gary Wheaton - Cyberstorm Loader.sndh */
TIMEDB_ENTRY( 12121edb,  0,   2400,          YM), /* 04481 - Mad Max - Return of Medusa (unused).sndh */
TIMEDB_ENTRY( 1215648d,  0,   3840,          YM), /* 00786 - Crawdaddy - Detonation.sndh */
TIMEDB_ENTRY( 1219e345,  0,   5712,       YM+TA), /* 04439 - Lotus - Synth Dream 2 - Wishful Thinking.sndh */
TIMEDB_ENTRY( 122d5a9a,  0,   4608,    YM+TA+TD), /* 02570 - !Cube-Aggression - Earth.sndh */
TIMEDB_ENTRY( 122fe23f,  0,  17019,       YM+TA), /* 03749 - Tufty - Kaos 1 Hyp.sndh */
TIMEDB_ENTRY( 123b4b2f,  0,  26592,          YM), /* 03477 - gwEm (Manou conversion) - RetroModern.sndh */
TIMEDB_ENTRY( 1281c6f4,  0,   6138, YM+TA+TB+TD), /* 00831 - Crazy Q - Timebomb ( 2003 ) - Phat & Phunky Part 1 (YM Rockerz).sndh */
TIMEDB_ENTRY( 129247c8,  0,   8832,          YM), /* 04848 - t1000 - Bits 42.sndh */
TIMEDB_ENTRY( 12c1b460,  0,   3645, YM+TA+TB+TD), /* 00073 - 505 - Lame Lups.sndh */
TIMEDB_ENTRY( 12ebad4e,  0,     53,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  1,    167,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  2,     25,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  3,    153,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  4,    133,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  6,     32,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  7,    122,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  8,    122,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e,  9,     60,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e, 10,     16,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e, 11,    711,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e, 12,    213,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 12ebad4e, 13,      9,          YM), /* 04694 - Ewald Greiml - Marble Islands.sndh */
TIMEDB_ENTRY( 13030213,  0,   2681,    YM+TB+TD), /* 02427 - 505-cpt^dox - sweet.sndh */
TIMEDB_ENTRY( 132348e5,  0,   6144,          YM), /* 00474 - Ultimatum - Choux C.sndh */
TIMEDB_ENTRY( 134d9229,  0,   5376,          YM), /* 03629 - Jari Nikula - Musixx.sndh */
TIMEDB_ENTRY( 13521ee8,  0,    450,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  1,    525,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  2,    626,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  3,    499,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  4,    652,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  5,    500,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  6,    615,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  7,    650,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  8,    710,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8,  9,    533,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8, 10,    823,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8, 11,    520,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8, 12,    795,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 13521ee8, 13,    476,          YM), /* 04538 - Darren Ithell - Fun School 2 - Over 8 yrs.sndh */
TIMEDB_ENTRY( 135f9504,  0,  42241,       YM+TA), /* 03727 - Spector - Manic Demo Intro.sndh */
TIMEDB_ENTRY( 1369cc10,  0,   5737,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  1,   6481,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  2,   2672,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  3,     91,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  4,    241,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  5,     97,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  6,    123,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  7,    293,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  8,   2471,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 1369cc10,  9,    169,          YM), /* 04152 - Unknown - Kampf.sndh */
TIMEDB_ENTRY( 136d361d,  0,   6911,          YM), /* 00505 - Tangens - Mindbomb - Foxx.sndh */
TIMEDB_ENTRY( 1374aa11,  0,     71,          YM), /* 04688 - V. Peschardt - Le jeu des sons no 1.sndh */
TIMEDB_ENTRY( 138e7dbf,  0,   6000,    YM+TA+TD), /* 00023 - Wahoo - Yahoo.sndh */
TIMEDB_ENTRY( 13cebc3b,  0,  26112, YM+TA+TB+TD), /* 02614 - STU - SAD DAY.sndh */
TIMEDB_ENTRY( 13d19fbf,  0,   1275,    YM+TA+TB), /* 04275 - 505-cpt^pdx - Lamer.sndh */
TIMEDB_ENTRY( 13e8f608,  0,   1510,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  1,   2110,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  2,   1726,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  3,   2110,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  4,   1697,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  5,   2211,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  6,   2403,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  7,   2014,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  8,   2014,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13e8f608,  9,    279,          YM), /* 01303 - Matt Furniss - Mad Professor Mariarti.sndh */
TIMEDB_ENTRY( 13edb6b6,  0,   3232,          YM), /* 01766 - Lap - Lap 6 - Jarre.sndh */
TIMEDB_ENTRY( 13f77939,  0,   6912,          YM), /* 00239 - Alx 33 - Sam.sndh */
TIMEDB_ENTRY( 14019d82,  0,    369,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  1,   4608,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  2,   7848,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  3,   8064,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  4,   9984,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  5,    384,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  6,   5120,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  7,   8192,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  8,  11552,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82,  9,   9989,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 14019d82, 10,   6735,          YM), /* 02218 - Mad Max - Leavin Teramis.sndh */
TIMEDB_ENTRY( 1401bf68,  0,  39936,    YM+TB+TD), /* 04853 - xFalcon - MOVE TO TRANCE.sndh */
TIMEDB_ENTRY( 141fa33b,  0,  10032,          YM), /* 01268 - Frazer - Jigsaw.sndh */
TIMEDB_ENTRY( 1440ad99,  0,    316,          YM), /* 02491 - gwEm - Oxygene Loader.sndh */
TIMEDB_ENTRY( 1445980b,  0,   4992,          YM), /* 00942 - Dark Angel - Pan Tau Kotzt!.sndh */
TIMEDB_ENTRY( 145cbd3b,  0,  10365, YM+TA+TB+TD), /* 01017 - DMA-SC - Dosk Opus - UCM 18.sndh */
TIMEDB_ENTRY( 14651f2d,  0,    791,          YM), /* 04508 - Unknown - Bullys Sporting Darts.sndh */
TIMEDB_ENTRY( 14651f2d,  1,    864,          YM), /* 04508 - Unknown - Bullys Sporting Darts.sndh */
TIMEDB_ENTRY( 14651f2d,  2,   1246,          YM), /* 04508 - Unknown - Bullys Sporting Darts.sndh */
TIMEDB_ENTRY( 1472c9e2,  0,  14977,          YM), /* 03622 - TOMCHI  NOEXTRA - No Monkey.sndh */
TIMEDB_ENTRY( 14752934,  0,   7296,          YM), /* 00238 - Alx 33 - Bulbing.sndh */
TIMEDB_ENTRY( 1478093a,  0,   3975,       YM+TA), /* 00052 - 505 - Glas.sndh */
TIMEDB_ENTRY( 1495071d,  0,   9616,          YM), /* 01023 - Doclands - Blastiny Bobs.sndh */
TIMEDB_ENTRY( 149a5ad1,  0,   5760,    YM+TA+TD), /* 02255 - Tao - Cream - Flow (Steps).sndh */
TIMEDB_ENTRY( 149fff39,  0,   9849,          YM), /* 01272 - Frazer - TSC 21.sndh */
TIMEDB_ENTRY( 14c5d927,  0, 120448,       YM+TA), /* 04037 - Roggie - Roggie 19.sndh */
TIMEDB_ENTRY( 14cfe380,  0,   3583, YM+TA+TB+TD), /* 00284 - Baggio - Evolution - Ode to inspiration.sndh */
TIMEDB_ENTRY( 14d0fae4,  0,  18090,      YM+STE), /* 04208 - Jess - Throw It in a Trash Can (STe).sndh */
TIMEDB_ENTRY( 14e8c428,  0,  10239,          YM), /* 00315 - Gary Wheaton - Liane.sndh */
TIMEDB_ENTRY( 14f413e3,  0,   3584,       YM+TA), /* 04584 - Crawdaddy - Pure Energy 1.sndh */
TIMEDB_ENTRY( 151710e3,  0,   4032,          YM), /* 02971 - Gary Wheaton - The Cocoons of Jusse.sndh */
TIMEDB_ENTRY( 152668f3,  0,   5817,       YM+TA), /* 01330 - Scavenger - Enchanted Land The Sequel.sndh */
TIMEDB_ENTRY( 1527d7e4,  0,   4992,          YM), /* 01207 - Excellence In Art - Talkload.sndh */
TIMEDB_ENTRY( 1549cf77,  0,   8024,          YM), /* 04723 - Benny Johansen - Yatzy.sndh */
TIMEDB_ENTRY( 1549cf77,  1,    153,          YM), /* 04723 - Benny Johansen - Yatzy.sndh */
TIMEDB_ENTRY( 1549cf77,  2,     84,          YM), /* 04723 - Benny Johansen - Yatzy.sndh */
TIMEDB_ENTRY( 1549cf77,  3,   3360,          YM), /* 04723 - Benny Johansen - Yatzy.sndh */
TIMEDB_ENTRY( 15581aa9,  0,   5565,       YM+TA), /* 04242 - 505-checkpoint - BALIBULU.sndh */
TIMEDB_ENTRY( 157390be,  0,   9309,          YM), /* 04193 - David Whittaker - Outrun - USA Version.sndh */
TIMEDB_ENTRY( 157390be,  1,   5665,          YM), /* 04193 - David Whittaker - Outrun - USA Version.sndh */
TIMEDB_ENTRY( 157390be,  2,   1464,          YM), /* 04193 - David Whittaker - Outrun - USA Version.sndh */
TIMEDB_ENTRY( 15912b80,  0,  11136,    YM+TA+TD), /* 02274 - Tao - Cream - Walk On Ice (Steps).sndh */
TIMEDB_ENTRY( 159bc317,  0,   2578,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  1,   1761,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  2,   2305,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  3,   1153,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  4,    961,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  5,   2049,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  6,   2521,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  7,    225,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  8,    980,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317,  9,   2082,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317, 10,   2113,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159bc317, 11,    386,          YM), /* 00393 - Wally Beben - R-Type.sndh */
TIMEDB_ENTRY( 159c7186,  0,   3647,          YM), /* 03472 - gwEm - ze japanim.sndh */
TIMEDB_ENTRY( 15b445e9,  0,  79899,       YM+TA), /* 01372 - Scavenger - Having a Period in 3D Space.sndh */
TIMEDB_ENTRY( 15c672dc,  0,  15360, YM+TA+TB+TD), /* 03428 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 15d4bdc9,  0,  14592,    YM+TA+TD), /* 02503 - Crazy Q - Creators^DHS - Chameleon.sndh */
TIMEDB_ENTRY( 15e0eb75,  0,   4810,          YM), /* 03131 - Unknown - O Little Town of Bethlehem.sndh */
TIMEDB_ENTRY( 15ecb8cc,  0,   9600,          YM), /* 04838 - t1000 - Bits 33.sndh */
TIMEDB_ENTRY( 160fbf10,  0,   7704,          YM), /* 00617 - Chrispy Noodle - Tristar.sndh */
TIMEDB_ENTRY( 161a542a,  0,   6079, YM+TA+TB+TD), /* 01152 - Tyan - Platz 1.sndh */
TIMEDB_ENTRY( 16228d20,  0,   3079,          YM), /* 02735 - Jeremy Das - Jungle Boy.sndh */
TIMEDB_ENTRY( 1650dd24,  0,  16144,          YM), /* 00089 - David Whittaker - Cosmic Pirates.sndh */
TIMEDB_ENTRY( 1650dd24,  1,  15360,          YM), /* 00089 - David Whittaker - Cosmic Pirates.sndh */
TIMEDB_ENTRY( 1650dd24,  2,     26,          YM), /* 00089 - David Whittaker - Cosmic Pirates.sndh */
TIMEDB_ENTRY( 1650dd24,  3,    607,          YM), /* 00089 - David Whittaker - Cosmic Pirates.sndh */
TIMEDB_ENTRY( 167e07f3,  0,   6527, YM+TA+TB+TD), /* 02981 - 505-paradox - disco slide.sndh */
TIMEDB_ENTRY( 1692a08e,  0,   5565, YM+TA+TB+TD), /* 00203 - 505 - Checkpoint - tripping.sndh */
TIMEDB_ENTRY( 16bfa1c0,  0,   3840,          YM), /* 01074 - Excellence In Art - Bellis Boot.sndh */
TIMEDB_ENTRY( 16d4bf68,  0,   2687, YM+TA+TB+TD), /* 02872 - 505 <nils feske> - sequences, 2005.sndh */
TIMEDB_ENTRY( 16daeec1,  0,   4984,          YM), /* 00571 - Jas C Brooke - Resolution 101.sndh */
TIMEDB_ENTRY( 16daeec1,  1,    201,          YM), /* 00571 - Jas C Brooke - Resolution 101.sndh */
TIMEDB_ENTRY( 16e1874a,  0,   6316,          YM), /* 00079 - David Whittaker - Back To The Future 2.sndh */
TIMEDB_ENTRY( 16e1874a,  1,   3254,          YM), /* 00079 - David Whittaker - Back To The Future 2.sndh */
TIMEDB_ENTRY( 16e1874a,  2,   2984,          YM), /* 00079 - David Whittaker - Back To The Future 2.sndh */
TIMEDB_ENTRY( 16e1874a,  3,   5017,          YM), /* 00079 - David Whittaker - Back To The Future 2.sndh */
TIMEDB_ENTRY( 16e1874a,  4,   6342,          YM), /* 00079 - David Whittaker - Back To The Future 2.sndh */
TIMEDB_ENTRY( 16e4d0e2,  0,   8568,          YM), /* 00147 - David Whittaker - Tetris.sndh */
TIMEDB_ENTRY( 16e4d0e2,  1,   1120,          YM), /* 00147 - David Whittaker - Tetris.sndh */
TIMEDB_ENTRY( 16e4d0e2,  2,     61,          YM), /* 00147 - David Whittaker - Tetris.sndh */
TIMEDB_ENTRY( 16e61312,  0,   4611,       YM+TA), /* 04409 - Dr Beat - Maziac Unused.sndh */
TIMEDB_ENTRY( 16fedf6d,  0,   8233,          YM), /* 00117 - David Whittaker - Millennium 2.2.sndh */
TIMEDB_ENTRY( 170813e0,  0,  28032,          YM), /* 02180 - Mad Max - Atomix (Cybermind).sndh */
TIMEDB_ENTRY( 171600c0,  0,  29536, YM+TA+TB+TD), /* 03358 - Marcer - MIGHT 382.sndh */
TIMEDB_ENTRY( 173021f6,  0,   3455, YM+TA+TB+TD), /* 02988 - 505-paradox - Jumper.sndh */
TIMEDB_ENTRY( 1734fa62,  0,   1152,          YM), /* 04616 - Timo Weggen - Plantonix.sndh */
TIMEDB_ENTRY( 1734fa62,  1,   2304,          YM), /* 04616 - Timo Weggen - Plantonix.sndh */
TIMEDB_ENTRY( 1734fa62,  2,   2688,          YM), /* 04616 - Timo Weggen - Plantonix.sndh */
TIMEDB_ENTRY( 177a0f9d,  0,   5413, YM+TA+TB+TD), /* 00055 - 505 - Heaven Sux.sndh */
TIMEDB_ENTRY( 177d17ab,  0,  16231,          YM), /* 03548 - Unknown - Knucklebusters.sndh */
TIMEDB_ENTRY( 177e6d27,  0,   9409, YM+TA+TB+TD), /* 03583 - Divag - Electropsychiatrique.sndh */
TIMEDB_ENTRY( 17836e5d,  0,  11168, YM+TA+TB+TD), /* 03852 - !Cube - Decisive Factor.sndh */
TIMEDB_ENTRY( 1793d65d,  0,  95719,       YM+TA), /* 04799 - Unknown - Confuzion.sndh */
TIMEDB_ENTRY( 179d1c88,  0,   8556, YM+TA+TB+TD), /* 04660 - Frequent-Ephidrena - Revers.sndh */
TIMEDB_ENTRY( 17a8e5a3,  0,  29283,       YM+TA), /* 04404 - Ninja One - Monkey (Hoby One Demo).sndh */
TIMEDB_ENTRY( 17c24cd9,  0,   1536,       YM+TA), /* 00920 - Dark Angel - C94cop.sndh */
TIMEDB_ENTRY( 17c7fb3e,  0,   5375, YM+TA+TB+TD), /* 03274 - Dubmood - Galax (orig. Ben Daglish).sndh */
TIMEDB_ENTRY( 17cfdfc7,  0,   7968,       YM+TA), /* 00807 - Crawdaddy - Poing.sndh */
TIMEDB_ENTRY( 17e037c4,  0,   9594,    YM+TA+TD), /* 01007 - DMA-SC - Pie Bill Gates - Game.sndh */
TIMEDB_ENTRY( 17e09dc9,  0,   6265,    YM+TA+TD), /* 01964 - MSG - Godpey - Stage.sndh */
TIMEDB_ENTRY( 17e0ec66,  0,   4415, YM+TA+TB+TD), /* 02996 - 505-paradox - phazer (STE only).sndh */
TIMEDB_ENTRY( 17e34c6a,  0,  12548,    YM+TA+TB), /* 01833 - Lotek Style - Artefakt.sndh */
TIMEDB_ENTRY( 17f47817,  0,   4608,          YM), /* 00496 - Ultimatum - Kidney Bean 12.sndh */
TIMEDB_ENTRY( 17f8abde,  0,   8058, YM+TA+TB+TD), /* 01371 - Scavenger - Chaos.sndh */
TIMEDB_ENTRY( 183e8fe0,  0,   8585,       YM+TA), /* 04432 - Lotus - Synth Dream 2 - Micro Kit.sndh */
TIMEDB_ENTRY( 1842c149,  0,  13440,          YM), /* 01178 - Sally - Stax 57.sndh */
TIMEDB_ENTRY( 184a5220,  0,   5754,    YM+TA+TD), /* 00063 - 505 - Joe Cool.sndh */
TIMEDB_ENTRY( 186c4ab7,  0,  29736,       YM+TD), /* 01463 - STU - DROP DA BOMB.sndh */
TIMEDB_ENTRY( 18707f17,  0,   3466,          YM), /* 03078 - Mark Quinn - Valse Venezolano, by A. Lauro.sndh */
TIMEDB_ENTRY( 1878be4e,  0,  11136,          YM), /* 00897 - Danny Oneway - Danny.sndh */
TIMEDB_ENTRY( 18a9400b,  0,  12670,       YM+TA), /* 04864 - Jovis - Baf.sndh */
TIMEDB_ENTRY( 18aadb4d,  0,   6050,          YM), /* 00366 - Steve Barrett - SAS Combat Simulator.sndh */
TIMEDB_ENTRY( 18c1972e,  0,  15232,          YM), /* 01663 - Jedi - Virtual Escape Hidden.sndh */
TIMEDB_ENTRY( 18c378fe,  0,   5376,          YM), /* 02659 - Stefan Jerowski - Blue Angel 69.sndh */
TIMEDB_ENTRY( 18c378fe,  1,  11420,          YM), /* 02659 - Stefan Jerowski - Blue Angel 69.sndh */
TIMEDB_ENTRY( 18c378fe,  2,    435,          YM), /* 02659 - Stefan Jerowski - Blue Angel 69.sndh */
TIMEDB_ENTRY( 18c378fe,  3,    414,          YM), /* 02659 - Stefan Jerowski - Blue Angel 69.sndh */
TIMEDB_ENTRY( 18e73b47,  0,  11807, YM+TA+TB+TD), /* 02702 - Dubmood - Publicea i Media beat.sndh */
TIMEDB_ENTRY( 18fd5ccc,  0,   8712,          YM), /* 00880 - Ben Daglish - The Running Man.sndh */
TIMEDB_ENTRY( 190849c1,  0,   2239,          YM), /* 01295 - Matt Furniss - Euro Football Champs.sndh */
TIMEDB_ENTRY( 1908add3,  0,  40320,          YM), /* 01470 - STU - MUGICHA.sndh */
TIMEDB_ENTRY( 190ea54c,  0,   6144,          YM), /* 02811 - Starbuck - Urban Shuffle.sndh */
TIMEDB_ENTRY( 19377722,  0,  12840,          YM), /* 00971 - Deneb - Kidney Bean - Data.sndh */
TIMEDB_ENTRY( 193886d5,  0,   3072,          YM), /* 00338 - Tony Williams - Chuckie Egg.sndh */
TIMEDB_ENTRY( 193886d5,  1,   7680,          YM), /* 00338 - Tony Williams - Chuckie Egg.sndh */
TIMEDB_ENTRY( 193886d5,  2,    171,          YM), /* 00338 - Tony Williams - Chuckie Egg.sndh */
TIMEDB_ENTRY( 193cfddc,  0,   6257, YM+TA+TB+TD), /* 04278 - 505 - Lousy.sndh */
TIMEDB_ENTRY( 19654ee5,  0,   2303,          YM), /* 02823 - COOPER, ORIGINAL BY CLAWZ - ZNAX HIGHSCORES.sndh */
TIMEDB_ENTRY( 196ad751,  0,   6938,          YM), /* 03124 - Unknown - I Am A Rock        P.Simon.sndh */
TIMEDB_ENTRY( 196c6bbd,  0,  10253,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 196c6bbd,  1,   1150,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 196c6bbd,  2,   3070,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 196c6bbd,  3,    144,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 196c6bbd,  4,    119,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 196c6bbd,  5,    766,          YM), /* 01296 - Matt Furniss - E.F.T.P.O.T.R.M..sndh */
TIMEDB_ENTRY( 199176ae,  0,   5794,       YM+TA), /* 00000 - Dhor - Uncomplicated.sndh */
TIMEDB_ENTRY( 19954286,  0,   1280,          YM), /* 01076 - Excellence In Art - Circles.sndh */
TIMEDB_ENTRY( 19a473a7,  0,   6912,    YM+TA+TD), /* 02326 - Tao - Cream - MV Jeu - Highscore (Steps).sndh */
TIMEDB_ENTRY( 19ae899a,  0,   3261, YM+TA+TB+TD), /* 00275 - Baggio - Compo.sndh */
TIMEDB_ENTRY( 19cb007c,  0,    768,       YM+TA), /* 01751 - Kanewood - Phantom Demo - Bonus B.sndh */
TIMEDB_ENTRY( 19ce95d7,  0,   1333,       YM+TA), /* 04424 - Lotus - Synth Dream 2 - Braindrops.sndh */
TIMEDB_ENTRY( 19e62de8,  0,   5631, YM+TA+TB+TD), /* 02979 - 505-paradox - chippy.sndh */
TIMEDB_ENTRY( 19f2d59e,  0,  10746,          YM), /* 01235 - Epic - Phototro.sndh */
TIMEDB_ENTRY( 19f4d71f,  0,   4864,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  1,   5760,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  2,   6144,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  3,   6144,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  4,   5632,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  5,   4608,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  6,   6400,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  7,   6765,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  8,  11904,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f,  9,   5889,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 10,   8160,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 11,   9184,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 12,   9984,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 13,  15192,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 14,   9912,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 15,   7800,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 16,   3520,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 17,   3840,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 18,   6144,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f4d71f, 19,  10944,      YM+STE), /* 02177 - Mad Max - Amberstar.sndh */
TIMEDB_ENTRY( 19f5d22e,  0,   7290,          YM), /* 00890 - Adrian Dalecki - Turtle Table Tennis.sndh */
TIMEDB_ENTRY( 19f5d22e,  1,    196,          YM), /* 00890 - Adrian Dalecki - Turtle Table Tennis.sndh */
TIMEDB_ENTRY( 1a16010c,  0,   6396,          YM), /* 01242 - Epic - Wizz.sndh */
TIMEDB_ENTRY( 1a1ce50e,  0,   3456,          YM), /* 02127 - Mad Max - Robocop Tune 2.sndh */
TIMEDB_ENTRY( 1a2334b7,  0,    963,          YM), /* 03510 - Nick Scarim - Spy Vs Spy.sndh */
TIMEDB_ENTRY( 1a25b77a,  0,   6144,    YM+TA+TD), /* 03317 - SH3-RG - Ingame.sndh */
TIMEDB_ENTRY( 1a2f66c9,  0,    448,    YM+TA+TD), /* 02504 - Crazy Q - DHS Disk Change 1 (Test).sndh */
TIMEDB_ENTRY( 1a3c29db,  0,   7161, YM+TA+TB+TD), /* 01381 - Schemantix - Zombie Apocalypse.sndh */
TIMEDB_ENTRY( 1a3c29db,  1,   4473, YM+TA+TB+TD), /* 01381 - Schemantix - Zombie Apocalypse.sndh */
TIMEDB_ENTRY( 1a3c29db,  2,   3576, YM+TA+TB+TD), /* 01381 - Schemantix - Zombie Apocalypse.sndh */
TIMEDB_ENTRY( 1a3c29db,  3,   6136, YM+TA+TB+TD), /* 01381 - Schemantix - Zombie Apocalypse.sndh */
TIMEDB_ENTRY( 1a3f6f91,  0,   4608,          YM), /* 03192 - n-a - Sports Day.sndh */
TIMEDB_ENTRY( 1a41cde6,  0,  15360,          YM), /* 01507 - Rob Hubbard - Thrust.sndh */
TIMEDB_ENTRY( 1a43fdf6,  0,  10232, YM+TA+TB+TD), /* 01829 - Lotek Style - Age.sndh */
TIMEDB_ENTRY( 1a4c351f,  0,  14268,    YM+TB+TD), /* 02848 - Helene - Carsong1 micromusic 4 the road.sndh */
TIMEDB_ENTRY( 1a5f9b1f,  0,   6656,          YM), /* 01655 - Jedi - Shadow.sndh */
TIMEDB_ENTRY( 1a621f36,  0,   5802,          YM), /* 00260 - Antichrist - Sad Tune.sndh */
TIMEDB_ENTRY( 1a6e73aa,  0,  13824,          YM), /* 02211 - Mad Max - Quizmaster.sndh */
TIMEDB_ENTRY( 1a7e4769,  0,   3023,    YM+TA+TD), /* 01939 - Modmate - Xmas.sndh */
TIMEDB_ENTRY( 1a93b511,  0,   7679,          YM), /* 04849 - !Cube - Glittering Waves.sndh */
TIMEDB_ENTRY( 1aa645f9,  0,   2367,          YM), /* 03208 - Unknown - International Soccer Challenge.sndh */
TIMEDB_ENTRY( 1ae0d0e0,  0,  10043,          YM), /* 00882 - Ben Daglish - Wizard Warz.sndh */
TIMEDB_ENTRY( 1ae0d0e0,  1,    236,          YM), /* 00882 - Ben Daglish - Wizard Warz.sndh */
TIMEDB_ENTRY( 1ae0d0e0,  2,    693,          YM), /* 00882 - Ben Daglish - Wizard Warz.sndh */
TIMEDB_ENTRY( 1ae0d0e0,  3,    192,          YM), /* 00882 - Ben Daglish - Wizard Warz.sndh */
TIMEDB_ENTRY( 1ae0d0e0,  4,    420,          YM), /* 00882 - Ben Daglish - Wizard Warz.sndh */
TIMEDB_ENTRY( 1ae10543,  0,  12672,       YM+TA), /* 00401 - Big Alec - Bangkok Knights Loader.sndh */
TIMEDB_ENTRY( 1ae6ad10,  0,   9595, YM+TA+TB+TD), /* 01865 - Lotek Style - tSCc^Vinyl Kartell - Rise Of The RI-Machines (YM Rockerz).sndh */
TIMEDB_ENTRY( 1af0890d,  0,   6197,          YM), /* 00354 - David Whittaker - Trantor.sndh */
TIMEDB_ENTRY( 1af0890d,  1,   1285,          YM), /* 00354 - David Whittaker - Trantor.sndh */
TIMEDB_ENTRY( 1af0890d,  2,    581,          YM), /* 00354 - David Whittaker - Trantor.sndh */
TIMEDB_ENTRY( 1af0890d,  3,    630,          YM), /* 00354 - David Whittaker - Trantor.sndh */
TIMEDB_ENTRY( 1b00ee68,  0,  10752,    YM+TB+TD), /* 02850 - Helene - eurotrance versjon.sndh */
TIMEDB_ENTRY( 1b042874,  0,   1280,          YM), /* 04726 - Beyonder - MIX number ONE demo - Intro.sndh */
TIMEDB_ENTRY( 1b042874,  1,   2080,          YM), /* 04726 - Beyonder - MIX number ONE demo - Intro.sndh */
TIMEDB_ENTRY( 1b042874,  2,     93,          YM), /* 04726 - Beyonder - MIX number ONE demo - Intro.sndh */
TIMEDB_ENTRY( 1b042874,  3,   2224,          YM), /* 04726 - Beyonder - MIX number ONE demo - Intro.sndh */
TIMEDB_ENTRY( 1b042874,  4,   2662,          YM), /* 04726 - Beyonder - MIX number ONE demo - Intro.sndh */
TIMEDB_ENTRY( 1b04a226,  0,   1152,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  1,   1920,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  2,   2016,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  3,   2212,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  4,   2320,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  5,   1449,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  6,   2324,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  7,    612,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  8,  22423,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226,  9,   2304,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04a226, 10,   3072,          YM), /* 04389 - Nigel Pritchard - Times of Lore.sndh */
TIMEDB_ENTRY( 1b04cc71,  0,  22848,          YM), /* 02039 - Mad Max - International Karate.sndh */
TIMEDB_ENTRY( 1b071dd3,  0,  41632,    YM+TA+TD), /* 02330 - Tao - Cream - Sillyshuffle (SV2K Compo).sndh */
TIMEDB_ENTRY( 1b199a9a,  0,   5098, YM+TA+TB+TD), /* 04201 - Paul Summers - Predator.sndh */
TIMEDB_ENTRY( 1b233e4c,  0,   4992,          YM), /* 01645 - Jedi - Thats The Way I Like It.sndh */
TIMEDB_ENTRY( 1b30ffba,  0,   3264,          YM), /* 01765 - Lap - Lap 5.sndh */
TIMEDB_ENTRY( 1b367860,  0,   1785,       YM+TA), /* 00223 - 7AN - Blip 4.sndh */
TIMEDB_ENTRY( 1b36c809,  0,   1346,          YM), /* 03286 - Philip Bishop - Dogfight.sndh */
TIMEDB_ENTRY( 1b4950a0,  0,   3336,          YM), /* 02719 - Chris Denman - Feudal Lords.sndh */
TIMEDB_ENTRY( 1b4d1ba1,  0,   1530, YM+TA+TB+TD), /* 02827 - Msg - Superfly tune 4.sndh */
TIMEDB_ENTRY( 1b739836,  0,   5518,       YM+TA), /* 03567 - 505-Paradox - Pozdrowienia do Gdanska.sndh */
TIMEDB_ENTRY( 1b7f470a,  0,   7680,          YM), /* 01094 - Techno - Omed.sndh */
TIMEDB_ENTRY( 1b86d25d,  0,   3584,          YM), /* 01555 - Rhino - Xmas 2003 Intro - Song 3.sndh */
TIMEDB_ENTRY( 1b893559,  0,   1488,          YM), /* 01382 - Chris Scudds - Chase HQ 2 Special Criminal Investigation.sndh */
TIMEDB_ENTRY( 1b893559,  1,   3936,          YM), /* 01382 - Chris Scudds - Chase HQ 2 Special Criminal Investigation.sndh */
TIMEDB_ENTRY( 1b893559,  2,   3080,          YM), /* 01382 - Chris Scudds - Chase HQ 2 Special Criminal Investigation.sndh */
TIMEDB_ENTRY( 1b893559,  3,   3430,          YM), /* 01382 - Chris Scudds - Chase HQ 2 Special Criminal Investigation.sndh */
TIMEDB_ENTRY( 1b92ca5a,  0,   3664,          YM), /* 03194 - n-a - Witches.sndh */
TIMEDB_ENTRY( 1b92d24d,  0,    209,          YM), /* 04509 - Unknown - Bumber Jacky.sndh */
TIMEDB_ENTRY( 1b92d24d,  1,    465,          YM), /* 04509 - Unknown - Bumber Jacky.sndh */
TIMEDB_ENTRY( 1ba3b22f,  0,   7968,       YM+TA), /* 02869 - Crawdaddy - Lame-Brain.sndh */
TIMEDB_ENTRY( 1bb19ba2,  0,   8952, YM+TA+TB+TD), /* 02860 - Helene - Version L.sndh */
TIMEDB_ENTRY( 1bb6db89,  0,   2026,          YM), /* 03886 - Unknown - Dobru Noc.sndh */
TIMEDB_ENTRY( 1bd0d72a,  0,   3920,       YM+TD), /* 02366 - Nexus 6 - Chinese Mead.sndh */
TIMEDB_ENTRY( 1bebdb7c,  0,   4359, YM+TA+TB+TD), /* 02523 - DUBMOOD - RaZoR 1911 - MONKEY ISLAND.sndh */
TIMEDB_ENTRY( 1c0016df,  0,   6255,          YM), /* 00115 - David Whittaker - Licence to Kill.sndh */
TIMEDB_ENTRY( 1c0016df,  1,   8192,          YM), /* 00115 - David Whittaker - Licence to Kill.sndh */
TIMEDB_ENTRY( 1c0016df,  2,   9216,          YM), /* 00115 - David Whittaker - Licence to Kill.sndh */
TIMEDB_ENTRY( 1c0016df,  3,    147,          YM), /* 00115 - David Whittaker - Licence to Kill.sndh */
TIMEDB_ENTRY( 1c0016df,  4,    668,          YM), /* 00115 - David Whittaker - Licence to Kill.sndh */
TIMEDB_ENTRY( 1c24767c,  0,  13056,          YM), /* 01649 - Jedi - Loader.sndh */
TIMEDB_ENTRY( 1c2f24ac,  0,   9560,          YM), /* 03126 - Unknown - I Just Called to Say I Love U.sndh */
TIMEDB_ENTRY( 1c3714a1,  0,  35506,          YM), /* 01473 - stu - punkey.sndh */
TIMEDB_ENTRY( 1c432308,  0,   3840,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  1,   6353,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  2,    432,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  3,   1768,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  4,   1920,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  5,    673,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  6,    673,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  7,   1344,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  8,    193,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308,  9,   1184,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c432308, 10,    129,          YM), /* 01055 - Jonathan Dunn - The Addams Family.sndh */
TIMEDB_ENTRY( 1c976fba,  0,  17594,    YM+TA+TD), /* 02626 - sh3-rg - teh won.sndh */
TIMEDB_ENTRY( 1ca333b3,  0,   3224,       YM+TA), /* 04304 - 505-checkpoint - Scheme.sndh */
TIMEDB_ENTRY( 1cbe11b8,  0,  18059,    YM+TB+TD), /* 04574 - Nemo - Undercover Action Agents.sndh */
TIMEDB_ENTRY( 1ccbc756,  0,   1736,          YM), /* 03089 - Unknown - Don't Stop Believing,"Journey".sndh */
TIMEDB_ENTRY( 1cf75009,  0,   3095, YM+TA+TB+TD), /* 03421 - Dubmood - Tempest - Acidjazzed Evenings (Dubmood).sndh */
TIMEDB_ENTRY( 1d005ee4,  0,   4771,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  1,   5376,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  2,   5184,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  3,   6528,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  4,   4992,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  5,   3619,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d005ee4,  6,   5376,          YM), /* 02033 - Mad Max - Gerry The Germ.sndh */
TIMEDB_ENTRY( 1d022c80,  0,  11568, YM+TA+TB+TD), /* 02793 - Frequent-Ephidrena (2006) - Dronning.sndh */
TIMEDB_ENTRY( 1d0c3960,  0,    135,          YM), /* 02307 - Tao - Cream - F6 v2 (EIL joke).sndh */
TIMEDB_ENTRY( 1d0c4127,  0,  12544,       YM+TD), /* 02016 - Jochen Hippel - Virtual Escape.sndh */
TIMEDB_ENTRY( 1d103200,  0,      1,         STE), /* 04646 - 505 - Saboteaser.sndh */
TIMEDB_ENTRY( 1d197f64,  0,   9759,          YM), /* 03540 - Allister Brimble - Goal.sndh */
TIMEDB_ENTRY( 1d23e0db,  0,  41471, YM+TA+TB+TD), /* 02795 - Frequent-Ephidrena - Vrien DMA.sndh */
TIMEDB_ENTRY( 1d30a779,  0,   7946,          YM), /* 00746 - Sean Conran - Videokid.sndh */
TIMEDB_ENTRY( 1d30a779,  1,   2331,          YM), /* 00746 - Sean Conran - Videokid.sndh */
TIMEDB_ENTRY( 1d30a779,  2,    298,          YM), /* 00746 - Sean Conran - Videokid.sndh */
TIMEDB_ENTRY( 1d410b20,  0,   8320,          YM), /* 00448 - Augustin Vidovic - Human Race 4.sndh */
TIMEDB_ENTRY( 1d4f7ec7,  0,  45116,       YM+TA), /* 03765 - Waldemar Belwon - Big Bang (Music Show 1).sndh */
TIMEDB_ENTRY( 1d6617bb,  0,   5375,          YM), /* 02968 - Unknown - Super Space Invaders.sndh */
TIMEDB_ENTRY( 1d6617bb,  1,    640,          YM), /* 02968 - Unknown - Super Space Invaders.sndh */
TIMEDB_ENTRY( 1d6617bb,  2,   1280,          YM), /* 02968 - Unknown - Super Space Invaders.sndh */
TIMEDB_ENTRY( 1d6617bb,  3,    767,          YM), /* 02968 - Unknown - Super Space Invaders.sndh */
TIMEDB_ENTRY( 1d6c637b,  0,   6906, YM+TA+TB+TD), /* 00904 - Danny Oneway - In Door.sndh */
TIMEDB_ENTRY( 1d6ca6e5,  0,   6914,          YM), /* 00244 - AN Cool - M-Demo 4.sndh */
TIMEDB_ENTRY( 1d804ac2,  0, 131669,       YM+TA), /* 04025 - Roggie - Dreams.sndh */
TIMEDB_ENTRY( 1d82271c,  0,   9174,       YM+TA), /* 01843 - Lotek Style - tSCc^ Vinyl Kartell ( 2003 ) - Dancing T-Girl (YM Rockerz).sndh */
TIMEDB_ENTRY( 1d8e7887,  0,  14209,    YM+TA+TD), /* 02339 - Tao - Cream - 13 (UCM 13).sndh */
TIMEDB_ENTRY( 1d8e7bfb,  0,   3168,          YM), /* 02371 - Nexus 6 - Daniela's Arsch.sndh */
TIMEDB_ENTRY( 1d93ed13,  0,   5376,       YM+TA), /* 04593 - Crawdaddy - The Joker.sndh */
TIMEDB_ENTRY( 1d9e9891,  0,   6528,          YM), /* 04399 - Anthony Lees - Fireball.sndh */
TIMEDB_ENTRY( 1d9e9891,  1,    288,          YM), /* 04399 - Anthony Lees - Fireball.sndh */
TIMEDB_ENTRY( 1d9e9891,  2,    768,          YM), /* 04399 - Anthony Lees - Fireball.sndh */
TIMEDB_ENTRY( 1d9e9891,  3,    160,          YM), /* 04399 - Anthony Lees - Fireball.sndh */
TIMEDB_ENTRY( 1da2a5d8,  0,   5760,          YM), /* 01160 - Tyan - Spin On This Megademo.sndh */
TIMEDB_ENTRY( 1dc4a1ea,  0,  26112, YM+TA+TB+TD), /* 02530 - Nexus 6 - Tuxtrip.sndh */
TIMEDB_ENTRY( 1ded33f5,  0,  22393,       YM+TA), /* 00582 - Buster - Impact - digi.sndh */
TIMEDB_ENTRY( 1df411ca,  0,   8640,          YM), /* 02742 - Unknown Composer - International Karate.sndh */
TIMEDB_ENTRY( 1e1016d5,  0,   3839, YM+TA+TB+TD), /* 03390 - Dubmood - Dbug^Razor1911 - Daddy Freddy - Bring Your Own Bottle.sndh */
TIMEDB_ENTRY( 1e1d7a1c,  0,  13033,          YM), /* 03041 - Ed Bogas - Brandenburg Concerto 2 by Bach.sndh */
TIMEDB_ENTRY( 1e1f129e,  0,   4863,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  1,   1593,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  2,   1610,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  3,   2788,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  4,   1689,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  5,   1474,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  6,    778,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  7,    803,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  8,    779,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e,  9,    559,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e1f129e, 10,   2635,          YM), /* 01804 - Barry Leitch - The Shoe People.sndh */
TIMEDB_ENTRY( 1e30fa3f,  0,   6136,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  1,   3064,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  2,   1565,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  3,   3064,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  4,   2044,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  5,   1528,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  6,   2302,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e30fa3f,  7,   1434,          YM), /* 02936 - Matt Furniss - Ultima VI.sndh */
TIMEDB_ENTRY( 1e3a28ce,  0,   4607, YM+TA+TB+TD), /* 02727 - maktone-fairlight - hero.sndh */
TIMEDB_ENTRY( 1e545f59,  0,   6915,          YM), /* 02228 - Tao - Dots Revenge (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 1e5bd747,  0,   9726,          YM), /* 02737 - Gary Wheaton - Trotsky Crazy.sndh */
TIMEDB_ENTRY( 1e5c095b,  0,   1036,          YM), /* 00465 - Augustin Vidovic - Troudbal.sndh */
TIMEDB_ENTRY( 1e626998,  0,   1942,          YM), /* 02124 - Mad Max - Cuddly Demos - No Name.sndh */
TIMEDB_ENTRY( 1e62974f,  0,   3065,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  1,    683,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  2,    381,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  3,   2665,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  4,   1943,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  5,    457,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  6,   3059,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e62974f,  7,   1420,          YM), /* 04615 - Al Lowe - Leisure Suit Larry - Lounge Lizards.sndh */
TIMEDB_ENTRY( 1e6a3f09,  0,   4027,          YM), /* 03108 - Unknown - ST. ELSEWHERE by Dave Grusin.sndh */
TIMEDB_ENTRY( 1e736ec8,  0,   7424,          YM), /* 04350 - YERZMYEY-AY-RIDERS - OUT OF THE MIST.sndh */
TIMEDB_ENTRY( 1e841691,  0,  39807, YM+TA+TB+TD), /* 03332 - Marcer - cpu slider preview.sndh */
TIMEDB_ENTRY( 1e8d2599,  0,  61750,       YM+TA), /* 04356 - X The Beermat - Drag.sndh */
TIMEDB_ENTRY( 1ea83810,  0,  21504,       YM+TA), /* 00772 - Count Zero - Decade Demo - Wheres My Willy.sndh */
TIMEDB_ENTRY( 1ee3509a,  0,  12537,       YM+TA), /* 01684 - Jess - Let ze Djoard Bells Ring.sndh */
TIMEDB_ENTRY( 1ef00b04,  0,   7679,          YM), /* 04649 - ActoDi - Friday.sndh */
TIMEDB_ENTRY( 1efb7a03,  0,   5530,          YM), /* 01237 - Epic - Shoot All Coders.sndh */
TIMEDB_ENTRY( 1efba80a,  0,   2304,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  1,    512,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  2,   2208,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  3,   2816,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  4,   3584,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  5,   2015,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  6,   1536,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  7,    223,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efba80a,  8,    192,          YM), /* 00750 - Mark Cooksey - Crack Down.sndh */
TIMEDB_ENTRY( 1efc3b9f,  0,   8453,    YM+TA+TB), /* 03416 - gwEm - PHF - Good Old Fashioned Blipp Blopper!.sndh */
TIMEDB_ENTRY( 1f08883c,  0,   5376,          YM), /* 02186 - Mad Max - Pro Tennis Tour 2.sndh */
TIMEDB_ENTRY( 1f11f49d,  0,   3840,          YM), /* 00747 - Sean Conran - Wacky Races.sndh */
TIMEDB_ENTRY( 1f221540,  0,   9600,          YM), /* 00785 - Crawdaddy - Heroes.sndh */
TIMEDB_ENTRY( 1f27f7d4,  0,   2304,    YM+TA+TB), /* 03412 - Dma-Sc - mazemenu.sndh */
TIMEDB_ENTRY( 1f2fa7c6,  0,   1789,          YM), /* 02475 - Iso - People Let Go.sndh */
TIMEDB_ENTRY( 1f36681d,  0,  43520,       YM+TD), /* 04860 - xFalcon - Christian Bunge - HAPPY CHIP.sndh */
TIMEDB_ENTRY( 1f3bf399,  0,   4415, YM+TA+TB+TD), /* 03667 - Starbuck - exercise wheel.sndh */
TIMEDB_ENTRY( 1f674ed8,  0,   7868,          YM), /* 00159 - David Whittaker - Xenon 2.sndh */
TIMEDB_ENTRY( 1f739f4e,  0,   7874,          YM), /* 00369 - Steve Barrett - Violator.sndh */
TIMEDB_ENTRY( 1f739f4e,  1,    218,          YM), /* 00369 - Steve Barrett - Violator.sndh */
TIMEDB_ENTRY( 1f739f4e,  2,    290,          YM), /* 00369 - Steve Barrett - Violator.sndh */
TIMEDB_ENTRY( 1f7a187f,  0,  13536,          YM), /* 00775 - Count Zero - Lost Boys Disk Maggie 2.sndh */
TIMEDB_ENTRY( 1f82847a,  0,   4368,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  1,   7393,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  2,   1296,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  3,   1536,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  4,   2784,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  5,   1152,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  6,   1120,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  7,   6924,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  8,  10368,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a,  9,    864,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1f82847a, 10,   5623,          YM), /* 01513 - Chris Huelsbeck - Jim Power in Mutant Planet.sndh */
TIMEDB_ENTRY( 1fa548f5,  0,   5370,    YM+TB+TD), /* 02426 - 505-cpt^dox - smiley.sndh */
TIMEDB_ENTRY( 1fbe7d94,  0,  11829,          YM), /* 02245 - Tao - Cream (orig. Rob Hubbard) - Spellbound (Stmygm).sndh */
TIMEDB_ENTRY( 1fcdf58d,  0,   9656,          YM), /* 00422 - Big Alec - PYM - Change Disks.sndh */
TIMEDB_ENTRY( 1ffc4cc1,  0,   4096,          YM), /* 01300 - Matt Furniss - Jahanghir Khan's Squash.sndh */
TIMEDB_ENTRY( 1ffc4cc1,  1,   5632,          YM), /* 01300 - Matt Furniss - Jahanghir Khan's Squash.sndh */
TIMEDB_ENTRY( 1ffc4cc1,  2,    197,          YM), /* 01300 - Matt Furniss - Jahanghir Khan's Squash.sndh */
TIMEDB_ENTRY( 1ffc4cc1,  3,    197,          YM), /* 01300 - Matt Furniss - Jahanghir Khan's Squash.sndh */
TIMEDB_ENTRY( 1fff055f,  0,   1919,          YM), /* 01917 - Jon Medek - Hard Nova.sndh */
TIMEDB_ENTRY( 202606bd,  0,   2930,          YM), /* 00741 - Sean Conran - Defenders of the Earth.sndh */
TIMEDB_ENTRY( 202606bd,  1,    541,          YM), /* 00741 - Sean Conran - Defenders of the Earth.sndh */
TIMEDB_ENTRY( 202606bd,  2,    543,          YM), /* 00741 - Sean Conran - Defenders of the Earth.sndh */
TIMEDB_ENTRY( 202606bd,  3,   3540,          YM), /* 00741 - Sean Conran - Defenders of the Earth.sndh */
TIMEDB_ENTRY( 20407076,  0,   9109, YM+TA+TB+TD), /* 03585 - Divag - Mental Fury.sndh */
TIMEDB_ENTRY( 204dfdd6,  0,   3918,          YM), /* 01288 - Matt Furniss - AV8B Harrier.sndh */
TIMEDB_ENTRY( 2057279c,  0,   9863,       YM+TA), /* 02329 - Tao - Cream - Ride The Sky (Digidrum-version).sndh */
TIMEDB_ENTRY( 205a2964,  0,   2834,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  1,   2633,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  2,   1755,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  3,   1536,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  4,   5266,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  5,    877,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  6,   1462,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  7,    877,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  8,   1462,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964,  9,   1616,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964, 10,    728,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 205a2964, 11,   1389,          YM), /* 00083 - David Whittaker - Chips Challenge.sndh */
TIMEDB_ENTRY( 20663a2a,  0,   1538,          YM), /* 00417 - Big Alec - Offbeat.sndh */
TIMEDB_ENTRY( 206a8bf9,  0,  12931,          YM), /* 00333 - Gary Wheaton - Boom.sndh */
TIMEDB_ENTRY( 20a6777b,  0,   5621,          YM), /* 01733 - Jookie - Xmas Intro.sndh */
TIMEDB_ENTRY( 20b4380c,  0,   4521,          YM), /* 03059 - Ed Bogas - Kyrie eleison-Mozart's REQUIEM.sndh */
TIMEDB_ENTRY( 20be5296,  0,   5151, YM+TA+TB+TD), /* 04243 - 505-Checkpoint (2000) - banga.sndh */
TIMEDB_ENTRY( 20d12ca5,  0,  42411,          YM), /* 01686 - Jess - See ya Again on Falcon.sndh */
TIMEDB_ENTRY( 20e00d95,  0,   4613,          YM), /* 02449 - David Whittaker - Solomons Key.sndh */
TIMEDB_ENTRY( 20e00d95,  1,   3749,          YM), /* 02449 - David Whittaker - Solomons Key.sndh */
TIMEDB_ENTRY( 20f1e35b,  0,  12662,          YM), /* 03409 - Allister Brimble - Sword and the Rose.sndh */
TIMEDB_ENTRY( 20f1e35b,  1,    272,          YM), /* 03409 - Allister Brimble - Sword and the Rose.sndh */
TIMEDB_ENTRY( 20f1e35b,  2,    342,          YM), /* 03409 - Allister Brimble - Sword and the Rose.sndh */
TIMEDB_ENTRY( 20f1e35b,  3,   1350,          YM), /* 03409 - Allister Brimble - Sword and the Rose.sndh */
TIMEDB_ENTRY( 211433b8,  0,  45604,       YM+TA), /* 04355 - X The Beermat - Deal.sndh */
TIMEDB_ENTRY( 2116eca1,  0,  26400,          YM), /* 01479 - STU - TOURIYANSE.sndh */
TIMEDB_ENTRY( 21333683,  0,   7536,       YM+TA), /* 00814 - Crawdaddy - Gods Theme.sndh */
TIMEDB_ENTRY( 213d81b7,  0,   4935,       YM+TA), /* 01933 - Miq - Hotliner v2.sndh */
TIMEDB_ENTRY( 214604ea,  0,   8648, YM+TA+TB+TD), /* 03474 - gwEm - Different World (dub mix).sndh */
TIMEDB_ENTRY( 2146cdf1,  0,   4218,          YM), /* 02880 - Beast - Snabblat.sndh */
TIMEDB_ENTRY( 21498355,  0,   6399,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  1,    359,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  2,     72,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  3,    192,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  4,    128,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  5,    576,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  6,    448,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  7,   5245,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 21498355,  8,   4405,          YM), /* 02445 - Tony Williams - Edd the Duck!.sndh */
TIMEDB_ENTRY( 2154f975,  0,   3839,          YM), /* 03532 - Jean Baudlot - Castle Warrior.sndh */
TIMEDB_ENTRY( 21588958,  0,   8256,          YM), /* 01083 - Techno - X-Show 1 - Intro.sndh */
TIMEDB_ENTRY( 216b8a9a,  0,   4992,          YM), /* 01175 - Sally - Sally 1.sndh */
TIMEDB_ENTRY( 216ba57f,  0,  20352,          YM), /* 02876 - Agrajag - 3.5k Megademo Menu.sndh */
TIMEDB_ENTRY( 218084c3,  0,   6339,       YM+TA), /* 04586 - Technoblip - Pure Energy 3.sndh */
TIMEDB_ENTRY( 21c43e70,  0,   3072,          YM), /* 01496 - Alex Herbert - Delta.sndh */
TIMEDB_ENTRY( 21d10358,  0,  55934,       YM+TA), /* 04808 - Unknown - Oh Yeah.sndh */
TIMEDB_ENTRY( 21d994de,  0,   6850,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  1,    480,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  2,     28,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  3,    961,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  4,    801,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  5,   1240,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  6,    777,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  7,    612,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  8,    763,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de,  9,    632,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de, 10,    961,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de, 11,    960,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21d994de, 12,    480,          YM), /* 00107 - David Whittaker - Grimblood.sndh */
TIMEDB_ENTRY( 21f4ea3f,  0,   3066,       YM+TA), /* 02410 - Miq - Exited.sndh */
TIMEDB_ENTRY( 21fb3379,  0,   8945,          YM), /* 03509 - Unknown - Indiana Jones & Temple of Doom.sndh */
TIMEDB_ENTRY( 2205b629,  0,   5688,          YM), /* 01406 - Skyline - Snork Intro - Main Theme.sndh */
TIMEDB_ENTRY( 22269ec8,  0,   6656,       YM+TA), /* 00266 - Antichrist - The Mystery.sndh */
TIMEDB_ENTRY( 2247e8b8,  0,  84033,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  1,  71458,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  2,    320,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  3,    325,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  4,    280,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  5,    355,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  6,   1457,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  7, 211995,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2247e8b8,  8,  31305,       YM+TA), /* 03814 - David Whittaker - Loopz.sndh */
TIMEDB_ENTRY( 2250063a,  0,   5435, YM+TA+TB+TD), /* 03224 - 505-checkpoint - Warmup (Black Sheep PHF).sndh */
TIMEDB_ENTRY( 2251acfe,  0,  13818, YM+TA+TB+TD), /* 01005 - DMA-SC - Odd Stuff end music.sndh */
TIMEDB_ENTRY( 225519c5,  0,   5760,    YM+TA+TD), /* 02327 - Tao - Cream - MV Jeu - Title (Steps).sndh */
TIMEDB_ENTRY( 22564cc6,  0,   5374,          YM), /* 03199 - Martin Iveson - Arcade Classix - Centipede.sndh */
TIMEDB_ENTRY( 2271789a,  1,   6144,          YM), /* 03273 - Haiko Ruttman - Twinworld.sndh */
TIMEDB_ENTRY( 2271789a,  2,   3840,          YM), /* 03273 - Haiko Ruttman - Twinworld.sndh */
TIMEDB_ENTRY( 2271789a,  3,   3072,          YM), /* 03273 - Haiko Ruttman - Twinworld.sndh */
TIMEDB_ENTRY( 2271789a,  4,   3507,          YM), /* 03273 - Haiko Ruttman - Twinworld.sndh */
TIMEDB_ENTRY( 227a9880,  0,  14202,       YM+TA), /* 01707 - Jess - Run for your Life.sndh */
TIMEDB_ENTRY( 2289e38c,  0, 177228,       YM+TA), /* 04805 - Unknown - House.sndh */
TIMEDB_ENTRY( 22a66a81,  0,   5233,          YM), /* 03174 - n-a - Studio Music by Davidow.sndh */
TIMEDB_ENTRY( 22bd0555,  0,   2194,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  1,   3680,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  2,   6582,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  3,   4608,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  4,   1600,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  5,   2560,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  6,   2590,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  7,    337,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  8,    150,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555,  9,     39,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 10,     36,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 11,     52,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 12,     41,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 13,      6,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 14,    153,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 15,     12,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 16,      4,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 17,     62,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 18,     27,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 19,     78,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22bd0555, 20,  12045,          YM), /* 01922 - The Mind - Dizzy Panic.sndh */
TIMEDB_ENTRY( 22d51000,  0,     36,          YM), /* 04539 - Darren Ithell - Fun school 3 - over 7 years - Word search.sndh */
TIMEDB_ENTRY( 22d51000,  1,     33,          YM), /* 04539 - Darren Ithell - Fun school 3 - over 7 years - Word search.sndh */
TIMEDB_ENTRY( 22d51000,  2,     51,          YM), /* 04539 - Darren Ithell - Fun school 3 - over 7 years - Word search.sndh */
TIMEDB_ENTRY( 22d51000,  3,    134,          YM), /* 04539 - Darren Ithell - Fun school 3 - over 7 years - Word search.sndh */
TIMEDB_ENTRY( 22e34b5d,  0,   2879,    YM+TA+TB), /* 02701 - dubmood - Pampinterlude.sndh */
TIMEDB_ENTRY( 22f07182,  0,  20112,          YM), /* 01923 - The Mind - Freddy's Revenge ST News 7.1.sndh */
TIMEDB_ENTRY( 22f2739a,  0,   3337,    YM+TA+TD), /* 00194 - 505 - Sure Trip 3.sndh */
TIMEDB_ENTRY( 22ff41b3,  0,   5681,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  1,   6144,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  2,     96,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  3,     63,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  4,     96,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  5,   6144,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  6,    283,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  7,    276,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 22ff41b3,  8,    102,          YM), /* 00133 - David Whittaker - Skateball.sndh */
TIMEDB_ENTRY( 230fae5c,  0, 101399,       YM+TA), /* 03719 - Spaz - Ooh Crikey - Chainsaw Massacre.sndh */
TIMEDB_ENTRY( 23160599,  0,   5745,    YM+TA+TD), /* 04325 - 505-checkpoint - 15kb one.sndh */
TIMEDB_ENTRY( 231b829d,  0,   7168,          YM), /* 03694 - Rhino - Sundown 2012.sndh */
TIMEDB_ENTRY( 23227fc4,  0,   3064,       YM+TA), /* 01354 - Scavenger - Only the Lonely.sndh */
TIMEDB_ENTRY( 23391902,  0,   6920,       YM+TA), /* 04012 - Mad Max - Jinks.sndh */
TIMEDB_ENTRY( 23498535,  0,   5839,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  1,   3459,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  2,   2307,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  3,    148,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  4,    118,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  5,     87,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  6,    456,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  7,    388,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  8,    164,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535,  9,    329,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 10,    181,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 11,    248,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 12,    390,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 13,    283,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 14,    217,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 15,    332,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 23498535, 16,   1767,          YM), /* 01799 - Barry Leitch - Microprose Soccer.sndh */
TIMEDB_ENTRY( 235a66f4,  0,  19134,          YM), /* 04081 - Gary Wheaton - Medley.sndh */
TIMEDB_ENTRY( 235a81ec,  0,   2720,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 235a81ec,  1,   2048,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 235a81ec,  2,   1536,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 235a81ec,  3,    640,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 235a81ec,  4,    960,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 235a81ec,  5,   1280,          YM), /* 01890 - Dave Lowe - Railroad Tycoon.sndh */
TIMEDB_ENTRY( 236072d7,  0,   6144,    YM+TA+TB), /* 00043 - 505 - Checkpoint - MMJ Xmas 03 - 1.sndh */
TIMEDB_ENTRY( 23642d41,  0,   1505,          YM), /* 04139 - Thomas Huth - Fantastic World 2.sndh */
TIMEDB_ENTRY( 23642d41,  1,    737,          YM), /* 04139 - Thomas Huth - Fantastic World 2.sndh */
TIMEDB_ENTRY( 23642d41,  2,   1288,          YM), /* 04139 - Thomas Huth - Fantastic World 2.sndh */
TIMEDB_ENTRY( 236c66f9,  0,  25761,       YM+TA), /* 04756 - Gilles Soulet - Sapiens.sndh */
TIMEDB_ENTRY( 236c66f9,  1,  17281,       YM+TA), /* 04756 - Gilles Soulet - Sapiens.sndh */
TIMEDB_ENTRY( 236de9cc,  0,  11904,          YM), /* 03662 - Jedi - So Lame.sndh */
TIMEDB_ENTRY( 23ab1ad4,  0,   2559, YM+TA+TB+TD), /* 02798 - Starbuck^Creative Mind - Every beginning is hard.sndh */
TIMEDB_ENTRY( 23ce900b,  0,   1040,          YM), /* 02416 - Nexus 6 - X-X-X-X-X.sndh */
TIMEDB_ENTRY( 23dca986,  0, 208901,       YM+TA), /* 03742 - Tufty - Don't Break My Heart (UB40).sndh */
TIMEDB_ENTRY( 23ec9f6a,  0,   6144,          YM), /* 03401 - Wally Beben - Time Machine.sndh */
TIMEDB_ENTRY( 23ec9f6a,  1,  10240,          YM), /* 03401 - Wally Beben - Time Machine.sndh */
TIMEDB_ENTRY( 23f0d816,  0,   2304,       YM+TA), /* 04294 - 505 - noix000f.sndh */
TIMEDB_ENTRY( 23f2706b,  0,  39767,       YM+TA), /* 03945 - Unknown - LT Fuge 2.sndh */
TIMEDB_ENTRY( 23f2a11c,  0,   5178,          YM), /* 01236 - Epic - Sauvage.sndh */
TIMEDB_ENTRY( 23fd2a3d,  0,   4986, YM+TA+TB+TD), /* 02453 - Timbral - DHS - Sanna suck my nectar.sndh */
TIMEDB_ENTRY( 240fa524,  0,   6144,          YM), /* 01134 - Tyan - Revolution - In Game.sndh */
TIMEDB_ENTRY( 241ff404,  0,   5188,    YM+TA+TD), /* 02322 - Tao - Cream (orig. Klaus Ude) - Magic Marble ingame 1.sndh */
TIMEDB_ENTRY( 24231b4c,  0,   7676,          YM), /* 00551 - Allister Brimble - The Final Day.sndh */
TIMEDB_ENTRY( 24231b4c,  1,   7996,          YM), /* 00551 - Allister Brimble - The Final Day.sndh */
TIMEDB_ENTRY( 24231b4c,  2,    763,          YM), /* 00551 - Allister Brimble - The Final Day.sndh */
TIMEDB_ENTRY( 2432f6ea,  0,  44351,    YM+TA+TB), /* 01355 - Scavenger - Pajaro - Synergy Megademo.sndh */
TIMEDB_ENTRY( 243525cc,  0,  16000,          YM), /* 00614 - Chrispy Noodle - Jet Set Willy.sndh */
TIMEDB_ENTRY( 24393c14,  0,  19200,          YM), /* 01652 - Jedi - NMM.sndh */
TIMEDB_ENTRY( 24447c00,  0,  10362,       YM+TD), /* 04786 - Intruder - Intruder 2.sndh */
TIMEDB_ENTRY( 2451ae16,  0,  13061,          YM), /* 02278 - Tao - Cream - Ice Age (Just Buggin).sndh */
TIMEDB_ENTRY( 245a43c9,  0,  10800, YM+TA+TB+TD), /* 02081 - Mad Max - Hunyer Patrol - Sid.sndh */
TIMEDB_ENTRY( 246f817c,  0,   2527,          YM), /* 03178 - Unknown - Two Part Invention #13 by Bach.sndh */
TIMEDB_ENTRY( 2471027f,  0,   1070,          YM), /* 04720 - Peter Hickman - Treasure Search.sndh */
TIMEDB_ENTRY( 24c48017,  0,  21755,          YM), /* 02240 - Tao - Cream (orig. Rob Hubbard) - Lightforce (Stmygm).sndh */
TIMEDB_ENTRY( 24d7f769,  0,   2957,          YM), /* 03160 - Unknown - Santa Claus is comin' to town!.sndh */
TIMEDB_ENTRY( 24dce34b,  0,   2690,          YM), /* 00436 - Big Alec - What Time is It? (Scrolly).sndh */
TIMEDB_ENTRY( 24de3165,  0,   1016,       YM+TA), /* 02619 - STU - SPOOKY.sndh */
TIMEDB_ENTRY( 24eaf9b6,  0,   2324,       YM+TA), /* 01332 - Scavenger - Gangstar.sndh */
TIMEDB_ENTRY( 24eb3117,  0,   9367, YM+TA+TB+TD), /* 03584 - Divag - Technorobotic.sndh */
TIMEDB_ENTRY( 24eb9818,  0,   1543,          YM), /* 03919 - Unknown - Holland.sndh */
TIMEDB_ENTRY( 24f4ec97,  0,    752,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  1,    872,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  2,    864,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  3,    940,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  4,    512,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  5,    366,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  6,     64,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  7,    149,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 24f4ec97,  8,   1600,          YM), /* 02960 - Unknown - Countout.sndh */
TIMEDB_ENTRY( 25191609,  0,  23456,          YM), /* 01510 - Rob Hubbard - Warhawk.sndh */
TIMEDB_ENTRY( 2521bedd,  0,   3580,       YM+TA), /* 04296 - 505 - Painsid.sndh */
TIMEDB_ENTRY( 25304018,  0,  37120,          YM), /* 02689 - Philip Linde - breakbeat.sndh */
TIMEDB_ENTRY( 2531e907,  0,   6652, YM+TA+TB+TD), /* 00280 - Baggio & Aio - Spritze.sndh */
TIMEDB_ENTRY( 2544e432,  0,   9096, YM+TA+TB+TD), /* 02973 - 505-paradox - shredder (STE only).sndh */
TIMEDB_ENTRY( 255ae2f1,  0,   1792,          YM), /* 01402 - Skyline - Snork Intro - Cuddly.sndh */
TIMEDB_ENTRY( 257ad548,  0,  10993,          YM), /* 03148 - Unknown - He Played Real Good For Free.sndh */
TIMEDB_ENTRY( 258b6b9f,  0,   7678,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  1,   4699,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  2,   5566,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  3,   3886,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  4,   2880,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  5,   2302,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  6,    304,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  7,    319,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258b6b9f,  8,   2864,          YM), /* 01323 - Matt Furniss - Torvak The Warrior.sndh */
TIMEDB_ENTRY( 258c9925,  0,   1536,          YM), /* 03251 - T1000 - The Bits Club 27.sndh */
TIMEDB_ENTRY( 25b4b8f3,  0,    620,          YM), /* 03971 - Geoff Camp - Maths X.sndh */
TIMEDB_ENTRY( 25b4b8f3,  1,    553,          YM), /* 03971 - Geoff Camp - Maths X.sndh */
TIMEDB_ENTRY( 25b4b8f3,  2,    207,          YM), /* 03971 - Geoff Camp - Maths X.sndh */
TIMEDB_ENTRY( 25b4b8f3,  3,   2708,          YM), /* 03971 - Geoff Camp - Maths X.sndh */
TIMEDB_ENTRY( 25b4b8f3,  4,   1560,          YM), /* 03971 - Geoff Camp - Maths X.sndh */
TIMEDB_ENTRY( 25b95675,  0,   5376,          YM), /* 00806 - Crawdaddy - Plazma.sndh */
TIMEDB_ENTRY( 25d9e06e,  0,    144,          YM), /* 00164 - Martin Walker - Devious Designs.sndh */
TIMEDB_ENTRY( 25d9e06e,  1,    385,          YM), /* 00164 - Martin Walker - Devious Designs.sndh */
TIMEDB_ENTRY( 25d9e06e,  2,   1312,          YM), /* 00164 - Martin Walker - Devious Designs.sndh */
TIMEDB_ENTRY( 25d9e06e,  3,    127,          YM), /* 00164 - Martin Walker - Devious Designs.sndh */
TIMEDB_ENTRY( 25d9e06e,  4,    159,          YM), /* 00164 - Martin Walker - Devious Designs.sndh */
TIMEDB_ENTRY( 25df222e,  0,   4662,       YM+TD), /* 03422 - Dubmood - Tempest - Acidjazzed Evenings (D-Bug).sndh */
TIMEDB_ENTRY( 25f635b6,  0,   1536,          YM), /* 02896 - Remalon - Unlimited Cho.sndh */
TIMEDB_ENTRY( 25ff6850,  0,   3648,       YM+TA), /* 00252 - Antichrist - Concours CSC2.sndh */
TIMEDB_ENTRY( 26020c80,  0,   4864,       YM+TA), /* 00796 - Crawdaddy - The Last Cow.sndh */
TIMEDB_ENTRY( 2611cb55,  0,   3456,       YM+TA), /* 00818 - Crawdaddy - in the Woods.sndh */
TIMEDB_ENTRY( 265194e3,  0,   8450,          YM), /* 01051 - Dr Beat - NTM - Zuul.sndh */
TIMEDB_ENTRY( 2678e222,  0,  14446,    YM+TA+TD), /* 02788 - Crazy Q - Phunky Shapeshifter 2149.sndh */
TIMEDB_ENTRY( 267c058d,  0,   6720,          YM), /* 00466 - Augustin Vidovic - What Name.sndh */
TIMEDB_ENTRY( 26ad342c,  0,   8448,          YM), /* 02031 - Mad Max - Formula 1.sndh */
TIMEDB_ENTRY( 26b9b1ed,  0,  41856,          YM), /* 01434 - Stax - DM.sndh */
TIMEDB_ENTRY( 26bdb013,  0,   8003,          YM), /* 01802 - Barry Leitch - Raiden.sndh */
TIMEDB_ENTRY( 26c406f4,  0,   1896,          YM), /* 04733 - Al Lowe - Donald Ducks Playground.sndh */
TIMEDB_ENTRY( 26c406f4,  1,   1143,          YM), /* 04733 - Al Lowe - Donald Ducks Playground.sndh */
TIMEDB_ENTRY( 26c406f4,  2,    699,          YM), /* 04733 - Al Lowe - Donald Ducks Playground.sndh */
TIMEDB_ENTRY( 26c406f4,  3,    255,          YM), /* 04733 - Al Lowe - Donald Ducks Playground.sndh */
TIMEDB_ENTRY( 270ddc54,  0,   7576, YM+TA+TB+TD), /* 00086 - 505 - Moesrave (Birthtro).sndh */
TIMEDB_ENTRY( 27199b05,  0,   8449,          YM), /* 01037 - Donald Fakk - Hooker.sndh */
TIMEDB_ENTRY( 271d689b,  0,  10027,          YM), /* 01538 - Ralph - Ralph 2.sndh */
TIMEDB_ENTRY( 272758d6,  0,   2400,          YM), /* 01880 - Dave Lowe - Darius.sndh */
TIMEDB_ENTRY( 272758d6,  1,   1152,          YM), /* 01880 - Dave Lowe - Darius.sndh */
TIMEDB_ENTRY( 272758d6,  2,   1536,          YM), /* 01880 - Dave Lowe - Darius.sndh */
TIMEDB_ENTRY( 2730c06f,  0,   5958,          YM), /* 04340 - Nexus 6 - Discover.sndh */
TIMEDB_ENTRY( 2732cc8e,  0,  11536, YM+TA+TB+TD), /* 02487 - gwEm - Coding on E.sndh */
TIMEDB_ENTRY( 273e1934,  0,  16352,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 273e1934,  1,  10240,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 273e1934,  2,  16384,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 273e1934,  3,  10752,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 273e1934,  4,  13824,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 273e1934,  5,  16384,          YM), /* 04452 - Wally Beben - Winter Olympiad 88.sndh */
TIMEDB_ENTRY( 2740541d,  0,   7680,       YM+TA), /* 03902 - Mad Max - Arkanoid (Digidrum Version).sndh */
TIMEDB_ENTRY( 275088e0,  0,  12672, YM+TA+TB+TD), /* 02104 - Mad Max - Wiz - Sid.sndh */
TIMEDB_ENTRY( 275088e0,  1,    321, YM+TA+TB+TD), /* 02104 - Mad Max - Wiz - Sid.sndh */
TIMEDB_ENTRY( 27511953,  0,   5380,       YM+TA), /* 01339 - Scavenger - Knuckle Joe.sndh */
TIMEDB_ENTRY( 275f64e1,  0,   8799,          YM), /* 01553 - Rhino - Xmas 2003 Intro - Song 1.sndh */
TIMEDB_ENTRY( 2769474f,  0,   8844,       YM+TA), /* 01755 - Eric Ksouri - Manciola - Melia.sndh */
TIMEDB_ENTRY( 2774ddce,  0,   1760,          YM), /* 01739 - Richard Joseph - Sensible Soccer.sndh */
TIMEDB_ENTRY( 277aa649,  0,  13887,          YM), /* 02385 - Nexus 6 - Nasty Song.sndh */
TIMEDB_ENTRY( 277aa649,  1,   9295,          YM), /* 02385 - Nexus 6 - Nasty Song.sndh */
TIMEDB_ENTRY( 277bad2e,  0,   3264,          YM), /* 02132 - Mad Max - Crazy Comets Game Over - LCD.sndh */
TIMEDB_ENTRY( 27893784,  0,  28160,       YM+TA), /* 03490 - Mad Max - Street Gang.sndh */
TIMEDB_ENTRY( 279a4e42,  0,  15360,       YM+TA), /* 04126 - Tao - The Cave - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 27a2881c,  0,   8450,       YM+TA), /* 04420 - EVE - Axel F.sndh */
TIMEDB_ENTRY( 27a3465f,  0,   2431,    YM+TA+TD), /* 04650 - Per Almered-XiA-Checkpoint - Prof E Gadd (NGC cover).sndh */
TIMEDB_ENTRY( 27a576b2,  0,  12284, YM+TA+TB+TD), /* 01006 - DMA-SC - Odd Stuff Main.sndh */
TIMEDB_ENTRY( 27b65471,  0,   9982,          YM), /* 01269 - Frazer - Reanimators Demo.sndh */
TIMEDB_ENTRY( 27d180ac,  0,   6240,          YM), /* 03478 - gwEm - 4gEtful.sndh */
TIMEDB_ENTRY( 27e00753,  0,  23289,       YM+TA), /* 01711 - Jess - SOS means 'Save Our ST'.sndh */
TIMEDB_ENTRY( 27e894ec,  0,   8960,          YM), /* 03193 - n-a - Stop Start.sndh */
TIMEDB_ENTRY( 27e9c494,  0,   4608,          YM), /* 00265 - Antichrist - Take Off.sndh */
TIMEDB_ENTRY( 27e9c53f,  0,   3584,          YM), /* 02935 - Chris Denman - Seige Master.sndh */
TIMEDB_ENTRY( 27fb194e,  0,  11712,          YM), /* 01433 - Stax - Aciidvsh.sndh */
TIMEDB_ENTRY( 2805c4a0,  0,  48386,    YM+TB+TD), /* 03333 - Marcer - ST KILLS RAMBO.sndh */
TIMEDB_ENTRY( 28125c34,  0,   7390,          YM), /* 03102 - n-a - Children's Corner # 4.sndh */
TIMEDB_ENTRY( 281bd3c9,  0,  62902,       YM+TA), /* 04557 - Simon O - Predator.sndh */
TIMEDB_ENTRY( 283bce5d,  0,  11648,       YM+TA), /* 04122 - Tao - Rise - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 284073f5,  0,  42496,       YM+TD), /* 04851 - xFalcon - DEEP THOUGHT.sndh */
TIMEDB_ENTRY( 28484b7f,  0,   9629, YM+TA+TB+TD), /* 02864 - Marcer - 2149 Final.sndh */
TIMEDB_ENTRY( 284af8bb,  0,   4207,          YM), /* 04083 - Gary Wheaton - Moondance.sndh */
TIMEDB_ENTRY( 285f472d,  0,   7355, YM+TA+TB+TD), /* 04248 - 505-checkpoint - Brotchen.sndh */
TIMEDB_ENTRY( 2869cc2e,  0,   3385,          YM), /* 04547 - Geoff Camp - Home Worker.sndh */
TIMEDB_ENTRY( 2869cc2e,  1,   3644,          YM), /* 04547 - Geoff Camp - Home Worker.sndh */
TIMEDB_ENTRY( 28736104,  0,   2708, YM+TA+TB+TD), /* 00233 - AIO - EIL 3 Invite.sndh */
TIMEDB_ENTRY( 287a3e7b,  0,   4928,          YM), /* 02289 - Tao - Cream - 128tros (DHS 128byte megademo).sndh */
TIMEDB_ENTRY( 288347ca,  0,   7168,          YM), /* 00122 - David Whittaker - Pandora.sndh */
TIMEDB_ENTRY( 288347ca,  1,   3072,          YM), /* 00122 - David Whittaker - Pandora.sndh */
TIMEDB_ENTRY( 288347ca,  2,   1792,          YM), /* 00122 - David Whittaker - Pandora.sndh */
TIMEDB_ENTRY( 288d8df0,  0,   6912,       YM+TA), /* 00472 - Viking - Viking - Sugar Fet 2 BBS.sndh */
TIMEDB_ENTRY( 289c2b6d,  0,   9215,       YM+TA), /* 03576 - 505-paradox - Finja.sndh */
TIMEDB_ENTRY( 28aa3438,  0, 143492,       YM+TA), /* 04022 - Roggie - Sonixx.sndh */
TIMEDB_ENTRY( 28b5acbc,  0,  10003, YM+TA+TB+TD), /* 01836 - Lotek Style - Beams are gonna blind me.sndh */
TIMEDB_ENTRY( 28d6c601,  0,   9407,    YM+TA+TD), /* 03445 - 505-checkpoint^paradox - visual unity (2009).sndh */
TIMEDB_ENTRY( 28e0642f,  0,   6906,          YM), /* 02881 - Beast - Suuuuuvk.sndh */
TIMEDB_ENTRY( 28e3c7ab,  0,   1633,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  1,    306,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  2,    564,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  3,   1969,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  4,   1335,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  5,    309,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  6,    753,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  7,    475,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  8,    250,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab,  9,    528,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28e3c7ab, 10,    968,          YM), /* 04734 - Al Lowe - Gold Rush.sndh */
TIMEDB_ENTRY( 28f3887b,  0,  16175,       YM+TA), /* 04793 - Mad Max - Bittner Rap.sndh */
TIMEDB_ENTRY( 28f5a73b,  0,  12415,    YM+TA+TD), /* 03455 - ULTRASYD - MERGING PRESENT AND PAST.sndh */
TIMEDB_ENTRY( 29009716,  0,  16319, YM+TA+TB+TD), /* 04656 - 505 - STNICCC 2015.sndh */
TIMEDB_ENTRY( 2931e3e3,  0,   9208,    YM+TA+TD), /* 01595 - Nemo - Harmless.sndh */
TIMEDB_ENTRY( 2934ae58,  0,  30788,          YM), /* 01800 - Barry Leitch - Pegasus.sndh */
TIMEDB_ENTRY( 2934ae58,  1,  24443,          YM), /* 01800 - Barry Leitch - Pegasus.sndh */
TIMEDB_ENTRY( 29544553,  0,  20928,          YM), /* 04473 - Charles Callet - Turlogh Le Rodeur.sndh */
TIMEDB_ENTRY( 296cb067,  0,   2301,       YM+TA), /* 00199 - 505 - Test.sndh */
TIMEDB_ENTRY( 2971aaca,  0,  13871,    YM+TA+TD), /* 02273 - Tao - Cream - The Powerman (Steps).sndh */
TIMEDB_ENTRY( 29740494,  0,   5116,    YM+TB+TD), /* 01014 - DMA-SC - Samtamagochi - Two In One.sndh */
TIMEDB_ENTRY( 29773ed0,  0,  15620,          YM), /* 00504 - Tangens - Lettrix.sndh */
TIMEDB_ENTRY( 29880d1b,  0,   3450,       YM+TA), /* 04449 - Scavenger - Sambase.sndh */
TIMEDB_ENTRY( 299daefc,  0,   1947,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  1,   1875,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  2,   1002,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  3,   1425,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  4,    579,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  5,    171,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  6,   1920,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  7,    369,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  8,    735,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc,  9,    486,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc, 10,    543,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc, 11,    591,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc, 12,    546,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 299daefc, 13,    651,          YM), /* 04737 - Al Lowe - Kings Quest 2 - Romancing the Throne.sndh */
TIMEDB_ENTRY( 29a3c8c7,  0,   1528,          YM), /* 00445 - Augustin Vidovic - Extensor.sndh */
TIMEDB_ENTRY( 29aa3524,  0,   5139,       YM+TA), /* 01039 - Donald Fakk - ICS Scum Of The Earth Intro.sndh */
TIMEDB_ENTRY( 29aae4da,  3,   2518,          YM), /* 03531 - Jean Baudlot - Operation Wolf.sndh */
TIMEDB_ENTRY( 29ae2ac9,  0,  10362,       YM+TA), /* 01280 - Frazer - TSC 82.sndh */
TIMEDB_ENTRY( 29c4c064,  0,  17469,          YM), /* 04871 - Jovis - Trash Buzz.sndh */
TIMEDB_ENTRY( 29d60b02,  0,   8639,          YM), /* 04390 - Dma Sc - It's A Girl 2.sndh */
TIMEDB_ENTRY( 29e93284,  0,    555,          YM), /* 02150 - Mad Max - Monty On The Run High - LSD.sndh */
TIMEDB_ENTRY( 2a00b4a1,  0,  12780,       YM+TA), /* 03898 - Skidz - Stone Age Menu.sndh */
TIMEDB_ENTRY( 2a04bc63,  0,   1505,          YM), /* 04512 - Jean-Claude Schopfer - Aventura I The Witch Curse.sndh */
TIMEDB_ENTRY( 2a13275d,  0,   9215,          YM), /* 00310 - Gary Wheaton - Honey Pie.sndh */
TIMEDB_ENTRY( 2a21698c,  0,  31872, YM+TA+TB+TD), /* 03593 - datavase - n-a.sndh */
TIMEDB_ENTRY( 2a2ec60d,  0,  12089,       YM+TA), /* 04369 - Jess - Zara (Bagpipes Buzz mix by Ben+Jess).sndh */
TIMEDB_ENTRY( 2a353411,  0,   7103, YM+TA+TB+TD), /* 03268 - Dubmood - Minipitt vs wotw intro 96(goto80 v wotw).sndh */
TIMEDB_ENTRY( 2a3f66f2,  0,   8505,          YM), /* 00848 - Ben Daglish - Deflektor.sndh */
TIMEDB_ENTRY( 2a405f9e,  0,   2304,          YM), /* 04003 - Blue Nun - Misspell.sndh */
TIMEDB_ENTRY( 2a4be55f,  0,   3072,          YM), /* 04594 - T1000 - Bits 59.sndh */
TIMEDB_ENTRY( 2a6eee16,  0,   6908, YM+TA+TB+TD), /* 01154 - Tyan - Platz 3.sndh */
TIMEDB_ENTRY( 2a6ffb4f,  0,  42704, YM+TA+TB+TD), /* 02789 - Marcer - The 5th Experiment.sndh */
TIMEDB_ENTRY( 2a7f3095,  0,  10745, YM+TA+TB+TD), /* 01951 - MSG - Chu Chu Rocket - Team Battle.sndh */
TIMEDB_ENTRY( 2a914b69,  0,   7681,          YM), /* 00387 - Wally Beben - I-Ball 2.sndh */
TIMEDB_ENTRY( 2ab97245,  0,   1024,          YM), /* 04692 - Arnauld Chevallier - Magenta.sndh */
TIMEDB_ENTRY( 2ab97245,  1,   1024,          YM), /* 04692 - Arnauld Chevallier - Magenta.sndh */
TIMEDB_ENTRY( 2ab97245,  8,   2816,          YM), /* 04692 - Arnauld Chevallier - Magenta.sndh */
TIMEDB_ENTRY( 2ab97245,  9,    223,          YM), /* 04692 - Arnauld Chevallier - Magenta.sndh */
TIMEDB_ENTRY( 2ab97245, 10,   2304,          YM), /* 04692 - Arnauld Chevallier - Magenta.sndh */
TIMEDB_ENTRY( 2ac1d816,  0,  34081,    YM+TB+TD), /* 03984 - XyNo - CYBORGS N' CLUBS.sndh */
TIMEDB_ENTRY( 2ac56364,  0,  11520,          YM), /* 00600 - Chris Mad - Overdrive Mainmenu.sndh */
TIMEDB_ENTRY( 2ad61ed6,  0,  17536,          YM), /* 02045 - Mad Max - Monty On The Run.sndh */
TIMEDB_ENTRY( 2ad61ed6,  1,   8192,          YM), /* 02045 - Mad Max - Monty On The Run.sndh */
TIMEDB_ENTRY( 2ad61ed6,  2,    901,          YM), /* 02045 - Mad Max - Monty On The Run.sndh */
TIMEDB_ENTRY( 2aeb113f,  0,  16130,          YM), /* 01896 - LTK - Env Try.sndh */
TIMEDB_ENTRY( 2af6c119,  0, 117537,       YM+TA), /* 04553 - Spaz - Killing Spree.sndh */
TIMEDB_ENTRY( 2b046516,  0,  12545,    YM+TA+TD), /* 02297 - Tao - Cream - Counted Out (Alive 5).sndh */
TIMEDB_ENTRY( 2b059920,  0,  15403,          YM), /* 01541 - Ralph - Ralph 4.sndh */
TIMEDB_ENTRY( 2b05da62,  0,   4352,          YM), /* 01535 - Iso - Stawpaws.sndh */
TIMEDB_ENTRY( 2b216eb9,  0,  12368,          YM), /* 01417 - Skyline - Snork Intro - Unused music!.sndh */
TIMEDB_ENTRY( 2b257e44,  0,  35136, YM+TA+TB+TD), /* 04331 - mOdmate-CheckPoint (orig. Adam Gilmore) - DRACONUST.sndh */
TIMEDB_ENTRY( 2b29df62,  0,  11520,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b29df62,  1,  15072,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b29df62,  2,  12960,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b29df62,  3,  12960,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b29df62,  4,  13061,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b29df62,  5,   9984,       YM+TA), /* 02206 - Mad Max - Ninja Remix.sndh */
TIMEDB_ENTRY( 2b2b878b,  0,   6272,          YM), /* 03635 - Rhino - Torment - Shorty.sndh */
TIMEDB_ENTRY( 2b3a27be,  0,   3199, YM+TA+TB+TD), /* 02833 - maktone-fairlight - 1st.sndh */
TIMEDB_ENTRY( 2b3b4b3f,  0,   3075,          YM), /* 03063 - Ed Bogas - Waltz.sndh */
TIMEDB_ENTRY( 2b4dfe23,  0,  18432,          YM), /* 02577 - STU - BLIP TERROR.sndh */
TIMEDB_ENTRY( 2b59cb79,  0,  52204,       YM+TA), /* 03797 - Tony Williams - Stampede 1.sndh */
TIMEDB_ENTRY( 2b5a715f,  0,   8826,    YM+TA+TD), /* 01848 - Lotek Style - Flainian.sndh */
TIMEDB_ENTRY( 2b642c52,  0,  68111,       YM+TA), /* 03813 - David Whittaker - Dan Dare III.sndh */
TIMEDB_ENTRY( 2b642c52,  1,  29749,       YM+TA), /* 03813 - David Whittaker - Dan Dare III.sndh */
TIMEDB_ENTRY( 2b96bb6f,  0,    457,         STE), /* 03453 - Per Almered (XiA) - $14.sndh */
TIMEDB_ENTRY( 2b97f341,  0,   8505,          YM), /* 04570 - Yerzmyey - SNDH v4.4 Update.sndh */
TIMEDB_ENTRY( 2b98362c,  0,  12288,       YM+TA), /* 04125 - Tao - Sunriders - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 2b9c2f51,  0,   2885,       YM+TA), /* 04238 - 505 - 22nd.sndh */
TIMEDB_ENTRY( 2b9dbfd5,  0,   3073,          YM), /* 02220 - Mad Max - Tom & Jerry.sndh */
TIMEDB_ENTRY( 2b9dbfd5,  1,   4994,          YM), /* 02220 - Mad Max - Tom & Jerry.sndh */
TIMEDB_ENTRY( 2b9dbfd5,  2,    962,          YM), /* 02220 - Mad Max - Tom & Jerry.sndh */
TIMEDB_ENTRY( 2ba03a26,  0,   6916,    YM+TA+TD), /* 02210 - Tao - Cream (orig. Jochen Hippel) - Madness (Thalion intro).sndh */
TIMEDB_ENTRY( 2ba4f988,  0,   8400,          YM), /* 01061 - Jonathan Dunn - Lost Patrol.sndh */
TIMEDB_ENTRY( 2ba4f988,  1,    142,          YM), /* 01061 - Jonathan Dunn - Lost Patrol.sndh */
TIMEDB_ENTRY( 2bb5e166,  0,   9280,    YM+TA+TD), /* 01391 - SH3-reservoir gods - PEA.sndh */
TIMEDB_ENTRY( 2bb80549,  0,   4497,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  1,   1645,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  2,   1828,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  3,   2048,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  4,   4294,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  5,   3809,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  6,    960,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  7,    752,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  8,   2666,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549,  9,   1672,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 10,   1360,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 11,   1852,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 12,   3840,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 13,    853,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 14,   4693,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 15,   1204,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 16,   1208,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 17,   1462,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 18,   1920,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 19,   1252,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 20,   3072,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 21,   1024,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 22,    914,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 23,   3572,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 24,   2964,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 25,    768,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 26,   3136,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 27,   1260,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 28,   1097,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 29,   3490,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bb80549, 30,   2304,          YM), /* 04711 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - The Corps.sndh */
TIMEDB_ENTRY( 2bca8dcb,  0,   1954,          YM), /* 03318 - MSG - Chariots of Fire.sndh */
TIMEDB_ENTRY( 2bd62c88,  0,   7487,          YM), /* 04302 - 505 - Relix (2013).sndh */
TIMEDB_ENTRY( 2c00be0f,  0,  12985, YM+TA+TB+TD), /* 01947 - MSG - Chu Chu Rocket - Puzzle.sndh */
TIMEDB_ENTRY( 2c07453b,  0,  12287, YM+TA+TB+TD), /* 03856 - dEViLOCk+!Cube - Outpost (Atari ST remix).sndh */
TIMEDB_ENTRY( 2c168cb0,  0,   5120,          YM), /* 02162 - Mad Max - Syntax Terror - Reset.sndh */
TIMEDB_ENTRY( 2c21aa4d,  0,   3071,    YM+TA+TB), /* 02994 - 505-paradox - paracon8.sndh */
TIMEDB_ENTRY( 2c29a738,  0,   3288,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  1,     69,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  2,    455,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  3,     72,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  4,    455,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  5,    322,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  6,   3120,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  7,     48,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  8,   1365,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c29a738,  9,    512,          YM), /* 00087 - David Whittaker - Cloud Kingdoms.sndh */
TIMEDB_ENTRY( 2c5d1243,  0,  10367,       YM+TA), /* 01781 - Lap - Bird Girl Mad Demo Menu Digi.sndh */
TIMEDB_ENTRY( 2c5e55b7,  0,   2235, YM+TA+TB+TD), /* 04310 - 505-checkpoint - ST_P.sndh */
TIMEDB_ENTRY( 2c5f33d7,  0,   4608,          YM), /* 01091 - Techno - Mathematiques Squizophrenes End.sndh */
TIMEDB_ENTRY( 2c61bcb6,  0,   1337, YM+TA+TB+TD), /* 01949 - MSG - Chu Chu Rocket - Slow Down.sndh */
TIMEDB_ENTRY( 2c681bcc,  0,   1536,          YM), /* 02636 - Miggy - In To Action (Unfinished).sndh */
TIMEDB_ENTRY( 2c753baf,  0,  88516,       YM+TA), /* 04558 - Simon O - Goldrunner.sndh */
TIMEDB_ENTRY( 2c7e574a,  0,   1530, YM+TA+TB+TD), /* 01946 - MSG - Chu Chu Rocket - Mousemania.sndh */
TIMEDB_ENTRY( 2c92323e,  0,   8448,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c92323e,  1,   9984,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c92323e,  2,   7520,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c92323e,  3,   7944,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c92323e,  4,   8574,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c92323e,  5,   7495,          YM), /* 01572 - Patrick Phelan - Lotus Turbo Challenge 3 The Ultimate Challenge.sndh */
TIMEDB_ENTRY( 2c969341,  0,   4362,       YM+TA), /* 02674 - Donald Fakk - Ko And Kossa.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  0,  85112,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  1,  45496,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  2,  62176,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  3,  42592,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  4,  99883,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  5,  35666,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  6,  39640,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  7, 137678,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2ca4fa1a,  8,  49985,       YM+TA), /* 03815 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 2cb9700c,  0,    811,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  1,   1156,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  2,    587,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  3,   1223,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  4,   2937,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  5,    258,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  6,     75,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cb9700c,  7,    753,          YM), /* 00871 - Ben Daglish - Rick Dangerous.sndh */
TIMEDB_ENTRY( 2cbfff0f,  0,   1824,          YM), /* 00451 - Augustin Vidovic - Music Mon Intro.sndh */
TIMEDB_ENTRY( 2cd6123b,  0,   6624,          YM), /* 03407 - Mark Cooksey - World Championship Soccer.sndh */
TIMEDB_ENTRY( 2cd6123b,  1,   4704,          YM), /* 03407 - Mark Cooksey - World Championship Soccer.sndh */
TIMEDB_ENTRY( 2cd6123b,  2,   1361,          YM), /* 03407 - Mark Cooksey - World Championship Soccer.sndh */
TIMEDB_ENTRY( 2cd6123b,  3,    192,          YM), /* 03407 - Mark Cooksey - World Championship Soccer.sndh */
TIMEDB_ENTRY( 2cd8b73d,  0,   4608, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  1,   4608, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  2,    171, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  3,    268, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  4,    396, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  5,    396, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  6,    299, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  7,    267, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  8,    394, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d,  9,    524, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d, 10,    428, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cd8b73d, 11,    492, YM+TA+TB+TD), /* 02074 - Mad Max - Delta - Sid.sndh */
TIMEDB_ENTRY( 2cdc85da,  0,   6717, YM+TA+TB+TD), /* 00078 - 505 - Louder1.sndh */
TIMEDB_ENTRY( 2cdcd0b2,  0,  17988, YM+TA+TB+TD), /* 03995 - XyNo - WAITING UPDATES.sndh */
TIMEDB_ENTRY( 2cdf9e25,  0,   1284,          YM), /* 04504 - Geoff Camp - Brain food - Pipeline.sndh */
TIMEDB_ENTRY( 2cdf9e25,  1,     24,          YM), /* 04504 - Geoff Camp - Brain food - Pipeline.sndh */
TIMEDB_ENTRY( 2cdf9e25,  2,   3144,          YM), /* 04504 - Geoff Camp - Brain food - Pipeline.sndh */
TIMEDB_ENTRY( 2cdf9e25,  3,    130,          YM), /* 04504 - Geoff Camp - Brain food - Pipeline.sndh */
TIMEDB_ENTRY( 2cdf9e25,  4,    288,          YM), /* 04504 - Geoff Camp - Brain food - Pipeline.sndh */
TIMEDB_ENTRY( 2ce651f2,  0,  18432,       YM+TD), /* 02574 - STU - ATARI FORCE.sndh */
TIMEDB_ENTRY( 2cec5651,  0,    196,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  1,    221,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  2,    180,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  3,    593,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  4,    526,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  5,    404,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  6,    240,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  7,    331,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  8,    912,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651,  9,     38,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 10,    996,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 12,   1205,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 14,    468,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 15,    347,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 16,    212,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 17,    536,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 18,    647,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 19,   1088,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 20,    848,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 21,   1853,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 22,   1862,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 24,    912,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 25,     68,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 26,    134,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 27,   1227,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 28,   1044,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 29,    887,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cec5651, 30,   1044,          YM), /* 04683 - Unknown - Kids' Academy - Alvin's puzzles.sndh */
TIMEDB_ENTRY( 2cedde16,  0,   6656,          YM), /* 02015 - Mad Max - Supersonic.sndh */
TIMEDB_ENTRY( 2cfd99f3,  0,   7540,          YM), /* 03070 - Chris French - Twelve Days of Christmas.sndh */
TIMEDB_ENTRY( 2d0618b9,  0,   3840,          YM), /* 02113 - Mad Max - Pro BMX Simulator A.sndh */
TIMEDB_ENTRY( 2d08da7c,  0,   1800,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  1,   3554,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  2,   3385,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  3,   3494,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  4,   4057,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  5,   4509,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  6,   2264,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  7,   3480,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  8,   3444,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c,  9,   5049,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 10,   3895,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 11,   3958,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 12,   3972,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 13,   3080,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 14,   3005,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d08da7c, 15,   3252,          YM), /* 02928 - Geoff Camp - Spider Spell.sndh */
TIMEDB_ENTRY( 2d0aad21,  0,   4992,          YM), /* 01133 - Tyan - Revolution - IchBin JaSo Cool.sndh */
TIMEDB_ENTRY( 2d1cb4e0,  0,   4986,       YM+TA), /* 01126 - Tyan - Bits 4.sndh */
TIMEDB_ENTRY( 2d330fa8,  0,   8576,          YM), /* 00811 - Crawdaddy - Streetkids.sndh */
TIMEDB_ENTRY( 2d3c9c67,  0,   5802,          YM), /* 00365 - Steve Barrett - Chronicles of Omega.sndh */
TIMEDB_ENTRY( 2d3c9c67,  1,   4994,          YM), /* 00365 - Steve Barrett - Chronicles of Omega.sndh */
TIMEDB_ENTRY( 2d3c9c67,  2,   1031,          YM), /* 00365 - Steve Barrett - Chronicles of Omega.sndh */
TIMEDB_ENTRY( 2d3c9c67,  3,    387,          YM), /* 00365 - Steve Barrett - Chronicles of Omega.sndh */
TIMEDB_ENTRY( 2d3c9c67,  4,    197,          YM), /* 00365 - Steve Barrett - Chronicles of Omega.sndh */
TIMEDB_ENTRY( 2d4a42cf,  0,   8449,       YM+TA), /* 04813 - Unknown - Outside.sndh */
TIMEDB_ENTRY( 2d5814e3,  0,  16511,          YM), /* 03688 - Dma-Sc - Saboteur 3 - In Game 1.sndh */
TIMEDB_ENTRY( 2d6596a3,  0,  17004,       YM+TA), /* 01545 - Ralph - Ralph Digi 2.sndh */
TIMEDB_ENTRY( 2d7f8cc2,  0,  43453,       YM+TA), /* 03956 - Scott Lemmon - Paranoid.sndh */
TIMEDB_ENTRY( 2d80a54d,  0,   4605,    YM+TB+TD), /* 00183 - 505 - Pilz (D-Bug 174).sndh */
TIMEDB_ENTRY( 2d97a73f,  0,  12800,          YM), /* 03234 - T1000 - The Beast Club - TBC 4.sndh */
TIMEDB_ENTRY( 2dc08e48,  0,  17424,          YM), /* 00318 - Gary Wheaton - So Watt.sndh */
TIMEDB_ENTRY( 2dc0e19e,  0,  23168,    YM+TB+TD), /* 04641 - EIZ-GRG - EIZ@WORK.sndh */
TIMEDB_ENTRY( 2dcb19ef,  0,   7182,    YM+TA+TD), /* 00195 - 505 - Checkpoint^Creators (2003) - Stage Completed (YM Rockerz).sndh */
TIMEDB_ENTRY( 2dcc758f,  0,   8400,          YM), /* 01293 - Matt Furniss - Championship Manager 93.sndh */
TIMEDB_ENTRY( 2de65596,  0,   2035, YM+TA+TB+TD), /* 04263 - 505 - Hack.sndh */
TIMEDB_ENTRY( 2e0fa0f3,  0,   8064,          YM), /* 01098 - Techno - PP NOEL 2.sndh */
TIMEDB_ENTRY( 2e1efe28,  0,   6522,          YM), /* 02882 - Beast - Xemispoy.sndh */
TIMEDB_ENTRY( 2e28fd1e,  0,   9409,          YM), /* 02018 - Mad Max - Clumsy Colin Action Biker.sndh */
TIMEDB_ENTRY( 2e28fd1e,  1,   3073,          YM), /* 02018 - Mad Max - Clumsy Colin Action Biker.sndh */
TIMEDB_ENTRY( 2e28fd1e,  2,    541,          YM), /* 02018 - Mad Max - Clumsy Colin Action Biker.sndh */
TIMEDB_ENTRY( 2e453b54,  0,   4599,    YM+TA+TD), /* 04295 - 505-checkpoint^paradox - Once.sndh */
TIMEDB_ENTRY( 2e4b8df5,  0,   6713,       YM+TD), /* 04108 - 505-checkpoint (2000) - december.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  0, 215543,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  1, 135248,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  2,  67201,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  3,  34817,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  4, 366823,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e4eb8b7,  5,   5575,       YM+TA), /* 03824 - TEC - TEC Demo.sndh */
TIMEDB_ENTRY( 2e5ac845,  0,   3039,       YM+TA), /* 03842 - Janette R. Armstrong - TimeRunner - Title.sndh */
TIMEDB_ENTRY( 2e837e42,  0,  11519,          YM), /* 04462 - Charles Callet - Hostages.sndh */
TIMEDB_ENTRY( 2e9179de,  0,   5120,       YM+TA), /* 00793 - Crawdaddy - Intro (cross the deserts).sndh */
TIMEDB_ENTRY( 2e9421ad,  0,   8448,          YM), /* 00216 - Markus Weichselbaum - The Apprentice.sndh */
TIMEDB_ENTRY( 2e9421ad,  1,   6400,          YM), /* 00216 - Markus Weichselbaum - The Apprentice.sndh */
TIMEDB_ENTRY( 2e9421ad,  2,   5376,          YM), /* 00216 - Markus Weichselbaum - The Apprentice.sndh */
TIMEDB_ENTRY( 2e9421ad,  3,   6080,          YM), /* 00216 - Markus Weichselbaum - The Apprentice.sndh */
TIMEDB_ENTRY( 2e99a92d,  0,   9856,          YM), /* 00910 - Danny Oneway - Scoopex.sndh */
TIMEDB_ENTRY( 2ea4d2f7,  0, 180559,       YM+TA), /* 04031 - Roggie - Interludes 1-4.sndh */
TIMEDB_ENTRY( 2eac31f3,  0,   3835,       YM+TA), /* 02637 - Miggy - Jeepers(Unfinished).sndh */
TIMEDB_ENTRY( 2eb03281,  0,   6824,          YM), /* 03166 - n-a - Soon It's Gonna Rain.sndh */
TIMEDB_ENTRY( 2ecfed70,  0,   1922,          YM), /* 00412 - Big Alec - What Time Is It - Main Menu.sndh */
TIMEDB_ENTRY( 2eec1b8f,  0,   6155,    YM+TA+TD), /* 02686 - Donald Fakk - Syntax Terror - Loader.sndh */
TIMEDB_ENTRY( 2f0c977c,  0,   1184,          YM), /* 04722 - Robert Groves - Yahtzee.sndh */
TIMEDB_ENTRY( 2f3c94b5,  0,   4096,          YM), /* 01727 - John Michael Phillips - Eliminator.sndh */
TIMEDB_ENTRY( 2f3c94b5,  1,   3072,          YM), /* 01727 - John Michael Phillips - Eliminator.sndh */
TIMEDB_ENTRY( 2f3c94b5,  2,    130,          YM), /* 01727 - John Michael Phillips - Eliminator.sndh */
TIMEDB_ENTRY( 2f3c94b5,  3,    117,          YM), /* 01727 - John Michael Phillips - Eliminator.sndh */
TIMEDB_ENTRY( 2f3c94b5,  4,    276,          YM), /* 01727 - John Michael Phillips - Eliminator.sndh */
TIMEDB_ENTRY( 2f468325,  0,   4610,          YM), /* 00425 - Big Alec - PYM - Intro (Prime Time).sndh */
TIMEDB_ENTRY( 2f49781f,  0,  36867, YM+TA+TB+TD), /* 04754 - xyce & 505 - SPACEDISKO.sndh */
TIMEDB_ENTRY( 2f59bcd4,  0,   6528,          YM), /* 03221 - Tomchi - 24 Tekbah.sndh */
TIMEDB_ENTRY( 2f5a69ae,  0,  16896,          YM), /* 01659 - Jedi - Telephone.sndh */
TIMEDB_ENTRY( 2f5ea4ee,  0,   6138,    YM+TA+TD), /* 04113 - 505-checkpoint - Sidarta.sndh */
TIMEDB_ENTRY( 2f8c3a8c,  0,   8064,          YM), /* 01097 - Techno - PP NOEL.sndh */
TIMEDB_ENTRY( 2f909756,  0,   5184,          YM), /* 00895 - Danny Oneway - Lamebrain 2 - Bubu.sndh */
TIMEDB_ENTRY( 2f949b8a,  0,   2698,       YM+TA), /* 02679 - Donald Fakk - Praeludium (unfinished).sndh */
TIMEDB_ENTRY( 2fa66e99,  0,   3649,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 2fa66e99,  1,    418,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 2fa66e99,  2,    193,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 2fa66e99,  3,    193,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 2fa66e99,  4,   1750,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 2fa66e99,  5,     97,          YM), /* 02049 - Mad Max - Sam Fox Strip Poker.sndh */
TIMEDB_ENTRY( 30076276,  0,   3195,          YM), /* 02602 - STU - LITTLE NINJA.sndh */
TIMEDB_ENTRY( 30151375,  0,   6505,    YM+TA+TD), /* 00204 - 505 - UCM 11.sndh */
TIMEDB_ENTRY( 3017e289,  0,   2486,          YM), /* 03085 - Harry Wootan - Fairest Lord Jesus.sndh */
TIMEDB_ENTRY( 3018c182,  0,   3456,          YM), /* 02171 - Mad Max - Pro BMX Simulator B.sndh */
TIMEDB_ENTRY( 30300443,  0,  14720,          YM), /* 01395 - Paul Shields - Jug.sndh */
TIMEDB_ENTRY( 30300443,  1,   8208,          YM), /* 01395 - Paul Shields - Jug.sndh */
TIMEDB_ENTRY( 30300443,  2,  18176,          YM), /* 01395 - Paul Shields - Jug.sndh */
TIMEDB_ENTRY( 30300443,  3,   1747,          YM), /* 01395 - Paul Shields - Jug.sndh */
TIMEDB_ENTRY( 3039f6ce,  0,   9425,          YM), /* 00158 - David Whittaker - Xenon.sndh */
TIMEDB_ENTRY( 3039f6ce,  1,   2560,          YM), /* 00158 - David Whittaker - Xenon.sndh */
TIMEDB_ENTRY( 3039f6ce,  2,   4803,          YM), /* 00158 - David Whittaker - Xenon.sndh */
TIMEDB_ENTRY( 303b3ca2,  0,  33280,       YM+TD), /* 01476 - STU - SHAMPOO CATTLE.sndh */
TIMEDB_ENTRY( 304a9ab5,  0,   8448,          YM), /* 02114 - Mad Max - JT In Space.sndh */
TIMEDB_ENTRY( 30622ebf,  0,   3450,    YM+TA+TD), /* 00991 - DMA-SC - Sidsound de Chocobo.sndh */
TIMEDB_ENTRY( 30688f23,  0,   3548,          YM), /* 02550 - Ben Daglish - Xybots.sndh */
TIMEDB_ENTRY( 30688f23,  1,    105,          YM), /* 02550 - Ben Daglish - Xybots.sndh */
TIMEDB_ENTRY( 306d7273,  0,   5759,    YM+TA+TB), /* 03294 - Dma-Sc - ATIPYK Developments - Manical Drop - Introduction.sndh */
TIMEDB_ENTRY( 30b5e9ba,  0,   4061,      YM+STE), /* 01638 - Jason - Jason Mix - STE only.sndh */
TIMEDB_ENTRY( 30c9d495,  0,   1536,          YM), /* 02482 - ENS Music - ENS 8.sndh */
TIMEDB_ENTRY( 30cf6273,  0,  13824,          YM), /* 01072 - Excellence In Art - Battle Tris.sndh */
TIMEDB_ENTRY( 30d39028,  0,  97586,       YM+TA), /* 03779 - Richard Joseph - International 3D Tennis Intro.sndh */
TIMEDB_ENTRY( 30fdfa0e,  0,   8253, YM+TA+TB+TD), /* 01009 - DMA-SC - Pocket Groove.sndh */
TIMEDB_ENTRY( 3117fc7b,  0,   9598, YM+TA+TB+TD), /* 03271 - Dubmood - Tempest fjortisfacials vs Zabutom.sndh */
TIMEDB_ENTRY( 3121658e,  0,   1151,    YM+TA+TD), /* 02825 - Msg - Superfly tune 2.sndh */
TIMEDB_ENTRY( 31241af4,  0,   7800,          YM), /* 00572 - Jas C Brooke - Revenge of the Mutant Camels.sndh */
TIMEDB_ENTRY( 3127ae07,  0,  16895, YM+TA+TB+TD), /* 00994 - Dma-Sc-Sector1^YM-Rockerz - Fantasia Tune 1 (ST Medley).sndh */
TIMEDB_ENTRY( 3141548d,  0,  19422,          YM), /* 01806 - Barry Leitch - Space Crusade.sndh */
TIMEDB_ENTRY( 3141548d,  1,    854,          YM), /* 01806 - Barry Leitch - Space Crusade.sndh */
TIMEDB_ENTRY( 3141548d,  2,   1501,          YM), /* 01806 - Barry Leitch - Space Crusade.sndh */
TIMEDB_ENTRY( 3141548d,  3,   9589,          YM), /* 01806 - Barry Leitch - Space Crusade.sndh */
TIMEDB_ENTRY( 3168eef8,  0,   8835,          YM), /* 01813 - Barry Leitch - Xenophobe.sndh */
TIMEDB_ENTRY( 316c52b9,  0,   8639,    YM+TA+TD), /* 03213 - marcer 2007 - we remember.sndh */
TIMEDB_ENTRY( 316e1a68,  0,   3834,       YM+TA), /* 01673 - Jess - Unfinished, Sorry AN Cool.sndh */
TIMEDB_ENTRY( 3176b0f3,  0,   6816,          YM), /* 00217 - Markus Weichselbaum & D Pleiss - Logical.sndh */
TIMEDB_ENTRY( 3176b0f3,  1,   5120,          YM), /* 00217 - Markus Weichselbaum & D Pleiss - Logical.sndh */
TIMEDB_ENTRY( 3176b0f3,  2,   3840,          YM), /* 00217 - Markus Weichselbaum & D Pleiss - Logical.sndh */
TIMEDB_ENTRY( 3176b0f3,  3,   6529,          YM), /* 00217 - Markus Weichselbaum & D Pleiss - Logical.sndh */
TIMEDB_ENTRY( 3176b0f3,  4,   3264,          YM), /* 00217 - Markus Weichselbaum & D Pleiss - Logical.sndh */
TIMEDB_ENTRY( 318448b0,  0,  13316,    YM+TA+TD), /* 02295 - Tao - Cream - Pulverize (Alive 3).sndh */
TIMEDB_ENTRY( 319185e0,  0,   7680,          YM), /* 00613 - Chrispy Noodle - Hybris.sndh */
TIMEDB_ENTRY( 31995c1b,  0, 112420,       YM+TA), /* 04822 - Twitch - Modelmix 91.sndh */
TIMEDB_ENTRY( 319daf8a,  0,    133,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  1,    157,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  2,   2816,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  3,    768,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  4,    141,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  5,    173,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  6,     39,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  7,     39,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  8,    254,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a,  9,     75,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 10,      9,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 11,     39,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 12,     44,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 13,    127,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 14,    768,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 319daf8a, 15,    768,          YM), /* 01904 - Mikael Lundberg - Rush.sndh */
TIMEDB_ENTRY( 31a35d34,  0,  17088,          YM), /* 02439 - Bodenstandig 2000 - HRVAT.sndh */
TIMEDB_ENTRY( 31a4d5b3,  0,  19016,          YM), /* 02283 - Tao - Cream - Cultivation (Just Buggin Main Menu).sndh */
TIMEDB_ENTRY( 31b317dd,  0,   8320,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  1,   1792,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  2,    381,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  3,    646,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  4,    239,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  5,    201,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b317dd,  6,    976,          YM), /* 01062 - Jonathan Dunn - Navy Seals.sndh */
TIMEDB_ENTRY( 31b72ca2,  0,  15460, YM+TA+TB+TD), /* 02816 - Marcer - Paradize - Nuclear Waste Dump - Ingame tune 3.sndh */
TIMEDB_ENTRY( 31bcee5a,  0,     36,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  1,    207,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  2,    120,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  3,     15,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  4,     69,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  5,     84,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  6,     15,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  7,     30,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  8,      9,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a,  9,     18,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 10,     18,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 11,   3370,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 12,    741,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 13,    496,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 14,    216,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 15,    150,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 17,     75,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 18,     18,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 19,    105,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 20,     81,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 21,     24,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 22,    123,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 23,     75,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 24,    165,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 25,    114,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 26,     78,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 27,    602,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 28,   1092,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 29,     45,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 30,     99,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 31,     63,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 33,     69,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 34,     42,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 35,     45,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 36,     15,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 37,     36,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 38,     42,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 39,     42,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 40,     78,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 41,     42,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 42,     63,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 43,    144,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 44,    123,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 45,    234,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 46,    216,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 47,    144,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31bcee5a, 48,     45,          YM), /* 04741 - Mark Crowe-Al Lowe - Space Quest Chapter 2 - Vohauls Revenge.sndh */
TIMEDB_ENTRY( 31ec5338,  0,   1215,          YM), /* 01704 - Jess - Please, no.sndh */
TIMEDB_ENTRY( 31f1d335,  0,   6336,          YM), /* 00301 - Gary Wheaton - Clementine.sndh */
TIMEDB_ENTRY( 3211a066,  0,   3072,          YM), /* 00981 - Chris Denman - Tornado Ground Attack.sndh */
TIMEDB_ENTRY( 3211a066,  1,   8192,          YM), /* 00981 - Chris Denman - Tornado Ground Attack.sndh */
TIMEDB_ENTRY( 3213e05d,  0,  23930,          YM), /* 01702 - Jess - The Legacy Nightmare.sndh */
TIMEDB_ENTRY( 321ef846,  0,  96982,       YM+TA), /* 04817 - Powapixel - Gyroscope.sndh */
TIMEDB_ENTRY( 323d8987,  0,   7454,          YM), /* 00093 - David Whittaker - Days of Thunder.sndh */
TIMEDB_ENTRY( 323d8987,  1,   1513,          YM), /* 00093 - David Whittaker - Days of Thunder.sndh */
TIMEDB_ENTRY( 323d8987,  2,   1755,          YM), /* 00093 - David Whittaker - Days of Thunder.sndh */
TIMEDB_ENTRY( 324b9b0f,  0,  36656, YM+TA+TB+TD), /* 03616 - Marcer - ITS MY PARADIZE.sndh */
TIMEDB_ENTRY( 3254498b,  0,   8640,          YM), /* 01975 - MSG - MSG 05.sndh */
TIMEDB_ENTRY( 3258fbe7,  0,  32008,    YM+TA+TD), /* 02354 - Tao - Cream - Groupie (YM Rockerz 3).sndh */
TIMEDB_ENTRY( 3262aef0,  0, 174671,       YM+TA), /* 04359 - X The Beermat - Next.sndh */
TIMEDB_ENTRY( 3265e0fb,  0,  17640, YM+TA+TB+TD), /* 02064 - Mad Max - Ace 2 - Sid.sndh */
TIMEDB_ENTRY( 3275a326,  0,   1845,          YM), /* 02154 - Mad Max - Ooh Crikey - Loader.sndh */
TIMEDB_ENTRY( 328a9d41,  0,  12641,       YM+TA), /* 03845 - Ingo Massalski - Tolteka - Title.sndh */
TIMEDB_ENTRY( 328d0d86,  0,   2304,       YM+TA), /* 00420 - Big Alec - Pursuit 2.sndh */
TIMEDB_ENTRY( 329b3c96,  0,  44924,    YM+TA+TD), /* 02786 - Tao-Cream - M143R1.sndh */
TIMEDB_ENTRY( 32a31884,  0,   4204,       YM+TD), /* 01988 - Mad Max - New Music.sndh */
TIMEDB_ENTRY( 32a5303c,  0,  10750,          YM), /* 01532 - Iso - Sheeeeep.sndh */
TIMEDB_ENTRY( 32a7f031,  0,   3200,          YM), /* 04002 - Blue Nun - Gnurk Demo Bobo.sndh */
TIMEDB_ENTRY( 32b60f4b,  0,  12481,    YM+TA+TD), /* 02319 - Tao - Cream - Lost.sndh */
TIMEDB_ENTRY( 32c4f870,  0,   6528,          YM), /* 00933 - Dark Angel - Masters Of Shit.sndh */
TIMEDB_ENTRY( 32e8616d,  0,   8442, YM+TA+TB+TD), /* 02648 - Dma-Sc - ExoCooper.sndh */
TIMEDB_ENTRY( 32e87446,  0,   2304,          YM), /* 01110 - Keith Tinman - Head Over Heels.sndh */
TIMEDB_ENTRY( 32e87446,  1,    118,          YM), /* 01110 - Keith Tinman - Head Over Heels.sndh */
TIMEDB_ENTRY( 32f3082d,  0,   7272, YM+TA+TB+TD), /* 00054 - 505 - Headaxe.sndh */
TIMEDB_ENTRY( 332b6163,  0,   5676,       YM+TA), /* 01918 - Mephistow - Ventura - Evil.sndh */
TIMEDB_ENTRY( 333b47b3,  0,   6526,          YM), /* 03198 - Martin Iveson - Arcade Classix - Breakout.sndh */
TIMEDB_ENTRY( 33526cb7,  0,  10370,    YM+TA+TD), /* 02311 - Tao - Cream - Intensity.sndh */
TIMEDB_ENTRY( 3355a037,  0,   8192, YM+TA+TB+TD), /* 03336 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 3356789d,  0,   6541,       YM+TA), /* 02155 - Mad Max - Ooh Crikey Wot a Scorcher Hidden.sndh */
TIMEDB_ENTRY( 335b93b1,  0,   8556, YM+TA+TB+TD), /* 04456 - Frequent-Ephidrena - Revers.sndh */
TIMEDB_ENTRY( 335f6906,  0,   8532, YM+TA+TB+TD), /* 01481 - Gary - UCM 4 Intro.sndh */
TIMEDB_ENTRY( 33625fef,  0,  26106,          YM), /* 03874 - Gary Wheaton - Sidelong.sndh */
TIMEDB_ENTRY( 3364a361,  0,   8064,          YM), /* 01131 - Tyan - Goldrunner.sndh */
TIMEDB_ENTRY( 337a575a,  0,  12171,       YM+TA), /* 02247 - Tao - Cream (orig. Rob Hubbard) - Zoolook (Stmygm).sndh */
TIMEDB_ENTRY( 337a9ba2,  0,   4984,          YM), /* 03114 - Unknown - Where Have All th Flowers Gone.sndh */
TIMEDB_ENTRY( 337fe6b7,  0,   8324,          YM), /* 02924 - Allister Brimble - Pro Boxing Simulator.sndh */
TIMEDB_ENTRY( 337fe6b7,  1,    244,          YM), /* 02924 - Allister Brimble - Pro Boxing Simulator.sndh */
TIMEDB_ENTRY( 3387177e,  0,   1285,          YM), /* 03980 - Claude Abromont - Killer Ball.sndh */
TIMEDB_ENTRY( 339a471b,  0,   8447, YM+TA+TB+TD), /* 03452 - Per Almered (XiA) 2009 - Handy-man (The Knife) 1.01.sndh */
TIMEDB_ENTRY( 339bc133,  0,     91,          YM), /* 00228 - ACC - Copy Star 1.8.sndh */
TIMEDB_ENTRY( 33cd3499,  0, 375414,       YM+TA), /* 03732 - Waxup - Humuskus (Ventura Reset).sndh */
TIMEDB_ENTRY( 33e21d35,  0,   3834,       YM+TA), /* 01128 - Tyan - Cruelana.sndh */
TIMEDB_ENTRY( 33e2f4a5,  0,   5759, YM+TA+TB+TD), /* 02729 - maktone-fairlight - jewel.sndh */
TIMEDB_ENTRY( 33e891ad,  0,   3520,          YM), /* 01071 - Excellence In Art - Barf.sndh */
TIMEDB_ENTRY( 3400c3dd,  0,   4991, YM+TA+TB+TD), /* 02630 - rubbish - ANOTHER TIME.sndh */
TIMEDB_ENTRY( 340925b9,  0,  20480,          YM), /* 00407 - Big Alec - Judgement Day (ST News 7.2).sndh */
TIMEDB_ENTRY( 341f26d6,  0,   6162,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 341f26d6,  1,  14018,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 341f26d6,  2,   8642,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 341f26d6,  3,   2881,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 341f26d6,  4,   4626,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 341f26d6,  5,   1538,          YM), /* 04198 - Big Alec - Audio Artisic Demo.sndh */
TIMEDB_ENTRY( 34216cde,  0,   2298,          YM), /* 01931 - Miq - Dream Work.sndh */
TIMEDB_ENTRY( 3421c628,  0,   4348,          YM), /* 04581 - Andrew Taylor - Big Run.sndh */
TIMEDB_ENTRY( 3421c628,  1,   3068,          YM), /* 04581 - Andrew Taylor - Big Run.sndh */
TIMEDB_ENTRY( 3421c628,  2,    207,          YM), /* 04581 - Andrew Taylor - Big Run.sndh */
TIMEDB_ENTRY( 3425043a,  0,   2466,          YM), /* 03932 - Unknown - Tango 001.sndh */
TIMEDB_ENTRY( 344b88ed,  0,  15937,          YM), /* 02137 - Mad Max - Masters Of Magic - LCD.sndh */
TIMEDB_ENTRY( 34613b0c,  0,   7847,    YM+TA+TB), /* 03200 - monk - disturbance bilocated.sndh */
TIMEDB_ENTRY( 347b809f,  0,   2304,          YM), /* 01762 - Lap - Lap 2.sndh */
TIMEDB_ENTRY( 348a8522,  0,   3328,       YM+TA), /* 04433 - Lotus - Synth Dream 2 - Pace.sndh */
TIMEDB_ENTRY( 34affc14,  0,   1024,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  1,   4160,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  2,   1280,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  3,   4224,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  4,   1536,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  5,   2688,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  6,   1152,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34affc14,  7,   9310,          YM), /* 03405 - Wally Beben - Ultimate Ride.sndh */
TIMEDB_ENTRY( 34ba9d61,  0,   2688,       YM+TA), /* 01692 - Jess - The name's Bond, Jess Bond.sndh */
TIMEDB_ENTRY( 34c46ff5,  0,    393,          YM), /* 02944 - Dave Munsie - Frogger.sndh */
TIMEDB_ENTRY( 34c46ff5,  1,    323,          YM), /* 02944 - Dave Munsie - Frogger.sndh */
TIMEDB_ENTRY( 34c46ff5,  2,    185,          YM), /* 02944 - Dave Munsie - Frogger.sndh */
TIMEDB_ENTRY( 34ca54fa,  0,  16128,          YM), /* 00898 - Danny Oneway - Lamebrain 1 - Danny die.sndh */
TIMEDB_ENTRY( 34cafd52,  0,   4639,    YM+TA+TD), /* 02688 - Donald Fakk - Young.sndh */
TIMEDB_ENTRY( 34ecd6e9,  0,   3761,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  1,   1600,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  2,   1280,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  3,    960,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  4,   1384,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  5,   1536,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  6,   2316,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  7,   3503,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34ecd6e9,  8,   1149,          YM), /* 03573 - Paul Tonge - Ancient Art of War in the Skies.sndh */
TIMEDB_ENTRY( 34f7b6a2,  0,   1248,          YM), /* 01879 - Dave Lowe - Count Duckula 2.sndh */
TIMEDB_ENTRY( 35073069,  0,   1170,          YM), /* 00132 - David Whittaker - Road Blasters.sndh */
TIMEDB_ENTRY( 35073069,  1,    630,          YM), /* 00132 - David Whittaker - Road Blasters.sndh */
TIMEDB_ENTRY( 35073069,  2,   1145,          YM), /* 00132 - David Whittaker - Road Blasters.sndh */
TIMEDB_ENTRY( 35073069,  3,   2184,          YM), /* 00132 - David Whittaker - Road Blasters.sndh */
TIMEDB_ENTRY( 35073069,  4,     54,          YM), /* 00132 - David Whittaker - Road Blasters.sndh */
TIMEDB_ENTRY( 3510042a,  0,   3498,          YM), /* 02406 - ACC - Badine.sndh */
TIMEDB_ENTRY( 3524e07c,  0,  10744,       YM+TA), /* 02418 - Scavenger - Unlimited Bobs.sndh */
TIMEDB_ENTRY( 35344e06,  0,   4558,          YM), /* 03111 - n-a - .sndh */
TIMEDB_ENTRY( 353d7660,  0,   5119,          YM), /* 01784 - Lap - Phaleon - Intro.sndh */
TIMEDB_ENTRY( 35470845,  0,   6542,          YM), /* 03136 - Unknown - Michelle-Lennon-McCartney.sndh */
TIMEDB_ENTRY( 354fbef4,  0,    329,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  1,    158,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  2,    271,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  3,   4096,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  4,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  5,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  6,  15408,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  7,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  8,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4,  9,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4, 10,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4, 11,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 354fbef4, 12,  15136,          YM), /* 00069 - David Whittaker - Archipelagos.sndh */
TIMEDB_ENTRY( 3558592a,  0,    288,          YM), /* 03512 - Unknown - Dragon Lord.sndh */
TIMEDB_ENTRY( 3558592a,  1,    256,          YM), /* 03512 - Unknown - Dragon Lord.sndh */
TIMEDB_ENTRY( 3558592a,  2,    144,          YM), /* 03512 - Unknown - Dragon Lord.sndh */
TIMEDB_ENTRY( 355cca36,  0,   3577, YM+TA+TB+TD), /* 02592 - STU - FOR REAL.sndh */
TIMEDB_ENTRY( 3574919f,  0,   4559,    YM+TA+TD), /* 00827 - Crazy Q - TiMeBoMb - Timebomb Loader II.sndh */
TIMEDB_ENTRY( 35bbea7a,  0,  12835,    YM+TA+TD), /* 04576 - Crazy Q - Corrupted Data File.sndh */
TIMEDB_ENTRY( 35c38980,  0,  54342,       YM+TA), /* 03761 - Tufty - Tuft2.sndh */
TIMEDB_ENTRY( 35dbbb7f,  0,  10499,       YM+TA), /* 03433 - Per Almered (XiA) 2009 - They pull me back in.sndh */
TIMEDB_ENTRY( 35e91fb5,  0, 130388,       YM+TA), /* 04820 - Twitch - Funkymix.sndh */
TIMEDB_ENTRY( 35f2c927,  0,  25257, YM+TA+TB+TD), /* 03201 - cerror (original by Pasta!) - River city open.sndh */
TIMEDB_ENTRY( 360831c5,  0,   3199,          YM), /* 02446 - Tony Williams - Judge Dredd.sndh */
TIMEDB_ENTRY( 360831c5,  1,   1599,          YM), /* 02446 - Tony Williams - Judge Dredd.sndh */
TIMEDB_ENTRY( 361939ff,  0,   8192,    YM+TB+TD), /* 03981 - XyNo - XYNO A LA PLAGE!.sndh */
TIMEDB_ENTRY( 361be82a,  0, 338641,       YM+TA), /* 04555 - Flix - What Time Is It?.sndh */
TIMEDB_ENTRY( 361be82a,  1,  52926,       YM+TA), /* 04555 - Flix - What Time Is It?.sndh */
TIMEDB_ENTRY( 362726aa,  0,   6654,          YM), /* 03553 - Jovis - Spider Ship.sndh */
TIMEDB_ENTRY( 362aaf32,  0,   8687,       YM+TB), /* 02875 - 505-Paradox [nils feske] - fullscreen digin' [2005].sndh */
TIMEDB_ENTRY( 363afdb3,  0,  75264,          YM), /* 02430 - Atom - TBC Intro 22.sndh */
TIMEDB_ENTRY( 363edd1a,  0,   4497,          YM), /* 03172 - n-a - Star Wars -- Main Theme.sndh */
TIMEDB_ENTRY( 363ede8b,  0,  11173,       YM+TA), /* 04218 - Karsten Obarksi - Window Wizard Digi.sndh */
TIMEDB_ENTRY( 36401902,  0,   3074,       YM+TD), /* 00237 - Alx 33 - BJ.sndh */
TIMEDB_ENTRY( 36822b7f,  0,   3834,       YM+TA), /* 04244 - 505 - Checkpoint (1999) - Beenden (unreleased).sndh */
TIMEDB_ENTRY( 3693d8b4,  0, 103005,       YM+TA), /* 04045 - Roggie - Pocket Warmer.sndh */
TIMEDB_ENTRY( 369a938e,  0,  14311,          YM), /* 03158 - Unknown - Goodnight Saigon by B. Joel.sndh */
TIMEDB_ENTRY( 369ab556,  0,   5755,    YM+TA+TB), /* 01327 - Scavenger - Crapman Sid.sndh */
TIMEDB_ENTRY( 36b91fbb,  0,   2298, YM+TA+TB+TD), /* 01130 - Tyan - Dont Know.sndh */
TIMEDB_ENTRY( 36c0b47c,  0,   1580,       YM+TA), /* 03847 - Unknown - Victory - Title.sndh */
TIMEDB_ENTRY( 36c15f4c,  0,   2555,    YM+TA+TD), /* 02771 - Morgan - Garcimore2 Rire reset.sndh */
TIMEDB_ENTRY( 36c1ce10,  0,   7680, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  1,   4416, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  2,   5184, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  3,   7488, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  4,   5760, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  5,   3840, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36c1ce10,  6,   5760, YM+TA+TB+TD), /* 02075 - Mad Max - Final Synth - Sid.sndh */
TIMEDB_ENTRY( 36e173f7,  0,    233,          YM), /* 01729 - Peter Johnson - Arkanoid.sndh */
TIMEDB_ENTRY( 36e173f7,  1,    129,          YM), /* 01729 - Peter Johnson - Arkanoid.sndh */
TIMEDB_ENTRY( 36e173f7,  2,    170,          YM), /* 01729 - Peter Johnson - Arkanoid.sndh */
TIMEDB_ENTRY( 36e173f7,  3,    769,          YM), /* 01729 - Peter Johnson - Arkanoid.sndh */
TIMEDB_ENTRY( 36e173f7,  4,     96,          YM), /* 01729 - Peter Johnson - Arkanoid.sndh */
TIMEDB_ENTRY( 37100241,  0,   1359,          YM), /* 00136 - David Whittaker - Snowstrike.sndh */
TIMEDB_ENTRY( 37100241,  1,   1225,          YM), /* 00136 - David Whittaker - Snowstrike.sndh */
TIMEDB_ENTRY( 37100241,  2,   1043,          YM), /* 00136 - David Whittaker - Snowstrike.sndh */
TIMEDB_ENTRY( 37100241,  3,   1474,          YM), /* 00136 - David Whittaker - Snowstrike.sndh */
TIMEDB_ENTRY( 3738c1fa,  0,   4473,       YM+TA), /* 01144 - Tyan - Loader.sndh */
TIMEDB_ENTRY( 375738fb,  0,   3589,          YM), /* 00293 - Gary Wheaton - Abhonk (aboriginal version).sndh */
TIMEDB_ENTRY( 376031f2,  0, 198325,       YM+TA), /* 03745 - Tufty - Hot Guitarin.sndh */
TIMEDB_ENTRY( 376335a6,  0,     40,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  1,    204,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  2,    353,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  3,    359,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  4,    366,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  5,    288,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 376335a6,  6,    287,          YM), /* 04496 - Charles E. Cullinan II - Batman - The Batwing.sndh */
TIMEDB_ENTRY( 3765fccc,  0,   6553,          YM), /* 00075 - David Whittaker - Beverly Hills Cop.sndh */
TIMEDB_ENTRY( 3786df73,  0,   9722,          YM), /* 00498 - Tangens - The 2nd Try - Dister.sndh */
TIMEDB_ENTRY( 37901b71,  0,  31098,       YM+TA), /* 04235 - 505 - 14th.sndh */
TIMEDB_ENTRY( 379f051d,  0,  36096, YM+TA+TB+TD), /* 04196 - Serpent - Compo Breaker.sndh */
TIMEDB_ENTRY( 37b04015,  0,   8832,          YM), /* 01915 - Max - Whatta Big Rat.sndh */
TIMEDB_ENTRY( 37b4bacb,  0, 200618,       YM+TA), /* 03718 - Spaz - Ooh Crikey - Intro.sndh */
TIMEDB_ENTRY( 37f988bd,  0,   2640,          YM), /* 01884 - Dave Lowe - Final Fight.sndh */
TIMEDB_ENTRY( 37f988bd,  1,   1632,          YM), /* 01884 - Dave Lowe - Final Fight.sndh */
TIMEDB_ENTRY( 380b3cec,  0,  12288,          YM), /* 02115 - Mad Max - Scout Remix.sndh */
TIMEDB_ENTRY( 382bc9dd,  0,   6144,          YM), /* 04825 - T1000 - Bits 53.sndh */
TIMEDB_ENTRY( 3838f30a,  0, 179792,       YM+TA), /* 04582 - Gary Wheaton - Shapeshifters.sndh */
TIMEDB_ENTRY( 3839299f,  0,   4986,       YM+TA), /* 04233 - 505 - 11th.sndh */
TIMEDB_ENTRY( 384551a3,  0,   4088,       YM+TA), /* 04761 - Ferox - State of Mind - Victory.sndh */
TIMEDB_ENTRY( 38565ed1,  0,   7829,          YM), /* 03186 - n-a - YANKEE DOODLE BOY.sndh */
TIMEDB_ENTRY( 38623bdd,  0,  11135,       YM+TA), /* 02493 - gwEm - Golden Years Over (used).sndh */
TIMEDB_ENTRY( 38678353,  0,  52096,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38678353,  1,  22471,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38678353,  2,  23247,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38678353,  3,  18145,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38678353,  4,  14581,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38678353,  5,  10001,       YM+TA), /* 04396 - Golden Otter - Hibernation Carol Demo 2.sndh */
TIMEDB_ENTRY( 38691dc2,  0,  30720, YM+TA+TB+TD), /* 02586 - STU - DATADANCE.sndh */
TIMEDB_ENTRY( 38834d4a,  0,   4989, YM+TA+TB+TD), /* 00283 - Baggio - Teddy - UCM 20 Intro.sndh */
TIMEDB_ENTRY( 3885b70a,  0,  19135,          YM), /* 00308 - Gary Wheaton - Goil.sndh */
TIMEDB_ENTRY( 388b52b5,  0,   8509,          YM), /* 03088 - n-a - She's Always A Woman.sndh */
TIMEDB_ENTRY( 389cb649,  0,   4921, YM+TA+TB+TD), /* 01956 - MSG - Godpey - Frontend.sndh */
TIMEDB_ENTRY( 38a97c4c,  0,   4798,          YM), /* 00969 - Demon X - Awesome.sndh */
TIMEDB_ENTRY( 38b24d90,  0,    224, YM+TA+TB+TD), /* 03624 - Excellence In Art - Sommarhack Report 2011 Pt 3.sndh */
TIMEDB_ENTRY( 38b9bdb1,  0,  28672,          YM), /* 02458 - Starbuck^Creative Mind - Memories.sndh */
TIMEDB_ENTRY( 38d851a2,  0,   6160,          YM), /* 02368 - Nexus 6 - Count One.sndh */
TIMEDB_ENTRY( 38eeac82,  0,  23897, YM+TA+TB+TD), /* 03300 - 505-checkpoint - morphonic.sndh */
TIMEDB_ENTRY( 3903ca97,  0,  13440,    YM+TA+TB), /* 01331 - Scavenger - Enchanted Jambala.sndh */
TIMEDB_ENTRY( 390fff1b,  0,  15648,          YM), /* 00267 - Antichrist - The Union.sndh */
TIMEDB_ENTRY( 39196440,  0,   1024,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  2,   1536,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  3,    249,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  4,  11009,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  5,   7482,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  6,   1536,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 39196440,  7,    862,          YM), /* 01567 - Jason Page - Realms.sndh */
TIMEDB_ENTRY( 393e5990,  0,   1914,          YM), /* 00444 - Augustin Vidovic - Ace 2.sndh */
TIMEDB_ENTRY( 397d0167,  0,   7709,          YM), /* 01036 - Donald Fakk - ITC Digi.sndh */
TIMEDB_ENTRY( 39cfda35,  0,   5832,          YM), /* 01972 - MSG - MSG 02.sndh */
TIMEDB_ENTRY( 39ddf67a,  0,  15092,       YM+TD), /* 01989 - Jochen Hippel - Solidify (tSCc's  Modcomp 15).sndh */
TIMEDB_ENTRY( 39de27ad,  0,  10745, YM+TA+TB+TD), /* 00829 - Crazy Q - TiMeBoMb - Live vers. Monster Phunk.sndh */
TIMEDB_ENTRY( 39e50cee,  0,  10181,          YM), /* 01665 - Stefan Jerowski - Mindbender.sndh */
TIMEDB_ENTRY( 3a18d56b,  0,   1480,          YM), /* 03895 - Unknown - Tango.sndh */
TIMEDB_ENTRY( 3a1b5564,  0,   4645,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a1b5564,  1,   3839,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a1b5564,  2,   3839,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a1b5564,  3,   2879,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a1b5564,  4,   5119,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a1b5564,  5,   2639,          YM), /* 01318 - Matt Furniss - Skull and Crossbones.sndh */
TIMEDB_ENTRY( 3a34fc47,  0,   8965,    YM+TA+TD), /* 01864 - Lotek Style - Rays of neon.sndh */
TIMEDB_ENTRY( 3a35dedc,  0, 246380,       YM+TA), /* 04818 - Powapixel - Prelude.sndh */
TIMEDB_ENTRY( 3a3b135b,  0,   3519,       YM+TB), /* 02868 - Monk - Smalltrip.sndh */
TIMEDB_ENTRY( 3a4bd0b6,  0,  11284, YM+TA+TB+TD), /* 01613 - No More - Face.sndh */
TIMEDB_ENTRY( 3a4e6052,  0,   9470,       YM+TA), /* 04671 - Unknown (orig Mike Alsop) - Wizardry.sndh */
TIMEDB_ENTRY( 3a4ea60b,  0,   5378,       YM+TD), /* 01898 - LTK - PS Harm.sndh */
TIMEDB_ENTRY( 3a5ca48e,  0,   2300,          YM), /* 02433 - Beast - TSC - Boring.sndh */
TIMEDB_ENTRY( 3a6042e4,  0,   1536,          YM), /* 01405 - Skyline - Snork Intro - Lightning.sndh */
TIMEDB_ENTRY( 3aa242af,  0,   8953,    YM+TA+TD), /* 01870 - Lotek Style - Silence.sndh */
TIMEDB_ENTRY( 3aac495c,  0,  29185,          YM), /* 00397 - Wally Beben - Total Eclipse.sndh */
TIMEDB_ENTRY( 3acbd283,  0,   7983,       YM+TB), /* 04104 - Per Almered-XiA-Checkpoint - SledgeBusters.sndh */
TIMEDB_ENTRY( 3aceb5d2,  0,  25450,       YM+TA), /* 04816 - Unknown - Elenor.sndh */
TIMEDB_ENTRY( 3ae209b5,  0,  11904,          YM), /* 03218 - Tomchi - Oldckool - Intro.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  0,  33248,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  1,   4736,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  2,   5136,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  3,   5120,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  4,   5088,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3ae5b6bb,  5,   8128,          YM), /* 01669 - Stefan Jerowski - Spherical.sndh */
TIMEDB_ENTRY( 3af52968,  0,   8190,          YM), /* 04073 - Gary Wheaton - Home Drunk.sndh */
TIMEDB_ENTRY( 3b01656b,  0,   2240,          YM), /* 02632 - Miggy - Credits(unfinished).sndh */
TIMEDB_ENTRY( 3b04eeac,  0,  20480, YM+TA+TB+TD), /* 02556 - Marcer - BYE BYE FRIEND.sndh */
TIMEDB_ENTRY( 3b216790,  0,  45312, YM+TA+TB+TD), /* 02528 - 505 - CPU Eater.sndh */
TIMEDB_ENTRY( 3b33b9f3,  0,   5760, YM+TA+TC+TD), /* 03664 - 505 - Outline 2010 Invite.sndh */
TIMEDB_ENTRY( 3b380ed1,  0,  50053,       YM+TA), /* 03964 - Scott Lemmon - I am the Walrus.sndh */
TIMEDB_ENTRY( 3b38fa87,  0,   3090,          YM), /* 03451 - Per Almered (XiA) 2009 - 1812 Overture(Tchaikovsky).sndh */
TIMEDB_ENTRY( 3b49a639,  0,   7680,          YM), /* 04764 - YERZMYEY-AY-RIDERS - EVELYNN.sndh */
TIMEDB_ENTRY( 3b5517d0,  0,  36992,          YM), /* 03861 - xyce - Cpp is Forbidden - Part 1.sndh */
TIMEDB_ENTRY( 3b62cb45,  0,   4332,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  1,   5168,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  2,   3286,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  3,   7326,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  4,   6084,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  5,  11568,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  6,   6746,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  7,   7315,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b62cb45,  8,   9155,          YM), /* 00863 - Ben Daglish - Motorhead.sndh */
TIMEDB_ENTRY( 3b6be469,  0,  32656, YM+TA+TB+TD), /* 03868 - Tao - Cream (2012) - Sillyride (SillyVenture 2k12).sndh */
TIMEDB_ENTRY( 3b76bf13,  0,   6913,          YM), /* 03970 - Eric Thoner - Super Senori.sndh */
TIMEDB_ENTRY( 3b93406c,  0,   3840,          YM), /* 02168 - Mad Max - Union Demo - Level 16.sndh */
TIMEDB_ENTRY( 3ba26fab,  0,   7680,          YM), /* 03191 - n-a - Organic Robots.sndh */
TIMEDB_ENTRY( 3baf6d79,  0,   2281,          YM), /* 03922 - Unknown - Marseill.sndh */
TIMEDB_ENTRY( 3bd36ec3,  0,    383,          YM), /* 00358 - Mark Wilson - Cloud Kingdom Dizzy.sndh */
TIMEDB_ENTRY( 3bd36ec3,  1,   6911,          YM), /* 00358 - Mark Wilson - Cloud Kingdom Dizzy.sndh */
TIMEDB_ENTRY( 3bd36ec3,  2,   4991,          YM), /* 00358 - Mark Wilson - Cloud Kingdom Dizzy.sndh */
TIMEDB_ENTRY( 3bd36ec3,  3,   4799,          YM), /* 00358 - Mark Wilson - Cloud Kingdom Dizzy.sndh */
TIMEDB_ENTRY( 3bd36ec3,  4,   4799,          YM), /* 00358 - Mark Wilson - Cloud Kingdom Dizzy.sndh */
TIMEDB_ENTRY( 3bd3bd4e,  0,  39767,       YM+TA), /* 04804 - Unknown - Fugue.sndh */
TIMEDB_ENTRY( 3bd95ceb,  0,   3839, YM+TA+TB+TD), /* 02990 - 505-paradox - long ago.sndh */
TIMEDB_ENTRY( 3bf50ca8,  0,  15344,    YM+TA+TD), /* 02275 - Tao - Cream - Wide Awake (Steps).sndh */
TIMEDB_ENTRY( 3bfe1a32,  0,   1920,          YM), /* 04608 - White Ninja - Farewell.sndh */
TIMEDB_ENTRY( 3c0e9f47,  0,   8448,          YM), /* 02022 - Mad Max - Bump Set And Spike.sndh */
TIMEDB_ENTRY( 3c0e9f47,  1,   9504,          YM), /* 02022 - Mad Max - Bump Set And Spike.sndh */
TIMEDB_ENTRY( 3c21902f,  0,  10177,          YM), /* 02109 - Mad Max - Esbe.sndh */
TIMEDB_ENTRY( 3c35d9fd,  0,   4356,          YM), /* 00319 - Gary Wheaton - The Monkey Man.sndh */
TIMEDB_ENTRY( 3c381868,  0,   6406,          YM), /* 04148 - Unknown - City Roads.sndh */
TIMEDB_ENTRY( 3c4609cc,  0,  58904,          YM), /* 01441 - Stax - Es Geht! Oder.sndh */
TIMEDB_ENTRY( 3c4c8046,  0,   9789, YM+TA+TB+TD), /* 03503 - 505 - Suretrip 2 - Main.sndh */
TIMEDB_ENTRY( 3c4d57d5,  0,   2080,          YM), /* 03170 - Unknown - Star-Spangled Banner - F.S.Key.sndh */
TIMEDB_ENTRY( 3c6d4ca1,  0,   2682,       YM+TA), /* 01337 - Scavenger - In Full Effect.sndh */
TIMEDB_ENTRY( 3c6fc76f,  0,   9216,          YM), /* 01576 - Powa Pixel - GBISTCP - Mega 4 A.sndh */
TIMEDB_ENTRY( 3c78197c,  0,   9262,       YM+TA), /* 01712 - Jess - Dyscomelodus.sndh */
TIMEDB_ENTRY( 3c833d4d,  0,  12075, YM+TA+TB+TD), /* 03437 - Dma-Sc - Happy Dancefloor.sndh */
TIMEDB_ENTRY( 3c96ac85,  0,   5904,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  1,   1535,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  2,   2063,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  3,   2303,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  4,   2303,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  5,   1535,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  6,     90,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  7,    141,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c96ac85,  8,   1535,          YM), /* 00356 - Tony Williams - Viz.sndh */
TIMEDB_ENTRY( 3c9e7164,  0,    383,          YM), /* 00359 - Mark Wilson - Round The Bend.sndh */
TIMEDB_ENTRY( 3c9e7164,  1,     95,          YM), /* 00359 - Mark Wilson - Round The Bend.sndh */
TIMEDB_ENTRY( 3c9e7164,  2,     96,          YM), /* 00359 - Mark Wilson - Round The Bend.sndh */
TIMEDB_ENTRY( 3c9e7164,  3,    132,          YM), /* 00359 - Mark Wilson - Round The Bend.sndh */
TIMEDB_ENTRY( 3c9e7164,  4,   1535,          YM), /* 00359 - Mark Wilson - Round The Bend.sndh */
TIMEDB_ENTRY( 3cad8c4c,  0,   2112,          YM), /* 04477 - Full FX - Sooty & Sweep.sndh */
TIMEDB_ENTRY( 3cb4ea89,  0,  20730,          YM), /* 01708 - Jess - Do you speak Russian.sndh */
TIMEDB_ENTRY( 3cba31e7,  0,   8640,          YM), /* 01179 - Excellence In Art - Frosty's Music.sndh */
TIMEDB_ENTRY( 3cd8c70a,  0,   5570,          YM), /* 01052 - Dr Beat - Imagination - Rollercoaster.sndh */
TIMEDB_ENTRY( 3cff211e,  0,    160,          YM), /* 04728 - Beyonder - MIX number ONE demo - Slide TSK.sndh */
TIMEDB_ENTRY( 3cff211e,  1,   1828,          YM), /* 04728 - Beyonder - MIX number ONE demo - Slide TSK.sndh */
TIMEDB_ENTRY( 3cff211e,  2,   1828,          YM), /* 04728 - Beyonder - MIX number ONE demo - Slide TSK.sndh */
TIMEDB_ENTRY( 3d061dce,  0,   2784, YM+TA+TB+TD), /* 03830 - Rhino - Torment (D Newman) - Airwolf (Sylvester Levay).sndh */
TIMEDB_ENTRY( 3d0d2149,  0,  12960,          YM), /* 01994 - Mad Max - Arpy Remix.sndh */
TIMEDB_ENTRY( 3d0f7efd,  0,  12703,          YM), /* 01760 - Lap - Expose Software Intro.sndh */
TIMEDB_ENTRY( 3d12a99a,  0,  17206,         STE), /* 03366 - 505 - Paradox - More Or Less Zero.sndh */
TIMEDB_ENTRY( 3d156e5f,  0,   4607,          YM), /* 04569 - Per Almered-XiA-Checkpoint - LOUDER! LOUDER!.sndh */
TIMEDB_ENTRY( 3d17e34a,  0,   6144,          YM), /* 01431 - Starbuck - Sleep Well.sndh */
TIMEDB_ENTRY( 3d1d30fc,  0,   3253,          YM), /* 03081 - Tim Shockley - Dance of the Candy Fairy.sndh */
TIMEDB_ENTRY( 3d1fc0d1,  0,  36096,       YM+TD), /* 01477 - STU - SPACE GUZZI.sndh */
TIMEDB_ENTRY( 3d2a9513,  0,  44544, YM+TA+TB+TD), /* 03393 - cerror - SALVATOR.sndh */
TIMEDB_ENTRY( 3d3164c3,  0,   9594,    YM+TA+TB), /* 00988 - DMA-SC - First Sid.sndh */
TIMEDB_ENTRY( 3d3ba81e,  0,  35937,       YM+TA), /* 03729 - Martin Walker - Defender 2.sndh */
TIMEDB_ENTRY( 3d3ccb16,  0,  11520,       YM+TA), /* 04119 - Tao - For Abyss - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 3d3da197,  0,   2314,       YM+TA), /* 02470 - Iso - Knuckle Joe.sndh */
TIMEDB_ENTRY( 3d43c43e,  0, 307273,       YM+TA), /* 03941 - Unknown - Acid1.sndh */
TIMEDB_ENTRY( 3d4e2b53,  0,  13440,          YM), /* 04748 - Mad Max - Warp Preview.sndh */
TIMEDB_ENTRY( 3d4e2b53,  1,   9216,          YM), /* 04748 - Mad Max - Warp Preview.sndh */
TIMEDB_ENTRY( 3d55bcc0,  0,   7753, YM+TA+TB+TD), /* 02713 - lotek style - tscc (2006) - farewell lil ssd.sndh */
TIMEDB_ENTRY( 3d61c948,  0,   6272,          YM), /* 00948 - Dark Angel - Slow Motion.sndh */
TIMEDB_ENTRY( 3d65e51b,  0,   6218,          YM), /* 03142 - Unknown - NIGHT MOVES by Marilyn Martin.sndh */
TIMEDB_ENTRY( 3d68b7ad,  0,    192,          YM), /* 01210 - Excellence In Art - Tra La La.sndh */
TIMEDB_ENTRY( 3d6ad039,  0,   4612,          YM), /* 03112 - n-a - .sndh */
TIMEDB_ENTRY( 3d7de33c,  0,   1536,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  1,    373,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  2,    342,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  3,    403,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  4,     57,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  5,     80,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  6,    216,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  7,    204,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7de33c,  8,     65,          YM), /* 00145 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( 3d7ea149,  0,     27,          YM), /* 04687 - Gary Cook, Robert Forsey - Kubes.sndh */
TIMEDB_ENTRY( 3d7ea149,  1,    940,          YM), /* 04687 - Gary Cook, Robert Forsey - Kubes.sndh */
TIMEDB_ENTRY( 3d7ea149,  2,     33,          YM), /* 04687 - Gary Cook, Robert Forsey - Kubes.sndh */
TIMEDB_ENTRY( 3d7ea149,  3,    127,          YM), /* 04687 - Gary Cook, Robert Forsey - Kubes.sndh */
TIMEDB_ENTRY( 3d990103,  0,   5375,       YM+TA), /* 04859 - Michel Winogradoff - Bumpy.sndh */
TIMEDB_ENTRY( 3da7acb5,  0,   9942,       YM+TA), /* 04429 - Lotus - Synth Dream 2 - Intro.sndh */
TIMEDB_ENTRY( 3db4ae01,  0,   3515, YM+TA+TB+TD), /* 01121 - Toodeloo - Test 2.sndh */
TIMEDB_ENTRY( 3dcae6b6,  0,  12864,          YM), /* 02450 - David Whittaker - Stormbringer.sndh */
TIMEDB_ENTRY( 3dcc854a,  0,   5115, YM+TA+TB+TD), /* 01252 - Floopy - Speeder.sndh */
TIMEDB_ENTRY( 3dd3a07e,  0,   7506,          YM), /* 01780 - Lap - Lap 33.sndh */
TIMEDB_ENTRY( 3dd6f8cd,  0,  27269,       YM+TA), /* 03763 - Tufty - Vangelis 2.sndh */
TIMEDB_ENTRY( 3dd94e9c,  0,   4810,          YM), /* 00754 - Mark Cooksey - Live and Let Die.sndh */
TIMEDB_ENTRY( 3dd94e9c,  1,   7296,          YM), /* 00754 - Mark Cooksey - Live and Let Die.sndh */
TIMEDB_ENTRY( 3de043eb,  0,   6720, YM+TA+TB+TD), /* 01606 - Nemo - Summer.sndh */
TIMEDB_ENTRY( 3df7689f,  0,  23680,    YM+TB+TD), /* 03992 - XyNo - That Fuckr.sndh */
TIMEDB_ENTRY( 3e0c5f1b,  0,   3200,          YM), /* 04599 - Rude Kid - Match of the Day.sndh */
TIMEDB_ENTRY( 3e13d051,  0,   3499,          YM), /* 04655 - Eric Kinkead - Nervana Quest.sndh */
TIMEDB_ENTRY( 3e165308,  0,   4927,          YM), /* 04096 - Gary Wheaton - Don't Let the Sun go down on me.sndh */
TIMEDB_ENTRY( 3e27512d,  0,   5166, YM+TA+TB+TD), /* 02752 - MSG - reservoir gods - omtrek-uit-lijn.sndh */
TIMEDB_ENTRY( 3e2a318d,  0,   4282,          YM), /* 00096 - David Whittaker - Emlyn Hughes Football.sndh */
TIMEDB_ENTRY( 3e2a47e1,  0,   2559,          YM), /* 02845 - Marcer - poker intro.sndh */
TIMEDB_ENTRY( 3e2ae99a,  0,   2808,          YM), /* 00475 - Ultimatum - Choux FL.sndh */
TIMEDB_ENTRY( 3e36efbe,  0,   8447, YM+TA+TB+TD), /* 02692 - DUBMOOD (JOGEIR & JUGI) - LOVEDEMO AND DOPE.sndh */
TIMEDB_ENTRY( 3e392c78,  0,   3960,          YM), /* 00488 - Ultimatum - Kidney Bean 4.sndh */
TIMEDB_ENTRY( 3e405883,  0,   2328,          YM), /* 02202 - Mad Max - Jinks.sndh */
TIMEDB_ENTRY( 3e405883,  1,    192,          YM), /* 02202 - Mad Max - Jinks.sndh */
TIMEDB_ENTRY( 3e45d068,  0,  20045, YM+TA+TB+TD), /* 02444 - Dubmood & Zabutom ! ! ! - SpaceJourney.sndh */
TIMEDB_ENTRY( 3e5ea92c,  0,   5313,          YM), /* 00329 - Gary Wheaton - Steel.sndh */
TIMEDB_ENTRY( 3e62c353,  0,  30013,       YM+TA), /* 03948 - Scott Lemmon - Axel F.sndh */
TIMEDB_ENTRY( 3e81befd,  0,   1535,       YM+TA), /* 03475 - gwEm (Follin conversion) - Gabba Bobble.sndh */
TIMEDB_ENTRY( 3eae2b14,  0,   4600,       YM+TA), /* 00028 - 505 - 6th.sndh */
TIMEDB_ENTRY( 3ecb3852,  0,  18286,       YM+TA), /* 03737 - David Whittaker - Fire & Brimstone.sndh */
TIMEDB_ENTRY( 3ecdf406,  0,   5824,       YM+TD), /* 01600 - Nemo - Nex.sndh */
TIMEDB_ENTRY( 3ed11e5c,  0,   2233,          YM), /* 04372 - Jess - Squeeze 11 (draft).sndh */
TIMEDB_ENTRY( 3ed3140e,  0,   8832,       YM+TA), /* 04115 - Tao - Androids - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 3ed78f3e,  0,   9520,          YM), /* 02975 - Dma-Sc - Bubble Days.sndh */
TIMEDB_ENTRY( 3ef913de,  0,   2298,          YM), /* 04785 - Intruder - Intruder 1.sndh */
TIMEDB_ENTRY( 3f2963c7,  0,   3264,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  1,   1536,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  2,   2688,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  3,   3456,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  4,    256,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  5,   3072,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  6,   3072,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f2963c7,  7,   1792,          YM), /* 04131 - Andi McGinty - The Simpsons - Bart vs The World.sndh */
TIMEDB_ENTRY( 3f386dc7,  0,   1536,          YM), /* 01197 - Excellence In Art - Prince Ig.sndh */
TIMEDB_ENTRY( 3f425b8f,  0,   4992,          YM), /* 01143 - Tyan - Revolution - Kiffer Song.sndh */
TIMEDB_ENTRY( 3f4b1e79,  0,   1287, YM+TA+TB+TD), /* 02797 - Starbuck^Creative Mind - Sonic - Bridge Zone.sndh */
TIMEDB_ENTRY( 3f614d44,  0,  17280, YM+TA+TB+TD), /* 03373 - STU - NIGHT JAM.sndh */
TIMEDB_ENTRY( 3f699102,  0,    600,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f699102,  1,   1080,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f699102,  2,   1164,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f699102,  3,   1251,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f699102,  4,   1060,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f699102,  5,   1140,          YM), /* 04153 - Unknown - Lauderbilder.sndh */
TIMEDB_ENTRY( 3f6c599b,  0,   8832,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  1,   4336,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  2,   2849,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  3,   1820,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  4,   5120,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  5,   3840,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  6,   3233,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f6c599b,  7,   2671,          YM), /* 00154 - David Whittaker - Weird Dreams.sndh */
TIMEDB_ENTRY( 3f82ce3a,  0,   9154,          YM), /* 04746 - Unknown - Beyond the Titanic 2.sndh */
TIMEDB_ENTRY( 3f870be2,  0,   3456,          YM), /* 00454 - Augustin Vidovic - Noisy Pillars.sndh */
TIMEDB_ENTRY( 3fb6c852,  0,   6590,          YM), /* 00506 - Tangens - Phaleon - Arsch.sndh */
TIMEDB_ENTRY( 3fe1891d,  0,   5456, YM+TA+TB+TD), /* 03599 - 505-checkpoint & paradox - Grado Groove.sndh */
TIMEDB_ENTRY( 3ffd8004,  0,   3291,          YM), /* 03155 - n-a - .sndh */
TIMEDB_ENTRY( 4005e38e,  0,  48384,          YM), /* 01440 - Stax - Match.sndh */
TIMEDB_ENTRY( 400d8c94,  0,   3840,       YM+TA), /* 04590 - Illusion - Pure Energy 6.sndh */
TIMEDB_ENTRY( 400dd2fb,  0,   5183, YM+TA+TB+TD), /* 00059 - 505 - Checkpoint (2002) - hysteria.sndh */
TIMEDB_ENTRY( 40132169,  0,   7680,       YM+TA), /* 01359 - Scavenger - Powercode.sndh */
TIMEDB_ENTRY( 4015556f,  0,  32732,          YM), /* 01472 - STU - PLAYER.sndh */
TIMEDB_ENTRY( 4021e89b,  0,   5370, YM+TA+TB+TD), /* 04276 - 505-cpt^pdx - laughter.sndh */
TIMEDB_ENTRY( 402d0e07,  0,  14652,    YM+TA+TD), /* 04575 - Crazy Q - Alive Team - Lunar Bounce.sndh */
TIMEDB_ENTRY( 40311d45,  0,   3453,    YM+TA+TD), /* 00213 - 505 - YM Pop (D-Bug 173).sndh */
TIMEDB_ENTRY( 40525f4c,  0,  20736,    YM+TA+TD), /* 02259 - Tao - Cream - Mine Sweeper ST (Steps).sndh */
TIMEDB_ENTRY( 406772da,  0,   7615,    YM+TA+TD), /* 04783 - gwEm - Trans-Debug Express (release edit).sndh */
TIMEDB_ENTRY( 40845105,  0,   2064,          YM), /* 01411 - Skyline - Snork Intro - PYM.sndh */
TIMEDB_ENTRY( 40912b3d,  0,    862,          YM), /* 04140 - Claudio Kronmuller - Kaiser.sndh */
TIMEDB_ENTRY( 40912b3d,  1,    364,          YM), /* 04140 - Claudio Kronmuller - Kaiser.sndh */
TIMEDB_ENTRY( 40b08c8f,  0,   1599,          YM), /* 04134 - Andreas Maier - Dizzy Wizard.sndh */
TIMEDB_ENTRY( 40b08c8f,  1,    858,          YM), /* 04134 - Andreas Maier - Dizzy Wizard.sndh */
TIMEDB_ENTRY( 40b0a6da,  0,  18048,          YM), /* 01542 - Ralph - Ralph 5.sndh */
TIMEDB_ENTRY( 40c0a003,  0,   9336, YM+TA+TB+TD), /* 03153 - gwEm - Garage Days Revisited.sndh */
TIMEDB_ENTRY( 40cbc1ce,  0,   2693,          YM), /* 02276 - Tao - Cream (orig. ??,C64) - Back Off Bitch (Just Buggin).sndh */
TIMEDB_ENTRY( 40e6dff2,  0,   3068, YM+TA+TB+TD), /* 01019 - DMA-SC - Wings of Death Remix.sndh */
TIMEDB_ENTRY( 40e6eb78,  0,  10057, YM+TA+TB+TD), /* 04636 - !Cube - Bullet Sequence.sndh */
TIMEDB_ENTRY( 40f113c5,  0,   7100,    YM+TA+TD), /* 03479 - Dma-Sc - Sector One - Hail to Sunny Days.sndh */
TIMEDB_ENTRY( 4114fb5d,  0,  10880,          YM), /* 01393 - Shaft - Xtream Dentro.sndh */
TIMEDB_ENTRY( 4117f3aa,  0,   5181,       YM+TA), /* 04224 - 505 - 4MC.sndh */
TIMEDB_ENTRY( 41641b13,  0,   5247,       YM+TB), /* 03620 - DMA-SC - Superstar.sndh */
TIMEDB_ENTRY( 416b9770,  0,   4096,          YM), /* 03660 - Jedi - Nothing Is More - Song 2.sndh */
TIMEDB_ENTRY( 416f7801,  0,   8442,       YM+TA), /* 01837 - Lotek Style - Blinded.sndh */
TIMEDB_ENTRY( 4186f72f,  0,  12160,          YM), /* 03417 - Dave Hasler - Steigar.sndh */
TIMEDB_ENTRY( 4186f72f,  1,   5120,          YM), /* 03417 - Dave Hasler - Steigar.sndh */
TIMEDB_ENTRY( 4197ebcf,  0,   3520,          YM), /* 02364 - Nexus 6 - Bin Doof.sndh */
TIMEDB_ENTRY( 419cdbc4,  0,  14784,          YM), /* 02037 - Mad Max - I-Ball.sndh */
TIMEDB_ENTRY( 41a0f845,  0,  14329, YM+TA+TB+TD), /* 01960 - MSG - Godpey - Normal.sndh */
TIMEDB_ENTRY( 41abaeae,  0,  11464,          YM), /* 02740 - Unknown Composer - The Karate Kid - Part II.sndh */
TIMEDB_ENTRY( 41b7ed8b,  0,   2044,          YM), /* 03551 - Jovis - Lame Bip.sndh */
TIMEDB_ENTRY( 41c77720,  0,   4200,          YM), /* 03245 - T1000 - The Beast Club - TBC 17.sndh */
TIMEDB_ENTRY( 41d2e0e9,  0,   1664,          YM), /* 04500 - Jon Wheatman - Bootiful Babe.sndh */
TIMEDB_ENTRY( 41f03310,  0,   4602,       YM+TA), /* 03556 - Stuart Trigg - Around The World.sndh */
TIMEDB_ENTRY( 41fbfde9,  0,  18816,          YM), /* 00945 - Dark Angel - Mathematica - Seasons.sndh */
TIMEDB_ENTRY( 42030931,  0,  61674,       YM+TA), /* 03717 - LTK - LTK Music Disk.sndh */
TIMEDB_ENTRY( 420af484,  0,   4229, YM+TA+TB+TD), /* 00908 - Danny Oneway - No Sheep In Mongolia (DIGI).sndh */
TIMEDB_ENTRY( 421b40a5,  0,   1146,       YM+TA), /* 01324 - Scavenger - Another Kind of Heaven.sndh */
TIMEDB_ENTRY( 422fc19d,  0,   5520,          YM), /* 01069 - Jonathan Dunn - Terminator 2.sndh */
TIMEDB_ENTRY( 422fc19d,  1,   1920,          YM), /* 01069 - Jonathan Dunn - Terminator 2.sndh */
TIMEDB_ENTRY( 422fc19d,  2,   2704,          YM), /* 01069 - Jonathan Dunn - Terminator 2.sndh */
TIMEDB_ENTRY( 4237532f,  0,   4992,          YM), /* 00800 - Crawdaddy - My Master.sndh */
TIMEDB_ENTRY( 425d6b8b,  0,   6399, YM+TA+TB+TD), /* 02997 - 505-paradox - psycho.sndh */
TIMEDB_ENTRY( 428d7eda,  0,   5425, YM+TA+TB+TD), /* 04250 - 505-checkpoint - Crapsong.sndh */
TIMEDB_ENTRY( 42a1a7d3,  0,   2106,    YM+TA+TD), /* 01938 - Modmate - Popcorn.sndh */
TIMEDB_ENTRY( 42a6e7d7,  0,   4992,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  1,     33,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  2,   1707,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  3,    315,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  4,    853,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  5,    763,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  6,    113,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  7,    611,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42a6e7d7,  8,     93,          YM), /* 00097 - David Whittaker - Empire Strikes Back.sndh */
TIMEDB_ENTRY( 42b355e4,  0,   9594,       YM+TA), /* 01756 - Eric Ksouri - Manciola - Etnos.sndh */
TIMEDB_ENTRY( 42cb41ab,  0,  13441,    YM+TA+TD), /* 02346 - Tao - Cream - Caught A Ray (UCM 22).sndh */
TIMEDB_ENTRY( 42d744dd,  0,   6140, YM+TA+TB+TD), /* 02755 - Dma-Sc - Sector One - Alive! 7 Intro.sndh */
TIMEDB_ENTRY( 42e2b05e,  0,   1920,          YM), /* 02310 - Tao - Cream - Ikari Union (1k-version).sndh */
TIMEDB_ENTRY( 42f07c44,  0,   8463,          YM), /* 01396 - Paul Shields - ST Soccer.sndh */
TIMEDB_ENTRY( 42f07c44,  1,   5485,          YM), /* 01396 - Paul Shields - ST Soccer.sndh */
TIMEDB_ENTRY( 42f07c44,  2,   5315,          YM), /* 01396 - Paul Shields - ST Soccer.sndh */
TIMEDB_ENTRY( 42f07c44,  3,   5642,          YM), /* 01396 - Paul Shields - ST Soccer.sndh */
TIMEDB_ENTRY( 42f33e2e,  0,  34560, YM+TA+TB+TD), /* 03598 - xyce - SES SEINS.sndh */
TIMEDB_ENTRY( 42f54684,  0,   9024,          YM), /* 00804 - Crawdaddy - No Way Out.sndh */
TIMEDB_ENTRY( 431cd63b,  0,  11768, YM+TA+TB+TD), /* 01142 - Tyan - Kanabis.sndh */
TIMEDB_ENTRY( 434ecd34,  0,   8016,    YM+TA+TD), /* 03577 - 505-checkpoint - Zyncyventure.sndh */
TIMEDB_ENTRY( 4351da35,  0,   5760,          YM), /* 02921 - 4-Mat - Heimdall.sndh */
TIMEDB_ENTRY( 43527b57,  0,   1161,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  1,   1739,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  2,   6047,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  3,   2027,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  4,   4229,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  5,   1733,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  6,   2404,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  7,   2891,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 43527b57,  8,   1733,          YM), /* 02549 - Ben Daglish - Dynamite Dux.sndh */
TIMEDB_ENTRY( 4365fca2,  0,   1858,          YM), /* 00501 - Tangens - The 2nd Try - Loader.sndh */
TIMEDB_ENTRY( 436ce977,  0,   2044,       YM+TB), /* 04303 - 505 - Checkpoint - retelling.sndh */
TIMEDB_ENTRY( 4376cf27,  0,  19135,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  1,  13668,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  2,   4005,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  3,   2384,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  4,   2032,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  5,   4670,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  6,   5313,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4376cf27,  7,   2565,          YM), /* 03867 - Gary Wheaton - Ace Invaders.sndh */
TIMEDB_ENTRY( 4389dc3b,  0,  37888,    YM+TB+TD), /* 02843 - Marcer - XMASMAN SID.sndh */
TIMEDB_ENTRY( 43981f38,  0,   8704,          YM), /* 01740 - Jovis - Lametrop - Beast.sndh */
TIMEDB_ENTRY( 439d0a4d,  0,  12291,    YM+TA+TD), /* 02341 - Tao - Cream - Monkey Cage (UCM 15).sndh */
TIMEDB_ENTRY( 439f139b,  0,   9984,          YM), /* 01770 - Lap - Lap 10.sndh */
TIMEDB_ENTRY( 43aeb26e,  0,  18048,    YM+TA+TD), /* 02345 - Tao - Cream - Eclipse (UCM 20).sndh */
TIMEDB_ENTRY( 43b169f3,  0,   7864,          YM), /* 00307 - Gary Wheaton - Beware of my Love.sndh */
TIMEDB_ENTRY( 43b78b1c,  0,  13056,       YM+TA), /* 02309 - Tao - Cream (orig. Jeroen Tel) - Ice Age (Digidrum-version).sndh */
TIMEDB_ENTRY( 43b9d655,  0,  29573,       YM+TA), /* 03755 - Tufty - Numan.sndh */
TIMEDB_ENTRY( 43c519f7,  0,    105,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  1,     12,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  2,    201,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  3,    673,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  4,    673,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  5,   2559,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  6,    177,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  7,     96,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  8,     12,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7,  9,     39,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 10,    147,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 11,    150,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 12,    114,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 14,    162,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 15,    201,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 16,   2566,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 17,     12,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 18,    240,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 19,    240,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 20,    111,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 21,      6,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c519f7, 22,    691,          YM), /* 04742 - Barry Murray - Manhunter II - San Francisco - B.sndh */
TIMEDB_ENTRY( 43c79c0e,  0,   3881,          YM), /* 03056 - Ed Bogas - Lead Vocal.sndh */
TIMEDB_ENTRY( 43cd6912,  0,   6020, YM+TA+TB+TD), /* 02991 - 505-paradox - nuance (STE only).sndh */
TIMEDB_ENTRY( 43e98c51,  0,   2017,          YM), /* 03894 - Unknown - Studie 9.sndh */
TIMEDB_ENTRY( 4408eb86,  0,   3456,          YM), /* 01202 - Excellence In Art - Stolen.sndh */
TIMEDB_ENTRY( 441d5b0b,  0,   1076,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  1,    559,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  2,     15,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  3,    174,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  4,    178,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  5,    275,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  6,    204,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  7,    153,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  8,    144,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b,  9,    144,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 10,    592,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 11,    144,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 12,    163,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 13,   3907,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 14,   3323,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d5b0b, 15,   2694,          YM), /* 02947 - Chris Parrish - Attack of Alien Gun Turrets.sndh */
TIMEDB_ENTRY( 441d9f8b,  0,     96,          YM), /* 03926 - Unknown - Richtig1.sndh */
TIMEDB_ENTRY( 442a4206,  0,   4992,          YM), /* 01184 - Excellence In Art - Hidden.sndh */
TIMEDB_ENTRY( 4438f589,  0,  38400, YM+TA+TB+TD), /* 02627 - schmx-awp - Bonebreaker.sndh */
TIMEDB_ENTRY( 4472aef8,  0,   4372,          YM), /* 04510 - Edwyn Thomas - Byte-man.sndh */
TIMEDB_ENTRY( 4472aef8,  1,    447,          YM), /* 04510 - Edwyn Thomas - Byte-man.sndh */
TIMEDB_ENTRY( 4472aef8,  2,    441,          YM), /* 04510 - Edwyn Thomas - Byte-man.sndh */
TIMEDB_ENTRY( 4477eb41,  0,  14070,          YM), /* 01746 - Jovis - Lametrop - Jet Set Willy.sndh */
TIMEDB_ENTRY( 44808679,  0,   6912,          YM), /* 00479 - Ultimatum - Choux K.sndh */
TIMEDB_ENTRY( 448093ee,  0,   4590,    YM+TB+TD), /* 04268 - 505 - Jelling.sndh */
TIMEDB_ENTRY( 4483c9f0,  0, 157385,       YM+TA), /* 03722 - Spaz - Ooh Crikey - What A Bummer.sndh */
TIMEDB_ENTRY( 4486eb53,  0,   5633,          YM), /* 00389 - Wally Beben - Liberator.sndh */
TIMEDB_ENTRY( 4486eb53,  1,    272,          YM), /* 00389 - Wally Beben - Liberator.sndh */
TIMEDB_ENTRY( 448a0597,  0,   3329,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  1,   2881,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  2,   2881,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  3,   3841,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  4,    645,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  5,    321,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448a0597,  6,    321,          YM), /* 02925 - Allister Brimble - Mean Machine.sndh */
TIMEDB_ENTRY( 448ef395,  0,   1536,          YM), /* 00101 - David Whittaker - Future Sports.sndh */
TIMEDB_ENTRY( 448ef395,  1,   3961,          YM), /* 00101 - David Whittaker - Future Sports.sndh */
TIMEDB_ENTRY( 448ef395,  2,     47,          YM), /* 00101 - David Whittaker - Future Sports.sndh */
TIMEDB_ENTRY( 448ef395,  3,     81,          YM), /* 00101 - David Whittaker - Future Sports.sndh */
TIMEDB_ENTRY( 448ef395,  4,    122,          YM), /* 00101 - David Whittaker - Future Sports.sndh */
TIMEDB_ENTRY( 449a8f2a,  0,   7783,          YM), /* 03064 - Chris French - Axel F  by Harold Faltermeyer.sndh */
TIMEDB_ENTRY( 449ce984,  0,   5006,          YM), /* 04444 - ISO - Non-Digified.sndh */
TIMEDB_ENTRY( 44bf284b,  0,   7199, YM+TA+TB+TD), /* 02974 - 505-paradox - cyborg clubbing.sndh */
TIMEDB_ENTRY( 44f3f5ec,  0,  39936,          YM), /* 02606 - STU - MY MELODY.sndh */
TIMEDB_ENTRY( 4510f9f1,  0,   8064,       YM+TB), /* 01844 - Lotek Style - Echoland.sndh */
TIMEDB_ENTRY( 451130c6,  0,   3449,          YM), /* 03134 - n-a - Away in a Manger.sndh */
TIMEDB_ENTRY( 45225afa,  0,   4555,          YM), /* 00459 - Augustin Vidovic - Pyramid.sndh */
TIMEDB_ENTRY( 453eb90b,  0,   6912,          YM), /* 01064 - Jonathan Dunn - Operation Thunderbolt.sndh */
TIMEDB_ENTRY( 453eb90b,  1,    761,          YM), /* 01064 - Jonathan Dunn - Operation Thunderbolt.sndh */
TIMEDB_ENTRY( 453fd01f,  0,   3623,          YM), /* 02969 - Unknown - Wolf Pack.sndh */
TIMEDB_ENTRY( 4540aaf8,  0,   4312,       YM+TA), /* 01279 - Frazer - TSC 79.sndh */
TIMEDB_ENTRY( 4545c9ea,  0,   8640,    YM+TB+TD), /* 02851 - Helene - FUCKTHEUNIVERSE.sndh */
TIMEDB_ENTRY( 455b626d,  0,   8091,          YM), /* 03671 - Dma-Sc - CYG Japan.sndh */
TIMEDB_ENTRY( 456297c7,  0,  10464,          YM), /* 02217 - Mad Max - Tangram.sndh */
TIMEDB_ENTRY( 456297c7,  1,  19776,          YM), /* 02217 - Mad Max - Tangram.sndh */
TIMEDB_ENTRY( 45657a96,  0,  16890,    YM+TA+TD), /* 01834 - Lotek Style - Ask me about chips.sndh */
TIMEDB_ENTRY( 4568c87a,  0,    768,    YM+TA+TD), /* 00405 - Big Alec - High Gear.sndh */
TIMEDB_ENTRY( 45698dbb,  0,   3840,          YM), /* 01045 - Dr Beat - Galtan 6 - Elite Intro 2.sndh */
TIMEDB_ENTRY( 457c457a,  0,   3528,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  1,   2688,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  2,   3528,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  3,   3696,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  4,   1638,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  5,   2184,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  6,    813,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 457c457a,  7,    840,          YM), /* 04220 - Holger Ghermann - Graffiti Man.sndh */
TIMEDB_ENTRY( 45a4f0f1,  0,  14777, YM+TA+TB+TD), /* 04518 - 505 - Alive!.sndh */
TIMEDB_ENTRY( 45a5b063,  0,   8448,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  1,   9600,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  2,   6272,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  3,   9552,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  4,   5632,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  5,   7808,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  6,   6336,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  7,   6528,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  8,   8320,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063,  9,   8320,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063, 10,   6120,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45a5b063, 11,   6960,          YM), /* 02743 - Unknown Composer - Lemmings 2 - The Tribes.sndh */
TIMEDB_ENTRY( 45abad6d,  0,   5375, YM+TA+TB+TD), /* 00201 - 505 - Too Hard.sndh */
TIMEDB_ENTRY( 45b4b0d5,  0,   5120,       YM+TA), /* 04434 - Lotus - Synth Dream 2 - Of All Places.sndh */
TIMEDB_ENTRY( 45b5ef21,  0, 182661,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  1, 183263,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  2,  44359,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  3, 105778,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  4,  85803,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  5,  43723,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  6,  49924,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  7, 160563,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  8,   7850,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21,  9,  64494,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21, 10,  40043,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45b5ef21, 11,   6014,       YM+TA), /* 04562 - Marcer - Marcer Tunes.sndh */
TIMEDB_ENTRY( 45e4090f,  0,   3064,          YM), /* 01284 - Frequent - Ephidrena - Deep blue.sndh */
TIMEDB_ENTRY( 45e95c53,  0,  54944,    YM+TB+TD), /* 03440 - TomChi  NoExtra - KID IS VIP.sndh */
TIMEDB_ENTRY( 46238a3c,  0,   4800,          YM), /* 04601 - Rude Kid - Samba 2.sndh */
TIMEDB_ENTRY( 46447c65,  0,  20165,          YM), /* 02964 - Unknown - Maze.sndh */
TIMEDB_ENTRY( 464ab741,  0,    896,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  1,     23,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  2,    263,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  3,     23,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  4,     43,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  5,    177,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  6,     33,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  7,     45,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  8,     39,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741,  9,    110,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 10,     41,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 11,    153,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 12,    134,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 13,    177,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 14,    177,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 15,     45,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 16,     34,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 464ab741, 17,     56,          YM), /* 04534 - Ewald Greiml - Fox Brothers.sndh */
TIMEDB_ENTRY( 46528a63,  0,  13433,      YM+STE), /* 04365 - Jess - The singing Green Frog (STe).sndh */
TIMEDB_ENTRY( 4652cd27,  0,   2575,       YM+TA), /* 03839 - I. Waugh - Frost Byte - Title.sndh */
TIMEDB_ENTRY( 465a6bdc,  0,  37633,          YM), /* 03529 - Wally Beben - Dark Side.sndh */
TIMEDB_ENTRY( 4681d442,  0,  42240,          YM), /* 01439 - Stax - Madness.sndh */
TIMEDB_ENTRY( 469137d5,  0,   1920,          YM), /* 00235 - AJT - Beachtro.sndh */
TIMEDB_ENTRY( 46a3d7b1,  0,   3199,          YM), /* 03513 - Dave Kelly - Return of the Jedi.sndh */
TIMEDB_ENTRY( 46a3d7b1,  1,    151,          YM), /* 03513 - Dave Kelly - Return of the Jedi.sndh */
TIMEDB_ENTRY( 46a3d7b1,  2,    865,          YM), /* 03513 - Dave Kelly - Return of the Jedi.sndh */
TIMEDB_ENTRY( 46a3d7b1,  3,   1153,          YM), /* 03513 - Dave Kelly - Return of the Jedi.sndh */
TIMEDB_ENTRY( 46a6ab5e,  0,   5762,          YM), /* 00432 - Big Alec - Poppycock - Stars.sndh */
TIMEDB_ENTRY( 46b04ed3,  0,  28736, YM+TA+TB+TD), /* 03389 - Tomchi - YM out.sndh */
TIMEDB_ENTRY( 46b8222c,  0,   2304,       YM+TA), /* 00951 - Dark Angel - Straight Shooter.sndh */
TIMEDB_ENTRY( 46ba2db5,  0,   1152,          YM), /* 01410 - Skyline - Snork Intro - Overdrive.sndh */
TIMEDB_ENTRY( 46d54b18,  0,   7631,    YM+TA+TB), /* 04659 - Frequent-Ephidrena - Lykkerus.sndh */
TIMEDB_ENTRY( 46df70e6,  0,   3353,          YM), /* 02716 - Chrispy Noodle - Tiger 2.sndh */
TIMEDB_ENTRY( 4711a6b6,  0,   8448,          YM), /* 00247 - AN Cool - So Watt - Techatron.sndh */
TIMEDB_ENTRY( 4722c6d4,  0,    935,          YM), /* 04549 - Simon Brown - Inferno.sndh */
TIMEDB_ENTRY( 4722c6d4,  1,   1870,          YM), /* 04549 - Simon Brown - Inferno.sndh */
TIMEDB_ENTRY( 4722c6d4,  2,   1008,          YM), /* 04549 - Simon Brown - Inferno.sndh */
TIMEDB_ENTRY( 4726dc9b,  0,   3834,       YM+TA), /* 04451 - Scavenger - SYM II.sndh */
TIMEDB_ENTRY( 472e2b73,  0,  10747, YM+TA+TB+TD), /* 01000 - DMA-SC - its live - alive 4.sndh */
TIMEDB_ENTRY( 472fd6cf,  0,   9216,          YM), /* 01095 - Techno - Omed 1.sndh */
TIMEDB_ENTRY( 47430808,  0,  10880,       YM+TD), /* 02006 - Mad Max ( ?-2002 ) - Faith (YM Rockerz).sndh */
TIMEDB_ENTRY( 4746395c,  0,  29184,       YM+TD), /* 02573 - STU - 8 BIT STYLE.sndh */
TIMEDB_ENTRY( 477203a1,  0,   9282,          YM), /* 03113 - Unknown - Flashdance...What A Feeling.sndh */
TIMEDB_ENTRY( 47828d84,  0,   9280,       YM+TA), /* 00264 - Antichrist - Stephanie.sndh */
TIMEDB_ENTRY( 47a7ec6e,  0,   2153,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  1,   2395,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  2,   1921,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  3,   1388,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  4,   1523,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  5,   2036,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47a7ec6e,  6,    437,          YM), /* 00085 - David Whittaker - Chronoquest.sndh */
TIMEDB_ENTRY( 47be14ac,  0,  10759,       YM+TA), /* 02492 - gwEm - Golden Years Over (orig).sndh */
TIMEDB_ENTRY( 47d552ec,  0,  16896,    YM+TB+TD), /* 02904 - Helene - n-a.sndh */
TIMEDB_ENTRY( 47d9555b,  0,   9210, YM+TA+TB+TD), /* 01831 - Lotek Style - Alpha Proxima.sndh */
TIMEDB_ENTRY( 47ec864a,  0,    910,       YM+TA), /* 01145 - Tyan - Noisy Pillars.sndh */
TIMEDB_ENTRY( 47f306f4,  0,    960,    YM+TA+TD), /* 02766 - Bobby Prince [ST by Dma-Sc] - Interlude.sndh */
TIMEDB_ENTRY( 47f3f9ea,  0,   2857,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  1,   7295,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  2,   2847,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  3,   5715,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  4,   2875,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  5,   3573,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  6,   2712,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  7,   1524,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 47f3f9ea,  8,   1686,          YM), /* 00156 - David Whittaker - Wrath of the Demon.sndh */
TIMEDB_ENTRY( 480b7552,  0,  14336,          YM), /* 02810 - Starbuck - Sunrise.sndh */
TIMEDB_ENTRY( 48318dc0,  0,   2331,          YM), /* 04531 - Andrew R. Oakley - Flipped.sndh */
TIMEDB_ENTRY( 4842a105,  0,   9600,          YM), /* 00763 - Count Zero - Decade Flexiscroll.sndh */
TIMEDB_ENTRY( 4864e718,  0,  47200, YM+TA+TB+TD), /* 02597 - STU - INSECTZ.sndh */
TIMEDB_ENTRY( 48651fff,  0,  10752,          YM), /* 00589 - Celtic - Starworx.sndh */
TIMEDB_ENTRY( 48887d4f,  0,  37120, YM+TA+TB+TD), /* 03343 - Marcer - I WISH ST VERS.sndh */
TIMEDB_ENTRY( 48ab9818,  0,   3836,          YM), /* 01286 - Full FX - The Jetsons.sndh */
TIMEDB_ENTRY( 48ad0e6f,  0,   1556,      YM+STE), /* 01211 - Elof - Elof Mix - STE only.sndh */
TIMEDB_ENTRY( 48b294b1,  0,  18444,       YM+TA), /* 01630 - No More - World of Wonders.sndh */
TIMEDB_ENTRY( 48dbdd9f,  0,  26624,    YM+TB+TD), /* 04852 - xFalcon - HURRY UP.sndh */
TIMEDB_ENTRY( 48f34eb8,  0,   7296,          YM), /* 01664 - Jedi - Vision.sndh */
TIMEDB_ENTRY( 48f850f0,  0,   7903, YM+TA+TB+TD), /* 03424 - Dubmood - Old Unreleased Jam.sndh */
TIMEDB_ENTRY( 49116698,  0, 179792,       YM+TA), /* 03806 - Sparky - Wave.sndh */
TIMEDB_ENTRY( 49176ce2,  0,  14666,          YM), /* 00076 - David Whittaker - Beyond The Ice Palace.sndh */
TIMEDB_ENTRY( 49176ce2,  1,   2688,          YM), /* 00076 - David Whittaker - Beyond The Ice Palace.sndh */
TIMEDB_ENTRY( 491f4cbb,  0,  11497, YM+TA+TB+TD), /* 04770 - V0yager - Takeover.sndh */
TIMEDB_ENTRY( 4925240e,  0,  16576,       YM+TD), /* 02359 - Nexus 6 - Ancient Files V2.sndh */
TIMEDB_ENTRY( 492c0d99,  0,   2664,          YM), /* 03934 - Unknown - Titel.sndh */
TIMEDB_ENTRY( 49416146,  0,  71629,       YM+TA), /* 03009 - Mad Max - Warp Title Music.sndh */
TIMEDB_ENTRY( 4965b274,  0,   7199,          YM), /* 03683 - Dma-Sc - STreet Art.sndh */
TIMEDB_ENTRY( 49790e75,  0,   1896,          YM), /* 01651 - Jedi - Melody Debut.sndh */
TIMEDB_ENTRY( 49c76abc,  0,   8442,    YM+TA+TD), /* 02829 - Msg - Superfly tune 6.sndh */
TIMEDB_ENTRY( 49ccd65a,  0,   8879,          YM), /* 02372 - Nexus 6 - Deep Hall.sndh */
TIMEDB_ENTRY( 49d37360,  0,  24960,          YM), /* 02806 - Starbuck - Oldschool Chip #1.sndh */
TIMEDB_ENTRY( 49d3bbe7,  0,   6522,    YM+TA+TD), /* 01855 - Lotek Style - Industrial Safe Disk 31.sndh */
TIMEDB_ENTRY( 49df71dd,  0,   2616,          YM), /* 01112 - Keith Tinman - Parasol Stars.sndh */
TIMEDB_ENTRY( 49df71dd,  1,    864,          YM), /* 01112 - Keith Tinman - Parasol Stars.sndh */
TIMEDB_ENTRY( 49df71dd,  2,    385,          YM), /* 01112 - Keith Tinman - Parasol Stars.sndh */
TIMEDB_ENTRY( 49df71dd,  3,   3360,          YM), /* 01112 - Keith Tinman - Parasol Stars.sndh */
TIMEDB_ENTRY( 49df71dd,  4,   1536,          YM), /* 01112 - Keith Tinman - Parasol Stars.sndh */
TIMEDB_ENTRY( 49e00f76,  0,   5754,          YM), /* 04101 - 505 - Blazing (YM-version).sndh */
TIMEDB_ENTRY( 49ea8fbd,  0,  22934,          YM), /* 01885 - Dave Lowe - IK+.sndh */
TIMEDB_ENTRY( 49eaeecf,  0,  10239,          YM), /* 00340 - Tony Williams - Future Basketball.sndh */
TIMEDB_ENTRY( 49eaeecf,  1,    650,          YM), /* 00340 - Tony Williams - Future Basketball.sndh */
TIMEDB_ENTRY( 49eaeecf,  2,   3039,          YM), /* 00340 - Tony Williams - Future Basketball.sndh */
TIMEDB_ENTRY( 49eaeecf,  3,    791,          YM), /* 00340 - Tony Williams - Future Basketball.sndh */
TIMEDB_ENTRY( 49f00d2d,  0,    975,    YM+TA+TD), /* 02476 - No More - Intro.sndh */
TIMEDB_ENTRY( 4a0c614b,  0,   7116,          YM), /* 03697 - Chris Mad - Overdrive Demo - Pharadist.sndh */
TIMEDB_ENTRY( 4a0d1fd9,  0,   3134,          YM), /* 01432 - Starbuck - YM Reggae.sndh */
TIMEDB_ENTRY( 4a21f916,  0,    288,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  1,   5688,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  2,   7140,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  3,   5280,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  4,  10944,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  5,   5616,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  6,    768,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  7,   3136,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a21f916,  8,   1344,          YM), /* 02961 - Unknown - Dick Tracy.sndh */
TIMEDB_ENTRY( 4a239905,  0,   7953,       YM+TA), /* 00196 - 505 - ST Tri Sound.sndh */
TIMEDB_ENTRY( 4a33c893,  0,   4771, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  1,   5376, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  2,   5184, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  3,   6528, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  4,   4992, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  5,   3619, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a33c893,  6,   5376, YM+TA+TB+TD), /* 02078 - Mad Max - Gerry The Germ - Sid.sndh */
TIMEDB_ENTRY( 4a3c5e50,  0,   7168,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  1,  15168,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  2,   2688,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  3,  18483,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  4,  16384,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  5,  10908,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a3c5e50,  6,  12160,      YM+STE), /* 02203 - Mad Max - Lethal Xcess (STe).sndh */
TIMEDB_ENTRY( 4a4d0981,  0,   6715, YM+TA+TB+TD), /* 01863 - Lotek Style - Photograph (Depech Mode).sndh */
TIMEDB_ENTRY( 4a6eaac1,  0,   7609, YM+TA+TB+TD), /* 01624 - No More - No More 2.sndh */
TIMEDB_ENTRY( 4a6f5ee3,  0,  27648,          YM), /* 01462 - STU - DEEP.sndh */
TIMEDB_ENTRY( 4a79bd65,  0,    241,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  1,   8448,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  2,   2260,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  3,   1280,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  4,   2184,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  5,   3840,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  6,    120,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  7,   2641,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a79bd65,  8,   1280,          YM), /* 00134 - David Whittaker - Slapfight.sndh */
TIMEDB_ENTRY( 4a82eb53,  0,   6913,       YM+TD), /* 01983 - MSG - Outline.sndh */
TIMEDB_ENTRY( 4a844c81,  0,   4224,          YM), /* 01164 - Tyan - Tyan 3.sndh */
TIMEDB_ENTRY( 4a941b05,  0,   6181,          YM), /* 03677 - Dma-Sc - Antiques - Physics Part.sndh */
TIMEDB_ENTRY( 4ab7ff41,  0,   1152,       YM+TA), /* 00913 - Dark Angel - All Is Lost!.sndh */
TIMEDB_ENTRY( 4ad3ea2a,  0,  10751, YM+TA+TB+TD), /* 04222 - lotek style-tscc - fujiology archivist.sndh */
TIMEDB_ENTRY( 4addffb9,  0,  24192,          YM), /* 01469 - STU - MOSQUITO.sndh */
TIMEDB_ENTRY( 4adf2953,  0,   1535,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  1,   1535,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  2,    306,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  3,   1156,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  4,    757,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  5,   1156,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  6,    574,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  7,   1727,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4adf2953,  8,   1012,          YM), /* 02959 - Andrew Taylor - Rodland.sndh */
TIMEDB_ENTRY( 4ae357f9,  0,   6138,       YM+TA), /* 01937 - Modmate - Impact.sndh */
TIMEDB_ENTRY( 4b0767c5,  0,   4997, YM+TA+TB+TD), /* 02865 - 505-cpt^pdx - xmas.sndh */
TIMEDB_ENTRY( 4b07caff,  0,   6400,       YM+TA), /* 01632 - Nexus - Dreamzone - Circle twist.sndh */
TIMEDB_ENTRY( 4b0bb5c9,  0,   4034, YM+TA+TB+TD), /* 02697 - Dubmood - Razor 1911 - Haschkaka LiveInterlude.sndh */
TIMEDB_ENTRY( 4b10e039,  0,  10745,       YM+TA), /* 02726 - Acid - Blue Boys Mod Compil 32.sndh */
TIMEDB_ENTRY( 4b1404c5,  0,   6528,          YM), /* 00794 - Crawdaddy - Japan.sndh */
TIMEDB_ENTRY( 4b1998d1,  0,   5435,    YM+TA+TB), /* 04750 - Bagarn - Spice Boys - XXX-Mas.sndh */
TIMEDB_ENTRY( 4b354c95,  0,  52224,          YM), /* 01435 - Stax - Living in a Dream.sndh */
TIMEDB_ENTRY( 4b448ccd,  0,   5115,    YM+TA+TB), /* 00064 - 505 - Jumps.sndh */
TIMEDB_ENTRY( 4b448eac,  0,   5760,          YM), /* 01195 - Excellence In Art - Offbeat.sndh */
TIMEDB_ENTRY( 4b497fd4,  0,   3095,    YM+TA+TD), /* 01106 - Timbral - DHS (2003) - Discoman.sndh */
TIMEDB_ENTRY( 4b4ee88e,  0,   9213,    YM+TA+TB), /* 01004 - DMA-SC - O-Style.sndh */
TIMEDB_ENTRY( 4b72e92b,  0,  17044,    YM+TA+TD), /* 02301 - Tao - Cream - Birth (.tSCc. Modcomp14).sndh */
TIMEDB_ENTRY( 4b9293ff,  0,   6144,          YM), /* 03214 - Tomchi - DCKolors.sndh */
TIMEDB_ENTRY( 4b9dd6fc,  0,   4228,          YM), /* 00576 - Jas C Brooke - Starquake.sndh */
TIMEDB_ENTRY( 4b9dd6fc,  1,   2410,          YM), /* 00576 - Jas C Brooke - Starquake.sndh */
TIMEDB_ENTRY( 4b9dd6fc,  2,   2446,          YM), /* 00576 - Jas C Brooke - Starquake.sndh */
TIMEDB_ENTRY( 4b9dd6fc,  3,   1184,          YM), /* 00576 - Jas C Brooke - Starquake.sndh */
TIMEDB_ENTRY( 4b9dd6fc,  4,   3178,          YM), /* 00576 - Jas C Brooke - Starquake.sndh */
TIMEDB_ENTRY( 4b9e68c8,  0,    893,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  1,    959,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  2,   1199,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  3,   1150,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  4,   1534,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  5,   1020,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  6,   1005,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  7,    640,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  8,   1565,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8,  9,    168,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8, 10,    308,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8, 11,   1588,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4b9e68c8, 12,   4606,          YM), /* 01304 - Matt Furniss - Manix 90.sndh */
TIMEDB_ENTRY( 4ba30d8f,  0,   3891,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  1,     68,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  2,     41,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  3,     28,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  4,     62,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  5,     77,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba30d8f,  6,     79,          YM), /* 04348 - David Whittaker - Emlyn Hughes Arcade Quiz.sndh */
TIMEDB_ENTRY( 4ba3b734,  0,   8836,       YM+TA), /* 01348 - Scavenger - Myth.sndh */
TIMEDB_ENTRY( 4bac0176,  0,   3456,          YM), /* 01254 - Floopy - Xmas 03 Tune 3.sndh */
TIMEDB_ENTRY( 4bb4cf79,  0,   2688,       YM+TA), /* 00819 - Crawdaddy - XXX-XXX.sndh */
TIMEDB_ENTRY( 4bb692c8,  0,   2560,          YM), /* 01660 - Jedi - Unfinished.sndh */
TIMEDB_ENTRY( 4bc6183a,  0,   4800,          YM), /* 03493 - Ben Daglish - Chubby Gristle.sndh */
TIMEDB_ENTRY( 4bc6183a,  1, 122400,          YM), /* 03493 - Ben Daglish - Chubby Gristle.sndh */
TIMEDB_ENTRY( 4bc6183a,  2, 122400,          YM), /* 03493 - Ben Daglish - Chubby Gristle.sndh */
TIMEDB_ENTRY( 4be36326,  0,   3361,          YM), /* 01735 - Richard Joseph - Chaos Engine.sndh */
TIMEDB_ENTRY( 4be36326,  1,   1921,          YM), /* 01735 - Richard Joseph - Chaos Engine.sndh */
TIMEDB_ENTRY( 4bea4113,  0,   2828, YM+TA+TB+TD), /* 00285 - Baggio - Evolution - Speed E is spooky.sndh */
TIMEDB_ENTRY( 4bf2f8a5,  0,   8467,    YM+TA+TD), /* 03829 - Frequent - Jernanger.sndh */
TIMEDB_ENTRY( 4bf639b9,  0,    137,          YM), /* 04505 - Chris Dukiet - Brain Jogging.sndh */
TIMEDB_ENTRY( 4bf639b9,  1,    268,          YM), /* 04505 - Chris Dukiet - Brain Jogging.sndh */
TIMEDB_ENTRY( 4bf639b9,  2,    175,          YM), /* 04505 - Chris Dukiet - Brain Jogging.sndh */
TIMEDB_ENTRY( 4bf639b9,  3,   1005,          YM), /* 04505 - Chris Dukiet - Brain Jogging.sndh */
TIMEDB_ENTRY( 4c014df9,  0,  33793,       YM+TA), /* 03764 - Tufty - Vangelis.sndh */
TIMEDB_ENTRY( 4c195cc7,  0,  23041,       YM+TA), /* 03837 - TPM - Lorna.sndh */
TIMEDB_ENTRY( 4c3c06bd,  0,   2306,          YM), /* 02172 - Mad Max - Scoop - That's The Way It Is.sndh */
TIMEDB_ENTRY( 4c3c06bd,  1,  21698,          YM), /* 02172 - Mad Max - Scoop - That's The Way It Is.sndh */
TIMEDB_ENTRY( 4c66ad56,  0,  27648, YM+TA+TB+TD), /* 02611 - STU - PSYCHOTIC.sndh */
TIMEDB_ENTRY( 4c7d722f,  0,   1608,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  1,   2937,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  5,     57,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  6,     96,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  7,     86,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  8,     75,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f,  9,     63,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f, 10,     36,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f, 11,     73,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f, 12,    105,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c7d722f, 13,     67,          YM), /* 01298 - Matt Furniss - Hard Drivin.sndh */
TIMEDB_ENTRY( 4c861a0e,  0,   6144,          YM), /* 02159 - Mad Max - So Watt - TCB.sndh */
TIMEDB_ENTRY( 4c99e00a,  0,   2047,          YM), /* 04090 - Gary Wheaton - Real Programmer.sndh */
TIMEDB_ENTRY( 4ca2f120,  0,  14941,    YM+TA+TB), /* 00418 - Big Alec - Prophecy (ST News 11.1).sndh */
TIMEDB_ENTRY( 4cb39ae7,  0,   9408,          YM), /* 01505 - Rob Hubbard - Goldrunner.sndh */
TIMEDB_ENTRY( 4cb89c38,  0,  18228,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  1,  16512,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  2,   9216,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  3,   6416,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  4,  12032,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  5,  25392,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  6,  19776,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cb89c38,  7,  15872,          YM), /* 04468 - Charles Callet - Passengers in the Wind 2.sndh */
TIMEDB_ENTRY( 4cbebadc,  0,    195,          YM), /* 03316 - SH3-RG - Fanfare.sndh */
TIMEDB_ENTRY( 4cc72c6e,  0,   2091,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cc72c6e,  1,     57,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cc72c6e,  2,     80,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cc72c6e,  3,     76,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cc72c6e,  4,     94,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cc72c6e,  5,   3780,          YM), /* 00110 - David Whittaker - Jaws.sndh */
TIMEDB_ENTRY( 4cd34442,  0,   5120,          YM), /* 01577 - Powa Pixel - GBISTCP - Mega 4 B.sndh */
TIMEDB_ENTRY( 4ce32d80,  0,   8084, YM+TA+TB+TD), /* 02999 - 505-paradox - Sinesinger.sndh */
TIMEDB_ENTRY( 4ce5b155,  0,   1920,          YM), /* 03285 - Chris Denman - Rorkes Drift.sndh */
TIMEDB_ENTRY( 4ce8caa8,  0,  11835, YM+TA+TB+TD), /* 01608 - Nemo - Tempest 2000 Mix.sndh */
TIMEDB_ENTRY( 4cf10c3a,  0,  64298,       YM+TA), /* 04036 - Roggie - Roggie 18.sndh */
TIMEDB_ENTRY( 4d0139e8,  0,   3840,          YM), /* 04443 - Iso - Easter Uncky.sndh */
TIMEDB_ENTRY( 4d017b7d,  0,   3584,          YM), /* 00604 - Chris Mad - Phalanx 3D.sndh */
TIMEDB_ENTRY( 4d192683,  0,   4502, YM+TA+TB+TD), /* 00066 - 505 - Kill Bill.sndh */
TIMEDB_ENTRY( 4d1f2f40,  0,   4607,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d1f2f40,  1,   3309,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d1f2f40,  2,   3095,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d1f2f40,  3,    416,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d1f2f40,  4,    200,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d1f2f40,  5,    559,          YM), /* 00353 - Tony Williams - Super Monaco Grand Prix.sndh */
TIMEDB_ENTRY( 4d23d41e,  0,   6818,          YM), /* 01047 - Dr Beat - NTM - Crusaders.sndh */
TIMEDB_ENTRY( 4d23fd12,  0,   3072,          YM), /* 00612 - Chrispy Noodle - Commando Hi-Score.sndh */
TIMEDB_ENTRY( 4d32f6ab,  0,  23104, YM+TA+TB+TD), /* 03372 - STU - KYOU WA KOKO MADE.sndh */
TIMEDB_ENTRY( 4d41504d,  0,   2213,          YM), /* 03937 - Mr Coke - Dasda.sndh */
TIMEDB_ENTRY( 4d4208c3,  0,   9840,          YM), /* 04480 - Dave Lowe - Raffles.sndh */
TIMEDB_ENTRY( 4d63d8b4,  0,  35640,       YM+TA), /* 03482 - Peter Jorgensen - Out For You.sndh */
TIMEDB_ENTRY( 4d68ce12,  0,  31488,          YM), /* 01458 - STU - AHARMAN.sndh */
TIMEDB_ENTRY( 4d6b8e91,  0,   5369, YM+TA+TB+TD), /* 04103 - lotek style-tscc - noisy again 2013.sndh */
TIMEDB_ENTRY( 4d75dfa2,  0,  11200,       YM+TD), /* 00360 - Wolfgang - ICS Intro.sndh */
TIMEDB_ENTRY( 4d8ebcb0,  0,   3088,          YM), /* 04394 - Ben Daglish - Saint & Greavsie.sndh */
TIMEDB_ENTRY( 4dae1219,  0,   5817,          YM), /* 03566 - DMA SC - Coast II Coast Edit.sndh */
TIMEDB_ENTRY( 4db771b0,  0,   6912,       YM+TD), /* 02413 - Nexus 6 - Tired Guitars.sndh */
TIMEDB_ENTRY( 4dbed7d0,  0,  13824,    YM+TB+TD), /* 02609 - STU - OLDSKOOL.sndh */
TIMEDB_ENTRY( 4dbf1a4b,  0,   3787,          YM), /* 03928 - Unknown - Sound 2.sndh */
TIMEDB_ENTRY( 4dc1767c,  0,    252,          YM), /* 01912 - Marcer - Trumorre.sndh */
TIMEDB_ENTRY( 4dc92979,  0,   5016,          YM), /* 03885 - Unknown - Blueberry Hill.sndh */
TIMEDB_ENTRY( 4dca576f,  0,   3072,          YM), /* 04828 - T1000 - Bits 55a.sndh */
TIMEDB_ENTRY( 4dcd8678,  0,   4029,          YM), /* 01533 - Iso - Broodje Shoarma.sndh */
TIMEDB_ENTRY( 4de0a2ec,  0, 105900,       YM+TA), /* 03725 - Spaz - Starball.sndh */
TIMEDB_ENTRY( 4e014863,  0,  11520,    YM+TB+TD), /* 03459 - Marcer - Blues.sndh */
TIMEDB_ENTRY( 4e068890,  0,  82938,       YM+TA), /* 04039 - Roggie - Roggie the baked bean on toast.sndh */
TIMEDB_ENTRY( 4e167adc,  0,  18432,          YM), /* 03249 - T1000 - The Bits Club 24.sndh */
TIMEDB_ENTRY( 4e209a72,  0,   9600,          YM), /* 03247 - T1000 - The Beast Club - TBC 19.sndh */
TIMEDB_ENTRY( 4e397598,  0,   6912,          YM), /* 03250 - T1000 - The Bits Club 25.sndh */
TIMEDB_ENTRY( 4e3fefd5,  0,   4219,          YM), /* 04653 - Beast - TSC - Unknown Blipblopper.sndh */
TIMEDB_ENTRY( 4e4db648,  0,    262,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  1,      2,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  2,    106,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  3,    220,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  4,    147,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  5,    160,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  6,    200,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  7,    226,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  8,    164,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648,  9,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 10,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 11,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 12,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 13,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 14,     31,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 15,     31,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 16,     31,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 17,     26,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 18,     54,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 19,   1344,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 20,   1280,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e4db648, 21,    826,          YM), /* 04682 - Ronan - Kass Kong.sndh */
TIMEDB_ENTRY( 4e58d311,  0,  11088,          YM), /* 01059 - Jonathan Dunn - Hook.sndh */
TIMEDB_ENTRY( 4e5947e5,  0,    831,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  1,    253,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  2,    247,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  3,    151,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  4,    152,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  5,    151,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  6,    333,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  7,    781,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  8,    239,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5,  9,    240,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 10,    310,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 11,    314,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 12,    273,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 13,    293,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 14,   1049,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 17,    195,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 18,    206,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 19,    195,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 20,    197,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 28,    330,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 29,    314,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 30,    545,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 31,    270,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 32,    450,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 33,    282,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 34,    273,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 35,    561,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 36,    678,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 37,    186,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 38,    186,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 39,    177,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 40,    167,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 41,    275,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 42,    278,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 43,    907,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 44,    176,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 45,    186,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 46,   1120,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 47,    323,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 48,    499,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 49,    200,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 50,    598,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 51,    266,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 52,    327,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 53,    691,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 54,    268,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 55,    171,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 56,    267,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 57,    171,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 58,    216,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 59,    199,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 60,    216,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 61,    200,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5947e5, 62,    270,          YM), /* 04725 - Unknown - Play and Read - Learn to Read with Prof.sndh */
TIMEDB_ENTRY( 4e5a4612,  0,   1152,       YM+TD), /* 02643 - Miggy - Sisco Sun (Unfinished).sndh */
TIMEDB_ENTRY( 4e946cbe,  0,  41855, YM+TA+TB+TD), /* 03651 - Rebb - French realtime.sndh */
TIMEDB_ENTRY( 4e982fc8,  0,    533,          YM), /* 04691 - Roderick Martin - Lunacy.sndh */
TIMEDB_ENTRY( 4e982fc8,  1,    533,          YM), /* 04691 - Roderick Martin - Lunacy.sndh */
TIMEDB_ENTRY( 4e982fc8,  2,    149,          YM), /* 04691 - Roderick Martin - Lunacy.sndh */
TIMEDB_ENTRY( 4e982fc8,  3,   1492,          YM), /* 04691 - Roderick Martin - Lunacy.sndh */
TIMEDB_ENTRY( 4eaae9f6,  0,   2817,          YM), /* 03896 - Unknown - Volkslied.sndh */
TIMEDB_ENTRY( 4ecaf12f,  0,   4147,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  1,   1015,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  2,     60,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  3,   1506,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  4,   1543,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  5,   2281,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  6,    813,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  7,     96,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  8,   4147,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f,  9,    775,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecaf12f, 10,   1686,          YM), /* 04137 - Arthur F Vosseler - Euroquiz.sndh */
TIMEDB_ENTRY( 4ecd1790,  0,   7167,    YM+TA+TB), /* 03435 - 4mat & malodix - Despair.sndh */
TIMEDB_ENTRY( 4ed01492,  0,  17408, YM+TA+TB+TD), /* 02091 - Mad Max - One Man and his Droid - Sid.sndh */
TIMEDB_ENTRY( 4eefb3ae,  0,   8448,          YM), /* 02179 - Mad Max - Atomino.sndh */
TIMEDB_ENTRY( 4eefb3ae,  1,  11904,          YM), /* 02179 - Mad Max - Atomino.sndh */
TIMEDB_ENTRY( 4eefb3ae,  2,   5376,          YM), /* 02179 - Mad Max - Atomino.sndh */
TIMEDB_ENTRY( 4eefb3ae,  3,   4608,          YM), /* 02179 - Mad Max - Atomino.sndh */
TIMEDB_ENTRY( 4f21bcd9,  0,  13376,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f21bcd9,  1,   2304,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f21bcd9,  2,   2304,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f21bcd9,  3,   5376,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f21bcd9,  4,   2688,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f21bcd9,  5,    247,          YM), /* 00150 - David Whittaker - Verminator.sndh */
TIMEDB_ENTRY( 4f34e37d,  0,   3300,          YM), /* 01290 - Matt Furniss - Hill Street Blues.sndh */
TIMEDB_ENTRY( 4f4086e1,  0,   7708,          YM), /* 00439 - Big Alec - Structure.sndh */
TIMEDB_ENTRY( 4f440b64,  0,  10800,          YM), /* 02036 - Mad Max - Hunter Patrol.sndh */
TIMEDB_ENTRY( 4f4fae18,  0,    556,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  1,    514,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  2,    202,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  3,   3183,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  4,   1152,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  5,    768,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  6,    498,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  7,     97,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  8,    103,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18,  9,    398,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18, 10,    275,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18, 11,    371,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18, 12,    159,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f4fae18, 13,    559,          YM), /* 04718 - William Anderson - The Galactic Miner.sndh */
TIMEDB_ENTRY( 4f68fcdf,  0,   3146,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  1,   1169,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  2,    270,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  3,    334,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  4,    196,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  5,    981,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  6,   1626,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  7,    344,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  8,    248,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f68fcdf,  9,    154,          YM), /* 01566 - Jason Page - Rainbow Islands.sndh */
TIMEDB_ENTRY( 4f6da7eb,  0,   1919,    YM+TA+TB), /* 03431 - Per Almered (XiA) 2009 - Hit the Road Jack-tutorial.sndh */
TIMEDB_ENTRY( 4f73e7bf,  0,  96207,       YM+TA), /* 03705 - Sparky - Fistfull of Dollars (Ennio Morricone).sndh */
TIMEDB_ENTRY( 4f7a867c,  0,  12288,          YM), /* 01028 - Doclands - Lightning Demo - Plasma.sndh */
TIMEDB_ENTRY( 4fa9f8d4,  0,   5196,          YM), /* 04388 - Paul Summers - Matrix Marauders.sndh */
TIMEDB_ENTRY( 4fa9f8d4,  1,   6140,          YM), /* 04388 - Paul Summers - Matrix Marauders.sndh */
TIMEDB_ENTRY( 4fa9f8d4,  2,   4800,          YM), /* 04388 - Paul Summers - Matrix Marauders.sndh */
TIMEDB_ENTRY( 4fad817d,  0,   5120,       YM+TA), /* 04592 - Crawdaddy - Jammers.sndh */
TIMEDB_ENTRY( 4fc24836,  0,   6138, YM+TA+TB+TD), /* 00989 - Dma-Sc-Sector1^YM-Rockerz - The First Blip-Blop Noel.sndh */
TIMEDB_ENTRY( 4fc33b4e,  0,    383,    YM+TA+TB), /* 03295 - Dma-Sc - ATIPYK Developments - Manical Drop - Menu.sndh */
TIMEDB_ENTRY( 4fc8ab31,  0,  39199,          YM), /* 03999 - Christophe Fevre - Blues Brothers.sndh */
TIMEDB_ENTRY( 4fc8ab31,  1,  27215,          YM), /* 03999 - Christophe Fevre - Blues Brothers.sndh */
TIMEDB_ENTRY( 4fc8ab31,  2,  19507,          YM), /* 03999 - Christophe Fevre - Blues Brothers.sndh */
TIMEDB_ENTRY( 4fc8ab31,  3,  34285,          YM), /* 03999 - Christophe Fevre - Blues Brothers.sndh */
TIMEDB_ENTRY( 4fd604b3,  0,  24870,       YM+TA), /* 04677 - Unknown - Demo 2.sndh */
TIMEDB_ENTRY( 4fdae614,  0,  12290,          YM), /* 01897 - LTK - Mage Mix - Phaleon - Mystic.sndh */
TIMEDB_ENTRY( 4ff09498,  0,  29184,          YM), /* 01465 - STU - FLOWER.sndh */
TIMEDB_ENTRY( 4ff0e760,  0,   1728,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  1,   1536,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  2,   2560,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  3,   1152,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  4,   2304,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  5,    512,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  6,    575,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  7,    384,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  8,    238,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760,  9,   2304,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 10,    192,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 11,   1920,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 12,   1728,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 13,    768,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 14,   1536,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 15,   2304,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff0e760, 16,    960,          YM), /* 00348 - Tony Williams - Pang.sndh */
TIMEDB_ENTRY( 4ff685cc,  0,   2048,          YM), /* 00799 - Crawdaddy - Looking For Peace.sndh */
TIMEDB_ENTRY( 4ff6dc2c,  0,   8442, YM+TA+TB+TD), /* 02831 - Msg - Superfly tune 8.sndh */
TIMEDB_ENTRY( 5000908c,  0,   5795, YM+TA+TB+TD), /* 03634 - Timbaland - Kisse Katten.sndh */
TIMEDB_ENTRY( 503196f8,  0,     85,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  1,    133,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  2,    517,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  3,     63,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  4,     27,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  5,     56,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  6,     42,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  7,     42,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  8,     79,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8,  9,     67,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503196f8, 10,   4952,          YM), /* 00851 - Ben Daglish - Footballer of the Year 2.sndh */
TIMEDB_ENTRY( 503d6010,  0,  10944,          YM), /* 02023 - Mad Max - Chimera.sndh */
TIMEDB_ENTRY( 503d6010,  1,   6144,          YM), /* 02023 - Mad Max - Chimera.sndh */
TIMEDB_ENTRY( 504c6e55,  0,  54981,          YM), /* 02111 - Mad Max - Knucklebusters.sndh */
TIMEDB_ENTRY( 5052dac8,  0,  22652, YM+TA+TB+TD), /* 02596 - STU - INEXISTENT.sndh */
TIMEDB_ENTRY( 50601a0e,  0,  15360,          YM), /* 01818 - Paul Loraine - Scooby Doo and Scrappy Doo.sndh */
TIMEDB_ENTRY( 50601a0e,  1,    897,          YM), /* 01818 - Paul Loraine - Scooby Doo and Scrappy Doo.sndh */
TIMEDB_ENTRY( 50601a0e,  2,   4475,          YM), /* 01818 - Paul Loraine - Scooby Doo and Scrappy Doo.sndh */
TIMEDB_ENTRY( 50601a0e,  3,   2047,          YM), /* 01818 - Paul Loraine - Scooby Doo and Scrappy Doo.sndh */
TIMEDB_ENTRY( 50654693,  0,  26240, YM+TA+TB+TD), /* 02909 - Helene - tecnotrance.sndh */
TIMEDB_ENTRY( 506e5380,  0,   4475, YM+TA+TB+TD), /* 01244 - Evil - Arpa.sndh */
TIMEDB_ENTRY( 50725811,  0,   6563, YM+TA+TB+TD), /* 02866 - 505-checkpoint^paradox - YM2006 (STE only).sndh */
TIMEDB_ENTRY( 507729d8,  0,   4032,          YM), /* 04004 - Celtic - Gnurk Snor Bobs.sndh */
TIMEDB_ENTRY( 50981e0d,  0,   7680,          YM), /* 01455 - Stax - Wutz.sndh */
TIMEDB_ENTRY( 5099c5e2,  0,   8188, YM+TA+TB+TD), /* 00008 - gwEm - Tubular Crusade.sndh */
TIMEDB_ENTRY( 50f9db32,  0,  13887,          YM), /* 02362 - Nexus 6 - Axmnpfnmumpf.sndh */
TIMEDB_ENTRY( 50fc9eb4,  0,   4566,    YM+TA+TD), /* 00034 - 505 - Bla.sndh */
TIMEDB_ENTRY( 5105bc1e,  0,   9520,    YM+TA+TD), /* 02438 - bob monkhouze - phoneomenon.sndh */
TIMEDB_ENTRY( 5109cc6d,  0,  14783, YM+TA+TB+TD), /* 02693 - Dubmood - Razor 1911 - 4Mats AnarchyMenu #3.sndh */
TIMEDB_ENTRY( 511553f6,  0,   6336,          YM), /* 00768 - Count Zero - Decade Demo - Parascroll.sndh */
TIMEDB_ENTRY( 511e2226,  0,   9600,          YM), /* 01508 - The Fate - 3615 Gen4 Demo (Thrust Remix).sndh */
TIMEDB_ENTRY( 513500c1,  0,   8666, YM+TA+TB+TD), /* 00234 - AIO - Revive Me.sndh */
TIMEDB_ENTRY( 513618ca,  0,  20730,       YM+TA), /* 01446 - Stax - Revotus - Digi.sndh */
TIMEDB_ENTRY( 5140d30f,  0,   5184,          YM), /* 03656 - Rhino - Xmas 2011.sndh */
TIMEDB_ENTRY( 5158f097,  0,  20065,       YM+TA), /* 04019 - Roggie - Phobia.sndh */
TIMEDB_ENTRY( 515e600b,  0,  28801,          YM), /* 02038 - Mad Max - International Karate.sndh */
TIMEDB_ENTRY( 516d61de,  0,    690,          YM), /* 04526 - Ewald Greiml - Eierkopf.sndh */
TIMEDB_ENTRY( 516d61de,  1,    133,          YM), /* 04526 - Ewald Greiml - Eierkopf.sndh */
TIMEDB_ENTRY( 516d61de,  2,   2364,          YM), /* 04526 - Ewald Greiml - Eierkopf.sndh */
TIMEDB_ENTRY( 51704069,  0,  17088,          YM), /* 02901 - Techno - Graaal Demo.sndh */
TIMEDB_ENTRY( 5177133b,  0,   4096,       YM+TA), /* 04591 - Technoblip - Bom Tsh Bap Bap.sndh */
TIMEDB_ENTRY( 5180476a,  0,  16705,       YM+TA), /* 03878 - Mad Max - Sanxion Loader.sndh */
TIMEDB_ENTRY( 51841b4d,  0,   3855,          YM), /* 04453 - Kafka - Beginning of the Ending Demo.sndh */
TIMEDB_ENTRY( 519ca005,  0,   6189,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  1,   1781,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  2,    230,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  3,   3491,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  4,   2509,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  5,   2337,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  6,   4660,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  7,    988,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 519ca005,  8,   1572,          YM), /* 00843 - Ben Daglish - California Games.sndh */
TIMEDB_ENTRY( 51aee08e,  0,  18816,          YM), /* 03248 - T1000 - The Bits Club 22.sndh */
TIMEDB_ENTRY( 51b18088,  0,   4559,          YM), /* 03119 - Unknown - Han Solo and the Princess.sndh */
TIMEDB_ENTRY( 51b339b1,  0,   3256,          YM), /* 02879 - Beast - Shitt.sndh */
TIMEDB_ENTRY( 51bc0c87,  0,  73137,       YM+TA), /* 04856 - Michel Winogradoff - Turbo Cup.sndh */
TIMEDB_ENTRY( 51bdedb7,  0,   3834,          YM), /* 04253 - 505 - Checkpoint - duckwalk.sndh */
TIMEDB_ENTRY( 51cc20a7,  0,  17080,          YM), /* 02030 - Mad Max - Flash Gordon.sndh */
TIMEDB_ENTRY( 51cd3d99,  0,   2212,          YM), /* 04194 - Unknown - Postman Pat 3.sndh */
TIMEDB_ENTRY( 51d0e0ed,  0,   4991,          YM), /* 00311 - Gary Wheaton - Jimi.sndh */
TIMEDB_ENTRY( 51e0d07b,  0,   4608,          YM), /* 02448 - David Whittaker - First Person Pinball.sndh */
TIMEDB_ENTRY( 51e0d07b,  1,   6144,          YM), /* 02448 - David Whittaker - First Person Pinball.sndh */
TIMEDB_ENTRY( 51e0d07b,  2,  10053,          YM), /* 02448 - David Whittaker - First Person Pinball.sndh */
TIMEDB_ENTRY( 51e0d07b,  3,   2389,          YM), /* 02448 - David Whittaker - First Person Pinball.sndh */
TIMEDB_ENTRY( 51eaa7ba,  0,   9853,          YM), /* 01902 - Luebke - Paracon 3 invitro.sndh */
TIMEDB_ENTRY( 51ef485a,  0,   8209,          YM), /* 01041 - Donald Fakk - Zzzlajm.sndh */
TIMEDB_ENTRY( 51fb2e63,  0,  23044,       YM+TA), /* 01741 - Jovis - Lametrop - 3D Bobs.sndh */
TIMEDB_ENTRY( 520064ac,  0,   4992,          YM), /* 04846 - t1000 - Bits 49.sndh */
TIMEDB_ENTRY( 5216af14,  0,   7165,          YM), /* 01118 - Paul Tonge - B17 Flying Fotress.sndh */
TIMEDB_ENTRY( 5220974f,  0,   6719,       YM+TA), /* 04663 - Dave Lowe - Starglider 2.sndh */
TIMEDB_ENTRY( 522156f8,  0,   9472,          YM), /* 02736 - Tony Williams - Chuckie Egg 2.sndh */
TIMEDB_ENTRY( 522156f8,  1,     97,          YM), /* 02736 - Tony Williams - Chuckie Egg 2.sndh */
TIMEDB_ENTRY( 52631c92,  0,   6656,          YM), /* 00255 - Antichrist - Low Tune.sndh */
TIMEDB_ENTRY( 5281ae1a,  0,   5085,          YM), /* 00246 - AN Cool - A.N. Bass.sndh */
TIMEDB_ENTRY( 5284ae81,  0,   5120,          YM), /* 03196 - 4-Mat - Arcade Classix - Hiscore.sndh */
TIMEDB_ENTRY( 52a8e499,  0,  10752,    YM+TA+TD), /* 02271 - Tao - Cream - The Dancer (Steps).sndh */
TIMEDB_ENTRY( 52b0c50f,  0,   8639,          YM), /* 00507 - Tangens - Synthetic Visions - Pulsing Balls.sndh */
TIMEDB_ENTRY( 52ca7216,  0,   7420, YM+TA+TB+TD), /* 03650 - 505-Checkpoint - Revision 2012.sndh */
TIMEDB_ENTRY( 52dc9635,  0,  32320, YM+TA+TB+TD), /* 02527 - Salkin-047 - Ko ko Kossan.sndh */
TIMEDB_ENTRY( 52dd204c,  0,   5861,       YM+TA), /* 04415 - EVE - Makelot.sndh */
TIMEDB_ENTRY( 52defff8,  0,    535,          YM), /* 04156 - Unknown - Pompom.sndh */
TIMEDB_ENTRY( 52e68367,  0,   1944,       YM+TA), /* 02633 - Miggy - DevilVer1(Unfinished).sndh */
TIMEDB_ENTRY( 52ee21bc,  0,   4842,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  1,    403,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  2,   1382,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  3,    495,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  4,   1355,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  5,    604,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  6,    453,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  7,   1322,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  8,   1489,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc,  9,   1923,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 10,   1388,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 11,    857,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 12,    816,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 13,    446,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 14,   1360,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 15,    402,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 16,    583,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 17,    944,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 18,   2034,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 52ee21bc, 19,   1796,          YM), /* 03977 - David Schroeder - War in Middle Earth.sndh */
TIMEDB_ENTRY( 53268dac,  0,   5369,       YM+TA), /* 02599 - STU - INTRO.sndh */
TIMEDB_ENTRY( 5330a665,  0,    576,          YM), /* 03411 - Phil Bak - Yolanda.sndh */
TIMEDB_ENTRY( 5330a665,  2,   7296,          YM), /* 03411 - Phil Bak - Yolanda.sndh */
TIMEDB_ENTRY( 5330a665,  3,   3131,          YM), /* 03411 - Phil Bak - Yolanda.sndh */
TIMEDB_ENTRY( 5330a665,  4,   1920,          YM), /* 03411 - Phil Bak - Yolanda.sndh */
TIMEDB_ENTRY( 5354a14f,  0,  11265,       YM+TA), /* 01035 - Donald Fakk - Digi Dick.sndh */
TIMEDB_ENTRY( 535d8acd,  0,   4701,          YM), /* 00249 - AN Cool - Zuzzexx!.sndh */
TIMEDB_ENTRY( 5389443c,  0,   4986,       YM+TA), /* 04232 - 505 - 10th.sndh */
TIMEDB_ENTRY( 538cf8c3,  0,   4416,          YM), /* 01032 - Dolby - Piece Of Shit.sndh */
TIMEDB_ENTRY( 53935695,  0,    768,          YM), /* 00978 - Chris Denman - International Championship Athletics.sndh */
TIMEDB_ENTRY( 53b03213,  0,  12898, YM+TA+TB+TD), /* 01854 - Lotek Style - Eiswelt.sndh */
TIMEDB_ENTRY( 53b505bc,  0,  20064,       YM+TA), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  1,   9213,       YM+TA), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  2,   8658,       YM+TA), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  3,  47324,       YM+TA), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  4,   7296,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  5,  24589,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  6,   8544,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  7,   3456,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  8,    240,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53b505bc,  9,  11392,          YM), /* 02184 - Mad Max (Jochen Hippel) - Chambers of Shaolin.sndh */
TIMEDB_ENTRY( 53be0a5b,  0,  43392,       YM+TD), /* 02582 - STU - CHIPRAPE.sndh */
TIMEDB_ENTRY( 53cea771,  0,   6272,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53cea771,  2,   7168,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53cea771,  3,  11520,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53cea771,  4,  12096,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53cea771,  5,   3840,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53cea771,  6,  15744,          YM), /* 03700 - Chris Mad - Overdrive Demo - Music Box.sndh */
TIMEDB_ENTRY( 53e01684,  0,   4025,    YM+TB+TD), /* 04286 - 505-cpt^pdx - Melting.sndh */
TIMEDB_ENTRY( 53e105e5,  0,   3286,          YM), /* 03925 - Unknown - Musik 002.sndh */
TIMEDB_ENTRY( 53ee72fb,  0,   5772,          YM), /* 04339 - Nexus 6 - Moskito Attack.sndh */
TIMEDB_ENTRY( 53f1f6e8,  0,   6144,          YM), /* 04406 - Dr Beat - Galtan 6 - Elite Intro 3.sndh */
TIMEDB_ENTRY( 540cb5f2,  0,  11520,    YM+TA+TD), /* 02256 - Tao - Cream - For Abyss (Steps).sndh */
TIMEDB_ENTRY( 540da2c3,  0,   9216,          YM), /* 01079 - Excellence In Art - DHS Intro.sndh */
TIMEDB_ENTRY( 5417b380,  0,   4037,          YM), /* 00045 - 505 - Clowns.sndh */
TIMEDB_ENTRY( 5421b25b,  0,   5760,          YM), /* 02485 - gwEm - Acidic Tears.sndh */
TIMEDB_ENTRY( 543e59ec,  0,   2386,          YM), /* 03866 - 505 - Chasey Intro.sndh */
TIMEDB_ENTRY( 544250d8,  0,   1275,          YM), /* 00467 - Augustin Vidovic - X-Out.sndh */
TIMEDB_ENTRY( 54425228,  0,   1565,          YM), /* 02465 - Iso - It's Happening.sndh */
TIMEDB_ENTRY( 544381a1,  0,   9783,          YM), /* 02140 - Mad Max - Commando - LSD.sndh */
TIMEDB_ENTRY( 5444fa7d,  0,   4025,    YM+TA+TD), /* 01157 - Tyan - Serious 2.sndh */
TIMEDB_ENTRY( 544ba2ae,  0,   3764,          YM), /* 04507 - Francois Lionet - Bullet Train.sndh */
TIMEDB_ENTRY( 5453fbd2,  0,   4929,          YM), /* 00888 - Adrian Dalecki - Rollin Ronny B.sndh */
TIMEDB_ENTRY( 545e3d52,  0,   6715,          YM), /* 00547 - Allister Brimble - Bignose The Caveman.sndh */
TIMEDB_ENTRY( 545e3d52,  1,    379,          YM), /* 00547 - Allister Brimble - Bignose The Caveman.sndh */
TIMEDB_ENTRY( 545e3d52,  2,   1531,          YM), /* 00547 - Allister Brimble - Bignose The Caveman.sndh */
TIMEDB_ENTRY( 545e3d52,  3,    119,          YM), /* 00547 - Allister Brimble - Bignose The Caveman.sndh */
TIMEDB_ENTRY( 546447a2,  0,   6559,          YM), /* 02417 - Nexus 6 - Zuck-Lohn.sndh */
TIMEDB_ENTRY( 546447a2,  1,   7008,          YM), /* 02417 - Nexus 6 - Zuck-Lohn.sndh */
TIMEDB_ENTRY( 546447a2,  2,   7839,          YM), /* 02417 - Nexus 6 - Zuck-Lohn.sndh */
TIMEDB_ENTRY( 546447a2,  3,    270,          YM), /* 02417 - Nexus 6 - Zuck-Lohn.sndh */
TIMEDB_ENTRY( 546447a2,  4,   7616,          YM), /* 02417 - Nexus 6 - Zuck-Lohn.sndh */
TIMEDB_ENTRY( 546844bc,  0,    980,          YM), /* 02765 - Bobby Prince [ST by Dma-Sc] - High Score.sndh */
TIMEDB_ENTRY( 546d4c97,  0, 166262,       YM+TA), /* 04051 - Roggie - Ameobas rule ok.sndh */
TIMEDB_ENTRY( 546e3caa,  0,   4752,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  1,   4416,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  2,   3456,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  3,   6240,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  4,   3258,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  6,   1920,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546e3caa,  7,   2688,          YM), /* 02920 - Richard Joseph - Wizkid.sndh */
TIMEDB_ENTRY( 546f51a2,  0,  10752,       YM+TA), /* 04118 - Tao - The Dancer - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 5473065c,  0,   1152,       YM+TA), /* 01367 - Scavenger - What the Fuck?.sndh */
TIMEDB_ENTRY( 54750c2b,  0,   3264,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  1,   4946,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  2,   7360,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  3,  10752,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  4,  14400,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  5,   3552,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  6,   4160,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  7,   2304,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  8,   3920,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b,  9,   3312,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 10,   3168,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 11,    672,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 12,    408,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 13,    480,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 14,    768,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 15,    768,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 16,   1176,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 17,    792,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 18,    384,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 19,    414,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 20,    768,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 21,    576,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 22,   1008,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 23,    336,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 24,    576,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 25,    384,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 26,    888,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 27,   1056,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 28,   1440,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 29,    540,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 30,    792,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 31,    768,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 32,    588,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 33,    600,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 54750c2b, 34,   1560,          YM), /* 04613 - Dave Lowe - The Games - Summer Edition.sndh */
TIMEDB_ENTRY( 5477020a,  0,  14602,          YM), /* 02063 - Mad Max - Zoolook.sndh */
TIMEDB_ENTRY( 54865e57,  0,  12078,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  1,  10032,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  2,   7922,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  3,   8616,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  4,   8906,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  5,   9868,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 54865e57,  6,  13183,       YM+TA), /* 02970 - WE Music - Captain Planet.sndh */
TIMEDB_ENTRY( 548cd2c7,  0,  12285,       YM+TA), /* 04870 - Jovis - Subir.sndh */
TIMEDB_ENTRY( 54a98c05,  0,   6396,       YM+TA), /* 04236 - 505 - 20th.sndh */
TIMEDB_ENTRY( 54c2c110,  0,   4218,       YM+TA), /* 01333 - Scavenger - HAL Ninethousand.sndh */
TIMEDB_ENTRY( 54c8a2f3,  0,   3450,       YM+TD), /* 01617 - No More - GSC 43 Sid.sndh */
TIMEDB_ENTRY( 54d91ce6,  0,   3450,          YM), /* 04371 - Jess - Routine 3.sndh */
TIMEDB_ENTRY( 54d99d64,  0,   3457,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  1,    161,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  2,    185,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  3,    185,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  4,    249,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  5,    225,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  6,    217,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54d99d64,  7,     27,          YM), /* 03284 - Unknown - Pub Trivia Simulator.sndh */
TIMEDB_ENTRY( 54e12509,  0,   6912,          YM), /* 00478 - Ultimatum - Choux J.sndh */
TIMEDB_ENTRY( 55025745,  0,  10368,          YM), /* 00413 - Big Alec - Crystal Clear (Mega Leif).sndh */
TIMEDB_ENTRY( 55080727,  0,  13440,    YM+TA+TD), /* 02267 - Tao - Cream - Steps (Steps).sndh */
TIMEDB_ENTRY( 5511a9fd,  0,  33792, YM+TA+TB+TD), /* 02559 - Marcer - DISKO SID.sndh */
TIMEDB_ENTRY( 552e100a,  0,  55296, YM+TA+TB+TD), /* 04614 - Marcer - BOOBS.sndh */
TIMEDB_ENTRY( 5532f99e,  0,  26062,          YM), /* 03633 - T1000 - Bits 54.sndh */
TIMEDB_ENTRY( 5536f3cf,  0,   2881,          YM), /* 00430 - Big Alec - Poppycock - Intro.sndh */
TIMEDB_ENTRY( 5539a3ec,  0,  14069,       YM+TA), /* 04230 - 505 - 8th.sndh */
TIMEDB_ENTRY( 553acdb6,  0,  23884,          YM), /* 03870 - Gary Wheaton - Cyberstorm Title.sndh */
TIMEDB_ENTRY( 555306f5,  0,   1530,    YM+TA+TB), /* 04298 - 505-checkpoint - Pestwurz.sndh */
TIMEDB_ENTRY( 55593e48,  0,   9216,    YM+TA+TD), /* 02325 - Tao - Cream - MV Jeu - Game (Steps).sndh */
TIMEDB_ENTRY( 557a9fc7,  0,  21945,       YM+TA), /* 01690 - Jess - Hardrock Ballad.sndh */
TIMEDB_ENTRY( 558740e6,  0,   4864,       YM+TA), /* 00779 - Crawdaddy - Bombsquad.sndh */
TIMEDB_ENTRY( 558a84e1,  0,   2755,          YM), /* 03061 - Ed Bogas - Tango.sndh */
TIMEDB_ENTRY( 5598ea22,  0,   2684,          YM), /* 04632 - Epic - Zik 11.sndh */
TIMEDB_ENTRY( 559c530e,  0,   5509, YM+TA+TB+TD), /* 04318 - 505-checkpoint (KB) - tammo (KBcover).sndh */
TIMEDB_ENTRY( 559f3f27,  0,   8987,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  1,  18773,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  2,  21458,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  3,  10666,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  4,  13241,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  5,  21333,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  6,  16340,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  7,  23184,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  8,  18285,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27,  9,  17808,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27, 10,  18561,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 559f3f27, 11,  24633,          YM), /* 03535 - Jean Baudlot - Operation Stealth.sndh */
TIMEDB_ENTRY( 55a354e5,  0,   3836,          YM), /* 01239 - Epic - TMS Dreams.sndh */
TIMEDB_ENTRY( 55a71ffe,  0,  18230,       YM+TA), /* 03791 - Roggie - Minuet 1.sndh */
TIMEDB_ENTRY( 55b124a8,  0,   7161, YM+TA+TB+TD), /* 01953 - MSG - Double Juggle Vegatabobble.sndh */
TIMEDB_ENTRY( 55b422c0,  0,  30720, YM+TA+TB+TD), /* 03637 - Starbuck - Dawn.sndh */
TIMEDB_ENTRY( 55d70968,  0,  18940, YM+TA+TB+TD), /* 03504 - 505 - Suretrip 2 - Main 2.sndh */
TIMEDB_ENTRY( 55df043e,  0,   3199,          YM), /* 01552 - Rhino - Twitching Flannels.sndh */
TIMEDB_ENTRY( 55f2952a,  0,   6948,       YM+TA), /* 01609 - No More - Amazine 4.sndh */
TIMEDB_ENTRY( 5623d508,  0,   9240,          YM), /* 03182 - n-a - WE ARE THE WORLD.sndh */
TIMEDB_ENTRY( 5628a985,  0,   9994,       YM+TA), /* 01859 - Lotek Style - Mourning.sndh */
TIMEDB_ENTRY( 564e0f84,  0,  14819,       YM+TA), /* 01167 - Tyan - Xtc.sndh */
TIMEDB_ENTRY( 566094fa,  0,   6577,          YM), /* 00163 - Matt Bates - Deuteros.sndh */
TIMEDB_ENTRY( 566094fa,  1,   1885,          YM), /* 00163 - Matt Bates - Deuteros.sndh */
TIMEDB_ENTRY( 566094fa,  2,   4315,          YM), /* 00163 - Matt Bates - Deuteros.sndh */
TIMEDB_ENTRY( 566094fa,  3,   2491,          YM), /* 00163 - Matt Bates - Deuteros.sndh */
TIMEDB_ENTRY( 567399c8,  0,   5042,          YM), /* 03074 - Peter Inzerillo - Piano Man.sndh */
TIMEDB_ENTRY( 5678b9e4,  0,   4609,          YM), /* 04145 - Unknown - Andante 6.sndh */
TIMEDB_ENTRY( 56844f54,  0,   4795,          YM), /* 03046 - Ed Bogas - Fifties.sndh */
TIMEDB_ENTRY( 56933435,  0,  23040,          YM), /* 02887 - Dr Kaya - No Limit Coding.sndh */
TIMEDB_ENTRY( 569634ec,  0,    432,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569634ec,  1,    504,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569634ec,  2,    103,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569634ec,  3,    180,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569634ec,  4,    524,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569634ec,  5,   2304,          YM), /* 00173 - Martin Walker - Teenage Mutant Hero Turtles.sndh */
TIMEDB_ENTRY( 569b83dd,  0,   1914,          YM), /* 01697 - Jess - For your Loader #3.sndh */
TIMEDB_ENTRY( 56a669f1,  0,   6835,          YM), /* 03097 - n-a - .sndh */
TIMEDB_ENTRY( 56c31160,  0,   5138,       YM+TA), /* 04630 - Mephistow - Sound.sndh */
TIMEDB_ENTRY( 56e307f1,  0,   9786,          YM), /* 01033 - Dolby - A New Great Sound.sndh */
TIMEDB_ENTRY( 56eb9f9b,  0,   5079,          YM), /* 03093 - n-a - .sndh */
TIMEDB_ENTRY( 56ec7baa,  0,  21492,       YM+TA), /* 04364 - Jess - Middle-Earth Theme (digit).sndh */
TIMEDB_ENTRY( 56f627a5,  0,   5984,          YM), /* 03693 - Yerzmyey - Get Mad.sndh */
TIMEDB_ENTRY( 56f878ff,  0,   6860,          YM), /* 00231 - ACC - Popcorn.sndh */
TIMEDB_ENTRY( 56fc33b0,  0,   7142,          YM), /* 01500 - Ashley Hogg - CJ In The USA.sndh */
TIMEDB_ENTRY( 56fc33b0,  1,   1002,          YM), /* 01500 - Ashley Hogg - CJ In The USA.sndh */
TIMEDB_ENTRY( 56fc33b0,  2,    374,          YM), /* 01500 - Ashley Hogg - CJ In The USA.sndh */
TIMEDB_ENTRY( 56fc33b0,  3,   1725,          YM), /* 01500 - Ashley Hogg - CJ In The USA.sndh */
TIMEDB_ENTRY( 570e1419,  0, 105032,       YM+TA), /* 03816 - Tufty - Dear Boy.sndh */
TIMEDB_ENTRY( 5727be83,  0,   6528,          YM), /* 00932 - Dark Angel - Magic Force.sndh */
TIMEDB_ENTRY( 57385c2c,  0,   2175,          YM), /* 02391 - Nexus 6 - Please Wait While Loading 2.sndh */
TIMEDB_ENTRY( 573c531c,  0,   8192,          YM), /* 04221 - Paul Shields - Crossland.sndh */
TIMEDB_ENTRY( 573c531c,  1,  23680,          YM), /* 04221 - Paul Shields - Crossland.sndh */
TIMEDB_ENTRY( 57570403,  0,   9216,          YM), /* 01176 - Sally - Sally 2.sndh */
TIMEDB_ENTRY( 576960b3,  0, 117647, YM+TA+TB+TD), /* 04217 - xyce - n-a.sndh */
TIMEDB_ENTRY( 5775383b,  0,   6169,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  1,   3841,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  2,   2329,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  3,   3097,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  4,   7640,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  5,   2941,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  6,   2713,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  7,   2301,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  8,   1897,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b,  9,   4641,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 10,   5153,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 11,   5153,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 12,   4521,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 13,    661,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 14,   2281,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 15,    194,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 16,    218,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 17,    306,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 18,    429,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 19,    266,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5775383b, 20,    354,          YM), /* 02884 - Wally Beben - Ultima V - Warriors of Destiny.sndh */
TIMEDB_ENTRY( 5778b48b,  0,   4845,          YM), /* 03508 - Unknown - Mousetrap.sndh */
TIMEDB_ENTRY( 579315cd,  0,   8448,          YM), /* 03212 - Crazy Q - DHS - Little sexy square.sndh */
TIMEDB_ENTRY( 57bde7bc,  0,  13440,          YM), /* 00484 - Ultimatum - Count Zero 6.sndh */
TIMEDB_ENTRY( 57c18f04,  0,    889,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  1,   4479,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  2,   3070,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  3,   1024,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  4,     25,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  5,    299,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  6,   1599,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  7,   1280,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57c18f04,  8,    319,          YM), /* 01319 - Matt Furniss - Space Harrier II.sndh */
TIMEDB_ENTRY( 57cd8977,  0,   3839,          YM), /* 00342 - Tony Williams - Narc.sndh */
TIMEDB_ENTRY( 57cf2288,  0,   2048,          YM), /* 01519 - Iso - Barst Allen.sndh */
TIMEDB_ENTRY( 5820dc63,  0,   4473,    YM+TB+TD), /* 00182 - 505 - Paradox Xmas 2003 #2.sndh */
TIMEDB_ENTRY( 5829be49,  0,   2331,    YM+TA+TD), /* 04284 - 505 - Loop09.sndh */
TIMEDB_ENTRY( 582af626,  0,  14592,       YM+TA), /* 03579 - Mad Max - Skate or Die!.sndh */
TIMEDB_ENTRY( 584046c7,  0,   6592,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  1,    960,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  2,    400,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  3,    640,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  4,   1280,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  5,    336,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  6,   1024,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  7,    832,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  8,    960,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7,  9,   1536,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 10,    462,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 11,   1152,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 12,    400,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 13,   1152,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 14,   3264,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 15,   1280,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 584046c7, 16,   3040,          YM), /* 01887 - Dave Lowe - Midwinter 2.sndh */
TIMEDB_ENTRY( 5844547d,  0,  13900,          YM), /* 00373 - Beast - TSC - If Pigs Could Fly Main Theme.sndh */
TIMEDB_ENTRY( 584c6418,  0,   3315,       YM+TD), /* 01108 - Timbral - DHS - Rudolphe (miniremix).sndh */
TIMEDB_ENTRY( 5863bafb,  0,   5629,          YM), /* 00557 - Allister Brimble - Robin Hood - Legend Quest.sndh */
TIMEDB_ENTRY( 5863bafb,  1,    253,          YM), /* 00557 - Allister Brimble - Robin Hood - Legend Quest.sndh */
TIMEDB_ENTRY( 58641f7d,  0,  10176,       YM+TA), /* 00788 - Crawdaddy - Escape From Space.sndh */
TIMEDB_ENTRY( 586670cc,  0,  12432,          YM), /* 01979 - MSG - MSG 10.sndh */
TIMEDB_ENTRY( 5873bea8,  0,  11638,          YM), /* 00248 - AN Cool - TCB Tracker.sndh */
TIMEDB_ENTRY( 58901c2c,  0,  76272, YM+TA+TB+TD), /* 03439 - Tomchi - Noextra - Wake up kid !.sndh */
TIMEDB_ENTRY( 589e514e,  0,  48576,          YM), /* 01557 - Robert Demming - Nostagic-O Demo.sndh */
TIMEDB_ENTRY( 58ad61b3,  0,   6911,          YM), /* 03646 - 505 - Shazz YM6channel 456.sndh */
TIMEDB_ENTRY( 58c89598,  0,   1337,          YM), /* 01679 - Jess - That's Illogical Captain.sndh */
TIMEDB_ENTRY( 58cf67dd,  0, 177988,       YM+TA), /* 03728 - Spector - Manic Demo LED Screen.sndh */
TIMEDB_ENTRY( 58d1dbee,  0,    532,          YM), /* 04698 - Dave Cobbledick - Mindlock.sndh */
TIMEDB_ENTRY( 58d1dbee,  1,   1346,          YM), /* 04698 - Dave Cobbledick - Mindlock.sndh */
TIMEDB_ENTRY( 58d1dbee,  2,    629,          YM), /* 04698 - Dave Cobbledick - Mindlock.sndh */
TIMEDB_ENTRY( 58de88c1,  0,     13,          YM), /* 04689 - A. G. Smith - Lettris.sndh */
TIMEDB_ENTRY( 58de88c1,  1,     51,          YM), /* 04689 - A. G. Smith - Lettris.sndh */
TIMEDB_ENTRY( 58de88c1,  2,     12,          YM), /* 04689 - A. G. Smith - Lettris.sndh */
TIMEDB_ENTRY( 58de88c1,  3,   4105,          YM), /* 04689 - A. G. Smith - Lettris.sndh */
TIMEDB_ENTRY( 58de88c1,  4,   4552,          YM), /* 04689 - A. G. Smith - Lettris.sndh */
TIMEDB_ENTRY( 58e58f33,  0,  46272,          YM), /* 01450 - Stax - Soba.sndh */
TIMEDB_ENTRY( 58e6767d,  0,   1614,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  1,   4703,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  2,   3647,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  3,   3456,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  4,   2880,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  5,    359,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 58e6767d,  6,    144,          YM), /* 01255 - Tim Follin - Bionic Commando.sndh */
TIMEDB_ENTRY( 59197d6a,  0,   1280,          YM), /* 00949 - Dark Angel - X-Out.sndh */
TIMEDB_ENTRY( 592acdc0,  0,   7062,          YM), /* 03043 - Ed Bogas - Pachelbel's Canon in D.sndh */
TIMEDB_ENTRY( 5942c097,  0,   8065,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  1,   2561,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  2,   2561,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  3,    561,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  4,   2561,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  5,    322,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 5942c097,  6,   5121,          YM), /* 00384 - Wally Beben - Hammerfist.sndh */
TIMEDB_ENTRY( 59446b8a,  0,   6533,          YM), /* 04159 - Unknown - I-Ball.sndh */
TIMEDB_ENTRY( 5950513d,  0,  23040, YM+TA+TB+TD), /* 02615 - STU - SANTA STU.sndh */
TIMEDB_ENTRY( 59574b1d,  0,   9071,          YM), /* 02655 - Dma-Sc - POOZ - In-Game.sndh */
TIMEDB_ENTRY( 5958b03d,  0,    484,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  1,   5114,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  2,    246,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  3,    120,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  4,   8574,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  5,   7392,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  6,   6930,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  7,   5664,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 5958b03d,  8,   9679,          YM), /* 01258 - Tim Follin - Ghouls N Ghosts.sndh */
TIMEDB_ENTRY( 595c4446,  0,  13785,          YM), /* 03253 - T1000 - The Bits Club 34.sndh */
TIMEDB_ENTRY( 595cc1b1,  0,    593,          YM), /* 02148 - Mad Max - Mad Max Tune 3 - LSD.sndh */
TIMEDB_ENTRY( 597fad6b,  0,   4224,          YM), /* 01218 - ENS Music - ENS 1.sndh */
TIMEDB_ENTRY( 598024f0,  0,  20128, YM+TA+TB+TD), /* 03853 - 505 & Mch - Mutative.sndh */
TIMEDB_ENTRY( 598b5167,  0,    953,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  1,    148,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  2,    177,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  3,      2,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  4,    935,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  5,     35,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  6,    803,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 598b5167,  7,    450,          YM), /* 04513 - Paul Dowers - Classroom maths.sndh */
TIMEDB_ENTRY( 59c38d29,  0,   9213, YM+TA+TB+TD), /* 01054 - Dubmood - Razor 1911 ( 2003 ) - Haschkaka (YM Rockerz).sndh */
TIMEDB_ENTRY( 59c7d1e4,  0,   9555,          YM), /* 00869 - Ben Daglish - Prison.sndh */
TIMEDB_ENTRY( 59cb6454,  0,   9177,          YM), /* 03173 - n-a - Stealing the Enterprise.sndh */
TIMEDB_ENTRY( 59f55101,  0,  58554,       YM+TA), /* 03714 - Steve Wetherill - Tusker.sndh */
TIMEDB_ENTRY( 59f78bfa,  0,   3561,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 59f78bfa,  1,   3252,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 59f78bfa,  2,   5266,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 59f78bfa,  3,   1755,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 59f78bfa,  4,    952,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 59f78bfa,  5,   2048,          YM), /* 00104 - David Whittaker - Golden Axe.sndh */
TIMEDB_ENTRY( 5a1b0007,  0,  19200, YM+TA+TB+TD), /* 02565 - Marcer - NEW TIME MARCER.sndh */
TIMEDB_ENTRY( 5a1b7e1d,  0,  98215,       YM+TA), /* 04626 - Michael Opel - Trinomic.sndh */
TIMEDB_ENTRY( 5a2fc291,  0,    762, YM+TA+TB+TD), /* 01942 - MSG - Chu Chu Rocket - Cat Mania.sndh */
TIMEDB_ENTRY( 5a33a74f,  0,  46080, YM+TA+TB+TD), /* 02560 - Marcer - DRAGONF MIX.sndh */
TIMEDB_ENTRY( 5a3b1f47,  0,  11036, YM+TA+TB+TD), /* 03005 - 505 - Antique.sndh */
TIMEDB_ENTRY( 5a3b27f6,  0,   7637, YM+TA+TB+TD), /* 00830 - Crazy Q - Timebomb - Outlaw (YM Rockerz).sndh */
TIMEDB_ENTRY( 5a4908be,  0,   6657,       YM+TA), /* 03820 - Tufty - Wunder.sndh */
TIMEDB_ENTRY( 5a4c8b67,  0,   2334,    YM+TA+TD), /* 02308 - Tao - Cream - Force (Amiga conversion).sndh */
TIMEDB_ENTRY( 5a4f1da6,  0,   2560,          YM), /* 02003 - Mad Max - Demo Music 9.sndh */
TIMEDB_ENTRY( 5a561c67,  0,    816,          YM), /* 04524 - Unknown - Early Learning Maths 2.sndh */
TIMEDB_ENTRY( 5a561c67,  1,   2944,          YM), /* 04524 - Unknown - Early Learning Maths 2.sndh */
TIMEDB_ENTRY( 5a561c67,  2,    489,          YM), /* 04524 - Unknown - Early Learning Maths 2.sndh */
TIMEDB_ENTRY( 5a614e63,  0,   1193,       YM+TA), /* 03843 - Unknown - Asterix im Morgenland - Title.sndh */
TIMEDB_ENTRY( 5a6fa9fc,  0,   6912,          YM), /* 04832 - T1000 - Bits 61.sndh */
TIMEDB_ENTRY( 5a9b3487,  0,  13434, YM+TA+TB+TD), /* 02480 - Floopy 2004 - Quasar.sndh */
TIMEDB_ENTRY( 5acc5761,  0,   3776,          YM), /* 00495 - Ultimatum - Kidney Bean 11.sndh */
TIMEDB_ENTRY( 5ad3d150,  0,  11808, YM+TA+TB+TD), /* 02070 - Mad Max - Commando - Sid.sndh */
TIMEDB_ENTRY( 5ad3d150,  1,   3072, YM+TA+TB+TD), /* 02070 - Mad Max - Commando - Sid.sndh */
TIMEDB_ENTRY( 5ad3d150,  2,    289, YM+TA+TB+TD), /* 02070 - Mad Max - Commando - Sid.sndh */
TIMEDB_ENTRY( 5af05644,  0,   8832,          YM), /* 02919 - Richard Joseph - Steel Empire.sndh */
TIMEDB_ENTRY( 5af1eb78,  0,  22393,          YM), /* 00581 - Buster - Impact - Chip.sndh */
TIMEDB_ENTRY( 5afa0232,  0,  40461,          YM), /* 04712 - Bob Bath - Real-stat Hockey.sndh */
TIMEDB_ENTRY( 5b04d0b6,  0,   3645, YM+TA+TB+TD), /* 04266 - 505-cpt^pdx - Impulse.sndh */
TIMEDB_ENTRY( 5b06affa,  0,   8508,          YM), /* 03167 - n-a - Spanish Eyes.sndh */
TIMEDB_ENTRY( 5b1eb16e,  0,   4986,    YM+TA+TD), /* 01177 - Sally - Sally Tri-Digi music.sndh */
TIMEDB_ENTRY( 5b38008d,  0,  59076,       YM+TA), /* 03796 - F Smadja - Back To The Golden Age.sndh */
TIMEDB_ENTRY( 5b5116e7,  0, 121893,       YM+TA), /* 04824 - Twitch - Altogether Now.sndh */
TIMEDB_ENTRY( 5b6c3f7c,  0,   5198,          YM), /* 01558 - Dave Rogers - Exolon.sndh */
TIMEDB_ENTRY( 5b6c3f7c,  1,    158,          YM), /* 01558 - Dave Rogers - Exolon.sndh */
TIMEDB_ENTRY( 5b6e2759,  0,   5120,       YM+TA), /* 01364 - Scavenger - Sait Evil.sndh */
TIMEDB_ENTRY( 5b845e71,  0,   7679,          YM), /* 01306 - Matt Furniss - Mig 29 Super Fulcrum.sndh */
TIMEDB_ENTRY( 5b8821ec,  0,   9504,          YM), /* 03873 - Gary Wheaton - JS Bach.sndh */
TIMEDB_ENTRY( 5b88bb81,  0,  94756,       YM+TA), /* 03967 - Unknown - Delta.sndh */
TIMEDB_ENTRY( 5b8ceb3a,  0,   6912,       YM+TA), /* 01749 - Kanewood - Faith.sndh */
TIMEDB_ENTRY( 5b8f84ed,  0,   1024,       YM+TA), /* 04762 - Ferox - State of Mind - Energy.sndh */
TIMEDB_ENTRY( 5baee0e2,  0,   1024,          YM), /* 01403 - Skyline - Snork Intro - European.sndh */
TIMEDB_ENTRY( 5bb7dda3,  0,   1337,       YM+TD), /* 04790 - Intruder - Intruder 6.sndh */
TIMEDB_ENTRY( 5bb94bdb,  0,  21725,       YM+TA), /* 03958 - Scott Lemmon - Piggies.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  0,   2879,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  1,   2879,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  2,   2879,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  3,   2879,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  4,   2879,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  5,   2685,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  6,   2429,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  7,   2301,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  8,   1919,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b,  9,   1919,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b, 10,   3454,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bbd5f2b, 11,   2750,          YM), /* 01308 - Matt Furniss - Outrun Europa.sndh */
TIMEDB_ENTRY( 5bc728b3,  0,   5440,          YM), /* 02440 - Jas C Brooke - Strike.sndh */
TIMEDB_ENTRY( 5bc76e91,  0,   2568,       YM+TA), /* 00923 - Dark Angel - Hop Jo Digi.sndh */
TIMEDB_ENTRY( 5bce1b51,  0,   4991, YM+TA+TB+TD), /* 02983 - 505-paradox - Erde.sndh */
TIMEDB_ENTRY( 5bd89943,  0,  11775,       YM+TD), /* 03605 - Per Almered (XiA) - TalkTalk2 Loader.sndh */
TIMEDB_ENTRY( 5bde0d87,  0,  15550,       YM+TA), /* 01399 - Skyline - Snork - Loader 2.sndh */
TIMEDB_ENTRY( 5be32494,  0,   4864,          YM), /* 00797 - Crawdaddy - Leavin Teramis.sndh */
TIMEDB_ENTRY( 5beb61e9,  0,   5755, YM+TA+TB+TD), /* 01839 - Lotek Style - Chips on Fire.sndh */
TIMEDB_ENTRY( 5bf3126c,  0,   3171,          YM), /* 00116 - David Whittaker - Menace.sndh */
TIMEDB_ENTRY( 5bf3126c,  1,   2312,          YM), /* 00116 - David Whittaker - Menace.sndh */
TIMEDB_ENTRY( 5bf3126c,  2,   7184,          YM), /* 00116 - David Whittaker - Menace.sndh */
TIMEDB_ENTRY( 5c041c45,  0,  11904,       YM+TD), /* 03220 - Tomchi - PTSPTR.sndh */
TIMEDB_ENTRY( 5c0bf856,  0,   2880,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  1,   6016,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  2,   4480,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  3,   5700,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  4,   5796,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  5,    255,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c0bf856,  6,   3168,          YM), /* 00751 - Mark Cooksey - Forgotten Worlds.sndh */
TIMEDB_ENTRY( 5c2d5f54,  0,   4583,          YM), /* 01311 - Matt Furniss - Revelation.sndh */
TIMEDB_ENTRY( 5c2d5f54,  1,    191,          YM), /* 01311 - Matt Furniss - Revelation.sndh */
TIMEDB_ENTRY( 5c2d5f54,  2,    204,          YM), /* 01311 - Matt Furniss - Revelation.sndh */
TIMEDB_ENTRY( 5c2d5f54,  3,    239,          YM), /* 01311 - Matt Furniss - Revelation.sndh */
TIMEDB_ENTRY( 5c385640,  0,   6048,          YM), /* 03024 - Alchemist - Rare (Main part).sndh */
TIMEDB_ENTRY( 5c471758,  0,   6265, YM+TA+TB+TD), /* 01962 - MSG - Godpey - Puzzle.sndh */
TIMEDB_ENTRY( 5c5abf0b,  0,  12096,    YM+TA+TD), /* 03050 - Crazy Q - Collage with a Moustache.sndh */
TIMEDB_ENTRY( 5c5ae98d,  0,   5127,    YM+TA+TD), /* 02324 - Tao - Cream (orig. Rob Hubbard) - Monty On The Run Highscore.sndh */
TIMEDB_ENTRY( 5c69cf28,  0,   6525,          YM), /* 00374 - Beast - TSC - TSC 22.sndh */
TIMEDB_ENTRY( 5c6d53d2,  0,    307,          YM), /* 04522 - Ian Chapman - Dizzy Fun Time.sndh */
TIMEDB_ENTRY( 5c6ec096,  0,  11516,    YM+TA+TD), /* 02511 - Crazy Q - DHS - Sex Droid.sndh */
TIMEDB_ENTRY( 5c6fdf77,  0,   4607, YM+TA+TB+TD), /* 04854 - Dubmood & Zabutom ! ! ! - Zabutoms Space Journey (D-Bug remix).sndh */
TIMEDB_ENTRY( 5c7fa268,  0,   8857, YM+TA+TB+TD), /* 04637 - MCH-GenesisProject^Dream - starlight.sndh */
TIMEDB_ENTRY( 5c8150d3,  0,   8098,          YM), /* 01277 - Frazer - TSC 52.sndh */
TIMEDB_ENTRY( 5c9f6920,  0,    504,    YM+TA+TB), /* 00834 - Crazy Q - TiMeBoMb - Toxid Tears pattern test.sndh */
TIMEDB_ENTRY( 5cab7ad6,  0,  10400,          YM), /* 01081 - Excellence In Art - Elftec.sndh */
TIMEDB_ENTRY( 5cb26475,  0,   7102,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb26475,  1,   7102,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb26475,  2,   6590,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb26475,  3,   6846,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb26475,  4,   7102,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb26475,  5,   4990,          YM), /* 01312 - Matt Furniss - Rogue Trooper.sndh */
TIMEDB_ENTRY( 5cb3cbd0,  0, 213967,       YM+TA), /* 03965 - Scott Lemmon - Work.sndh */
TIMEDB_ENTRY( 5cb75581,  0,     16,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  1,     14,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  2,     51,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  3,     36,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  4,     22,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  5,    483,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  6,     18,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  7,     35,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  8,     88,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581,  9,     39,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581, 10,     36,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cb75581, 11,     35,          YM), /* 04703 - M Franks - Pixie & Dixie.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  0,  13103,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  1,   2730,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  2,   1045,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  3,   4681,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  4,     97,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cc8ca9b,  5,    151,          YM), /* 00090 - David Whittaker - Custodian.sndh */
TIMEDB_ENTRY( 5cd611c2,  0,    864,       YM+TA), /* 04430 - Lotus - Synth Dream 2 - Loader.sndh */
TIMEDB_ENTRY( 5ce03029,  0,   8320,          YM), /* 01497 - Alex Herbert - Guardian.sndh */
TIMEDB_ENTRY( 5ce03029,  1,   9344,          YM), /* 01497 - Alex Herbert - Guardian.sndh */
TIMEDB_ENTRY( 5ce03029,  2,  16896,          YM), /* 01497 - Alex Herbert - Guardian.sndh */
TIMEDB_ENTRY( 5ce03029,  3,   1536,          YM), /* 01497 - Alex Herbert - Guardian.sndh */
TIMEDB_ENTRY( 5ce45268,  0,   8644,          YM), /* 00272 - Atomic - Last Hero (Phaleon - Albedo).sndh */
TIMEDB_ENTRY( 5ce57cce,  0,   4486,       YM+TA), /* 02242 - Tao - Cream (orig. Rob Hubbard) - Monty On The Run Highscore (Stmygm).sndh */
TIMEDB_ENTRY( 5ce7c5f2,  0,   2289,          YM), /* 03891 - Unknown - Risvegli.sndh */
TIMEDB_ENTRY( 5ce8e2e7,  0,   1785, YM+TA+TB+TD), /* 01957 - MSG - Godpey - Hiscore.sndh */
TIMEDB_ENTRY( 5ced344b,  0,   6528,          YM), /* 02902 - Techno - Plucked Goose Dune 2.sndh */
TIMEDB_ENTRY( 5ced4178,  0,   5892,          YM), /* 00447 - Augustin Vidovic - Heidi.sndh */
TIMEDB_ENTRY( 5cf67318,  0,   3604,          YM), /* 00157 - David Whittaker - Wreckers.sndh */
TIMEDB_ENTRY( 5cf67318,  1,   2520,          YM), /* 00157 - David Whittaker - Wreckers.sndh */
TIMEDB_ENTRY( 5cf67318,  2,    176,          YM), /* 00157 - David Whittaker - Wreckers.sndh */
TIMEDB_ENTRY( 5cf67318,  3,    314,          YM), /* 00157 - David Whittaker - Wreckers.sndh */
TIMEDB_ENTRY( 5cff4909,  0,  59459,       YM+TA), /* 03827 - JO - TEC Demo.sndh */
TIMEDB_ENTRY( 5cff4909,  1, 103284,       YM+TA), /* 03827 - JO - TEC Demo.sndh */
TIMEDB_ENTRY( 5cff4909,  2,  52998,       YM+TA), /* 03827 - JO - TEC Demo.sndh */
TIMEDB_ENTRY( 5cff4909,  3,  76033,       YM+TA), /* 03827 - JO - TEC Demo.sndh */
TIMEDB_ENTRY( 5d1ede86,  0,   6927, YM+TA+TB+TD), /* 01156 - Tyan - Serious 1.sndh */
TIMEDB_ENTRY( 5d207e67,  0,  44438, YM+TA+TB+TD), /* 02613 - STU - RETINA.sndh */
TIMEDB_ENTRY( 5d2b73cc,  0,   3888,          YM), /* 03399 - Sean Conran - Ruff and Ready.sndh */
TIMEDB_ENTRY( 5d2b73cc,  1,    183,          YM), /* 03399 - Sean Conran - Ruff and Ready.sndh */
TIMEDB_ENTRY( 5d2d3b46,  0,  14365,          YM), /* 04752 - MSG - Hallucinations.sndh */
TIMEDB_ENTRY( 5d30e01b,  0,   6639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  1,   1510,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  2,    766,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  3,    669,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  4,   1341,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  5,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  6,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  7,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  8,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b,  9,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b, 10,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b, 11,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d30e01b, 12,    639,          YM), /* 01313 - Matt Furniss - Rugby - The World Cup.sndh */
TIMEDB_ENTRY( 5d3bbc34,  0,   4552,          YM), /* 04088 - Gary Wheaton - Peewee's Big Adventure.sndh */
TIMEDB_ENTRY( 5d4599d2,  0,    767,          YM), /* 04097 - Gary Wheaton - Tri Heli.sndh */
TIMEDB_ENTRY( 5d562a89,  0,    553,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  1,    596,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  2,    385,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  3,    222,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  4,    320,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  5,    248,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  6,    128,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  7,    321,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  8,    121,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89,  9,     27,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89, 10,     28,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89, 11,     53,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89, 12,    382,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89, 13,    104,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d562a89, 14,     27,          YM), /* 01816 - Paul Loraine - Arcade Trivia Quiz.sndh */
TIMEDB_ENTRY( 5d6d8726,  0,   4429,    YM+TA+TD), /* 04252 - 505-cpt^pdx - Destruct.sndh */
TIMEDB_ENTRY( 5d784c68,  0,   8320,          YM), /* 02817 - Crazy Q - DHS - . Mindwarp ..sndh */
TIMEDB_ENTRY( 5d8e0999,  0,   7433,          YM), /* 03031 - Ric Clayton - Witchy Woman - Eagles (1972).sndh */
TIMEDB_ENTRY( 5dbf134d,  0,   5471,          YM), /* 00015 - 4-Mat - Magic Dizzy.sndh */
TIMEDB_ENTRY( 5dbf134d,  1,    113,          YM), /* 00015 - 4-Mat - Magic Dizzy.sndh */
TIMEDB_ENTRY( 5dbf134d,  2,    135,          YM), /* 00015 - 4-Mat - Magic Dizzy.sndh */
TIMEDB_ENTRY( 5dbf134d,  3,   2305,          YM), /* 00015 - 4-Mat - Magic Dizzy.sndh */
TIMEDB_ENTRY( 5dbf134d,  4,   3074,          YM), /* 00015 - 4-Mat - Magic Dizzy.sndh */
TIMEDB_ENTRY( 5dc28293,  0,  48458,       YM+TA), /* 03743 - Tufty - C64.sndh */
TIMEDB_ENTRY( 5dcb165f,  0,  76128,       YM+TA), /* 04020 - Roggie - Prelude.sndh */
TIMEDB_ENTRY( 5de01097,  0,    352,          YM), /* 04487 - Andrei Ellman - 3D Ski.sndh */
TIMEDB_ENTRY( 5de01097,  1,    288,          YM), /* 04487 - Andrei Ellman - 3D Ski.sndh */
TIMEDB_ENTRY( 5de5027f,  0,   2816,          YM), /* 04530 - Andrew Gower - Fleas.sndh */
TIMEDB_ENTRY( 5de5027f,  1,   2816,          YM), /* 04530 - Andrew Gower - Fleas.sndh */
TIMEDB_ENTRY( 5dfb9bff,  0,   4104,          YM), /* 01117 - Titan - GSC 62.sndh */
TIMEDB_ENTRY( 5e1ab844,  0,  37888,       YM+TD), /* 02616 - STU - SHIN.sndh */
TIMEDB_ENTRY( 5e292a9d,  0,  17450, YM+TA+TC+TD), /* 04368 - Skaven (ST adaption by Dma-Sc) - Second Reality 2013 end.sndh */
TIMEDB_ENTRY( 5e3529e2,  0,  16416,          YM), /* 02001 - Mad Max - Demo Music 6.sndh */
TIMEDB_ENTRY( 5e531be3,  0,   3136,          YM), /* 02191 - Tao - Cream - ...do things (loader).sndh */
TIMEDB_ENTRY( 5e553c9f,  0,  10941,       YM+TA), /* 04867 - Jovis - No Dash.sndh */
TIMEDB_ENTRY( 5e58f156,  0,   6522,          YM), /* 00025 - 505 - 17th.sndh */
TIMEDB_ENTRY( 5e7de824,  0,   5500,          YM), /* 04729 - Jason Sheppard - BSS Jane Seymour.sndh */
TIMEDB_ENTRY( 5e9ead81,  0,   1442,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  1,    974,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  2,   1058,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  3,   1330,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  4,   1410,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  5,   1442,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  7,   1380,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  8,    899,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81,  9,   1152,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 10,   1748,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 11,    861,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 12,    778,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 13,   1417,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 14,   1116,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 15,   1472,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 16,   1281,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 17,    693,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 20,    813,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 21,   1574,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 22,    837,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 24,    355,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 25,    487,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 26,    473,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 27,    581,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 28,    744,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 29,    708,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 30,    349,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 31,    360,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 32,    180,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 33,    429,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 34,    309,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 35,    408,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 36,    392,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 37,    416,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 38,    558,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 39,    212,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5e9ead81, 40,     86,          YM), /* 04735 - Amenda Lombardo - Mixed Up Mother Goose.sndh */
TIMEDB_ENTRY( 5ebfa147,  0,   5116,          YM), /* 00865 - Ben Daglish - Munsters.sndh */
TIMEDB_ENTRY( 5ed476e3,  0,  37116,       YM+TD), /* 02620 - STU - STEREOID DIGI.sndh */
TIMEDB_ENTRY( 5ee39ada,  0,  33699,       YM+TA), /* 04675 - Unknown - BIO C.sndh */
TIMEDB_ENTRY( 5eec5619,  0,   3199,          YM), /* 02151 - Mad Max - Monty On The Run High - LSD.sndh */
TIMEDB_ENTRY( 5eed22bd,  0,   4608,          YM), /* 03494 - Shamen - Certified Dope - Attitude Problem.sndh */
TIMEDB_ENTRY( 5f0e8e84,  0,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  1,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  2,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  4,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  5,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  6,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  7,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  8,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84,  9,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 10,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 11,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 12,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 13,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 14,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 15,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 16,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 17,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f0e8e84, 18,  12543,          YM), /* 01491 - Fred Gray - Where Time Stood Still.sndh */
TIMEDB_ENTRY( 5f1170cd,  0,  10452, YM+TA+TB+TD), /* 02757 - Dma-Sc - Sector One - CRAUFunk.sndh */
TIMEDB_ENTRY( 5f168890,  0,  15216,       YM+TA), /* 01753 - Kanewood - Phantom Demo - Main.sndh */
TIMEDB_ENTRY( 5f2cbdd9,  0,  11852,          YM), /* 00102 - David Whittaker - Garfield.sndh */
TIMEDB_ENTRY( 5f2cbdd9,  1,    182,          YM), /* 00102 - David Whittaker - Garfield.sndh */
TIMEDB_ENTRY( 5f2cbdd9,  2,    274,          YM), /* 00102 - David Whittaker - Garfield.sndh */
TIMEDB_ENTRY( 5f307aaf,  0,  29184,          YM), /* 02534 - Remo - Bouncy.sndh */
TIMEDB_ENTRY( 5f361be2,  0,   3464,          YM), /* 01336 - Scavenger - Hotliner.sndh */
TIMEDB_ENTRY( 5f51c888,  0,   4224,       YM+TA), /* 01089 - Techno - Puppet Show Intro digi.sndh */
TIMEDB_ENTRY( 5f671ce9,  0,   5657,          YM), /* 02134 - Mad Max - Flash Gordon - LCD.sndh */
TIMEDB_ENTRY( 5f6cdfc1,  0,   5369,       YM+TD), /* 04792 - Intruder - Intruder 5.sndh */
TIMEDB_ENTRY( 5f6d3827,  0,   8505, YM+TA+TB+TD), /* 01847 - Lotek Style - Fading Twilight Intro Edit.sndh */
TIMEDB_ENTRY( 5f83c750,  0,   4480,    YM+TA+TD), /* 00041 - 505 - Checkpoint - MMJ Xmas 03 - 1.sndh */
TIMEDB_ENTRY( 5f8eacef,  0,  35712,          YM), /* 01478 - STU - STEREOID.sndh */
TIMEDB_ENTRY( 5f93d42c,  0,  12672,          YM), /* 03276 - Unknown - ESD G-Loc Intro.sndh */
TIMEDB_ENTRY( 5f95cad7,  0,  13377,    YM+TA+TD), /* 02357 - Tao - Cream (orig. Hubbard-Jarre) - Zoolook.sndh */
TIMEDB_ENTRY( 5fb25bb2,  0,   7680,          YM), /* 01654 - Jedi - Sernsima.sndh */
TIMEDB_ENTRY( 5fb26a8a,  0,   1164,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  1,    518,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  2,    731,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  3,    341,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  4,    112,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  5,    160,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  6,   1280,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  7,   1828,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  8,   1828,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a,  9,   1219,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a, 11,    194,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a, 12,    335,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a, 14,    103,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fb26a8a, 15,    278,          YM), /* 04495 - Darren Ithell - Balls up.sndh */
TIMEDB_ENTRY( 5fcdea08,  0,   7679, YM+TA+TB+TD), /* 02708 - dubmood & Zabutom - ym2149 oh sexy thing edit.sndh */
TIMEDB_ENTRY( 5fd4be5a,  0,   8955,       YM+TB), /* 00057 - 505 - Nostalgia - Toys.sndh */
TIMEDB_ENTRY( 5fd8dde9,  0,  21492,          YM), /* 01701 - Jess - Middle-Earth Theme (chip).sndh */
TIMEDB_ENTRY( 5fe3d3ca,  0,   2848,          YM), /* 00591 - Cerror - Outline.sndh */
TIMEDB_ENTRY( 5fe4bfb4,  0,   5760,          YM), /* 00245 - AN Cool - So Watt - 2 Life Crew.sndh */
TIMEDB_ENTRY( 5fe5097b,  0,  43521, YM+TA+TB+TD), /* 03367 - STU - BOMBABAZOOKA.sndh */
TIMEDB_ENTRY( 5ff1f0b5,  0,  19584, YM+TA+TB+TD), /* 02531 - Bodenst ndig 2000 - Die Wahrheit.sndh */
TIMEDB_ENTRY( 5ff35036,  0,  12801,       YM+TA), /* 03955 - Scott Lemmon - Ivan 3.sndh */
TIMEDB_ENTRY( 6002d68c,  0,   3834,       YM+TA), /* 04448 - Scavenger - Loaders.sndh */
TIMEDB_ENTRY( 6011f87f,  0,  13824,          YM), /* 01002 - DMA-SC - Mr Boomer Intro.sndh */
TIMEDB_ENTRY( 601bcaa3,  0,   8413,          YM), /* 02050 - Mad Max - Sanxion Title.sndh */
TIMEDB_ENTRY( 602ad458,  0,  11009,       YM+TA), /* 04768 - Unknown - Gore 2.sndh */
TIMEDB_ENTRY( 6041e614,  0,   1785,       YM+TD), /* 04788 - Intruder - Intruder 4.sndh */
TIMEDB_ENTRY( 607b46c5,  0,   5663,          YM), /* 02656 - Dma-Sc (Rockman style  ) - ST 20th Ann. Megademo Menu.sndh */
TIMEDB_ENTRY( 6081f79b,  0,   3092,          YM), /* 03048 - Ed Bogas - Folk Song.sndh */
TIMEDB_ENTRY( 60944764,  0,   3584,          YM), /* 02225 - Tao - Cream - 4rce (DHS 4k-demo).sndh */
TIMEDB_ENTRY( 60c727ff,  0,   8442,          YM), /* 01919 - Mephistow - Ventura - Hystanders.sndh */
TIMEDB_ENTRY( 60c8284f,  0,   5952,          YM), /* 00936 - Dark Angel - Mix.sndh */
TIMEDB_ENTRY( 60c93af9,  0,    967,          YM), /* 04684 - Unknown - Kids' Academy - Alvin's puzzles - titles.sndh */
TIMEDB_ENTRY( 60c93af9, 16,    967,          YM), /* 04684 - Unknown - Kids' Academy - Alvin's puzzles - titles.sndh */
TIMEDB_ENTRY( 60e83f4d,  0,  15360,    YM+TA+TD), /* 02248 - Tao - Cream - 150 Mph.sndh */
TIMEDB_ENTRY( 60eda8b4,  0,   4656,          YM), /* 03655 - Jedi - Nothing Is More - Song 4.sndh */
TIMEDB_ENTRY( 60f60dd8,  0,   4033, YM+TA+TB+TD), /* 02995 - 505-paradox - parkbank (STE only).sndh */
TIMEDB_ENTRY( 611973ab,  0,  24576, YM+TA+TB+TD), /* 02566 - Marcer - NOISY 05 MARCER.sndh */
TIMEDB_ENTRY( 6119bafd,  0,  13445, YM+TA+TB+TD), /* 01270 - Frazer - The Syndicate ( 2003 ) - Return Of The Lapplander (YM Rockerz).sndh */
TIMEDB_ENTRY( 612de492,  0,  10752,          YM), /* 01974 - MSG - MSG 04.sndh */
TIMEDB_ENTRY( 616a92b9,  0,   3578,          YM), /* 00561 - Allister Brimble - Super Seymour.sndh */
TIMEDB_ENTRY( 616a92b9,  1,   8506,          YM), /* 00561 - Allister Brimble - Super Seymour.sndh */
TIMEDB_ENTRY( 616e8445,  0,  23040, YM+TA+TB+TD), /* 02568 - Marcer - TRUST - MARCER.sndh */
TIMEDB_ENTRY( 617b96f0,  0,   8959, YM+TA+TB+TD), /* 04363 - V0yager-Lamers - One Day Wonder.sndh */
TIMEDB_ENTRY( 6180066f,  0,  25600, YM+TA+TB+TD), /* 02815 - Marcer - Paradize - Nuclear Waste Dump - Ingame tune 2.sndh */
TIMEDB_ENTRY( 6183ea9f,  0,   5200, YM+TA+TB+TD), /* 00033 - 505 - Bambi.sndh */
TIMEDB_ENTRY( 618a1a6d,  0,  18424,       YM+TD), /* 01029 - Doclands - Ventura Demo - Main Menu.sndh */
TIMEDB_ENTRY( 619efd39,  0,    635,          YM), /* 04274 - 505 - lame02.sndh */
TIMEDB_ENTRY( 61a592b4,  0,   6906,          YM), /* 01231 - Epic - Japanese Sunlight.sndh */
TIMEDB_ENTRY( 61ab308e,  0,    288,       YM+TA), /* 00406 - Big Alec - Jazz.sndh */
TIMEDB_ENTRY( 61c69dc2,  0,   4960,          YM), /* 01387 - Andy Severn - Electraglide.sndh */
TIMEDB_ENTRY( 61c69dc2,  1,   4960,          YM), /* 01387 - Andy Severn - Electraglide.sndh */
TIMEDB_ENTRY( 61c69dc2,  2,   4960,          YM), /* 01387 - Andy Severn - Electraglide.sndh */
TIMEDB_ENTRY( 61c69dc2,  3,   6720,          YM), /* 01387 - Andy Severn - Electraglide.sndh */
TIMEDB_ENTRY( 61c69dc2,  4,   5440,          YM), /* 01387 - Andy Severn - Electraglide.sndh */
TIMEDB_ENTRY( 61d360ea,  0,  16640,       YM+TD), /* 02584 - STU - CRIME TIME.sndh */
TIMEDB_ENTRY( 61e85111,  0,  10753, YM+TA+TB+TD), /* 03586 - Divag - medievol.sndh */
TIMEDB_ENTRY( 62020414,  0,   8058, YM+TA+TB+TD), /* 01358 - Scavenger - Please Wait.sndh */
TIMEDB_ENTRY( 6207fb61,  0,   3129,    YM+TA+TD), /* 04287 - 505-cpt^pdx - messing.sndh */
TIMEDB_ENTRY( 622d4c6c,  0,   4599,    YM+TA+TD), /* 02481 - Floopy - Teletubb.sndh */
TIMEDB_ENTRY( 6235abd7,  0,   1040,      YM+STE), /* 01999 - Mad Max - Bomber (STe).sndh */
TIMEDB_ENTRY( 6242f4be,  0,  15797,          YM), /* 02720 - Christian Haupt - Wizmo.sndh */
TIMEDB_ENTRY( 6242f4be,  1,   3949,          YM), /* 02720 - Christian Haupt - Wizmo.sndh */
TIMEDB_ENTRY( 6242f4be,  2,   5920,          YM), /* 02720 - Christian Haupt - Wizmo.sndh */
TIMEDB_ENTRY( 62760220,  0,   1914,       YM+TA), /* 02639 - Miggy - DevilVer3(Unfinished).sndh */
TIMEDB_ENTRY( 6289981e,  0,   2304,          YM), /* 04609 - White Ninja - How forget....sndh */
TIMEDB_ENTRY( 628f08ca,  0,   5951, YM+TA+TB+TD), /* 03857 - Yerzmyey+!Cube - Dizzzruptor part 3 (Atari ST remix).sndh */
TIMEDB_ENTRY( 629aae6c,  0,  17408,          YM), /* 02047 - Mad Max - One Man and His Droid.sndh */
TIMEDB_ENTRY( 62a84672,  0,    767,          YM), /* 04772 - Dma-Sc - Lovetro intro.sndh */
TIMEDB_ENTRY( 62acc74a,  0,   8831,    YM+TB+TD), /* 02657 - DIVAG - La potato.sndh */
TIMEDB_ENTRY( 62afbe50,  0,   4026,          YM), /* 01243 - Epic - Wonderful Scroll.sndh */
TIMEDB_ENTRY( 62f59ad0,  0,  46080, YM+TA+TB+TD), /* 02535 - Remo - Fluffy Bunny.sndh */
TIMEDB_ENTRY( 62fc134d,  0,  13096,    YM+TA+TD), /* 01851 - Lotek Style - Hard Sid.sndh */
TIMEDB_ENTRY( 62fe5f12,  0, 178183,       YM+TA), /* 03953 - Scott Lemmon - Instod5.sndh */
TIMEDB_ENTRY( 63152838,  0,  12288,          YM), /* 03262 - Remalon - Requiem 2.sndh */
TIMEDB_ENTRY( 63166e49,  0,   5412,          YM), /* 01253 - Floopy - Vivevent.sndh */
TIMEDB_ENTRY( 632cda62,  0,   3071,    YM+TA+TD), /* 02802 - Starbuck^Creative Mind - Insomnia (Cracktro styled).sndh */
TIMEDB_ENTRY( 6335a459,  0,   2463,          YM), /* 01245 - FFT - Phaleon - Beast 1.sndh */
TIMEDB_ENTRY( 634d837a,  0,   3200,          YM), /* 04206 - Steve Bak - Lands of Havoc.sndh */
TIMEDB_ENTRY( 6357527a,  0,  10257,          YM), /* 02136 - Mad Max - Human Race Level 4 - LCD.sndh */
TIMEDB_ENTRY( 6361d67f,  0,     64,          YM), /* 04546 - Peter M. Caddock, Gary P. Tinsley - Hedge Hogger.sndh */
TIMEDB_ENTRY( 6361d67f,  1,    232,          YM), /* 04546 - Peter M. Caddock, Gary P. Tinsley - Hedge Hogger.sndh */
TIMEDB_ENTRY( 6361d67f,  2,    307,          YM), /* 04546 - Peter M. Caddock, Gary P. Tinsley - Hedge Hogger.sndh */
TIMEDB_ENTRY( 6361d67f,  3,   3643,          YM), /* 04546 - Peter M. Caddock, Gary P. Tinsley - Hedge Hogger.sndh */
TIMEDB_ENTRY( 6361d67f,  4,     59,          YM), /* 04546 - Peter M. Caddock, Gary P. Tinsley - Hedge Hogger.sndh */
TIMEDB_ENTRY( 63875879,  0,  37242,          YM), /* 00889 - Adrian Dalecki - Soccer Manager Plus.sndh */
TIMEDB_ENTRY( 63875879,  1, 106746,          YM), /* 00889 - Adrian Dalecki - Soccer Manager Plus.sndh */
TIMEDB_ENTRY( 638e66d5,  0,  24976, YM+TA+TB+TD), /* 03299 - STU - DARK ANGEL.sndh */
TIMEDB_ENTRY( 639137c6,  0,  17465,       YM+TA), /* 03550 - Jovis - Fun Sad.sndh */
TIMEDB_ENTRY( 6396d176,  0,   2770,          YM), /* 00226 - ACC - Beguine.sndh */
TIMEDB_ENTRY( 63a3a65c,  0,  11808,          YM), /* 02024 - Mad Max - Commando.sndh */
TIMEDB_ENTRY( 63a3a65c,  1,   3072,          YM), /* 02024 - Mad Max - Commando.sndh */
TIMEDB_ENTRY( 63a3a65c,  2,    289,          YM), /* 02024 - Mad Max - Commando.sndh */
TIMEDB_ENTRY( 63a4f145,  0,   6822,          YM), /* 04341 - Nexus 6 - Beyond Eden (Nino...).sndh */
TIMEDB_ENTRY( 63a9a3bc,  0,   6048,          YM), /* 00929 - Dark Angel - Snork - Logic Force.sndh */
TIMEDB_ENTRY( 63ac3769,  0,  12244, YM+TA+TB+TD), /* 03581 - Divag - Atari Evil Circus.sndh */
TIMEDB_ENTRY( 63b7828a,  0,  32768,       YM+TD), /* 02532 - Remo - Bass Bass Bass.sndh */
TIMEDB_ENTRY( 63b80bfb,  0,  11684,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  1,   6927,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  2,   5568,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  3,   9600,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  4,  12096,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  5,  11896,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  6,  12545,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  7,   9600,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  8,  18816,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63b80bfb,  9,  11904,          YM), /* 03246 - T1000 - The Beast Club - TBC 18.sndh */
TIMEDB_ENTRY( 63bac55c,  0,   4105,       YM+TA), /* 03484 - Peter Jorgensen - Klatmix.sndh */
TIMEDB_ENTRY( 63c04c7e,  0,  53788, YM+TA+TB+TD), /* 02781 - Stu - Audiosphere.sndh */
TIMEDB_ENTRY( 63cdc140,  0,   3584,          YM), /* 00609 - Chris Mad - Whattaheck - Ghost.sndh */
TIMEDB_ENTRY( 63d6a88f,  0, 141872,       YM+TA), /* 03721 - Spaz - Ooh Crikey - Homische.sndh */
TIMEDB_ENTRY( 63e3acd9,  0,  22080,          YM), /* 02169 - Mad Max - Mega Apocalpyse.sndh */
TIMEDB_ENTRY( 63ec18b6,  0,   2681,       YM+TA), /* 01930 - Miq - Axel Foley.sndh */
TIMEDB_ENTRY( 63f7b876,  0,  11522,          YM), /* 04189 - Tao - For Abyss.sndh */
TIMEDB_ENTRY( 6420fd5d,  0,   6912,       YM+TD), /* 00433 - Big Alec - Reality.sndh */
TIMEDB_ENTRY( 6420ff6f,  0,   8959,    YM+TA+TD), /* 03434 - Per Almered (XiA) 2009 - Trans-Y VideoLife.sndh */
TIMEDB_ENTRY( 642e0540,  0,   2674,          YM), /* 02744 - Unknown Composer - Vampire's Empire.sndh */
TIMEDB_ENTRY( 64409956,  0,   5568,          YM), /* 00128 - David Whittaker - Renaissance.sndh */
TIMEDB_ENTRY( 64549b22,  0,   1536,          YM), /* 00979 - Chris Denman - ProSoccer 2190.sndh */
TIMEDB_ENTRY( 645fd0d2,  0,   8832,          YM), /* 02510 - Crazy Q - Creators - Parallax Lover.sndh */
TIMEDB_ENTRY( 646e645a,  0,   5919,          YM), /* 00129 - David Whittaker - Renegade - Andrew Tather 88 Remix.sndh */
TIMEDB_ENTRY( 6472e23e,  0,   1516, YM+TA+TB+TD), /* 04640 - mOdmate-CheckPoint - Yanartas SV Thank You.sndh */
TIMEDB_ENTRY( 64804aff,  0,   4993,       YM+TA), /* 01363 - Scavenger - Row, Row, Row Your Boat.sndh */
TIMEDB_ENTRY( 64872fc0,  0,   1788,       YM+TA), /* 01351 - Scavenger - Now Loading (And Packing).sndh */
TIMEDB_ENTRY( 648af9ef,  0,   5817,    YM+TB+TD), /* 01838 - Lotek Style - Centauri.sndh */
TIMEDB_ENTRY( 6496bb3d,  0,   5184,          YM), /* 03901 - Frantisek Fuka - Belegost.sndh */
TIMEDB_ENTRY( 649e2a02,  0,  18624,          YM), /* 02216 - Mad Max - Stormlord.sndh */
TIMEDB_ENTRY( 649e2a02,  1,    494,          YM), /* 02216 - Mad Max - Stormlord.sndh */
TIMEDB_ENTRY( 649e2a02,  2,    206,          YM), /* 02216 - Mad Max - Stormlord.sndh */
TIMEDB_ENTRY( 649e2a02,  3,  24576,          YM), /* 02216 - Mad Max - Stormlord.sndh */
TIMEDB_ENTRY( 649f7c4b,  0,  21381,       YM+TA), /* 04560 - Rob Povey - Quartet v1.5.sndh */
TIMEDB_ENTRY( 649f7c4b,  1,  39530,       YM+TA), /* 04560 - Rob Povey - Quartet v1.5.sndh */
TIMEDB_ENTRY( 649f7c4b,  2,  14884,       YM+TA), /* 04560 - Rob Povey - Quartet v1.5.sndh */
TIMEDB_ENTRY( 64a27951,  0,   3456,          YM), /* 00490 - Ultimatum - Kidney Bean 6.sndh */
TIMEDB_ENTRY( 64ab1c59,  0,   4096,          YM), /* 03217 - Tomchi - 4 Screen.sndh */
TIMEDB_ENTRY( 64becfa4,  0,   4608,          YM), /* 01514 - Chris Huelsbeck - The Adventures of Quik & Silva.sndh */
TIMEDB_ENTRY( 64becfa4,  1,   4352,          YM), /* 01514 - Chris Huelsbeck - The Adventures of Quik & Silva.sndh */
TIMEDB_ENTRY( 64becfa4,  2,   2816,          YM), /* 01514 - Chris Huelsbeck - The Adventures of Quik & Silva.sndh */
TIMEDB_ENTRY( 64c634bc,  0,  10047,          YM), /* 01646 - Jedi - Sprite Ball record.sndh */
TIMEDB_ENTRY( 64ee1ded,  0,  12089,          YM), /* 02494 - gwEm - Love Washed Away With The Tide.sndh */
TIMEDB_ENTRY( 650104a6,  0,  13665,    YM+TA+TD), /* 02209 - Tao - Cream - Stella Polaris (Madness reset).sndh */
TIMEDB_ENTRY( 650176bc,  0,    960,          YM), /* 04479 - Thomas the Tank Engine 2 - Paul Tankard.sndh */
TIMEDB_ENTRY( 650176bc,  1,    193,          YM), /* 04479 - Thomas the Tank Engine 2 - Paul Tankard.sndh */
TIMEDB_ENTRY( 650176bc,  2,    192,          YM), /* 04479 - Thomas the Tank Engine 2 - Paul Tankard.sndh */
TIMEDB_ENTRY( 650176bc,  3,    307,          YM), /* 04479 - Thomas the Tank Engine 2 - Paul Tankard.sndh */
TIMEDB_ENTRY( 6502f195,  0,   3150,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  1,    120,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  2,    375,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  3,    138,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  4,    129,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  5,    128,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 6502f195,  6,    102,          YM), /* 00062 - David Whittaker - Advanced Rugby Simulator.sndh */
TIMEDB_ENTRY( 65067faf,  0,   6166,       YM+TA), /* 00408 - Big Alec - Killer Digi.sndh */
TIMEDB_ENTRY( 65094efa,  0,   2664,          YM), /* 03912 - Unknown - Dampf.sndh */
TIMEDB_ENTRY( 6512cfbb,  0,   8596,          YM), /* 01612 - No More - AM Dentro.sndh */
TIMEDB_ENTRY( 6522b5e3,  0,  16512,       YM+TA), /* 00259 - Antichrist - Oh No!!...sndh */
TIMEDB_ENTRY( 652af93c,  0,  20176,          YM), /* 03661 - Jedi - Shoot Preview.sndh */
TIMEDB_ENTRY( 652cbfee,  0,   3196,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652cbfee,  1,   3196,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652cbfee,  2,   1916,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652cbfee,  3,   2956,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652cbfee,  4,   3596,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652cbfee,  5,   1531,          YM), /* 00554 - Allister Brimble - Kwik Snax.sndh */
TIMEDB_ENTRY( 652e2cbb,  0,  58416, YM+TA+TB+TD), /* 04334 - mOdmate-CheckPoint (orig. Djamm-eKO) - almoST real (ENtRACte cHiP).sndh */
TIMEDB_ENTRY( 65362b13,  0,   3072,          YM), /* 03308 - Tomchi - Hap.sndh */
TIMEDB_ENTRY( 65543089,  0,   2939,       YM+TA), /* 04292 - 505 - newu.sndh */
TIMEDB_ENTRY( 656490af,  0,   2685,          YM), /* 02473 - Iso - Lefthanded.sndh */
TIMEDB_ENTRY( 657e825f,  0,   9024,       YM+TA), /* 00803 - Crawdaddy - No Way Out Digi.sndh */
TIMEDB_ENTRY( 658107b4,  0,  21488,          YM), /* 03515 - Richard Joseph - Barbarian.sndh */
TIMEDB_ENTRY( 658107b4,  1,    704,          YM), /* 03515 - Richard Joseph - Barbarian.sndh */
TIMEDB_ENTRY( 658107b4,  2,   4785,          YM), /* 03515 - Richard Joseph - Barbarian.sndh */
TIMEDB_ENTRY( 65b25ff0,  0,   1536,    YM+TA+TD), /* 02819 - Soulmate #1 - Znax ingame tune 2.sndh */
TIMEDB_ENTRY( 65c65b22,  0,  18432,          YM), /* 01449 - Stax - Simple.sndh */
TIMEDB_ENTRY( 65c6be3b,  0,  12993,       YM+TA), /* 03738 - David Whittaker - International Soccer Challenge.sndh */
TIMEDB_ENTRY( 65d3843c,  0,  13304, YM+TA+TB+TD), /* 01015 - DMA-SC - Spiral Zone - UCM 16.sndh */
TIMEDB_ENTRY( 65d849af,  0,   8956,          YM), /* 01273 - Frazer - TSC 27.sndh */
TIMEDB_ENTRY( 65e9574a,  0,   1222,          YM), /* 03180 - n-a - Up on the House Top.sndh */
TIMEDB_ENTRY( 65f86d4d,  0,  39936, YM+TA+TB+TD), /* 02585 - STU - DANGER.sndh */
TIMEDB_ENTRY( 65fd957b,  0,  15275,          YM), /* 02025 - Mad Max - Confuzion.sndh */
TIMEDB_ENTRY( 661d9cf4,  0,   4602,          YM), /* 00887 - Adrian Dalecki - Rollin Ronny A.sndh */
TIMEDB_ENTRY( 6642e0c3,  0,  29952, YM+TA+TB+TD), /* 02598 - STU - INSTREAMINTRO.sndh */
TIMEDB_ENTRY( 6660ae1a,  0,   6154, YM+TA+TB+TD), /* 00180 - 505 - old crap - UCM 22.sndh */
TIMEDB_ENTRY( 66662f74,  0,  64695,       YM+TA), /* 03962 - Scott Lemmon - Thrash 1.sndh */
TIMEDB_ENTRY( 668a6c70,  0,   4169,          YM), /* 04089 - Gary Wheaton - Permanent Waves.sndh */
TIMEDB_ENTRY( 6695e64c,  0,   2206,          YM), /* 00866 - Ben Daglish - Northstar.sndh */
TIMEDB_ENTRY( 6695e64c,  1,    830,          YM), /* 00866 - Ben Daglish - Northstar.sndh */
TIMEDB_ENTRY( 6695e64c,  2,  10387,          YM), /* 00866 - Ben Daglish - Northstar.sndh */
TIMEDB_ENTRY( 6697b220,  0,   3835,    YM+TB+TD), /* 00222 - 7AN - Blip 3.sndh */
TIMEDB_ENTRY( 669c22a9,  0,   6222,          YM), /* 03673 - Dma-Sc - Antiques - 3D Part.sndh */
TIMEDB_ENTRY( 66a16c32,  0,  59977,       YM+TA), /* 04812 - Unknown - Vincent.sndh */
TIMEDB_ENTRY( 66bd7052,  0,   6927,          YM), /* 03238 - T1000 - The Beast Club - TBC 10.sndh */
TIMEDB_ENTRY( 66d1653f,  0,   1715,          YM), /* 04541 - JAWX - Funfares.sndh */
TIMEDB_ENTRY( 66f95843,  0,    288,          YM), /* 01080 - Excellence In Art - Drumz.sndh */
TIMEDB_ENTRY( 66fd72bf,  0,  11362,          YM), /* 03282 - Sparky - Evil Force 121.sndh */
TIMEDB_ENTRY( 6704938d,  0,  14208,          YM), /* 00761 - Count Zero - Decade Demo - 3D Mania.sndh */
TIMEDB_ENTRY( 6725d6fa,  0,   2309,          YM), /* 03871 - Gary Wheaton - Clod Hopper.sndh */
TIMEDB_ENTRY( 67300f54,  0,  11901, YM+TA+TB+TD), /* 01020 - DMA-SC - Wonderland Speedway.sndh */
TIMEDB_ENTRY( 673d587d,  0,  11699,          YM), /* 02246 - Tao - Cream (orig. Rob Hubbard) - Thundercats (Stmygm).sndh */
TIMEDB_ENTRY( 67591bed,  0,   9595,    YM+TB+TD), /* 04661 - Nemo - YM Rocker.sndh */
TIMEDB_ENTRY( 67678c83,  0,   8064,    YM+TA+TD), /* 04210 - Tao - Cream - Just Feel It (Steps).sndh */
TIMEDB_ENTRY( 676f2107,  0,  53052,       YM+TA), /* 04819 - Ace - Demo Tune.sndh */
TIMEDB_ENTRY( 6778bf2e,  0,   1536,          YM), /* 01412 - Skyline - Snork Intro - So Watt.sndh */
TIMEDB_ENTRY( 679131f6,  0,   4608,       YM+TA), /* 00820 - !Cube - Japanese (Sinful Sinuses).sndh */
TIMEDB_ENTRY( 67a3f3db,  0,   5461,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  1,   1306,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  2,   5915,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  3,    767,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  4,    308,       YM+TA), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  5,   9943,       YM+TA), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  6,  10736,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  7,   5075,       YM+TA), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  8,  14972,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67a3f3db,  9,   7438,    YM+TA+TD), /* 01575 - Jurgen Piscol - X-Out.sndh */
TIMEDB_ENTRY( 67aec997,  0,   9360,       YM+TD), /* 02381 - Nexus 6 - Lethal Xcess Level 3.sndh */
TIMEDB_ENTRY( 67aec997,  1,    800,       YM+TD), /* 02381 - Nexus 6 - Lethal Xcess Level 3.sndh */
TIMEDB_ENTRY( 67aec997,  2,   2087,       YM+TD), /* 02381 - Nexus 6 - Lethal Xcess Level 3.sndh */
TIMEDB_ENTRY( 67b3ebf9,  0,  11777,       YM+TA), /* 00258 - Antichrist - Oceans.sndh */
TIMEDB_ENTRY( 67c3acd0,  0,   8315, YM+TA+TB+TD), /* 00999 - DMA-SC - Heaven Is A Place On Earth.sndh */
TIMEDB_ENTRY( 67cb3b1e,  0,    635,    YM+TA+TD), /* 04281 - 505 - Loop04.sndh */
TIMEDB_ENTRY( 67d46062,  0,   7680,          YM), /* 01423 - Sonic - Rough.sndh */
TIMEDB_ENTRY( 67d7a40b,  0,   5025,          YM), /* 00230 - ACC - Michella.sndh */
TIMEDB_ENTRY( 67dc492e,  0,   2233,       YM+TA), /* 04277 - 505 - Loading.sndh */
TIMEDB_ENTRY( 67e7f2a8,  0,   5632,          YM), /* 04482 - Mad Max - Axel F - TEX Cracktro.sndh */
TIMEDB_ENTRY( 68112d77,  0,  86044,       YM+TA), /* 04796 - Unknown - Axel F.sndh */
TIMEDB_ENTRY( 6814db57,  0,  66817,       YM+TA), /* 03818 - TUFTY - PYRO.sndh */
TIMEDB_ENTRY( 6827389a,  0,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  1,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  2,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  3,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  4,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  5,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  6,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  7,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  8,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a,  9,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a, 10,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a, 11,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a, 12,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a, 13,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 6827389a, 14,   4992,          YM), /* 01443 - Stax - Noisy Pillars.sndh */
TIMEDB_ENTRY( 683f9a14,  0,  20594,          YM), /* 01895 - LTK - Complete.sndh */
TIMEDB_ENTRY( 685ffc0f,  0,  18816,       YM+TA), /* 00946 - Dark Angel - Mathematica - Seasons Digi.sndh */
TIMEDB_ENTRY( 688eb0c2,  0,   5833, YM+TA+TB+TD), /* 00058 - 505 - Huzt.sndh */
TIMEDB_ENTRY( 689208de,  0,  17664,          YM), /* 01093 - Techno - Mathematiques Squizophrenes Main.sndh */
TIMEDB_ENTRY( 68abea93,  0,   1553,          YM), /* 00862 - Ben Daglish - Monty Python's Flying Circus.sndh */
TIMEDB_ENTRY( 68ad370f,  0,   3835,    YM+TA+TD), /* 00996 - Dma-Sc-Sector1^YM-Rockerz - Fantasia Tune 3 (End Part).sndh */
TIMEDB_ENTRY( 68b06314,  0,   3517,       YM+TA), /* 02776 - 505-paradox (2006) - 96shades.sndh */
TIMEDB_ENTRY( 68b92abf,  0,  10758,          YM), /* 04426 - Lotus - Synth Dream 2 - End.sndh */
TIMEDB_ENTRY( 68cc4dd8,  0,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  1,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  2,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  3,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  4,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  5,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  6,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  7,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  8,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8,  9,    147,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8, 10,    388,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8, 11,    973,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8, 12,    512,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68cc4dd8, 13,  44429,          YM), /* 04488 - Flipside - Alphabet mix.sndh */
TIMEDB_ENTRY( 68e43d5d,  0,   3129, YM+TA+TB+TD), /* 03644 - 505-checkpoint - No Stress.sndh */
TIMEDB_ENTRY( 69079f07,  0,  18495,       YM+TA), /* 01747 - Jovis - Lametrop - Minitext.sndh */
TIMEDB_ENTRY( 69244f58,  0,   4992,          YM), /* 00901 - Danny Oneway - Eat This.sndh */
TIMEDB_ENTRY( 69296b00,  0,  36897,          YM), /* 00009 - gwEm - Tune 4 My Broken Atari (re-version).sndh */
TIMEDB_ENTRY( 692c239e,  0,   9602,       YM+TA), /* 02677 - Donald Fakk - Leif (Chip).sndh */
TIMEDB_ENTRY( 693a839e,  0,   9983,       YM+TA), /* 02496 - gwEm - Outlined.sndh */
TIMEDB_ENTRY( 6950e7cf,  0,  22016,          YM), /* 03359 - Marcer - TECHNO ORION.sndh */
TIMEDB_ENTRY( 69557458,  0,   9023, YM+TA+TB+TD), /* 03525 - Timbral 2005-2010 - So alone... (Live ver.).sndh */
TIMEDB_ENTRY( 69656cbe,  0,   3072,          YM), /* 00977 - Chris Denman - Fort Apache.sndh */
TIMEDB_ENTRY( 6985b24d,  0,   7035, YM+TA+TB+TD), /* 04324 - 505-Checkpoint - Weird.sndh */
TIMEDB_ENTRY( 69aa5838,  0,  14560,          YM), /* 02198 - Mad Max - Insects In Space.sndh */
TIMEDB_ENTRY( 69ad3e79,  0,  14689,       YM+TA), /* 04809 - Unknown - Mission Impossible.sndh */
TIMEDB_ENTRY( 69da012c,  0,    622,          YM), /* 03905 - Unknown - Atahym.sndh */
TIMEDB_ENTRY( 6a022c8c,  0,   8452,    YM+TA+TD), /* 02328 - Tao - Cream - Quarterback (C64 conv.).sndh */
TIMEDB_ENTRY( 6a16fb10,  0,  21528,    YM+TB+TD), /* 03986 - XyNo - n-a.sndh */
TIMEDB_ENTRY( 6a286f56,  0,   1181,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  1,    518,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  2,    731,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  3,    341,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  4,    112,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  5,    160,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  6,   1280,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  7,   1828,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  8,   1828,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a286f56,  9,   1219,          YM), /* 03325 - Darren Ithell - Mouth Trap.sndh */
TIMEDB_ENTRY( 6a2c6a9f,  0,   7680,          YM), /* 01127 - Tyan - Revolution - Car Racing.sndh */
TIMEDB_ENTRY( 6a334ab5,  0,   2167,          YM), /* 03054 - Ed Bogas - Harmony.sndh */
TIMEDB_ENTRY( 6a5e21bb,  0,   4096,          YM), /* 00731 - Kevin Collier - Harley Davison.sndh */
TIMEDB_ENTRY( 6a5e21bb,  1,    182,          YM), /* 00731 - Kevin Collier - Harley Davison.sndh */
TIMEDB_ENTRY( 6a5e21bb,  2,    153,          YM), /* 00731 - Kevin Collier - Harley Davison.sndh */
TIMEDB_ENTRY( 6a5e21bb,  3,    307,          YM), /* 00731 - Kevin Collier - Harley Davison.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  0,  16415,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  1,    122,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  2,   5475,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  3,  35983,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  4,   8842,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  5,   3015,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  6,   2667,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  7,  18124,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  8,  61739,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab,  9,   9062,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab, 10,   8622,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab, 11,    924,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab, 12,  53642,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a6bd1ab, 13,   2870,          YM), /* 04393 - Unknown - Space Harrier.sndh */
TIMEDB_ENTRY( 6a73db9c,  0,   8703, YM+TA+TB+TD), /* 03831 - Rhino - Torment (D Newman) - Bumble Beezer.sndh */
TIMEDB_ENTRY( 6a77e267,  0,  16801,       YM+TA), /* 04029 - Roggie - Spellbound.sndh */
TIMEDB_ENTRY( 6a79cf52,  0,   1360,          YM), /* 04548 - Iain Young - Hop.sndh */
TIMEDB_ENTRY( 6abe41d0,  0,   6534,       YM+TA), /* 04563 - Scavenger - Knuckle Joe - Extended.sndh */
TIMEDB_ENTRY( 6ac1f37c,  0,   8960,    YM+TB+TD), /* 02905 - heleenh - juuujuu.sndh */
TIMEDB_ENTRY( 6ad0abf2,  0, 119726,       YM+TA), /* 04021 - Roggie - Quasar.sndh */
TIMEDB_ENTRY( 6aeffc55,  0,  79648,       YM+TA), /* 03825 - TEC - TEC Demo - Intro.sndh */
TIMEDB_ENTRY( 6af6a6ae,  0,   7680,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  1,   7974,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  2,   7440,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  3,   9545,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  4,   8832,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  5,  10304,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6af6a6ae,  6,   8611,          YM), /* 01377 - Justin Scharvona - Last Ninja 3.sndh */
TIMEDB_ENTRY( 6afddd06,  0,   5975,          YM), /* 00330 - Gary Wheaton - Tangerine Dream 1.sndh */
TIMEDB_ENTRY( 6b0ab29b,  0,   3072,          YM), /* 01183 - Excellence In Art - Heathers.sndh */
TIMEDB_ENTRY( 6b0e6a79,  0,    127,          YM), /* 04520 - Vivian Peschardt - Denver Je Decouvre Les Animaux.sndh */
TIMEDB_ENTRY( 6b12fd37,  0,  12288,    YM+TA+TD), /* 02268 - Tao - Cream - Sunriders (Steps).sndh */
TIMEDB_ENTRY( 6b26103c,  0,   2240,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  1,   6496,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  2,    321,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  3,    320,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  4,   4960,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  5,    448,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  6,   2670,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  7,   4000,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  8,   1280,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c,  9,   2688,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b26103c, 10,   2400,          YM), /* 01502 - Dave Hasler - Ninja Spirit.sndh */
TIMEDB_ENTRY( 6b3041dd,  0,  13818,          YM), /* 01027 - Doclands - Frenetic.sndh */
TIMEDB_ENTRY( 6b46576e,  0,   2105, YM+TA+TB+TD), /* 01165 - Tyan - Ulk.sndh */
TIMEDB_ENTRY( 6b60c6cd,  0,    229,          YM), /* 04523 - Richard Brough - Drip.sndh */
TIMEDB_ENTRY( 6b60c6cd,  1,   1745,          YM), /* 04523 - Richard Brough - Drip.sndh */
TIMEDB_ENTRY( 6b60c6cd,  2,    903,          YM), /* 04523 - Richard Brough - Drip.sndh */
TIMEDB_ENTRY( 6b62e9cc,  0,  54286,       YM+TA), /* 04027 - Roggie - Inferno.sndh */
TIMEDB_ENTRY( 6b780fd1,  0,  32348,       YM+TA), /* 03783 - Miggy - Mig.sndh */
TIMEDB_ENTRY( 6b81bbea,  0,    319,          YM), /* 02471 - Iso - Kop Houwe!.sndh */
TIMEDB_ENTRY( 6b85ea13,  0,   5567, YM+TA+TB+TD), /* 03002 - 505-paradox - thrust.sndh */
TIMEDB_ENTRY( 6b94f3db,  0,  36864, YM+TA+TB+TD), /* 03425 - cerror - DATASS.sndh */
TIMEDB_ENTRY( 6babbb1e,  0,  55634,       YM+TA), /* 04810 - Unknown - Tear Drop.sndh */
TIMEDB_ENTRY( 6bb5a80b,  0,   3240,          YM), /* 00562 - Allister Brimble - Terry's Big Adventure.sndh */
TIMEDB_ENTRY( 6bb5a80b,  1,    384,          YM), /* 00562 - Allister Brimble - Terry's Big Adventure.sndh */
TIMEDB_ENTRY( 6bb5a80b,  2,    128,          YM), /* 00562 - Allister Brimble - Terry's Big Adventure.sndh */
TIMEDB_ENTRY( 6bb5a80b,  3,    240,          YM), /* 00562 - Allister Brimble - Terry's Big Adventure.sndh */
TIMEDB_ENTRY( 6bb8e642,  0,  10752,    YM+TA+TD), /* 00832 - Crazy Q - - TiMeBoMb - - Swingers Club.sndh */
TIMEDB_ENTRY( 6bb98b4e,  0,  11521,          YM), /* 00765 - Count Zero - Decade Demo - Intro.sndh */
TIMEDB_ENTRY( 6bc7cbc1,  0,   9220,    YM+TA+TD), /* 02302 - Tao - Cream - Breath (MSB EIL99 demo).sndh */
TIMEDB_ENTRY( 6bce9bb7,  0,   1792,          YM), /* 01504 - Rob Hubbard - Battleships.sndh */
TIMEDB_ENTRY( 6bce9bb7,  1,    289,          YM), /* 01504 - Rob Hubbard - Battleships.sndh */
TIMEDB_ENTRY( 6c287539,  0,   7296,          YM), /* 01205 - Excellence In Art - Sweep.sndh */
TIMEDB_ENTRY( 6c287f4e,  0,   6526,          YM), /* 01321 - Matt Furniss - Subbuteo.sndh */
TIMEDB_ENTRY( 6c2c0a56,  0,   3847,          YM), /* 02721 - Jeremy Das & Nik Sen - Locomotion.sndh */
TIMEDB_ENTRY( 6c3c6768,  0,   1634,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  1,   2178,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  2,    339,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  3,    228,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  4,    181,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  5,   2497,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  6,    176,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  7,    135,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  8,    126,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768,  9,   3183,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 10,   3138,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 11,   2560,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 12,    482,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 13,    229,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 14,    369,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 15,    264,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 17,   2688,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 18,    205,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 19,    419,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 21,    512,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c3c6768, 22,   2285,          YM), /* 01492 - Christopher Guard - Skate Tribe.sndh */
TIMEDB_ENTRY( 6c4168b0,  0,    360,          YM), /* 01715 - Jess - The Music Bootsector.sndh */
TIMEDB_ENTRY( 6c442139,  0,   3839, YM+TA+TB+TD), /* 03518 - D-Force - Shure Of One Thing.sndh */
TIMEDB_ENTRY( 6c51974e,  0,    775,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  1,    288,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  2,   2872,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  3,    605,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  4,   1542,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  5,     11,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  6,    235,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  7,     31,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  8,     34,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e,  9,     34,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e, 10,     68,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e, 11,     31,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c51974e, 12,     30,          YM), /* 00845 - Ben Daglish - Chase H.Q..sndh */
TIMEDB_ENTRY( 6c5fd6de,  0,  12800, YM+TA+TB+TD), /* 02092 - Mad Max - Phantoms of the Asteroid - Sid.sndh */
TIMEDB_ENTRY( 6c5fd6de,  1,   6912, YM+TA+TB+TD), /* 02092 - Mad Max - Phantoms of the Asteroid - Sid.sndh */
TIMEDB_ENTRY( 6c614484,  0,  23679, YM+TA+TB+TD), /* 02978 - 505-paradox - certain rage (STE only).sndh */
TIMEDB_ENTRY( 6c85ad49,  0,  72549,       YM+TA), /* 03803 - Sparky - Pompey Pirates Menu 84.sndh */
TIMEDB_ENTRY( 6ca221af,  0,  11545,          YM), /* 01775 - Lap - Lap 15.sndh */
TIMEDB_ENTRY( 6cab176f,  0,   3848,       YM+TA), /* 00906 - Danny Oneway - Lamebrain 2 - Loader.sndh */
TIMEDB_ENTRY( 6cb9071a,  0, 105727,       YM+TA), /* 03708 - Sparky - Magnetic Fields 2 (Jean-Michel Jarre).sndh */
TIMEDB_ENTRY( 6cd391e6,  0,  24320,          YM), /* 01452 - Stax - Stax 12.sndh */
TIMEDB_ENTRY( 6cda38e9,  0,   6912,          YM), /* 03663 - Jedi - The Ultimate Boothunt.sndh */
TIMEDB_ENTRY( 6ce115b5,  0,   2304,          YM), /* 01136 - Tyan - Revolution -John Player Theme.sndh */
TIMEDB_ENTRY( 6cf1bfac,  0,  16896,    YM+TB+TD), /* 02854 - Helene - Noname (auf h's tagebuch).sndh */
TIMEDB_ENTRY( 6cffdc1b,  0,   4605, YM+TA+TB+TD), /* 00030 - 505 - Tribal.sndh */
TIMEDB_ENTRY( 6d060e40,  0,   8902, YM+TA+TB+TD), /* 03004 - 505 - Nowhere.sndh */
TIMEDB_ENTRY( 6d189261,  0,   9360,          YM), /* 02365 - Nexus 6 - Bubbling Overall.sndh */
TIMEDB_ENTRY( 6d323657,  0,  14388,       YM+TA), /* 02681 - Donald Fakk - Earth is Rotten (Chip).sndh */
TIMEDB_ENTRY( 6d5eb2eb,  0,  31320, YM+TA+TB+TD), /* 03630 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 6d5f8793,  0,  74161,       YM+TA), /* 04802 - Unknown - Killer.sndh */
TIMEDB_ENTRY( 6d63857c,  0,   4991,    YM+TA+TB), /* 03364 - forestcreature - bois de la biere.sndh */
TIMEDB_ENTRY( 6d66fa57,  0,  14822,       YM+TA), /* 02667 - Donald Fakk - Echoing (Chip).sndh */
TIMEDB_ENTRY( 6d68490c,  0,   1730,          YM), /* 00383 - Wally Beben aka Hagar - Foundation Waste.sndh */
TIMEDB_ENTRY( 6d68490c,  1,   9217,          YM), /* 00383 - Wally Beben aka Hagar - Foundation Waste.sndh */
TIMEDB_ENTRY( 6d6889f0,  0,  12722, YM+TA+TB+TD), /* 03640 - tomchi    NoeXtra - ltc 08 sheep n chip.sndh */
TIMEDB_ENTRY( 6d6b1568,  0,   7484,          YM), /* 02649 - Dma-Sc - Sector One - Ninja Action Fighter.sndh */
TIMEDB_ENTRY( 6d781d57,  0,   5026,          YM), /* 04154 - Unknown - Michelle.sndh */
TIMEDB_ENTRY( 6d7df9cb,  0,   5628,          YM), /* 00857 - Ben Daglish - Gary Lineker's Hot Shots.sndh */
TIMEDB_ENTRY( 6d7df9cb,  1,    266,          YM), /* 00857 - Ben Daglish - Gary Lineker's Hot Shots.sndh */
TIMEDB_ENTRY( 6d7df9cb,  2,   2935,          YM), /* 00857 - Ben Daglish - Gary Lineker's Hot Shots.sndh */
TIMEDB_ENTRY( 6d87e6e4,  0,   7392, YM+TA+TB+TD), /* 02080 - Mad Max - Human Race - Sid.sndh */
TIMEDB_ENTRY( 6d87e6e4,  1,   3456, YM+TA+TB+TD), /* 02080 - Mad Max - Human Race - Sid.sndh */
TIMEDB_ENTRY( 6d87e6e4,  2,   3840, YM+TA+TB+TD), /* 02080 - Mad Max - Human Race - Sid.sndh */
TIMEDB_ENTRY( 6d87e6e4,  3,   8064, YM+TA+TB+TD), /* 02080 - Mad Max - Human Race - Sid.sndh */
TIMEDB_ENTRY( 6d87e6e4,  4,   5184, YM+TA+TB+TD), /* 02080 - Mad Max - Human Race - Sid.sndh */
TIMEDB_ENTRY( 6d99a2d5,  0,   3180,          YM), /* 01073 - Excellence In Art - Bebop.sndh */
TIMEDB_ENTRY( 6da4b105,  0,   1570,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6da4b105,  1,   1172,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6da4b105,  2,    789,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6da4b105,  3,     78,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6da4b105,  4,    668,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6da4b105,  5,    603,          YM), /* 00854 - Ben Daglish - Hotrod.sndh */
TIMEDB_ENTRY( 6dc56de0,  0,   5370, YM+TA+TB+TD), /* 00983 - D-Force - Plink.sndh */
TIMEDB_ENTRY( 6de426b0,  0,   3996,          YM), /* 03324 - Unknown - Medway Boys 11.sndh */
TIMEDB_ENTRY( 6e0f68c6,  0,   3468,          YM), /* 03082 - Jim Thomson - Four Leaf Clover.sndh */
TIMEDB_ENTRY( 6e2d7de0,  0,  17653,          YM), /* 02290 - Tao - Cream (orig. Rob Hubbard) - Ace II.sndh */
TIMEDB_ENTRY( 6e34397e,  0,   9600,          YM), /* 00972 - Deneb - Plucked Goose - Fantasy.sndh */
TIMEDB_ENTRY( 6e63bcd2,  0,  18240,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  1,    470,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  2,    599,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  3,    226,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  4,    370,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  5,    406,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  6,    447,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  7,    396,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  8,    299,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2,  9,    466,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2, 10,    442,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2, 11,    310,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e63bcd2, 12,    875,          YM), /* 02019 - Mad Max - Aufweidersehen Monty.sndh */
TIMEDB_ENTRY( 6e92fc60,  0,   1820,          YM), /* 00142 - David Whittaker - Startrek.sndh */
TIMEDB_ENTRY( 6e967386,  0,  14208,          YM), /* 01875 - Dave Lowe - Afterburner.sndh */
TIMEDB_ENTRY( 6e967386,  1,  12416,          YM), /* 01875 - Dave Lowe - Afterburner.sndh */
TIMEDB_ENTRY( 6e967386,  2,   7400,          YM), /* 01875 - Dave Lowe - Afterburner.sndh */
TIMEDB_ENTRY( 6e967386,  3,   4480,          YM), /* 01875 - Dave Lowe - Afterburner.sndh */
TIMEDB_ENTRY( 6e967386,  4,   3840,          YM), /* 01875 - Dave Lowe - Afterburner.sndh */
TIMEDB_ENTRY( 6ebc3157,  0,  13050,       YM+TA), /* 01453 - Stax - Tauris - Digi.sndh */
TIMEDB_ENTRY( 6ebd51f9,  0,   7879,          YM), /* 03161 - n-a - .sndh */
TIMEDB_ENTRY( 6edf2455,  0,   5697,          YM), /* 00382 - Wally Beben - Elite.sndh */
TIMEDB_ENTRY( 6ee1a637,  0,  12096,          YM), /* 00596 - Chris Mad - Overdrive Dots.sndh */
TIMEDB_ENTRY( 6ee9c6f1,  0,   3072,          YM), /* 04005 - Riff Raff - Protector.sndh */
TIMEDB_ENTRY( 6ef4d709,  0,   2044,    YM+TA+TB), /* 04280 - 505 - Loop03.sndh */
TIMEDB_ENTRY( 6f0dbf68,  0,   6144,          YM), /* 04830 - T1000 - Bits 55c.sndh */
TIMEDB_ENTRY( 6f19a42b,  0,  15361,       YM+TA), /* 04807 - Unknown - Oh No.sndh */
TIMEDB_ENTRY( 6f19d560,  0,  70627,          YM), /* 01797 - Barry Leitch - Impossamole.sndh */
TIMEDB_ENTRY( 6f19d560,  1,  75267,          YM), /* 01797 - Barry Leitch - Impossamole.sndh */
TIMEDB_ENTRY( 6f19d560,  2,  24579,          YM), /* 01797 - Barry Leitch - Impossamole.sndh */
TIMEDB_ENTRY( 6f1bf9fb,  0,   1535, YM+TA+TB+TD), /* 02710 - WOTW . Remake by Dubmood - intro 101.sndh */
TIMEDB_ENTRY( 6f2274c6,  0,   5766,          YM), /* 02000 - Mad Max - Car Race.sndh */
TIMEDB_ENTRY( 6f2ceb7a,  0,  38912,          YM), /* 02805 - Starbuck - A new Beginning.sndh */
TIMEDB_ENTRY( 6f3f1de7,  0,   9984,       YM+TD), /* 00585 - Celtic - I'm The Deiffel! - TFU End.sndh */
TIMEDB_ENTRY( 6f68a236,  0,   5958,       YM+TA), /* 01349 - Scavenger - Nadaaaaaaaaaahhh!.sndh */
TIMEDB_ENTRY( 6f74f173,  0,   6207, YM+TA+TB+TD), /* 02561 - marcer - hybris remix.sndh */
TIMEDB_ENTRY( 6f941621,  0,   9409,          YM), /* 00424 - Big Alec - PYM - Copperkaahbaahnaah.sndh */
TIMEDB_ENTRY( 6f9c6fe4,  0,    762, YM+TA+TB+TD), /* 04110 - 505-cpt^dox - jingle01.sndh */
TIMEDB_ENTRY( 6fa98978,  0,   3360,          YM), /* 04695 - Dean Sharples - Mario's Quest.sndh */
TIMEDB_ENTRY( 6fa98978,  1,    335,          YM), /* 04695 - Dean Sharples - Mario's Quest.sndh */
TIMEDB_ENTRY( 6fa98978,  2,   1828,          YM), /* 04695 - Dean Sharples - Mario's Quest.sndh */
TIMEDB_ENTRY( 6fa98978,  3,    630,          YM), /* 04695 - Dean Sharples - Mario's Quest.sndh */
TIMEDB_ENTRY( 6fb7b5be,  0,   2977,    YM+TA+TB), /* 00210 - 505 - UCM 11 Intro Greets.sndh */
TIMEDB_ENTRY( 6fc13ab4,  0,   1152,          YM), /* 00964 - Mike Davies - Italia 90.sndh */
TIMEDB_ENTRY( 6fc13ab4,  1,    473,          YM), /* 00964 - Mike Davies - Italia 90.sndh */
TIMEDB_ENTRY( 6fc13ab4,  2,    768,          YM), /* 00964 - Mike Davies - Italia 90.sndh */
TIMEDB_ENTRY( 6fc13ab4,  3,   5376,          YM), /* 00964 - Mike Davies - Italia 90.sndh */
TIMEDB_ENTRY( 6fcaeefd,  0,   6912, YM+TA+TB+TD), /* 02085 - Mad Max - The Last V8 - Sid.sndh */
TIMEDB_ENTRY( 6fde453e,  0,    241,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e,  1,    473,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e,  2,    390,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e,  3,    310,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e,  4,    248,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e, 29,     75,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 6fde453e, 30,     75,          YM), /* 04714 - Terry Mancey and Kevin Hearson - Reels of Fortune.sndh */
TIMEDB_ENTRY( 700163c3,  0,   2240,          YM), /* 01109 - Keith Tinman - Batman - The Caped Crusader.sndh */
TIMEDB_ENTRY( 7027d331,  0,   5376,          YM), /* 01204 - Excellence In Art - Sweat.sndh */
TIMEDB_ENTRY( 702e32b4,  0,  15899, YM+TA+TB+TD), /* 02794 - Frequent-Ephidrena (2006) - Lambo.sndh */
TIMEDB_ENTRY( 703de794,  0,   7040,          YM), /* 00126 - David Whittaker - Rampage.sndh */
TIMEDB_ENTRY( 703de794,  1,    570,          YM), /* 00126 - David Whittaker - Rampage.sndh */
TIMEDB_ENTRY( 70605074,  0,  82397,       YM+TA), /* 04043 - Roggie - Haggies Man.sndh */
TIMEDB_ENTRY( 7060eecb,  0,   7035, YM+TA+TB+TD), /* 00084 - 505 - Modulation 2.sndh */
TIMEDB_ENTRY( 706b10c4,  0,   6342,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  1,   9228,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  2,   5772,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  3,   6413,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  4,   3213,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  5,  12179,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706b10c4,  6,    559,          YM), /* 00853 - Ben Daglish - Hate.sndh */
TIMEDB_ENTRY( 706ccaaa,  0,   5448,       YM+TD), /* 04315 - 505 - Switch.sndh */
TIMEDB_ENTRY( 7075ae3d,  0,  10405, YM+TA+TB+TD), /* 01590 - Nemo - Milliways - Online Compo.sndh */
TIMEDB_ENTRY( 707ad6e5,  0,   1728,          YM), /* 02208 - Tao - Cream - Madness (loader).sndh */
TIMEDB_ENTRY( 7081fb45,  0,   4147,          YM), /* 03935 - Unknown - Udssr.sndh */
TIMEDB_ENTRY( 708b613b,  0,   1792,          YM), /* 04836 - t1000 - Bits 30.sndh */
TIMEDB_ENTRY( 708bf8fa,  0,   9984,          YM), /* 00415 - Big Alec - Musical Wonder.sndh */
TIMEDB_ENTRY( 70ac75cf,  0,   9985,    YM+TA+TD), /* 02349 - Tao-Cream - Cutie (UCM 25).sndh */
TIMEDB_ENTRY( 70af9e1c,  0,   4575,          YM), /* 03151 - n-a - .sndh */
TIMEDB_ENTRY( 70b8e905,  0,  18241,       YM+TA), /* 04055 - Roggie - Rissoto supreme & custard.sndh */
TIMEDB_ENTRY( 70be9cba,  0,  17671,          YM), /* 00437 - Big Alec - Stand By.sndh */
TIMEDB_ENTRY( 70e87410,  0,   9394,          YM), /* 03099 - n-a - Children's Corner # 1.sndh */
TIMEDB_ENTRY( 7109672b,  0,   3623,          YM), /* 00165 - Martin Walker - D-Generation.sndh */
TIMEDB_ENTRY( 71144213,  0,   4106,          YM), /* 03110 - Unknown - The Entertainer-Scott Joplin.sndh */
TIMEDB_ENTRY( 711628dd,  0,   6335,          YM), /* 04082 - Gary Wheaton - Miner 2049er.sndh */
TIMEDB_ENTRY( 711f39f2,  0,   4995,          YM), /* 03543 - David Whittaker - Graham Gooch World Class Cricket.sndh */
TIMEDB_ENTRY( 713a75a3,  0,   3535,          YM), /* 01559 - Dave Rogers - Ranarama.sndh */
TIMEDB_ENTRY( 713a75a3,  1,    235,          YM), /* 01559 - Dave Rogers - Ranarama.sndh */
TIMEDB_ENTRY( 713a75a3,  2,    157,          YM), /* 01559 - Dave Rogers - Ranarama.sndh */
TIMEDB_ENTRY( 713a75a3,  3,    190,          YM), /* 01559 - Dave Rogers - Ranarama.sndh */
TIMEDB_ENTRY( 713dd020,  0,   3949,          YM), /* 01487 - Holger Gehrmann - Space Port.sndh */
TIMEDB_ENTRY( 713dd020,  1,   2880,          YM), /* 01487 - Holger Gehrmann - Space Port.sndh */
TIMEDB_ENTRY( 713dd020,  2,    504,          YM), /* 01487 - Holger Gehrmann - Space Port.sndh */
TIMEDB_ENTRY( 7166323c,  0,   8450,          YM), /* 04628 - tomchi  noextra - ltc03 0ompa main.sndh */
TIMEDB_ENTRY( 716939e6,  0,   3839, YM+TA+TB+TD), /* 02704 - DUBMOOD - SPEEDLOADER II.sndh */
TIMEDB_ENTRY( 717aa801,  0,   8386,          YM), /* 00875 - Ben Daglish - Switchblade.sndh */
TIMEDB_ENTRY( 717aa801,  1,  10111,          YM), /* 00875 - Ben Daglish - Switchblade.sndh */
TIMEDB_ENTRY( 717aa801,  2,    137,          YM), /* 00875 - Ben Daglish - Switchblade.sndh */
TIMEDB_ENTRY( 719aad6e,  0,   1337,       YM+TA), /* 01352 - Scavenger - Ode to Stash.sndh */
TIMEDB_ENTRY( 719ba61d,  0,  17899,       YM+TA), /* 03739 - David Whittaker - Snow Strike.sndh */
TIMEDB_ENTRY( 71aacc37,  0,   1620,          YM), /* 03071 - Chris French - We Wish You A Merry Christmas.sndh */
TIMEDB_ENTRY( 71b214ce,  0,  57025,    YM+TB+TD), /* 02714 - naPO - AC 2006.sndh */
TIMEDB_ENTRY( 71c1575b,  0,   9984,          YM), /* 01562 - Rude Kid - Knuff - Golden Dawn 52.sndh */
TIMEDB_ENTRY( 71ca34bd,  0,   9381, YM+TA+TB+TD), /* 00032 - 505 - Aragorn - UCM 22.sndh */
TIMEDB_ENTRY( 71e67240,  0,   3200,          YM), /* 02437 - Gari Biasillo - Slayer.sndh */
TIMEDB_ENTRY( 71e67240,  1,    443,          YM), /* 02437 - Gari Biasillo - Slayer.sndh */
TIMEDB_ENTRY( 71e67240,  2,   4224,          YM), /* 02437 - Gari Biasillo - Slayer.sndh */
TIMEDB_ENTRY( 7201cbb3,  0,  12544,          YM), /* 01498 - Alex Herbert - Turbo.sndh */
TIMEDB_ENTRY( 7201cbb3,  1,   3968,          YM), /* 01498 - Alex Herbert - Turbo.sndh */
TIMEDB_ENTRY( 7201cbb3,  2,   6144,          YM), /* 01498 - Alex Herbert - Turbo.sndh */
TIMEDB_ENTRY( 7201cbb3,  3,  17676,          YM), /* 01498 - Alex Herbert - Turbo.sndh */
TIMEDB_ENTRY( 720c8d9b,  0,   9594, YM+TA+TB+TD), /* 01010 - DMA-SC - Pre Midnight Pizza.sndh */
TIMEDB_ENTRY( 7211ab87,  0,  17080, YM+TA+TB+TD), /* 02076 - Mad Max - Flash Gordon - Sid.sndh */
TIMEDB_ENTRY( 721728d7,  0,    896,          YM), /* 00958 - Dark Angel - Your Turn.sndh */
TIMEDB_ENTRY( 723e22d3,  0,   8057,       YM+TD), /* 01622 - No More - Night Walk.sndh */
TIMEDB_ENTRY( 724620eb,  0,   7336,          YM), /* 03121 - n-a - Homage to ANDRES SEGOVIA.sndh */
TIMEDB_ENTRY( 725e3a5a,  0,  20544, YM+TA+TB+TD), /* 02526 - Salkin-047 - Hero.sndh */
TIMEDB_ENTRY( 7269fca9,  0,  94672,       YM+TA), /* 03726 - Spaz - Dark Side of the Spoon - TLB.sndh */
TIMEDB_ENTRY( 7271e4a4,  0,   8446,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  1,  16131,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  2,   1824,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  3,  94983,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  4,    901,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  5,   7559,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  6,  29187,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 7271e4a4,  7,   4611,          YM), /* 04395 - Barry Leitch - Back to the Future Part III.sndh */
TIMEDB_ENTRY( 727b0b84,  0,   2304,          YM), /* 00485 - Ultimatum - Kidney Bean 1.sndh */
TIMEDB_ENTRY( 727f9716,  0,  87916,       YM+TA), /* 04670 - Kijop - Raaap.sndh */
TIMEDB_ENTRY( 729567dc,  0,   7648,          YM), /* 03128 - n-a - Jump.sndh */
TIMEDB_ENTRY( 72994953,  0,   6858,          YM), /* 00123 - David Whittaker - Panther.sndh */
TIMEDB_ENTRY( 72b9137c,  0,   9217,          YM), /* 00431 - Big Alec - Poppycock - Sprites.sndh */
TIMEDB_ENTRY( 72c1d405,  0,  11332,          YM), /* 03608 - Tomchi  Noextra - AC 2010 invit.sndh */
TIMEDB_ENTRY( 72e1cb0a,  0,   6136, YM+TA+TB+TD), /* 00274 - Baggio - BAJS2.sndh */
TIMEDB_ENTRY( 72e28105,  0,   4991, YM+TA+TB+TD), /* 03021 - dubmood - RaZoR 1911 - Vains we are da indians.sndh */
TIMEDB_ENTRY( 731b4358,  0,  10752,          YM), /* 04391 - Tomchi - Treasue Island intro.sndh */
TIMEDB_ENTRY( 731c0738,  0,   4608,          YM), /* 00801 - Crawdaddy - Whatta Nerd!.sndh */
TIMEDB_ENTRY( 733a2c97,  0,  47081, YM+TA+TB+TD), /* 03993 - XyNo - WALEBRKS!.sndh */
TIMEDB_ENTRY( 73515f35,  0,   4352,          YM), /* 00289 - Steve Wheterill - Flimbo's Quest.sndh */
TIMEDB_ENTRY( 73515f35,  1,   6032,          YM), /* 00289 - Steve Wheterill - Flimbo's Quest.sndh */
TIMEDB_ENTRY( 73515f35,  2,   2688,          YM), /* 00289 - Steve Wheterill - Flimbo's Quest.sndh */
TIMEDB_ENTRY( 736af191,  0,   1120,          YM), /* 04679 - Andy Kauling - Ininity One.sndh */
TIMEDB_ENTRY( 736af191,  1,   2419,          YM), /* 04679 - Andy Kauling - Ininity One.sndh */
TIMEDB_ENTRY( 736af191,  2,   2315,          YM), /* 04679 - Andy Kauling - Ininity One.sndh */
TIMEDB_ENTRY( 7380fc26,  0,  10362,       YM+TA), /* 01932 - Miq - First Song.sndh */
TIMEDB_ENTRY( 7383fbd6,  0,   7840,          YM), /* 00106 - David Whittaker - Goldrunner 2.sndh */
TIMEDB_ENTRY( 7383fbd6,  1,   2352,          YM), /* 00106 - David Whittaker - Goldrunner 2.sndh */
TIMEDB_ENTRY( 7383fbd6,  2,   1280,          YM), /* 00106 - David Whittaker - Goldrunner 2.sndh */
TIMEDB_ENTRY( 73857d7b,  0,  11641,       YM+TD), /* 04791 - Intruder - Intruder 7.sndh */
TIMEDB_ENTRY( 7387cfd9,  0,  41491,    YM+TA+TD), /* 02312 - Tao - Cream - Intensity (200Hz).sndh */
TIMEDB_ENTRY( 73a7adbe,  0,   5436,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73a7adbe,  1,   3067,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73a7adbe,  2,     75,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73a7adbe,  3,     71,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73a7adbe,  4,     91,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73a7adbe,  5,    154,          YM), /* 00555 - Allister Brimble - Mig 29.sndh */
TIMEDB_ENTRY( 73b5d83c,  0,   3584,          YM), /* 01886 - Dave Lowe - Microprose Formula One Grand Prix.sndh */
TIMEDB_ENTRY( 73bd8ff6,  0,   2107,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  1,   2604,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  2,   1174,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  3,   5124,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  4,   2074,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  5,   2254,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  6,   1174,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73bd8ff6,  7,  59098,          YM), /* 04727 - Beyonder - MIX number ONE demo - Sicmu.sndh */
TIMEDB_ENTRY( 73c42f88,  0,   5370, YM+TA+TB+TD), /* 00909 - Danny Oneway - No Sheep In Mongolia (SID).sndh */
TIMEDB_ENTRY( 73c98d09,  0,   5762,          YM), /* 00508 - Tangens - Synthetic Visions - Dister.sndh */
TIMEDB_ENTRY( 73d9805c,  0,   4754,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  1,   2217,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  2,   1828,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  3,   2048,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  4,   1536,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  5,   3072,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  6,    960,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  7,    752,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  8,   3040,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c,  9,   1664,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 10,   1360,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 11,   1852,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 12,   3840,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 13,    853,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 14,   4848,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73d9805c, 15,    387,          YM), /* 04709 - Vesa, Emphii, Bad Boy - Razor's Mega Demo - Screen 4.sndh */
TIMEDB_ENTRY( 73ee82dc,  0,  25536, YM+TA+TB+TD), /* 02102 - Mad Max - War - Sid.sndh */
TIMEDB_ENTRY( 73f88dc8,  0,    512,    YM+TA+TD), /* 02812 - Crazy Q - DHS - GAME OVER TEST.sndh */
TIMEDB_ENTRY( 74190da8,  0,  23034,       YM+TA), /* 01706 - Jess - Protect your Environment.sndh */
TIMEDB_ENTRY( 741c5eb1,  0,   6139,          YM), /* 04373 - Jess - Tears On My Keyboard (alpha).sndh */
TIMEDB_ENTRY( 7423d024,  0,    767, YM+TA+TB+TD), /* 02705 - DUBBIE - STILL D.U.B.sndh */
TIMEDB_ENTRY( 742963cd,  0,   3066,       YM+TA), /* 02623 - STU - ULTIMATE.sndh */
TIMEDB_ENTRY( 744b14ae,  0,   4992,    YM+TA+TD), /* 04245 - 505-checkpoint^paradox - Blark.sndh */
TIMEDB_ENTRY( 7451cebe,  0,   1785, YM+TA+TB+TD), /* 01963 - MSG - Godpey - Results.sndh */
TIMEDB_ENTRY( 7458dea7,  0,   4281, YM+TA+TB+TD), /* 04326 - 505 - Checkpoint - Yaa Taaru (unreleased).sndh */
TIMEDB_ENTRY( 7462b4a6,  0,  21375, YM+TA+TB+TD), /* 03023 - Dubmood - RZR^Dbug^YMR - Detta ar DDR!.sndh */
TIMEDB_ENTRY( 74754556,  0,   5952,          YM), /* 00816 - Crawdaddy - The Village.sndh */
TIMEDB_ENTRY( 747eae7e,  0,   9472,       YM+TA), /* 00810 - Crawdaddy - Seekers (Cross The Deserts).sndh */
TIMEDB_ENTRY( 7485c843,  0,   7674,          YM), /* 01238 - Epic - Stunt Priest.sndh */
TIMEDB_ENTRY( 7494a084,  0,   6672,          YM), /* 00905 - Danny Oneway - Lamebrain 2 - Song.sndh */
TIMEDB_ENTRY( 74ae2b50,  0,  12288,          YM), /* 03256 - T1000 - The Beast Club - TS 4.sndh */
TIMEDB_ENTRY( 74b82316,  0,   1926,          YM), /* 03382 - Unknown - Bach.sndh */
TIMEDB_ENTRY( 74bc3732,  0,   7296,          YM), /* 01641 - JB 007 - Climax - Plastro.sndh */
TIMEDB_ENTRY( 74e07e3c,  0,   2304,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  1,    191,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  2,     81,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  3,    255,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  4,    512,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  5,    648,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  6,    288,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  7,     86,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74e07e3c,  8,    576,          YM), /* 01256 - Tim Follin - Bubble Bobble.sndh */
TIMEDB_ENTRY( 74ecaa22,  0,  14680,          YM), /* 02110 - Mad Max - Galway Mix.sndh */
TIMEDB_ENTRY( 74f2ab37,  0,   9216,    YM+TA+TD), /* 01360 - Scavenger - Primus.sndh */
TIMEDB_ENTRY( 750cba32,  0,   7680,          YM), /* 01550 - Rhino - GBISTCP - Torment.sndh */
TIMEDB_ENTRY( 750f2a72,  0,   8280,          YM), /* 00740 - Connor - UCM 2 Intro.sndh */
TIMEDB_ENTRY( 75176418,  0,  12287, YM+TA+TB+TD), /* 02486 - gwEm (M. Gray conversion) - Bangin' Nights.sndh */
TIMEDB_ENTRY( 75478a17,  0,  41216, YM+TA+TB+TD), /* 03426 - cerror - A BEGINNING.sndh */
TIMEDB_ENTRY( 7549d578,  0,   5095,          YM), /* 03141 - Unknown - Music Box Dancer - Frank Mills.sndh */
TIMEDB_ENTRY( 7562aff4,  0,  14665,       YM+TB), /* 01031 - Doclands - Zoomer Preview - FOF.sndh */
TIMEDB_ENTRY( 7563ab89,  0,   5120,          YM), /* 02118 - Mad Max - Warhawk Remix.sndh */
TIMEDB_ENTRY( 756f8880,  0,   4602,       YM+TA), /* 01350 - Scavenger - No Muls, Rapido?.sndh */
TIMEDB_ENTRY( 7577cc92,  0,  47528, YM+TA+TB+TD), /* 03427 - cerror - VERMOUTH.sndh */
TIMEDB_ENTRY( 757c0a86,  0,   6336,          YM), /* 03204 - Steve Wetherill - Tusker.sndh */
TIMEDB_ENTRY( 757c0a86,  1,   3840,          YM), /* 03204 - Steve Wetherill - Tusker.sndh */
TIMEDB_ENTRY( 757c4fb6,  0,   6527,    YM+TA+TD), /* 02821 - marcer - first tYMe (Znax credits-scores).sndh */
TIMEDB_ENTRY( 757e009b,  0,   2239,          YM), /* 01783 - Lap - Phaleon - Change Disks.sndh */
TIMEDB_ENTRY( 759f714f,  0,  43344, YM+TA+TB+TD), /* 03369 - STU - DISCO VER ATARI.sndh */
TIMEDB_ENTRY( 75a2df71,  0,    706,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  1,    110,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  2,    180,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  3,    187,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  4,    120,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  5,     69,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  6,   1632,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  7,    384,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  8,    206,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75a2df71,  9,      6,          YM), /* 01511 - Chris Huelsbeck - Carl Lewis Challenge.sndh */
TIMEDB_ENTRY( 75c72475,  0,   9408,          YM), /* 04475 - Charles Callet - Championship Water Skiing.sndh */
TIMEDB_ENTRY( 75d17603,  0,   8642,          YM), /* 01772 - Lap - Lap 12.sndh */
TIMEDB_ENTRY( 75e3431c,  0,    383,    YM+TA+TB), /* 03380 - crazy Pissed off Q - fucking untight shit !.sndh */
TIMEDB_ENTRY( 75eaa993,  0,  75293,       YM+TA), /* 04030 - Roggie - Spooked.sndh */
TIMEDB_ENTRY( 75f4089b,  0,  12540, YM+TA+TB+TD), /* 02452 - Timbral (2003) - Mandarine Street.sndh */
TIMEDB_ENTRY( 75f4293e,  0,    447,    YM+TA+TD), /* 04774 - dhor-mec - Bore.sndh */
TIMEDB_ENTRY( 76188b95,  0,   8192,          YM), /* 03657 - Jedi - DNT Paper 4 intro - Song 1.sndh */
TIMEDB_ENTRY( 76226833,  0,   5825,          YM), /* 01503 - Dave Hasler - The Punisher.sndh */
TIMEDB_ENTRY( 76226833,  1,   1251,          YM), /* 01503 - Dave Hasler - The Punisher.sndh */
TIMEDB_ENTRY( 7624935e,  0,   5376,          YM), /* 00595 - Chris Mad - Delirious Mainmenu.sndh */
TIMEDB_ENTRY( 76411619,  0,   8064,          YM), /* 01976 - MSG - MSG 06.sndh */
TIMEDB_ENTRY( 764c58b6,  0,   7609,          YM), /* 02429 - 505-cpt^dox - x2004.sndh */
TIMEDB_ENTRY( 764cf3c5,  0,   9210, YM+TA+TB+TD), /* 02759 - Dma-Sc - Sector One - Forgotten Figures.sndh */
TIMEDB_ENTRY( 764dd191,  0,   6254,          YM), /* 01206 - Excellence In Art - Sweep 2.sndh */
TIMEDB_ENTRY( 765ee6e5,  0,  15360,    YM+TA+TD), /* 02270 - Tao - Cream - The Cave (Steps).sndh */
TIMEDB_ENTRY( 765fada5,  0,   3584,       YM+TA), /* 00784 - Crawdaddy - The Desert.sndh */
TIMEDB_ENTRY( 766c708a,  0,  11904,          YM), /* 04837 - t1000 - Bits 31.sndh */
TIMEDB_ENTRY( 7677b922,  0,  80581,       YM+TA), /* 03724 - Spaz - Punish Your Machine - Reset.sndh */
TIMEDB_ENTRY( 76916ffc,  0,   5760,          YM), /* 04603 - Rude Kid - TV Themes.sndh */
TIMEDB_ENTRY( 76928172,  0,   8704, YM+TA+TB+TD), /* 04192 - Doclands - Transbeauce Demo 2 Hidden.sndh */
TIMEDB_ENTRY( 76b3d6b8,  0,   1792,       YM+TA), /* 00947 - Dark Angel - Shadows Of Death.sndh */
TIMEDB_ENTRY( 76b80b74,  0,   6525, YM+TA+TB+TD), /* 04335 - 505 - Chase.sndh */
TIMEDB_ENTRY( 76c7c9d2,  0,  15360, YM+TA+TB+TD), /* 02552 - Marcer - AT HOUSE SID.sndh */
TIMEDB_ENTRY( 76ce5c69,  0,   1694,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  1,   1907,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  2,    538,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  3,    555,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  4,    283,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  5,    397,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  6,    582,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  7,    617,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  8,    281,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69,  9,    120,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 10,    846,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 11,    873,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 12,    632,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 13,    917,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 14,    515,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 15,    699,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 16,    861,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 17,    610,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 18,    580,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 19,    473,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 20,    405,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 21,    384,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 22,    815,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 23,    408,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 24,   2024,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 25,    675,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 26,    300,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 27,    434,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 28,    493,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 29,    487,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 30,    831,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 31,     28,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 32,    161,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 33,    550,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 34,    461,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 35,    155,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 36,    952,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 37,    243,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 38,   1294,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 39,     82,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 40,    404,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 41,    768,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 42,   1801,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 43,   2722,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 44,    617,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 45,    109,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 46,    284,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 47,    540,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 48,    123,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 49,    301,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 50,    906,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 51,   1483,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 52,     18,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 53,    285,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 54,   1857,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 55,   1805,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 56,    976,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 57,   1036,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 58,    897,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 59,    995,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 60,   1110,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 61,   1864,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76ce5c69, 62,   1840,          YM), /* 04739 - William Goldstein - Kings Quest IV - The Perils of Rosella.sndh */
TIMEDB_ENTRY( 76d16088,  0,  13056,          YM), /* 02158 - Mad Max - So Watt - Nordik Coders.sndh */
TIMEDB_ENTRY( 76f03168,  0,  28416, YM+TA+TB+TD), /* 02912 - Helene - n-a.sndh */
TIMEDB_ENTRY( 77205c6a,  0,  11532,          YM), /* 03939 - Andreas Maier - Puzzlepuzzle.sndh */
TIMEDB_ENTRY( 7733372f,  0,   2304,          YM), /* 01429 - Starbuck - Praying Hands remix.sndh */
TIMEDB_ENTRY( 7745ec71,  0,   2048,          YM), /* 02931 - Chris Denman - Charge of the Light Brigade.sndh */
TIMEDB_ENTRY( 774f4078,  0,  13312,       YM+TD), /* 01984 - Jochen Hippel - "1".sndh */
TIMEDB_ENTRY( 7759b9fe,  0,  37898,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  1,  89397,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  2,  14351,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  3,  39361,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  4,   1665,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  5,  59777,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  6,  32769,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  7,  35845,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  8,  27460,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe,  9,  18574,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7759b9fe, 10,  55297,       YM+TA), /* 04352 - Edd the Duck - Sonixx Music demo.sndh */
TIMEDB_ENTRY( 7772a662,  0,   7168,       YM+TA), /* 00823 - !Cube - Chrome.sndh */
TIMEDB_ENTRY( 778a20d7,  0,  23856, YM+TA+TB+TD), /* 02587 - STU - DROP DA BOMB LIVE INTRO.sndh */
TIMEDB_ENTRY( 779bc20f,  0,   4880,          YM), /* 02367 - Nexus 6 - Colorclash.sndh */
TIMEDB_ENTRY( 779eef82,  0,   8320,       YM+TA), /* 01636 - Nexus - Maniax.sndh */
TIMEDB_ENTRY( 77b3ba68,  0,  10193,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  1,   2049,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  2,   3073,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  3,   3073,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  4,   2561,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  5,   2049,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b3ba68,  6,   4097,          YM), /* 00395 - Wally Beben - Summer Olympiad.sndh */
TIMEDB_ENTRY( 77b6a8f6,  0, 133057,       YM+TA), /* 03793 - Roggie - Ripped Off.sndh */
TIMEDB_ENTRY( 77be6a62,  0,   6401,          YM), /* 01779 - Lap - Lap 29.sndh */
TIMEDB_ENTRY( 77e0afdf,  0,   3584,          YM), /* 02938 - Insomnia - Plax Atax.sndh */
TIMEDB_ENTRY( 780319fe,  0,  10490,         STE), /* 03526 - 505 - Appendix.sndh */
TIMEDB_ENTRY( 781943c1,  0,   5375,    YM+TB+TD), /* 03696 - gwEm (Nick Scarim conversion) - Spy vs Spy.sndh */
TIMEDB_ENTRY( 781c78d4,  0,  43509, YM+TA+TB+TD), /* 02783 - Dubmood - Destroy the Template.sndh */
TIMEDB_ENTRY( 78204309,  0,  10656,          YM), /* 00736 - Sean Connolly - Moving On.sndh */
TIMEDB_ENTRY( 78258491,  0,   7103, YM+TA+TB+TD), /* 03641 - Timbral-DHS-YMR - Polish waitress.sndh */
TIMEDB_ENTRY( 7831c996,  0,  10392, YM+TA+TB+TD), /* 00193 - 505 - Sure Trip 2.sndh */
TIMEDB_ENTRY( 784a6f4e,  0,  43744,       YM+TA), /* 00002 - gwEm - Hardcore Will Never Die.sndh */
TIMEDB_ENTRY( 785b5c73,  0,   9220,    YM+TA+TD), /* 02332 - Tao - Cream (orig. Karsten Obarski) - Sleepwalk.sndh */
TIMEDB_ENTRY( 788dfc4e,  0,  18529,          YM), /* 04843 - t1000 - Bits 38.sndh */
TIMEDB_ENTRY( 78a8fed2,  0,   3583,          YM), /* 00219 - Tony Williams - Smash TV.sndh */
TIMEDB_ENTRY( 78a8fed2,  1,    575,          YM), /* 00219 - Tony Williams - Smash TV.sndh */
TIMEDB_ENTRY( 78ae279e,  0,   4210,         STE), /* 03454 - Per Almered (XiA) - Save The Earth Intro.sndh */
TIMEDB_ENTRY( 78b86c80,  0,  11840, YM+TA+TB+TD), /* 03395 - ULTRASYD - SCENE SPIRIT.sndh */
TIMEDB_ENTRY( 78c38480,  0,   1785,       YM+TA), /* 01934 - Miq - Hotliner.sndh */
TIMEDB_ENTRY( 78d65c71,  0,   9210,    YM+TB+TD), /* 01585 - Nemo - 2 lives left.sndh */
TIMEDB_ENTRY( 78f74af9,  0,  10753,    YM+TA+TD), /* 04204 - Tao - Cream - Nostalgic-O-Demo Fractal Blast (2 Sid Voice).sndh */
TIMEDB_ENTRY( 790768e3,  0,  17228,       YM+TD), /* 01986 - Jochen Hippel - "Ajh99".sndh */
TIMEDB_ENTRY( 790e4d93,  0,  16128,          YM), /* 00739 - Connor - TFC Spaceballs Demo.sndh */
TIMEDB_ENTRY( 790f5fb6,  0, 141313,       YM+TA), /* 03752 - Tufty - Knucklebusters.sndh */
TIMEDB_ENTRY( 7913a832,  0,   2682,          YM), /* 04213 - Jess - The name's Bond, Jess Bond (chip).sndh */
TIMEDB_ENTRY( 7913db4d,  0,   2737,          YM), /* 03137 - Unknown - It Came Upon a Midnight Clear.sndh */
TIMEDB_ENTRY( 7926f696,  0,  18240, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  1,    470, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  2,    600, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  3,    227, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  4,    370, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  5,    406, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  6,    448, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  7,    395, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  8,    299, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696,  9,    466, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696, 10,    443, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696, 11,    310, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 7926f696, 12,    876, YM+TA+TB+TD), /* 02065 - Mad Max - Auf Wiedersehen Monty - Sid.sndh */
TIMEDB_ENTRY( 79355a03,  0,  17640,          YM), /* 02017 - Mad Max - Ace 2.sndh */
TIMEDB_ENTRY( 793968aa,  0,  13824,          YM), /* 02133 - Mad Max - Crazy Comets  - LCD.sndh */
TIMEDB_ENTRY( 79400daf,  0,   2273,          YM), /* 03154 - n-a - THE ROCK ISLAND LINE.sndh */
TIMEDB_ENTRY( 7946ba85,  0,   1536,          YM), /* 04347 - Chris Denman - Cohort - Fight for Rome.sndh */
TIMEDB_ENTRY( 7953aa0e,  0,    243,          YM), /* 02141 - Mad Max - Commando Levels - LSD.sndh */
TIMEDB_ENTRY( 795a029e,  0,   2304,      YM+STE), /* 02952 - Chris Scudds - Cisco Heat (STe).sndh */
TIMEDB_ENTRY( 795a029e,  1,   2688,      YM+STE), /* 02952 - Chris Scudds - Cisco Heat (STe).sndh */
TIMEDB_ENTRY( 795eaaae,  0,   8916,          YM), /* 03032 - Debbie Davis - I Was a Fool (To Let You Go).sndh */
TIMEDB_ENTRY( 79683b0b,  0,   3505,          YM), /* 01731 - Peter Johnson - Wizball.sndh */
TIMEDB_ENTRY( 79683b0b,  1,   2697,          YM), /* 01731 - Peter Johnson - Wizball.sndh */
TIMEDB_ENTRY( 79683b0b,  2,   1153,          YM), /* 01731 - Peter Johnson - Wizball.sndh */
TIMEDB_ENTRY( 79683b0b,  3,   1778,          YM), /* 01731 - Peter Johnson - Wizball.sndh */
TIMEDB_ENTRY( 79683b0b,  4,    138,          YM), /* 01731 - Peter Johnson - Wizball.sndh */
TIMEDB_ENTRY( 7970cb7a,  0,   3734,          YM), /* 03080 - Paul Schuh - Cagney & Lacey Theme - B.Conti.sndh */
TIMEDB_ENTRY( 798a20b2,  0,   8970,          YM), /* 03090 - Unknown - Ben Hur Chariot Race March.sndh */
TIMEDB_ENTRY( 798bc04e,  0,   6110,          YM), /* 04346 - Nexus 6 - Nexus of Sominus (Theme).sndh */
TIMEDB_ENTRY( 799757ee,  0,   4602,       YM+TA), /* 02478 - No More - Unexpected.sndh */
TIMEDB_ENTRY( 79a266bf,  0,   4480,          YM), /* 02890 - Illusion - Betty (Golden Dawn 40a).sndh */
TIMEDB_ENTRY( 79a7ca4d,  0,   9840,    YM+TA+TD), /* 02508 - Crazy Q - DHS - Mad Monkey Funk-Fu.sndh */
TIMEDB_ENTRY( 79d46eb6,  0,  48049,       YM+TA), /* 03943 - Unknown - Under The Bridge.sndh */
TIMEDB_ENTRY( 79e39f24,  0,  15167, YM+TA+TB+TD), /* 04779 - Ultrasyd - Thunderdome.sndh */
TIMEDB_ENTRY( 79eeee20,  0,   8058,          YM), /* 00886 - Adrian Dalecki - Leonardo 2.sndh */
TIMEDB_ENTRY( 79eeee20,  1,    769,          YM), /* 00886 - Adrian Dalecki - Leonardo 2.sndh */
TIMEDB_ENTRY( 79ef500d,  0,   4260,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  1,   3700,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  2,   3936,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  3,   2640,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  4,   1536,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  5,   4032,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79ef500d,  6,    265,          YM), /* 02661 - Dave Lowe - Galaxy Force II.sndh */
TIMEDB_ENTRY( 79f3a147,  0,  18529,          YM), /* 03257 - T1000 - The Beast Club - TS 5.sndh */
TIMEDB_ENTRY( 79fa5755,  0,   9594,       YM+TA), /* 01627 - No More - Too Hot.sndh */
TIMEDB_ENTRY( 7a0ff461,  0,  19718,    YM+TA+TD), /* 02342 - Tao - Cream - Body Language (UCM 16).sndh */
TIMEDB_ENTRY( 7a13364e,  0,   1664,          YM), /* 00294 - Gary Wheaton - Alb.sndh */
TIMEDB_ENTRY( 7a2c43f5,  0,   8006, YM+TA+TB+TD), /* 02421 - 505-cpt+pdx [nils feske] - dosk [2005].sndh */
TIMEDB_ENTRY( 7a481c5d,  0,  47051,       YM+TD), /* 00006 - Gwem - Song 4 Hanah (YM Rockerz).sndh */
TIMEDB_ENTRY( 7a4be950,  0,   6464,          YM), /* 02389 - Nexus 6 - Once Again.sndh */
TIMEDB_ENTRY( 7a56b89c,  0,  49440, YM+TA+TB+TD), /* 03377 - STU - ST-GOD BACK FROM HELL.sndh */
TIMEDB_ENTRY( 7a722bf8,  0,   4607, YM+TA+TB+TD), /* 02982 - 505-paradox - easy.sndh */
TIMEDB_ENTRY( 7a725569,  0,   4880,          YM), /* 02374 - Nexus 6 - Dum Dum Zak.sndh */
TIMEDB_ENTRY( 7a79da45,  0,  15360,       YM+TD), /* 02610 - STU - PANIC INTRO.sndh */
TIMEDB_ENTRY( 7a8e4251,  0,  28160, YM+TA+TB+TD), /* 02621 - STU - TEKNOID BIGBEAT.sndh */
TIMEDB_ENTRY( 7a95db3e,  0,   6143,    YM+TB+TD), /* 03292 - Dma-Sc - ATIPYK Developments - Manical Drop - 1 Player music #1.sndh */
TIMEDB_ENTRY( 7a99cc64,  0,   1850,          YM), /* 00594 - Paul Summers - P47 The Freedom Fighter.sndh */
TIMEDB_ENTRY( 7a99cc64,  1,   3976,          YM), /* 00594 - Paul Summers - P47 The Freedom Fighter.sndh */
TIMEDB_ENTRY( 7a99cc64,  2,   2765,          YM), /* 00594 - Paul Summers - P47 The Freedom Fighter.sndh */
TIMEDB_ENTRY( 7a99cc64,  3,   2305,          YM), /* 00594 - Paul Summers - P47 The Freedom Fighter.sndh */
TIMEDB_ENTRY( 7a99cc64,  4,   3328,          YM), /* 00594 - Paul Summers - P47 The Freedom Fighter.sndh */
TIMEDB_ENTRY( 7aa07393,  0,   2317,          YM), /* 01734 - Richard Joseph - Cannon Fodder.sndh */
TIMEDB_ENTRY( 7aa07393,  1,   3763,          YM), /* 01734 - Richard Joseph - Cannon Fodder.sndh */
TIMEDB_ENTRY( 7aa07393,  2,   1345,          YM), /* 01734 - Richard Joseph - Cannon Fodder.sndh */
TIMEDB_ENTRY( 7aa07393,  3,   5041,          YM), /* 01734 - Richard Joseph - Cannon Fodder.sndh */
TIMEDB_ENTRY( 7aa07393,  4,   1025,          YM), /* 01734 - Richard Joseph - Cannon Fodder.sndh */
TIMEDB_ENTRY( 7aa1c085,  0,   7134,          YM), /* 03045 - Ed Bogas - Echo Rock.sndh */
TIMEDB_ENTRY( 7aa1f6bf,  0,  63588,       YM+TA), /* 04358 - X The Beermat - Mud.sndh */
TIMEDB_ENTRY( 7aa72754,  0,   4219,          YM), /* 00377 - 7AN - Nibbel Is Small Kermit Is Fat.sndh */
TIMEDB_ENTRY( 7aacfcc2,  0,  16128,    YM+TA+TD), /* 02257 - Tao - Cream  (orig. Jeroen Tel) - Ice Age (Steps).sndh */
TIMEDB_ENTRY( 7ac51aff,  0,   6652, YM+TA+TB+TD), /* 00189 - 505 - Sid Rave 3.sndh */
TIMEDB_ENTRY( 7aca079b,  0,   3839, YM+TA+TB+TD), /* 00060 - 505 - Ice Storm.sndh */
TIMEDB_ENTRY( 7ad13084,  0,   7164, YM+TA+TB+TD), /* 03321 - Kenet - Oldiez.sndh */
TIMEDB_ENTRY( 7ad4ffeb,  0,  13818,          YM), /* 01683 - Jess - Psychosonorous Disorder.sndh */
TIMEDB_ENTRY( 7ad8a222,  0,   1986,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  1,   1766,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  2,      9,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  3,     13,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  4,     17,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  5,     39,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7ad8a222,  6,      9,          YM), /* 04135 - Michael Haar - Defend The Earth 2.sndh */
TIMEDB_ENTRY( 7aed45b4,  0,  20616, YM+TA+TB+TD), /* 02711 - Marcer - SORN DESTINY (DIGI).sndh */
TIMEDB_ENTRY( 7b065ea8,  0,   8704,          YM), /* 00462 - Augustin Vidovic - Starpaws.sndh */
TIMEDB_ENTRY( 7b08f9ed,  0, 141401,       YM+TA), /* 03949 - Scott Lemmon - Elenor.sndh */
TIMEDB_ENTRY( 7b0c8115,  0,   7952,          YM), /* 04567 - Excellence in Art - The Chicken Song 4K.sndh */
TIMEDB_ENTRY( 7b1de588,  0,    240,          YM), /* 02190 - Mad Max - Dugger.sndh */
TIMEDB_ENTRY( 7b1de588,  1,   2304,          YM), /* 02190 - Mad Max - Dugger.sndh */
TIMEDB_ENTRY( 7b1de588,  2,   3072,          YM), /* 02190 - Mad Max - Dugger.sndh */
TIMEDB_ENTRY( 7b1de588,  3,    768,          YM), /* 02190 - Mad Max - Dugger.sndh */
TIMEDB_ENTRY( 7b300939,  0,  11136,          YM), /* 01146 - Tyan - Revolution - Main Menu.sndh */
TIMEDB_ENTRY( 7b33d44e,  0,     37,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  1,     39,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  2,     44,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  3,     73,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  4,     27,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  5,     36,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  6,     38,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  7,     41,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  8,    345,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b33d44e,  9,   6135,          YM), /* 02946 - Andrew Oakley - Square Off.sndh */
TIMEDB_ENTRY( 7b35b2e2,  0,   6720,          YM), /* 01980 - MSG - MSG 11.sndh */
TIMEDB_ENTRY( 7b3b1fb9,  0,   5378,          YM), /* 02157 - Mad Max - So Watt - Main Menu.sndh */
TIMEDB_ENTRY( 7b3dc813,  0,  45820, YM+TA+TB+TD), /* 02575 - STU - ATARI UNDERGROUND CHIPTUNE RESISTANCE.sndh */
TIMEDB_ENTRY( 7b568abe,  0,   3072,       YM+TA), /* 00922 - Dark Angel - Party In Hell!.sndh */
TIMEDB_ENTRY( 7b5f880f,  0,  11519,          YM), /* 01521 - Iso - Enchanted Lands.sndh */
TIMEDB_ENTRY( 7b62d825,  0,  23552, YM+TA+TB+TD), /* 02542 - SCHMX & STU - TAKE THIS.sndh */
TIMEDB_ENTRY( 7b7834ab,  0,   3450,          YM), /* 04374 - Jess - Vulx (draft).sndh */
TIMEDB_ENTRY( 7b8256b0,  0,  39168, YM+TA+TB+TD), /* 03672 - xyce - n-a.sndh */
TIMEDB_ENTRY( 7b8853a3,  0,   2304,          YM), /* 01427 - Starbuck - Atari Party.sndh */
TIMEDB_ENTRY( 7b9c5d58,  0,   4535,          YM), /* 01788 - Barry Leitch - Airborne Ranger.sndh */
TIMEDB_ENTRY( 7b9c5d58,  1,    503,          YM), /* 01788 - Barry Leitch - Airborne Ranger.sndh */
TIMEDB_ENTRY( 7ba6257e,  0,    319,    YM+TB+TD), /* 04185 - Dubmood - Lost Refound Unfinished Dub Session.sndh */
TIMEDB_ENTRY( 7bb1d48b,  0,  38525, YM+TA+TB+TD), /* 04552 - Factor 6 - Give Me Time.sndh */
TIMEDB_ENTRY( 7bd0c66a,  0,   2560,          YM), /* 01882 - Dave Lowe - E-Swat.sndh */
TIMEDB_ENTRY( 7bd0c66a,  1,    640,          YM), /* 01882 - Dave Lowe - E-Swat.sndh */
TIMEDB_ENTRY( 7bd0c66a,  2,     73,          YM), /* 01882 - Dave Lowe - E-Swat.sndh */
TIMEDB_ENTRY( 7bd54594,  0,   1015,          YM), /* 03916 - Unknown - England.sndh */
TIMEDB_ENTRY( 7bdf8712,  0,   2349,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  1,    235,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  2,    216,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  3,    199,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  4,    217,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  5,    184,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  6,    162,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  7,    241,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  8,    222,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712,  9,    258,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 10,   1580,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 11,   1178,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 12,   2335,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 13,   1585,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 14,   1582,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 15,   1197,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 16,   1576,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 17,   1098,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bdf8712, 18,     59,          YM), /* 01564 - Jason Page - Ivan Ironman Stewarts Super Off Road.sndh */
TIMEDB_ENTRY( 7bf5c286,  0,  34560, YM+TA+TB+TD), /* 02555 - Marcer - BUBBLE2005.sndh */
TIMEDB_ENTRY( 7c04319f,  0,   6400,       YM+TD), /* 03223 - Tomchi - Vertical Scroller.sndh */
TIMEDB_ENTRY( 7c0f7cb8,  0,   6080,          YM), /* 00091 - David Whittaker - Army Moves.sndh */
TIMEDB_ENTRY( 7c0f7cb8,  1,   5184,          YM), /* 00091 - David Whittaker - Army Moves.sndh */
TIMEDB_ENTRY( 7c0f7cb8,  2,     46,          YM), /* 00091 - David Whittaker - Army Moves.sndh */
TIMEDB_ENTRY( 7c0f7cb8,  3,    161,          YM), /* 00091 - David Whittaker - Army Moves.sndh */
TIMEDB_ENTRY( 7c174ffc,  0,  40311, YM+TA+TB+TD), /* 03524 - STU - BOMBABAZOOKA.sndh */
TIMEDB_ENTRY( 7c2f1ead,  0,   4608,          YM), /* 04607 - White Ninja - Grounded.sndh */
TIMEDB_ENTRY( 7c2f33ee,  0,   5663,    YM+TA+TD), /* 02762 - Dma-Sc (Rockman style  ) - ST 20th Ann. Megademo Menu.sndh */
TIMEDB_ENTRY( 7c3e07d1,  0,   6015,          YM), /* 04074 - Gary Wheaton - Honk.sndh */
TIMEDB_ENTRY( 7c5342a5,  0,   1910,          YM), /* 03938 - Mr Coke - Yet Another 4K Intro.sndh */
TIMEDB_ENTRY( 7c5342a5,  1,    767,          YM), /* 03938 - Mr Coke - Yet Another 4K Intro.sndh */
TIMEDB_ENTRY( 7c5986cb,  0,  43020,    YM+TA+TD), /* 02353 - Tao - Cream - Da Chipshit (YM Rockerz 2).sndh */
TIMEDB_ENTRY( 7c5fd02b,  0,   2261,          YM), /* 04686 - Robert Groves - Kingdom.sndh */
TIMEDB_ENTRY( 7c5fd02b,  1,    314,          YM), /* 04686 - Robert Groves - Kingdom.sndh */
TIMEDB_ENTRY( 7c5fd02b,  2,    391,          YM), /* 04686 - Robert Groves - Kingdom.sndh */
TIMEDB_ENTRY( 7c5fd02b,  3,    311,          YM), /* 04686 - Robert Groves - Kingdom.sndh */
TIMEDB_ENTRY( 7c73c7ce,  0,  24384, YM+TA+TB+TD), /* 04333 - mOdmate-CheckPoint (orig. Adam Gilmore) - ZYBEX ENSHRINED.sndh */
TIMEDB_ENTRY( 7ca60e91,  0,  30928,       YM+TA), /* 04797 - Unknown - Axel F.sndh */
TIMEDB_ENTRY( 7cb0fb7b,  0,  16387, YM+TA+TB+TD), /* 03632 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 7cb0fef2,  0,   9511,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cb0fef2,  1,    377,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cb0fef2,  2,    129,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cb0fef2,  3,    117,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cb0fef2,  4,     76,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cb0fef2,  5,     97,          YM), /* 00872 - Ben Daglish - Super Scramble Simulator.sndh */
TIMEDB_ENTRY( 7cbe3fa9,  0,   9595, YM+TA+TB+TD), /* 01830 - Lotek Style - Altair 4 - Stone Tower.sndh */
TIMEDB_ENTRY( 7cbed450,  0,  16960,       YM+TD), /* 01475 - STU - SCOOTER.sndh */
TIMEDB_ENTRY( 7cde2999,  0,   6912,          YM), /* 03309 - Tomchi - PI Bass.sndh */
TIMEDB_ENTRY( 7ce087f3,  0,  12051,       YM+TA), /* 03899 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( 7ce087f3,  1,   5121,       YM+TA), /* 03899 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( 7ce087f3,  2,  16898,       YM+TA), /* 03899 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( 7ced1e01,  0,    184,          YM), /* 03974 - Ian Blair - Detonate.sndh */
TIMEDB_ENTRY( 7ced1e01,  1,    160,          YM), /* 03974 - Ian Blair - Detonate.sndh */
TIMEDB_ENTRY( 7ced1e01,  2,    261,          YM), /* 03974 - Ian Blair - Detonate.sndh */
TIMEDB_ENTRY( 7d05ee31,  0,   6144,          YM), /* 04840 - t1000 - Bits 35.sndh */
TIMEDB_ENTRY( 7d0ad5a2,  0,  22859,          YM), /* 02239 - Tao - Cream (orig. Rob Hubbard) - IK+ (Stmygm).sndh */
TIMEDB_ENTRY( 7d0c5fa4,  0,   3613,          YM), /* 04455 - Ian Maxwell - Dragon Scape.sndh */
TIMEDB_ENTRY( 7d0c5fa4,  1,   7169,          YM), /* 04455 - Ian Maxwell - Dragon Scape.sndh */
TIMEDB_ENTRY( 7d17868c,  0,  38400,          YM), /* 04105 - Elrinth - n-a.sndh */
TIMEDB_ENTRY( 7d25d2f0,  0,   8287,          YM), /* 01777 - LAP - Tune 27.sndh */
TIMEDB_ENTRY( 7d2e02af,  0,  11647,    YM+TA+TD), /* 02563 - marcer - i wish v2.sndh */
TIMEDB_ENTRY( 7d694af5,  0,   6138,       YM+TA), /* 01344 - Scavenger - Synergy Love Theme.sndh */
TIMEDB_ENTRY( 7d7c95a0,  0,  29700,       YM+TA), /* 03817 - Tufty - Red New 3.sndh */
TIMEDB_ENTRY( 7d86db7b,  0,   3070,          YM), /* 02474 - Iso - Patat met Kroket.sndh */
TIMEDB_ENTRY( 7da088cb,  0,  16896,       YM+TD), /* 04861 - xFalcon - Christian Bunge - In'a country style.sndh */
TIMEDB_ENTRY( 7dc2f713,  0,   2771,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  1,   8658,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  2,   3499,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  3,     92,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  4,   4822,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  5,   2664,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  6,   3022,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  7,   8719,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  8,   5026,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713,  9,   6861,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 10,   2488,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 11,   8484,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 12,   3787,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 13,   3286,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 14,   2676,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7dc2f713, 15,   2386,          YM), /* 04143 - Unknown - Title Designer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  0,   6624,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  1,   2688,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  2,    193,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  3,   2240,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  4,    448,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  5,     44,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  6,    135,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  7,    106,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  8,   2304,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae,  9,   3264,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7ddaeeae, 10,   3072,          YM), /* 02205 - Mad Max - Master Blazer.sndh */
TIMEDB_ENTRY( 7de5eee1,  0,   3839,          YM), /* 04638 - dhor - Happy2.sndh */
TIMEDB_ENTRY( 7e0c09a7,  0,  59509,       YM+TA), /* 03877 - Mad Max - Ace 2.sndh */
TIMEDB_ENTRY( 7e4c2abc,  0,   9216,          YM), /* 02055 - Mad Max - Starpaws.sndh */
TIMEDB_ENTRY( 7e4c2abc,  1,   2688,          YM), /* 02055 - Mad Max - Starpaws.sndh */
TIMEDB_ENTRY( 7e4c2abc,  2,    384,          YM), /* 02055 - Mad Max - Starpaws.sndh */
TIMEDB_ENTRY( 7e4c2abc,  3,     11,          YM), /* 02055 - Mad Max - Starpaws.sndh */
TIMEDB_ENTRY( 7e705d5d,  0,  20480,          YM), /* 02046 - Mad Max - Nemesis The Warlock.sndh */
TIMEDB_ENTRY( 7e87073c,  0,   6912,          YM), /* 02108 - Mad Max - Arkanoid.sndh */
TIMEDB_ENTRY( 7e9147b9,  0,   6903, YM+TA+TB+TD), /* 00178 - 505 - Mystic Bytes 2.sndh */
TIMEDB_ENTRY( 7ec89c40,  0,   6656,       YM+TA), /* 04408 - Dr Beat - Maziac.sndh */
TIMEDB_ENTRY( 7ec89c40,  1,   6656,       YM+TA), /* 04408 - Dr Beat - Maziac.sndh */
TIMEDB_ENTRY( 7ec89c40,  2,   3241,       YM+TA), /* 04408 - Dr Beat - Maziac.sndh */
TIMEDB_ENTRY( 7ec89c40,  3,   6528,       YM+TA), /* 04408 - Dr Beat - Maziac.sndh */
TIMEDB_ENTRY( 7ec89c40,  4,   4096,       YM+TA), /* 04408 - Dr Beat - Maziac.sndh */
TIMEDB_ENTRY( 7ecc8753,  0,   8718,          YM), /* 00224 - ACC - ACC Song.sndh */
TIMEDB_ENTRY( 7ed223e3,  0,  11520,          YM), /* 01981 - MSG - MSG 13.sndh */
TIMEDB_ENTRY( 7ed6be79,  0,   4992,          YM), /* 03255 - T1000 - The Beast Club - TS 1.sndh */
TIMEDB_ENTRY( 7ef2d75c,  0,  31992,          YM), /* 03211 - Michel Winogradoff - Builder Land.sndh */
TIMEDB_ENTRY( 7ef2d75c,  1,  11352,          YM), /* 03211 - Michel Winogradoff - Builder Land.sndh */
TIMEDB_ENTRY( 7ef2d75c,  2,    571,          YM), /* 03211 - Michel Winogradoff - Builder Land.sndh */
TIMEDB_ENTRY( 7ef2d75c,  3,    497,          YM), /* 03211 - Michel Winogradoff - Builder Land.sndh */
TIMEDB_ENTRY( 7ef2d75c,  4,   1074,          YM), /* 03211 - Michel Winogradoff - Builder Land.sndh */
TIMEDB_ENTRY( 7f01cbd8,  0,  36864,       YM+TD), /* 03329 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 7f1c7ce7,  0,   4175,          YM), /* 00011 - 4-Mat - Bubble Dizzy.sndh */
TIMEDB_ENTRY( 7f1c7ce7,  1,    767,          YM), /* 00011 - 4-Mat - Bubble Dizzy.sndh */
TIMEDB_ENTRY( 7f1c7ce7,  2,   2207,          YM), /* 00011 - 4-Mat - Bubble Dizzy.sndh */
TIMEDB_ENTRY( 7f1c7ce7,  3,   5567,          YM), /* 00011 - 4-Mat - Bubble Dizzy.sndh */
TIMEDB_ENTRY( 7f1d73ee,  0,   8519,          YM), /* 04086 - Gary Wheaton - Old Man.sndh */
TIMEDB_ENTRY( 7f215061,  0,   1152,          YM), /* 01086 - Techno - Dentro 1a.sndh */
TIMEDB_ENTRY( 7f2e1bf1,  0,  10494,    YM+TA+TD), /* 00892 - damo - resgodz - ym rucka.sndh */
TIMEDB_ENTRY( 7f2e346b,  0,   3072,          YM), /* 04411 - Dr Beat - Niet AF 1.sndh */
TIMEDB_ENTRY( 7f3281eb,  0,   2880,          YM), /* 00745 - Sean Conran - Top Cat.sndh */
TIMEDB_ENTRY( 7f3281eb,  1,    361,          YM), /* 00745 - Sean Conran - Top Cat.sndh */
TIMEDB_ENTRY( 7f3a98de,  0,   6144,          YM), /* 01163 - Tyan - Tyan 2.sndh */
TIMEDB_ENTRY( 7f3d3bad,  0,  33280, YM+TA+TB+TD), /* 04573 - schmx-awp - SCHM-X-CALIBUR.sndh */
TIMEDB_ENTRY( 7f53b518,  0,  35328,          YM), /* 01451 - Stax - Stax 1.sndh */
TIMEDB_ENTRY( 7f84242c,  0,   7657,          YM), /* 01822 - Lord - Ambition Menu 2.sndh */
TIMEDB_ENTRY( 7f8516c9,  0,   3524,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  1,   3524,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  2,   3331,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  3,   3204,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  4,   2051,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  5,   2306,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  6,   4804,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  7,   3524,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  8,   3843,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9,  9,   2896,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 10,   3597,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 11,   4484,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 12,   3077,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 13,   3845,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 14,   4613,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 15,   4229,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 16,   3461,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 17,   3077,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 18,   3077,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 19,   2693,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8516c9, 20,   3845,          YM), /* 01376 - Justin Scharvona - Lemmings.sndh */
TIMEDB_ENTRY( 7f8c1d77,  0,   6475,          YM), /* 00323 - Gary Wheaton - New Religion.sndh */
TIMEDB_ENTRY( 7f92b81d,  0,   2240,       YM+TA), /* 00955 - Dark Angel - To be continued 2.sndh */
TIMEDB_ENTRY( 7fa94136,  0,    768,          YM), /* 04000 - Blue Nun - Gnurk Loader.sndh */
TIMEDB_ENTRY( 7fb2afe4,  0,  16517,    YM+TA+TD), /* 02340 - Tao - Cream - Building A Ruin (UCM 14).sndh */
TIMEDB_ENTRY( 7fb938bc,  0,   1785, YM+TA+TB+TD), /* 01966 - MSG - Godpey - Story 1.sndh */
TIMEDB_ENTRY( 7fc62b9e,  0,   8676,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  1,     64,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  2,     29,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  3,    160,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  4,   3384,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  5,    257,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  6,     97,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  7,     96,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  8,    161,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fc62b9e,  9,     29,          YM), /* 02183 - Mad Max - Circus Attractions.sndh */
TIMEDB_ENTRY( 7fd7219b,  0,   1920,       YM+TA), /* 00921 - Dark Angel - Golem Fever.sndh */
TIMEDB_ENTRY( 7feb1b63,  0,   7165,    YM+TA+TB), /* 03574 - 505-checkpoint^paradox - moodswinger.sndh */
TIMEDB_ENTRY( 7ff6d22e,  0,   1040,       YM+TA), /* 01998 - Mad Max - Bomber (ST-Falc).sndh */
TIMEDB_ENTRY( 7ff74252,  0,  10525,          YM), /* 01527 - Iso - Horrorscope.sndh */
TIMEDB_ENTRY( 8016e2b8,  0,    199,          YM), /* 01050 - Dr Beat - NTM - New.sndh */
TIMEDB_ENTRY( 802725bf,  0,   4620,       YM+TA), /* 01149 - Tyan - No Name.sndh */
TIMEDB_ENTRY( 8027fa17,  0,   3584,       YM+TA), /* 00808 - Crawdaddy - Rezonati.sndh */
TIMEDB_ENTRY( 802c9128,  0,  25094,       YM+TA), /* 04050 - Roggie - Three pints of milk please.sndh */
TIMEDB_ENTRY( 803eb91c,  0,   4609,          YM), /* 03404 - Allister Brimble - Pro Powerboat Simulator.sndh */
TIMEDB_ENTRY( 803f0458,  0,   1408, YM+TA+TB+TD), /* 02497 - gwEm (Ian Howe conversion) - Ravin'.sndh */
TIMEDB_ENTRY( 804d9fc1,  0,  11898, YM+TA+TB+TD), /* 02758 - Dma-Sc - Sector One - Fighting His Fears.sndh */
TIMEDB_ENTRY( 804ee466,  0,   5767,       YM+TA), /* 02773 - Skidz - Hydroxid Dentro.sndh */
TIMEDB_ENTRY( 8054e3b8,  0,   6255,       YM+TA), /* 02664 - Donald Fakk - Beat Nick mix.sndh */
TIMEDB_ENTRY( 807e9415,  0,   1068,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 807e9415,  1,   1069,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 807e9415,  2,    211,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 807e9415,  3,    934,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 807e9415,  4,    210,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 807e9415,  5,   2399,          YM), /* 00138 - David Whittaker - Spitting Image.sndh */
TIMEDB_ENTRY( 8086197d,  0,  14765,       YM+TA), /* 03756 - Tufty - Numero.sndh */
TIMEDB_ENTRY( 809533e6,  0,  14596,    YM+TA+TD), /* 02791 - SH3 - Reservoir Gods - Jobil.sndh */
TIMEDB_ENTRY( 80977f6f,  0,  48481,       YM+TA), /* 03805 - Sparky - Twitch.sndh */
TIMEDB_ENTRY( 80a8b3eb,  0,   6149,       YM+TD), /* 04407 - Dr Beat - Galtan 6 - Elite Intro 3 digi.sndh */
TIMEDB_ENTRY( 80aaad6b,  0,   2239,          YM), /* 02846 - Marcer - Poker Menu.sndh */
TIMEDB_ENTRY( 80adb17b,  0,  44544,          YM), /* 02691 - Philip Linde - Spraymaali.sndh */
TIMEDB_ENTRY( 80b2cf54,  0,  44160, YM+TA+TB+TD), /* 04007 - mOdmate-CheckPoint (orig. Uwe Loos-TGC) - TRIBUTE TO MY CHILDHOOD HEROES.sndh */
TIMEDB_ENTRY( 80bf9c60,  0,  11228,          YM), /* 04551 - Yerzmyey - Sommarhack 2014 Invite.sndh */
TIMEDB_ENTRY( 80cd4c07,  0,  10908,    YM+TA+TD), /* 04299 - 505-checkpoint^paradox - Pieces.sndh */
TIMEDB_ENTRY( 80d049d0,  0,   6239,          YM), /* 01534 - Iso - Sinking.sndh */
TIMEDB_ENTRY( 80dc0ddd,  0,   1693,          YM), /* 04502 - Geoff Camp - Brain food - Deuce.sndh */
TIMEDB_ENTRY( 80de0b59,  0,   4184,          YM), /* 00867 - Ben Daglish - Pacmania.sndh */
TIMEDB_ENTRY( 80de0b59,  1,    199,          YM), /* 00867 - Ben Daglish - Pacmania.sndh */
TIMEDB_ENTRY( 80de0b59,  2,   3346,          YM), /* 00867 - Ben Daglish - Pacmania.sndh */
TIMEDB_ENTRY( 80de0b59,  3,   2677,          YM), /* 00867 - Ben Daglish - Pacmania.sndh */
TIMEDB_ENTRY( 80eedbdd,  0,   9599, YM+TA+TB+TD), /* 02785 - Timbral-dhs^ymr - Junkie2149.sndh */
TIMEDB_ENTRY( 80ef941c,  0,   1152,    YM+TA+TD), /* 00400 - Big Alec - Action.sndh */
TIMEDB_ENTRY( 80f81d19,  0,  30037,       YM+TA), /* 03881 - Mad Max - Phantom of the Asteroid.sndh */
TIMEDB_ENTRY( 80f81d19,  1,  20704,       YM+TA), /* 03881 - Mad Max - Phantom of the Asteroid.sndh */
TIMEDB_ENTRY( 810de267,  0,   8640,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  1,    128,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  2,    257,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  3,    640,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  4,    288,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  5,   1024,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 810de267,  6,    257,          YM), /* 01786 - Anthony Lees - Brat.sndh */
TIMEDB_ENTRY( 81317f3f,  0,   8063,    YM+TA+TD), /* 02557 - marcer - comic baker maxer.sndh */
TIMEDB_ENTRY( 8136a17c,  0,   9983, YM+TA+TB+TD), /* 03611 - Timbral-DHS-YMR - Fresh Bitches.sndh */
TIMEDB_ENTRY( 814b39bc,  0,  12288,          YM), /* 04834 - T1000 - Bits 28.sndh */
TIMEDB_ENTRY( 814f3f2e,  0,   7171,       YM+TA), /* 01150 - Tyan - Old.sndh */
TIMEDB_ENTRY( 817b29ae,  0,  21504,          YM), /* 01467 - STU - FUNKEY.sndh */
TIMEDB_ENTRY( 81852f5e,  0,   4848,          YM), /* 03406 - Dave Lowe - Line of Fire.sndh */
TIMEDB_ENTRY( 81852f5e,  1,   2304,          YM), /* 03406 - Dave Lowe - Line of Fire.sndh */
TIMEDB_ENTRY( 8194b266,  0,    488,          YM), /* 04542 - Albert Baggetta - Ghost Mine.sndh */
TIMEDB_ENTRY( 81eb526a,  0,   4512,       YM+TD), /* 02380 - Nexus 6 - Lethal Xcess Level 2.sndh */
TIMEDB_ENTRY( 81eb526a,  1,    768,       YM+TD), /* 02380 - Nexus 6 - Lethal Xcess Level 2.sndh */
TIMEDB_ENTRY( 81eb526a,  2,    560,       YM+TD), /* 02380 - Nexus 6 - Lethal Xcess Level 2.sndh */
TIMEDB_ENTRY( 8206808c,  0,   3967, YM+TA+TB+TD), /* 00088 - 505 - More Sun.sndh */
TIMEDB_ENTRY( 820ffe29,  0,   7840,    YM+TA+TD), /* 02181 - Tao - Cream - ..do things (install).sndh */
TIMEDB_ENTRY( 821c23b4,  0, 170068,       YM+TA), /* 03968 - Unknown - Son of God.sndh */
TIMEDB_ENTRY( 82884d6e,  0,   4080,    YM+TA+TD), /* 02745 - Crazy Q - DHS - . G-SpOt ..sndh */
TIMEDB_ENTRY( 828877c1,  0,   6527,          YM), /* 02396 - Nexus 6 - Silly But Happy.sndh */
TIMEDB_ENTRY( 8289f87a,  0,   6132,       YM+TA), /* 00914 - Dark Angel - Angel In Hell.sndh */
TIMEDB_ENTRY( 828f12cb,  0,  51827,    YM+TA+TB), /* 01353 - Scavenger - Symbiosis - Synergy Megademo.sndh */
TIMEDB_ENTRY( 82925534,  0,   2497,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  1,    365,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  2,   2285,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  3,    752,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  4,    233,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  5,    218,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82925534,  6,    914,          YM), /* 04543 - William Anderson - Gopher Hunt.sndh */
TIMEDB_ENTRY( 82a10bd4,  0,  12576,       YM+TD), /* 00410 - Big Alec - Locomotion.sndh */
TIMEDB_ENTRY( 82a1a072,  0,  15275, YM+TA+TB+TD), /* 02071 - Mad Max - Confuzion - Sid.sndh */
TIMEDB_ENTRY( 82ab9458,  0,  12288,          YM), /* 02116 - Mad Max - Scout.sndh */
TIMEDB_ENTRY( 82ab9458,  1,   1536,          YM), /* 02116 - Mad Max - Scout.sndh */
TIMEDB_ENTRY( 82aee9e4,  0,   7062,          YM), /* 03096 - n-a - Pachelbel's Canon in D.sndh */
TIMEDB_ENTRY( 82b08ab6,  0,  17663,          YM), /* 02962 - Unknown - Final Command.sndh */
TIMEDB_ENTRY( 82bb663c,  0,   5954,          YM), /* 01046 - Dr Beat - Imagination - Many Minis.sndh */
TIMEDB_ENTRY( 82cdfc07,  0,   8256,          YM), /* 03177 - n-a - 76 Trombones.sndh */
TIMEDB_ENTRY( 82df2d73,  0,  13785,          YM), /* 04839 - t1000 - Bits 34.sndh */
TIMEDB_ENTRY( 82e4b442,  0,   1536,       YM+TA), /* 00950 - Dark Angel - It's Spring Time.sndh */
TIMEDB_ENTRY( 82ebfe04,  0,   5120,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82ebfe04,  1,    883,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82ebfe04,  2,   1536,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82ebfe04,  3,    395,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82ebfe04,  4,    371,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82ebfe04,  5,    307,          YM), /* 00980 - Chris Denman - Warriors Of Releyne.sndh */
TIMEDB_ENTRY( 82fbb472,  0,   6144,          YM), /* 01219 - ENS Music - ENS 2.sndh */
TIMEDB_ENTRY( 830a7e7e,  0,   5184,          YM), /* 02718 - Chris Denman - Legend of the Lost.sndh */
TIMEDB_ENTRY( 830a7e7e,  1,   2400,          YM), /* 02718 - Chris Denman - Legend of the Lost.sndh */
TIMEDB_ENTRY( 830a7e7e,  2,   4032,          YM), /* 02718 - Chris Denman - Legend of the Lost.sndh */
TIMEDB_ENTRY( 8310eb1b,  0,     32,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  1,      9,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  2,     16,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  3,      7,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  4,     27,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  5,     39,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  6,     57,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  7,     35,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  8,     32,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b,  9,     43,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 10,     74,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 11,    914,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 12,    144,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 13,      7,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 14,     88,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 15,     33,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 16,      5,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 17,    217,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 18,    179,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 19,   4000,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 8310eb1b, 20,    147,          YM), /* 04706 - Unknown - Prehistorik.sndh */
TIMEDB_ENTRY( 83230484,  0,    671,    YM+TA+TB), /* 03623 - Excellence In Art - Sommarkhack Report Pt 1.sndh */
TIMEDB_ENTRY( 83300f99,  0,  24639, YM+TA+TB+TD), /* 02695 - dubmood (breakmekanix) - Alla vill till himmelen.sndh */
TIMEDB_ENTRY( 833f75a0,  0,   8303,       YM+TA), /* 03669 - Per Almered-XiA-Checkpoint - Sunshine.sndh */
TIMEDB_ENTRY( 835c49fc,  0,   4602,       YM+TA), /* 00825 - Crazy Q - DHS - Nightmare at Beatstreet.sndh */
TIMEDB_ENTRY( 83652ab4,  0,   6226,          YM), /* 01927 - The Mind - Genesys Remix - Zeppelin.sndh */
TIMEDB_ENTRY( 839e8649,  0,   7895,          YM), /* 02460 - Ross Goodley - Bombuzal.sndh */
TIMEDB_ENTRY( 83b8a41f,  0,   4795, YM+TA+TB+TD), /* 00049 - 505 - Focky2.sndh */
TIMEDB_ENTRY( 83b9ee01,  0,  31299,       YM+TA), /* 03762 - Tufty - Tuft2a.sndh */
TIMEDB_ENTRY( 83c6f446,  0,   2612,          YM), /* 03039 - Ed Bogas - Boogie.sndh */
TIMEDB_ENTRY( 83e8d142,  0,   1280,          YM), /* 04478 - Thomas the Tank Engine - Paul Tankard.sndh */
TIMEDB_ENTRY( 83ebd9c3,  0,   4353,          YM), /* 00379 - Wally Beben - CDS Chess.sndh */
TIMEDB_ENTRY( 83ebd9c3,  1,   5365,          YM), /* 00379 - Wally Beben - CDS Chess.sndh */
TIMEDB_ENTRY( 83ebd9c3,  3,   9473,          YM), /* 00379 - Wally Beben - CDS Chess.sndh */
TIMEDB_ENTRY( 83f4c7f6,  0,   4608,       YM+TA), /* 00253 - Antichrist - Coolbass.sndh */
TIMEDB_ENTRY( 840a28cc,  0,   1152,          YM), /* 00777 - Crawdaddy - Bedtime.sndh */
TIMEDB_ENTRY( 840b9a90,  0,   8520,       YM+TA), /* 04015 - Mad Max - Axel F.sndh */
TIMEDB_ENTRY( 84303fe0,  0,  11432,          YM), /* 01022 - Doclands - Lightning Demo - Beast.sndh */
TIMEDB_ENTRY( 843e061f,  0,  26213,    YM+TB+TD), /* 04618 - Laserball 2014 - xFalcon.sndh */
TIMEDB_ENTRY( 843e061f,  1,  33026,    YM+TB+TD), /* 04618 - Laserball 2014 - xFalcon.sndh */
TIMEDB_ENTRY( 843e061f,  2,  26626,    YM+TB+TD), /* 04618 - Laserball 2014 - xFalcon.sndh */
TIMEDB_ENTRY( 843e061f,  3,   1077,       YM+TD), /* 04618 - Laserball 2014 - xFalcon.sndh */
TIMEDB_ENTRY( 843e061f,  4,    436,          YM), /* 04618 - Laserball 2014 - xFalcon.sndh */
TIMEDB_ENTRY( 8446cdb5,  0,   6911,       YM+TD), /* 04353 - gwEm - Fuji Sunset.sndh */
TIMEDB_ENTRY( 846ab268,  0,   4604, YM+TA+TB+TD), /* 00035 - 505 - Blue Snow (D-Bug 184a).sndh */
TIMEDB_ENTRY( 846e1b3b,  0,   2688,          YM), /* 03495 - Shamen - Certified Dope - Filled 3D.sndh */
TIMEDB_ENTRY( 846f92cb,  0,   1543,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  1,    621,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  2,   1371,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  3,     68,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  4,    187,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  5,    193,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 846f92cb,  6,   1518,          YM), /* 00960 - Mike Davies - Alien Storm.sndh */
TIMEDB_ENTRY( 848f1879,  0,   9600,          YM), /* 00569 - David Whittaker - Outrun.sndh */
TIMEDB_ENTRY( 848f1879,  1,   5832,          YM), /* 00569 - David Whittaker - Outrun.sndh */
TIMEDB_ENTRY( 84969b36,  0,  12288,          YM), /* 01648 - Jedi - Kidney 8 Remix.sndh */
TIMEDB_ENTRY( 84be4f62,  0,   1535,       YM+TD), /* 03288 - Dma-Sc - ATIPYK Developments - Manical Drop - 2 Players music #1.sndh */
TIMEDB_ENTRY( 8502b722,  0,   8363,          YM), /* 00169 - Martin Walker - Frenetic.sndh */
TIMEDB_ENTRY( 851a3afa,  0,   2670,          YM), /* 03058 - Ed Bogas - Prelude.sndh */
TIMEDB_ENTRY( 851d7ab9,  0,   9146,          YM), /* 03528 - Matt Furniss - Laser Squad.sndh */
TIMEDB_ENTRY( 85272569,  0,   5545,          YM), /* 03429 - Per Almered (XiA) 2009 - Cheap (in terms of CPU).sndh */
TIMEDB_ENTRY( 853a01d8,  0,   2559,          YM), /* 04380 - Britelite - One Trick Pony.sndh */
TIMEDB_ENTRY( 855c682b,  0,   2174, YM+TA+TB+TD), /* 00068 - 505 - Lajbkalg.sndh */
TIMEDB_ENTRY( 856484b0,  0,   2608,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  1,   3072,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  2,    145,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  3,    145,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  4,    128,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  5,    128,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  6,    193,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  7,    193,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  8,     97,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 856484b0,  9,     97,          YM), /* 00072 - David Whittaker - Bad Company.sndh */
TIMEDB_ENTRY( 85755c2d,  0,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  1,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  2,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  3,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  4,      9,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  5,     15,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  6,    135,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  7,     30,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  8,    240,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d,  9,    102,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 10,    120,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 11,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 12,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 14,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 15,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 16,     36,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 17,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 18,    313,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 19,    406,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 20,    802,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 21,    360,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 22,   2054,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 23,    303,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 24,    310,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 25,    285,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 26,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 27,    388,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 28,      6,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 29,      9,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 30,     18,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 31,      6,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 32,    310,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 33,   1933,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 34,    730,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85755c2d, 35,   1917,          YM), /* 04745 - Margaret Lowe - Police Quest - In the Pursuit of Death Angel.sndh */
TIMEDB_ENTRY( 85793842,  0,   7674,       YM+TA), /* 03506 - 505 - Suretrip 2 - Digi.sndh */
TIMEDB_ENTRY( 8580c999,  0,   8858, YM+TA+TB+TD), /* 03638 - Dma-Sc - ATARI  subarashii.sndh */
TIMEDB_ENTRY( 85830820,  0,  33023, YM+TA+TB+TD), /* 02985 - 505-paradox - Factored.sndh */
TIMEDB_ENTRY( 858d453e,  0,   6336,          YM), /* 04827 - T1000 - Bits 58.sndh */
TIMEDB_ENTRY( 8591da8c,  0,   4043, YM+TA+TB+TD), /* 04297 - 505-Checkpoint (2000) - PaniC.sndh */
TIMEDB_ENTRY( 8594f198,  0,  46512,       YM+TA), /* 03819 - Tufty - Wierd.sndh */
TIMEDB_ENTRY( 85a86719,  0,   8064,    YM+TA+TD), /* 02265 - Tao - Cream - State Of Confusion (Steps).sndh */
TIMEDB_ENTRY( 85c0e6c8,  0,  16128,          YM), /* 03310 - Tomchi - 1st Extra.sndh */
TIMEDB_ENTRY( 85c8d991,  0,   5142,          YM), /* 01438 - Stax - Lazer.sndh */
TIMEDB_ENTRY( 85d5f3b8,  0,   5567, YM+TA+TB+TD), /* 03418 - dubmood - RaZoR 1911 - D-Bug - Heatbeat vs WoTW.sndh */
TIMEDB_ENTRY( 85dd5c95,  0,   4608,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  1,    560,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  2,    218,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  3,    145,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  4,    512,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  5,     21,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  6,     29,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85dd5c95,  7,    768,          YM), /* 00966 - Mike Davies - Mercs.sndh */
TIMEDB_ENTRY( 85de9990,  0,   6336,          YM), /* 02384 - Nexus 6 - My Delight.sndh */
TIMEDB_ENTRY( 85e9d93c,  0,  12288,          YM), /* 01540 - Ralph - Ralph 3.sndh */
TIMEDB_ENTRY( 85f9dbbf,  0,   9401, YM+TA+TB+TD), /* 01961 - MSG - Godpey - Practice.sndh */
TIMEDB_ENTRY( 85fb312e,  0,  50330, YM+TA+TB+TD), /* 04336 - !Cube - Blipblop Memories.sndh */
TIMEDB_ENTRY( 86011a1a,  0,   8188, YM+TA+TB+TD), /* 00036 - 505 - Another Brick In The Wall.sndh */
TIMEDB_ENTRY( 8628401b,  0,  38153,          YM), /* 00849 - Ben Daglish - Flintstones.sndh */
TIMEDB_ENTRY( 8628401b,  1,  33302,          YM), /* 00849 - Ben Daglish - Flintstones.sndh */
TIMEDB_ENTRY( 862e79b4,  0,  14266,       YM+TA), /* 01742 - Jovis - Lametrop - End.sndh */
TIMEDB_ENTRY( 862f0c04,  0,   8512,          YM), /* 02352 - Tao - Cream - Crazy Train (YM Rockerz 5).sndh */
TIMEDB_ENTRY( 8647ab46,  0,  89852,       YM+TA), /* 04057 - Roggie - Spleen music.sndh */
TIMEDB_ENTRY( 865c215f,  0,   4608,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  1,   4608,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  2,    171,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  3,    268,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  4,    395,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  5,    395,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  6,    299,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  7,    267,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  8,    395,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 865c215f,  9,    524,          YM), /* 02028 - Mad Max - Delta.sndh */
TIMEDB_ENTRY( 8670176a,  0,   9216,       YM+TA), /* 04858 - Michel Winogradoff - Disc.sndh */
TIMEDB_ENTRY( 86712ac2,  0,   8355,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  1,   5888,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  2,   6720,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  3,   8160,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  4,   6656,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  5,  10880,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  6,   5280,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  7,   6048,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  8,   9856,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2,  9,     81,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2, 10,    201,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86712ac2, 11,   8632,          YM), /* 01803 - Barry Leitch - Rock & Roll.sndh */
TIMEDB_ENTRY( 86877312,  0,   4992,          YM), /* 01147 - Tyan - Revolution - Melody.sndh */
TIMEDB_ENTRY( 868f4905,  0,   7168,          YM), /* 00607 - Chris Mad - Phalanx Main Menu.sndh */
TIMEDB_ENTRY( 86a16e5f,  0,   1534,          YM), /* 02466 - Iso - Holland is OK.sndh */
TIMEDB_ENTRY( 86a590d4,  0,   2112,          YM), /* 00449 - Augustin Vidovic - Incommunicardo.sndh */
TIMEDB_ENTRY( 86a9b159,  0,    959,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  1,    528,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  2,    170,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  3,    118,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  4,     96,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  5,    168,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  6,   3071,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  7,   3071,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  8,   4479,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159,  9,   3071,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86a9b159, 10,   4608,          YM), /* 00357 - Mark Wilson - Bonaza Brothers.sndh */
TIMEDB_ENTRY( 86be6826,  0,   6528,          YM), /* 00113 - David Whittaker - Knightmare.sndh */
TIMEDB_ENTRY( 86c4eac7,  0,  11630,    YM+TA+TB), /* 00074 - 505 - leave it - alive 4.sndh */
TIMEDB_ENTRY( 86c60eb4,  0,    514,    YM+TA+TD), /* 02233 - Tao - Cream - Loader (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 86e3fc9b,  0,  14603, YM+TA+TB+TD), /* 02106 - Mad Max - Zoolook - Sid.sndh */
TIMEDB_ENTRY( 86ed0e65,  0,    256,          YM), /* 04544 - Grant Magnuson - Granshot.sndh */
TIMEDB_ENTRY( 86ed0e65,  1,    640,          YM), /* 04544 - Grant Magnuson - Granshot.sndh */
TIMEDB_ENTRY( 86f8c6b0,  0,   6465,       YM+TA), /* 04417 - EVE - Rock & Roll.sndh */
TIMEDB_ENTRY( 86fbb490,  0,   3924,          YM), /* 03924 - Unknown - Musik 001.sndh */
TIMEDB_ENTRY( 8716d8eb,  0,   6725,          YM), /* 00020 - Martin Iveson - Wolfchild.sndh */
TIMEDB_ENTRY( 872fca5b,  0, 109257,       YM+TA), /* 04060 - Hybrid - Swirl.sndh */
TIMEDB_ENTRY( 876c3fe0,  0, 195903,       YM+TA), /* 04040 - Roggie - Who stood on the caterpillar?.sndh */
TIMEDB_ENTRY( 8777606c,  0,   3823,          YM), /* 04070 - Gary Wheaton - Fractal.sndh */
TIMEDB_ENTRY( 8779fed0,  0,   5584,          YM), /* 01230 - Epic - Another Illusion.sndh */
TIMEDB_ENTRY( 87a3e62f,  0,  14208,          YM), /* 00762 - Count Zero - Decade Demo - Boot.sndh */
TIMEDB_ENTRY( 87a70bcb,  0,  10456,          YM), /* 03091 - n-a - Both Sides Now.sndh */
TIMEDB_ENTRY( 87a8a5b0,  0,  81189,       YM+TA), /* 03702 - Sparky - Bachola - From a Bach 3-part inv..sndh */
TIMEDB_ENTRY( 87aa164b,  0,   3241,          YM), /* 03281 - Thothy - Stille.sndh */
TIMEDB_ENTRY( 87b13082,  0,   1799,          YM), /* 02723 - Jeremy Das & Nik Sen - Orbit 2000.sndh */
TIMEDB_ENTRY( 87b13082,  1,    128,          YM), /* 02723 - Jeremy Das & Nik Sen - Orbit 2000.sndh */
TIMEDB_ENTRY( 87ba7407,  0,   7267,    YM+TA+TD), /* 03432 - Per Almered (XiA) 2009 - The Last Song.sndh */
TIMEDB_ENTRY( 87c9aef3,  0,   9756,          YM), /* 04565 - Stan Mercury - Innuendo.sndh */
TIMEDB_ENTRY( 87d9730c,  0,  10766,          YM), /* 00510 - Tangens - Synthetic Visions - Magic Devil.sndh */
TIMEDB_ENTRY( 87e8abf0,  0,     49,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  1,     42,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  2,    775,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  3,   1822,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  4,   2280,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  5,    149,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  6,   1492,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87e8abf0,  7,    629,          YM), /* 04707 - P.F. Rees - Pushover.sndh */
TIMEDB_ENTRY( 87ee8140,  0,   8075,          YM), /* 02431 - Beast - Blaackel.sndh */
TIMEDB_ENTRY( 87f1a3c3,  0,   1025,          YM), /* 02939 - Ken Legace - Knights of the Sky.sndh */
TIMEDB_ENTRY( 87f1a3c3,  1,   1021,          YM), /* 02939 - Ken Legace - Knights of the Sky.sndh */
TIMEDB_ENTRY( 87f1a3c3,  2,   1027,          YM), /* 02939 - Ken Legace - Knights of the Sky.sndh */
TIMEDB_ENTRY( 8809a088,  0,   6713,       YM+TA), /* 04102 - 505-cpt^pdx - Blazing.sndh */
TIMEDB_ENTRY( 8816f97e,  0,   8428,          YM), /* 01571 - Patrick Phelan - Zool.sndh */
TIMEDB_ENTRY( 8816f97e,  1,   6000,          YM), /* 01571 - Patrick Phelan - Zool.sndh */
TIMEDB_ENTRY( 8816f97e,  2,   5060,          YM), /* 01571 - Patrick Phelan - Zool.sndh */
TIMEDB_ENTRY( 8816f97e,  3,   4356,          YM), /* 01571 - Patrick Phelan - Zool.sndh */
TIMEDB_ENTRY( 8816f97e,  4,  10560,          YM), /* 01571 - Patrick Phelan - Zool.sndh */
TIMEDB_ENTRY( 882b5a87,  0,   3824,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 882b5a87,  1,   3824,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 882b5a87,  2,   3464,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 882b5a87,  3,   3824,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 882b5a87,  4,   3824,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 882b5a87,  5,   7553,          YM), /* 01490 - Fred Gray - Eco.sndh */
TIMEDB_ENTRY( 883c5219,  0,   2533,          YM), /* 02889 - Captain Headcrash - Total Vision Inc 3.5k.sndh */
TIMEDB_ENTRY( 8842bf29,  0,   5120,          YM), /* 02631 - Nemo - DHS - Bobby.sndh */
TIMEDB_ENTRY( 885918eb,  0,   7169,          YM), /* 00391 - Wally Beben - Quartz.sndh */
TIMEDB_ENTRY( 885a67aa,  0,   1536,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  1,    883,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  2,    505,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  3,    417,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  4,    358,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  5,    437,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  6,    221,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  7,     28,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  8,     23,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa,  9,     63,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa, 10,    892,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa, 11,      7,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa, 12,     17,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 885a67aa, 13,    295,          YM), /* 04536 - Darren Ithell - Fun School 4 (5-7 years).sndh */
TIMEDB_ENTRY( 8862706d,  0,  11819, YM+TA+TB+TD), /* 01283 - Frequent - Ephidrena ( 2003 ) - Asfalt (YM Rockerz).sndh */
TIMEDB_ENTRY( 88706463,  0,   7872,          YM), /* 02942 - Andi Mcginty - WWF European Rampage.sndh */
TIMEDB_ENTRY( 88706463,  1,   1901,          YM), /* 02942 - Andi Mcginty - WWF European Rampage.sndh */
TIMEDB_ENTRY( 88706463,  2,    642,          YM), /* 02942 - Andi Mcginty - WWF European Rampage.sndh */
TIMEDB_ENTRY( 88706463,  3,    403,          YM), /* 02942 - Andi Mcginty - WWF European Rampage.sndh */
TIMEDB_ENTRY( 887b01d2,  0,   2954,          YM), /* 03066 - Chris French - The Little Drummer Boy.sndh */
TIMEDB_ENTRY( 888f17ae,  0,   9640, YM+TA+TB+TD), /* 01846 - Lotek Style - Elfworld.sndh */
TIMEDB_ENTRY( 8894f6cd,  0,   9210,       YM+TB), /* 01845 - Lotek Style - Eilig.sndh */
TIMEDB_ENTRY( 889e8ff3,  0,    268,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  1,   4437,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  2,   5916,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  3,   4681,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  4,   6096,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  5,   4096,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  6,   1812,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  7,   5140,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 889e8ff3,  8,   1524,          YM), /* 00105 - David Whittaker - Gold of the Aztecs.sndh */
TIMEDB_ENTRY( 88a522f4,  0,  13440,          YM), /* 01726 - Jet Power - AGITRS - MAIN MENU.sndh */
TIMEDB_ENTRY( 88b71630,  0,   7290,          YM), /* 01720 - Jess - The Undead Horror Show.sndh */
TIMEDB_ENTRY( 88cf05aa,  0,   5119, YM+TA+TB+TD), /* 02834 - maktone-fairlight - goldfish.sndh */
TIMEDB_ENTRY( 88e1a889,  0,   5120,       YM+TA), /* 04588 - Illusion - Pure Energy 4.sndh */
TIMEDB_ENTRY( 890722c7,  0,   4032,    YM+TA+TD), /* 00442 - Big Alec - Vortex.sndh */
TIMEDB_ENTRY( 8907a108,  0,   7678,          YM), /* 01265 - Frazer - Bertram.sndh */
TIMEDB_ENTRY( 8927bb05,  0, 281597,       YM+TA), /* 03785 - Nexus 6 - Rising 1.sndh */
TIMEDB_ENTRY( 89340a3e,  0,  13056,          YM), /* 00473 - Ultimatum - Big Kid.sndh */
TIMEDB_ENTRY( 89370a2b,  0,  21504,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  1,   1665,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  2,  14400,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  3,   1961,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  4,    769,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  5,   2580,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89370a2b,  6,   7200,          YM), /* 00753 - Mark Cooksey - Gremlins 2.sndh */
TIMEDB_ENTRY( 89419650,  0,   6255,          YM), /* 02941 - Archer Maclean - Jimmy Whites Whirlwind Snooker.sndh */
TIMEDB_ENTRY( 8952507b,  0,  37632,          YM), /* 03862 - xyce - Cpp is Forbidden - Part 2.sndh */
TIMEDB_ENTRY( 895907a8,  0,   8179,          YM), /* 00288 - Andrew Barnabas - SWIV (SILKWORM IV).sndh */
TIMEDB_ENTRY( 895907a8,  1,   8640,          YM), /* 00288 - Andrew Barnabas - SWIV (SILKWORM IV).sndh */
TIMEDB_ENTRY( 8964eb98,  0,  25338,          YM), /* 01698 - Jess - Not Soporific Lullaby.sndh */
TIMEDB_ENTRY( 896fbc6d,  0,   2079,          YM), /* 03692 - Dma-Sc - Saboteur 3 - Title Screen.sndh */
TIMEDB_ENTRY( 898a612a,  0,  18136,       YM+TA), /* 04017 - Roggie - Bach.sndh */
TIMEDB_ENTRY( 898af7b9,  0,   4096,          YM), /* 03499 - Shamen - Certified Dope - Shape it Up.sndh */
TIMEDB_ENTRY( 899af032,  0,   5375,          YM), /* 04094 - Gary Wheaton - So Watt remix.sndh */
TIMEDB_ENTRY( 89b4327d,  0,   5633,          YM), /* 01506 - Rob Hubbard - Jupiter Probe.sndh */
TIMEDB_ENTRY( 89c22e74,  0,  24576,          YM), /* 01400 - Skyline - Snork - Rave YM.sndh */
TIMEDB_ENTRY( 89c7d192,  0,  30432, YM+TA+TB+TD), /* 03375 - STU - READY TO SCORE.sndh */
TIMEDB_ENTRY( 89ee07ec,  0,   3066,          YM), /* 01696 - Jess - For your Loader #2.sndh */
TIMEDB_ENTRY( 89f069cd,  0,   5520,          YM), /* 00153 - David Whittaker - World Class Rugby.sndh */
TIMEDB_ENTRY( 89f069cd,  1,    769,          YM), /* 00153 - David Whittaker - World Class Rugby.sndh */
TIMEDB_ENTRY( 89f069cd,  2,    491,          YM), /* 00153 - David Whittaker - World Class Rugby.sndh */
TIMEDB_ENTRY( 89f069cd,  3,    242,          YM), /* 00153 - David Whittaker - World Class Rugby.sndh */
TIMEDB_ENTRY( 89f069cd,  4,     71,          YM), /* 00153 - David Whittaker - World Class Rugby.sndh */
TIMEDB_ENTRY( 89fad116,  0,   4032,          YM), /* 02235 - Tao - Cream - Reset (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 8a05829a,  0,   6399, YM+TA+TB+TD), /* 02993 - 505-paradox - p7action (STE only).sndh */
TIMEDB_ENTRY( 8a177b0d,  0,  10752,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  1,   3072,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  2,   8000,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  3,   4160,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  4,   8960,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  5,   1536,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  6,  12288,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  7,  10368,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  8,    793,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d,  9,  10496,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a177b0d, 10,   9792,       YM+TA), /* 02224 - Mad Max - Wings of Death.sndh */
TIMEDB_ENTRY( 8a209682,  0,   3216,    YM+TA+TD), /* 04251 - 505-checkpoint - Dark.sndh */
TIMEDB_ENTRY( 8a424ece,  0,  12838,       YM+TA), /* 01042 - Bodenst ndig 2000 - Landflucht.sndh */
TIMEDB_ENTRY( 8a6360ab,  0,   8663,          YM), /* 00873 - Ben Daglish - Skidz.sndh */
TIMEDB_ENTRY( 8a855118,  0,   1920,          YM), /* 01103 - Techno - Techno 8.sndh */
TIMEDB_ENTRY( 8abf584e,  0,   9517, YM+TA+TB+TD), /* 02753 - Dma-Sc - Street Robots.sndh */
TIMEDB_ENTRY( 8acc81b1,  0,  32270,          YM), /* 04685 - John Wangler - Killing Time.sndh */
TIMEDB_ENTRY( 8af0536e,  0,    804,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af0536e,  1,    522,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af0536e,  2,    777,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af0536e,  3,    492,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af0536e,  4,    566,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af0536e,  5,    513,          YM), /* 02032 - Mad Max - Geoff Capes Strongman.sndh */
TIMEDB_ENTRY( 8af193c9,  0,  17536, YM+TA+TB+TD), /* 02089 - Mad Max - Monty on the Run - Sid.sndh */
TIMEDB_ENTRY( 8af193c9,  1,   8192, YM+TA+TB+TD), /* 02089 - Mad Max - Monty on the Run - Sid.sndh */
TIMEDB_ENTRY( 8af193c9,  2,    901, YM+TA+TB+TD), /* 02089 - Mad Max - Monty on the Run - Sid.sndh */
TIMEDB_ENTRY( 8b055b20,  0,   5632,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  1,   7296,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  2,   1536,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  3,   7680,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  4,   9400,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  5,   9040,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b055b20,  6,   6272,          YM), /* 00758 - Count Zero - Awesome.sndh */
TIMEDB_ENTRY( 8b0fed98,  0,   4428,          YM), /* 00858 - Ben Daglish - Lotus Turbo Espirit.sndh */
TIMEDB_ENTRY( 8b0fed98,  1,   5444,          YM), /* 00858 - Ben Daglish - Lotus Turbo Espirit.sndh */
TIMEDB_ENTRY( 8b0fed98,  2,   4167,          YM), /* 00858 - Ben Daglish - Lotus Turbo Espirit.sndh */
TIMEDB_ENTRY( 8b0fed98,  3,   6212,          YM), /* 00858 - Ben Daglish - Lotus Turbo Espirit.sndh */
TIMEDB_ENTRY( 8b0fed98,  4,   5831,          YM), /* 00858 - Ben Daglish - Lotus Turbo Espirit.sndh */
TIMEDB_ENTRY( 8b1281b1,  0,   4992,          YM), /* 01173 - Sally - Groovy.sndh */
TIMEDB_ENTRY( 8b25b76f,  0,   9603,       YM+TA), /* 04216 - Big Alec - Best Part of Creation.sndh */
TIMEDB_ENTRY( 8b2795f7,  0,   7680,       YM+TA), /* 04413 - EVE - F Nightmare.sndh */
TIMEDB_ENTRY( 8b53314e,  0,   8863,          YM), /* 03116 - n-a - Entry of the Gladiators.sndh */
TIMEDB_ENTRY( 8b7693b4,  0,  15583,          YM), /* 04149 - Unknown - Danza.sndh */
TIMEDB_ENTRY( 8b87fb58,  0,  16128,    YM+TA+TD), /* 02351 - Tao - Cream - Shifter (YM Rockerz 1).sndh */
TIMEDB_ENTRY( 8bac11ad,  0,   6331,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  1,   3196,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  2,   3835,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  3,   4603,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  4,   2556,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  5,     91,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bac11ad,  6,    379,          YM), /* 00550 - Alistair Brimble - Fast Food.sndh */
TIMEDB_ENTRY( 8bba7351,  0,   7200,       YM+TD), /* 02387 - Nexus 6 - Never Say Die!.sndh */
TIMEDB_ENTRY( 8bc623bf,  0, 119050,       YM+TA), /* 03735 - David Whittaker - The Killing Cloud.sndh */
TIMEDB_ENTRY( 8bc623bf,  1,  85936,       YM+TA), /* 03735 - David Whittaker - The Killing Cloud.sndh */
TIMEDB_ENTRY( 8bc623bf,  2,  35785,       YM+TA), /* 03735 - David Whittaker - The Killing Cloud.sndh */
TIMEDB_ENTRY( 8bcea812,  0,  16608,          YM), /* 00733 - Sean Connolly - Euro Soccer.sndh */
TIMEDB_ENTRY( 8bcea812,  1,    486,          YM), /* 00733 - Sean Connolly - Euro Soccer.sndh */
TIMEDB_ENTRY( 8bcea812,  2,    257,          YM), /* 00733 - Sean Connolly - Euro Soccer.sndh */
TIMEDB_ENTRY( 8bcea812,  3,    377,          YM), /* 00733 - Sean Connolly - Euro Soccer.sndh */
TIMEDB_ENTRY( 8be0ecd1,  0,    768,       YM+TA), /* 00941 - Dark Angel - On The Road.sndh */
TIMEDB_ENTRY( 8be8f290,  0,   9026,       YM+TA), /* 00434 - Big Alec - Sanxion Loader Digi.sndh */
TIMEDB_ENTRY( 8c29a260,  0,   7141,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  1,  12153,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  2,  71534,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  3,  16645,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  4,  11525,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  5, 145145,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  6,  36386,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  7,  25345,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  8,  42889,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260,  9,  73340,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260, 10,  48487,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260, 11,  10361,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260, 12,  29649,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260, 13,  31385,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c29a260, 14,  29377,       YM+TA), /* 04554 - Flix - Mega Musix Demo.sndh */
TIMEDB_ENTRY( 8c2fb4ec,  0,  11808,       YM+TD), /* 02370 - Nexus 6 - Cyclone's Memory Theme.sndh */
TIMEDB_ENTRY( 8c37b1ae,  0,   3330,          YM), /* 03181 - n-a - .sndh */
TIMEDB_ENTRY( 8c4525aa,  0,   5632,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  1,   4480,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  2,  13824,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  3,   7168,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  4,   7680,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  5,   3072,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c4525aa,  6,   5568,          YM), /* 00124 - David Whittaker - Platoon.sndh */
TIMEDB_ENTRY( 8c48beb4,  0,   6937, YM+TA+TB+TD), /* 01954 - MSG - EIL entry.sndh */
TIMEDB_ENTRY( 8c50cdbf,  0,  18432,          YM), /* 04644 - Xrwwr - Dentro.sndh */
TIMEDB_ENTRY( 8c5607ee,  0,  68141,          YM), /* 01810 - Barry Leitch - Treasure Trap.sndh */
TIMEDB_ENTRY( 8c77c4d5,  0,   3515, YM+TA+TB+TD), /* 01120 - Toodeloo - Test 1.sndh */
TIMEDB_ENTRY( 8c7baee8,  0,   4479, YM+TA+TB+TD), /* 00278 - Baggio - PMP Sug.sndh */
TIMEDB_ENTRY( 8c8d30f4,  0,   4604, YM+TA+TB+TD), /* 04288 - 505 - Checkpoint - Midfinger (Unrel.).sndh */
TIMEDB_ENTRY( 8c8dcdc0,  0,   7680,          YM), /* 03485 - Nigel Pritchard - Solider 2000.sndh */
TIMEDB_ENTRY( 8cb0143e,  0,    672,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0143e,  1,    281,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0143e,  2,    361,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0143e,  3,    378,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0143e,  4,    445,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0143e,  5,   3366,          YM), /* 00168 - Martin Walker - Final Blow.sndh */
TIMEDB_ENTRY( 8cb0d7a4,  0,   5632,       YM+TA), /* 04121 - Tao - Prelude - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 8cbc3d18,  0,   6724,          YM), /* 00464 - Augustin Vidovic - Thalion Intro.sndh */
TIMEDB_ENTRY( 8cbedd59,  0,  39443,       YM+TA), /* 03775 - Stewart Gardner - Stampede 4.sndh */
TIMEDB_ENTRY( 8ce171f3,  0,   5567, YM+TA+TB+TD), /* 03269 - Dubmood - Mr Deaths Omega Illusion.sndh */
TIMEDB_ENTRY( 8ce23e7b,  0,  38912,       YM+TD), /* 02624 - STU - THE UNEXPECTED.sndh */
TIMEDB_ENTRY( 8ce2eab8,  0, 205169,       YM+TA), /* 03712 - Charles Callet - Metal Masters.sndh */
TIMEDB_ENTRY( 8cf7cc14,  0,    320,          YM), /* 03687 - Dma-Sc - Saboteur 3 - Game Over.sndh */
TIMEDB_ENTRY( 8d1af88d,  0,   1150,          YM), /* 02472 - Iso - Last Yell.sndh */
TIMEDB_ENTRY( 8d1c0533,  0,  23040,       YM+TD), /* 01461 - STU - DARK SEASON.sndh */
TIMEDB_ENTRY( 8d210821,  0,   3807, YM+TA+TB+TD), /* 03328 - Dubmood & Zabutom - 2006 Concrt Intro - Zabutom - Kenny Star.sndh */
TIMEDB_ENTRY( 8d2c0721,  0,   9212,    YM+TB+TD), /* 01594 - Nemo - Game Over.sndh */
TIMEDB_ENTRY( 8d2eb1f2,  0,  17920,          YM), /* 01631 - Carsten Neubauer - Beam.sndh */
TIMEDB_ENTRY( 8d3d8d54,  0,   2687, YM+TA+TB+TD), /* 03591 - Dma-Sc - Frankiki.sndh */
TIMEDB_ENTRY( 8d45cf8a,  0,   7883,       YM+TA), /* 02672 - Donald Fakk - Jag Rullar.sndh */
TIMEDB_ENTRY( 8d48f5b2,  0,   7807,       YM+TA), /* 01757 - Lap - Bird Girl Mad Demo Apart.sndh */
TIMEDB_ENTRY( 8d4d32dd,  0,   3456,          YM), /* 00919 - Dark Angel - Thalion Intro.sndh */
TIMEDB_ENTRY( 8d63c0db,  0,   3712,          YM), /* 03037 - Ed Bogas - Aisle Dance.sndh */
TIMEDB_ENTRY( 8d6663f1,  0,   1008,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6663f1,  1,    401,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6663f1,  2,   4223,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6663f1,  3,   1440,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6663f1,  4,    325,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6663f1,  5,   1920,          YM), /* 00351 - Tony Williams - Soldier of Light.sndh */
TIMEDB_ENTRY( 8d6949f9,  0,   4224,          YM), /* 04833 - T1000 - Bits 60.sndh */
TIMEDB_ENTRY( 8d7895af,  0,   1193,    YM+TA+TD), /* 02678 - Donald Fakk - Porkchop (Logical mix).sndh */
TIMEDB_ENTRY( 8d939c23,  0,  99220,       YM+TA), /* 04035 - Roggie - Roggie 17.sndh */
TIMEDB_ENTRY( 8dbd7740,  0,   2568,          YM), /* 00924 - Dark Angel - Hop Jo.sndh */
TIMEDB_ENTRY( 8dbe6511,  0, 137540,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  1,  26113,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  3,  61987,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  4,  46081,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  5,  66886,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  6, 148483,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  7, 174029,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  8, 254294,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511,  9,  20737,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 10,  49281,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 12,  62977,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 13, 289262,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 14,  84291,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 15,  85581,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 16,  98850,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dbe6511, 17,  46081,       YM+TA), /* 03769 - Count Zero - Decade Demo - Quartet.sndh */
TIMEDB_ENTRY( 8dd544db,  0,   8925,          YM), /* 03101 - n-a - .sndh */
TIMEDB_ENTRY( 8ddc7e6d,  0,   7680,          YM), /* 02917 - Richard Joseph - Megalomania.sndh */
TIMEDB_ENTRY( 8decf627,  0,   6048,          YM), /* 02395 - Nexus 6 - Rainy Days.sndh */
TIMEDB_ENTRY( 8decf627,  1,    419,          YM), /* 02395 - Nexus 6 - Rainy Days.sndh */
TIMEDB_ENTRY( 8dedec82,  0,  11136,          YM), /* 02164 - Mad Max - Syntax Terror - TLB.sndh */
TIMEDB_ENTRY( 8df3ff4b,  0,   5370,    YM+TA+TB), /* 01589 - Nemo - DHS - Clocks (miniremix).sndh */
TIMEDB_ENTRY( 8e063c55,  0,  10080,          YM), /* 02122 - Mad Max - Cuddly Demos - Big Sprite.sndh */
TIMEDB_ENTRY( 8e0efde2,  0,    227,          YM), /* 04492 - Christof Dukiet - Arc.sndh */
TIMEDB_ENTRY( 8e0efde2,  1,     84,          YM), /* 04492 - Christof Dukiet - Arc.sndh */
TIMEDB_ENTRY( 8e0efde2,  2,     33,          YM), /* 04492 - Christof Dukiet - Arc.sndh */
TIMEDB_ENTRY( 8e0efde2,  3,    207,          YM), /* 04492 - Christof Dukiet - Arc.sndh */
TIMEDB_ENTRY( 8e0efde2,  4,    183,          YM), /* 04492 - Christof Dukiet - Arc.sndh */
TIMEDB_ENTRY( 8e2e2598,  0,  84708,       YM+TA), /* 04042 - Roggie - F.O.A.D. ya bam.sndh */
TIMEDB_ENTRY( 8e349d20,  0,   7424,          YM), /* 02170 - Mad Max - Pandora.sndh */
TIMEDB_ENTRY( 8e36011b,  0,  12672,    YM+TA+TD), /* 02294 - Tao - Cream - Sense (Alive 2).sndh */
TIMEDB_ENTRY( 8e378611,  0,   1933,    YM+TA+TD), /* 04205 - Tao - Cream (orig. Jeroen Tel) - Nostalgic-O-Demo Ikari Union (2 Sid Voice).sndh */
TIMEDB_ENTRY( 8e4e7df4,  0,   7348,          YM), /* 01817 - Paul Loraine - Huckleberry Hound in Hollywood Capers.sndh */
TIMEDB_ENTRY( 8e4e7df4,  1,   6671,          YM), /* 01817 - Paul Loraine - Huckleberry Hound in Hollywood Capers.sndh */
TIMEDB_ENTRY( 8e4e7df4,  2,   8523,          YM), /* 01817 - Paul Loraine - Huckleberry Hound in Hollywood Capers.sndh */
TIMEDB_ENTRY( 8e4e7df4,  3,   4600,          YM), /* 01817 - Paul Loraine - Huckleberry Hound in Hollywood Capers.sndh */
TIMEDB_ENTRY( 8e4f2328,  0,   8543,    YM+TA+TD), /* 02647 - Dma-Sc - Disco Fizz.sndh */
TIMEDB_ENTRY( 8e5379b3,  0,  19968, YM+TA+TB+TD), /* 02591 - STU - FIRST SID.sndh */
TIMEDB_ENTRY( 8e7fd9cc,  0,  13439, YM+TA+TB+TD), /* 02490 - gwEm (Jeff Wayne remix) - Forever 'Ardcore.sndh */
TIMEDB_ENTRY( 8e95a515,  0,   5175,       YM+TA), /* 01123 - Tooseb - Bob Morane (Indochine).sndh */
TIMEDB_ENTRY( 8eac0fea,  0,   9214,          YM), /* 01024 - Doclands - Delirious Demo 2 - Reset.sndh */
TIMEDB_ENTRY( 8ebe66b0,  0,   4979,    YM+TB+TD), /* 00181 - 505 - Paradox Xmas 2003 #1.sndh */
TIMEDB_ENTRY( 8eca3208,  0,   8386,          YM), /* 01267 - Frazer - If Pigs Could Fly End.sndh */
TIMEDB_ENTRY( 8ef3c08c,  0,   5375,          YM), /* 04062 - Gary Wheaton - Backlash.sndh */
TIMEDB_ENTRY( 8ef8b91d,  0,   2488,          YM), /* 03385 - Unknown - Prelude.sndh */
TIMEDB_ENTRY( 8f05f777,  0,   9604,    YM+TA+TD), /* 02676 - Donald Fakk - Leif.sndh */
TIMEDB_ENTRY( 8f074b95,  0,   9594, YM+TA+TB+TD), /* 04249 - 505 - Byeing.sndh */
TIMEDB_ENTRY( 8f12df12,  0,   3264,          YM), /* 01113 - Keith Tinman - Recoil.sndh */
TIMEDB_ENTRY( 8f12df12,  1,   3072,          YM), /* 01113 - Keith Tinman - Recoil.sndh */
TIMEDB_ENTRY( 8f171ada,  0,   5688,          YM), /* 03202 - Nick Scarim - Spy vs Spy - The Island Caper.sndh */
TIMEDB_ENTRY( 8f2cd337,  0,  13815, YM+TA+TB+TD), /* 01832 - Lotek Style - Arche - UCM 13 part 3.sndh */
TIMEDB_ENTRY( 8f2ce871,  0, 106739,       YM+TA), /* 04033 - Roggie - Roggie 15.sndh */
TIMEDB_ENTRY( 8f59b130,  0,  17280,          YM), /* 02057 - Mad Max - Thrust.sndh */
TIMEDB_ENTRY( 8f6772c4,  0,  11026,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f6772c4,  1,    256,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f6772c4,  2,    826,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f6772c4,  3,    296,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f6772c4,  4,   1921,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f6772c4,  5,   1537,          YM), /* 00385 - Wally Beben - Hawkeye.sndh */
TIMEDB_ENTRY( 8f8babad,  0,   6960,    YM+TA+TD), /* 00891 - damo - resgodz - funktion.sndh */
TIMEDB_ENTRY( 8f9064ce,  0,   2233,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  1,   2380,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  2,   2100,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  3,   5068,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  4,   4846,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  5,    396,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f9064ce,  6,   1481,          YM), /* 02915 - Richard Joseph - Lure of the Temptress.sndh */
TIMEDB_ENTRY( 8f90a02d,  0,   9986,          YM), /* 04780 - tomchi  noextra - ltc16 smoke town.sndh */
TIMEDB_ENTRY( 8f959e76,  0,   9728,          YM), /* 00570 - Jas C Brooke - Overlander.sndh */
TIMEDB_ENTRY( 8f959e76,  1,   1536,          YM), /* 00570 - Jas C Brooke - Overlander.sndh */
TIMEDB_ENTRY( 8f959e76,  2,    164,          YM), /* 00570 - Jas C Brooke - Overlander.sndh */
TIMEDB_ENTRY( 8f959e76,  3,    274,          YM), /* 00570 - Jas C Brooke - Overlander.sndh */
TIMEDB_ENTRY( 8f959e76,  4,   2832,          YM), /* 00570 - Jas C Brooke - Overlander.sndh */
TIMEDB_ENTRY( 8fb5ca0b,  0,   6146,          YM), /* 00403 - Big Alec - Feardrop.sndh */
TIMEDB_ENTRY( 8fc80c16,  0,   3083,          YM), /* 01761 - Lap - Lap 1 - Batman.sndh */
TIMEDB_ENTRY( 8fd390e1,  0,  69605,       YM+TA), /* 03879 - Mad Max - Starpaws.sndh */
TIMEDB_ENTRY( 8fe9358b,  0,  10875, YM+TA+TB+TD), /* 04290 - 505-checkpoint^paradox - Myth.sndh */
TIMEDB_ENTRY( 9000c3ab,  0,   7874,          YM), /* 00362 - Steve Barrett - Adv Skiing Simulator.sndh */
TIMEDB_ENTRY( 9000c3ab,  1,    290,          YM), /* 00362 - Steve Barrett - Adv Skiing Simulator.sndh */
TIMEDB_ENTRY( 9000c3ab,  2,    218,          YM), /* 00362 - Steve Barrett - Adv Skiing Simulator.sndh */
TIMEDB_ENTRY( 9002cb90,  0,   3200,    YM+TA+TD), /* 02253 - Tao - Cream  (orig. Rob Hubbard) - Delta (Steps).sndh */
TIMEDB_ENTRY( 900306d7,  0,   9288,    YM+TA+TD), /* 02652 - Dma-Sc - POOZ - Intro & Menu.sndh */
TIMEDB_ENTRY( 900cca7a,  0,  13667,          YM), /* 00313 - Gary Wheaton - Cuddly Demos Main Menu.sndh */
TIMEDB_ENTRY( 90181e30,  0,  14008,          YM), /* 00738 - Sean Connolly - Rainy Day.sndh */
TIMEDB_ENTRY( 901b8ed0,  0,  14338,    YM+TA+TD), /* 02344 - Tao - Cream - Double Fat (UCM 18).sndh */
TIMEDB_ENTRY( 903338c6,  0,   9258,       YM+TA), /* 04226 - 505 - 5th.sndh */
TIMEDB_ENTRY( 904819f7,  0,   6144,          YM), /* 01186 - Excellence In Art - I Wonder.sndh */
TIMEDB_ENTRY( 9052a143,  0,  17282,          YM), /* 02188 - Mad Max - Cybernoid 2.sndh */
TIMEDB_ENTRY( 9052a143,  1,    292,          YM), /* 02188 - Mad Max - Cybernoid 2.sndh */
TIMEDB_ENTRY( 905b4e7c,  0,   3520,    YM+TA+TD), /* 00042 - 505 - Checkpoint - MMJ Xmas 03 - 1.sndh */
TIMEDB_ENTRY( 90638a9d,  0,   3936,          YM), /* 01819 - Lord - Ambition Intro 1.sndh */
TIMEDB_ENTRY( 9070f526,  0,  12206,       YM+TD), /* 02684 - Donald Fakk - Spanking (Chip).sndh */
TIMEDB_ENTRY( 90950882,  0,   1440,          YM), /* 03507 - Unknown - Veteran.sndh */
TIMEDB_ENTRY( 909d7c6a,  0,   3070,          YM), /* 02432 - Beast - TSC - Bla Bla Bla.sndh */
TIMEDB_ENTRY( 90a07127,  0,   4992,          YM), /* 01174 - Sally - Oh Yeah.sndh */
TIMEDB_ENTRY( 90c5579c,  0,  10848,          YM), /* 01978 - MSG - MSG 08.sndh */
TIMEDB_ENTRY( 90d3e897,  0,   5952,          YM), /* 00566 - Jas C Brooke - Heavy Metal.sndh */
TIMEDB_ENTRY( 90d748b8,  0,   6344,          YM), /* 01789 - Barry Leitch - Championship Manager.sndh */
TIMEDB_ENTRY( 90d7fb7a,  0,  11008,       YM+TA), /* 02895 - Remalon - Bitume.sndh */
TIMEDB_ENTRY( 90de30f7,  0,  13792,          YM), /* 04784 - Barry Leitch - Postman Pat.sndh */
TIMEDB_ENTRY( 90de30f7,  1,    721,          YM), /* 04784 - Barry Leitch - Postman Pat.sndh */
TIMEDB_ENTRY( 90e8ecd2,  0,    217,          YM), /* 01070 - Jonathan Dunn - Wildwheels.sndh */
TIMEDB_ENTRY( 90e8ecd2,  1,    217,          YM), /* 01070 - Jonathan Dunn - Wildwheels.sndh */
TIMEDB_ENTRY( 90e8ecd2,  2,     98,          YM), /* 01070 - Jonathan Dunn - Wildwheels.sndh */
TIMEDB_ENTRY( 90e8ecd2,  3,    133,          YM), /* 01070 - Jonathan Dunn - Wildwheels.sndh */
TIMEDB_ENTRY( 90e8ecd2,  4,    157,          YM), /* 01070 - Jonathan Dunn - Wildwheels.sndh */
TIMEDB_ENTRY( 90eca942,  0,  11044,       YM+TA), /* 04423 - Lotus - Synth Dream 2 - Blow Jam.sndh */
TIMEDB_ENTRY( 90f254be,  0,   7107,       YM+TA), /* 04585 - Crawdaddy - Pure Energy 2.sndh */
TIMEDB_ENTRY( 90f2e8d8,  0,   3584,    YM+TB+TD), /* 00048 - 505 - Fabo.sndh */
TIMEDB_ENTRY( 9124ef78,  0,   6480,    YM+TA+TD), /* 02775 - Crazy Q - DHS - . G-SpOt . (final version).sndh */
TIMEDB_ENTRY( 91283c69,  0,   1075,          YM), /* 02922 - CS Adams - Head Coach.sndh */
TIMEDB_ENTRY( 912dcccc,  0,   5120, YM+TA+TB+TD), /* 04311 - 505-checkpoint - Stream.sndh */
TIMEDB_ENTRY( 912fa995,  0,   1536,          YM), /* 01401 - Skyline - Snork Intro - BIG.sndh */
TIMEDB_ENTRY( 91310c49,  0,  12540,          YM), /* 00773 - Count Zero - Go For It - Maggie 8.sndh */
TIMEDB_ENTRY( 91363e06,  0,   4475,    YM+TA+TB), /* 01605 - Nemo - Short Cuts.sndh */
TIMEDB_ENTRY( 9136adf9,  0,   4935,       YM+TD), /* 04787 - Intruder - Intruder 3.sndh */
TIMEDB_ENTRY( 91399987,  0,   4922,          YM), /* 02926 - Allister Brimble - Seymour goes to Hollywood.sndh */
TIMEDB_ENTRY( 91399987,  1,   8506,          YM), /* 02926 - Allister Brimble - Seymour goes to Hollywood.sndh */
TIMEDB_ENTRY( 91521990,  0,   9221,    YM+TA+TD), /* 02197 - Ultra & Tao - Cream - ..do things (reset).sndh */
TIMEDB_ENTRY( 91718563,  0,   8442, YM+TA+TB+TD), /* 03415 - lotek style (orig. timbral) - electronica re-shox.sndh */
TIMEDB_ENTRY( 91792432,  0,  19776,          YM), /* 02178 - Mad Max - Astaroth.sndh */
TIMEDB_ENTRY( 91792432,  1,   6912,          YM), /* 02178 - Mad Max - Astaroth.sndh */
TIMEDB_ENTRY( 91792432,  2,   6144,          YM), /* 02178 - Mad Max - Astaroth.sndh */
TIMEDB_ENTRY( 91792432,  3,  15616,          YM), /* 02178 - Mad Max - Astaroth.sndh */
TIMEDB_ENTRY( 91857058,  0,   3221,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91857058,  1,   1933,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91857058,  2,    181,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91857058,  3,   3039,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91857058,  4,     81,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91857058,  5,     31,          YM), /* 03517 - Fred Gray - Victory Road.sndh */
TIMEDB_ENTRY( 91935cc7,  0,   9986,    YM+TA+TD), /* 02303 - Tao - Cream - Breath 2000.sndh */
TIMEDB_ENTRY( 91a0b1c5,  0,   1152,       YM+TA), /* 04447 - Scavenger - Bomb.sndh */
TIMEDB_ENTRY( 91c5ee9a,  0,   4686,          YM), /* 01299 - Matt Furniss - John Barnes Euro Football.sndh */
TIMEDB_ENTRY( 91c5ee9a,  1,   4702,          YM), /* 01299 - Matt Furniss - John Barnes Euro Football.sndh */
TIMEDB_ENTRY( 91c5ee9a,  2,   4638,          YM), /* 01299 - Matt Furniss - John Barnes Euro Football.sndh */
TIMEDB_ENTRY( 91c5ee9a,  3,    637,          YM), /* 01299 - Matt Furniss - John Barnes Euro Football.sndh */
TIMEDB_ENTRY( 91c5ee9a,  4,     17,          YM), /* 01299 - Matt Furniss - John Barnes Euro Football.sndh */
TIMEDB_ENTRY( 91c7d7e4,  0, 260108,       YM+TA), /* 04669 - Piotr Szerszen - Last Man.sndh */
TIMEDB_ENTRY( 91d09de8,  0,  12672,          YM), /* 03263 - Jedi - Zoomer.sndh */
TIMEDB_ENTRY( 91db48fb,  0,   6787,       YM+TA), /* 04437 - Lotus - Synth Dream 2 - State of the World.sndh */
TIMEDB_ENTRY( 91dccdd3,  0,   2688,          YM), /* 02760 - Dma-Sc - Sector One - Oric Demo.sndh */
TIMEDB_ENTRY( 91f7e381,  0,   1048,          YM), /* 01909 - Marcer - Nnooo.sndh */
TIMEDB_ENTRY( 91fe5aa6,  0,   8826, YM+TA+TB+TD), /* 02830 - Msg - Superfly tune 7.sndh */
TIMEDB_ENTRY( 91ff9259,  0,   8442,       YM+TA), /* 04231 - 505 - 9th.sndh */
TIMEDB_ENTRY( 9218c31e,  0,   2547,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  1,   2747,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  2,   1929,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  3,   2810,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  4,   2976,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  5,   2295,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  6,   3981,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 9218c31e,  7,   1967,    YM+TA+TD), /* 01825 - Pierre-Eric Loriaux - Ivanhoe.sndh */
TIMEDB_ENTRY( 92220034,  0,   4992,          YM), /* 01570 - Patrick Phelan - Premier Manager.sndh */
TIMEDB_ENTRY( 923228e7,  0,  39516, YM+STE+TA+TD), /* 03610 - damo-resgods - deep blue.sndh */
TIMEDB_ENTRY( 92359c63,  0,   4944,          YM), /* 01482 - Holger Gehrmann - Domination.sndh */
TIMEDB_ENTRY( 92359c63,  1,   2880,          YM), /* 01482 - Holger Gehrmann - Domination.sndh */
TIMEDB_ENTRY( 924044dd,  0,    184,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  1,    244,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  2,    266,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  3,    670,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  4,   1534,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  5,   1341,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  6,   6592,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  7,   1921,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  8,   6144,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd,  9,   6784,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd, 10,   6144,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd, 11,   5086,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd, 12,   7678,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 924044dd, 13,  12990,          YM), /* 01582 - Karsten Obarski - Dyter 7.sndh */
TIMEDB_ENTRY( 92507639,  0,  14672,       YM+TA), /* 01573 - Jurgen Piscol - Hard 'n' Heavy.sndh */
TIMEDB_ENTRY( 92507639,  1,  13701,    YM+TA+TD), /* 01573 - Jurgen Piscol - Hard 'n' Heavy.sndh */
TIMEDB_ENTRY( 92507639,  2,  22495,    YM+TA+TD), /* 01573 - Jurgen Piscol - Hard 'n' Heavy.sndh */
TIMEDB_ENTRY( 92507639,  3,   9187,       YM+TA), /* 01573 - Jurgen Piscol - Hard 'n' Heavy.sndh */
TIMEDB_ENTRY( 92544fb7,  0,  30903,          YM), /* 00511 - Tangens - Synthetic Visions - Menu.sndh */
TIMEDB_ENTRY( 925c0055,  0,  44752, YM+TA+TB+TD), /* 03612 - STU - INSANE LEVEL.sndh */
TIMEDB_ENTRY( 925da54f,  0,   5568,       YM+TD), /* 02379 - Nexus 6 - Lethal Xcess Level 1.sndh */
TIMEDB_ENTRY( 925da54f,  1,    880,       YM+TD), /* 02379 - Nexus 6 - Lethal Xcess Level 1.sndh */
TIMEDB_ENTRY( 925da54f,  2,   1120,       YM+TD), /* 02379 - Nexus 6 - Lethal Xcess Level 1.sndh */
TIMEDB_ENTRY( 925da54f,  3,    800,       YM+TD), /* 02379 - Nexus 6 - Lethal Xcess Level 1.sndh */
TIMEDB_ENTRY( 9267624e,  0, 315218,       YM+TA), /* 04634 - TFC - Disco I.sndh */
TIMEDB_ENTRY( 926e5f9c,  0,  44544,          YM), /* 01445 - Stax - Old.sndh */
TIMEDB_ENTRY( 927aaaf0,  0,    768,          YM), /* 01891 - Dave Lowe - Street Fighter II - The World Warrior.sndh */
TIMEDB_ENTRY( 927aaaf0,  1,   3520,          YM), /* 01891 - Dave Lowe - Street Fighter II - The World Warrior.sndh */
TIMEDB_ENTRY( 927aaaf0,  2,   2128,          YM), /* 01891 - Dave Lowe - Street Fighter II - The World Warrior.sndh */
TIMEDB_ENTRY( 927aaaf0,  3,   3136,          YM), /* 01891 - Dave Lowe - Street Fighter II - The World Warrior.sndh */
TIMEDB_ENTRY( 927aaaf0,  4,   2320,          YM), /* 01891 - Dave Lowe - Street Fighter II - The World Warrior.sndh */
TIMEDB_ENTRY( 92918c4f,  0,  10280,          YM), /* 04472 - Charles Callet - TNT.sndh */
TIMEDB_ENTRY( 92918c4f,  1,  11776,          YM), /* 04472 - Charles Callet - TNT.sndh */
TIMEDB_ENTRY( 92918c4f,  2,  14376,          YM), /* 04472 - Charles Callet - TNT.sndh */
TIMEDB_ENTRY( 92918c4f,  3,  14016,          YM), /* 04472 - Charles Callet - TNT.sndh */
TIMEDB_ENTRY( 92af0dc4,  0,   4794, YM+TA+TB+TD), /* 01611 - No More - Cool Days.sndh */
TIMEDB_ENTRY( 92b271a9,  0,  13056,          YM), /* 01101 - Techno - Techno 6.sndh */
TIMEDB_ENTRY( 92b4df1f,  0,  33260,          YM), /* 01444 - Stax - Nully.sndh */
TIMEDB_ENTRY( 92c04749,  0,   5127, YM+TA+TB+TD), /* 00214 - 505 - You Bore (D-Bug 180).sndh */
TIMEDB_ENTRY( 92c48488,  0,   9226,       YM+TA), /* 02670 - Donald Fakk - Human Race - tune 4 (Chip).sndh */
TIMEDB_ENTRY( 9302f392,  0,   4032,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  1,   6528,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  2,   3019,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  3,   6657,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  4,   2400,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  5,   5378,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  6,     49,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  7,    181,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  8,    192,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9302f392,  9,     49,          YM), /* 01056 - Jonathan Dunn - Batman The Movie.sndh */
TIMEDB_ENTRY( 9313ee1f,  0,   7168,          YM), /* 03408 - Unknown - European Soccer Challenge.sndh */
TIMEDB_ENTRY( 931da92a,  0,   6533,          YM), /* 01378 - Justin Scharvona - Predator 2.sndh */
TIMEDB_ENTRY( 931da92a,  1,   8069,          YM), /* 01378 - Justin Scharvona - Predator 2.sndh */
TIMEDB_ENTRY( 931da92a,  2,   6149,          YM), /* 01378 - Justin Scharvona - Predator 2.sndh */
TIMEDB_ENTRY( 931da92a,  3,  10757,          YM), /* 01378 - Justin Scharvona - Predator 2.sndh */
TIMEDB_ENTRY( 93249f24,  0,   1296,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  1,   1393,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  2,    299,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  3,    217,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  4,   1092,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  5,   2880,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  6,   4376,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  7,   4632,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  8,   1034,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24,  9,    384,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24, 10,    201,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 93249f24, 11,    257,          YM), /* 02956 - Rudi Stember - Dynablaster.sndh */
TIMEDB_ENTRY( 9326708f,  0,   8064,          YM), /* 03410 - Riff Raff - T Bird.sndh */
TIMEDB_ENTRY( 9326708f,  1,   9408,          YM), /* 03410 - Riff Raff - T Bird.sndh */
TIMEDB_ENTRY( 9375f2d8,  0,  18432,          YM), /* 01485 - Holger Gehrmann - Hollywood Poker Pro.sndh */
TIMEDB_ENTRY( 9375f2d8,  1,   3840,          YM), /* 01485 - Holger Gehrmann - Hollywood Poker Pro.sndh */
TIMEDB_ENTRY( 938d8ca8,  0,  15432, YM+TA+TB+TD), /* 03320 - Marcer - n-a.sndh */
TIMEDB_ENTRY( 9393b118,  0,  11136,       YM+TA), /* 04127 - Tao - Walk on Ice - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 93a5ed13,  0,  12803,          YM), /* 00764 - Count Zero - Decade Demo - Gigadist.sndh */
TIMEDB_ENTRY( 93a68ee3,  0,  24576, YM+TA+TB+TD), /* 03990 - XyNo - POCAILLE BURGER!.sndh */
TIMEDB_ENTRY( 93a934e0,  0,   4608,          YM), /* 01172 - Sally - Great.sndh */
TIMEDB_ENTRY( 93b4af45,  0,   8191,          YM), /* 00306 - Gary Wheaton - Fresh.sndh */
TIMEDB_ENTRY( 93db17a9,  0,   6672,          YM), /* 03872 - Gary Wheaton - Flying.sndh */
TIMEDB_ENTRY( 93db49a4,  0,   4866,          YM), /* 01778 - Lap - Lap 28.sndh */
TIMEDB_ENTRY( 93eabe05,  0,   1541,          YM), /* 02280 - Tao - Cream ( orig. ??, C64) - Just Feel It 2 (Just Buggin).sndh */
TIMEDB_ENTRY( 93f41167,  0,  15360, YM+TA+TB+TD), /* 02814 - Marcer - Paradize - Nuclear Waste Dump - Ingame tune 1.sndh */
TIMEDB_ENTRY( 93fcbb92,  0,   9696,          YM), /* 00877 - Ben Daglish - Terramex.sndh */
TIMEDB_ENTRY( 940a5311,  0,   9690,          YM), /* 00121 - David Whittaker - Obliterator.sndh */
TIMEDB_ENTRY( 940a5311,  1,   6912,          YM), /* 00121 - David Whittaker - Obliterator.sndh */
TIMEDB_ENTRY( 940a5311,  2,   1153,          YM), /* 00121 - David Whittaker - Obliterator.sndh */
TIMEDB_ENTRY( 940a5311,  3,   3840,          YM), /* 00121 - David Whittaker - Obliterator.sndh */
TIMEDB_ENTRY( 941290a2,  0,   3281,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  1,   4608,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  2,    768,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  3,   3840,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  4,   3456,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  5,    768,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941290a2,  6,   2304,          YM), /* 02958 - Paul Tankard - Dr Who.sndh */
TIMEDB_ENTRY( 941e7120,  0,   3576,       YM+TA), /* 01125 - Tyan - Bits 2.sndh */
TIMEDB_ENTRY( 94335b15,  0,  11774,          YM), /* 04071 - Gary Wheaton - Galaxia.sndh */
TIMEDB_ENTRY( 9441b4a6,  0,   8826,    YM+TB+TD), /* 01012 - DMA-SC - Sun Burns - SV2k Intro.sndh */
TIMEDB_ENTRY( 94447618,  0,   3071,          YM), /* 02009 - Mad Max - Jazzy.sndh */
TIMEDB_ENTRY( 9449913a,  0,   3841,          YM), /* 00399 - Big Alec - A Case For Two.sndh */
TIMEDB_ENTRY( 94594d5f,  0,    384,          YM), /* 02130 - Mad Max - Life's A Bitch - Loader.sndh */
TIMEDB_ENTRY( 9471e78e,  0,  13685,       YM+TA), /* 04863 - Jovis - 5 Tunes.sndh */
TIMEDB_ENTRY( 9476a9f7,  0,  10747,          YM), /* 00559 - Allister Brimble - Spellbound Dizzy.sndh */
TIMEDB_ENTRY( 9476a9f7,  1,    144,          YM), /* 00559 - Allister Brimble - Spellbound Dizzy.sndh */
TIMEDB_ENTRY( 9476a9f7,  2,    143,          YM), /* 00559 - Allister Brimble - Spellbound Dizzy.sndh */
TIMEDB_ENTRY( 94855d20,  0,   5568,          YM), /* 02014 - Mad Max - ST News 6.1.sndh */
TIMEDB_ENTRY( 94975963,  0,  12384,          YM), /* 01425 - Spector - Manic Demo - Greets.sndh */
TIMEDB_ENTRY( 949bc29d,  0,   8832,       YM+TA), /* 00789 - Crawdaddy - Freedom.sndh */
TIMEDB_ENTRY( 949e8058,  0,   1069,       YM+TA), /* 03846 - Unknown - Mewilo - Title.sndh */
TIMEDB_ENTRY( 94ae535a,  0,  12031,          YM), /* 03456 - 505-checkpoint^paradox - Sommarhack invite 2010.sndh */
TIMEDB_ENTRY( 94c6dcd9,  0,  15743,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  1,  10045,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  2,    959,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  3,   4639,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  4,   4799,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  5,    215,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  6,   1919,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  7,   9215,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  8,  19151,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9,  9,   9215,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9, 10,  11679,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94c6dcd9, 11,    215,          YM), /* 04461 - Charles Callet - Bobo.sndh */
TIMEDB_ENTRY( 94d7ca31,  0, 246018,    YM+TA+TD), /* 02350 - Tao - Cream - YM-Jam ( UMD 8730 ).sndh */
TIMEDB_ENTRY( 94eca603,  0,   2720,          YM), /* 01732 - Nick Jones - First Samurai.sndh */
TIMEDB_ENTRY( 94eca603,  1,   1280,          YM), /* 01732 - Nick Jones - First Samurai.sndh */
TIMEDB_ENTRY( 94eca603,  2,    319,          YM), /* 01732 - Nick Jones - First Samurai.sndh */
TIMEDB_ENTRY( 94eca603,  3,    499,          YM), /* 01732 - Nick Jones - First Samurai.sndh */
TIMEDB_ENTRY( 95081672,  0,   3286,          YM), /* 03386 - Unknown - Sailor.sndh */
TIMEDB_ENTRY( 950bdefd,  0,   6720,          YM), /* 04068 - Gary Wheaton - Douglas Rockmoor 1.sndh */
TIMEDB_ENTRY( 95388ec7,  0,   7913,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  1,   3440,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  2,   1304,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  3,    617,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  4,     32,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  5,     80,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95388ec7,  6,  11710,          YM), /* 03882 - Arnauld Chevallier - RGB.sndh */
TIMEDB_ENTRY( 95415f16,  0,  58063,       YM+TA), /* 04023 - Roggie - The Realm.sndh */
TIMEDB_ENTRY( 9552015e,  0,   2566,          YM), /* 04465 - Charles Callet - La Marque Jaune.sndh */
TIMEDB_ENTRY( 9552015e,  1,   4608,          YM), /* 04465 - Charles Callet - La Marque Jaune.sndh */
TIMEDB_ENTRY( 9552015e,  2,   5444,          YM), /* 04465 - Charles Callet - La Marque Jaune.sndh */
TIMEDB_ENTRY( 955241c3,  0,     55,          YM), /* 02407 - Augustin Vidovic - Ploing.sndh */
TIMEDB_ENTRY( 9563057b,  0,  30457,          YM), /* 01717 - Jess - Tralali and Tralala.sndh */
TIMEDB_ENTRY( 95ab22e0,  0,  10944,          YM), /* 04471 - Charles Callet - Sidewalk.sndh */
TIMEDB_ENTRY( 95c01bf7,  0,  10846,       YM+TA), /* 04781 - LSL - NPRDKT.sndh */
TIMEDB_ENTRY( 95c8b74d,  0,  37202,       YM+TA), /* 03596 - 505-paradox - hushrush (STE).sndh */
TIMEDB_ENTRY( 95db7840,  0,   1595,       YM+TA), /* 02644 - miggy - Stackers(Unfinished).sndh */
TIMEDB_ENTRY( 961366b6,  0,   3577, YM+TA+TB+TD), /* 01940 - MSG - Chu Chu Rocket - Awards.sndh */
TIMEDB_ENTRY( 961a7e6f,  0,  11519, YM+TA+TB+TD), /* 03017 - dubmood - Jeron tels Alloyrun.sndh */
TIMEDB_ENTRY( 96415d60,  0,  16000,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  1,   6144,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  2,  12480,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  3,  17280,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  4,  11732,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  5,   8383,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96415d60,  6,   8192,       YM+TA), /* 02193 - Mad Max - Ghostbattle.sndh */
TIMEDB_ENTRY( 96568789,  0,  29952,          YM), /* 00852 - Ben Daglish - 3d Galax.sndh */
TIMEDB_ENTRY( 9676dae3,  0,  29121,    YM+TA+TD), /* 02338 - Tao - Cream - Magnetic (UCM 12).sndh */
TIMEDB_ENTRY( 9688321a,  0,  12000,       YM+TD), /* 02360 - Nexus 6 - Arie EXX.sndh */
TIMEDB_ENTRY( 968efd32,  0,   7180,    YM+TB+TD), /* 04384 - Christophe Zurfluh - Baby Jo in Going Home.sndh */
TIMEDB_ENTRY( 968efd32,  1,   4208,    YM+TB+TD), /* 04384 - Christophe Zurfluh - Baby Jo in Going Home.sndh */
TIMEDB_ENTRY( 968efd32,  2,   6211,    YM+TB+TD), /* 04384 - Christophe Zurfluh - Baby Jo in Going Home.sndh */
TIMEDB_ENTRY( 968efd32,  3,   4015,    YM+TB+TD), /* 04384 - Christophe Zurfluh - Baby Jo in Going Home.sndh */
TIMEDB_ENTRY( 968efd32,  4,   1899,    YM+TB+TD), /* 04384 - Christophe Zurfluh - Baby Jo in Going Home.sndh */
TIMEDB_ENTRY( 969aa546,  0,  16129,          YM), /* 00378 - Wally Beben - The Ballgame.sndh */
TIMEDB_ENTRY( 969f3663,  0,   2785,          YM), /* 03978 - Paul Summers - Sonic Boom.sndh */
TIMEDB_ENTRY( 969f3663,  1,   1921,          YM), /* 03978 - Paul Summers - Sonic Boom.sndh */
TIMEDB_ENTRY( 969f3663,  2,   2941,          YM), /* 03978 - Paul Summers - Sonic Boom.sndh */
TIMEDB_ENTRY( 969f3663,  3,   2689,          YM), /* 03978 - Paul Summers - Sonic Boom.sndh */
TIMEDB_ENTRY( 969f3663,  4,   2304,          YM), /* 03978 - Paul Summers - Sonic Boom.sndh */
TIMEDB_ENTRY( 96a2d403,  0,  33280, YM+TA+TB+TD), /* 04642 - Jakub Husak - n-a.sndh */
TIMEDB_ENTRY( 96bf9825,  0,   2879,          YM), /* 01310 - Matt Furniss - Race Drivin.sndh */
TIMEDB_ENTRY( 96e7cd5d,  0,  12672,          YM), /* 02020 - Mad Max - Balloon Challenge.sndh */
TIMEDB_ENTRY( 96e7cd5d,  1,    321,          YM), /* 02020 - Mad Max - Balloon Challenge.sndh */
TIMEDB_ENTRY( 96e80ca2,  0,   6144,          YM), /* 00491 - Ultimatum - Kidney Bean 7.sndh */
TIMEDB_ENTRY( 97182e59,  0,   3840,       YM+TA), /* 04428 - Lotus - Synth Dream 2 - Inna Groovy Mood.sndh */
TIMEDB_ENTRY( 971a913a,  0,   6079,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  1,    250,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  2,   4222,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  3,   2367,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  4,    155,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  5,   2686,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  6,   3166,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  7,   1739,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  8,   1919,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a,  9,   2239,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a, 10,   2879,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a, 11,    766,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a, 12,    138,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a, 13,   1147,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971a913a, 14,    716,          YM), /* 04130 - Matt Furniss - Shadow Warriors.sndh */
TIMEDB_ENTRY( 971fb4d9,  0,  33280, YM+TA+TB+TD), /* 03335 - Marcer - LATINA - NWD Ingame - Extended V..sndh */
TIMEDB_ENTRY( 97285509,  0,  10368,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  1,   8194,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  2,   8327,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  3,   9600,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  4,  12480,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  5,  10874,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  6,   6912,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  7,   9984,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  8,   9104,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509,  9,   5376,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509, 11,   5376,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 97285509, 12,   5506,       YM+TA), /* 02660 - Lotus-Hotline - No Buddies Land.sndh */
TIMEDB_ENTRY( 9728a920,  0,   1440,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 9728a920,  1,    891,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 9728a920,  2,   1920,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 9728a920,  3,    320,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 9728a920,  4,    960,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 9728a920,  5,    972,          YM), /* 01878 - Dave Lowe - California Games 2.sndh */
TIMEDB_ENTRY( 97431cba,  0,   2304,       YM+TD), /* 00587 - Celtic - The! Loader Track.sndh */
TIMEDB_ENTRY( 97510d6f,  0,   8460,       YM+TA), /* 01278 - Frazer - TSC 60.sndh */
TIMEDB_ENTRY( 97702443,  0,   5880,          YM), /* 01260 - Tim Follin - Renegade.sndh */
TIMEDB_ENTRY( 97872dee,  0,  16896,    YM+TB+TD), /* 03458 - Marcer - Axel F.sndh */
TIMEDB_ENTRY( 978a8d03,  0,   6720,       YM+TA), /* 00938 - Dark Angel - Mr X Must Die.sndh */
TIMEDB_ENTRY( 97ae217f,  0,   1536,          YM), /* 03240 - T1000 - The Beast Club - TBC 12.sndh */
TIMEDB_ENTRY( 97c8362b,  0,   7700,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  1,   9276,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  2,   5756,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  3,   7516,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  4,   7996,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  5,   1276,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  6,    278,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  8,   1276,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b,  9,   2075,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b, 10,    156,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97c8362b, 11,    156,          YM), /* 00558 - Allister Brimble - Slightly Magic.sndh */
TIMEDB_ENTRY( 97cc1c60,  0,   9216, YM+TA+TB+TD), /* 02098 - Mad Max - Starpaws - Sid.sndh */
TIMEDB_ENTRY( 97cc1c60,  1,   2688, YM+TA+TB+TD), /* 02098 - Mad Max - Starpaws - Sid.sndh */
TIMEDB_ENTRY( 97cc1c60,  2,    384, YM+TA+TB+TD), /* 02098 - Mad Max - Starpaws - Sid.sndh */
TIMEDB_ENTRY( 97eb9731,  0,   2928,          YM), /* 00453 - Augustin Vidovic - Motogoto.sndh */
TIMEDB_ENTRY( 97f3ab2e,  0,   3840,       YM+TA), /* 00952 - Dark Angel - The Devil In You.sndh */
TIMEDB_ENTRY( 9805afa1,  0,   5376,       YM+TA), /* 04440 - Lotus - Synth Dream 2 - You Know Lotus.sndh */
TIMEDB_ENTRY( 981baf1f,  0,   1117,          YM), /* 01536 - Iso - Waking Up.sndh */
TIMEDB_ENTRY( 9828011a,  0,   4986,          YM), /* 00899 - Danny Oneway - Danny O Sidsong.sndh */
TIMEDB_ENTRY( 98294817,  0,   6528,          YM), /* 04845 - t1000 - Bits 57.sndh */
TIMEDB_ENTRY( 98376a21,  0,   5376,       YM+TD), /* 02011 - Jochen Hippel - Loading.sndh */
TIMEDB_ENTRY( 984089bf,  0,   1536,          YM), /* 02213 - Mad Max - Rollout.sndh */
TIMEDB_ENTRY( 984089bf,  1,  10677,          YM), /* 02213 - Mad Max - Rollout.sndh */
TIMEDB_ENTRY( 984266e9,  0,   3881,       YM+TA), /* 03787 - Sean O'Connor - Guy Fawkes.sndh */
TIMEDB_ENTRY( 9867eb14,  0,    240,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  1,    372,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  2,    336,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  3,    336,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  4,    372,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  5,    336,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  6,    264,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  7,    336,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  8,   1119,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14,  9,    512,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14, 10,    503,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14, 11,    768,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14, 12,    570,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14, 13,   1200,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9867eb14, 14,     40,          YM), /* 00341 - Tony Williams - International Ice Hockey.sndh */
TIMEDB_ENTRY( 9874059f,  0,   9226,          YM), /* 00426 - Big Alec - PYM - Isido Way Of The Stones.sndh */
TIMEDB_ENTRY( 9880a1dd,  0,  14495,          YM), /* 02375 - Nexus 6 - Hans A. aus Dortmund.sndh */
TIMEDB_ENTRY( 988903d2,  0,   3838,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  1,   6079,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  2,   8319,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  3,   5759,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  4,   8319,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  5,   8062,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  6,   5439,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 988903d2,  7,   6079,          YM), /* 04219 - Karsten Obarksi - Window Wizard.sndh */
TIMEDB_ENTRY( 9898be70,  0,  24832,          YM), /* 02042 - Mad Max - Lightforce.sndh */
TIMEDB_ENTRY( 98d188fc,  0,  10240,          YM), /* 02589 - STU - DRUMBASE.sndh */
TIMEDB_ENTRY( 98d78363,  0,  17913,       YM+TA), /* 04223 - 505 - 3rd.sndh */
TIMEDB_ENTRY( 98da9c76,  0,   7168,          YM), /* 01264 - Frazer - Apex.sndh */
TIMEDB_ENTRY( 98ee413e,  0,  13871,       YM+TA), /* 04120 - Tao - Powerman - STMYGM 2 version.sndh */
TIMEDB_ENTRY( 98f2466f,  0,   5796, YM+TA+TB+TD), /* 03430 - Per Almered (XiA) 2009 - Brit.sndh */
TIMEDB_ENTRY( 98fd4dbd,  0,   2497,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  1,   1542,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  2,   1634,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  3,   2178,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  4,    339,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  5,    228,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  6,    181,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  7,   2497,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  8,    176,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd,  9,    135,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 10,    126,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 11,   3183,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 12,   3138,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 13,   2560,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 14,    482,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 15,    229,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 16,    369,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 17,    264,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 19,   2688,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 20,    205,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 21,    419,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 23,    512,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 98fd4dbd, 24,   2285,          YM), /* 04506 - Unknown - Bubblefield screen.sndh */
TIMEDB_ENTRY( 990a94c4,  0,   2309,    YM+TA+TD), /* 02230 - Tao - Cream - Chased-3D Cube (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( 9911c535,  0, 226959,       YM+TA), /* 04672 - Unknown - Sweet Dreams.sndh */
TIMEDB_ENTRY( 99327bfa,  0,   1152,          YM), /* 01087 - Techno - Dentro 2.sndh */
TIMEDB_ENTRY( 99671248,  0,   3450,          YM), /* 01342 - Scavenger - Leif Rular.sndh */
TIMEDB_ENTRY( 997d6e29,  0,  27648, YM+TA+TB+TD), /* 03376 - STU - SINGING ROBOT.sndh */
TIMEDB_ENTRY( 999d9ffe,  0,   7296,       YM+TA), /* 04422 - Lotus - Synth Dream 2 - Bionic Fields.sndh */
TIMEDB_ENTRY( 99b0b0a8,  0,   7168,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b0b0a8,  1,  15168,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b0b0a8,  2,   4180,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b0b0a8,  3,  18483,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b0b0a8,  5,  10908,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b0b0a8,  6,  12160,       YM+TA), /* 02204 - Mad Max - Lethal Xcess (ST-Falc).sndh */
TIMEDB_ENTRY( 99b17b4f,  0,    112,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  1,    103,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  2,   3113,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  3,   2497,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  4,    181,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  5,    160,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  6,   2285,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b17b4f,  7,   1146,          YM), /* 04715 - Simon Wilsher - Spaced Weller.sndh */
TIMEDB_ENTRY( 99b3dad9,  0,   8853,          YM), /* 01971 - MSG - MSG 01.sndh */
TIMEDB_ENTRY( 99b563c1,  0,    494,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  1,    142,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  2,    180,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  3,    198,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  4,    355,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  5,    169,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99b563c1,  6,    199,          YM), /* 04493 - Ralph Effemey - Arthur of the Britons.sndh */
TIMEDB_ENTRY( 99cea52d,  0,    871,          YM), /* 00098 - David Whittaker - Enduro Racer.sndh */
TIMEDB_ENTRY( 99cea52d,  1,   8533,          YM), /* 00098 - David Whittaker - Enduro Racer.sndh */
TIMEDB_ENTRY( 99cea52d,  2,    853,          YM), /* 00098 - David Whittaker - Enduro Racer.sndh */
TIMEDB_ENTRY( 99d0693a,  0,  13405,          YM), /* 03893 - Unknown - Sevilla.sndh */
TIMEDB_ENTRY( 99f889ef,  0,   5435, YM+TA+TB+TD), /* 04568 - Bagarn - Spice Boys - Jag fyller Assn till bredden.sndh */
TIMEDB_ENTRY( 99ff9925,  0,   7680,          YM), /* 03215 - Tomchi - Checkboard.sndh */
TIMEDB_ENTRY( 9a025a97,  0,  18080,    YM+TB+TD), /* 02604 - STU - MAD CREATURE.sndh */
TIMEDB_ENTRY( 9a04d1ff,  0,   7913,          YM), /* 00876 - Ben Daglish - Thunderbirds.sndh */
TIMEDB_ENTRY( 9a1dc958,  0,   9689,          YM), /* 03033 - Debbie Davis - Lay Me Down.sndh */
TIMEDB_ENTRY( 9a315c32,  0,  32256, YM+TA+TB+TD), /* 02536 - Remo - Mia Mia.sndh */
TIMEDB_ENTRY( 9a321c58,  0,  10368,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a321c58,  1,   6144,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a321c58,  2,    161,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a321c58,  3,    121,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a321c58,  4,   8320,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a321c58,  5,   3840,          YM), /* 01067 - Jonathan Dunn - Rambo III.sndh */
TIMEDB_ENTRY( 9a35c9be,  0,  18705,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  1,   6826,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  2,     93,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  3,    135,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  4,     79,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  5,   8577,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  6,    332,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9a35c9be,  7,    380,          YM), /* 00094 - David Whittaker - Dogs of War.sndh */
TIMEDB_ENTRY( 9ad9d815,  0,    413,          YM), /* 03203 - Unknown - World Trophy Soccer.sndh */
TIMEDB_ENTRY( 9ad9d815,  1,    459,          YM), /* 03203 - Unknown - World Trophy Soccer.sndh */
TIMEDB_ENTRY( 9ad9d815,  2,    312,          YM), /* 03203 - Unknown - World Trophy Soccer.sndh */
TIMEDB_ENTRY( 9ae3a822,  0,  32256, YM+TA+TB+TD), /* 03996 - XyNo - XC3N BDAY 48H.sndh */
TIMEDB_ENTRY( 9b14e618,  0,  13058, YM+TA+TB+TD), /* 00206 - 505 - UCM 12 2.sndh */
TIMEDB_ENTRY( 9b1fc0cc,  0,   2559,          YM), /* 04209 - Dave Lowe - Ghostbusters II.sndh */
TIMEDB_ENTRY( 9b2b2f8a,  0,   4224,       YM+TA), /* 04419 - EVE - Sewer 8.sndh */
TIMEDB_ENTRY( 9b391918,  0,   4092, YM+TA+TB+TD), /* 04328 - 505-Checkpoint - zuuuuuuu.sndh */
TIMEDB_ENTRY( 9b3cda13,  0, 268756,       YM+TA), /* 03757 - Tufty - Pandemonium - Acid Demo.sndh */
TIMEDB_ENTRY( 9b427428,  0,   9967, YM+TA+TB+TD), /* 04351 - HYLST - H CHIP EWPERIMENT 4 SV2K13.sndh */
TIMEDB_ENTRY( 9b4f3100,  0,  41554,       YM+TA), /* 04059 - Roggie - The Works.sndh */
TIMEDB_ENTRY( 9b57347c,  0, 153412,       YM+TA), /* 03711 - Wally Beben - Time Machine.sndh */
TIMEDB_ENTRY( 9b66fd97,  0,  25552, YM+TA+TB+TD), /* 02861 - Helene - whatever.sndh */
TIMEDB_ENTRY( 9b7a9b87,  0,   1934,          YM), /* 00926 - Dark Angel - Invaders.sndh */
TIMEDB_ENTRY( 9b82e7ae,  0,  18816,          YM), /* 03259 - T1000 - The Beast Club - TS 8.sndh */
TIMEDB_ENTRY( 9b855101,  0,  10752,          YM), /* 01026 - Doclands - European Demos Main Menu.sndh */
TIMEDB_ENTRY( 9b8e4792,  0,   4638,          YM), /* 00838 - Ben Daglish - Action Fighter.sndh */
TIMEDB_ENTRY( 9b8e4792,  1,    709,          YM), /* 00838 - Ben Daglish - Action Fighter.sndh */
TIMEDB_ENTRY( 9ba23b31,  0, 114186,       YM+TA), /* 04056 - Roggie - Shoo and away ye go.sndh */
TIMEDB_ENTRY( 9ba720f0,  0,   1600,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  1,     57,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  2,     52,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  3,     49,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  4,     51,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  5,     50,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9ba720f0,  6,   1021,          YM), /* 04497 - A.C. Porritt - Battle Zone.sndh */
TIMEDB_ENTRY( 9bc97355,  0,  67569,       YM+TA), /* 03834 - Charles Callet - Pick N Pile.sndh */
TIMEDB_ENTRY( 9bc97355,  1,  64431,       YM+TA), /* 03834 - Charles Callet - Pick N Pile.sndh */
TIMEDB_ENTRY( 9bd0d1e4,  0,   4048, YM+TA+TB+TD), /* 02813 - Marcer - NWD INTRO.sndh */
TIMEDB_ENTRY( 9bda4dde,  0,   4095,          YM), /* 01785 - Lap - Phaleon - Krom.sndh */
TIMEDB_ENTRY( 9bf79fd9,  0,   6912,          YM), /* 01189 - Excellence In Art - Lynx Tune.sndh */
TIMEDB_ENTRY( 9bfc2d55,  0,  49728,    YM+TA+TB), /* 01675 - Jess - Best Part  Ze 16 Mins Hit.sndh */
TIMEDB_ENTRY( 9c00376c,  0,   8832,    YM+TA+TD), /* 02249 - Tao - Cream - Androids.sndh */
TIMEDB_ENTRY( 9c1a5c7a,  0,   4498,          YM), /* 03143 - Unknown - The First Noel the Angel ....sndh */
TIMEDB_ENTRY( 9c1ab919,  0,  11787,    YM+TA+TD), /* 02251 - Tao - Cream - Chicago (Steps).sndh */
TIMEDB_ENTRY( 9c66c9d0,  0,   4352,       YM+TA), /* 00813 - Crawdaddy - The Cave.sndh */
TIMEDB_ENTRY( 9c72511a,  0,   1536,          YM), /* 02004 - Mad Max - Demo Shorty.sndh */
TIMEDB_ENTRY( 9c8bb406,  0,   8442,    YM+TA+TD), /* 01140 - Tyan - Julia 4.sndh */
TIMEDB_ENTRY( 9c92d01b,  0,   1365,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  1,    420,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  2,   1072,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  3,   2048,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  4,    193,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  5,   2389,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  6,   1150,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  7,   3360,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  8,   2507,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b,  9,   1337,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 10,   3510,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 11,   1671,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 12,   1310,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 13,   4555,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 14,   7105,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 15,     87,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9c92d01b, 16,    335,          YM), /* 00146 - David Whittaker - Super Wonder Boy.sndh */
TIMEDB_ENTRY( 9cb76943,  0,   5120,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cb76943,  1,   1216,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cb76943,  2,    319,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cb76943,  3,    214,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cb76943,  4,    608,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cb76943,  5,    512,          YM), /* 00344 - Tony Williams - Nitro.sndh */
TIMEDB_ENTRY( 9cc37862,  0,   8042,          YM), /* 02145 - Mad Max - International Karate 1 - LSD.sndh */
TIMEDB_ENTRY( 9cc737cc,  0,  27136,          YM), /* 04643 - xrwwr - WINGS OF LIFE.sndh */
TIMEDB_ENTRY( 9cc8b789,  0,   1920,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  1,   1152,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  2,   1152,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  3,   2976,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  4,   1198,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  5,   5902,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  6,   2300,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  7,   1536,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9cc8b789,  8,   3264,          YM), /* 01111 - Keith Tinman - Midnight Resistance.sndh */
TIMEDB_ENTRY( 9ccc87fe,  0,    496,          YM), /* 04133 - Unknown - Carry All.sndh */
TIMEDB_ENTRY( 9ccc87fe,  1,    483,          YM), /* 04133 - Unknown - Carry All.sndh */
TIMEDB_ENTRY( 9cd87c01,  0,   6412,          YM), /* 00855 - Ben Daglish - Kingmaker.sndh */
TIMEDB_ENTRY( 9cd87c01,  1,   3790,          YM), /* 00855 - Ben Daglish - Kingmaker.sndh */
TIMEDB_ENTRY( 9cd87c01,  2,    402,          YM), /* 00855 - Ben Daglish - Kingmaker.sndh */
TIMEDB_ENTRY( 9cd87c01,  3,    324,          YM), /* 00855 - Ben Daglish - Kingmaker.sndh */
TIMEDB_ENTRY( 9cd87c01,  4,    447,          YM), /* 00855 - Ben Daglish - Kingmaker.sndh */
TIMEDB_ENTRY( 9cf60b4b,  0,    550,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  1,    629,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  2,    568,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  3,    532,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  4,    592,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  5,   1492,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  6,   1866,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  7,   1346,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  8,    309,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b,  9,    272,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cf60b4b, 10,    255,          YM), /* 04540 - JAWX - Funfares (STOS accessories disk).sndh */
TIMEDB_ENTRY( 9cfdf4ac,  0,   9983,       YM+TA), /* 03619 - 505-paradox - Blue.sndh */
TIMEDB_ENTRY( 9d0dc76f,  0,   8704,          YM), /* 00580 - Ladislav John Brovosky - Hades Nebula.sndh */
TIMEDB_ENTRY( 9d0dc76f,  1,    236,          YM), /* 00580 - Ladislav John Brovosky - Hades Nebula.sndh */
TIMEDB_ENTRY( 9d0dc76f,  2,    831,          YM), /* 00580 - Ladislav John Brovosky - Hades Nebula.sndh */
TIMEDB_ENTRY( 9d2a8769,  0,   1536,          YM), /* 04717 - Unknown - The Big STOS demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  0,   7680,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  1,   4416,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  2,   5184,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  3,   7488,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  4,   5760,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  5,   3840,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d2eaeb3,  6,   5760,          YM), /* 02034 - Mad Max - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( 9d3bad89,  0,    403,          YM), /* 04077 - Gary Wheaton - Jingle.sndh */
TIMEDB_ENTRY( 9d448bf8,  0,   8328,    YM+TB+TD), /* 03589 - Divag - Boomjasz.sndh */
TIMEDB_ENTRY( 9d50b1cc,  0,  13248,          YM), /* 02007 - Mad Max - Gambol.sndh */
TIMEDB_ENTRY( 9d50b1cc,  1,   3082,          YM), /* 02007 - Mad Max - Gambol.sndh */
TIMEDB_ENTRY( 9d50b1cc,  2,   7680,          YM), /* 02007 - Mad Max - Gambol.sndh */
TIMEDB_ENTRY( 9d5ee854,  0,  22848, YM+TA+TB+TD), /* 02084 - Mad Max - IK+ - Sid.sndh */
TIMEDB_ENTRY( 9d6a644a,  0,   3648,          YM), /* 03514 - Gominolas - Pepsi - Mad Mix Game.sndh */
TIMEDB_ENTRY( 9d6c1090,  0,  20730,       YM+TA), /* 01722 - Jess - Another Western Legend.sndh */
TIMEDB_ENTRY( 9d70cdba,  0,   3072,          YM), /* 02131 - Mad Max - Axel F - LCD.sndh */
TIMEDB_ENTRY( 9d8449ea,  0,  12202, YM+TA+TB+TD), /* 03642 - e^pi-Tristesse - silly cone.sndh */
TIMEDB_ENTRY( 9d96a836,  0,  15546, YM+TA+TB+TD), /* 00990 - DMA-SC - Chipster.sndh */
TIMEDB_ENTRY( 9da43427,  0,   1759,          YM), /* 03115 - Unknown - FUR ELISE by L. V. Beethoven.sndh */
TIMEDB_ENTRY( 9df47e85,  0,   4226,          YM), /* 01228 - Epic - Feed My Frankenstein.sndh */
TIMEDB_ENTRY( 9e23b2be,  0,   6976,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  1,   7680,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  2,   7680,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  3,    506,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  4,   9216,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  5,   8704,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  6,   7936,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e23b2be,  7,   7680,          YM), /* 01375 - Justin Scharvona & Andy Severn - Kid Gloves 2.sndh */
TIMEDB_ENTRY( 9e2769bd,  0,   3456,          YM), /* 03659 - Jedi - Nothing Is More - Song 1.sndh */
TIMEDB_ENTRY( 9e32e31d,  0,   6136,       YM+TA), /* 01335 - Scavenger - Heaven.sndh */
TIMEDB_ENTRY( 9e42a039,  0,   4511,          YM), /* 00332 - Gary Wheaton - Greetings.sndh */
TIMEDB_ENTRY( 9e53db7d,  0,  10610,    YM+TA+TD), /* 02750 - sh3 - house of the rising sun.sndh */
TIMEDB_ENTRY( 9e5a0592,  0,   2402,          YM), /* 01791 - Barry Leitch - Ferrari Formula One.sndh */
TIMEDB_ENTRY( 9e5fedcd,  0,  38400,          YM), /* 01448 - Stax - Scandal.sndh */
TIMEDB_ENTRY( 9e69482f,  0,  13433,          YM), /* 01687 - Jess - The singing Green Frog.sndh */
TIMEDB_ENTRY( 9e776126,  0,  26800,          YM), /* 01723 - Jess - If It Works Good !.sndh */
TIMEDB_ENTRY( 9e8a8d67,  0,   2569,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  1,   3853,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  2,   3849,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  3,   4621,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  4,   4047,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  5,   3085,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  6,   2569,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  7,   3175,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  8,   5393,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9e8a8d67,  9,   4915,          YM), /* 04187 - Jezz Woodroffe - Philip Nixon - Elvira II - The Jaws of Cerberus.sndh */
TIMEDB_ENTRY( 9ea21686,  0,   6048,       YM+TA), /* 00915 - Dark Angel - Mathematica 1.sndh */
TIMEDB_ENTRY( 9ea94a32,  0,  18433,       YM+TA), /* 04800 - Unknown - Delta.sndh */
TIMEDB_ENTRY( 9ebbca94,  0,   4322,          YM), /* 00243 - AN Cool - Atari-Baby.sndh */
TIMEDB_ENTRY( 9ec32bc8,  0,   7680,       YM+TB), /* 00040 - 505 - Checkpoint - MMJ Xmas 03 - 1.sndh */
TIMEDB_ENTRY( 9ec94363,  0,   3456,       YM+TA), /* 00414 - Big Alec - Melancho.sndh */
TIMEDB_ENTRY( 9edc4972,  0, 106299,       YM+TA), /* 03753 - Tufty - Mirage.sndh */
TIMEDB_ENTRY( 9ee36e8a,  0,   1736,          YM), /* 04696 - P Blay - Mars Maze.sndh */
TIMEDB_ENTRY( 9ee36e8a,  1,    800,          YM), /* 04696 - P Blay - Mars Maze.sndh */
TIMEDB_ENTRY( 9ee36e8a,  2,   1306,          YM), /* 04696 - P Blay - Mars Maze.sndh */
TIMEDB_ENTRY( 9ee36e8a,  3,    879,          YM), /* 04696 - P Blay - Mars Maze.sndh */
TIMEDB_ENTRY( 9eeab97a,  0,  17701,          YM), /* 00916 - Dark Angel - Aura Mix.sndh */
TIMEDB_ENTRY( 9ef29429,  0,   3328,          YM), /* 00003 - gwEm - micromusic radio jingle.sndh */
TIMEDB_ENTRY( 9efd158b,  0,   8058,       YM+TA), /* 04450 - Scavenger - Sambase Digi.sndh */
TIMEDB_ENTRY( 9f094e8e,  0,   1275,          YM), /* 00026 - 505 - 18th.sndh */
TIMEDB_ENTRY( 9f144684,  0,   5787,          YM), /* 01232 - Epic - Lost in Universe.sndh */
TIMEDB_ENTRY( 9f148304,  0,  35840,          YM), /* 01468 - STU - HEDGEHOG.sndh */
TIMEDB_ENTRY( 9f1d23bd,  0,   4120,    YM+TA+TD), /* 02260 - Tao - Cream (orig. Rob Hubbard) - Ninja (Steps).sndh */
TIMEDB_ENTRY( 9f265e3d,  0,   7674,          YM), /* 02455 - Timbral - Y'M still here.sndh */
TIMEDB_ENTRY( 9f2aeb7c,  0,  11065, YM+TA+TB+TD), /* 02100 - Mad Max - Thing on a Spring - Sid.sndh */
TIMEDB_ENTRY( 9f373ded,  0,  20601,          YM), /* 01672 - Jess - Akira VS Sangoku.sndh */
TIMEDB_ENTRY( 9f633da5,  0,  16512,          YM), /* 04457 - Charles Callet - Final Assault.sndh */
TIMEDB_ENTRY( 9f783efb,  0,   4223,       YM+TA), /* 00200 - 505 - TK Machine.sndh */
TIMEDB_ENTRY( 9f92ee0c,  0,   8448,          YM), /* 01556 - Riff - Bomb Fusion.sndh */
TIMEDB_ENTRY( 9f94483d,  0,  12288,       YM+TD), /* 03997 - XyNo - n-a.sndh */
TIMEDB_ENTRY( 9f9606fe,  0,   5376,          YM), /* 00967 - Mike Davies - Strider.sndh */
TIMEDB_ENTRY( 9f9606fe,  1,    912,          YM), /* 00967 - Mike Davies - Strider.sndh */
TIMEDB_ENTRY( 9f9606fe,  2,    265,          YM), /* 00967 - Mike Davies - Strider.sndh */
TIMEDB_ENTRY( 9f9606fe,  3,    289,          YM), /* 00967 - Mike Davies - Strider.sndh */
TIMEDB_ENTRY( 9f9606fe,  4,   2944,          YM), /* 00967 - Mike Davies - Strider.sndh */
TIMEDB_ENTRY( 9fa83622,  0,  24570,       YM+TA), /* 01691 - Jess - Pentatonik Illusion.sndh */
TIMEDB_ENTRY( 9fa890e4,  0,  12672,    YM+TA+TB), /* 01325 - Scavenger - Avenger.sndh */
TIMEDB_ENTRY( 9faed090,  0,  13568,       YM+TA), /* 02717 - Lee Lee - Automation 391.sndh */
TIMEDB_ENTRY( 9fb5b4eb,  0,   2688, YM+TA+TB+TD), /* 01721 - Jess - Wave Mania is my Passion.sndh */
TIMEDB_ENTRY( 9fbc887d,  0,   5763,          YM), /* 01758 - Lap - Bird Girl Mad Demo Menu.sndh */
TIMEDB_ENTRY( 9fc7da3a,  0,   4607, YM+TA+TB+TD), /* 02840 - maktone-fairlight - varberga.sndh */
TIMEDB_ENTRY( 9fd943e9,  0,   2458,          YM), /* 03123 - n-a - .sndh */
TIMEDB_ENTRY( 9fe1c921,  0,   6401,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  1,    105,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  2,    163,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  3,    643,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  4,    163,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  5,    145,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  6,    135,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  7,    145,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe1c921,  8,    265,          YM), /* 00361 - Steve Barrett - Adv Fruit Machine Sim.sndh */
TIMEDB_ENTRY( 9fe29d4d,  0,  12031,          YM), /* 03022 - DUBMOOD ET GOTO80 - DBUG MENU 197.sndh */
TIMEDB_ENTRY( 9ffd1026,  0,   8057, YM+TA+TB+TD), /* 00191 - 505 - Space.sndh */
TIMEDB_ENTRY( a008ebcd,  0,   5799, YM+TA+TB+TD), /* 03519 - D-Force - Time Ticking By.sndh */
TIMEDB_ENTRY( a00a62ab,  0,   8636,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  1,     76,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  2,    291,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  3,    128,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  4,    371,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  5,    636,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  6,    283,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a00a62ab,  7,    283,          YM), /* 00546 - Allister Brimble - 1st Division Manager.sndh */
TIMEDB_ENTRY( a014ee7d,  0,   7447,          YM), /* 00127 - David Whittaker - The Real Ghostbusters.sndh */
TIMEDB_ENTRY( a014ee7d,  1,     38,          YM), /* 00127 - David Whittaker - The Real Ghostbusters.sndh */
TIMEDB_ENTRY( a014ee7d,  2,    123,          YM), /* 00127 - David Whittaker - The Real Ghostbusters.sndh */
TIMEDB_ENTRY( a014ee7d,  3,    303,          YM), /* 00127 - David Whittaker - The Real Ghostbusters.sndh */
TIMEDB_ENTRY( a02678df,  0,  60994,       YM+TA), /* 04556 - Simon O - Bangkok Knights Loader.sndh */
TIMEDB_ENTRY( a0308e31,  0,   8930,          YM), /* 00109 - David Whittaker - Ironlord.sndh */
TIMEDB_ENTRY( a0308e31,  1,   5071,          YM), /* 00109 - David Whittaker - Ironlord.sndh */
TIMEDB_ENTRY( a0308e31,  2,   7606,          YM), /* 00109 - David Whittaker - Ironlord.sndh */
TIMEDB_ENTRY( a0390153,  0,   5372, YM+TA+TB+TD), /* 03521 - D-Force - FUN - STNICCC.sndh */
TIMEDB_ENTRY( a03c807c,  0,   4608,          YM), /* 01159 - Tyan - Some Sleep.sndh */
TIMEDB_ENTRY( a0583f2f,  0,  16896, YM+TA+TB+TD), /* 02594 - STU - GRAVE DISCO.sndh */
TIMEDB_ENTRY( a0689dfd,  0,    889,       YM+TD), /* 01958 - MSG - Godpey - Intro.sndh */
TIMEDB_ENTRY( a07ef087,  0,  95684,       YM+TA), /* 03703 - Sparky - Badd Boy - Lode Runner Forever.sndh */
TIMEDB_ENTRY( a09bee55,  0,  27508,       YM+TA), /* 03744 - Tufty - C64 Caz.sndh */
TIMEDB_ENTRY( a0a62086,  0,  26624,          YM), /* 04458 - Charles Callet - Bob Morane - Jungle.sndh */
TIMEDB_ENTRY( a0a62086,  1,   3537,          YM), /* 04458 - Charles Callet - Bob Morane - Jungle.sndh */
TIMEDB_ENTRY( a0ab174b,  0,  54502,       YM+TA), /* 03833 - Charles Callet - Advantage Tennis.sndh */
TIMEDB_ENTRY( a0ab174b,  1, 191166,       YM+TA), /* 03833 - Charles Callet - Advantage Tennis.sndh */
TIMEDB_ENTRY( a0ab174b,  2,  85210,       YM+TA), /* 03833 - Charles Callet - Advantage Tennis.sndh */
TIMEDB_ENTRY( a0ae25f0,  0,   9409,    YM+TA+TD), /* 04202 - Tao - Cream - Nostalgic-O-Demo Main Menu (2 Sid Voice).sndh */
TIMEDB_ENTRY( a0afd16f,  0,  11279,          YM), /* 02284 - Tao - Cream - Ride The Sky (Just Buggin).sndh */
TIMEDB_ENTRY( a0c219dd,  0, 312206,       YM+TA), /* 03794 - Jim Smelley & Steve Schrade - Disciples of Steel.sndh */
TIMEDB_ENTRY( a0c8ef8f,  0,  10464,          YM), /* 02152 - Mad Max - Ooh Crikey - Ashtray.sndh */
TIMEDB_ENTRY( a0df0320,  0,   6306,          YM), /* 03027 - Alice Amore - AULD LANG SYNE, arr.by A.Amore.sndh */
TIMEDB_ENTRY( a0e5eff7,  0,   4254,          YM), /* 04343 - Nexus 6 - Ga!.sndh */
TIMEDB_ENTRY( a0e9f82d,  0,  46810,       YM+TA), /* 04360 - X The Beermat - Night.sndh */
TIMEDB_ENTRY( a0f1e6ed,  0, 103897,       YM+TA), /* 03790 - Roggie - Disk Mag.sndh */
TIMEDB_ENTRY( a1159697,  0,   5584,          YM), /* 01929 - The Mind - Repatriated Nucleus.sndh */
TIMEDB_ENTRY( a129df5f,  0,  29532,    YM+TB+TD), /* 03855 - YERZMYEY-AY-RIDERS - *WAKE UP!* YERZ.sndh */
TIMEDB_ENTRY( a136bb47,  0,   5119,    YM+TA+TB), /* 02703 - Dubmood - Speedloader I.sndh */
TIMEDB_ENTRY( a17c886e,  0,   5375, YM+TA+TB+TD), /* 01016 - DMA-SC - UCM 16 Intro.sndh */
TIMEDB_ENTRY( a184cbea,  0,   2125, YM+TA+TB+TD), /* 03695 - Per Almered (XiA) 2009 - Injector.sndh */
TIMEDB_ENTRY( a1a6ba29,  0,   2560,          YM), /* 00968 - Mike Davies - UN Squadron.sndh */
TIMEDB_ENTRY( a1a6ba29,  1,    864,          YM), /* 00968 - Mike Davies - UN Squadron.sndh */
TIMEDB_ENTRY( a1a6ba29,  2,   3456,          YM), /* 00968 - Mike Davies - UN Squadron.sndh */
TIMEDB_ENTRY( a1af75f3,  0,     12,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  1,     57,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  2,     60,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  3,    138,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  4,     48,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  5,    876,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  6,     66,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  7,    165,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  8,    240,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3,  9,     81,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 10,     18,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 11,      6,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 12,    180,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 13,    117,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 14,    240,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 15,     21,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 16,     36,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 17,     39,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 18,     30,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 19,    240,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 20,     12,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 21,      6,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 22,     12,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 23,    150,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 24,    519,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 25,     72,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 26,     33,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 27,     12,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 28,     18,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 29,     39,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 30,    234,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 31,     48,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 32,    180,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 33,     45,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 34,     33,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 35,     18,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 36,     60,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 37,     39,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 38,     21,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 39,      9,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 40,    799,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 41,     12,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 42,    359,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 43,     39,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 44,     27,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 45,    436,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 46,     18,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 47,      9,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 49,     18,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 50,     39,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 51,    208,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 52,    405,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 53,    117,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 54,    172,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 55,     36,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 56,    222,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 57,    387,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 58,    972,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 59,     30,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 60,    354,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 61,    156,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1af75f3, 62,    376,          YM), /* 04740 - Barry Murray - Manhunter - New York.sndh */
TIMEDB_ENTRY( a1c3c7cd,  0,    639,          YM), /* 04087 - Gary Wheaton - Pacman.sndh */
TIMEDB_ENTRY( a1cd2fe3,  0,   3072,          YM), /* 03252 - T1000 - The Bits Club 32.sndh */
TIMEDB_ENTRY( a1d04ddb,  0,   3590,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  1,   3400,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  2,   3316,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  3,   4228,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  4,   2800,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  5,   3345,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  6,   3480,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  7,   3893,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  8,    576,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1d04ddb,  9,   1141,          YM), /* 04503 - Geoff Camp - Brain food - Letters.sndh */
TIMEDB_ENTRY( a1e592e4,  0,  17684, YM+TA+TB+TD), /* 02625 - STU - YMETAL.sndh */
TIMEDB_ENTRY( a1e8f58b,  0,   6144,          YM), /* 00268 - Antichrist - Verity.sndh */
TIMEDB_ENTRY( a205e610,  0,     92,          YM), /* 03929 - Unknown - Sound 3.sndh */
TIMEDB_ENTRY( a20ed582,  0,   8658,          YM), /* 04144 - Unknown - Alture.sndh */
TIMEDB_ENTRY( a21825a8,  0,  71723,       YM+TA), /* 04048 - Roggie - STC End.sndh */
TIMEDB_ENTRY( a21c1160,  0,   6647,          YM), /* 01924 - The Mind - Genesys Remix - Dynamika.sndh */
TIMEDB_ENTRY( a2408d27,  0,   1600,          YM), /* 00610 - Chrispy Noodle - Chrispy Noodle 1.sndh */
TIMEDB_ENTRY( a24ee054,  0,  14480, YM+TA+TB+TD), /* 03883 - LSL - Morphonic Lab XI.sndh */
TIMEDB_ENTRY( a259b297,  0,   2560,          YM), /* 01921 - The Mind - Phaleon - Aenigmatica.sndh */
TIMEDB_ENTRY( a2658ad7,  0,   4608, YM+TA+TB+TD), /* 02583 - STU - CRAPTRO.sndh */
TIMEDB_ENTRY( a27fabe6,  0,   9215, YM+TA+TB+TD), /* 02489 - gwEm (Hubbard conversion) - Disco Snax.sndh */
TIMEDB_ENTRY( a295d8c3,  0,  14976, YM+TA+TB+TD), /* 01604 - Nemo - Real Arcade.sndh */
TIMEDB_ENTRY( a2b87858,  0,    205,          YM), /* 02916 - Richard Joseph - Magic Pockets.sndh */
TIMEDB_ENTRY( a2bc8307,  0,  21520,    YM+TA+TD), /* 03051 - STU - DROPDABOMB - Chillout.sndh */
TIMEDB_ENTRY( a2d97155,  0,  35840,       YM+TA), /* 02588 - STU - DEAD HACKER.sndh */
TIMEDB_ENTRY( a2e23f33,  0,   6200,    YM+TA+TD), /* 04775 - dhor-mec - hurry harry, hurry.sndh */
TIMEDB_ENTRY( a2f94f3d,  0,   1440,          YM), /* 01250 - Janko Mrsic Flogel - Warzone.sndh */
TIMEDB_ENTRY( a31ddc78,  0,  15559,    YM+TA+TD), /* 03859 - tomchi  noextra - MUDA.sndh */
TIMEDB_ENTRY( a32f7822,  0,   5632,       YM+TA), /* 00262 - Antichrist - Slow.sndh */
TIMEDB_ENTRY( a34a0911,  0,  14469,          YM), /* 03030 - Ric Clayton - That's The Way   (Led Zepplin).sndh */
TIMEDB_ENTRY( a363e0cc,  0,   4320, YM+TA+TB+TD), /* 03001 - 505-paradox - Thing.sndh */
TIMEDB_ENTRY( a3739bfc,  0,   7104,          YM), /* 00781 - Crawdaddy - Chinese heroes.sndh */
TIMEDB_ENTRY( a39fc801,  0,   1921,          YM), /* 02292 - Tao - Cream - A-t-'s Rendermachine.sndh */
TIMEDB_ENTRY( a3c4adf4,  0,   7202,          YM), /* 01048 - Dr Beat - NTM - Dynamic Duo.sndh */
TIMEDB_ENTRY( a3c69210,  0,   7035, YM+TA+TB+TD), /* 00281 - Baggio - Rape Me.sndh */
TIMEDB_ENTRY( a3dac9cf,  0,  13442,          YM), /* 01044 - Dr Beat - Galtan 6 - Elite Intro.sndh */
TIMEDB_ENTRY( a3f7a518,  0,   2304,       YM+TA), /* 00261 - Antichrist - Sandra's Land.sndh */
TIMEDB_ENTRY( a4027d15,  0,   4100, YM+TA+TB+TD), /* 03156 - Starbuck-Code Monkz - maxYMum Time Pressure.sndh */
TIMEDB_ENTRY( a404d61f,  0,   5142,          YM), /* 01217 - Energizer - Lazer.sndh */
TIMEDB_ENTRY( a41968fa,  0,   1919,          YM), /* 01289 - Matt Furniss - Badlands.sndh */
TIMEDB_ENTRY( a41968fa,  1,   1920,          YM), /* 01289 - Matt Furniss - Badlands.sndh */
TIMEDB_ENTRY( a41968fa,  2,   5760,          YM), /* 01289 - Matt Furniss - Badlands.sndh */
TIMEDB_ENTRY( a425d896,  0,    768,          YM), /* 01151 - Tyan - Revolution - Piep Piep Piep.sndh */
TIMEDB_ENTRY( a428f37f,  0,  17920, YM+TA+TB+TD), /* 03362 - Marcer - n-a.sndh */
TIMEDB_ENTRY( a42a477d,  0,  12288,          YM), /* 01459 - STU - BOMBERMAN.sndh */
TIMEDB_ENTRY( a434d439,  0,  21504,       YM+TD), /* 02612 - STU - RACE PART TWO.sndh */
TIMEDB_ENTRY( a4497ccf,  0,     12,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  1,     99,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  2,     57,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  3,     60,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  4,     36,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  5,     12,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  6,     84,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  7,    297,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  8,     36,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf,  9,      6,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 10,      6,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 11,    129,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 12,    135,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 13,     15,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 14,     48,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 15,     57,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 16,    201,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 17,     12,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 18,    117,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 19,      6,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 20,    117,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 21,    240,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 22,     33,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 23,     81,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 24,     30,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 25,    180,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 26,     45,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 27,     27,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 28,    186,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 29,      9,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 30,      9,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 31,      9,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 32,     21,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 33,    180,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 34,     18,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 35,     27,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 36,    138,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 37,    360,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 38,     21,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 39,     42,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 40,     42,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 41,     18,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 42,    114,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 43,    108,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 44,    144,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 45,    240,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 46,    234,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 47,     54,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 48,    105,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 49,     30,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 50,     84,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 51,    357,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 52,     18,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 53,    129,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 54,    117,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 55,     66,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 56,     27,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 57,     12,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 58,     45,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 59,     15,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 60,    663,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 61,    180,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a4497ccf, 62,     15,          YM), /* 04743 - Barry Murray - Manhunter II - San Francisco - A.sndh */
TIMEDB_ENTRY( a44cae7f,  0,  12985, YM+TA+TB+TD), /* 01601 - Nemo - Pirate Sunset.sndh */
TIMEDB_ENTRY( a4672f98,  0,   2304,       YM+TA), /* 00471 - Viking - North.sndh */
TIMEDB_ENTRY( a46c1730,  0,   2816,          YM), /* 01397 - Paul Shields - Tetraquest.sndh */
TIMEDB_ENTRY( a46c1730,  1,  25088,          YM), /* 01397 - Paul Shields - Tetraquest.sndh */
TIMEDB_ENTRY( a46c1730,  2,   6272,          YM), /* 01397 - Paul Shields - Tetraquest.sndh */
TIMEDB_ENTRY( a46c1730,  3,    150,          YM), /* 01397 - Paul Shields - Tetraquest.sndh */
TIMEDB_ENTRY( a46c1730,  4,    149,          YM), /* 01397 - Paul Shields - Tetraquest.sndh */
TIMEDB_ENTRY( a484459b,  0,  12110,    YM+TA+TD), /* 02685 - Donald Fakk - Spanking.sndh */
TIMEDB_ENTRY( a484c828,  0,  12289,          YM), /* 00386 - Wally Beben aka Hagar - Hyperdrome.sndh */
TIMEDB_ENTRY( a484c828,  1,    622,          YM), /* 00386 - Wally Beben aka Hagar - Hyperdrome.sndh */
TIMEDB_ENTRY( a4880e66,  0,   5435, YM+TA+TB+TD), /* 00984 - D-Force - Some How.sndh */
TIMEDB_ENTRY( a498b5bd,  0,   6906, YM+TA+TB+TD), /* 02756 - Dma-Sc - Sector One - Burton Tribute Project.sndh */
TIMEDB_ENTRY( a49eb8c8,  0,  90046,       YM+TA), /* 04038 - Roggie - Roggie 20.sndh */
TIMEDB_ENTRY( a4a0c69d,  0,   2969,          YM), /* 03055 - Ed Bogas - Latin.sndh */
TIMEDB_ENTRY( a4be8bf8,  0,   2630,          YM), /* 03171 - n-a - Theme From "Star Trek".sndh */
TIMEDB_ENTRY( a4c88cba,  0,   6144,          YM), /* 01543 - Ralph - Ralph 6.sndh */
TIMEDB_ENTRY( a4cfbfa7,  0,   2041,          YM), /* 03691 - Dma-Sc - Saboteur 3 - The Mummy Rises.sndh */
TIMEDB_ENTRY( a4d22f03,  0,  14592, YM+TA+TB+TD), /* 01105 - Timbral - Before The Raw.sndh */
TIMEDB_ENTRY( a4e339a3,  0,   2970,          YM), /* 04597 - Rude Kid - Fergus.sndh */
TIMEDB_ENTRY( a4f1a7f7,  0,   4986, YM+TA+TB+TD), /* 04254 - 505-Checkpoint - dudelich.sndh */
TIMEDB_ENTRY( a509c3dc,  0,    512,          YM), /* 04545 - Magnus Adolphzon - Groggy balldric.sndh */
TIMEDB_ENTRY( a509c3dc,  1,    320,          YM), /* 04545 - Magnus Adolphzon - Groggy balldric.sndh */
TIMEDB_ENTRY( a50f1db7,  0,   3269, YM+TA+TB+TD), /* 04300 - 505-checkpoint - Reaching.sndh */
TIMEDB_ENTRY( a512f2d6,  0,   3450,          YM), /* 03077 - Tom Marvin - THUNDERER          J.P.Sousa.sndh */
TIMEDB_ENTRY( a51df7b2,  0,  11520,          YM), /* 00337 - Paul Tonge - Bloody Money.sndh */
TIMEDB_ENTRY( a529bfa1,  0,   6844,          YM), /* 03104 - n-a - Children's Corner # 6.sndh */
TIMEDB_ENTRY( a5407018,  0,   7691,          YM), /* 02125 - Mad Max - Cuddly Demos - Reset.sndh */
TIMEDB_ENTRY( a5596fe4,  0,  10746,    YM+TA+TD), /* 01369 - Scavenger - Wicked Polygons.sndh */
TIMEDB_ENTRY( a55d69b8,  0,  25977,          YM), /* 01681 - Jess - Cubase VS Notator.sndh */
TIMEDB_ENTRY( a55fe2d6,  0,   5236,          YM), /* 03109 - n-a - The Imperial Death March.sndh */
TIMEDB_ENTRY( a562de36,  0,  15344,       YM+TA), /* 04128 - Tao - Wide Awake - STMYGM 2 version.sndh */
TIMEDB_ENTRY( a573e77b,  0,  23809,    YM+TA+TD), /* 02293 - Tao - Cream - Alive (Birth, part 2 - Alive 1).sndh */
TIMEDB_ENTRY( a5762364,  0,   1152,          YM), /* 02390 - Nexus 6 - Please Wait While Loading 1.sndh */
TIMEDB_ENTRY( a57c8c98,  0,   2664,          YM), /* 04147 - Unknown - Bumerang.sndh */
TIMEDB_ENTRY( a57c8c98,  1,    109,          YM), /* 04147 - Unknown - Bumerang.sndh */
TIMEDB_ENTRY( a583dc63,  0,  37155,          YM), /* 01317 - Matt Furniss - Shadow Worlds.sndh */
TIMEDB_ENTRY( a59030f9,  0,    640,          YM), /* 02147 - Mad Max - Mad Max Tune 2 - LSD.sndh */
TIMEDB_ENTRY( a5914f29,  0,  25600, YM+TA+TB+TD), /* 03985 - XyNo - DARKROOMS.sndh */
TIMEDB_ENTRY( a5a557ae,  0,   1247,    YM+TA+TD), /* 02761 - H.Tanaka [ST by Dma-Sc] - Super Mario Land.sndh */
TIMEDB_ENTRY( a5c011df,  0,    620,          YM), /* 03973 - Geoff Camp - Maths X 2.sndh */
TIMEDB_ENTRY( a5c011df,  1,    553,          YM), /* 03973 - Geoff Camp - Maths X 2.sndh */
TIMEDB_ENTRY( a5c011df,  2,    207,          YM), /* 03973 - Geoff Camp - Maths X 2.sndh */
TIMEDB_ENTRY( a5c011df,  3,   2708,          YM), /* 03973 - Geoff Camp - Maths X 2.sndh */
TIMEDB_ENTRY( a5c011df,  4,   1560,          YM), /* 03973 - Geoff Camp - Maths X 2.sndh */
TIMEDB_ENTRY( a5cb8483,  0,   6048, YM+TA+TB+TD), /* 02087 - Mad Max - Locomotion - Sid.sndh */
TIMEDB_ENTRY( a5e54bdd,  0,    959,          YM), /* 01525 - Iso - Gekkuhwerruk.sndh */
TIMEDB_ENTRY( a5ea5ae4,  0,  21492,       YM+TA), /* 01700 - Jess - Middle-Earth Theme.sndh */
TIMEDB_ENTRY( a612e913,  0,   9246,          YM), /* 03062 - J Green & Ed Bogas - Walker.sndh */
TIMEDB_ENTRY( a63e3855,  0,   1920,          YM), /* 02898 - Remalon - Requiem.sndh */
TIMEDB_ENTRY( a64425df,  0,   1275,      YM+STE), /* 01908 - Marcer - New Marcer - STE only.sndh */
TIMEDB_ENTRY( a65395a6,  0,  10038,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  1,   8062,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  2,   8475,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  3,  10670,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  4,   2710,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  5,   1617,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  6,   3049,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  7,  10716,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  8,   4071,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6,  9,   6954,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 10,   5711,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 11,   7571,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 12,   2574,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 13,   2758,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 14,   8363,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 15,   3324,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 16,   2898,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 17,   4443,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a65395a6, 18,   6083,          YM), /* 00850 - Ben Daglish - Federation of Free Traders.sndh */
TIMEDB_ENTRY( a657eca3,  0,    635,    YM+TA+TB), /* 04269 - 505-cpt^pdx - Jingle02.sndh */
TIMEDB_ENTRY( a65934b0,  0,   9216,       YM+TA), /* 01634 - Nexus - Dreamzone - Filled 3D.sndh */
TIMEDB_ENTRY( a65d7cfa,  0,   3284,          YM), /* 04069 - Gary Wheaton - Dream 10.sndh */
TIMEDB_ENTRY( a66242ca,  0,   2905,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  1,   5120,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  2,   4000,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  3,   2327,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  4,   2176,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  5,   2953,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  6,   3741,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  7,   1505,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66242ca,  8,    406,          YM), /* 03205 - Darren Ithell - Flipit & Magnose.sndh */
TIMEDB_ENTRY( a66fecfd,  0,   2594,          YM), /* 04705 - Daryl Kiley - Poker Dice.sndh */
TIMEDB_ENTRY( a66fecfd,  1,    296,          YM), /* 04705 - Daryl Kiley - Poker Dice.sndh */
TIMEDB_ENTRY( a66fecfd,  2,   2561,          YM), /* 04705 - Daryl Kiley - Poker Dice.sndh */
TIMEDB_ENTRY( a671058b,  0,   2685, YM+TA+TB+TD), /* 04314 - 505-checkpoint - Supadupa.sndh */
TIMEDB_ENTRY( a6aa9d05,  0,  12778,          YM), /* 00860 - Ben Daglish - Masters of the Universe.sndh */
TIMEDB_ENTRY( a6aa9d05,  1,   2244,          YM), /* 00860 - Ben Daglish - Masters of the Universe.sndh */
TIMEDB_ENTRY( a6aa9d05,  2,    369,          YM), /* 00860 - Ben Daglish - Masters of the Universe.sndh */
TIMEDB_ENTRY( a6be89ed,  0,   8064,          YM), /* 04195 - Rhino - Compofiller 2013.sndh */
TIMEDB_ENTRY( a6dbffc0,  0,   1942,          YM), /* 02215 - Mad Max - Spaceball.sndh */
TIMEDB_ENTRY( a6dbffc0,  1,   1537,          YM), /* 02215 - Mad Max - Spaceball.sndh */
TIMEDB_ENTRY( a6dbffc0,  2,    769,          YM), /* 02215 - Mad Max - Spaceball.sndh */
TIMEDB_ENTRY( a6dbffc0,  3,    140,          YM), /* 02215 - Mad Max - Spaceball.sndh */
TIMEDB_ENTRY( a6dbffc0,  4,   5690,          YM), /* 02215 - Mad Max - Spaceball.sndh */
TIMEDB_ENTRY( a6e104c4,  0,   4632,    YM+TB+TD), /* 02858 - Helene - STRAYKITTENS1.sndh */
TIMEDB_ENTRY( a6ee5e2e,  0,   5447,          YM), /* 02715 - Chrispy Noodle - Tiger 1.sndh */
TIMEDB_ENTRY( a6f1c155,  0,  10752, YM+TA+TB+TD), /* 03304 - Marcer - COCO-COOOPER.sndh */
TIMEDB_ENTRY( a6f824c6,  0,   5284,          YM), /* 03129 - n-a - La Virgen de la Macarena.sndh */
TIMEDB_ENTRY( a7151188,  0,  25080,          YM), /* 01215 - Energizer - Floating - Chip.sndh */
TIMEDB_ENTRY( a7188afa,  0,   9984,          YM), /* 01821 - Lord - Ambition Menu.sndh */
TIMEDB_ENTRY( a735aab7,  0,   2304,          YM), /* 00837 - Crawdaddy - Ninja.sndh */
TIMEDB_ENTRY( a7621359,  0,   1792,          YM), /* 01196 - Excellence In Art - Oxygene 4.sndh */
TIMEDB_ENTRY( a78239d5,  0,   5280,          YM), /* 00590 - Celtic - Unknown Trial.sndh */
TIMEDB_ENTRY( a7986dac,  0,   6144,    YM+TA+TD), /* 03595 - 505-cpt^pdx - Triplex.sndh */
TIMEDB_ENTRY( a7c6aad7,  0,  37888,          YM), /* 02842 - Marcer - XMASSNOWMAN.sndh */
TIMEDB_ENTRY( a7e7b254,  0,   2560,          YM), /* 00334 - Gary Wheaton - Wilbu.sndh */
TIMEDB_ENTRY( a8029a43,  0,   7812,       YM+TA), /* 04438 - Lotus - Synth Dream 2 - Tidal Waves.sndh */
TIMEDB_ENTRY( a807f6bd,  0,  12800,          YM), /* 00144 - David Whittaker - Stormtrooper.sndh */
TIMEDB_ENTRY( a807f6bd,  1,   5376,          YM), /* 00144 - David Whittaker - Stormtrooper.sndh */
TIMEDB_ENTRY( a807f6bd,  2,     66,          YM), /* 00144 - David Whittaker - Stormtrooper.sndh */
TIMEDB_ENTRY( a818b3bd,  0,   4384,          YM), /* 03139 - n-a - Minuet in G by J.S. Bach.sndh */
TIMEDB_ENTRY( a819a501,  0,   4606, YM+TA+TB+TD), /* 04285 - 505 - Checkpoint - Lovesong (Unrel.).sndh */
TIMEDB_ENTRY( a81da630,  0,   2730,          YM), /* 00118 - David Whittaker - Mr Heli.sndh */
TIMEDB_ENTRY( a81da630,  1,    560,          YM), /* 00118 - David Whittaker - Mr Heli.sndh */
TIMEDB_ENTRY( a81da630,  2,    247,          YM), /* 00118 - David Whittaker - Mr Heli.sndh */
TIMEDB_ENTRY( a81da630,  3,    114,          YM), /* 00118 - David Whittaker - Mr Heli.sndh */
TIMEDB_ENTRY( a81da630,  4,    350,          YM), /* 00118 - David Whittaker - Mr Heli.sndh */
TIMEDB_ENTRY( a8226058,  0,   8448,          YM), /* 01973 - MSG - MSG 03.sndh */
TIMEDB_ENTRY( a82290ab,  0,   3583,       YM+TB), /* 01857 - Lotek Style - Lip Gloss.sndh */
TIMEDB_ENTRY( a84fdaaa,  0,   3456,          YM), /* 00976 - Chris Denman - Cool World.sndh */
TIMEDB_ENTRY( a84fdaaa,  1,   5758,          YM), /* 00976 - Chris Denman - Cool World.sndh */
TIMEDB_ENTRY( a850f552,  0,   7035,    YM+TB+TD), /* 01587 - Nemo - Bambooka.sndh */
TIMEDB_ENTRY( a852fdee,  0,   1535,          YM), /* 04730 - Ralph - Coming Back After 15 Years.sndh */
TIMEDB_ENTRY( a85b48d4,  0,  27884,          YM), /* 01471 - STU - NERVOUS.sndh */
TIMEDB_ENTRY( a85da635,  0,   3336,    YM+TB+TD), /* 02712 - Marcer - SIMPLY LARRY.sndh */
TIMEDB_ENTRY( a89eb139,  0,  12282, YM+TA+TB+TD), /* 02828 - Msg - Superfly tune 5.sndh */
TIMEDB_ENTRY( a8a808e3,  0,   9984,          YM), /* 03219 - Tomchi - Parallax.sndh */
TIMEDB_ENTRY( a8e71704,  0,   3092,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  1,     47,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  2,    187,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  3,    227,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  5,   1593,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  6,   1541,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  7,    921,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  8,    521,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704,  9,   1082,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e71704, 10,   1602,          YM), /* 00870 - Ben Daglish - Rick Dangerous 2.sndh */
TIMEDB_ENTRY( a8e800cb,  0,   7680,          YM), /* 01124 - Tyan - Bad Boys.sndh */
TIMEDB_ENTRY( a8ed8659,  0,   9862,          YM), /* 00726 - Tebirod - No Cooper Plasma.sndh */
TIMEDB_ENTRY( a9059a8e,  0,  39936, YM+TA+TB+TD), /* 03337 - Marcer - OXYGEN8 YM.sndh */
TIMEDB_ENTRY( a90b38fe,  0, 106994,       YM+TA), /* 03759 - Tufty - Rev1.sndh */
TIMEDB_ENTRY( a93fb805,  0,   1600,          YM), /* 04532 - Mike Davis - Forest Fire.sndh */
TIMEDB_ENTRY( a9558038,  0,  41320, YM+TA+TB+TD), /* 02580 - STU - BOMBTRO.sndh */
TIMEDB_ENTRY( a96cf208,  0,   7290,          YM), /* 00970 - Demon X - Dreamzone - vectorball.sndh */
TIMEDB_ENTRY( a981ef77,  0, 106431,       YM+TA), /* 03733 - David Whittaker - Shadow of the Beast.sndh */
TIMEDB_ENTRY( a9840451,  0,  17828,         STE), /* 03448 - 505 - Paradox - Cernit Trandafir.sndh */
TIMEDB_ENTRY( a98f4a8a,  0,   1047,       YM+TD), /* 02665 - Donald Fakk - Countdown.sndh */
TIMEDB_ENTRY( a9a4e1e2,  0,   2560,          YM), /* 04604 - White Ninja - Unchanged.sndh */
TIMEDB_ENTRY( a9a63e99,  0,  10486,          YM), /* 03241 - T1000 - The Beast Club - TBC 13.sndh */
TIMEDB_ENTRY( a9a79d0b,  0,   7199, YM+TA+TB+TD), /* 03832 - Rhino - Torment (D Newman) - Three out of Four.sndh */
TIMEDB_ENTRY( a9bf9dca,  0,   6144,          YM), /* 03254 - T1000 - The Bits Club 35.sndh */
TIMEDB_ENTRY( a9df42c6,  0,   8736,          YM), /* 01392 - Shaft - Ninja Spirit.sndh */
TIMEDB_ENTRY( aa09bd9d,  0,   3680,          YM), /* 03261 - Justin Ward - Deadland.sndh */
TIMEDB_ENTRY( aa0c2a3a,  0,    904,          YM), /* 03383 - Unknown - Ballade.sndh */
TIMEDB_ENTRY( aa1bc80a,  0,   4096,          YM), /* 02934 - Chris Denman - Discovery.sndh */
TIMEDB_ENTRY( aa390706,  0,  10944, YM+TA+TB+TD), /* 02068 - Mad Max - Chimera - Sid.sndh */
TIMEDB_ENTRY( aa390706,  1,   6144, YM+TA+TB+TD), /* 02068 - Mad Max - Chimera - Sid.sndh */
TIMEDB_ENTRY( aa417285,  0,   2891,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  1,    828,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  2,   3840,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  3,   3648,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  4,    648,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  5,   3840,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  6,    832,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa417285,  7,   5387,          YM), /* 00095 - David Whittaker - Defender of the Crown.sndh */
TIMEDB_ENTRY( aa45734b,  0,   4992,          YM), /* 02415 - Nexus 6 -  fta ma wat neues.sndh */
TIMEDB_ENTRY( aa592647,  0,   2681,          YM), /* 04092 - Gary Wheaton - The Earth dies Screaming.sndh */
TIMEDB_ENTRY( aa6340a9,  0,  11712,          YM), /* 01969 - MSG - MSG 12 - Maggie 24 Intro.sndh */
TIMEDB_ENTRY( aa703432,  0,   4029, YM+TA+TB+TD), /* 00276 - Baggio - Demo Test.sndh */
TIMEDB_ENTRY( aa79026a,  0,   1280,          YM), /* 03544 - Dean Evans - One Step Beyond.sndh */
TIMEDB_ENTRY( aa90541a,  0,   4351,    YM+TA+TD), /* 02564 - marcer - lotus 1 title remix.sndh */
TIMEDB_ENTRY( aa9b1cb0,  0,   3839, YM+TA+TB+TD), /* 03998 - XyNo - YS1 MINEA TOWN ST COVER.sndh */
TIMEDB_ENTRY( aaa45375,  0,   6162,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  1,    245,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  2,    297,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  3,    331,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  4,    264,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  5,    265,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  6,    223,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  7,    200,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  8,    235,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375,  9,    500,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 10,    122,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 11,     56,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 12,    164,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 13,    195,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 14,    206,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 15,    203,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 16,    253,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 17,   1502,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaa45375, 18,    293,          YM), /* 01291 - Matt Furniss - Botics.sndh */
TIMEDB_ENTRY( aaacd8b9,  0,  50887,       YM+TA), /* 04361 - X The Beermat - Now.sndh */
TIMEDB_ENTRY( aab58399,  0,   6912,          YM), /* 02734 - Acid - The Blue Boys - Font Menu 3.sndh */
TIMEDB_ENTRY( aabb428f,  0,   3839, YM+TA+TB+TD), /* 03420 - Dubmood - Jesters Chipmunks.sndh */
TIMEDB_ENTRY( aabcb0a8,  0,   1728,    YM+TA+TD), /* 04267 - 505-checkpoint^paradox - idiot.sndh */
TIMEDB_ENTRY( aaeb85e3,  0,  33024,    YM+TA+TD), /* 03049 - Marcer - Elite - Traxxer 2.sndh */
TIMEDB_ENTRY( aaf096c2,  0,     99,      YM+STE), /* 03231 - Mad Max - Chambers of Shaolin - STe.sndh */
TIMEDB_ENTRY( ab084016,  0,    398,          YM), /* 03013 - Unknown - Heroes of the Lance.sndh */
TIMEDB_ENTRY( ab1c924b,  0,   1558,          YM), /* 03117 - n-a - .sndh */
TIMEDB_ENTRY( ab25346b,  0,  23348,    YM+TB+TD), /* 03340 - Marcer - TURRICANA.sndh */
TIMEDB_ENTRY( ab2725b6,  0,   4607,          YM), /* 01305 - Matt Furniss - Manchester United.sndh */
TIMEDB_ENTRY( ab3cf2f3,  0,   9449,          YM), /* 03133 - n-a - Maggie May.sndh */
TIMEDB_ENTRY( ab412ce5,  0,   1600,          YM), /* 04769 - blind io - borderline bacon.sndh */
TIMEDB_ENTRY( ab452992,  0,   5765, YM+TA+TB+TD), /* 01153 - Tyan - Platz 2.sndh */
TIMEDB_ENTRY( ab471e96,  0,   1792,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab471e96,  1,   1632,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab471e96,  2,   3072,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab471e96,  3,   4320,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab471e96,  4,    240,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab471e96,  5,     73,          YM), /* 01893 - Dave Lowe - Time Scanner.sndh */
TIMEDB_ENTRY( ab47a128,  0,  10481, YM+TA+TB+TD), /* 02495 - gwEm - maxYMise.sndh */
TIMEDB_ENTRY( ab4b55ca,  0,  24073,       YM+TA), /* 04674 - Unknown (orig Karsten Obarksi) - End Theme.sndh */
TIMEDB_ENTRY( ab4fde3b,  0,   3068,       YM+TA), /* 01531 - Iso - Particoloured.sndh */
TIMEDB_ENTRY( ab59c75b,  0,   7680,    YM+TB+TD), /* 03460 - Marcer - Gamet.sndh */
TIMEDB_ENTRY( ab9f613f,  0,   9216,    YM+TA+TD), /* 02348 - Tao - Cream - High Times (UCM 24).sndh */
TIMEDB_ENTRY( abaca928,  0,   7744, YM+TA+TB+TD), /* 02873 - gwEm and Dma Sc - Chipping for Ca$h.sndh */
TIMEDB_ENTRY( abb5b4e0,  0,  89681,       YM+TA), /* 03747 - Tufty - I Feel Love - Donna Summer.sndh */
TIMEDB_ENTRY( abb78bc7,  0,    412,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  1,    640,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  2,    189,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  3,   3264,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  4,   3264,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  5,    991,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  6,   3264,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abb78bc7,  7,    978,          YM), /* 01116 - Keith Tinman - The Untouchables.sndh */
TIMEDB_ENTRY( abc94b3f,  0,  33024, YM+TA+TB+TD), /* 02906 - helene - NEWEDITT.sndh */
TIMEDB_ENTRY( abceba69,  0,  19039,       YM+TA), /* 04869 - Jovis - Speedrit.sndh */
TIMEDB_ENTRY( abd81942,  0,   5382,    YM+TA+TB), /* 00822 - !Cube - Admirable's Mod Compil #32.sndh */
TIMEDB_ENTRY( abf5cf75,  0,  15396,    YM+TB+TD), /* 03983 - XyNo - Cheeselame.sndh */
TIMEDB_ENTRY( ac0bf2f8,  0,    378,          YM), /* 01905 - Marcer - Ho Ho Ho.sndh */
TIMEDB_ENTRY( ac3ab56c,  0,   6390,          YM), /* 00364 - Steve Barrett - Kelly X.sndh */
TIMEDB_ENTRY( ac51d303,  0,   9599, YM+TA+TB+TD), /* 02862 - Evil - Enola Gay.sndh */
TIMEDB_ENTRY( ac542b66,  0,   3136,          YM), /* 01251 - Floopy - Rapide 01.sndh */
TIMEDB_ENTRY( ac5f527d,  0,   9915,          YM), /* 00322 - Gary Wheaton - Not quite a new car.sndh */
TIMEDB_ENTRY( ac60b1f1,  0,   5762,          YM), /* 01774 - Lap - Lap 14.sndh */
TIMEDB_ENTRY( ac68d080,  0,  12000,          YM), /* 01426 - Spector - Manic Demo - Main Menu.sndh */
TIMEDB_ENTRY( ac8da535,  0,   3433,          YM), /* 03047 - Ed Bogas - Five Four.sndh */
TIMEDB_ENTRY( ac9c9eda,  0,   1380,          YM), /* 03388 - Al Lowe - Kings Quest 2.sndh */
TIMEDB_ENTRY( acac9885,  0,   9494,          YM), /* 03094 - n-a - They Call It The Blues.sndh */
TIMEDB_ENTRY( acb77857,  0,   5934,          YM), /* 00567 - Jas C Brooke - Hellfire.sndh */
TIMEDB_ENTRY( acb77857,  1,    512,          YM), /* 00567 - Jas C Brooke - Hellfire.sndh */
TIMEDB_ENTRY( acb77857,  2,    310,          YM), /* 00567 - Jas C Brooke - Hellfire.sndh */
TIMEDB_ENTRY( acbaca50,  0,  18636,          YM), /* 01928 - The Mind - Malleus Mallificorum - ST News 6.1.sndh */
TIMEDB_ENTRY( acc170ef,  0,   9471, YM+TA+TB+TD), /* 04378 - dhor-mec - Crash test.sndh */
TIMEDB_ENTRY( acd8a58e,  0,   7098,          YM), /* 01241 - Epic - 'Vixit' In His Bath.sndh */
TIMEDB_ENTRY( ace1a20f,  0,  21882,          YM), /* 01705 - Jess - The Popple's Revenge.sndh */
TIMEDB_ENTRY( acebab64,  0,   5120,          YM), /* 02138 - Mad Max - Monty On The Run - LCD.sndh */
TIMEDB_ENTRY( acf02a6f,  0,   7118,          YM), /* 01805 - Barry Leitch - Silkworm.sndh */
TIMEDB_ENTRY( ad149c96,  0,   4992,       YM+TA), /* 02314 - Tao - Cream (orig. Klaus Ude) - Lin Wu's Challenge 1.sndh */
TIMEDB_ENTRY( ad31113e,  0,    360,       YM+TA), /* 01716 - Jess - The Music Bootsector (remix).sndh */
TIMEDB_ENTRY( ad55b3f3,  0,  14512,    YM+TA+TD), /* 02680 - Donald Fakk - Earth is Rotten.sndh */
TIMEDB_ENTRY( ad56e774,  0,  25296, YM+TA+TB+TD), /* 02908 - Helene - sndh4Christian.sndh */
TIMEDB_ENTRY( ad5bbfb5,  0,  15359,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  1,  14669,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  2,  18506,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  3,  16516,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  4,  16225,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  5,  10355,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  6,  12799,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  7,  13395,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  8,  27613,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5,  9,  10127,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5, 10,  18567,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad5bbfb5, 11,  18019,          YM), /* 03536 - Jean Baudlot - Cruise For A Corpse.sndh */
TIMEDB_ENTRY( ad6f982d,  0,   5120,    YM+TA+TD), /* 03447 - 505 - Checkpoint - Humming.sndh */
TIMEDB_ENTRY( ada9d399,  0,   6395,    YM+TA+TD), /* 02422 - 505-cpt^dox - gloomy.sndh */
TIMEDB_ENTRY( adc8f88c,  0,   2304,          YM), /* 00120 - David Whittaker - Nitro Boost Challenge.sndh */
TIMEDB_ENTRY( adc8f88c,  1,    129,          YM), /* 00120 - David Whittaker - Nitro Boost Challenge.sndh */
TIMEDB_ENTRY( adc8f88c,  2,    129,          YM), /* 00120 - David Whittaker - Nitro Boost Challenge.sndh */
TIMEDB_ENTRY( adccd12a,  0,  13446,    YM+TA+TD), /* 02343 - Tao - Cream - Spontex (UCM 17).sndh */
TIMEDB_ENTRY( add5b0f9,  0,   6912,          YM), /* 00161 - Martin Walker - 9 Lives.sndh */
TIMEDB_ENTRY( ade32b57,  0,   2237,          YM), /* 01529 - Iso - Instinctively.sndh */
TIMEDB_ENTRY( adfc67fe,  0,  38400, YM+TA+TB+TD), /* 03897 - Pinokio - Bzdet.sndh */
TIMEDB_ENTRY( ae0cc76a,  0,   9920,          YM), /* 01185 - Excellence In Art - Hit Me.sndh */
TIMEDB_ENTRY( ae3726d1,  0, 103425,       YM+TA), /* 03963 - Scott Lemmon - Thrash 5.sndh */
TIMEDB_ENTRY( ae3bdef7,  0,   9471,          YM), /* 01759 - Lap - Phaleon - Dungeon.sndh */
TIMEDB_ENTRY( ae505cee,  0,   3840,          YM), /* 02112 - Mad Max - Noisy Pillars.sndh */
TIMEDB_ENTRY( ae668413,  0,   3072,          YM), /* 01085 - Techno - Dentro 1.sndh */
TIMEDB_ENTRY( ae732abc,  0,  19008,          YM), /* 02212 - Mad Max - Rings Of Medusa.sndh */
TIMEDB_ENTRY( ae732abc,  1,   4800,          YM), /* 02212 - Mad Max - Rings Of Medusa.sndh */
TIMEDB_ENTRY( ae7b8d79,  0,   2688,          YM), /* 01169 - Sally - Stax 53 - Tune 1.sndh */
TIMEDB_ENTRY( ae8e4d6b,  0,   3584,          YM), /* 00563 - Jas C Brooke - 1943.sndh */
TIMEDB_ENTRY( aef26aa2,  0,   4609, YM+TA+TB+TD), /* 03643 - 505-checkpoint - Stress.sndh */
TIMEDB_ENTRY( af5dfc0d,  0,  12192, YM+TA+TB+TD), /* 02073 - Mad Max - Deep Strike - Sid.sndh */
TIMEDB_ENTRY( af682d9c,  0,    431,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  1,   2155,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  2,    516,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  3,   1325,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  4,    481,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  5,    342,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  6,    248,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af682d9c,  7,     92,          YM), /* 00067 - David Whittaker - APB.sndh */
TIMEDB_ENTRY( af74630c,  0,   3584,    YM+TA+TD), /* 02764 - Bobby Prince [ST by Dma-Sc] - Get Them....sndh */
TIMEDB_ENTRY( afa96128,  0,  14472, YM+TA+TB+TD), /* 02500 - gwEm - YM303 Ruffneck.sndh */
TIMEDB_ENTRY( afad8ead,  0,   9600,       YM+TA), /* 02409 - Kanewood - Phantom Demo - Bonus A.sndh */
TIMEDB_ENTRY( afb4a50b,  0,  13440,       YM+TA), /* 04124 - Tao - Steps - STMYGM 2 version.sndh */
TIMEDB_ENTRY( afb80fb7,  0,   9299,          YM), /* 04645 - V0yager-Lamers - Cartridge Generation.sndh */
TIMEDB_ENTRY( afcf21dc,  0,   8315, YM+TA+TB+TD), /* 01841 - Lotek Style - Destroya.sndh */
TIMEDB_ENTRY( afd23fc7,  0,   5754, YM+TA+TB+TD), /* 02622 - STU - TWENTY-FIVE.sndh */
TIMEDB_ENTRY( afd55952,  0,   4214,          YM), /* 03069 - Chris French - Silent Night.sndh */
TIMEDB_ENTRY( aff68460,  0,  16384,       YM+TA), /* 04412 - Lotus - Wall PD.sndh */
TIMEDB_ENTRY( affb8647,  0,  11648,    YM+TA+TD), /* 02263 - Tao - Cream - Rise (Steps).sndh */
TIMEDB_ENTRY( b002fd49,  0, 236193,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  1,  90818,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  2, 132258,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  3, 459791,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  4, 174754,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  5, 321223,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  6,  22748,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  7,  31778,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  8,  47300,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49,  9,  11525,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b002fd49, 10, 163326,       YM+TA), /* 03788 - The Phantom - Insignificant Demo - Quartet.sndh */
TIMEDB_ENTRY( b014305f,  0,  58354,       YM+TA), /* 04559 - Simon O - Gremlin Music Demo.sndh */
TIMEDB_ENTRY( b01a35e5,  0,   6405,          YM), /* 00227 - ACC - City Road.sndh */
TIMEDB_ENTRY( b0461eba,  0,  10799,          YM), /* 03858 - !Cube - Renaissance.sndh */
TIMEDB_ENTRY( b04f5361,  0,  64727,          YM), /* 01811 - Barry Leitch - Utopia.sndh */
TIMEDB_ENTRY( b04f5361,  1,  63748,          YM), /* 01811 - Barry Leitch - Utopia.sndh */
TIMEDB_ENTRY( b04f5361,  2,  76035,          YM), /* 01811 - Barry Leitch - Utopia.sndh */
TIMEDB_ENTRY( b04f5361,  3,  50691,          YM), /* 01811 - Barry Leitch - Utopia.sndh */
TIMEDB_ENTRY( b052d02c,  0,   6144,       YM+TD), /* 00583 - Celtic - Audial Overdrive.sndh */
TIMEDB_ENTRY( b054596f,  0,  15103,          YM), /* 03969 - DMA-SC - Circus Backstage.sndh */
TIMEDB_ENTRY( b056712f,  0,   3456,       YM+TA), /* 02892 - Kanewood - Memorial Demo.sndh */
TIMEDB_ENTRY( b0610733,  0,  27648,       YM+TD), /* 02600 - STU - THE INVASION.sndh */
TIMEDB_ENTRY( b06a78e6,  0,   5376,          YM), /* 00795 - Crawdaddy - Kruz Theme.sndh */
TIMEDB_ENTRY( b086370c,  0,   9812,       YM+TA), /* 03232 - Mad Max - Chambers of Shaolin - FM.sndh */
TIMEDB_ENTRY( b0ac4fe4,  0,   1344,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  1,   1536,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  2,    512,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  3,    384,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  4,    384,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  5,    384,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  6,    256,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  7,   1152,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  8,   2304,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4,  9,   2400,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4, 10,    193,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0ac4fe4, 12,     81,          YM), /* 00748 - Mark Cooksey - Mighty Bombjack.sndh */
TIMEDB_ENTRY( b0c2fb90,  0,   6736, YM+TA+TB+TD), /* 03371 - STU - ICE CREAM DREAM.sndh */
TIMEDB_ENTRY( b0df1620,  0,   4600,       YM+TA), /* 04763 - Ferox - State of Mind - Tuesday.sndh */
TIMEDB_ENTRY( b10c4711,  0,   1729,       YM+TB), /* 03473 - gwEm - Real Friends.sndh */
TIMEDB_ENTRY( b10c7de8,  0,   1342,          YM), /* 02462 - Iso - Buzzzzzzzing.sndh */
TIMEDB_ENTRY( b14cd84f,  0,   3840,    YM+TA+TD), /* 02323 - Tao - Cream (orig. Klaus Ude) - Magic Marble ingame 2.sndh */
TIMEDB_ENTRY( b14f0777,  0,   9216,          YM), /* 02163 - Mad Max - Syntax Terror - TEX.sndh */
TIMEDB_ENTRY( b158b1e4,  0,  11454,       YM+TA), /* 01524 - Iso - Fun!!!.sndh */
TIMEDB_ENTRY( b159c8f7,  0,  13439,         STE), /* 03381 - Gargamel - DHS - Tyranny.sndh */
TIMEDB_ENTRY( b15ea782,  0,  87922,       YM+TA), /* 03758 - Tufty - Paradise - Sade.sndh */
TIMEDB_ENTRY( b16a8bc3,  0,  39936,          YM), /* 01454 - Stax - Wild.sndh */
TIMEDB_ENTRY( b17f1c81,  0,  10560,       YM+TD), /* 01985 - Jochen Hippel - "2".sndh */
TIMEDB_ENTRY( b1819cf9,  0,   1292,          YM), /* 00010 - 4-Mat - Beastbusters.sndh */
TIMEDB_ENTRY( b1819cf9,  1,    378,          YM), /* 00010 - 4-Mat - Beastbusters.sndh */
TIMEDB_ENTRY( b1819cf9,  2,    481,          YM), /* 00010 - 4-Mat - Beastbusters.sndh */
TIMEDB_ENTRY( b1819cf9,  3,   1292,          YM), /* 00010 - 4-Mat - Beastbusters.sndh */
TIMEDB_ENTRY( b188dadd,  0,  71343,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b188dadd,  1,  33810,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b188dadd,  2,  95175,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b188dadd,  3,  33281,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b188dadd,  4, 153482,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b188dadd,  5,  31630,       YM+TA), /* 03823 - Spaz - Mindbomb - Digi Synth.sndh */
TIMEDB_ENTRY( b192764f,  0,   5672,          YM), /* 02238 - Tao - Cream (orig. Rob Hubbard) - Flash Gordon (Stmygm).sndh */
TIMEDB_ENTRY( b198ba4a,  0, 184885,       YM+TA), /* 03786 - Nexus 6 - Rising 2.sndh */
TIMEDB_ENTRY( b1acc0eb,  0,  13056,          YM), /* 01643 - Jedi - Digital.sndh */
TIMEDB_ENTRY( b1b25226,  0,   1824,      YM+STE), /* 01212 - Elof - Farvalse - STE only.sndh */
TIMEDB_ENTRY( b1b755b4,  0,  16320,       YM+TA), /* 04199 - Furax - Equinox & Hemoroids - Virtual Escape Theme (enhanced version).sndh */
TIMEDB_ENTRY( b1c1bfaf,  0,   7675,    YM+TB+TD), /* 01871 - Lotek Style - Space Pirate.sndh */
TIMEDB_ENTRY( b1f6ff32,  0,  48603,       YM+TA), /* 03741 - Tufty - Aces High (orig Romeo Knight).sndh */
TIMEDB_ENTRY( b202b6c6,  0,   2881,          YM), /* 02142 - Mad Max - Commando High - LSD.sndh */
TIMEDB_ENTRY( b22e35a0,  0,  83269,       YM+TA), /* 04053 - Roggie - Ooooer my digits fell off.sndh */
TIMEDB_ENTRY( b24bc173,  0,  20352,          YM), /* 02107 - Mad Max - Alloy Run.sndh */
TIMEDB_ENTRY( b24bc173,  1,   2304,          YM), /* 02107 - Mad Max - Alloy Run.sndh */
TIMEDB_ENTRY( b2810651,  0,    504,          YM), /* 03680 - Dma-Sc - Antiques - Loader 1.sndh */
TIMEDB_ENTRY( b282048d,  0,   9856,          YM), /* 01161 - Tyan - STAX Hat keinen Plan.sndh */
TIMEDB_ENTRY( b28d3b2f,  0,   3465,          YM), /* 00321 - Gary Wheaton - Mystic.sndh */
TIMEDB_ENTRY( b29116e1,  0,   8832,          YM), /* 01220 - ENS Music - ENS 3.sndh */
TIMEDB_ENTRY( b2a8d219,  0,  99085,       YM+TA), /* 04054 - Roggie - Give me the rice pudding.sndh */
TIMEDB_ENTRY( b2ac6c99,  0,   6656,          YM), /* 00730 - Kevin Collier - Advanced Tactical Fighter 2.sndh */
TIMEDB_ENTRY( b2b75a57,  0,  12096,          YM), /* 00119 - David Whittaker - Ninja Mission.sndh */
TIMEDB_ENTRY( b2c0ab4c,  0,  14777,    YM+TA+TD), /* 00232 - Acid - Blue Boys Stuff 11.sndh */
TIMEDB_ENTRY( b2c91b33,  0,   9216,          YM), /* 01488 - Holger Gehrmann - Terra Nova.sndh */
TIMEDB_ENTRY( b2cf58e8,  0,   2065,       YM+TA), /* 04662 - Dave Lowe - Carrier Command.sndh */
TIMEDB_ENTRY( b2d356bc,  0, 105073,          YM), /* 04527 - Jim Cuomo - Expressing.sndh */
TIMEDB_ENTRY( b2fd791e,  0,  39914,    YM+TA+TD), /* 02356 - Tao - Cream - Warryorz (YM Rockerz 4).sndh */
TIMEDB_ENTRY( b3355200,  0,  14976,          YM), /* 04476 - Popeye 2 - Paul Loraine.sndh */
TIMEDB_ENTRY( b3355200,  1,  11247,          YM), /* 04476 - Popeye 2 - Paul Loraine.sndh */
TIMEDB_ENTRY( b3355200,  2,  10464,          YM), /* 04476 - Popeye 2 - Paul Loraine.sndh */
TIMEDB_ENTRY( b3355200,  3,   6480,          YM), /* 04476 - Popeye 2 - Paul Loraine.sndh */
TIMEDB_ENTRY( b33cf165,  0,   6096,          YM), /* 00577 - Jas C Brooke - Thunderstrike.sndh */
TIMEDB_ENTRY( b33cf165,  1,    175,          YM), /* 00577 - Jas C Brooke - Thunderstrike.sndh */
TIMEDB_ENTRY( b33cf165,  2,     86,          YM), /* 00577 - Jas C Brooke - Thunderstrike.sndh */
TIMEDB_ENTRY( b33cf165,  3,     59,          YM), /* 00577 - Jas C Brooke - Thunderstrike.sndh */
TIMEDB_ENTRY( b33cf165,  4,    152,          YM), /* 00577 - Jas C Brooke - Thunderstrike.sndh */
TIMEDB_ENTRY( b34bb126,  0,   3456,    YM+TA+TD), /* 04270 - 505 - Kickdown.sndh */
TIMEDB_ENTRY( b34f88d5,  0, 139347,       YM+TA), /* 04625 - Michael Opel - ACIA.sndh */
TIMEDB_ENTRY( b35c0808,  0,  10793,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  1,    204,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  2,    190,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  3,    255,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  4,    143,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  5,    409,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  6,    140,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  7,    228,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35c0808,  8,    256,          YM), /* 00840 - Ben Daglish - Axel's Magic Hammer.sndh */
TIMEDB_ENTRY( b35f679a,  0,   6880,          YM), /* 02388 - Nexus 6 - Oh wie sch n.sndh */
TIMEDB_ENTRY( b3637b32,  0,   8508,          YM), /* 04344 - Nexus 6 - Johnny B. (The Hooters).sndh */
TIMEDB_ENTRY( b36ff00f,  0,   8512,          YM), /* 03699 - Chris Mad - Overdrive Demo - Times.sndh */
TIMEDB_ENTRY( b38062b5,  0,   5503,          YM), /* 04079 - Gary Wheaton - A8 Menu.sndh */
TIMEDB_ENTRY( b381de0c,  0,   6720,       YM+TA), /* 00254 - Antichrist - Enigma.sndh */
TIMEDB_ENTRY( b389ea27,  0,   4608,          YM), /* 00489 - Ultimatum - Kidney Bean 5.sndh */
TIMEDB_ENTRY( b39fb491,  0,   2304,       YM+TA), /* 00419 - Big Alec - Pursuit.sndh */
TIMEDB_ENTRY( b3a13281,  0,   3456,       YM+TA), /* 02288 - Tao - Cream - Just Loadin' (Steps).sndh */
TIMEDB_ENTRY( b3a80925,  0, 169102,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  1, 128798,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  2,  39899,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  3, 109194,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  4,  67271,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  5,  77764,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  6,  33016,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  7,  14337,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  8,  50512,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925,  9,  61060,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925, 10,  71457,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a80925, 11,  27649,       YM+TA), /* 03826 - BB - TEC Demo.sndh */
TIMEDB_ENTRY( b3a8b83c,  0,     60,          YM), /* 03917 - Unknown - Falsch1.sndh */
TIMEDB_ENTRY( b3b7342a,  0,  78546,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  1,  57411,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  2,  97083,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  3, 134953,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  4,  58272,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  5,  91705,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  6, 106551,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  7,  60886,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  8,  95858,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a,  9,  86124,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a, 10,  11332,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3b7342a, 11,  59454,       YM+TA), /* 03812 - Martin Walker - Thunderhawk AH73-M Levels.sndh */
TIMEDB_ENTRY( b3c5a7a9,  0,  15995, YM+TA+TB+TD), /* 01598 - Nemo - Matrix - UCM 19.sndh */
TIMEDB_ENTRY( b3cf4287,  0,     28,          YM), /* 04700 - Robin Edwards - Mr. Dig.sndh */
TIMEDB_ENTRY( b3cf4287,  1,     32,          YM), /* 04700 - Robin Edwards - Mr. Dig.sndh */
TIMEDB_ENTRY( b3cf4287,  2,     39,          YM), /* 04700 - Robin Edwards - Mr. Dig.sndh */
TIMEDB_ENTRY( b3cf4287,  3,     77,          YM), /* 04700 - Robin Edwards - Mr. Dig.sndh */
TIMEDB_ENTRY( b3cf4287,  4,     36,          YM), /* 04700 - Robin Edwards - Mr. Dig.sndh */
TIMEDB_ENTRY( b3fce5fa,  0,   5376, YM+TA+TB+TD), /* 03505 - 505 - Suretrip 2 - End.sndh */
TIMEDB_ENTRY( b405094f,  0,   5376,          YM), /* 00480 - Ultimatum - Choux M.sndh */
TIMEDB_ENTRY( b4273de8,  0,  37248, YM+TA+TB+TD), /* 02548 - cerror - Make some sense.sndh */
TIMEDB_ENTRY( b42bb328,  0,   8064,    YM+TA+TD), /* 01867 - Lotek Style - Sense.net.sndh */
TIMEDB_ENTRY( b42f4e44,  0,   5368,       YM+TA), /* 03483 - Peter Jorgensen - Boesseti.sndh */
TIMEDB_ENTRY( b4625ffd,  0,  29185,       YM+TA), /* 04357 - X The Beermat - Dunno.sndh */
TIMEDB_ENTRY( b467d863,  0,   3097,          YM), /* 03280 - Thothy - Jingle.sndh */
TIMEDB_ENTRY( b4685082,  0,  15675,    YM+TB+TD), /* 01107 - Timbral - DHS - RI Rebellion (YM Rockerz).sndh */
TIMEDB_ENTRY( b46b9acd,  0,   4928,          YM), /* 04215 - Jess - My ST suntans in Brazil (chip).sndh */
TIMEDB_ENTRY( b46f56a9,  0,   6720,          YM), /* 00957 - Dark Angel - What's Your Name.sndh */
TIMEDB_ENTRY( b4758e21,  0,   6727,       YM+TA), /* 03554 - Juri - Juriman.sndh */
TIMEDB_ENTRY( b4987a9a,  0,  12577, YM+TA+TB+TD), /* 01828 - Lotek Style - 80 years.sndh */
TIMEDB_ENTRY( b4ba3887,  0,   1536,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  1,    267,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  2,    215,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  3,    174,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  4,     80,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  5,    150,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  6,    270,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4ba3887,  7,    394,          YM), /* 04704 - Kay Downes - Platform Capers.sndh */
TIMEDB_ENTRY( b4c17f79,  0,   4602, YM+TA+TB+TD), /* 02826 - Msg - Superfly tune 3.sndh */
TIMEDB_ENTRY( b4d2b8ca,  0,   4607,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4d2b8ca,  1,   2303,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4d2b8ca,  2,   3071,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4d2b8ca,  3,   4607,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4d2b8ca,  4,   6559,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4d2b8ca,  5,   3839,          YM), /* 00346 - Tony Williams - Oh No! More Lemmings.sndh */
TIMEDB_ENTRY( b4ebda37,  0,   6239,          YM), /* 02363 - Nexus 6 - Big Jelly Fish V1.sndh */
TIMEDB_ENTRY( b4ee1085,  0,   2878,       YM+TA), /* 01522 - Iso - Family Dog.sndh */
TIMEDB_ENTRY( b506f527,  0,   8194,       YM+TA), /* 04441 - Joard - Synth Dream 2 - Joardy Blitter.sndh */
TIMEDB_ENTRY( b52e4463,  0,  83316,       YM+TA), /* 03961 - Scott Lemmon - The Song.sndh */
TIMEDB_ENTRY( b52f08db,  0,  53760, YM+TA+TB+TD), /* 02593 - STU - GHOSTRIDER.sndh */
TIMEDB_ENTRY( b537451e,  0,   9216,          YM), /* 02459 - Starbuck^Creative Mind - DHS Compo Shorty.sndh */
TIMEDB_ENTRY( b5519818,  0,  11755,          YM), /* 01583 - Outlander - Awesome 22.sndh */
TIMEDB_ENTRY( b552a01d,  0,    262,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  1,      2,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  2,    106,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  3,    220,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  4,    147,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  5,    160,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  6,    200,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  7,    226,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  8,    164,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d,  9,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 10,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 11,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 12,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 13,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 14,     31,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 15,     31,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 16,     31,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 17,     26,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 18,     54,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 19,   1113,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 20,   1520,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b552a01d, 21,    826,          YM), /* 04680 - Ronan - Iraky.sndh */
TIMEDB_ENTRY( b56213b1,  0,   8337, YM+TA+TB+TD), /* 00192 - 505 - Sure Trip 1.sndh */
TIMEDB_ENTRY( b575051a,  0, 100046,       YM+TA), /* 04032 - Roggie - Interludes 5-7.sndh */
TIMEDB_ENTRY( b5928e97,  0,   5759,          YM), /* 00303 - Gary Wheaton - Batman.sndh */
TIMEDB_ENTRY( b5a3e4a1,  0,   3837,    YM+TA+TB), /* 00286 - Baggio - Evolution - Xmas sux.sndh */
TIMEDB_ENTRY( b5bc0d6d,  0,  20480, YM+TA+TB+TD), /* 02618 - STU - SPACE INTRO.sndh */
TIMEDB_ENTRY( b5c59aa2,  0,  13434, YM+TA+TB+TD), /* 01001 - DMA-SC - Magnetic Mind.sndh */
TIMEDB_ENTRY( b5c9b447,  0,   4096,       YM+TA), /* 00798 - Crawdaddy - Legality.sndh */
TIMEDB_ENTRY( b5f57c60,  0,   2057,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  1,    382,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  2,    274,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  3,   3751,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  4,   3751,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  5,   3885,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  6,    256,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b5f57c60,  7,   1015,          YM), /* 02954 - Yann Sevin - Cool Croc Twins.sndh */
TIMEDB_ENTRY( b6020665,  0,   4319,    YM+TA+TD), /* 02976 - Forestcreature - Sorry Sorry Sorry.sndh */
TIMEDB_ENTRY( b60b3e3d,  0,  17390,       YM+TA), /* 03900 - Mad Max - MUDS.sndh */
TIMEDB_ENTRY( b60b3e3d,  1,   2082,       YM+TA), /* 03900 - Mad Max - MUDS.sndh */
TIMEDB_ENTRY( b60b3e3d,  2,   2882,       YM+TA), /* 03900 - Mad Max - MUDS.sndh */
TIMEDB_ENTRY( b611e966,  0,   4181,          YM), /* 01484 - Holger Gehrmann - Hollywood Poker.sndh */
TIMEDB_ENTRY( b611e966,  1,   1493,          YM), /* 01484 - Holger Gehrmann - Hollywood Poker.sndh */
TIMEDB_ENTRY( b611e966,  2,    129,          YM), /* 01484 - Holger Gehrmann - Hollywood Poker.sndh */
TIMEDB_ENTRY( b611e966,  3,     89,          YM), /* 01484 - Holger Gehrmann - Hollywood Poker.sndh */
TIMEDB_ENTRY( b611e966,  4,    115,          YM), /* 01484 - Holger Gehrmann - Hollywood Poker.sndh */
TIMEDB_ENTRY( b612927e,  0,   3839,          YM), /* 04647 - ActoDi - No Timer SV'14.sndh */
TIMEDB_ENTRY( b648c266,  0,   4927,          YM), /* 03621 - Dma-Sc - SV2011INTRO.sndh */
TIMEDB_ENTRY( b64a9732,  0,  15356,          YM), /* 01748 - Jovis - Lametrop - Starfield.sndh */
TIMEDB_ENTRY( b665d7fe,  0,  16524,          YM), /* 02061 - Mad Max - Wiz.sndh */
TIMEDB_ENTRY( b665d7fe,  1,   9384,          YM), /* 02061 - Mad Max - Wiz.sndh */
TIMEDB_ENTRY( b665d7fe,  2,    302,          YM), /* 02061 - Mad Max - Wiz.sndh */
TIMEDB_ENTRY( b6780cd4,  0,  12545, YM+TA+TB+TD), /* 04109 - 505-checkpoint - hopping.sndh */
TIMEDB_ENTRY( b6796f2e,  0,   5952,          YM), /* 02662 - Dave Lowe - Turbo Outrun.sndh */
TIMEDB_ENTRY( b6796f2e,  1,   4480,          YM), /* 02662 - Dave Lowe - Turbo Outrun.sndh */
TIMEDB_ENTRY( b6796f2e,  2,    576,          YM), /* 02662 - Dave Lowe - Turbo Outrun.sndh */
TIMEDB_ENTRY( b6796f2e,  3,   1600,          YM), /* 02662 - Dave Lowe - Turbo Outrun.sndh */
TIMEDB_ENTRY( b67fb95a,  0,   9849,       YM+TA), /* 02741 - Unknown Composer - Jigsaw.sndh */
TIMEDB_ENTRY( b69262a3,  0,   1786,    YM+TA+TB), /* 04279 - 505 - Loop00.sndh */
TIMEDB_ENTRY( b6b816bd,  0,  41601,       YM+TA), /* 03754 - Tufty - NSynth.sndh */
TIMEDB_ENTRY( b6c1f2ed,  0,  17280,          YM), /* 00602 - Chris Mad - Overdrive Snurkle.sndh */
TIMEDB_ENTRY( b6e84eda,  0,   2067,          YM), /* 03042 - n-a - Calypso by Ed Bogas.sndh */
TIMEDB_ENTRY( b6f3f526,  0,  13760,          YM), /* 01790 - Barry Leitch - Combo Racer.sndh */
TIMEDB_ENTRY( b6f4299a,  0,  53356,       YM+TA), /* 03807 - Spaz - Dead Orgy 2.sndh */
TIMEDB_ENTRY( b72736bd,  0,  38725, YM+TA+TB+TD), /* 03394 - cerror - Calculon.sndh */
TIMEDB_ENTRY( b72839f4,  0,  13055,          YM), /* 00347 - Tony Williams - Onslaught.sndh */
TIMEDB_ENTRY( b72839f4,  1,   2559,          YM), /* 00347 - Tony Williams - Onslaught.sndh */
TIMEDB_ENTRY( b72839f4,  2,   2559,          YM), /* 00347 - Tony Williams - Onslaught.sndh */
TIMEDB_ENTRY( b72d8c60,  0,   3840,          YM), /* 01224 - ENS Music - ENS 7.sndh */
TIMEDB_ENTRY( b72d9e98,  0,   2592,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  1,    192,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  2,   2304,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  3,   4800,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  4,   3264,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  5,   1152,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b72d9e98,  6,   1152,          YM), /* 02194 - Mad Max - Great Giana Sisters.sndh */
TIMEDB_ENTRY( b7465bbe,  0,   9215,       YM+TA), /* 03575 - 505-paradox - Finja (STE native mode).sndh */
TIMEDB_ENTRY( b7519b13,  0,   9216,          YM), /* 01407 - Skyline - Snork - Main Menu.sndh */
TIMEDB_ENTRY( b754c2bf,  0,   5017,          YM), /* 03547 - Allister Brimble - Sky High Stuntman.sndh */
TIMEDB_ENTRY( b7595386,  0,   1914,       YM+TA), /* 01368 - Scavenger - Who cares ? (Elite Intro).sndh */
TIMEDB_ENTRY( b75f06c7,  0,  24346,          YM), /* 01699 - Jess - Super Mario Bros Family.sndh */
TIMEDB_ENTRY( b75f8ae7,  0,  28800, YM+TA+TB+TD), /* 02083 - Mad Max - International Karate - Sid.sndh */
TIMEDB_ENTRY( b762df58,  0,   6719, YM+TA+TB+TD), /* 04760 - Shinobi - Outgoing.sndh */
TIMEDB_ENTRY( b76f5fb4,  0,   9849,       YM+TA), /* 01694 - Jess - Lemon Squeezer's Dream.sndh */
TIMEDB_ENTRY( b7b182ac,  0,   5160, YM+TA+TB+TD), /* 03876 - Hylst - Murdered.sndh */
TIMEDB_ENTRY( b7b64257,  0,   8914,       YM+TA), /* 04425 - Lotus - Synth Dream 2 - Big Bubble.sndh */
TIMEDB_ENTRY( b7bb39fc,  0,   2688,    YM+TA+TD), /* 02315 - Tao - Cream (orig. Klaus Ude) - Lin Wu's Challenge 2.sndh */
TIMEDB_ENTRY( b7d0f956,  0,   6335, YM+TA+TB+TD), /* 04100 - Per Almered-XiA-Checkpoint - Smakrik (medley).sndh */
TIMEDB_ENTRY( b7d663ab,  0,   1536,          YM), /* 03272 - Chris Denman - Samurai - The Way of the Warrior.sndh */
TIMEDB_ENTRY( b7d663ab,  1,   3072,          YM), /* 03272 - Chris Denman - Samurai - The Way of the Warrior.sndh */
TIMEDB_ENTRY( b7da1000,  0,   2616,          YM), /* 00143 - David Whittaker - Starwars.sndh */
TIMEDB_ENTRY( b7dddacd,  0,   2240,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  1,   2048,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  2,   3584,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  3,   1728,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  4,   2560,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  5,    256,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  6,   3072,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  7,   2048,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  8,   1536,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd,  9,    145,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd, 10,   3520,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd, 11,    218,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7dddacd, 12,    256,          YM), /* 01916 - Andi McGinty - McDonald Land.sndh */
TIMEDB_ENTRY( b7edf654,  0,   9984,    YM+TA+TD), /* 02512 - Crazy Q - DHS - Snake Funk.sndh */
TIMEDB_ENTRY( b7ef0a5a,  0,  15360, YM+TA+TB+TD), /* 02841 - Marcer - TESTING3 - unfinished.sndh */
TIMEDB_ENTRY( b7fb8030,  0,   5628,    YM+TA+TD), /* 00220 - 7AN - Blip 1.sndh */
TIMEDB_ENTRY( b7fd73b1,  0,  43008, YM+TA+TB+TD), /* 04766 - Garvalf - n-a.sndh */
TIMEDB_ENTRY( b7fd89a5,  0,   5891,    YM+TA+TD), /* 02313 - Tao - Cream (orig. Rob Hubbard) - Jupiter Probe.sndh */
TIMEDB_ENTRY( b80b6df6,  0,   1637,    YM+TA+TB), /* 03606 - Per Almered (XiA) - Jim Parker  Midsomer Theme.sndh */
TIMEDB_ENTRY( b81434a6,  0, 128910,       YM+TA), /* 03811 - Martin Walker - Thunderhawk AH-73M Title.sndh */
TIMEDB_ENTRY( b81564c7,  0,   4558,          YM), /* 01040 - Donald Fakk - Who Cares.sndh */
TIMEDB_ENTRY( b8196d28,  0,   2855,          YM), /* 01200 - Excellence In Art - Sonik.sndh */
TIMEDB_ENTRY( b83c54a0,  0,  11808,          YM), /* 00611 - Chrispy Noodle - Commando.sndh */
TIMEDB_ENTRY( b8498af6,  0,  28608,       YM+TA), /* 03636 - Mad Max - International Karate Digi.sndh */
TIMEDB_ENTRY( b8500cbe,  0,   1599,    YM+TA+TB), /* 02653 - Dma-Sc - POOZ - Mini-Games.sndh */
TIMEDB_ENTRY( b891b124,  0,   3676, YM+TA+TB+TD), /* 04272 - 505 - Checkpoint - Kurts.sndh */
TIMEDB_ENTRY( b894d191,  0,   2653,          YM), /* 01315 - Matt Furniss - Scramble Spirits.sndh */
TIMEDB_ENTRY( b894d191,  1,   2703,          YM), /* 01315 - Matt Furniss - Scramble Spirits.sndh */
TIMEDB_ENTRY( b894d191,  2,    602,          YM), /* 01315 - Matt Furniss - Scramble Spirits.sndh */
TIMEDB_ENTRY( b894d191,  3,    537,          YM), /* 01315 - Matt Furniss - Scramble Spirits.sndh */
TIMEDB_ENTRY( b894d191,  4,   1656,          YM), /* 01315 - Matt Furniss - Scramble Spirits.sndh */
TIMEDB_ENTRY( b8ab5f28,  0,  18436,          YM), /* 00456 - Augustin Vidovic - Prime Time.sndh */
TIMEDB_ENTRY( b8b72dd6,  0,   1337,       YM+TA), /* 01935 - Miq - Miq's Acid - Na Na Ni Ni.sndh */
TIMEDB_ENTRY( b8b8236e,  0,   8833,          YM), /* 00390 - Wally Beben - Phantasm.sndh */
TIMEDB_ENTRY( b8b8236e,  1,   2050,          YM), /* 00390 - Wally Beben - Phantasm.sndh */
TIMEDB_ENTRY( b8b8236e,  2,   1537,          YM), /* 00390 - Wally Beben - Phantasm.sndh */
TIMEDB_ENTRY( b8b8236e,  3,  13153,          YM), /* 00390 - Wally Beben - Phantasm.sndh */
TIMEDB_ENTRY( b8d715dc,  0,   2509,       YM+TA), /* 02683 - Donald Fakk - Skit (Delta Remix).sndh */
TIMEDB_ENTRY( b8dd9ce0,  0,   1536,          YM), /* 01418 - Skyline - Snork Intro - Vodka.sndh */
TIMEDB_ENTRY( b8e70c7e,  0,  14208,          YM), /* 02120 - Mad Max - Bangkok Nights - Loader.sndh */
TIMEDB_ENTRY( b8fc4054,  0,   8069,    YM+TA+TD), /* 02320 - Tao - Cream - Mad Vision (never used).sndh */
TIMEDB_ENTRY( b90c66ef,  0,   6095,          YM), /* 03087 - n-a - 12 Bar Blues-IRK.sndh */
TIMEDB_ENTRY( b91f61fe,  0,   8352,          YM), /* 01996 - Mad Max - Axel F.sndh */
TIMEDB_ENTRY( b9267165,  0,  10752,          YM), /* 01623 - No More - No More.sndh */
TIMEDB_ENTRY( b9589a0f,  0,   8058,       YM+TA), /* 00896 - Danny Oneway - C64.sndh */
TIMEDB_ENTRY( b97102f1,  0,   5115, YM+TA+TB+TD), /* 02595 - STU - HOTCOM.sndh */
TIMEDB_ENTRY( b9796766,  0, 119703,       YM+TA), /* 03835 - P Webb - Hoversprint.sndh */
TIMEDB_ENTRY( b98f03fb,  0,  15842,       YM+TA), /* 04767 - Dave Lowe - After Burner digi.sndh */
TIMEDB_ENTRY( b9b50cb1,  0,   7947,          YM), /* 01038 - Donald Fakk - Pressure.sndh */
TIMEDB_ENTRY( b9c04132,  0,   2688,       YM+TA), /* 00250 - Antichrist - Chelou No!!.sndh */
TIMEDB_ENTRY( b9f1d758,  0,   5046,          YM), /* 03106 - n-a - Daybreaker-ELO.sndh */
TIMEDB_ENTRY( b9fde005,  0,   3072,          YM), /* 03654 - Jedi - Nothing Is More - Song 3.sndh */
TIMEDB_ENTRY( ba2c50f8,  0,   6138,          YM), /* 03287 - Adrian Dalecki - Logo.sndh */
TIMEDB_ENTRY( ba2c50f8,  1,   8826,          YM), /* 03287 - Adrian Dalecki - Logo.sndh */
TIMEDB_ENTRY( ba2c50f8,  2,   2682,          YM), /* 03287 - Adrian Dalecki - Logo.sndh */
TIMEDB_ENTRY( ba2f914f,  0,   7279, YM+TA+TB+TD), /* 03392 - Ultrasyd - Bouncing Banana.sndh */
TIMEDB_ENTRY( ba4c288d,  0,   5754,          YM), /* 04260 - 505 - Gamesong 1.sndh */
TIMEDB_ENTRY( ba574393,  0,   3200,          YM), /* 04466 - Charles Callet - Meurt En Serie.sndh */
TIMEDB_ENTRY( ba5bf3fb,  0,   4860,       YM+TA), /* 01629 - No More - Vectronix.sndh */
TIMEDB_ENTRY( ba6a8ef8,  0,    379,          YM), /* 02945 - Dave Munsie - Kid Kong.sndh */
TIMEDB_ENTRY( ba6a8ef8,  1,    435,          YM), /* 02945 - Dave Munsie - Kid Kong.sndh */
TIMEDB_ENTRY( ba6a8ef8,  2,     73,          YM), /* 02945 - Dave Munsie - Kid Kong.sndh */
TIMEDB_ENTRY( ba6f3e4b,  0,   5340,          YM), /* 03631 - Marcer - n-a.sndh */
TIMEDB_ENTRY( ba708458,  0,   1920,          YM), /* 01057 - Jonathan Dunn - Darkman.sndh */
TIMEDB_ENTRY( ba708458,  1,   2304,          YM), /* 01057 - Jonathan Dunn - Darkman.sndh */
TIMEDB_ENTRY( ba708458,  2,    385,          YM), /* 01057 - Jonathan Dunn - Darkman.sndh */
TIMEDB_ENTRY( ba708458,  3,   5376,          YM), /* 01057 - Jonathan Dunn - Darkman.sndh */
TIMEDB_ENTRY( ba71e92e,  0,   5809,          YM), /* 03163 - Unknown - SOUND OF SILENCE By Paul Simon.sndh */
TIMEDB_ENTRY( ba886f7d,  0,  40729,          YM), /* 00007 - gwEm - Space Trip (dub mix).sndh */
TIMEDB_ENTRY( baa002c0,  0,   9724,    YM+TA+TB), /* 00212 - 505 - Vagrant.sndh */
TIMEDB_ENTRY( bab4b43b,  0,  20992,          YM), /* 04459 - Charles Callet - Bob Morane - Chevalerie.sndh */
TIMEDB_ENTRY( bab4b43b,  1,   3392,          YM), /* 04459 - Charles Callet - Bob Morane - Chevalerie.sndh */
TIMEDB_ENTRY( bac8350c,  0,   7919,          YM), /* 01316 - Matt Furniss - Shadowlands.sndh */
TIMEDB_ENTRY( bad7e342,  0,   1676,          YM), /* 03487 - Unknown - Trailblazer.sndh */
TIMEDB_ENTRY( bb183018,  0,   6410,          YM), /* 03028 - Alice Amore - ETUDE (Opus25,No.1)by F.Chopin.sndh */
TIMEDB_ENTRY( bb188322,  0,  15177, YM+TA+TB+TD), /* 02105 - Mad Max - Zoids - Sid.sndh */
TIMEDB_ENTRY( bb1c8a06,  0,    521,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb1c8a06,  1,    290,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb1c8a06,  2,    146,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb1c8a06,  3,    130,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb1c8a06,  4,   1585,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb1c8a06,  5,     96,          YM), /* 04654 - Peter Johnson - Arkanoid - Revenge of Doh.sndh */
TIMEDB_ENTRY( bb317266,  0,  10752,       YM+TA), /* 02156 - Mad Max - Ooh Crikey Wot a Scorcher Main Menu.sndh */
TIMEDB_ENTRY( bb425ad7,  0,   3455, YM+TA+TB+TD), /* 03052 - Orbiter and Timbral - Dubmoods MILF (in da park).sndh */
TIMEDB_ENTRY( bb470a6d,  0,  14506,          YM), /* 02040 - Mad Max - Saboteur 2.sndh */
TIMEDB_ENTRY( bb65fe65,  0,  11059,    YM+TA+TD), /* 01852 - Lotek Style - Her Face.sndh */
TIMEDB_ENTRY( bb765c7e,  0,   8442,       YM+TA), /* 02634 - Miggy - EELectric Game Music(90%).sndh */
TIMEDB_ENTRY( bb7fffa1,  0,  29184, YM+TA+TB+TD), /* 03469 - cerror - Loader tune.sndh */
TIMEDB_ENTRY( bb8432f7,  0,   2964,          YM), /* 03144 - n-a - OLD JOE CLARK'S BOOGIE.sndh */
TIMEDB_ENTRY( bb8625e1,  0,  19578,       YM+TA), /* 01678 - Jess - Sharpness Buzztone.sndh */
TIMEDB_ENTRY( bb875700,  0,   9600,          YM), /* 03260 - T1000 - The Beast Club - TS 11.sndh */
TIMEDB_ENTRY( bb96bca5,  0,   3480,       YM+TA), /* 03601 - Excellence in Art - Sommarhack 2011 Invite.sndh */
TIMEDB_ENTRY( bbac27d1,  0,   8058,    YM+TA+TB), /* 02774 - Gary - UCM 5 Intro.sndh */
TIMEDB_ENTRY( bbb25894,  0,   1280,       YM+TA), /* 00925 - Dark Angel - I Hate Techo.sndh */
TIMEDB_ENTRY( bbb47b36,  0,   5370,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  1,   5112,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  2,   5387,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  3,   6204,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  4,   6361,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  5,   6905,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbb47b36,  6,    180,          YM), /* 00874 - Ben Daglish - Supercars.sndh */
TIMEDB_ENTRY( bbbbd013,  0,   2469,          YM), /* 04731 - Juha Terho - AGI Quest 1.sndh */
TIMEDB_ENTRY( bbbbd013,  1,    693,          YM), /* 04731 - Juha Terho - AGI Quest 1.sndh */
TIMEDB_ENTRY( bbbbd013,  2,    588,          YM), /* 04731 - Juha Terho - AGI Quest 1.sndh */
TIMEDB_ENTRY( bbbbd013,  3,    742,          YM), /* 04731 - Juha Terho - AGI Quest 1.sndh */
TIMEDB_ENTRY( bbbbd013,  4,    672,          YM), /* 04731 - Juha Terho - AGI Quest 1.sndh */
TIMEDB_ENTRY( bbcb5276,  0,   1248,          YM), /* 02955 - Skinnybone - Special Forces.sndh */
TIMEDB_ENTRY( bbe35e3b,  0,   1179,          YM), /* 04499 - C.R.G. Johnson - Boing.sndh */
TIMEDB_ENTRY( bbe35e3b,  1,   1280,          YM), /* 04499 - C.R.G. Johnson - Boing.sndh */
TIMEDB_ENTRY( bbed6592,  0,   6537,          YM), /* 03165 - n-a - Somewhere, My Love.sndh */
TIMEDB_ENTRY( bbf23b4e,  0,   4112,          YM), /* 01668 - Stefan Jerowski - Sarakon.sndh */
TIMEDB_ENTRY( bbf23b4e,  1,   5376,          YM), /* 01668 - Stefan Jerowski - Sarakon.sndh */
TIMEDB_ENTRY( bbf23b4e,  2,   2560,          YM), /* 01668 - Stefan Jerowski - Sarakon.sndh */
TIMEDB_ENTRY( bc031a4c,  0,  25216,       YM+TD), /* 03994 - XyNo - Why Emm?.sndh */
TIMEDB_ENTRY( bc0e28fe,  0,  24768,          YM), /* 02578 - STU - BOMBERMEN.sndh */
TIMEDB_ENTRY( bc0ea2f0,  0,  19578,          YM), /* 04366 - Jess - Sharpness Buzztone (chip).sndh */
TIMEDB_ENTRY( bc161558,  0,  29440,          YM), /* 01474 - STU - RUN.sndh */
TIMEDB_ENTRY( bc18a985,  0,   4992,          YM), /* 01135 - Tyan - Revolution - James Brown.sndh */
TIMEDB_ENTRY( bc1b9871,  0,   2518,    YM+TA+TD), /* 02669 - Donald Fakk - Flum Hum.sndh */
TIMEDB_ENTRY( bc46b8cd,  0,  68544,          YM), /* 01442 - Stax - New Happy Wave.sndh */
TIMEDB_ENTRY( bc57dc94,  0,   1920,       YM+TA), /* 00469 - Viking - Ending.sndh */
TIMEDB_ENTRY( bc58e7d0,  0,   8928,          YM), /* 00962 - Mike Davies - Dynasty Wars.sndh */
TIMEDB_ENTRY( bc58e7d0,  1,   3072,          YM), /* 00962 - Mike Davies - Dynasty Wars.sndh */
TIMEDB_ENTRY( bc58e7d0,  2,   1584,          YM), /* 00962 - Mike Davies - Dynasty Wars.sndh */
TIMEDB_ENTRY( bc58e7d0,  3,    105,          YM), /* 00962 - Mike Davies - Dynasty Wars.sndh */
TIMEDB_ENTRY( bc58e7d0,  4,    153,          YM), /* 00962 - Mike Davies - Dynasty Wars.sndh */
TIMEDB_ENTRY( bc6ef6f9,  0,  10943,    YM+TA+TD), /* 02696 - Dubmood - RazoR 1911 - Botkyrka VS ST-STYLE.sndh */
TIMEDB_ENTRY( bc857fb3,  0,   3503, YM+TA+TB+TD), /* 02731 - maktone-fairlight - reason.sndh */
TIMEDB_ENTRY( bc874908,  0,  15354, YM+TA+TB+TD), /* 01018 - DMA-SC - vistors from dreams.sndh */
TIMEDB_ENTRY( bc8ba5b8,  0,  31864,       YM+TA), /* 03008 - Mad Max - Stormlord Title Music.sndh */
TIMEDB_ENTRY( bc9380cc,  0,  11684,          YM), /* 03233 - T1000 - The Beast Club - TBC 3.sndh */
TIMEDB_ENTRY( bc9c82ce,  0,   6265, YM+TA+TB+TD), /* 00051 - 505 - Gauloise - Posh Demo original.sndh */
TIMEDB_ENTRY( bcb0814b,  0,  23230,    YM+TA+TD), /* 03413 - Tao - Cream (orig. Jochen Hippel) - Lethal Xcess Intro (Edit).sndh */
TIMEDB_ENTRY( bcbcb5fb,  0,   9849, YM+TA+TB+TD), /* 01952 - MSG - Chu Chu Rocket - Title.sndh */
TIMEDB_ENTRY( bcbda481,  0,    576,       YM+TA), /* 00937 - Dark Angel - Mixpart Remix.sndh */
TIMEDB_ENTRY( bcc07af1,  0,   5518,       YM+TA), /* 00056 - 505 - Hmmm One - EIL 3 invite.sndh */
TIMEDB_ENTRY( bcc24215,  0,  31559,          YM), /* 00545 - Bodenst ndig 2000 - Highscore (YM Rockerz).sndh */
TIMEDB_ENTRY( bcd3fb1c,  0,   9215, YM+TA+TB+TD), /* 02524 - DUBMOOD (quasar) - HYBRID FUNKY STARS.sndh */
TIMEDB_ENTRY( bcd9b2e1,  0,   1541,          YM), /* 02282 - Tao - Cream - Just Loadin' (Just Buggin).sndh */
TIMEDB_ENTRY( bcd9b2e1,  1,    389,          YM), /* 02282 - Tao - Cream - Just Loadin' (Just Buggin).sndh */
TIMEDB_ENTRY( bce196c9,  0,  54981,       YM+TA), /* 02126 - Mad Max - Knucklebusters.sndh */
TIMEDB_ENTRY( bd03e011,  0,   3117,          YM), /* 03279 - Thothy - Heuthier.sndh */
TIMEDB_ENTRY( bd21f512,  0,  36380,       YM+TA), /* 03716 - Richard Joseph - Speedball 2.sndh */
TIMEDB_ENTRY( bd24a414,  0,   2048,          YM), /* 01414 - Skyline - Snork Intro - Transbeauce.sndh */
TIMEDB_ENTRY( bd3181a0,  0, 123322,       YM+TA), /* 03772 - Exarch - Psychedika.sndh */
TIMEDB_ENTRY( bd3181a0,  1, 158045,       YM+TA), /* 03772 - Exarch - Psychedika.sndh */
TIMEDB_ENTRY( bd3181a0,  2, 104074,       YM+TA), /* 03772 - Exarch - Psychedika.sndh */
TIMEDB_ENTRY( bd3181a0,  3,  42113,       YM+TA), /* 03772 - Exarch - Psychedika.sndh */
TIMEDB_ENTRY( bd3181a0,  4,  71252,       YM+TA), /* 03772 - Exarch - Psychedika.sndh */
TIMEDB_ENTRY( bd3ba2ea,  0,  17982, YM+TA+TB+TD), /* 02567 - marcer - outrun 2005 remix.sndh */
TIMEDB_ENTRY( bd485b20,  0,    120,          YM), /* 04529 - Andreaz - Fia.sndh */
TIMEDB_ENTRY( bd485b20,  1,    282,          YM), /* 04529 - Andreaz - Fia.sndh */
TIMEDB_ENTRY( bd485b20,  2,    183,          YM), /* 04529 - Andreaz - Fia.sndh */
TIMEDB_ENTRY( bd485b20,  3,   1280,          YM), /* 04529 - Andreaz - Fia.sndh */
TIMEDB_ENTRY( bd4dad7d,  0,   5120,          YM), /* 04596 - Rude Kid - In Air.sndh */
TIMEDB_ENTRY( bd530e2b,  0,   1922,          YM), /* 01262 - Aaron Fothergill - Sky Strike.sndh */
TIMEDB_ENTRY( bd530e2b,  1,   1038,          YM), /* 01262 - Aaron Fothergill - Sky Strike.sndh */
TIMEDB_ENTRY( bd530e2b,  2,   4456,          YM), /* 01262 - Aaron Fothergill - Sky Strike.sndh */
TIMEDB_ENTRY( bd5738da,  0,   3022,          YM), /* 03887 - Unknown - Hommage.sndh */
TIMEDB_ENTRY( bd5b476a,  0,   3072, YM+TA+TB+TD), /* 02099 - Mad Max - Thanatos - Sid.sndh */
TIMEDB_ENTRY( bd619255,  0,   1280,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  1,    126,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  2,     44,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  3,     18,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  4,     30,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  5,     41,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  6,     39,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd619255,  7,    104,          YM), /* 04516 - Ross McNaughton - Confusion.sndh */
TIMEDB_ENTRY( bd640e59,  0,  22272, YM+TA+TB+TD), /* 02441 - cerror (or.by.4mat.c64) - VERY BLAND.sndh */
TIMEDB_ENTRY( bd75fdd2,  0,   8384, YM+TA+TB+TD), /* 02079 - Mad Max - Harvey Smiths Show Jumping - Sid.sndh */
TIMEDB_ENTRY( bdac181f,  0,  10399,          YM), /* 03122 - n-a - I Honestly Love You.sndh */
TIMEDB_ENTRY( bdad5be6,  0,  11904,          YM), /* 01633 - Nexus - Dreamzone - Cool J.sndh */
TIMEDB_ENTRY( bdb1ca4d,  0,   3775,          YM), /* 01246 - FFT - Phaleon - Beast 2.sndh */
TIMEDB_ENTRY( bdc295cc,  0,  11089,          YM), /* 01182 - Excellence In Art - Get into The Groove.sndh */
TIMEDB_ENTRY( bdcb7563,  0,  13942,          YM), /* 00125 - David Whittaker - Quadrilian.sndh */
TIMEDB_ENTRY( bdcb7563,  1,  11106,          YM), /* 00125 - David Whittaker - Quadrilian.sndh */
TIMEDB_ENTRY( bdeb5cd0,  0,  27845,       YM+TA), /* 04354 - X The Beermat - Carrot.sndh */
TIMEDB_ENTRY( be03a1d8,  0,  11193, YM+TA+TB+TD), /* 01013 - DMA-SC - Sweety End.sndh */
TIMEDB_ENTRY( be12afd4,  0,  10240,    YM+TA+TD), /* 02269 - Tao - Cream - The Arrival (Steps).sndh */
TIMEDB_ENTRY( be23c6ad,  0, 107030,       YM+TA), /* 03766 - Waldemar Belwon - Hey (Music Show 1).sndh */
TIMEDB_ENTRY( be24fa83,  0, 174530,       YM+TA), /* 03849 - Chris of TGE - NTM - TGE.sndh */
TIMEDB_ENTRY( be34d10d,  0,  11721,          YM), /* 00019 - 4-Mat - Warzone.sndh */
TIMEDB_ENTRY( be43c9fb,  0,   8452, YM+TA+TB+TD), /* 00993 - Dma Sc - Sector One ( 2003 ) - Cybernetic Dancer (YM Rockerz).sndh */
TIMEDB_ENTRY( be534838,  0,  46337,       YM+TA), /* 03773 - Exarch - Scientia.sndh */
TIMEDB_ENTRY( be534838,  1, 217109,       YM+TA), /* 03773 - Exarch - Scientia.sndh */
TIMEDB_ENTRY( be534838,  2, 201444,       YM+TA), /* 03773 - Exarch - Scientia.sndh */
TIMEDB_ENTRY( be534838,  3, 109257,       YM+TA), /* 03773 - Exarch - Scientia.sndh */
TIMEDB_ENTRY( be534838,  4,  70255,       YM+TA), /* 03773 - Exarch - Scientia.sndh */
TIMEDB_ENTRY( be564c23,  0,   5742,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  1,   5742,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  2,   6080,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  3,   4130,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  4,   5600,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  5,   9530,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  6,   2076,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  7,   5355,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23,  9,    922,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be564c23, 10,   4764,          YM), /* 01913 - Jose A. Martin - Risky Woods.sndh */
TIMEDB_ENTRY( be71d4b7,  0,   6568,          YM), /* 00375 - Beast - TSC - TSC 31.sndh */
TIMEDB_ENTRY( be75a4a1,  0,   8718,          YM), /* 01271 - Frazer - Tjosan.sndh */
TIMEDB_ENTRY( be7893e1,  0,  10088,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be7893e1,  1,    245,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be7893e1,  2,    170,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be7893e1,  3,    195,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be7893e1,  4,    201,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be7893e1,  5,    113,          YM), /* 03516 - Peter Johnson - Daley Thompson Olympic Challenge.sndh */
TIMEDB_ENTRY( be80306e,  0,  38281,       YM+TA), /* 04028 - Roggie - Odyssey.sndh */
TIMEDB_ENTRY( be8284f9,  0,   9594,       YM+TA), /* 02893 - Kenet - Gameboy.sndh */
TIMEDB_ENTRY( be93edb9,  0,   8273,          YM), /* 02505 - Crazy Q - Crushed Insect.sndh */
TIMEDB_ENTRY( be9e7cd2,  0,    381,    YM+TA+TD), /* 04283 - 505 - Loop07.sndh */
TIMEDB_ENTRY( bea40b51,  0,  12282,          YM), /* 01688 - Jess - Generation 2.0.sndh */
TIMEDB_ENTRY( beb1517f,  0,     25,          YM), /* 01357 - Scavenger - Password.sndh */
TIMEDB_ENTRY( beb89041,  0,   2303,       YM+TA), /* 03423 - Dubmood - D-Bug 197 Jam Version.sndh */
TIMEDB_ENTRY( bec0d438,  0,  24372,          YM), /* 01092 - Techno - Mathematiques Squizophrenes Intro.sndh */
TIMEDB_ENTRY( bec625b1,  0,  11315,          YM), /* 00879 - Ben Daglish - Ultimate Golf.sndh */
TIMEDB_ENTRY( bedc45d2,  0,   7488,          YM), /* 03209 - Jonathan Dunn - Battle Command.sndh */
TIMEDB_ENTRY( bef28cde,  0,  13443,          YM), /* 02243 - Tao - Cream (orig. Rob Hubbard) - Rasputin (Stmygm).sndh */
TIMEDB_ENTRY( bef28cde,  1,   2179,          YM), /* 02243 - Tao - Cream (orig. Rob Hubbard) - Rasputin (Stmygm).sndh */
TIMEDB_ENTRY( befa3934,  0,  11066,          YM), /* 02058 - Mad Max - Thing On A Spring.sndh */
TIMEDB_ENTRY( bf0e1fae,  0,  26463,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  1,  39429,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  2,  34309,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  3,  26625,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  4,  55297,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  5,  33797,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  6, 106497,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf0e1fae,  7, 131443,       YM+TA), /* 03808 - Spherical - Quartet Music Demo II.sndh */
TIMEDB_ENTRY( bf3fce4e,  0,   9208, YM+TA+TB+TD), /* 01874 - Lotek Style - Zeitloch.sndh */
TIMEDB_ENTRY( bf4dd448,  0,   9600,          YM), /* 02121 - Mad Max - Comic Bakery.sndh */
TIMEDB_ENTRY( bf4de37d,  0,   9059,    YM+TA+TD), /* 03480 - Dma-Sc - Sector One - Spring Party.sndh */
TIMEDB_ENTRY( bf6004d3,  0,   1536,          YM), /* 01415 - Skyline - Snork Intro - ULM DSOTS.sndh */
TIMEDB_ENTRY( bf721338,  0,    337,          YM), /* 01792 - Barry Leitch - Frankenstein.sndh */
TIMEDB_ENTRY( bf721338,  1,   1573,          YM), /* 01792 - Barry Leitch - Frankenstein.sndh */
TIMEDB_ENTRY( bf721338,  2,    229,          YM), /* 01792 - Barry Leitch - Frankenstein.sndh */
TIMEDB_ENTRY( bf721338,  3,   1921,          YM), /* 01792 - Barry Leitch - Frankenstein.sndh */
TIMEDB_ENTRY( bf9c5998,  0,   2880,          YM), /* 00778 - Crawdaddy - Blazer.sndh */
TIMEDB_ENTRY( bfb64ee2,  0,    768,          YM), /* 00934 - Dark Angel - Mathematica Intro.sndh */
TIMEDB_ENTRY( bfb698ca,  0,   3000,          YM), /* 03147 - n-a - Peter Gunn by JFS.sndh */
TIMEDB_ENTRY( bfd90bda,  0,   5772,       YM+TA), /* 01619 - No More - GSC 56.sndh */
TIMEDB_ENTRY( bfdcbcb2,  0,   6940, YM+TA+TB+TD), /* 01869 - Lotek Style - Seraphim.sndh */
TIMEDB_ENTRY( bffa5268,  0,   4369,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  1,  24575,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  2,    149,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  3,    231,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  4,    630,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  5,    372,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  6,    987,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  7,    987,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffa5268,  8,   1228,          YM), /* 00099 - David Whittaker - Eye of Horus.sndh */
TIMEDB_ENTRY( bffeb734,  0,   1535,       YM+TD), /* 03290 - Dma-Sc - ATIPYK Developments - Manical Drop - 2 Players music #3.sndh */
TIMEDB_ENTRY( c01bb09d,  0,  41600, YM+TA+TB+TD), /* 02629 - schmx-awp - Schmx It.sndh */
TIMEDB_ENTRY( c03b9a4e,  0,   1408,          YM), /* 04381 - Chris Denman - Caesar.sndh */
TIMEDB_ENTRY( c045e331,  0,   3835,       YM+TD), /* 01380 - Schemantix - Enola Gay.sndh */
TIMEDB_ENTRY( c0495c5a,  0,  30860, YM+TA+TB+TD), /* 04271 - 505 - Kitchentable.sndh */
TIMEDB_ENTRY( c04cb940,  0,   8064,          YM), /* 01724 - Jet Power - AGITRS - RABUBA SAHNE TEE.sndh */
TIMEDB_ENTRY( c059ec25,  0,   4989,    YM+TB+TD), /* 01282 - Frequent - Asbest.sndh */
TIMEDB_ENTRY( c05e2f85,  0,  16512,    YM+TA+TD), /* 02258 - Tao - Cream - Line Of Fire (Steps).sndh */
TIMEDB_ENTRY( c097612e,  0,   7728,          YM), /* 01977 - MSG - MSG 07.sndh */
TIMEDB_ENTRY( c09b74ed,  0,   3759,          YM), /* 02943 - Dave Munsie - Evader.sndh */
TIMEDB_ENTRY( c0a4682e,  0,   3072,       YM+TA), /* 00256 - Antichrist - Melody.sndh */
TIMEDB_ENTRY( c0ba0415,  0,   9210,       YM+TA), /* 04225 - 505 - 4th.sndh */
TIMEDB_ENTRY( c0bfa7c8,  0,  33061,       YM+TA), /* 04814 - Unknown - Tin.sndh */
TIMEDB_ENTRY( c0c5ae91,  0,   1926,          YM), /* 03179 - Unknown - Two Part Invention by J. Bach.sndh */
TIMEDB_ENTRY( c0c689b3,  0,   6630,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  1,   7416,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  2,    141,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  3,    128,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  4,  10496,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  5,    140,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  6,    140,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  7,    598,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0c689b3,  8,  14254,          YM), /* 00574 - Jas C Brooke - Savage.sndh */
TIMEDB_ENTRY( c0d0d761,  0,   2303, YM+TA+TB+TD), /* 02989 - 505-paradox - krabbencocktail.sndh */
TIMEDB_ENTRY( c0d9d308,  0,   5260,       YM+TA), /* 04631 - David Whittaker - Xenon 2 Title.sndh */
TIMEDB_ENTRY( c10e2b29,  0,   6144,       YM+TA), /* 04410 - Dr Beat - ENS 1 - SD player.sndh */
TIMEDB_ENTRY( c11d69b3,  0,   7680,          YM), /* 01554 - Rhino - Xmas 2003 Intro - Song 2.sndh */
TIMEDB_ENTRY( c125f0b1,  0,  26490,          YM), /* 01680 - Jess - Throw The Confettis.sndh */
TIMEDB_ENTRY( c12837c7,  0,   1536,          YM), /* 01388 - Andy Severn - Joe Blade.sndh */
TIMEDB_ENTRY( c133576c,  0,   6424,          YM), /* 03176 - Unknown -        Today   Randy Sparks.sndh */
TIMEDB_ENTRY( c14407ec,  0,   2486,    YM+TA+TB), /* 00202 - 505 - Trace.sndh */
TIMEDB_ENTRY( c162d957,  0,   9216,          YM), /* 01099 - Techno - Rigolo.sndh */
TIMEDB_ENTRY( c1ae24ca,  0,   1920,          YM), /* 02167 - Mad Max - Ikari Union.sndh */
TIMEDB_ENTRY( c1b840c6,  0,   2688,          YM), /* 01398 - Skyline - Snork - Loader 1.sndh */
TIMEDB_ENTRY( c1ba284b,  0,   6176,          YM), /* 01066 - Jonathan Dunn - Puzznic.sndh */
TIMEDB_ENTRY( c1bd8522,  0,   2681, YM+TA+TB+TD), /* 01965 - MSG - Godpey - Story 0.sndh */
TIMEDB_ENTRY( c1bdf638,  0,  13568,          YM), /* 00927 - Dark Angel - Kaskades.sndh */
TIMEDB_ENTRY( c1d303aa,  0,   1857,          YM), /* 00339 - Tony Williams - Espana 92.sndh */
TIMEDB_ENTRY( c1d303aa,  1,   2689,          YM), /* 00339 - Tony Williams - Espana 92.sndh */
TIMEDB_ENTRY( c1d303aa,  2,    196,          YM), /* 00339 - Tony Williams - Espana 92.sndh */
TIMEDB_ENTRY( c1d303aa,  3,    226,          YM), /* 00339 - Tony Williams - Espana 92.sndh */
TIMEDB_ENTRY( c1d303aa,  4,    130,          YM), /* 00339 - Tony Williams - Espana 92.sndh */
TIMEDB_ENTRY( c1d60aca,  0,   7388,    YM+TA+TD), /* 02746 - Marcer - Bellanotte Chip - O-2006-Invit Demo.sndh */
TIMEDB_ENTRY( c1dce608,  0,   3072,       YM+TA), /* 00953 - Dark Angel - Techno Stormer.sndh */
TIMEDB_ENTRY( c1e48f02,  0,   4608,          YM), /* 01647 - Jedi - Jedi Fun.sndh */
TIMEDB_ENTRY( c1f006f2,  0,  13249,    YM+TA+TD), /* 02226 - Tao - Cream - Toxygen - Intro (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( c1f3233f,  0,   8057,    YM+TA+TD), /* 01840 - Lotek Style - Crystal.sndh */
TIMEDB_ENTRY( c1f36a07,  0,   6144,          YM), /* 02176 - Mad Max - In 80 Days Around the World.sndh */
TIMEDB_ENTRY( c1f36a07,  1,   4704,          YM), /* 02176 - Mad Max - In 80 Days Around the World.sndh */
TIMEDB_ENTRY( c1f36a07,  2,   6336,          YM), /* 02176 - Mad Max - In 80 Days Around the World.sndh */
TIMEDB_ENTRY( c1f36a07,  3,   6528,          YM), /* 02176 - Mad Max - In 80 Days Around the World.sndh */
TIMEDB_ENTRY( c1f36a07,  4,   6336,          YM), /* 02176 - Mad Max - In 80 Days Around the World.sndh */
TIMEDB_ENTRY( c1fec1c6,  0,   6144,          YM), /* 02161 - Mad Max - Syntax Terror - Loader.sndh */
TIMEDB_ENTRY( c2221d2d,  0,   1920,          YM), /* 02144 - Mad Max - International Karate - LSD.sndh */
TIMEDB_ENTRY( c28d0598,  0,   4096,          YM), /* 01997 - Mad Max - Bach The Future.sndh */
TIMEDB_ENTRY( c2972abe,  0,   2879, YM+TA+TB+TD), /* 02780 - 505-cpt^pdx - 052.sndh */
TIMEDB_ENTRY( c2a12b71,  0,   5472,          YM), /* 00114 - David Whittaker - Leathernecks.sndh */
TIMEDB_ENTRY( c2a12b71,  1,   5376,          YM), /* 00114 - David Whittaker - Leathernecks.sndh */
TIMEDB_ENTRY( c2a12b71,  2,    270,          YM), /* 00114 - David Whittaker - Leathernecks.sndh */
TIMEDB_ENTRY( c2a979fb,  0,   6940,          YM), /* 01868 - Lotek Style - Seraphim (Chip).sndh */
TIMEDB_ENTRY( c2bb978b,  0,   5413,          YM), /* 00380 - Wally Beben - Circus Games.sndh */
TIMEDB_ENTRY( c2bb978b,  2,   1537,          YM), /* 00380 - Wally Beben - Circus Games.sndh */
TIMEDB_ENTRY( c2bb978b,  3,   3073,          YM), /* 00380 - Wally Beben - Circus Games.sndh */
TIMEDB_ENTRY( c2bb978b,  4,   9553,          YM), /* 00380 - Wally Beben - Circus Games.sndh */
TIMEDB_ENTRY( c2c9c987,  0,    896,          YM), /* 01194 - Excellence In Art - No.sndh */
TIMEDB_ENTRY( c2ccc6ff,  0,  16517,          YM), /* 02281 - Tao - Cream - Line Of Fire (Just Buggin).sndh */
TIMEDB_ENTRY( c2d49b76,  0,   5568,          YM), /* 00452 - Augustin Vidovic - Music Mon Mix.sndh */
TIMEDB_ENTRY( c2d7c987,  0,   3072,          YM), /* 01198 - Excellence In Art - Runner.sndh */
TIMEDB_ENTRY( c2df6c6b,  0,  14225,       YM+TD), /* 01384 - Scy - TSCC Mod Compil 5.sndh */
TIMEDB_ENTRY( c2ecaab7,  0,   3842,          YM), /* 00427 - Big Alec - PYM - Loader.sndh */
TIMEDB_ENTRY( c2ef6ee4,  0,   3455, YM+TA+TB+TD), /* 04778 - DMA-SC - Lovetro Main Theme.sndh */
TIMEDB_ENTRY( c2fa1a91,  0,   1544,          YM), /* 04010 - Danny Oneway - The Year After - Risk #1.sndh */
TIMEDB_ENTRY( c2fe8b3d,  0,   1149,          YM), /* 00724 - Andrew Taylor - Double Dragon 3 - The Rosetta Stone.sndh */
TIMEDB_ENTRY( c2fe8b3d,  1,   3645,          YM), /* 00724 - Andrew Taylor - Double Dragon 3 - The Rosetta Stone.sndh */
TIMEDB_ENTRY( c2fe8b3d,  2,   1533,          YM), /* 00724 - Andrew Taylor - Double Dragon 3 - The Rosetta Stone.sndh */
TIMEDB_ENTRY( c312cf0f,  0,   2040,       YM+TD), /* 00470 - Viking - Fairy.sndh */
TIMEDB_ENTRY( c3310a79,  0,   9600,          YM), /* 01661 - Jedi - Unfinished 3.sndh */
TIMEDB_ENTRY( c33f9df3,  0,   2912,       YM+TA), /* 00438 - Big Alec - Sticks.sndh */
TIMEDB_ENTRY( c35a93d7,  0,   8069,          YM), /* 02279 - Tao - Cream - Just Feel It 1 (Just Buggin).sndh */
TIMEDB_ENTRY( c35ba437,  0,   7160, YM+TA+TB+TD), /* 04301 - 505-checkpoint - Render.sndh */
TIMEDB_ENTRY( c36a424a,  0,   8976,          YM), /* 01549 - Rhino - GBISTCP - Electronic Images.sndh */
TIMEDB_ENTRY( c36c51b1,  0,   8192,          YM), /* 01656 - Jedi - Sonar.sndh */
TIMEDB_ENTRY( c38b053f,  0,   8448,          YM), /* 01578 - Powa Pixel - Mega 4 Disk Mag.sndh */
TIMEDB_ENTRY( c38dfe7f,  0,  10239,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c38dfe7f,  1,    491,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c38dfe7f,  2,    256,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c38dfe7f,  3,   7168,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c38dfe7f,  4,   4096,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c38dfe7f,  5,   3584,          YM), /* 02207 - Mad Max - Pablo Gold Of Montezuma.sndh */
TIMEDB_ENTRY( c39436dd,  0,  53704,       YM+TA), /* 04013 - Mad Max - Thalion Fever.sndh */
TIMEDB_ENTRY( c3af3bf5,  0,   7209,          YM), /* 03511 - Unknown - Cybernoid.sndh */
TIMEDB_ENTRY( c3af3bf5,  1,   2693,          YM), /* 03511 - Unknown - Cybernoid.sndh */
TIMEDB_ENTRY( c3c98efc,  0,   3600,          YM), /* 02377 - Nexus 6 - Headbomb.sndh */
TIMEDB_ENTRY( c3d0833a,  0,  11867,          YM), /* 02054 - Mad Max - Spellbound.sndh */
TIMEDB_ENTRY( c3fa1d2f,  0,   2880,          YM), /* 01515 - Illusion - Blutack.sndh */
TIMEDB_ENTRY( c3fd98f9,  0,   5802,          YM), /* 04386 - Paul Summers - Stryx.sndh */
TIMEDB_ENTRY( c404f606,  0,   8448,          YM), /* 00584 - Celtic - Collision - TFU Intro.sndh */
TIMEDB_ENTRY( c4267165,  0,   1785, YM+TA+TB+TD), /* 01959 - MSG - Godpey - Losing.sndh */
TIMEDB_ENTRY( c42df652,  0,   1080,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  1,    204,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  2,     40,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  3,   1536,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  4,    175,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  5,    268,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  6,   1408,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  7,     85,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  8,    166,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42df652,  9,   1108,          YM), /* 00752 - Mark Cooksey - Ghost N Goblins.sndh */
TIMEDB_ENTRY( c42e22ed,  0,   4224,          YM), /* 01888 - Dave Lowe - Night Shift.sndh */
TIMEDB_ENTRY( c42e22ed,  1,   8000,          YM), /* 01888 - Dave Lowe - Night Shift.sndh */
TIMEDB_ENTRY( c42e22ed,  2,    256,          YM), /* 01888 - Dave Lowe - Night Shift.sndh */
TIMEDB_ENTRY( c45c2d02,  0,   1920,          YM), /* 00815 - Crawdaddy - Tough Guys in Sweden.sndh */
TIMEDB_ENTRY( c49a99bb,  0,   3204,          YM), /* 03152 - Unknown - Believe It or Not!    Mancini.sndh */
TIMEDB_ENTRY( c49c7137,  0,  15755,          YM), /* 02012 - Mad Max - Nearly Spherical.sndh */
TIMEDB_ENTRY( c4a74563,  0,  11867, YM+TA+TB+TD), /* 02097 - Mad Max - Spellbound - Sid.sndh */
TIMEDB_ENTRY( c4b42ef2,  0,  25984,    YM+TA+TD), /* 02264 - Tao - Cream - Stars (Steps).sndh */
TIMEDB_ENTRY( c4c24475,  0,   4236,       YM+TA), /* 01628 - No More - TRSI.sndh */
TIMEDB_ENTRY( c4c5a4f7,  0,   1873,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  1,    749,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  2,    946,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  3,    889,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  4,   2373,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  5,    608,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  6,   1873,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  7,    312,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4c5a4f7,  8,    405,    YM+TA+TD), /* 01823 - Pierre-Eric Loriaux - Cabal.sndh */
TIMEDB_ENTRY( c4d14c90,  0,   9216,       YM+TA), /* 02331 - Tao - Cream (orig. Karsten Obarski) - Sleepwalk.sndh */
TIMEDB_ENTRY( c4d76130,  0,  26544, YM+TA+TB+TD), /* 03370 - STU - FUCK MACHINE.sndh */
TIMEDB_ENTRY( c4da11c2,  0,   2303,       YM+TD), /* 03291 - Dma-Sc - ATIPYK Developments - Manical Drop - 2 Players music #4.sndh */
TIMEDB_ENTRY( c50bb9dd,  0,   2109,       YM+TA), /* 04306 - 505-cpt^pdx - something.sndh */
TIMEDB_ENTRY( c51dee77,  0,  28032, YM+TA+TB+TD), /* 03982 - XyNo - BLIPSKILL.sndh */
TIMEDB_ENTRY( c51e0093,  0,   2293,    YM+TA+TB), /* 04289 - 505-cpt^pdx - Muell.sndh */
TIMEDB_ENTRY( c53fc73e,  0,  15744,          YM), /* 00597 - Chris Mad - Overdrive Dragon.sndh */
TIMEDB_ENTRY( c543754d,  0,   6945,          YM), /* 00140 - David Whittaker - Starblaze.sndh */
TIMEDB_ENTRY( c543754d,  1,  17604,          YM), /* 00140 - David Whittaker - Starblaze.sndh */
TIMEDB_ENTRY( c543754d,  2,     97,          YM), /* 00140 - David Whittaker - Starblaze.sndh */
TIMEDB_ENTRY( c543754d,  3,     97,          YM), /* 00140 - David Whittaker - Starblaze.sndh */
TIMEDB_ENTRY( c543754d,  4,    129,          YM), /* 00140 - David Whittaker - Starblaze.sndh */
TIMEDB_ENTRY( c552efdf,  0,    385,          YM), /* 00965 - Mike Davies - Last Duel.sndh */
TIMEDB_ENTRY( c552efdf,  1,  24577,          YM), /* 00965 - Mike Davies - Last Duel.sndh */
TIMEDB_ENTRY( c552efdf,  2,    481,          YM), /* 00965 - Mike Davies - Last Duel.sndh */
TIMEDB_ENTRY( c552efdf,  3,  10560,          YM), /* 00965 - Mike Davies - Last Duel.sndh */
TIMEDB_ENTRY( c56a1afe,  0,   4223, YM+TA+TB+TD), /* 02706 - Dubmood - Toffelskater Interlude.sndh */
TIMEDB_ENTRY( c56d9fa5,  0,  13600,       YM+TA), /* 04652 - Brain - Supremacy.sndh */
TIMEDB_ENTRY( c56de18c,  0,   1920,          YM), /* 02411 - Nexus 6 - Theme 1993.sndh */
TIMEDB_ENTRY( c587bd0e,  0,   4973,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  1,    101,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  2,    389,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  3,    208,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  4,    324,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  5,   2884,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c587bd0e,  6,   2051,          YM), /* 01738 - Richard Joseph - Knightmare.sndh */
TIMEDB_ENTRY( c594a311,  0,  19968,          YM), /* 01658 - Jedi - Stephanie.sndh */
TIMEDB_ENTRY( c5af9dd8,  0,   4389,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  1,   4399,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  2,   3003,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  3,   2934,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  4,   2568,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  5,   2184,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  6,   4368,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  7,   2688,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  8,   6048,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8,  9,    180,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 10,   2880,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 11,    453,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 12,    270,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 13,    192,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 14,    504,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 15,    972,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5af9dd8, 16,   2509,          YM), /* 00167 - Martin Walker - Dragon Breed.sndh */
TIMEDB_ENTRY( c5b5dc5e,  0,  12869,          YM), /* 02949 - Paul Robotham - Space 1889.sndh */
TIMEDB_ENTRY( c5c398c4,  0,   2688,          YM), /* 01725 - Jet Power - AGITRS - LOADER.sndh */
TIMEDB_ENTRY( c5c9741d,  0,  39169,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  1,  39711,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  2,  18412,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  3,  48482,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  4,  43265,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  5,  43265,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  6,  51005,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  7,  24549,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5c9741d,  8,  24129,       YM+TA), /* 03851 - The Archmage - MSI Sound Demo.sndh */
TIMEDB_ENTRY( c5d8d8b6,  0,   4098,          YM), /* 00295 - Gary Wheaton - British Anthem - West Indian Version.sndh */
TIMEDB_ENTRY( c5daac93,  0,   5903,          YM), /* 00345 - Tony Williams - Obitus.sndh */
TIMEDB_ENTRY( c5ef82c7,  0,   3453, YM+TA+TB+TD), /* 02871 - 505 - Zymplix.sndh */
TIMEDB_ENTRY( c630704d,  0,   5370,       YM+TA), /* 03555 - Stax - Shark.sndh */
TIMEDB_ENTRY( c66ec609,  0,   5785,          YM), /* 03015 - Skinnybone - Trex Warrior.sndh */
TIMEDB_ENTRY( c66ec609,  1,   1584,          YM), /* 03015 - Skinnybone - Trex Warrior.sndh */
TIMEDB_ENTRY( c66ec609,  2,  10752,          YM), /* 03015 - Skinnybone - Trex Warrior.sndh */
TIMEDB_ENTRY( c67289c8,  0,   5376,          YM), /* 01078 - Excellence In Art - Deserted.sndh */
TIMEDB_ENTRY( c6a2be86,  0,  12384,    YM+TA+TD), /* 01987 - Jochen Hippel aka Mad Max - Buzzer.sndh */
TIMEDB_ENTRY( c6aa274c,  0,   3450,          YM), /* 00884 - Adrian Dalecki - Clownomania.sndh */
TIMEDB_ENTRY( c6aa274c,  1,   6522,          YM), /* 00884 - Adrian Dalecki - Clownomania.sndh */
TIMEDB_ENTRY( c6b7c2bb,  0,   9218,          YM), /* 02165 - Mad Max - Children's Song.sndh */
TIMEDB_ENTRY( c6d142da,  0,  13433,    YM+TA+TD), /* 01329 - Scavenger - DBA 6 Intro.sndh */
TIMEDB_ENTRY( c6dbb53f,  0,   8321,          YM), /* 04158 - Unknown - Rumores.sndh */
TIMEDB_ENTRY( c6ea90a6,  0,    640,          YM), /* 01248 - Janko Mrsic Flogel - Fireblaster.sndh */
TIMEDB_ENTRY( c6ea90a6,  1,    640,          YM), /* 01248 - Janko Mrsic Flogel - Fireblaster.sndh */
TIMEDB_ENTRY( c6fc56df,  0,    860,      YM+STE), /* 01907 - Marcer - Marcer 79 - STE only.sndh */
TIMEDB_ENTRY( c6fd617f,  0,   9218,          YM), /* 02419 - Tangens - The 2nd Try - Multi Colour.sndh */
TIMEDB_ENTRY( c711e291,  0,  13275,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  1,     17,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  2,     49,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  3,     49,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  4,     38,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  5,     25,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  6,    126,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  7,     75,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  8,     37,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291,  9,     14,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291, 10,    129,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291, 11,    385,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291, 12,    161,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291, 13,    124,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c711e291, 14,    681,          YM), /* 02738 - Tony Williams & The Spoon - Nevermind.sndh */
TIMEDB_ENTRY( c7147a34,  0,   1685,          YM), /* 03169 - n-a - Stand by Me.sndh */
TIMEDB_ENTRY( c71a18ad,  0,   8735, YM+TA+TB+TD), /* 03396 - Ultrasyd - CORSAIR (mB cover).sndh */
TIMEDB_ENTRY( c71ef6ab,  0,   8703,          YM), /* 00309 - Gary Wheaton - Hey Hey, My My.sndh */
TIMEDB_ENTRY( c722956b,  0,   4782,          YM), /* 00013 - 4-Mat - Captain Dynamo.sndh */
TIMEDB_ENTRY( c722956b,  1,   2300,          YM), /* 00013 - 4-Mat - Captain Dynamo.sndh */
TIMEDB_ENTRY( c722956b,  2,   6080,          YM), /* 00013 - 4-Mat - Captain Dynamo.sndh */
TIMEDB_ENTRY( c7279020,  0,   4223, YM+TA+TB+TD), /* 02977 - 505-paradox - Attack.sndh */
TIMEDB_ENTRY( c73ad30c,  0,   6400,          YM), /* 04001 - Blue Nun - Gnurk Intro.sndh */
TIMEDB_ENTRY( c74e296a,  0,  10746,       YM+TD), /* 01620 - No More - GSC 57.sndh */
TIMEDB_ENTRY( c74f7013,  0,  13824, YM+TA+TB+TD), /* 02072 - Mad Max - Crazy Comets - Sid.sndh */
TIMEDB_ENTRY( c74f7013,  1,   3264, YM+TA+TB+TD), /* 02072 - Mad Max - Crazy Comets - Sid.sndh */
TIMEDB_ENTRY( c757f90e,  0,  53760, YM+TA+TB+TD), /* 04099 - mOdmate-CheckPoint (orig. Rob says  Wow!) - SILLY JET WILLY.sndh */
TIMEDB_ENTRY( c7670d6d,  0,   8359,          YM), /* 03562 - YERZMYEY-AY-RIDERS - STandard Crap.sndh */
TIMEDB_ENTRY( c78b1435,  0,   1923, YM+TA+TB+TD), /* 01166 - Tyan - Unten.sndh */
TIMEDB_ENTRY( c799aa1e,  0,   9600,          YM), /* 02804 - Starbuck - Mislead.sndh */
TIMEDB_ENTRY( c7a3bc19,  0,  23302,    YM+TA+TD), /* 02304 - Tao - Cream - Bumpsong (DHS-compo).sndh */
TIMEDB_ENTRY( c7a3fc91,  0,   6126,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  1,   3584,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  2,    610,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  3,    427,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  4,    512,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  5,    104,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  6,    105,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  7,    104,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  8,    107,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91,  9,    282,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91, 10,    282,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7a3fc91, 11,    512,          YM), /* 00111 - David Whittaker - Jupiter Master Drive.sndh */
TIMEDB_ENTRY( c7aa5a55,  0,  13816, YM+TA+TB+TD), /* 01602 - Nemo - Polarnoid.sndh */
TIMEDB_ENTRY( c7c162f8,  0,   3515,          YM), /* 04533 - Greg George - Four-F.sndh */
TIMEDB_ENTRY( c7c162f8,  1,    149,          YM), /* 04533 - Greg George - Four-F.sndh */
TIMEDB_ENTRY( c7c162f8,  2,    210,          YM), /* 04533 - Greg George - Four-F.sndh */
TIMEDB_ENTRY( c7c162f8,  3,    960,          YM), /* 04533 - Greg George - Four-F.sndh */
TIMEDB_ENTRY( c7c162f8,  4,   1660,          YM), /* 04533 - Greg George - Four-F.sndh */
TIMEDB_ENTRY( c7ca5444,  0,   6265,          YM), /* 04370 - Jess - Rock (draft).sndh */
TIMEDB_ENTRY( c7eb6d07,  0,   4320,          YM), /* 03302 - CRAZY Q - TWISTED LIFE SYLE.sndh */
TIMEDB_ENTRY( c7ec04eb,  0,   7679,    YM+TA+TB), /* 02698 - Dubmood (The Herbalizer) - Herbalizer Live Interlude.sndh */
TIMEDB_ENTRY( c7fc24fe,  0,   6138,    YM+TB+TD), /* 01021 - Dma Sc - Sector One - YM Racing (YM Rockerz).sndh */
TIMEDB_ENTRY( c80edc08,  0,   5281,          YM), /* 02135 - Mad Max - Human Race Level 1 - LCD.sndh */
TIMEDB_ENTRY( c81b26e2,  0,   8848,          YM), /* 01223 - ENS - ENS 6.sndh */
TIMEDB_ENTRY( c820c8a4,  0,  11022,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  1,    333,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  2,    289,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  3,   4216,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  4,   2688,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  5,   3696,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  6,   2688,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  7,   1536,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  8,   2048,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4,  9,   3584,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 10,   3072,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 11,   4608,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 12,   7296,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 13,   3200,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 14,   1578,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 15,    601,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 16,    501,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 17,   1152,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 18,   3840,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 19,    197,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c820c8a4, 20,   1920,          YM), /* 01563 - Jason Page - Fire and Ice.sndh */
TIMEDB_ENTRY( c834bd87,  0,   5372,          YM), /* 01240 - Epic - Toxic O Mania.sndh */
TIMEDB_ENTRY( c83fb5d0,  0,   6432,          YM), /* 00450 - Augustin Vidovic - Magic Force.sndh */
TIMEDB_ENTRY( c863fd43,  0,   2160,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  1,     48,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  2,    481,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  3,   1669,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  4,   4609,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  5,   2304,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  6,    361,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  7,    120,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c863fd43,  8,    109,          YM), /* 01730 - Peter Johnson - Robocop.sndh */
TIMEDB_ENTRY( c86f9022,  0,   9905, YM+TA+TB+TD), /* 04657 - Dma-Sc - Return of the Fighters.sndh */
TIMEDB_ENTRY( c87748f4,  0,  11195, YM+TA+TB+TD), /* 00997 - DMA-SC - Flying Over The Earth.sndh */
TIMEDB_ENTRY( c87dfa1f,  0,   4821,          YM), /* 00229 - ACC - King of the Road.sndh */
TIMEDB_ENTRY( c8859a00,  0,  59887,       YM+TA), /* 03771 - Exarch - Excited.sndh */
TIMEDB_ENTRY( c887513e,  0,  15360,          YM), /* 02800 - Starbuck - Her 7.sndh */
TIMEDB_ENTRY( c896a887,  0,  17271,    YM+TA+TD), /* 02562 - marcer - impossamol mix.sndh */
TIMEDB_ENTRY( c8a17152,  0,   3456,    YM+TA+TD), /* 02818 - Soulmate #1 - Znax ingame tune 1.sndh */
TIMEDB_ENTRY( c8ba0fce,  0,  14030,          YM), /* 03533 - Jean Baudlot - Dragon Ninja.sndh */
TIMEDB_ENTRY( c8ba0fce,  1,  11815,          YM), /* 03533 - Jean Baudlot - Dragon Ninja.sndh */
TIMEDB_ENTRY( c8ba0fce,  2,   2953,          YM), /* 03533 - Jean Baudlot - Dragon Ninja.sndh */
TIMEDB_ENTRY( c8ba0fce,  3,    738,          YM), /* 03533 - Jean Baudlot - Dragon Ninja.sndh */
TIMEDB_ENTRY( c8ba3168,  0,   4243,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  1,   3661,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  2,   4694,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  3,   1796,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  4,    821,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  5,   4685,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  6,   2077,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  7,   5245,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  8,   3233,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168,  9,   4243,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 10,   2117,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 11,   4458,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 12,    418,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 13,    581,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 14,    461,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 15,    389,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 16,   1134,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 17,   1015,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 18,    371,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 19,    773,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 20,    773,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 21,    789,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 22,   1067,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 23,    389,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 24,    659,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 25,    726,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 26,    541,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 27,    773,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8ba3168, 28,    420,          YM), /* 00881 - Ben Daglish - The Games  - Winter Edition.sndh */
TIMEDB_ENTRY( c8c2490b,  0,   6911,          YM), /* 02754 - Dma-Sc - Stars of the Night.sndh */
TIMEDB_ENTRY( c8caeb3d,  0,   6522,          YM), /* 04246 - 505-checkpoint - Bounce.sndh */
TIMEDB_ENTRY( c901f2fb,  0,   5375,          YM), /* 00021 - 4-Mat - Dizzy  Prince Of Yolkfolk.sndh */
TIMEDB_ENTRY( c901f2fb,  1,   1596,          YM), /* 00021 - 4-Mat - Dizzy  Prince Of Yolkfolk.sndh */
TIMEDB_ENTRY( c901f2fb,  2,   1151,          YM), /* 00021 - 4-Mat - Dizzy  Prince Of Yolkfolk.sndh */
TIMEDB_ENTRY( c90809a1,  0,   4224,          YM), /* 00241 - Alx 33 - Yng.sndh */
TIMEDB_ENTRY( c9118963,  0,   1729,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  1,   2112,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  2,   1536,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  3,    320,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  4,    780,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  5,    481,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  6,    889,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  7,    307,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  8,   1920,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c9118963,  9,     78,          YM), /* 00749 - Mark Cooksey - Commando.sndh */
TIMEDB_ENTRY( c91600e6,  0,   5392, YM+TA+TB+TD), /* 00197 - 505 - Sunny.sndh */
TIMEDB_ENTRY( c927db85,  0,  39370,       YM+TA), /* 04046 - Roggie - R.F.C. Ok.sndh */
TIMEDB_ENTRY( c92fb6cf,  0,   5567,          YM), /* 00335 - Gary Wheaton - Womble Song.sndh */
TIMEDB_ENTRY( c9376313,  0,  10092,    YM+TA+TD), /* 02513 - Crazy Q - Bouncing TITS.sndh */
TIMEDB_ENTRY( c940fc5b,  0,   5921,          YM), /* 03305 - O Cornelius & G Kennett - Enterprise.sndh */
TIMEDB_ENTRY( c940fc5b,  1,   1791,          YM), /* 03305 - O Cornelius & G Kennett - Enterprise.sndh */
TIMEDB_ENTRY( c94ae0a1,  0,   3836,       YM+TA), /* 00221 - 7AN - Blip 2.sndh */
TIMEDB_ENTRY( c96243ab,  0,   7167,    YM+TA+TB), /* 02837 - maktone-fairlight - husselhof.sndh */
TIMEDB_ENTRY( c9666046,  0,   7008,          YM), /* 04345 - Nexus 6 - Red Movement.sndh */
TIMEDB_ENTRY( c979c886,  0,   3064,          YM), /* 04320 - 505-cpt^pdx - Tell me.sndh */
TIMEDB_ENTRY( c98600ce,  0,   5668,    YM+TA+TD), /* 02658 - ear_dis - peagasm.sndh */
TIMEDB_ENTRY( c996d5e0,  0,  19201,          YM), /* 00398 - Wally Beben - Wicked.sndh */
TIMEDB_ENTRY( c996d5e0,  1,   6145,          YM), /* 00398 - Wally Beben - Wicked.sndh */
TIMEDB_ENTRY( c996d5e0,  2,   4097,          YM), /* 00398 - Wally Beben - Wicked.sndh */
TIMEDB_ENTRY( c996d5e0,  3,    408,          YM), /* 00398 - Wally Beben - Wicked.sndh */
TIMEDB_ENTRY( c996d5e0,  4,   3073,          YM), /* 00398 - Wally Beben - Wicked.sndh */
TIMEDB_ENTRY( c9ad41c0,  0,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  1,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  2,     30,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  4,     36,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  5,    108,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  6,    150,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  7,     63,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  8,     45,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0,  9,      9,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 10,   2880,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 11,     45,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 12,     66,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 13,    240,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 14,     63,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 15,     30,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 16,     39,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 17,    117,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 18,     99,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 19,     72,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 20,     12,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 21,    150,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 24,    240,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 25,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 26,      9,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 27,    156,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 28,     57,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 29,    240,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 30,     99,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 31,     90,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 33,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 34,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 35,     15,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 36,    180,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 37,     60,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 38,     12,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 39,     99,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 40,     12,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 41,      9,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 42,     18,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 43,     45,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 44,    240,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 45,     60,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 46,    117,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 47,    850,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9ad41c0, 49,    788,          YM), /* 04744 - Mark Crowe - Space Quest Chapter 1- The Sarien Encounter.sndh */
TIMEDB_ENTRY( c9bdaa9b,  0,  58395,       YM+TA), /* 04633 - TFC - Sunbeat.sndh */
TIMEDB_ENTRY( c9ccf33e,  0,   5568,          YM), /* 03236 - T1000 - The Beast Club - TBC 7.sndh */
TIMEDB_ENTRY( c9f17485,  0,   3836,          YM), /* 01862 - Lotek Style - Perfect Chipperman.sndh */
TIMEDB_ENTRY( c9f924b5,  0,  72963,          YM), /* 01808 - Barry Leitch - Super Cars 2.sndh */
TIMEDB_ENTRY( c9f924b5,  1, 112643,          YM), /* 01808 - Barry Leitch - Super Cars 2.sndh */
TIMEDB_ENTRY( ca048487,  0,  11916,       YM+TA), /* 01618 - No More - GSC 50.sndh */
TIMEDB_ENTRY( ca2ddb75,  0,   1536,          YM), /* 03239 - T1000 - The Beast Club - TBC 11.sndh */
TIMEDB_ENTRY( ca33de93,  0,  11662,          YM), /* 00502 - Tangens - The 2nd Try - Menu.sndh */
TIMEDB_ENTRY( ca3ec858,  0,  23936,    YM+TB+TD), /* 03989 - XyNo - NOT A 0110100101.sndh */
TIMEDB_ENTRY( ca525c6a,  0,   6325,          YM), /* 00446 - Augustin Vidovic - Facile.sndh */
TIMEDB_ENTRY( ca5a32ae,  0,   3456,          YM), /* 02646 - Miggy - Target (Unfinished).sndh */
TIMEDB_ENTRY( ca5e4691,  0,   1935,          YM), /* 04773 - BriteLite - Hammarsock.sndh */
TIMEDB_ENTRY( ca6a8733,  0,  16859, YM+TA+TB+TD), /* 04247 - 505 - BRN 2010.sndh */
TIMEDB_ENTRY( ca6c1df0,  0,   6528,          YM), /* 01769 - Lap - Lap 9.sndh */
TIMEDB_ENTRY( ca7a4bed,  0,   2131,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  1,    787,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  2,   7390,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  3,   7390,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  4,   6942,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  5,   7198,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  6,   7390,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  7,     66,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7a4bed,  9,    247,          YM), /* 01301 - Matt Furniss - Klax.sndh */
TIMEDB_ENTRY( ca7f6569,  0,   3057,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  1,     19,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  2,      9,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  3,   3560,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  4,     15,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  5,    103,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca7f6569,  6,     40,          YM), /* 04136 - Michael Haar - Der Wurm 2.sndh */
TIMEDB_ENTRY( ca847b86,  0,  27462,          YM), /* 01914 - Mate - Just Buggin - Xenex.sndh */
TIMEDB_ENTRY( ca87983a,  0, 144625,       YM+TA), /* 03810 - Martin Walker - Spindizzy Worlds.sndh */
TIMEDB_ENTRY( ca87983a,  1, 121746,       YM+TA), /* 03810 - Martin Walker - Spindizzy Worlds.sndh */
TIMEDB_ENTRY( ca87983a,  2,  44817,       YM+TA), /* 03810 - Martin Walker - Spindizzy Worlds.sndh */
TIMEDB_ENTRY( ca87983a,  3,  70042,       YM+TA), /* 03810 - Martin Walker - Spindizzy Worlds.sndh */
TIMEDB_ENTRY( caa63e4b,  0,  37632, YM+TA+TB+TD), /* 02543 - SCHMX & STU - TEMPLE OF SHID.sndh */
TIMEDB_ENTRY( cabca9c7,  0,   3840,          YM), /* 01132 - Tyan - Harm.sndh */
TIMEDB_ENTRY( cac3b1bb,  0,   2383,          YM), /* 03149 - n-a - Rainbows by Davidow.sndh */
TIMEDB_ENTRY( cac60fdd,  0,   2233,       YM+TA), /* 01547 - Redzone - TSC 65.sndh */
TIMEDB_ENTRY( cacb2888,  0,   4992,          YM), /* 01192 - Excellence In Art - New.sndh */
TIMEDB_ENTRY( cacd3efd,  0,   6528,    YM+TA+TD), /* 03188 - Crazy Q and dIRTY 120o - SEXGAMES.sndh */
TIMEDB_ENTRY( cad3c9c5,  0,    448,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  1,   1335,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  2,   2448,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  3,    400,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  4,    480,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  5,    139,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  6,    148,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  7,    256,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  8,     43,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5,  9,    112,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5, 10,    132,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cad3c9c5, 11,     58,          YM), /* 01063 - Jonathan Dunn - New Zealand Story.sndh */
TIMEDB_ENTRY( cade30a5,  0,   9215, YM+TA+TB+TD), /* 03848 - lotek style-.tscc.^ymr - approaching the endzeit.sndh */
TIMEDB_ENTRY( cae010b8,  0,   8719,          YM), /* 04150 - Unknown - GGGra125.sndh */
TIMEDB_ENTRY( caebabb8,  0,   5887,       YM+TD), /* 03226 - Peace <-> Oxyron - Oxyron Party v2.sndh */
TIMEDB_ENTRY( caf45579,  0,   9665,          YM), /* 03538 - Dave Lowe - Black Hornet.sndh */
TIMEDB_ENTRY( caf45579,  1,   3328,          YM), /* 03538 - Dave Lowe - Black Hornet.sndh */
TIMEDB_ENTRY( caf45579,  2,   3680,          YM), /* 03538 - Dave Lowe - Black Hornet.sndh */
TIMEDB_ENTRY( caf45579,  3,   2560,          YM), /* 03538 - Dave Lowe - Black Hornet.sndh */
TIMEDB_ENTRY( caf45579,  4,    113,          YM), /* 03538 - Dave Lowe - Black Hornet.sndh */
TIMEDB_ENTRY( cb0f5d99,  0,   8410,    YM+TA+TD), /* 03326 - CRAZY Q - PUNK ASS MF.sndh */
TIMEDB_ENTRY( cb1abe75,  0,   6144,          YM), /* 00893 - Tomas Danko - Danko.sndh */
TIMEDB_ENTRY( cb21aad7,  0,   9594,          YM), /* 00468 - Augustin Vidovic - Zvo.sndh */
TIMEDB_ENTRY( cb230ceb,  0,   4610,          YM), /* 02927 - Allister Brimble - Wacky Darts.sndh */
TIMEDB_ENTRY( cb230ceb,  1,    386,          YM), /* 02927 - Allister Brimble - Wacky Darts.sndh */
TIMEDB_ENTRY( cb277471,  0,   6912,          YM), /* 04831 - T1000 - Bits 55e.sndh */
TIMEDB_ENTRY( cb27c767,  0,  11292, YM+TA+TB+TD), /* 03665 - The Estate - Blackbird.sndh */
TIMEDB_ENTRY( cb45266b,  0,  14016,          YM), /* 02173 - Mad Max - Thundercats.sndh */
TIMEDB_ENTRY( cb484820,  0,   5376,       YM+TA), /* 00824 - !Cube - Shazam (Cross The Deserts).sndh */
TIMEDB_ENTRY( cb4da967,  0,   5759, YM+TA+TB+TD), /* 02839 - maktone-fairight - shotgun.sndh */
TIMEDB_ENTRY( cb67c425,  0,   8912,          YM), /* 00297 - Gary Wheaton - Belgium.sndh */
TIMEDB_ENTRY( cb72e4a4,  0,  19007, YM+TA+TB+TD), /* 03443 - ULTRASYD - LONESOME ROBOT.sndh */
TIMEDB_ENTRY( cb75192c,  0,  10848,    YM+TB+TD), /* 02855 - Helene - NONONO.sndh */
TIMEDB_ENTRY( cb75a000,  0,  11264,          YM), /* 00776 - Count Zero - Willy Preview.sndh */
TIMEDB_ENTRY( cb7fd9fa,  0,   7355,    YM+TB+TD), /* 01593 - Nemo - Excess.sndh */
TIMEDB_ENTRY( cb92ad8e,  0,   2685,          YM), /* 02468 - Iso - Juicy Alcohol.sndh */
TIMEDB_ENTRY( cb99d602,  0,   8239,          YM), /* 00376 - Beast - TSC - TSC 99.sndh */
TIMEDB_ENTRY( cb9d3083,  0,  17520,          YM), /* 04460 - Charles Callet - Bob Morane - Science Fiction.sndh */
TIMEDB_ENTRY( cba8b8b4,  0,  16507, YM+TA+TB+TD), /* 03275 - Dubmood - Kompis tog fram en fet skiva och korde.sndh */
TIMEDB_ENTRY( cbbfe050,  0,   2308,       YM+TA), /* 04377 - Dhor - Jumping squares.sndh */
TIMEDB_ENTRY( cbeedfb2,  0,   6144,          YM), /* 03497 - Shamen - Certified Dope - Intro.sndh */
TIMEDB_ENTRY( cc012723,  0,   5376,          YM), /* 00809 - Crawdaddy - Rings Of Medusa Mix.sndh */
TIMEDB_ENTRY( cc0267b2,  0,  12480,          YM), /* 00131 - David Whittaker - Return to Genesis.sndh */
TIMEDB_ENTRY( cc0267b2,  1,     81,          YM), /* 00131 - David Whittaker - Return to Genesis.sndh */
TIMEDB_ENTRY( cc0267b2,  2,    960,          YM), /* 00131 - David Whittaker - Return to Genesis.sndh */
TIMEDB_ENTRY( cc0267b2,  3,    256,          YM), /* 00131 - David Whittaker - Return to Genesis.sndh */
TIMEDB_ENTRY( cc1020eb,  0,   1199,          YM), /* 01670 - Jess - 970 Bytes.sndh */
TIMEDB_ENTRY( cc2b24cc,  0,   7429,          YM), /* 04257 - 505 - Checkpoint (2002) - Etudial.sndh */
TIMEDB_ENTRY( cc34ca47,  0,   7392,          YM), /* 02035 - Mad Max - Human Race.sndh */
TIMEDB_ENTRY( cc34ca47,  1,   3456,          YM), /* 02035 - Mad Max - Human Race.sndh */
TIMEDB_ENTRY( cc34ca47,  2,   3840,          YM), /* 02035 - Mad Max - Human Race.sndh */
TIMEDB_ENTRY( cc34ca47,  3,   8064,          YM), /* 02035 - Mad Max - Human Race.sndh */
TIMEDB_ENTRY( cc34ca47,  4,   5184,          YM), /* 02035 - Mad Max - Human Race.sndh */
TIMEDB_ENTRY( cc3d7ed0,  0,    867,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  1,    975,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  2,    741,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  3,   1155,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  4,    240,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  5,    153,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  6,   1407,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  7,    318,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc3d7ed0,  8,    195,          YM), /* 04732 - Al Lowe - The Black Cauldron.sndh */
TIMEDB_ENTRY( cc6dee92,  0,   5568, YM+TA+TB+TD), /* 00184 - 505 - Posh - Part 1.sndh */
TIMEDB_ENTRY( cc8053db,  0,   4392,          YM), /* 01801 - Barry Leitch - President Is Missing.sndh */
TIMEDB_ENTRY( cc8610c2,  0,  36864,       YM+TD), /* 01460 - STU - BOOMBAYA.sndh */
TIMEDB_ENTRY( ccbc07a9,  0,   8069,          YM), /* 02286 - Tao - Cream - State Of Confusion (Just Buggin).sndh */
TIMEDB_ENTRY( ccceb6ca,  0,  84102,       YM+TA), /* 03748 - Tufty - Kaos 1.sndh */
TIMEDB_ENTRY( ccdabfaf,  0,   1531,          YM), /* 04724 - Kevin J. Grantham - Zedoinoid.sndh */
TIMEDB_ENTRY( ccdabfaf,  1,    244,          YM), /* 04724 - Kevin J. Grantham - Zedoinoid.sndh */
TIMEDB_ENTRY( ccff96c0,  0,  30578,       YM+TA), /* 04798 - Unknown - Big in Japan.sndh */
TIMEDB_ENTRY( cd1ff097,  0,    280,          YM), /* 03075 - Nick Jones - Xevious.sndh */
TIMEDB_ENTRY( cd1ff097,  1,    312,          YM), /* 03075 - Nick Jones - Xevious.sndh */
TIMEDB_ENTRY( cd1ff097,  2,    254,          YM), /* 03075 - Nick Jones - Xevious.sndh */
TIMEDB_ENTRY( cd23e9ac,  0,   5100,          YM), /* 04611 - Richard Joseph - Demoniak.sndh */
TIMEDB_ENTRY( cd23e9ac,  1,     95,          YM), /* 04611 - Richard Joseph - Demoniak.sndh */
TIMEDB_ENTRY( cd5e2909,  0,    388,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  1,    377,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  2,   1300,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  3,    199,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  4,    480,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  5,    213,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  6,    193,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5e2909,  7,   1024,          YM), /* 02965 - Unknown - Noddy's Playtime.sndh */
TIMEDB_ENTRY( cd5fceb7,  0,   8696, YM+TA+TB+TD), /* 04262 - 505-cpt^pdx - Ground.sndh */
TIMEDB_ENTRY( cd6c584d,  0,   2951,       YM+TA), /* 03841 - Unknown - Into The Eagles Nest - Title.sndh */
TIMEDB_ENTRY( cd794ce5,  0,   8640,       YM+TA), /* 00787 - Crawdaddy - Electric Experience.sndh */
TIMEDB_ENTRY( cd7db35c,  0,   6910,          YM), /* 00371 - Beast - TSC - Beast New.sndh */
TIMEDB_ENTRY( cd832f48,  0,   1562,          YM), /* 03888 - Unknown - Kobyly.sndh */
TIMEDB_ENTRY( cd8ea7fd,  0,  23296, YM+TA+TB+TD), /* 03597 - NebulaH - Crontab.sndh */
TIMEDB_ENTRY( cd9fe1ea,  0, 135637,       YM+TA), /* 04049 - Roggie - Bag of tomato crisps.sndh */
TIMEDB_ENTRY( cda19cc4,  0,   5376,    YM+TA+TB), /* 01362 - Scavenger - Credits - Synergy Megademo.sndh */
TIMEDB_ENTRY( cda645e6,  0,  13248,          YM), /* 01782 - Lap - Yo Demo Main Menu.sndh */
TIMEDB_ENTRY( cdac174e,  0,   1686,          YM), /* 03931 - Unknown - Sterrei.sndh */
TIMEDB_ENTRY( cdb73ec6,  0,   2688,          YM), /* 01483 - Holger Gehrmann - Extensor.sndh */
TIMEDB_ENTRY( cdb73ec6,  1,   1792,          YM), /* 01483 - Holger Gehrmann - Extensor.sndh */
TIMEDB_ENTRY( cdb73ec6,  2,    308,          YM), /* 01483 - Holger Gehrmann - Extensor.sndh */
TIMEDB_ENTRY( cdb73ec6,  3,    224,          YM), /* 01483 - Holger Gehrmann - Extensor.sndh */
TIMEDB_ENTRY( cdb73ec6,  4,   2464,          YM), /* 01483 - Holger Gehrmann - Extensor.sndh */
TIMEDB_ENTRY( cdba023f,  0,    241,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdba023f,  1,     98,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdba023f,  2,   3201,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdba023f,  3,   1281,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdba023f,  4,   3073,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdba023f,  5,    147,          YM), /* 00381 - Wally Beben - Do Run Run.sndh */
TIMEDB_ENTRY( cdcf4d29,  0,   8064,       YM+TA), /* 00790 - Crawdaddy - Hardwired.sndh */
TIMEDB_ENTRY( cdd32584,  0,  18809,          YM), /* 01714 - Jess - Tears on my Keyboard.sndh */
TIMEDB_ENTRY( cddd5893,  0,    583,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  1,    217,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  2,   1159,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  3,    773,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  4,    389,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  5,    261,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  6,    391,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  7,    775,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  8,    517,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893,  9,   3081,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893, 10,   2053,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893, 11,    775,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893, 12,   1159,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893, 13,   1029,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cddd5893, 14,    901,          YM), /* 04006 - Timo Bauer - StarTrash.sndh */
TIMEDB_ENTRY( cdeb73dc,  0,  16228,    YM+TA+TD), /* 01574 - Jurgen Piscol - Oxxonian.sndh */
TIMEDB_ENTRY( cdeb73dc,  1,  15860,    YM+TA+TD), /* 01574 - Jurgen Piscol - Oxxonian.sndh */
TIMEDB_ENTRY( cdeb73dc,  2,  15807,    YM+TA+TD), /* 01574 - Jurgen Piscol - Oxxonian.sndh */
TIMEDB_ENTRY( cdeb73dc,  3,  15808,    YM+TA+TD), /* 01574 - Jurgen Piscol - Oxxonian.sndh */
TIMEDB_ENTRY( ce081658,  0,   1321,          YM), /* 03014 - Unknown - Wanderer 3D.sndh */
TIMEDB_ENTRY( ce081658,  1,   4603,          YM), /* 03014 - Unknown - Wanderer 3D.sndh */
TIMEDB_ENTRY( ce081658,  2,    263,          YM), /* 03014 - Unknown - Wanderer 3D.sndh */
TIMEDB_ENTRY( ce081658,  3,    894,          YM), /* 03014 - Unknown - Wanderer 3D.sndh */
TIMEDB_ENTRY( ce081658,  4,   1988,          YM), /* 03014 - Unknown - Wanderer 3D.sndh */
TIMEDB_ENTRY( ce10e412,  0,   1536,          YM), /* 01416 - Skyline - Snork Intro - Union.sndh */
TIMEDB_ENTRY( ce1d88e7,  0,  10362,       YM+TA), /* 01341 - Scavenger - Leif Rular - PYM.sndh */
TIMEDB_ENTRY( ce2943e0,  0,   7676,          YM), /* 01281 - Frazer - TSC 98.sndh */
TIMEDB_ENTRY( ce513a50,  0,  12864,          YM), /* 04463 - Charles Callet - Iznogoud.sndh */
TIMEDB_ENTRY( ce513a50,  1,   3392,          YM), /* 04463 - Charles Callet - Iznogoud.sndh */
TIMEDB_ENTRY( ce7065c1,  0,   8184, YM+TA+TB+TD), /* 01860 - Lotek Style - In my Heart.sndh */
TIMEDB_ENTRY( ce761369,  0,  37120,          YM), /* 01466 - STU - FLEX FLICKER.sndh */
TIMEDB_ENTRY( ce7b0f53,  0,  95553,       YM+TA), /* 03723 - Spaz - Ooh Crikey - Reset.sndh */
TIMEDB_ENTRY( ce7cf8c8,  0, 106349,       YM+TA), /* 03730 - Martin Walker - Hunter.sndh */
TIMEDB_ENTRY( ce91991c,  0,  38460,          YM), /* 04612 - Richard Joseph - Dragons Breath.sndh */
TIMEDB_ENTRY( ce91991c,  1,  45108,          YM), /* 04612 - Richard Joseph - Dragons Breath.sndh */
TIMEDB_ENTRY( ce91991c,  2,  12480,          YM), /* 04612 - Richard Joseph - Dragons Breath.sndh */
TIMEDB_ENTRY( cea96759,  0,  12112,          YM), /* 04442 - Iso - Checkmate.sndh */
TIMEDB_ENTRY( cec3ab5d,  0,  14601,          YM), /* 04138 - Unknown - Fantastic World.sndh */
TIMEDB_ENTRY( cec3ab5d,  1,   3724,          YM), /* 04138 - Unknown - Fantastic World.sndh */
TIMEDB_ENTRY( cec3ab5d,  2,   5026,          YM), /* 04138 - Unknown - Fantastic World.sndh */
TIMEDB_ENTRY( ced67746,  0,   4992,    YM+TA+TB), /* 00186 - 505 - Posh - Part 3.sndh */
TIMEDB_ENTRY( cee8f2ab,  0,  16849,    YM+TA+TD), /* 02300 - Tao - Cream - Quest (Alive 8).sndh */
TIMEDB_ENTRY( cef367b7,  0,   8320,          YM), /* 04602 - Rude Kid - Weirdo 2.sndh */
TIMEDB_ENTRY( cef628ae,  0,   1920,          YM), /* 00912 - Dark Angel - Ace 2.sndh */
TIMEDB_ENTRY( cef7431e,  0,   7870,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  1,   2239,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  2,   3070,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  3,    320,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  4,    269,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  5,   3583,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  6,   2302,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  7,   3135,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cef7431e,  8,   2522,          YM), /* 01314 - Matt Furniss - Sabre Team.sndh */
TIMEDB_ENTRY( cefd9efc,  0,   9720,    YM+TA+TB), /* 01366 - Scavenger - Synical.sndh */
TIMEDB_ENTRY( cf05d18a,  0,   5592,          YM), /* 00847 - Ben Daglish - Dark Fusion.sndh */
TIMEDB_ENTRY( cf1f811e,  0,   6155,          YM), /* 02005 - Mad Max - Dark Side Of The Spoon - TLB.sndh */
TIMEDB_ENTRY( cf1f811e,  1,    768,          YM), /* 02005 - Mad Max - Dark Side Of The Spoon - TLB.sndh */
TIMEDB_ENTRY( cf353947,  0,   8063,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  1,   6911,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  2,   2680,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  3,   1535,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  4,   1151,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  5,    761,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  6,   2687,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf353947,  7,    377,          YM), /* 01894 - Jason Sheppard - Venus the Flytrap.sndh */
TIMEDB_ENTRY( cf55417a,  0,   9722,          YM), /* 01275 - Frazer - TSC 36.sndh */
TIMEDB_ENTRY( cf56db36,  0,  10955, YM+TA+TB+TD), /* 03283 - 505-checkpoint (2003) - ahead.sndh */
TIMEDB_ENTRY( cf823424,  0,   1182,          YM), /* 03690 - Dma-Sc - Saboteur 3 - Introduction.sndh */
TIMEDB_ENTRY( cf9c3df9,  0,   3840,       YM+TA), /* 01750 - Kanewood - Paradise.sndh */
TIMEDB_ENTRY( cfb71ab0,  0,   2333,          YM), /* 03120 - Unknown - Hey Hey,My My - Neil Young.sndh */
TIMEDB_ENTRY( cfb75681,  0,  11134,       YM+TA), /* 02464 - Iso - Crazy Thing - DBA 9.sndh */
TIMEDB_ENTRY( cfcd90d1,  0,  20448, YM+TA+TB+TD), /* 02859 - Helene - tull.sndh */
TIMEDB_ENTRY( cfec4101,  0, 110212,       YM+TA), /* 03960 - Scott Lemmon - SRL4.sndh */
TIMEDB_ENTRY( cff247aa,  0,    762, YM+TA+TB+TD), /* 02424 - 505-cpt^dox - jingle01.sndh */
TIMEDB_ENTRY( d0094c23,  0,  23124,    YM+TB+TD), /* 03987 - XyNo - INSECURE.sndh */
TIMEDB_ENTRY( d05d8524,  0,  11904,          YM), /* 04598 - Rude Kid - Funeral.sndh */
TIMEDB_ENTRY( d064049d,  0,  14592,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  1,  10272,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  2,   6144,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  3,  17472,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  4,   8640,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  5,  11520,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d064049d,  6,  10080,          YM), /* 00509 - Tangens - Sythetic Visions - Hidden Screen.sndh */
TIMEDB_ENTRY( d0721025,  0,   3577,       YM+TA), /* 04229 - 505 - 7th.sndh */
TIMEDB_ENTRY( d08a23ab,  0,   3888,          YM), /* 00744 - Sean Conran - Megatwins.sndh */
TIMEDB_ENTRY( d08a23ab,  1,    384,          YM), /* 00744 - Sean Conran - Megatwins.sndh */
TIMEDB_ENTRY( d08a23ab,  2,    264,          YM), /* 00744 - Sean Conran - Megatwins.sndh */
TIMEDB_ENTRY( d08b8000,  0,  12288,          YM), /* 00483 - Ultimatum - Count Zero 3.sndh */
TIMEDB_ENTRY( d09368cb,  0,   9987,          YM), /* 00771 - Count Zero - Decade Demo - Reset.sndh */
TIMEDB_ENTRY( d09cce18,  0,   5120,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  1,   8914,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  2,   7812,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  3,   9995,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  4,   8643,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  5,  11911,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  6,   3712,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d09cce18,  7,   7296,       YM+TA), /* 02539 - Lotus-Hotline - Monster Business.sndh */
TIMEDB_ENTRY( d0a0478f,  0,  18431, YM+TA+TB+TD), /* 03441 - ULTRASYD - THE DAY YM KILLED FAKEBIT.sndh */
TIMEDB_ENTRY( d0a543c5,  0,   9985,    YM+TA+TD), /* 02299 - Tao - Cream - Still Alive (Alive 7).sndh */
TIMEDB_ENTRY( d0a77676,  0,   2564,          YM), /* 00291 - Gary Wheaton - Bandit.sndh */
TIMEDB_ENTRY( d0c202f0,  0,   7153,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  1,    205,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  2,    185,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  3,    348,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  4,    206,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  5,    374,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  6,   2821,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0c202f0,  7,   2188,          YM), /* 04385 - Paul Summers - Everton FC Intelligensia.sndh */
TIMEDB_ENTRY( d0d741c8,  0,   3840,          YM), /* 00931 - Dark Angel - Noisy Pillars.sndh */
TIMEDB_ENTRY( d0e4aa01,  0,  16032, YM+TA+TB+TD), /* 02088 - Mad Max - Master of Magic - Sid.sndh */
TIMEDB_ENTRY( d0e4aa01,  1,    575, YM+TA+TB+TD), /* 02088 - Mad Max - Master of Magic - Sid.sndh */
TIMEDB_ENTRY( d0e4aa01,  2,   1152, YM+TA+TB+TD), /* 02088 - Mad Max - Master of Magic - Sid.sndh */
TIMEDB_ENTRY( d0f55f18,  0,  42985, YM+TA+TB+TD), /* 03563 - TomChi  Noextra - CHIPVENTURE.sndh */
TIMEDB_ENTRY( d1054f32,  0, 107072,       YM+TA), /* 03767 - Waldemar Belwon - Power (Music Show 1).sndh */
TIMEDB_ENTRY( d1126a02,  0,   4416,          YM), /* 01768 - Lap - Lap 8.sndh */
TIMEDB_ENTRY( d119ec19,  0,   8250,          YM), /* 01234 - Epic - Epilsons' Nightmare.sndh */
TIMEDB_ENTRY( d12b6cdd,  0,   1440,          YM), /* 03227 - sh3 - loader rewind.sndh */
TIMEDB_ENTRY( d12bdbdb,  0,   3086,    YM+TA+TD), /* 02305 - Tao - Cream (orig. Rob Hubbard) - Commando.sndh */
TIMEDB_ENTRY( d130e95d,  0,   2731, YM+TA+TB+TD), /* 04265 - 505-cpt^pdx - hectic.sndh */
TIMEDB_ENTRY( d13fd641,  0, 278553,       YM+TA), /* 04673 - Unknown - I Just Cant Get Enough.sndh */
TIMEDB_ENTRY( d1460223,  0,   7723,       YM+TA), /* 01544 - Ralph - Ralph Digi 1.sndh */
TIMEDB_ENTRY( d158574e,  0,   2112,          YM), /* 00077 - David Whittaker - BMX Simulator.sndh */
TIMEDB_ENTRY( d158574e,  1,    225,          YM), /* 00077 - David Whittaker - BMX Simulator.sndh */
TIMEDB_ENTRY( d158574e,  2,    193,          YM), /* 00077 - David Whittaker - BMX Simulator.sndh */
TIMEDB_ENTRY( d15fed50,  0,   6176,          YM), /* 00327 - Gary Wheaton - Ram on.sndh */
TIMEDB_ENTRY( d192467c,  0,  63799,       YM+TA), /* 04624 - Andre Hollweg - Explorer.sndh */
TIMEDB_ENTRY( d1a32da4,  0,   6080,    YM+TA+TD), /* 02863 - marcer - souns of 1985-2006.sndh */
TIMEDB_ENTRY( d1a7dd4b,  0,   3119,          YM), /* 01292 - Matt Furniss - Carvup.sndh */
TIMEDB_ENTRY( d1a7dd4b,  1,   3199,          YM), /* 01292 - Matt Furniss - Carvup.sndh */
TIMEDB_ENTRY( d1a7dd4b,  2,   3199,          YM), /* 01292 - Matt Furniss - Carvup.sndh */
TIMEDB_ENTRY( d1a7dd4b,  3,   3359,          YM), /* 01292 - Matt Furniss - Carvup.sndh */
TIMEDB_ENTRY( d1b00941,  0,    691,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  1,    863,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  2,    623,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  3,    687,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  4,    887,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  5,    623,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  6,    431,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  7,    496,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  8,    604,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941,  9,    728,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 10,   1001,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 11,    614,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 12,    863,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 13,    623,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 14,    687,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 15,    887,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 16,    623,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 17,    431,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 18,    496,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 19,    614,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 20,    622,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1b00941, 21,    216,          YM), /* 04521 - Vivian Peschardt - Denver - Je decouvre les chiffres.sndh */
TIMEDB_ENTRY( d1c7fc9f,  0,   3855,          YM), /* 01322 - Matt Furniss - Toobin.sndh */
TIMEDB_ENTRY( d1c7fc9f,  1,   1928,          YM), /* 01322 - Matt Furniss - Toobin.sndh */
TIMEDB_ENTRY( d1c7fc9f,  2,   2460,          YM), /* 01322 - Matt Furniss - Toobin.sndh */
TIMEDB_ENTRY( d1c7fc9f,  3,   3087,          YM), /* 01322 - Matt Furniss - Toobin.sndh */
TIMEDB_ENTRY( d1c7fc9f,  4,   2247,          YM), /* 01322 - Matt Furniss - Toobin.sndh */
TIMEDB_ENTRY( d1cc615d,  0,   2563,       YM+TA), /* 01662 - Jedi - Virtual Escape Intro.sndh */
TIMEDB_ENTRY( d1dbd514,  0,   3920,          YM), /* 01201 - Excellence In Art - Spunge.sndh */
TIMEDB_ENTRY( d1e364b8,  0,   2688,          YM), /* 03653 - Tomchi - Pakakaka.sndh */
TIMEDB_ENTRY( d1e644cf,  0,  42311,       YM+TA), /* 03740 - Tufty - Acappella Choir.sndh */
TIMEDB_ENTRY( d1e68c94,  0,  13945,    YM+TA+TD), /* 01591 - Nemo ( 2003 ) - Double Danger Action Agents (YM Rockerz).sndh */
TIMEDB_ENTRY( d1e87fdb,  0,   7693,          YM), /* 01129 - Tyan - Revolution - Distomatic.sndh */
TIMEDB_ENTRY( d1ed13d1,  0,   5370,          YM), /* 01695 - Jess - For your Loader #1.sndh */
TIMEDB_ENTRY( d1ffb61d,  0,   7094, YM+TA+TB+TD), /* 00211 - 505 - Useless.sndh */
TIMEDB_ENTRY( d200eaae,  0,  31047, YM+TA+TB+TD), /* 03558 - Marcer - No Ecuador.sndh */
TIMEDB_ENTRY( d212a72f,  0,   7387,          YM), /* 00560 - Allister Brimble - Spellfire The Sorceror.sndh */
TIMEDB_ENTRY( d212a72f,  1,    379,          YM), /* 00560 - Allister Brimble - Spellfire The Sorceror.sndh */
TIMEDB_ENTRY( d212a72f,  2,     91,          YM), /* 00560 - Allister Brimble - Spellfire The Sorceror.sndh */
TIMEDB_ENTRY( d2224e7c,  0,  43201,       YM+TA), /* 04664 - Przybyszewski - Anthrax.sndh */
TIMEDB_ENTRY( d2473448,  0,   8057,       YM+TA), /* 01713 - Jess - My ST suntans in Brazil.sndh */
TIMEDB_ENTRY( d24de703,  0,   5957, YM+TA+TB+TD), /* 04483 - corpsicle - olympian.sndh */
TIMEDB_ENTRY( d25aaebf,  0,   9952,          YM), /* 04470 - Charles Callet - Les Ripoux.sndh */
TIMEDB_ENTRY( d268a213,  0,  11321,          YM), /* 04132 - Wally Beben - Superman - The Man of Steel.sndh */
TIMEDB_ENTRY( d268a213,  1,   1920,          YM), /* 04132 - Wally Beben - Superman - The Man of Steel.sndh */
TIMEDB_ENTRY( d268a213,  2,   5360,          YM), /* 04132 - Wally Beben - Superman - The Man of Steel.sndh */
TIMEDB_ENTRY( d26ab283,  0,   2095,          YM), /* 00312 - Gary Wheaton - Junk.sndh */
TIMEDB_ENTRY( d27f555c,  0,   8733,       YM+TA), /* 01526 - Iso - Hiya Placez.sndh */
TIMEDB_ENTRY( d2932e8e,  0,   6912,          YM), /* 01003 - DMA-SC - Mr Boomer.sndh */
TIMEDB_ENTRY( d29f8d1e,  0,    416,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  1,    256,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  2,    488,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  3,    512,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  4,    512,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  5,    248,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d29f8d1e,  6,    512,          YM), /* 01247 - Janko Mrsic Flogel - Bombjack.sndh */
TIMEDB_ENTRY( d2b42ac8,  0,  12201,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b42ac8,  1,  24575,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b42ac8,  2,   5119,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b42ac8,  3,   4607,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b42ac8,  4,   3692,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b42ac8,  5,   3459,          YM), /* 03530 - Jean Baudlot - Beach Volley.sndh */
TIMEDB_ENTRY( d2b46a8c,  0,   6713, YM+TA+TB+TD), /* 00277 - Baggio - Magmakt.sndh */
TIMEDB_ENTRY( d2c50ea8,  0,   8320, YM+TA+TB+TD), /* 03331 - Marcer - n-a.sndh */
TIMEDB_ENTRY( d2c9a559,  0,   6272,          YM), /* 00603 - Chris Mad - Overdrive Vector.sndh */
TIMEDB_ENTRY( d2caa33f,  0,   1275, YM+TA+TB+TD), /* 01950 - MSG - Chu Chu Rocket - Speed Up.sndh */
TIMEDB_ENTRY( d2cae2b6,  0,   8370,          YM), /* 03140 - n-a - Money for Nothing.sndh */
TIMEDB_ENTRY( d2df67b7,  0,   4992, YM+TA+TB+TD), /* 00185 - 505 - Posh - Part 2.sndh */
TIMEDB_ENTRY( d2e36b56,  0, 123011,       YM+TA), /* 03776 - Stewart Gardner - Stampede 5.sndh */
TIMEDB_ENTRY( d2edab52,  0,   9212, YM+TA+TB+TD), /* 04112 - 505 - Checkpoint - mysery.sndh */
TIMEDB_ENTRY( d2f3c6bb,  0,  10080,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  1,   1536,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  2,  10624,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  3,   2048,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  4,   1152,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  5,    154,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  6,    150,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  7,   9728,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  8,   2496,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb,  9,   9216,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 10,   1152,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 11,   2688,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 12,    171,          YM), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 13,    512,          YM), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 14,   6456,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 15,   1536,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 16,   5568,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 17,   1440,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 18,   1920,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 19,   1885,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 20,  10496,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 21,   2402,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 22,   6144,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 23,  15600,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 24,   2816,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 25,   1080,          YM), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 26,   2816,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 27,   5472,          YM), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2f3c6bb, 28,   2560,       YM+TA), /* 02221 - Mad Max - Turrican 2.sndh */
TIMEDB_ENTRY( d2fc8516,  0,  26731,    YM+TB+TD), /* 02443 - cerror - Taoism.sndh */
TIMEDB_ENTRY( d3098d04,  0,   4708,          YM), /* 00316 - Gary Wheaton - Lisa.sndh */
TIMEDB_ENTRY( d31c66cf,  0,   6403,       YM+TA), /* 02675 - Donald Fakk - Kvant Krull Kompott.sndh */
TIMEDB_ENTRY( d31de3bf,  0,   7530,          YM), /* 00012 - 4-Mat - Chuck Rock.sndh */
TIMEDB_ENTRY( d33eddd9,  0, 177228,       YM+TA), /* 03942 - Unknown - Adamski 7.sndh */
TIMEDB_ENTRY( d3473ca6,  0,   3937,          YM), /* 04151 - Unknown - Infanzia.sndh */
TIMEDB_ENTRY( d362e9ae,  0,   8448,          YM), /* 01077 - Excellence In Art - Concert.sndh */
TIMEDB_ENTRY( d366b673,  0,   9216,    YM+TA+TD), /* 02671 - Donald Fakk - Human Race - tune 4.sndh */
TIMEDB_ENTRY( d3802740,  0,    889,    YM+TA+TD), /* 04282 - 505 - Loop06.sndh */
TIMEDB_ENTRY( d384127f,  0,    752,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  1,    872,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  2,    864,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  3,    940,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  4,    512,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  5,    366,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  6,     64,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  7,    149,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  8,   1600,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f,  9,    845,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f, 10,    845,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f, 11,    845,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d384127f, 12,    533,          YM), /* 04702 - Unknown - Nova Cam.sndh */
TIMEDB_ENTRY( d38512c2,  0,    775,          YM), /* 03911 - Unknown - D1 MLike.sndh */
TIMEDB_ENTRY( d3a1a667,  0,  22528,       YM+TD), /* 02605 - STU - MUSHROOM.sndh */
TIMEDB_ENTRY( d3a23289,  0,   2592,          YM), /* 00575 - Jas C Brooke - Sidearms.sndh */
TIMEDB_ENTRY( d3a23289,  1,    512,          YM), /* 00575 - Jas C Brooke - Sidearms.sndh */
TIMEDB_ENTRY( d3a23289,  2,    288,          YM), /* 00575 - Jas C Brooke - Sidearms.sndh */
TIMEDB_ENTRY( d3ab72da,  0,   6144,          YM), /* 04829 - T1000 - Bits 55b.sndh */
TIMEDB_ENTRY( d3b0e310,  0,  28736, YM+TA+TB+TD), /* 03468 - Marcer - YM Out.sndh */
TIMEDB_ENTRY( d3b4c932,  0,  46416,       YM+TA), /* 03792 - Roggie - Minuet 2.sndh */
TIMEDB_ENTRY( d3b9562f,  0,   6272,       YM+TA), /* 01710 - Jess - A 60's Slow.sndh */
TIMEDB_ENTRY( d3c2d60e,  0,    649,          YM), /* 01906 - Marcer - Hu.sndh */
TIMEDB_ENTRY( d3d07c52,  0,   3073,          YM), /* 03890 - Unknown - Menuett.sndh */
TIMEDB_ENTRY( d3e32337,  0,   4602, YM+TA+TB+TD), /* 02824 - Msg - Superfly tune 1.sndh */
TIMEDB_ENTRY( d3ec14f9,  0,   2752,          YM), /* 01249 - Janko Mrsic Flogel - Metropolis.sndh */
TIMEDB_ENTRY( d3ec14f9,  1,   1024,          YM), /* 01249 - Janko Mrsic Flogel - Metropolis.sndh */
TIMEDB_ENTRY( d3f3b079,  0,   3584,          YM), /* 03628 - Unknown - Sommarhack 2011 Unknown.sndh */
TIMEDB_ENTRY( d3fccfc6,  0,   3936,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  1,   4416,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  2,   4416,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  3,   3520,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  4,   3072,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  5,   2048,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  6,   7093,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fccfc6,  7,    660,          YM), /* 01876 - Dave Lowe - Altered Beast.sndh */
TIMEDB_ENTRY( d3fee8a0,  0,    813,          YM), /* 03923 - Unknown - Musik.sndh */
TIMEDB_ENTRY( d406f0ef,  0,   7120,          YM), /* 01266 - Frazer - Frantic.sndh */
TIMEDB_ENTRY( d40bc48a,  0,  11898,    YM+TB+TD), /* 03520 - Nemo - DHS - Aqua Polka.sndh */
TIMEDB_ENTRY( d4211ea2,  0,   6912, YM+TA+TB+TD), /* 00070 - 505 - Lala.sndh */
TIMEDB_ENTRY( d437770c,  0,  11272,    YM+TA+TD), /* 02262 - Tao - Cream - Ride The Sky (Steps).sndh */
TIMEDB_ENTRY( d437eec6,  0,   1914,       YM+TA), /* 01343 - Scavenger - Let's Talk About Max.sndh */
TIMEDB_ENTRY( d4401668,  0,  19074, YM+TA+TC+TD), /* 04367 - Purple Motion (ST adaption by Dma-Sc) - Second Reality 2013 main.sndh */
TIMEDB_ENTRY( d4404f79,  0,  13228,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  1,   8839,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  2,  17144,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  3,  11769,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  4,  12976,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  5,  14890,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4404f79,  6,  12948,          YM), /* 04398 - Paul Summers - Rainbow Warriors.sndh */
TIMEDB_ENTRY( d4591664,  0,   4608,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d4591664,  1,    512,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d4591664,  2,    512,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d4591664,  3,   2304,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d4591664,  4,   1280,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d4591664,  5,    512,          YM), /* 00974 - Chris Denman - Airbucks.sndh */
TIMEDB_ENTRY( d45eefd3,  0,  25984,    YM+TA+TD), /* 02333 - Tao - Cream - Stars (Stmygm2).sndh */
TIMEDB_ENTRY( d479e1e3,  0,  38755,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  1,   4612,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  2,   6147,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  3,   5123,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  4,   6147,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  5,   7683,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  6,   4611,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  7,   4611,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  8,   4611,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3,  9,   6147,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 10,   6147,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 11,  29284,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 12,  28675,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 13,  35475,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 14,  40323,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 15,  35331,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 16,  30820,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 17,  28803,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 18,  33027,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 19,  27651,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 20,  61443,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d479e1e3, 21,   4612,          YM), /* 01812 - Barry Leitch & Sean Connolly - Prophecy I - Viking Child.sndh */
TIMEDB_ENTRY( d4842722,  0,  12800,          YM), /* 02048 - Mad Max - Phantom Of The Asteroid.sndh */
TIMEDB_ENTRY( d4842722,  1,   6912,          YM), /* 02048 - Mad Max - Phantom Of The Asteroid.sndh */
TIMEDB_ENTRY( d48c913d,  0,   8704,          YM), /* 00014 - 4-Mat - G-Loc.sndh */
TIMEDB_ENTRY( d4985978,  0,   4992,          YM), /* 01199 - Excellence In Art - Smile.sndh */
TIMEDB_ENTRY( d4a595ef,  0,   7700,    YM+TA+TD), /* 04782 - gwEm - Trans-Debug Express (orig edit).sndh */
TIMEDB_ENTRY( d4c06138,  0,   6265, YM+TA+TB+TD), /* 00282 - Baggio - Finale.sndh */
TIMEDB_ENTRY( d4c07d72,  0,  10158,          YM), /* 00759 - Count Zero - Count Zero Again - Maggie 8.sndh */
TIMEDB_ENTRY( d4d70b62,  0,   3084,          YM), /* 03860 - Dma-Sc - Skit 4kb.sndh */
TIMEDB_ENTRY( d4f18263,  0,  10027,          YM), /* 01539 - Ralph - Ralph 7.sndh */
TIMEDB_ENTRY( d50127c9,  0,   6153,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  1,   8320,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  2,   9344,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  3,  16896,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  4,   1536,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  5,  12544,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  6,   3968,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  7,   6144,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  8,  17676,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d50127c9,  9,   2096,          YM), /* 01495 - Alex Herbert - Bombz.sndh */
TIMEDB_ENTRY( d5060e28,  0,   6857,          YM), /* 01302 - Matt Furniss - Lords Of Chaos.sndh */
TIMEDB_ENTRY( d514da56,  0,    378,          YM), /* 01911 - Marcer - Ojoj.sndh */
TIMEDB_ENTRY( d52ed309,  0,  13056,          YM), /* 01653 - Jedi - Nuisance.sndh */
TIMEDB_ENTRY( d5309031,  0,   3072,       YM+TA), /* 00792 - Crawdaddy - Indian.sndh */
TIMEDB_ENTRY( d530f65e,  0,   1532,       YM+TA), /* 04240 - 505 - Anfang.sndh */
TIMEDB_ENTRY( d535ed42,  0,   7290,          YM), /* 00458 - Augustin Vidovic - Punish Your Machine Loader.sndh */
TIMEDB_ENTRY( d536d94e,  0,    261,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  1,    279,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  2,    262,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  3,    265,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  4,    179,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  5,    191,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  6,    215,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  7,    229,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  8,    234,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e,  9,    239,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 10,    285,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 11,    230,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 12,    411,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 13,    265,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 14,    239,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d536d94e, 15,   1076,          YM), /* 04693 - Dan Sharpe - Makers match.sndh */
TIMEDB_ENTRY( d5830c35,  0,  22527, YM+TA+TB+TD), /* 03003 - 505-paradox - upbuzzer.sndh */
TIMEDB_ENTRY( d5840c66,  0,   2328, YM+TA+TB+TD), /* 04264 - 505-cpt^pdx - Hall of Fame.sndh */
TIMEDB_ENTRY( d5885164,  0,   7035,    YM+TB+TD), /* 01599 - Nemo - Mischief.sndh */
TIMEDB_ENTRY( d5975740,  0,  10944,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  1,   3971,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  2,   3456,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  3,   4736,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  4,   7392,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  5,    982,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  6,    288,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5975740,  7,   4608,          YM), /* 02223 - Mad Max - Warp.sndh */
TIMEDB_ENTRY( d5a9e389,  0,   1920,          YM), /* 02149 - Mad Max - Mad Max Tune 4 - LSD.sndh */
TIMEDB_ENTRY( d5e45f05,  0,  38418, YM+TA+TB+TD), /* 02748 - Marcer - BREATH OUTLINE.sndh */
TIMEDB_ENTRY( d603c7a0,  0,    441,       YM+TA), /* 00828 - Crazy Q -- TiMeBoMb - - Mono minded shit!.sndh */
TIMEDB_ENTRY( d60fdf9d,  0,   5760,          YM), /* 01162 - Tyan - Tyan 1.sndh */
TIMEDB_ENTRY( d6246ebb,  0,   6522,    YM+TA+TB), /* 00903 - Danny Oneway - Hidden 2.sndh */
TIMEDB_ENTRY( d626f1cf,  0,  20480,    YM+TB+TD), /* 03463 - Marcer - Paradize.sndh */
TIMEDB_ENTRY( d62c56ca,  0,   1294,          YM), /* 04606 - White Ninja - Zoolookologic.sndh */
TIMEDB_ENTRY( d6571e3c,  0,   2978,          YM), /* 03036 - Ed Bogas - Ballad.sndh */
TIMEDB_ENTRY( d661fadf,  0,  47056,    YM+TA+TD), /* 03414 - Tao - Cream (orig. Lotek Style-tSCc) - Sweety (Main Tune).sndh */
TIMEDB_ENTRY( d668a0b5,  0,  60930,       YM+TA), /* 03736 - David Whittaker - Supremacy.sndh */
TIMEDB_ENTRY( d66b346f,  0,  11138,          YM), /* 00429 - Big Alec - PYM - Sickest So Far.sndh */
TIMEDB_ENTRY( d68e33ce,  0,  15375, YM+TA+TB+TD), /* 03270 - Dubmood - Still DRE Alla Man borde ringt live vers.sndh */
TIMEDB_ENTRY( d69b35bd,  0,  69520,       YM+TA), /* 03704 - Sparky - Dogs Life - For PP Sledgehammer.sndh */
TIMEDB_ENTRY( d6a5a6d7,  0,  12545,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  1,  22802,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  2,  36183,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  3,   6145,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  4,   9271,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  5,  84099,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  6,  40925,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  7,  36902,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  8,   8577,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6a5a6d7,  9,   7237,       YM+TA), /* 04392 - Golden Otter - Hibernation Carol Demo.sndh */
TIMEDB_ENTRY( d6af7069,  0,   6272,          YM), /* 01188 - Excellence In Art - Kitten.sndh */
TIMEDB_ENTRY( d6d5d0d5,  0,   3649, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6d5d0d5,  1,    418, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6d5d0d5,  2,    193, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6d5d0d5,  3,    193, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6d5d0d5,  4,   1750, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6d5d0d5,  5,     96, YM+TA+TB+TD), /* 02094 - Mad Max - Sam Fox Strip Poker - Sid.sndh */
TIMEDB_ENTRY( d6e2e254,  0, 142913,       YM+TA), /* 03836 - Martin Walker - Devious Designs Digi.sndh */
TIMEDB_ENTRY( d6e2e254,  1,  58788,       YM+TA), /* 03836 - Martin Walker - Devious Designs Digi.sndh */
TIMEDB_ENTRY( d6e994aa,  0,   2503,       YM+TA), /* 04421 - EVE - Stack Row.sndh */
TIMEDB_ENTRY( d6fc04a8,  0,  12993,    YM+TA+TD), /* 02296 - Tao - Cream - Thing (Alive 4).sndh */
TIMEDB_ENTRY( d71de918,  0,  20616, YM+TA+TB+TD), /* 03345 - Marcer - SORN DESTINY.sndh */
TIMEDB_ENTRY( d72ee71f,  0,   3410,          YM), /* 02143 - Mad Max - Flash Gordon - LSD.sndh */
TIMEDB_ENTRY( d730c8db,  0,  13254,          YM), /* 01424 - Spector - Manic Demo - 3D Balls.sndh */
TIMEDB_ENTRY( d745b1fa,  0,   1100,          YM), /* 03684 - Dma-Sc - Saboteur 3 - Crew Logo.sndh */
TIMEDB_ENTRY( d762ac76,  0,  11744,          YM), /* 01877 - Dave Lowe - Bangkok Knights.sndh */
TIMEDB_ENTRY( d762ac76,  1,   1024,          YM), /* 01877 - Dave Lowe - Bangkok Knights.sndh */
TIMEDB_ENTRY( d762ac76,  2,  10624,          YM), /* 01877 - Dave Lowe - Bangkok Knights.sndh */
TIMEDB_ENTRY( d7760970,  0,  10745, YM+TA+TB+TD), /* 01944 - MSG - Chu Chu Rocket - Credits.sndh */
TIMEDB_ENTRY( d78b99f6,  0,   6175, YM+TA+TB+TD), /* 02732 - maktone-fairlight - shaman.sndh */
TIMEDB_ENTRY( d7a32db1,  0,   2448,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  1,   2688,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  2,    235,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  3,    208,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  4,   1056,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  5,   2688,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  6,    378,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  7,    144,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a32db1,  8,   1843,          YM), /* 04387 - Paul Summers - Fighting Soccer.sndh */
TIMEDB_ENTRY( d7a88c17,  0, 121743,       YM+TA), /* 04362 - X The Beermat - Yours.sndh */
TIMEDB_ENTRY( d7cb3159,  0,   4032,       YM+TA), /* 00817 - Crawdaddy - Walkman.sndh */
TIMEDB_ENTRY( d7ce8025,  0,    768,       YM+TA), /* 03864 - Mr Coke - Chasey Loader.sndh */
TIMEDB_ENTRY( d7d8b1ba,  0,  81053,       YM+TA), /* 03782 - Max Headroom - Aquanaut.sndh */
TIMEDB_ENTRY( d7dc1900,  0,   2687, YM+TA+TB+TD), /* 02730 - maktone-fairlight - new start.sndh */
TIMEDB_ENTRY( d7e95dae,  0,   5568, YM+TA+TB+TD), /* 02853 - Helene - hmmmm.sndh */
TIMEDB_ENTRY( d7f5df4f,  0,   8413, YM+TA+TB+TD), /* 02096 - Mad Max - Sanxion Title - Sid.sndh */
TIMEDB_ENTRY( d7fbc1cf,  0,   7161, YM+TA+TB+TD), /* 01967 - MSG - Godpey - Story 2.sndh */
TIMEDB_ENTRY( d7fc2a5e,  0, 123873,       YM+TA), /* 04026 - Roggie - Echoes.sndh */
TIMEDB_ENTRY( d7fc628c,  0,  28896, YM+TA+TB+TD), /* 02910 - Helene - Temp.sndh */
TIMEDB_ENTRY( d815cdbe,  0,   2684,          YM), /* 02434 - Beast - TSC - Evert.sndh */
TIMEDB_ENTRY( d816e5e6,  0,   1920,          YM), /* 01088 - Techno - Dentro 3.sndh */
TIMEDB_ENTRY( d81df61c,  0,  24024,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  1,  21976,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  2,  24534,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  3,  24021,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  4,  28120,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  5,  24664,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  6,  10972,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  7,   9812,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  8,   2692,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c,  9,  11095,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 10,  16937,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 11,  17186,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 12,   3138,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 13,   5678,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 14,   7613,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 15,   2597,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81df61c, 16,   7578,          YM), /* 01326 - Scavenger - Crapman.sndh */
TIMEDB_ENTRY( d81efd02,  0,   5698,          YM), /* 00503 - Tangens - The 2nd Try - Star Demo.sndh */
TIMEDB_ENTRY( d8240c6e,  0,   5187,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  1,   4352,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  2,    960,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  3,    103,          YM), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  4,   6480,          YM), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  5,   6656,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  6,   6740,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  7,   1409,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  8,     85,          YM), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e,  9,   2592,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 10,   5376,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 11,   7520,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 12,   7488,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 13,   4176,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 14,   1760,       YM+TA), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 15,     91,          YM), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d8240c6e, 16,   7424,          YM), /* 02222 - Mad Max - Turrican.sndh */
TIMEDB_ENTRY( d84e4888,  0,   7576, YM+TA+TB+TD), /* 00198 - 505 - Surprise.sndh */
TIMEDB_ENTRY( d85d8e21,  0,   3200,    YM+TA+TD), /* 02767 - Bobby Prince [ST by Dma-Sc] - Wondering About My Loved....sndh */
TIMEDB_ENTRY( d86e3d04,  0,   8255,          YM), /* 03446 - 505-paradox - Cubes.sndh */
TIMEDB_ENTRY( d8867f29,  0,   3640,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  1,   2160,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  2,   1922,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  3,    125,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  4,     68,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  5,     68,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d8867f29,  6,    125,          YM), /* 00961 - Mike Davies - Die Hard 2.sndh */
TIMEDB_ENTRY( d888616e,  0,   5795,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  1,   2112,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  2,   3840,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  3,   6528,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  4,   5952,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  5,   3072,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  6,   4560,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d888616e,  7,   6720,          YM), /* 00939 - Dark Angel - Music Mon Tunes.sndh */
TIMEDB_ENTRY( d88d14fc,  0,  12282, YM+TA+TB+TD), /* 01588 - Nemo - Beaucoup Sun.sndh */
TIMEDB_ENTRY( d89721cc,  0,     65,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  1,     51,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  2,     51,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  3,     32,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  4,     35,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  5,     38,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  6,      9,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  7,     36,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  8,     26,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc,  9,     26,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 10,    575,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 11,     21,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 12,     36,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 13,     53,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 14,     46,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 15,     62,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 16,     23,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 17,     23,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 18,     23,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 19,     22,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 20,      9,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 21,   4224,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 22,  12160,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d89721cc, 23,   6144,          YM), /* 00774 - Count Zero - Killing Game Show.sndh */
TIMEDB_ENTRY( d8b66be0,  0, 137838,       YM+TA), /* 04338 - Rob Povey - Quartet 1.0 demo tunes.sndh */
TIMEDB_ENTRY( d8b66be0,  1,  39858,       YM+TA), /* 04338 - Rob Povey - Quartet 1.0 demo tunes.sndh */
TIMEDB_ENTRY( d8b66be0,  2,  59664,       YM+TA), /* 04338 - Rob Povey - Quartet 1.0 demo tunes.sndh */
TIMEDB_ENTRY( d8b66be0,  3,  48072,       YM+TA), /* 04338 - Rob Povey - Quartet 1.0 demo tunes.sndh */
TIMEDB_ENTRY( d8b66be0,  4,  44029,       YM+TA), /* 04338 - Rob Povey - Quartet 1.0 demo tunes.sndh */
TIMEDB_ENTRY( d8c67ef6,  0,  19008,          YM), /* 02119 - Mad Max - Bangkok Nights.sndh */
TIMEDB_ENTRY( d8cc7658,  0,   6912,          YM), /* 04826 - T1000 - Bits 45.sndh */
TIMEDB_ENTRY( d8d3be34,  0,   1363,          YM), /* 01900 - Luebke - IRC Simulator.sndh */
TIMEDB_ENTRY( d8ee76ed,  0,   4223,    YM+TA+TD), /* 02498 - gwEm - U.L.S. Loader.sndh */
TIMEDB_ENTRY( d8fd86e3,  0,   6914,          YM), /* 01043 - Dr Beat - Imagination - Enola Gay.sndh */
TIMEDB_ENTRY( d90609e2,  0,  58112, YM+TA+TB+TD), /* 02663 - STU - MEGABLAST.sndh */
TIMEDB_ENTRY( d920221b,  0,   2189,          YM), /* 00963 - Mike Davies - England Championship Special.sndh */
TIMEDB_ENTRY( d9255810,  0,   2177,          YM), /* 00499 - Tangens - The 2nd Try - Doc Screen.sndh */
TIMEDB_ENTRY( d92d0991,  0,   1341,          YM), /* 02467 - Iso - Introtune.sndh */
TIMEDB_ENTRY( d9409c35,  0,   7487, YM+TA+TB+TD), /* 03018 - Dubmood - Zabutoms Bice4.sndh */
TIMEDB_ENTRY( d966d13d,  0,  13951, YM+TA+TB+TD), /* 02525 - dubmood Marcer & Nightbeat - Muslichip.sndh */
TIMEDB_ENTRY( d976aeb0,  0,  17017,    YM+TA+TD), /* 01137 - Tyan - Julia 1.sndh */
TIMEDB_ENTRY( d97cf52b,  0,  19008,    YM+TA+TD), /* 02252 - Tao - Cream - Cultivation (Steps).sndh */
TIMEDB_ENTRY( d98d6e29,  0,  85811,       YM+TA), /* 03709 - Sparky - Sledge - For PP Sledgehammer.sndh */
TIMEDB_ENTRY( d9989e36,  0,   6720, YM+TA+TB+TD), /* 00985 - D-Force - Summer.sndh */
TIMEDB_ENTRY( d9b6a11b,  0,   3645, YM+TA+TB+TD), /* 04293 - 505 - Checkpoint - noir.sndh */
TIMEDB_ENTRY( d9d16c83,  0,   6138,    YM+TA+TD), /* 01141 - Tyan - Julia 5.sndh */
TIMEDB_ENTRY( d9e3f9db,  0,   7110,          YM), /* 03010 - Fred Gray - Star Goose.sndh */
TIMEDB_ENTRY( d9e3f9db,  2,    447,          YM), /* 03010 - Fred Gray - Star Goose.sndh */
TIMEDB_ENTRY( d9e49cb5,  0,   2307,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  1,   3145,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  2,   4403,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  3,    579,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  4,    643,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  5,   4035,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  6,    291,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  7,   7858,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  8,   1200,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5,  9,   4395,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 10,    579,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 11,   3843,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 12,   2115,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 13,   3459,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 14,    835,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 15,    451,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 16,   1155,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 17,   1155,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 18,   3201,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9e49cb5, 19,   1539,          YM), /* 01793 - Barry Leitch - Gemini Wing.sndh */
TIMEDB_ENTRY( d9f1537f,  0,  10656,          YM), /* 02153 - Mad Max - Ooh Crikey - No Sheep.sndh */
TIMEDB_ENTRY( d9f719cd,  0,  14604,       YM+TA), /* 02682 - Donald Fakk - Steam Dream.sndh */
TIMEDB_ENTRY( da11e462,  0,  21928, YM+TA+TB+TD), /* 03464 - Marcer - Popcorn.sndh */
TIMEDB_ENTRY( da202c20,  0,   1631,       YM+TD), /* 03470 - gwEm (Mad Max conversion) - Wings of Death #6.sndh */
TIMEDB_ENTRY( da21a346,  0,  10362,    YM+TA+TD), /* 02479 - Floopy - Cassiop e.sndh */
TIMEDB_ENTRY( da22a0bb,  0,   6656,          YM), /* 03658 - Jedi - DNT Paper 4 intro - Song 2.sndh */
TIMEDB_ENTRY( da303bc9,  0,   4991,          YM), /* 01309 - Matt Furniss - Pitfighter.sndh */
TIMEDB_ENTRY( da382412,  0,   8000,          YM), /* 04842 - t1000 - Bits 37.sndh */
TIMEDB_ENTRY( da390a6f,  0,   9309, YM+TA+TB+TD), /* 03580 - Divag - Drum n Buzz.sndh */
TIMEDB_ENTRY( da512160,  0,   2420,          YM), /* 03086 - Unknown - Metro Cross.sndh */
TIMEDB_ENTRY( da512160,  1,  15367,          YM), /* 03086 - Unknown - Metro Cross.sndh */
TIMEDB_ENTRY( da512160,  2,    619,          YM), /* 03086 - Unknown - Metro Cross.sndh */
TIMEDB_ENTRY( da512160,  3,    299,          YM), /* 03086 - Unknown - Metro Cross.sndh */
TIMEDB_ENTRY( da59c334,  0,   1260,          YM), /* 01990 - Mad Max - Hexer.sndh */
TIMEDB_ENTRY( da5b1349,  0,   2880,          YM), /* 00975 - Chris Denman - Cohort 2 - Fighting For Rome.sndh */
TIMEDB_ENTRY( da688877,  0,   9603,    YM+TA+TD), /* 02355 - Tao - Cream - Sidchimes (YM Rockerz 1).sndh */
TIMEDB_ENTRY( da6ab68c,  0,   2559, YM+TA+TB+TD), /* 03267 - Dubmood - KOLOBKI VODKA DANCE.sndh */
TIMEDB_ENTRY( da7841b3,  0,   1792,          YM), /* 03207 - Chris Denman - Striker.sndh */
TIMEDB_ENTRY( da85a916,  0,  35328, YM+TA+TB+TD), /* 03330 - Marcer - n-a.sndh */
TIMEDB_ENTRY( da8b4f1d,  0,   7771,          YM), /* 00864 - Ben Daglish - Motor Massacre.sndh */
TIMEDB_ENTRY( da8b4f1d,  1,    936,          YM), /* 00864 - Ben Daglish - Motor Massacre.sndh */
TIMEDB_ENTRY( da8b4f1d,  2,    359,          YM), /* 00864 - Ben Daglish - Motor Massacre.sndh */
TIMEDB_ENTRY( da8b4f1d,  3,    257,          YM), /* 00864 - Ben Daglish - Motor Massacre.sndh */
TIMEDB_ENTRY( da8b4f1d,  4,    658,          YM), /* 00864 - Ben Daglish - Motor Massacre.sndh */
TIMEDB_ENTRY( da986208,  0,   7148, YM+TA+TB+TD), /* 02779 - Alchemist - TUC Compo Entry (BFP 06).sndh */
TIMEDB_ENTRY( daa08cbd,  0,   3456,          YM), /* 01764 - Lap - Lap 4.sndh */
TIMEDB_ENTRY( dabd0f6d,  0,  14018,          YM), /* 00440 - Big Alec - Traffic.sndh */
TIMEDB_ENTRY( dac05279,  0,    384,          YM), /* 03496 - Shamen - Certified Dope - In Space.sndh */
TIMEDB_ENTRY( dac31e15,  0,  12666,          YM), /* 00500 - Tangens - The 2nd Try - Landscape.sndh */
TIMEDB_ENTRY( dad1894d,  0,  11904,       YM+TA), /* 02129 - Mad Max - Life's A Bitch - GigaTex.sndh */
TIMEDB_ENTRY( dadac5e6,  0,   8112,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  1,    315,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  2,   6528,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  3,  12672,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  4,   1120,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  5,   2048,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  6,   1152,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  7,    864,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  8,   1280,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6,  9,   1008,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dadac5e6, 10,    960,          YM), /* 01798 - Barry Leitch - Lotus Turbo Challenge 2.sndh */
TIMEDB_ENTRY( dae63ffc,  0,   3066,       YM+TA), /* 02640 - Miggy - DevilVer2(Unfinished).sndh */
TIMEDB_ENTRY( daf97772,  0,  70614,       YM+TA), /* 04635 - TFC - Disco II.sndh */
TIMEDB_ENTRY( db0412b8,  0,  25088,          YM), /* 01420 - Sonic - Hype.sndh */
TIMEDB_ENTRY( db09e17e,  0,  16896,          YM), /* 00487 - Ultimatum - Kidney Bean 3.sndh */
TIMEDB_ENTRY( db1fd431,  0,   2262,          YM), /* 03067 - Chris French - Good King Wenceslas.sndh */
TIMEDB_ENTRY( db20f6e5,  0,  15680,    YM+TA+TB), /* 01866 - Lotek Style - Sad.sndh */
TIMEDB_ENTRY( db35c13a,  0,  10362, YM+TA+TB+TD), /* 01008 - DMA-SC - Pie Bill Gates - Menu.sndh */
TIMEDB_ENTRY( db376cf4,  0,  23041,       YM+TA), /* 04803 - Unknown - Flying Shark.sndh */
TIMEDB_ENTRY( db561d60,  0,   8327,          YM), /* 03118 - Unknown - Goodbye Yellow Brick Road.sndh */
TIMEDB_ENTRY( db572f78,  0,   8826, YM+TA+TB+TD), /* 03195 - Carbon - UCM 18.sndh */
TIMEDB_ENTRY( db595e25,  0,  23552, YM+TA+TB+TD), /* 03523 - BIOSCILLATOR - ROTTERDAM ATARI ACID.sndh */
TIMEDB_ENTRY( db59b2e8,  0,  20736,          YM), /* 00441 - Big Alec - Twilight Beyond.sndh */
TIMEDB_ENTRY( db5d3c5a,  0,   6222, YM+TA+TB+TD), /* 04578 - MSG - RG (2003) - EIL 3.sndh */
TIMEDB_ENTRY( db5dc23c,  0,   8829,    YM+TA+TD), /* 02335 - Tao - Cream - Torture (DHS online ssd compo).sndh */
TIMEDB_ENTRY( db62b2b3,  0,   8953, YM+TA+TB+TD), /* 00080 - 505 - Checkpoint^Creators ( 2003 ) - Manic (YM Rockerz).sndh */
TIMEDB_ENTRY( db6679ae,  0,   3262,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  1,   4219,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  2,   4349,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  3,   3036,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  4,   4316,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  5,   5725,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db6679ae,  6,   5725,          YM), /* 01824 - Pierre-Eric Loriaux - Elf.sndh */
TIMEDB_ENTRY( db686050,  0,   3449,          YM), /* 04072 - Gary Wheaton - Handle With Care.sndh */
TIMEDB_ENTRY( db71cda5,  0,   3839,          YM), /* 03592 - Dma-Sc - Happy 2011!!!!!!!!!!!!!!!!.sndh */
TIMEDB_ENTRY( db86a92b,  0,   7872,          YM), /* 00783 - Crawdaddy - Enigma Force.sndh */
TIMEDB_ENTRY( db8cdec5,  0, 117126, YM+TA+TB+TD), /* 04256 - 505 - Checkpoint - EIL99 testsong.sndh */
TIMEDB_ENTRY( db8dadca,  0,  12673,          YM), /* 02021 - Mad Max - Battle Of Britain.sndh */
TIMEDB_ENTRY( db98a346,  0,   4218,       YM+TA), /* 01155 - Tyan - Senseless.sndh */
TIMEDB_ENTRY( db993294,  0,   4896,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  1,    289,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  2,     97,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  3,     73,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  4,    256,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  5,   1601,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  6,   1536,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  7,    129,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  8,    512,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294,  9,    768,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 10,   1872,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 11,    593,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 12,    359,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 13,    976,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 14,    713,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 15,    541,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 16,   1044,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 17,   1044,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 18,    761,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 19,    760,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 20,    736,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 21,    553,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 22,    761,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 23,    377,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 24,    400,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 25,    449,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 26,    760,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 27,    304,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 28,    256,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( db993294, 29,   1120,          YM), /* 01892 - Dave Lowe - Summer Games.sndh */
TIMEDB_ENTRY( dbb04250,  0,  11137,    YM+TA+TD), /* 01872 - Lotek Style - Sweety.sndh */
TIMEDB_ENTRY( dbb6421c,  0,  43008,       YM+TD), /* 02537 - Remo - Tough Mothafucka.sndh */
TIMEDB_ENTRY( dbc40773,  0,   2112,          YM), /* 02219 - Mad Max - Thalion.sndh */
TIMEDB_ENTRY( dbc40773,  1,   7872,          YM), /* 02219 - Mad Max - Thalion.sndh */
TIMEDB_ENTRY( dbc40773,  2,   5047,          YM), /* 02219 - Mad Max - Thalion.sndh */
TIMEDB_ENTRY( dbc40773,  3,    192,          YM), /* 02219 - Mad Max - Thalion.sndh */
TIMEDB_ENTRY( dbc40773,  4,    480,          YM), /* 02219 - Mad Max - Thalion.sndh */
TIMEDB_ENTRY( dbd2e70f,  0,   6144,          YM), /* 01430 - Starbuck - Rocket.sndh */
TIMEDB_ENTRY( dbd3c7e1,  0, 235825,       YM+TA), /* 03701 - Sparky - Absentia.sndh */
TIMEDB_ENTRY( dbd90140,  0,  10068,          YM), /* 01486 - Holger Gehrmann - Oil Imperium.sndh */
TIMEDB_ENTRY( dbd90140,  1,   5076,          YM), /* 01486 - Holger Gehrmann - Oil Imperium.sndh */
TIMEDB_ENTRY( dc20e456,  0,   5214,          YM), /* 03105 - n-a - I Can See Clearly Now.sndh */
TIMEDB_ENTRY( dc24995a,  0,   2682,       YM+TA), /* 04323 - 505-cpt^pdx - Tsrups.sndh */
TIMEDB_ENTRY( dc2a5e38,  0,   9726,          YM), /* 04065 - Gary Wheaton - Chuckie Egg 2.sndh */
TIMEDB_ENTRY( dc2fea6d,  0,    959,    YM+TA+TD), /* 02654 - Dma-Sc - POOZ - The Lab.sndh */
TIMEDB_ENTRY( dc331f15,  0,   7428,    YM+TA+TB), /* 00207 - 505 - UCM 14.sndh */
TIMEDB_ENTRY( dc40a8d8,  0,  15840,       YM+TA), /* 01644 - Jedi - For Jeff.sndh */
TIMEDB_ENTRY( dc828d00,  0,  13439, YM+TA+TB+TD), /* 04758 - Shinobi - Drop Zoner.sndh */
TIMEDB_ENTRY( dc8654c8,  0,  11519, YM+TA+TB+TD), /* 02784 - gwEm - Future Gone Wrong.sndh */
TIMEDB_ENTRY( dc902a39,  0,  18624,       YM+TA), /* 02128 - Mad Max - Beat Nick.sndh */
TIMEDB_ENTRY( dc95372a,  0,   8827,          YM), /* 00553 - Allister Brimble - Kamikaze.sndh */
TIMEDB_ENTRY( dc95372a,  1,   1915,          YM), /* 00553 - Allister Brimble - Kamikaze.sndh */
TIMEDB_ENTRY( dca12676,  0,  18240,       YM+TD), /* 01995 - Mad Max - Auf Weidersehen Monty - digidrums.sndh */
TIMEDB_ENTRY( dca399f9,  0,   2875,          YM), /* 01639 - Jason - Silly - STE only.sndh */
TIMEDB_ENTRY( dca4cc63,  0,   5115, YM+TA+TB+TD), /* 00190 - 505 - Smurfy.sndh */
TIMEDB_ENTRY( dca92951,  0,  47631,    YM+TA+TD), /* 03564 - Tao - Cream (2010) - Heftig (Sillyventure 2010).sndh */
TIMEDB_ENTRY( dcb3ba9b,  0,   4298,          YM), /* 03164 - Unknown - Silent Night - Franz Gruber.sndh */
TIMEDB_ENTRY( dcc31ecc,  0, 227495,       YM+TA), /* 04806 - Unknown - Magnetic Fields.sndh */
TIMEDB_ENTRY( dcc68a4d,  0,  12160,          YM), /* 03649 - Chris Mad - David Whittaker Demo.sndh */
TIMEDB_ENTRY( dcd09be2,  0,   7493,       YM+TA), /* 04414 - EVE - Loading & Decrunching.sndh */
TIMEDB_ENTRY( dcd0e02d,  0,  52036,       YM+TA), /* 04676 - Unknown - Idea.sndh */
TIMEDB_ENTRY( dcd274f9,  0,  23034,          YM), /* 01677 - Jess - My First Resonant Burp (STf).sndh */
TIMEDB_ENTRY( dcd4953e,  0,  83072,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  1,  62140,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  2,  89961,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  3,  40970,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  4, 234749,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  5, 311586,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  6,  74669,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd4953e,  7,  10001,       YM+TA), /* 03770 - Steve Wetherill - Projectyle.sndh */
TIMEDB_ENTRY( dcd9b691,  0,   3840,          YM), /* 00907 - Danny Oneway - Logan.sndh */
TIMEDB_ENTRY( dcd9ee5b,  0,   6400,          YM), /* 00606 - Chris Mad - Phalanx Mega Dister.sndh */
TIMEDB_ENTRY( dce50e74,  0,   1917,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  1,    956,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  2,    139,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  3,    139,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  4,    129,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  5,    272,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  6,     76,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  7,    187,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce50e74,  8,    378,          YM), /* 00556 - Allister Brimble - Rampart.sndh */
TIMEDB_ENTRY( dce72ffd,  0,  74161,       YM+TA), /* 03840 - Tufty - Killer.sndh */
TIMEDB_ENTRY( dcf9b333,  0,  11040,          YM), /* 00588 - Celtic - Space Invaders - TFU Reset.sndh */
TIMEDB_ENTRY( dcfb13ca,  0,   8256,       YM+TA), /* 00263 - Antichrist - Speedzak.sndh */
TIMEDB_ENTRY( dd1d07f0,  0,   3744,          YM), /* 01383 - Chris Scudds - Hydra.sndh */
TIMEDB_ENTRY( dd1d07f0,  1,   3456,          YM), /* 01383 - Chris Scudds - Hydra.sndh */
TIMEDB_ENTRY( dd1d07f0,  2,   3072,          YM), /* 01383 - Chris Scudds - Hydra.sndh */
TIMEDB_ENTRY( dd1d07f0,  3,   1152,          YM), /* 01383 - Chris Scudds - Hydra.sndh */
TIMEDB_ENTRY( dd1d5f5f,  0,  22020,          YM), /* 00767 - Count Zero - Decade Demo - Mainmenu.sndh */
TIMEDB_ENTRY( dd27750b,  0,   7708,          YM), /* 01565 - Jason Page - Paradroid 90.sndh */
TIMEDB_ENTRY( dd27750b,  1,   2716,          YM), /* 01565 - Jason Page - Paradroid 90.sndh */
TIMEDB_ENTRY( dd27750b,  2,    337,          YM), /* 01565 - Jason Page - Paradroid 90.sndh */
TIMEDB_ENTRY( dd27750b,  3,    412,          YM), /* 01565 - Jason Page - Paradroid 90.sndh */
TIMEDB_ENTRY( dd27750b,  4,    406,          YM), /* 01565 - Jason Page - Paradroid 90.sndh */
TIMEDB_ENTRY( dd2c8fe2,  0,   8058,       YM+TA), /* 01365 - Scavenger - Sample Animal is Alive!.sndh */
TIMEDB_ENTRY( dd396e4e,  0,  11560,          YM), /* 03402 - Unknown - Tiger Road.sndh */
TIMEDB_ENTRY( dd411d05,  0,  10745, YM+TA+TB+TD), /* 00982 - D-Force - Bytes.sndh */
TIMEDB_ENTRY( dd4b0659,  0,  15177,          YM), /* 02062 - Mad Max - Zoids.sndh */
TIMEDB_ENTRY( dd4c0fea,  0,  12800,          YM), /* 01637 - Nexus - PETE - Wild Boys.sndh */
TIMEDB_ENTRY( dd4ddf9f,  0,   1920,          YM), /* 00486 - Ultimatum - Kidney Bean 2.sndh */
TIMEDB_ENTRY( dd603d4c,  0,   5619,          YM), /* 01320 - Matt Furniss - The Spy Who Loved Me.sndh */
TIMEDB_ENTRY( dd603d4c,  1,   7494,          YM), /* 01320 - Matt Furniss - The Spy Who Loved Me.sndh */
TIMEDB_ENTRY( dd6922cc,  0,    889, YM+TA+TB+TD), /* 01948 - MSG - Chu Chu Rocket - Results.sndh */
TIMEDB_ENTRY( dd6a5647,  0,   3396,          YM), /* 01925 - The Mind - Genesys Remix - Main Menu.sndh */
TIMEDB_ENTRY( dd71184e,  0,  19744,          YM), /* 00343 - Tony Williams - Ninja Warriors.sndh */
TIMEDB_ENTRY( dd7ee208,  0,  41371,       YM+TA), /* 04044 - Roggie - Is he a mule?.sndh */
TIMEDB_ENTRY( dd9df0b2,  0,   9120,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  1,    145,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  2,     66,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  3,     31,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  4,     31,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  5,     81,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( dd9df0b2,  6,     11,          YM), /* 01787 - Anthony Lees - Incredible Shrinking Sphere.sndh */
TIMEDB_ENTRY( ddc18eeb,  0,  10746, YM+TA+TB+TD), /* 00998 - DMA-SC - Gr00vix.sndh */
TIMEDB_ENTRY( ddc3e416,  0,   3840,          YM), /* 01096 - Techno - Plucked Goose.sndh */
TIMEDB_ENTRY( ddc4b469,  0,   7200,          YM), /* 04469 - Charles Callet - Prohibition.sndh */
TIMEDB_ENTRY( ddc62fc9,  0,   3834,          YM), /* 01229 - Epic - Got No Respect For The Law.sndh */
TIMEDB_ENTRY( dddca2cb,  0,   3750,          YM), /* 03492 - Simon Berry - Backlash.sndh */
TIMEDB_ENTRY( de0f51a4,  0,   4218,       YM+TA), /* 01340 - Scavenger - Knuckle Joe 2.sndh */
TIMEDB_ENTRY( de20d9ef,  0,   4352,          YM), /* 01075 - Excellence In Art - Blommor.sndh */
TIMEDB_ENTRY( de2c8be5,  0,   6911,          YM), /* 03645 - 505 - Shazz YM6channel 123.sndh */
TIMEDB_ENTRY( de2c9fb9,  0,   9249,          YM), /* 00842 - Ben Daglish - Butcher Hill.sndh */
TIMEDB_ENTRY( de2c9fb9,  1,   8783,          YM), /* 00842 - Ben Daglish - Butcher Hill.sndh */
TIMEDB_ENTRY( de2c9fb9,  2,   9252,          YM), /* 00842 - Ben Daglish - Butcher Hill.sndh */
TIMEDB_ENTRY( de2c9fb9,  3,   8319,          YM), /* 00842 - Ben Daglish - Butcher Hill.sndh */
TIMEDB_ENTRY( de2c9fb9,  6,  10561,          YM), /* 00842 - Ben Daglish - Butcher Hill.sndh */
TIMEDB_ENTRY( de2e1e47,  0,   5760,          YM), /* 01489 - Fred Gray - Black Lamp.sndh */
TIMEDB_ENTRY( de2e1e47,  1,   2880,          YM), /* 01489 - Fred Gray - Black Lamp.sndh */
TIMEDB_ENTRY( de2e1e47,  2,    437,          YM), /* 01489 - Fred Gray - Black Lamp.sndh */
TIMEDB_ENTRY( de341604,  0,  14352,       YM+TA), /* 04403 - Scavenger - Anarchy.sndh */
TIMEDB_ENTRY( de361082,  0,   3360,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de361082,  1,    335,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de361082,  2,   1828,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de361082,  3,   1828,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de361082,  4,   1280,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de361082,  5,    518,          YM), /* 04485 - Ralph Effemey - 3D aSTeroids.sndh */
TIMEDB_ENTRY( de50676e,  0,  13824, YM+TA+TB+TD), /* 03457 - Marcer - Thalion Intro.sndh */
TIMEDB_ENTRY( de57adb7,  0,   9603,          YM), /* 00421 - Big Alec - PYM - Best Part Of Creation.sndh */
TIMEDB_ENTRY( de5f08a6,  0,  50703,    YM+TA+TD), /* 03617 - Tao-Cream (2011) - Ultraflow.sndh */
TIMEDB_ENTRY( de729049,  0,    984,    YM+TA+TD), /* 02507 - Q - intro v2.sndh */
TIMEDB_ENTRY( de78d317,  0,   7710, YM+TA+TB+TD), /* 02998 - 505-paradox - reasonless.sndh */
TIMEDB_ENTRY( de9041a5,  0,   8447, YM+TA+TB+TD), /* 03000 - 505-paradox - sphere.sndh */
TIMEDB_ENTRY( de9dfebf,  0,   9026,          YM), /* 00435 - Big Alec - Sanxion Loader.sndh */
TIMEDB_ENTRY( de9fed79,  0,   1018,          YM), /* 00973 - Deneb - Kidney Bean - Present.sndh */
TIMEDB_ENTRY( dec1c446,  0,   6265,       YM+TA), /* 00024 - 505 - 13th.sndh */
TIMEDB_ENTRY( ded32f1c,  0,  14827,    YM+TA+TD), /* 02668 - Donald Fakk - Echoing.sndh */
TIMEDB_ENTRY( ded49b0f,  0,  35548, YM+TA+TB+TD), /* 03298 - STU - ELECTROX.sndh */
TIMEDB_ENTRY( dee9caa6,  0,  93993,       YM+TA), /* 03784 - Mr Micro - Magic Fly.sndh */
TIMEDB_ENTRY( deebacbe,  0,    508,       YM+TA), /* 02645 - Miggy - stinky(Unfinished).sndh */
TIMEDB_ENTRY( def1319f,  0,  51783,       YM+TA), /* 04052 - Roggie - Beef Barbeque.sndh */
TIMEDB_ENTRY( def9a728,  0,   3819,          YM), /* 03103 - n-a - Children's Corner # 5.sndh */
TIMEDB_ENTRY( df0f0e71,  0,   3577, YM+TA+TB+TD), /* 00279 - Baggio - Softly.sndh */
TIMEDB_ENTRY( df112436,  0,  15335,    YM+TA+TD), /* 02772 - Chromex - Hydroxid Dentro.sndh */
TIMEDB_ENTRY( df183024,  0,   1535,          YM), /* 04085 - Gary Wheaton - Necro.sndh */
TIMEDB_ENTRY( df1d8073,  0,   6344,          YM), /* 01569 - Patrick Phelan - Nigel Mansell's World Championship.sndh */
TIMEDB_ENTRY( df44b4e9,  0,  15087, YM+TA+TB+TD), /* 02867 - Dubmood - POSTapoclyphse.sndh */
TIMEDB_ENTRY( df47a02c,  0,  14660, YM+TA+TB+TD), /* 03442 - ULTRASYD - BRAIN DAMAGING ARPS.sndh */
TIMEDB_ENTRY( df636b47,  0,   3072,          YM), /* 03500 - Shamen - Certified Dope - Sources.sndh */
TIMEDB_ENTRY( df72b5f0,  0,  14784, YM+TA+TB+TD), /* 02082 - Mad Max - I Ball - Sid.sndh */
TIMEDB_ENTRY( df7310fa,  0,   7720, YM+TA+TB+TD), /* 00031 - 505 - Phlegm Bleepz.sndh */
TIMEDB_ENTRY( df850308,  0,   3428,          YM), /* 00103 - David Whittaker - Gazza 2.sndh */
TIMEDB_ENTRY( df850308,  1,   1117,          YM), /* 00103 - David Whittaker - Gazza 2.sndh */
TIMEDB_ENTRY( df850308,  2,    509,          YM), /* 00103 - David Whittaker - Gazza 2.sndh */
TIMEDB_ENTRY( df850308,  3,    481,          YM), /* 00103 - David Whittaker - Gazza 2.sndh */
TIMEDB_ENTRY( df971bda,  0,   5400,          YM), /* 00242 - AN Cool - 45-King.sndh */
TIMEDB_ENTRY( df9c0855,  0,  23418,          YM), /* 01671 - Jess - Concerto for Acid Band.sndh */
TIMEDB_ENTRY( dfb00c35,  0,  25536,          YM), /* 02059 - Mad Max - Make Love Not W.A.R..sndh */
TIMEDB_ENTRY( dfb138ee,  0,   6166,          YM), /* 00443 - Big Alec - Wicked.sndh */
TIMEDB_ENTRY( dfedc994,  0,   3916,          YM), /* 04066 - Gary Wheaton - Circus.sndh */
TIMEDB_ENTRY( e013b117,  0,  53120,       YM+TD), /* 01464 - STU - DYMANITE.sndh */
TIMEDB_ENTRY( e014c8cd,  0,  11486,       YM+TA), /* 03954 - Scott Lemmon - Ivan 2.sndh */
TIMEDB_ENTRY( e0196bcd,  0,    339,          YM), /* 00868 - Ben Daglish - Passing Shot.sndh */
TIMEDB_ENTRY( e0196bcd,  1,    375,          YM), /* 00868 - Ben Daglish - Passing Shot.sndh */
TIMEDB_ENTRY( e0196bcd,  2,    305,          YM), /* 00868 - Ben Daglish - Passing Shot.sndh */
TIMEDB_ENTRY( e0196bcd,  3,    394,          YM), /* 00868 - Ben Daglish - Passing Shot.sndh */
TIMEDB_ENTRY( e0196bcd,  4,   2260,          YM), /* 00868 - Ben Daglish - Passing Shot.sndh */
TIMEDB_ENTRY( e021b94c,  0,   5613,          YM), /* 03029 - The Assasin - Fun Fun.sndh */
TIMEDB_ENTRY( e037009e,  0,  17949,          YM), /* 00770 - Count Zero - Decade Demo - Rasters.sndh */
TIMEDB_ENTRY( e0457cb6,  0,   4909,          YM), /* 03542 - Justin Scharvona - The Godfather.sndh */
TIMEDB_ENTRY( e0457cb6,  2,   1926,          YM), /* 03542 - Justin Scharvona - The Godfather.sndh */
TIMEDB_ENTRY( e06a7958,  0,   8688,    YM+TA+TD), /* 02782 - Lotek Style - Crushed The Majors.sndh */
TIMEDB_ENTRY( e076a58f,  0,    181,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  1,    194,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  2,   1828,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  3,   2288,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  4,   1645,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  5,   2080,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e076a58f,  6,   1905,          YM), /* 03972 - Geoff Camp - 1st Maths.sndh */
TIMEDB_ENTRY( e079951f,  0,   3743,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  1,   3927,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  2,    117,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  3,     71,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  4,     43,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  5,     11,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  6,     83,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e079951f,  7,     19,          YM), /* 04142 - Michael Haar - Tetris X.sndh */
TIMEDB_ENTRY( e08a92f7,  0,    389,          YM), /* 00729 - Tebirod - No Cooper Vertical Rasters.sndh */
TIMEDB_ENTRY( e09a157d,  0,  22581,          YM), /* 00616 - Chrispy Noodle - Sapiens 2.sndh */
TIMEDB_ENTRY( e0b125f6,  0,   1536,          YM), /* 02378 - Nexus 6 - Hopscotch.sndh */
TIMEDB_ENTRY( e0b46bce,  0,   4992,       YM+TA), /* 00943 - Dark Angel - Hallo Pan Tau.sndh */
TIMEDB_ENTRY( e0c51fef,  0,   2305, YM+TA+TB+TD), /* 03676 - Dma-Sc - Antiques - Intro.sndh */
TIMEDB_ENTRY( e0e0e3e9,  0,   6248,          YM), /* 03084 - Harry Wootan - Desafinado.sndh */
TIMEDB_ENTRY( e0ecf617,  0,   9690,          YM), /* 02214 - Mad Max - Return Of Medusa.sndh */
TIMEDB_ENTRY( e0ecf617,  1,    768,          YM), /* 02214 - Mad Max - Return Of Medusa.sndh */
TIMEDB_ENTRY( e0ecf617,  2,  26631,          YM), /* 02214 - Mad Max - Return Of Medusa.sndh */
TIMEDB_ENTRY( e124bd14,  0,  18152,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  1,   7679,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  2,   3134,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  3,   5036,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  4,   8694,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  5,   7144,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  6,   5907,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  7,   5036,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  8,   8302,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e124bd14,  9,  17655,          YM), /* 03534 - Jean Baudlot - Future Wars - Time Travellers.sndh */
TIMEDB_ENTRY( e1276741,  0,    831,          YM), /* 04627 - tomchi  noextra - ltc03 0ompa intro.sndh */
TIMEDB_ENTRY( e14f1aa3,  0,  33792,       YM+TD), /* 01480 - STU - TYMPANUM.sndh */
TIMEDB_ENTRY( e1564b76,  0,  18090,          YM), /* 01718 - Jess - Throw It in a Trash Can (STf).sndh */
TIMEDB_ENTRY( e161778f,  0,    731,          YM), /* 04708 - Vesa, Emphii, Bad Boy - Razor Megademo Intro.sndh */
TIMEDB_ENTRY( e161778f,  1,    370,          YM), /* 04708 - Vesa, Emphii, Bad Boy - Razor Megademo Intro.sndh */
TIMEDB_ENTRY( e161778f,  2,    416,          YM), /* 04708 - Vesa, Emphii, Bad Boy - Razor Megademo Intro.sndh */
TIMEDB_ENTRY( e161778f,  3,    256,          YM), /* 04708 - Vesa, Emphii, Bad Boy - Razor Megademo Intro.sndh */
TIMEDB_ENTRY( e17de217,  0,   7953,       YM+TA), /* 04313 - 505 - ST Tri Sound.sndh */
TIMEDB_ENTRY( e1a66b55,  0,    413,          YM), /* 02963 - Unknown - Henriettas Book of Spells.sndh */
TIMEDB_ENTRY( e1a66b55,  1,    256,          YM), /* 02963 - Unknown - Henriettas Book of Spells.sndh */
TIMEDB_ENTRY( e1a66b55,  2,    320,          YM), /* 02963 - Unknown - Henriettas Book of Spells.sndh */
TIMEDB_ENTRY( e1a66b55,  3,    643,          YM), /* 02963 - Unknown - Henriettas Book of Spells.sndh */
TIMEDB_ENTRY( e1cbb5fa,  0,   2654,          YM), /* 03125 - n-a - I Heard the Bells.sndh */
TIMEDB_ENTRY( e1d5a4e3,  0,   2864,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  1,   3644,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  2,   1800,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  3,   3554,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  4,   3385,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  5,   3494,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  6,   4057,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  7,   4509,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  8,   2264,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3,  9,   3480,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 10,   3444,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 11,   5049,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 12,   3895,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 13,   3958,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 14,   3972,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 15,   3080,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 16,   3005,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 17,   3252,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 29,   9040,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1d5a4e3, 30,    288,          YM), /* 04716 - Geoff Camp - Spider Spell 2.sndh */
TIMEDB_ENTRY( e1f17eba,  0,    672,          YM), /* 02361 - Nexus 6 - Audiopac.sndh */
TIMEDB_ENTRY( e1feabd9,  0,  20480, YM+TA+TB+TD), /* 02090 - Mad Max - Nemesis - Sid.sndh */
TIMEDB_ENTRY( e207c574,  0,    896,       YM+TA), /* 00954 - Dark Angel - The Walk.sndh */
TIMEDB_ENTRY( e20c2c4b,  0,   3447,       YM+TA), /* 01745 - Jovis - Lametrop - Intro.sndh */
TIMEDB_ENTRY( e2614bdc,  0,  10745,    YM+TA+TB), /* 01592 - Nemo - Endpart M2V.sndh */
TIMEDB_ENTRY( e26bf07b,  0,  14400,       YM+TA), /* 01287 - Furax - Virtual Escape Theme.sndh */
TIMEDB_ENTRY( e275ea20,  0,  59848,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  1, 154618,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  2,  30627,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  3,  27269,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  4,  66620,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  5,  71019,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  6,  74309,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e275ea20,  7,  29569,       YM+TA), /* 03822 - Tufty - My Digi 2.sndh */
TIMEDB_ENTRY( e2811199,  0,   9060,          YM), /* 00769 - Count Zero - Decade Demo - Pipeworx.sndh */
TIMEDB_ENTRY( e28aa01f,  0,   4225, YM+TA+TB+TD), /* 04583 - mOdmate-Point de Controle (orig. David Whittaker) - COLONYST (PROOF OF POKEY CONCEPT).sndh */
TIMEDB_ENTRY( e2a3859f,  0,   5759, YM+TA+TB+TD), /* 03150 - Rubbish - Almost Happy.sndh */
TIMEDB_ENTRY( e2a56417,  0,   8326,          YM), /* 01568 - Jason Page - Simulcra.sndh */
TIMEDB_ENTRY( e2a56417,  1,   3209,          YM), /* 01568 - Jason Page - Simulcra.sndh */
TIMEDB_ENTRY( e2bbc9e5,  0,   2525,          YM), /* 03135 - Unknown - Have Yourself-Merry Christmas.sndh */
TIMEDB_ENTRY( e2db2940,  0,   8960,    YM+TB+TD), /* 02849 - Helene - Carsong2.sndh */
TIMEDB_ENTRY( e2dce743,  0,   4799,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  1,   3519,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  2,   3519,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  3,   2369,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  4,   2239,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  5,   2879,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2dce743,  6,   3839,          YM), /* 00352 - Tony Williams - Shadow of the Beast 2.sndh */
TIMEDB_ENTRY( e2df9ac0,  0,  12880, YM+TA+TB+TD), /* 02768 - cerror (or.by.radix) - Imaginary friend (st-version).sndh */
TIMEDB_ENTRY( e2e0bc3f,  0,   6750,       YM+TA), /* 01920 - Mephistow - Librarie.sndh */
TIMEDB_ENTRY( e2e8ac75,  0,  26490,       YM+TA), /* 04579 - 505 - Hardsync 01I.sndh */
TIMEDB_ENTRY( e2f1e8f5,  0,   2560,          YM), /* 04382 - Chris Denman - Great Napoleonic Battles.sndh */
TIMEDB_ENTRY( e3029117,  0,  18432, YM+TA+TB+TD), /* 02576 - STU - BEE.sndh */
TIMEDB_ENTRY( e30425fe,  0,  10724, YM+TA+TB+TD), /* 03559 - Evil - noname (Amiga cover).sndh */
TIMEDB_ENTRY( e30e58b8,  0,  12898,    YM+TA+TD), /* 02347 - Tao - Cream - Starchild (UCM 23).sndh */
TIMEDB_ENTRY( e3161b72,  0,   7297,    YM+TA+TD), /* 02318 - Tao - Cream - Looniez.sndh */
TIMEDB_ENTRY( e3162d86,  0,  13824,          YM), /* 01263 - Frazer - Amazine 2.sndh */
TIMEDB_ENTRY( e31838d4,  0,  59664,       YM+TA), /* 04815 - Unknown - Bob P.sndh */
TIMEDB_ENTRY( e32a0f5c,  0,  23040,          YM), /* 01422 - Sonic - Ramp.sndh */
TIMEDB_ENTRY( e3361e8b,  0,  11936,          YM), /* 02940 - Archer Maclean - Archer Macleans Pool.sndh */
TIMEDB_ENTRY( e33f0401,  0,  11641,       YM+TA), /* 04617 - The Byte - BSW 88.sndh */
TIMEDB_ENTRY( e3481e5a,  0,   6528,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e3481e5a,  1,   2688,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e3481e5a,  2,    384,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e3481e5a,  3,   5376,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e3481e5a,  4,   1968,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e3481e5a,  5,   5760,          YM), /* 01060 - Jonathan Dunn - Lethal Weapon.sndh */
TIMEDB_ENTRY( e398ab48,  0,  17313,       YM+TA), /* 03746 - Tufty - Hypnotic.sndh */
TIMEDB_ENTRY( e3af26e5,  0,   2304,          YM), /* 04084 - Gary Wheaton - Munsters.sndh */
TIMEDB_ENTRY( e3b9c8d4,  0,   8658,          YM), /* 03930 - Unknown - Sound1.sndh */
TIMEDB_ENTRY( e3c81a39,  0,   4127, YM+TA+TB+TD), /* 02728 - maktone-fairlight - glory.sndh */
TIMEDB_ENTRY( e3e0e278,  0,   2240,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  1,   1440,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  2,   1152,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  3,   1440,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  4,   1536,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  5,   1728,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  6,    864,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  7,   1536,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e0e278,  8,   1836,          YM), /* 02914 - Richard Joseph - Legend.sndh */
TIMEDB_ENTRY( e3e82dbb,  0,  15512,    YM+TB+TD), /* 03303 - Marcer - ROM TITLE MIX.sndh */
TIMEDB_ENTRY( e3ea9120,  0,  14213,          YM), /* 00269 - Archimede - Gen 4 3615 Demo.sndh */
TIMEDB_ENTRY( e411e968,  0,   2873,          YM), /* 02937 - Ash Hogg - Spikey in Transylvania.sndh */
TIMEDB_ENTRY( e411e968,  1,   6642,          YM), /* 02937 - Ash Hogg - Spikey in Transylvania.sndh */
TIMEDB_ENTRY( e411e968,  2,    170,          YM), /* 02937 - Ash Hogg - Spikey in Transylvania.sndh */
TIMEDB_ENTRY( e4260543,  0,   1544,          YM), /* 00935 - Dark Angel - Metropolitan.sndh */
TIMEDB_ENTRY( e426ae9f,  0,   3238, YM+TA+TB+TD), /* 04309 - 505-checkpoint - ST_O.sndh */
TIMEDB_ENTRY( e4298532,  0,   1096,          YM), /* 04690 - Unknown - Life on the Ocean Waves.sndh */
TIMEDB_ENTRY( e43f7d3c,  0,   1280,          YM), /* 01404 - Skyline - Snork Intro - Hemosound.sndh */
TIMEDB_ENTRY( e4604a81,  0,   2079,          YM), /* 01297 - Matt Furniss - Face Off.sndh */
TIMEDB_ENTRY( e4660401,  0,   7683,          YM), /* 00428 - Big Alec - PYM - Main Menu.sndh */
TIMEDB_ENTRY( e469ae57,  0,   8105,          YM), /* 01276 - Frazer - TSC 42.sndh */
TIMEDB_ENTRY( e4742717,  0,   5119,          YM), /* 04063 - Gary Wheaton - Bass.sndh */
TIMEDB_ENTRY( e47d718d,  0,   8424,          YM), /* 02972 - Gary Wheaton - Escapade (Scarborough Fair).sndh */
TIMEDB_ENTRY( e483468f,  0,  51840,          YM), /* 03400 - Carsten Neubauer - Nightdawn.sndh */
TIMEDB_ENTRY( e4985a0c,  0,   5773,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  1,   1272,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  2,   2677,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  3,   5229,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  4,   1955,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  5,   4059,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  6,    322,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  7,    145,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  8,    585,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c,  9,    472,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 10,    771,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 11,    613,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 12,    411,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 13,    490,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 14,    352,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 15,    375,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 16,    307,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 17,    754,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 18,   1570,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 19,    250,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 20,    930,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 21,    392,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 22,    920,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 23,   1164,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 24,    361,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4985a0c, 25,    540,          YM), /* 03975 - Christopher Barker - Spirit of Excalibur.sndh */
TIMEDB_ENTRY( e4a0a277,  0,   1536,          YM), /* 01889 - Dave Lowe - Power Drift.sndh */
TIMEDB_ENTRY( e4a0a277,  1,   1547,          YM), /* 01889 - Dave Lowe - Power Drift.sndh */
TIMEDB_ENTRY( e4b18b94,  0,   5057,          YM), /* 01379 - Justin Scharvona - Slackskin & Flint.sndh */
TIMEDB_ENTRY( e4b18b94,  1,   1281,          YM), /* 01379 - Justin Scharvona - Slackskin & Flint.sndh */
TIMEDB_ENTRY( e4b18b94,  2,   2241,          YM), /* 01379 - Justin Scharvona - Slackskin & Flint.sndh */
TIMEDB_ENTRY( e4b18b94,  3,    769,          YM), /* 01379 - Justin Scharvona - Slackskin & Flint.sndh */
TIMEDB_ENTRY( e4b88228,  0,   5116,          YM), /* 00549 - Allister Brimble - Fantasy World Dizzy.sndh */
TIMEDB_ENTRY( e4b88228,  1,   6043,          YM), /* 00549 - Allister Brimble - Fantasy World Dizzy.sndh */
TIMEDB_ENTRY( e4b88228,  2,   2876,          YM), /* 00549 - Allister Brimble - Fantasy World Dizzy.sndh */
TIMEDB_ENTRY( e4e000d3,  0,  11751,          YM), /* 00755 - Mark Cooksey - Paperboy.sndh */
TIMEDB_ENTRY( e4e000d3,  1,   1248,          YM), /* 00755 - Mark Cooksey - Paperboy.sndh */
TIMEDB_ENTRY( e4e2ed49,  0,   2239,       YM+TA), /* 01520 - Iso - Eindhoven CS.sndh */
TIMEDB_ENTRY( e5004bf0,  0,  10496, YM+TA+TB+TD), /* 02103 - Mad Max - Warhawk - Sid.sndh */
TIMEDB_ENTRY( e50e1386,  0,  16296,    YM+TB+TD), /* 02852 - Helene - hah.sndh */
TIMEDB_ENTRY( e5112240,  0,   3933,          YM), /* 00568 - Jas C Brooke - Ikari Warriors.sndh */
TIMEDB_ENTRY( e5112240,  1,   3936,          YM), /* 00568 - Jas C Brooke - Ikari Warriors.sndh */
TIMEDB_ENTRY( e5112240,  2,   1152,          YM), /* 00568 - Jas C Brooke - Ikari Warriors.sndh */
TIMEDB_ENTRY( e5112240,  3,    150,          YM), /* 00568 - Jas C Brooke - Ikari Warriors.sndh */
TIMEDB_ENTRY( e54bd79f,  0,   2304,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  1,   3840,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  2,    192,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  3,   1728,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  4,     64,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  5,    215,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  6,   2187,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  7,   1920,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  8,    121,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f,  9,     25,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f, 10,     48,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f, 11,     16,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f, 12,     88,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e54bd79f, 13,     54,          YM), /* 02918 - Richard Joseph - The Adventures of Robin Hood.sndh */
TIMEDB_ENTRY( e553a69e,  0,   6909, YM+TA+TB+TD), /* 00047 - 505 - EIL2 Intro.sndh */
TIMEDB_ENTRY( e56e957a,  0,   5597,          YM), /* 01518 - Iso - Arf Arf - Miauw.sndh */
TIMEDB_ENTRY( e579c04d,  0,   8460,          YM), /* 02929 - Sean Conran - WWF - Wrestle Mania.sndh */
TIMEDB_ENTRY( e579c04d,  1,    573,          YM), /* 02929 - Sean Conran - WWF - Wrestle Mania.sndh */
TIMEDB_ENTRY( e57bd95d,  0,   6304,    YM+TA+TD), /* 02554 - Marcer - Broken.sndh */
TIMEDB_ENTRY( e5834ea9,  0,   3680,          YM), /* 04600 - Rude Kid - Playtime.sndh */
TIMEDB_ENTRY( e598f127,  0,   4608,    YM+TA+TD), /* 00833 - Crazy Q - DHS - TAC-2 Memories.sndh */
TIMEDB_ENTRY( e5a3c2c6,  0,  39892, YM+TA+TB+TD), /* 03561 - !Cube - Blink.sndh */
TIMEDB_ENTRY( e5ff165b,  0,   5120,          YM), /* 02799 - Starbuck - Her 5.sndh */
TIMEDB_ENTRY( e62c9af8,  0,    320,          YM), /* 01208 - Excellence In Art - Tarzan.sndh */
TIMEDB_ENTRY( e633cfd4,  0,  39936,    YM+TB+TD), /* 02607 - STU - MY MELODY SID.sndh */
TIMEDB_ENTRY( e65f1def,  0,   7672, YM+TA+TB+TD), /* 00544 - Blindio - Decadent.sndh */
TIMEDB_ENTRY( e67eb69f,  0,   3066,       YM+TA), /* 00902 - Danny Oneway - Frust.sndh */
TIMEDB_ENTRY( e69c3ee3,  0,  37120, YM+TA+TB+TD), /* 03341 - Marcer - n-a.sndh */
TIMEDB_ENTRY( e6ab9b25,  0,  35154,       YM+TA), /* 03781 - Jose A Martin - Narco Police.sndh */
TIMEDB_ENTRY( e6b3f254,  0,   2299,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  1,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  2,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  3,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  4,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  5,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  6,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  7,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  8,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6b3f254,  9,  16219,          YM), /* 01826 - Pierre-Eric Loriaux - Toki.sndh */
TIMEDB_ENTRY( e6bbb5a2,  0,   5760,          YM), /* 00493 - Ultimatum - Kidney Bean 9.sndh */
TIMEDB_ENTRY( e6c049a0,  0,  11904,    YM+TA+TD), /* 02250 - Tao - Cream - Call Me (Steps).sndh */
TIMEDB_ENTRY( e6c20336,  0,   6396,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6c20336,  1,   2298,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6c20336,  2,    762,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6c20336,  3,   2298,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6c20336,  4,   9772,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6c20336,  5,   6906,          YM), /* 00835 - Neil Crossley - Hunt For Red October.sndh */
TIMEDB_ENTRY( e6d22082,  0, 123322,       YM+TA), /* 04061 - Hybrid - Hybrid.sndh */
TIMEDB_ENTRY( e6e43013,  0,   6912,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  1,   8449,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  2,  10304,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  3,   7296,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  4,   7168,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  5,   8064,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  6,   1920,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  7,   1280,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  8,    768,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e6e43013,  9,   4928,          YM), /* 01065 - Jonathan Dunn - Dean Evans - Keith Tinman - Pushover.sndh */
TIMEDB_ENTRY( e70e705e,  0,  10048,          YM), /* 01030 - Doclands - Zak.sndh */
TIMEDB_ENTRY( e744958f,  0,   6331,    YM+TB+TD), /* 00082 - 505 - Melodie.sndh */
TIMEDB_ENTRY( e74a3255,  0,  76272,    YM+TB+TD), /* 03461 - Marcer - Kid.sndh */
TIMEDB_ENTRY( e76ad285,  0,   4853,    YM+TA+TD), /* 02822 - Marcer - Greasy (Znax Menu).sndh */
TIMEDB_ENTRY( e7754910,  0,  13096,          YM), /* 01850 - Lotek Style - Hard Rock.sndh */
TIMEDB_ENTRY( e775cf70,  0,   7242, YM+TA+TB+TD), /* 01873 - Lotek Style - Das Sidhaeuser Tor.sndh */
TIMEDB_ENTRY( e77ec0a9,  0,   4800,          YM), /* 03242 - T1000 - The Beast Club - TBC 14.sndh */
TIMEDB_ENTRY( e781ebf3,  0,   5759,       YM+TA), /* 04111 - 505-PARADOX [NILS FESKE] - MIST [2005].sndh */
TIMEDB_ENTRY( e782759a,  0,   7491,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  1,   1539,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  2,    401,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  3,   2211,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  4,    282,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  5,    802,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  6,    452,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e782759a,  7,   1443,          YM), /* 01794 - Barry Leitch - Gilbert - Escape from Drill.sndh */
TIMEDB_ENTRY( e79dcd5c,  0,    663,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  1,    430,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  2,    369,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  3,    341,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  4,    417,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  5,    207,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  6,    489,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  7,    183,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  8,    509,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c,  9,    322,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c, 10,    315,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c, 11,   1777,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e79dcd5c, 12,     95,          YM), /* 00846 - Ben Daglish - Continental Circus.sndh */
TIMEDB_ENTRY( e7a7464a,  0,   8960,          YM), /* 01421 - Sonic - It Goes On.sndh */
TIMEDB_ENTRY( e7a79994,  0,  38501,    YM+TB+TD), /* 02533 - Remo - Beatz.sndh */
TIMEDB_ENTRY( e7ba3248,  0,   7488,          YM), /* 00911 - Danny Oneway - Lamebrain 1 - Techmoc.sndh */
TIMEDB_ENTRY( e7c6841e,  0,  11521, YM+TA+TB+TD), /* 02790 - Dma-Sc - Trials.sndh */
TIMEDB_ENTRY( e7d19bb3,  0,     11,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  1,     14,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  2,     53,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  3,     53,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  4,     29,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  5,    101,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  6,    212,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  7,     44,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  8,     53,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3,  9,     44,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3, 10,     13,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3, 12,   2689,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3, 13,    238,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7d19bb3, 14,   6913,          YM), /* 01581 - Nigel Pritchard - Thunderwing.sndh */
TIMEDB_ENTRY( e7e0229c,  0,  55223,       YM+TA), /* 03778 - Stewart Gardner - Stampede 7.sndh */
TIMEDB_ENTRY( e7e2200f,  0,   4511,          YM), /* 04080 - Gary Wheaton - Masters of Magic.sndh */
TIMEDB_ENTRY( e7ef8c50,  0,  12799,    YM+TA+TD), /* 03486 - gwEm - Fantasy.sndh */
TIMEDB_ENTRY( e8061997,  0,   4986,    YM+TA+TB), /* 01596 - Nemo - Loading - Space Invader.sndh */
TIMEDB_ENTRY( e80744aa,  0,  18529,          YM), /* 04841 - t1000 - Bits 36.sndh */
TIMEDB_ENTRY( e81015af,  0,   8933,          YM), /* 03034 - Barry Deutsch - The Next Time I Fall.sndh */
TIMEDB_ENTRY( e8140729,  0,  11648,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  1,  14592,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  2,  18912,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  3,  12809,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  4,  13824,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  5,  15904,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  6,   2560,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  7,   3072,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  8,   2837,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729,  9,    289,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8140729, 10,   5376,          YM), /* 02192 - Mad Max - Enchanted Land.sndh */
TIMEDB_ENTRY( e8152538,  0,  11522,    YM+TA+TD), /* 02334 - Tao - Cream - Torture (Magic Synth vers.).sndh */
TIMEDB_ENTRY( e825e877,  0,  15743,          YM), /* 04564 - Charles Callet - The Toyottes.sndh */
TIMEDB_ENTRY( e8337ccb,  0,  10880,          YM), /* 02801 - Starbuck - Inside.sndh */
TIMEDB_ENTRY( e836372d,  0,   6914,    YM+TA+TD), /* 04203 - Tao - Cream - Nostalgic-O-Demo Dots Revenge (2 Sid Voice).sndh */
TIMEDB_ENTRY( e864f155,  0,   5385,          YM), /* 02160 - Mad Max - Syntax Terror - Level 16.sndh */
TIMEDB_ENTRY( e86a731e,  0,  16384, YM+TA+TB+TD), /* 02581 - STU - CHASER.sndh */
TIMEDB_ENTRY( e876f525,  0,  10304,    YM+TB+TD), /* 01607 - Nemo - tambumbi.sndh */
TIMEDB_ENTRY( e87936c3,  0,  30080, YM+TA+TB+TD), /* 03652 - Xyce - Saucage 3D.sndh */
TIMEDB_ENTRY( e87c457c,  0,  10763,          YM), /* 01561 - Paul Rowbotham - Whirligig.sndh */
TIMEDB_ENTRY( e8807874,  0,   7936,       YM+TA), /* 02897 - Remalon - Don't Stop.sndh */
TIMEDB_ENTRY( e882f484,  0,   6566,          YM), /* 02883 - Beast - Yopsimex.sndh */
TIMEDB_ENTRY( e88505be,  0,   1341,          YM), /* 02469 - Iso - Kamagurka.sndh */
TIMEDB_ENTRY( e8912ce9,  0,   2241,    YM+TA+TD), /* 02317 - Tao - Cream (orig. Klaus Ude) - LinWu's Challenge 4.sndh */
TIMEDB_ENTRY( e89ac254,  0,   9094,          YM), /* 03243 - T1000 - The Beast Club - TBC 15.sndh */
TIMEDB_ENTRY( e8af3fa1,  0,  16121,       YM+TA), /* 01213 - Energizer - Avante.sndh */
TIMEDB_ENTRY( e8bf2061,  0,    504,          YM), /* 03681 - Dma-Sc - Antiques - Loader 2.sndh */
TIMEDB_ENTRY( e8c3ba44,  0,  44137,          YM), /* 00004 - gwEm - Raid The Tower (non-looping).sndh */
TIMEDB_ENTRY( e8d26c77,  0,   2640,          YM), /* 00743 - Sean Conran - Hong Kong Phooey.sndh */
TIMEDB_ENTRY( e8d26c77,  1,    304,          YM), /* 00743 - Sean Conran - Hong Kong Phooey.sndh */
TIMEDB_ENTRY( e8e24519,  0,   6528,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  1,    480,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  2,    768,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  3,    432,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  4,   2048,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  5,    192,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  6,   7645,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e24519,  7,    133,          YM), /* 01736 - Richard Joseph - James Pond.sndh */
TIMEDB_ENTRY( e8e98938,  0,   6880,          YM), /* 02358 - Nexus 6 - 5-16-38.sndh */
TIMEDB_ENTRY( e8e9b785,  0,   1491,          YM), /* 03138 - n-a - Mini Dance.sndh */
TIMEDB_ENTRY( e8f066f3,  0,   1624,          YM), /* 00841 - Ben Daglish - Blasteroids.sndh */
TIMEDB_ENTRY( e8f066f3,  1,   3448,          YM), /* 00841 - Ben Daglish - Blasteroids.sndh */
TIMEDB_ENTRY( e8f3e611,  0,    768,       YM+TA), /* 04436 - Lotus - Synth Dream 2 - Shuffle Express.sndh */
TIMEDB_ENTRY( e8fc8215,  0,  77071,       YM+TA), /* 04016 - Mad Max - Defcon.sndh */
TIMEDB_ENTRY( e911b901,  0,    374,          YM), /* 00455 - Augustin Vidovic - Pinochio.sndh */
TIMEDB_ENTRY( e927d75f,  0,   6912,    YM+TA+TD), /* 02569 - !Cube-Aggression - Wind.sndh */
TIMEDB_ENTRY( e9330639,  0,  10751,          YM), /* 04064 - Gary Wheaton - Chuckie Egg II remix.sndh */
TIMEDB_ENTRY( e9557f7d,  0,   4004,          YM), /* 00331 - Gary Wheaton - Tang mus 2.sndh */
TIMEDB_ENTRY( e95b5837,  0,   4027,          YM), /* 00548 - Allister Brimble - CJ's Elephant Antics.sndh */
TIMEDB_ENTRY( e95b5837,  1,   9209,          YM), /* 00548 - Allister Brimble - CJ's Elephant Antics.sndh */
TIMEDB_ENTRY( e95b5837,  2,   1531,          YM), /* 00548 - Allister Brimble - CJ's Elephant Antics.sndh */
TIMEDB_ENTRY( e95b5837,  3,   1531,          YM), /* 00548 - Allister Brimble - CJ's Elephant Antics.sndh */
TIMEDB_ENTRY( e95b5837,  4,    475,          YM), /* 00548 - Allister Brimble - CJ's Elephant Antics.sndh */
TIMEDB_ENTRY( e95e4156,  0,  17408, YM+TA+TB+TD), /* 02617 - STU - SID BUZZY.sndh */
TIMEDB_ENTRY( e9622a3d,  0,   6720,          YM), /* 02891 - Illusion - Fucker (Golden Dawn 38).sndh */
TIMEDB_ENTRY( e978fba5,  0,   2740, YM+TA+TB+TD), /* 04307 - 505-checkpoint - ST_B.sndh */
TIMEDB_ENTRY( e98225f9,  0,    384,          YM), /* 00599 - Chris Mad - Overdrive Loader.sndh */
TIMEDB_ENTRY( e98a7b23,  0,   6538,       YM+TA), /* 02501 - Lotek Style - Datapirate.sndh */
TIMEDB_ENTRY( e9a3a81f,  0,   8064,          YM), /* 03277 - Crawdaddy - The Forest.sndh */
TIMEDB_ENTRY( e9af2ea1,  0,   3072,          YM), /* 02029 - Mad Max - Education.sndh */
TIMEDB_ENTRY( e9bb98a1,  0,   1528,       YM+TA), /* 01338 - Scavenger - Jungle.sndh */
TIMEDB_ENTRY( e9c08462,  0,   2292,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c08462,  1,     77,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c08462,  2,    403,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c08462,  3,    185,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c08462,  4,    208,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c08462,  5,    183,          YM), /* 03979 - Claude Abromont - Iron Trackers.sndh */
TIMEDB_ENTRY( e9c5ceb5,  0,   7680,          YM), /* 04629 - Tarzan Boy - Hier Sind Noch Zweu Greetings.sndh */
TIMEDB_ENTRY( e9efb3f5,  0,  15216, YM+TA+TB+TD), /* 02857 - Helene - Rescued.sndh */
TIMEDB_ENTRY( e9f6bccf,  0,  32256,          YM), /* 03374 - STU - PIRATE SHOWDOWN.sndh */
TIMEDB_ENTRY( e9fe50e9,  0, 105124,       YM+TA), /* 03710 - Sparky - War and Pieces (Nuke Iraq?).sndh */
TIMEDB_ENTRY( ea0387cf,  0,    384,       YM+TA), /* 00956 - Dark Angel - To be continued.sndh */
TIMEDB_ENTRY( ea0ae5a3,  0,   5130,       YM+TB), /* 03527 - !Cube - Crystallized.sndh */
TIMEDB_ENTRY( ea1d9a19,  0,   2304,          YM), /* 01193 - Excellence In Art - Nice.sndh */
TIMEDB_ENTRY( ea296a9d,  0,   9216,          YM), /* 01187 - Excellence In Art - Jingle.sndh */
TIMEDB_ENTRY( ea417dd3,  0,   5984,          YM), /* 00861 - Ben Daglish - Mickey Mouse.sndh */
TIMEDB_ENTRY( ea77c957,  0,   9849,          YM), /* 04214 - Jess - Lemon Squeezer's Dream (chip).sndh */
TIMEDB_ENTRY( ea7cf4b0,  0,   1536,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  1,   1536,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  2,   4608,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  3,    480,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  4,     96,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  5,   2880,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea7cf4b0,  6,   2720,          YM), /* 01993 - Mad Max - Archon.sndh */
TIMEDB_ENTRY( ea807e2f,  0,   7296,          YM), /* 02877 - Arishem - Rigolo Demo (Abstract).sndh */
TIMEDB_ENTRY( ea83d2c5,  0,  13154, YM+TA+TB+TD), /* 04517 - Ultrasyd - Staring at the Sea.sndh */
TIMEDB_ENTRY( ea88b50c,  0,   2304,          YM), /* 03206 - Chris Denman - Web of Terror.sndh */
TIMEDB_ENTRY( ea90e2d0,  0,   9905,          YM), /* 04200 - Paul Summers - Espionage.sndh */
TIMEDB_ENTRY( ea9a97c5,  0,   5112,       YM+TA), /* 01685 - Jess - Echelle.sndh */
TIMEDB_ENTRY( ea9bd562,  0,   5850,          YM), /* 00894 - Danny Oneway - Lamebrain 2 - Bomben.sndh */
TIMEDB_ENTRY( eaa4afd7,  0,  96673,       YM+TA), /* 03715 - Richard Joseph - Cadaver.sndh */
TIMEDB_ENTRY( eaac2a5f,  0,   2009,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  1,    319,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  2,   5472,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  3,   7062,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  4,   5121,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  5,   8962,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  6,   7389,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  7,    177,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  8,   2160,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eaac2a5f,  9,   2884,          YM), /* 02948 - Carlo Perconti - Titus the Fox.sndh */
TIMEDB_ENTRY( eab3cd98,  0,    662,          YM), /* 00461 - Augustin Vidovic - Robotor.sndh */
TIMEDB_ENTRY( eac5e0ed,  0,   8061,          YM), /* 01903 - Luebke - Reverse Pong.sndh */
TIMEDB_ENTRY( ead4a93a,  0,   8826,    YM+TA+TD), /* 01139 - Tyan - Julia 3.sndh */
TIMEDB_ENTRY( eaefed23,  0,   6680,          YM), /* 02123 - Mad Max - Cuddly Demos - Main Menu.sndh */
TIMEDB_ENTRY( eafa7d19,  0,  11039, YM+TA+TB+TD), /* 04757 - Shinobi - Biomechanoid Locomotion.sndh */
TIMEDB_ENTRY( eb185808,  0,   8505,          YM), /* 03076 - Tom Marvin - CLOSE ENCOUNT.OF THE 3rd KIND.sndh */
TIMEDB_ENTRY( eb21c301,  0,   1761,          YM), /* 04771 - V0yager - Boobswatch theme.sndh */
TIMEDB_ENTRY( eb6719a5,  0,  10176,          YM), /* 03006 - New Frequency - ZPC TLT Demo.sndh */
TIMEDB_ENTRY( eb6719a5,  1,   2339,          YM), /* 03006 - New Frequency - ZPC TLT Demo.sndh */
TIMEDB_ENTRY( eb683fa4,  0,   8448, YM+TA+TB+TD), /* 02077 - Mad Max - Formula One - Sid.sndh */
TIMEDB_ENTRY( eb6a3a4c,  0,    129,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  1,     91,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  2,     64,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  3,    147,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  4,   1289,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  5,   1127,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  6,    612,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  7,    345,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb6a3a4c,  8,   2048,          YM), /* 00100 - David Whittaker - Fire and Brimstone.sndh */
TIMEDB_ENTRY( eb89bc02,  0,  17305,          YM), /* 02244 - Tao - Cream (orig. Rob Hubbard) - Sanxion Loader (Stmygm).sndh */
TIMEDB_ENTRY( eb919aff,  0,  28895,    YM+TA+TD), /* 02747 - Marcer - Bellanotte chippy 200hz.sndh */
TIMEDB_ENTRY( eba9311a,  0,   2305,          YM), /* 00394 - Wally Beben - Street Fighter.sndh */
TIMEDB_ENTRY( eba9311a,  1,   3841,          YM), /* 00394 - Wally Beben - Street Fighter.sndh */
TIMEDB_ENTRY( eba9311a,  2,    236,          YM), /* 00394 - Wally Beben - Street Fighter.sndh */
TIMEDB_ENTRY( eba9311a,  3,    146,          YM), /* 00394 - Wally Beben - Street Fighter.sndh */
TIMEDB_ENTRY( eba9311a,  4,    322,          YM), /* 00394 - Wally Beben - Street Fighter.sndh */
TIMEDB_ENTRY( ebb24e39,  0,  70983,       YM+TA), /* 03780 - Richard Joseph - International 3D Tennis Main.sndh */
TIMEDB_ENTRY( ebc7a679,  0,   1205, YM+TA+TB+TD), /* 03025 - Alchemist - Rare (Endpart).sndh */
TIMEDB_ENTRY( ebd6394d,  0,  30208, YM+TA+TB+TD), /* 03398 - cerror - CRANBERRY PT2.sndh */
TIMEDB_ENTRY( ebe46c52,  0,   4602,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  1,   2010,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  2,   2298,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  3,   5368,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  4,   5370,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  5,   4410,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  6,   1530,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe46c52,  7,   2352,          YM), /* 00016 - 4-Mat - Shadow Dancer.sndh */
TIMEDB_ENTRY( ebe4ce3b,  0,   6036,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  1,  31320,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  2,  10842,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  3,   9960,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  5,  23304,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  6,  18876,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  7,   9216,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  8,   9216,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b,  9,  24696,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b, 10,  11000,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b, 12,  16896,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b, 13,  17664,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b, 14,  42840,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ebe4ce3b, 15,  17664,          YM), /* 00732 - Sean Connolly - Elvira - The Arcade Game.sndh */
TIMEDB_ENTRY( ec047ff8,  0,   2564,          YM), /* 00112 - Kid Gloves - David Whittaker.sndh */
TIMEDB_ENTRY( ec047ff8,  1,    260,          YM), /* 00112 - Kid Gloves - David Whittaker.sndh */
TIMEDB_ENTRY( ec047ff8,  2,     97,          YM), /* 00112 - Kid Gloves - David Whittaker.sndh */
TIMEDB_ENTRY( ec047ff8,  3,    592,          YM), /* 00112 - Kid Gloves - David Whittaker.sndh */
TIMEDB_ENTRY( ec05b698,  0,   9696,          YM), /* 00460 - Augustin Vidovic - Rainbow.sndh */
TIMEDB_ENTRY( ec1f4791,  0,   3455, YM+TA+TB+TD), /* 02733 - maktone-fairlight - total.sndh */
TIMEDB_ENTRY( ec1f8fe4,  0,  23520,    YM+TB+TD), /* 02558 - Marcer - CONAN REMIX.sndh */
TIMEDB_ENTRY( ec3216b6,  0,  71585,       YM+TA), /* 04047 - Roggie - STC.sndh */
TIMEDB_ENTRY( ec33206c,  0,  12640,          YM), /* 01509 - Rob Hubbard - Thundercats.sndh */
TIMEDB_ENTRY( ec4e9b57,  0,   4223,          YM), /* 00370 - Beast - TSC - Basstep 2.sndh */
TIMEDB_ENTRY( ec85260c,  0,   4100,          YM), /* 01926 - The Mind - Genesys Remix - Monochrome.sndh */
TIMEDB_ENTRY( ec9c9cd3,  0,    511,          YM), /* 04093 - Gary Wheaton - Sea Fisherman.sndh */
TIMEDB_ENTRY( ecae2cec,  0,  11520,    YM+TA+TD), /* 03315 - SH3-RG - Blakkel.sndh */
TIMEDB_ENTRY( ecb560f8,  0,  36864,          YM), /* 01456 - Stax - X Large.sndh */
TIMEDB_ENTRY( ecb7bebc,  0,  31744,          YM), /* 02803 - Starbuck - Logical Lvl.3.sndh */
TIMEDB_ENTRY( ecce9105,  0,   1437,       YM+TD), /* 02382 - Nexus 6 - Lethal Xcess Level 4.sndh */
TIMEDB_ENTRY( ecce9105,  1,   8480,       YM+TD), /* 02382 - Nexus 6 - Lethal Xcess Level 4.sndh */
TIMEDB_ENTRY( ecce9105,  2,   2176,       YM+TD), /* 02382 - Nexus 6 - Lethal Xcess Level 4.sndh */
TIMEDB_ENTRY( ecce9105,  3,   1152,       YM+TD), /* 02382 - Nexus 6 - Lethal Xcess Level 4.sndh */
TIMEDB_ENTRY( ecd449f5,  0,   3120,          YM), /* 01068 - Jonathon Dunn - Simpsons - Bart vs Space Mutants.sndh */
TIMEDB_ENTRY( ece92dba,  0,  14332,       YM+TA), /* 01285 - Frequent - Evila.sndh */
TIMEDB_ENTRY( ecf43517,  0,   9408,          YM), /* 01970 - MSG - MSG 09 - Maggie 27 Intro.sndh */
TIMEDB_ENTRY( ecf9567a,  0,  13818, YM+TA+TB+TD), /* 00986 - D-Force - Treat.sndh */
TIMEDB_ENTRY( ed033f60,  0,   5168, YM+TA+TB+TD), /* 00179 - 505 - Mystic Bytes 3.sndh */
TIMEDB_ENTRY( ed2f2895,  0,   2096,          YM), /* 01499 - Alex Herbert - Xrun.sndh */
TIMEDB_ENTRY( ed33fd77,  0,   3839, YM+TA+TB+TD), /* 03020 - dubmood - RaZoR 1911 - Nukes Chiptune.sndh */
TIMEDB_ENTRY( ed35355e,  0,   8446,          YM), /* 00372 - Beast - TSC - Beast 2.sndh */
TIMEDB_ENTRY( ed440180,  0,  14397,       YM+TA), /* 04868 - Jovis - Plucking.sndh */
TIMEDB_ENTRY( ed4ac7a6,  0,  16696,          YM), /* 02051 - Mad Max - Sanxion Loader.sndh */
TIMEDB_ENTRY( ed5b98db,  0,    667,       YM+TA), /* 03838 - Fabrice Decroix - Altair Title.sndh */
TIMEDB_ENTRY( ed704505,  0,  19988,          YM), /* 03237 - T1000 - The Beast Club - TBC 9.sndh */
TIMEDB_ENTRY( ed76828a,  0,  14507, YM+TA+TB+TD), /* 02093 - Mad Max - Saboteur 2 - Sid.sndh */
TIMEDB_ENTRY( eda03d6e,  0,    314,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  1,    587,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  2,    139,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  3,    156,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  4,    156,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  5,    795,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  6,    890,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  7,    826,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  8,    173,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e,  9,    201,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e, 10,    570,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( eda03d6e, 11,    186,          YM), /* 01294 - Matt Furniss - Cyberball.sndh */
TIMEDB_ENTRY( edc369e1,  0,   5119, YM+TA+TB+TD), /* 02835 - maktone-fairlight - granit.sndh */
TIMEDB_ENTRY( edc4b3cb,  0,   1536,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  1,   1152,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  2,    422,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  3,    205,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  4,   2304,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  5,   1920,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  6,   2688,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  7,     91,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  8,    121,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb,  9,    125,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb, 10,   3072,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb, 11,     70,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb, 12,   1556,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb, 13,   1728,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edc4b3cb, 14,   1716,          YM), /* 02792 - Keith Tinman - Hudson Hawk.sndh */
TIMEDB_ENTRY( edccb89c,  0,   2280,       YM+TB), /* 02499 - gwEm (Prodigy conversion) - Underground... Like the Wombles.sndh */
TIMEDB_ENTRY( edf8769a,  0,   9216,          YM), /* 01345 - Scavenger - Menu (Megaleif Memories) - Synergy Megademo.sndh */
TIMEDB_ENTRY( ee1f4380,  0,   2120,          YM), /* 01807 - Barry Leitch - Stratego.sndh */
TIMEDB_ENTRY( ee31ab5f,  0,   3840,       YM+TA), /* 00404 - Big Alec - Galaxy.sndh */
TIMEDB_ENTRY( ee5b2513,  0,   1791,       YM+TD), /* 03289 - Dma-Sc - ATIPYK Developments - Manical Drop - 2 Players music #2.sndh */
TIMEDB_ENTRY( ee692bca,  0,   2304,       YM+TA), /* 02951 - Chris Scudds - Cisco Heat (STfm).sndh */
TIMEDB_ENTRY( ee692bca,  1,   2688,       YM+TA), /* 02951 - Chris Scudds - Cisco Heat (STfm).sndh */
TIMEDB_ENTRY( ee7330c7,  0,  27644,          YM), /* 01709 - Jess - I Want U 2 Shoot'em UP.sndh */
TIMEDB_ENTRY( ee7447f1,  0,   8320,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  1,   2496,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  2,   1600,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  3,   1293,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  4,   2730,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  5,   2497,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  6,   1542,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  7,   1634,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  8,   2178,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1,  9,   2497,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 10,   3138,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 11,   2560,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 12,   2688,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 13,    512,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 14,   2560,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 15,    752,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 16,    872,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 17,    940,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 18,    512,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 19,   1536,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 20,   1828,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 21,   1828,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 22,   1280,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 23,   1536,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 24,    590,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 25,    369,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 26,    335,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 27,    320,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 28,    149,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee7447f1, 29,     64,          YM), /* 04699 - D. Haylett - Moneymania 2.sndh */
TIMEDB_ENTRY( ee852a96,  0,   7161, YM+TA+TB+TD), /* 04239 - 505-checkpoint - AEH.sndh */
TIMEDB_ENTRY( ee941dca,  0,  21520, YM+TA+TB+TD), /* 03988 - XyNo - n-a.sndh */
TIMEDB_ENTRY( eebde3ab,  0,  15350,          YM), /* 01744 - Jovis - Lametrop - Intro Part 2.sndh */
TIMEDB_ENTRY( eec4e1ee,  0,   3834,    YM+TA+TB), /* 01346 - Scavenger - Meltdis.sndh */
TIMEDB_ENTRY( eec642e3,  0,  27216, YM+TA+TB+TD), /* 03578 - Marcer - n-a.sndh */
TIMEDB_ENTRY( eef0e1d0,  0,   2386,          YM), /* 03387 - Unknown - Tango.sndh */
TIMEDB_ENTRY( eef1808d,  0,  14081, YM+TA+TB+TD), /* 04332 - mOdmate-CheckPoint (orig. David Whittaker) - STORM REVISITED.sndh */
TIMEDB_ENTRY( ef047ba0,  0,  51520, YM+TA+TB+TD), /* 03312 - TomChi - Noextra - PICROCCINELLE.sndh */
TIMEDB_ENTRY( ef139890,  0,   3712,          YM), /* 01512 - Chris Huelsbeck - Gem'X.sndh */
TIMEDB_ENTRY( ef139890,  1,   4096,          YM), /* 01512 - Chris Huelsbeck - Gem'X.sndh */
TIMEDB_ENTRY( ef139890,  2,   4096,          YM), /* 01512 - Chris Huelsbeck - Gem'X.sndh */
TIMEDB_ENTRY( ef139890,  3,    267,          YM), /* 01512 - Chris Huelsbeck - Gem'X.sndh */
TIMEDB_ENTRY( ef1693ba,  0,   9213,       YM+TA), /* 04865 - Jovis - Cannon F.sndh */
TIMEDB_ENTRY( ef255e42,  0,   8257,          YM), /* 03884 - Unknown - Andalusa.sndh */
TIMEDB_ENTRY( ef4650fc,  0,  21332,       YM+TA), /* 01743 - Jovis - Lametrop - End Part 2.sndh */
TIMEDB_ENTRY( ef4de347,  0,  21792, YM+TA+TB+TD), /* 04400 - Skaven (ST adaption by mOdmate) - Second Reality 2013 intro.sndh */
TIMEDB_ENTRY( ef63b160,  0,   5654,          YM), /* 03035 - Barry Deutsch - SANTA CLAUS Arr. by B. Deutsch.sndh */
TIMEDB_ENTRY( ef770201,  0,   5375, YM+TA+TB+TD), /* 02699 - Dubmood & Zabutom - Pampradion livebeat.sndh */
TIMEDB_ENTRY( ef92d231,  0,   5760,          YM), /* 01820 - Lord - Ambition Intro 2.sndh */
TIMEDB_ENTRY( efaca7a4,  0,   2194,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  1,   3680,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  2,   6582,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  3,   4608,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  4,   1600,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  5,   2560,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  6,   2590,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  7,    337,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  8,    150,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4,  9,     39,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4, 10,     36,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4, 11,     52,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efaca7a4, 12,     41,          YM), /* 01493 - Hard N Soft - Steg The Slug.sndh */
TIMEDB_ENTRY( efcc5dd3,  0,   3840,          YM), /* 00601 - Chris Mad - Overdrive Nybble.sndh */
TIMEDB_ENTRY( effb0c76,  0,  13377, YM+TA+TB+TD), /* 04184 - Dubmood - Liberators - Kick De Bucket.sndh */
TIMEDB_ENTRY( effc756e,  0,   5119,    YM+TB+TD), /* 04759 - Shinobi - Gems in Montana.sndh */
TIMEDB_ENTRY( f0101dd3,  0,  10236, YM+TA+TB+TD), /* 00992 - DMA-SC - Clean Closet - UCM 19.sndh */
TIMEDB_ENTRY( f0127a5b,  0,   2306,          YM), /* 00728 - Tebirod - No Cooper Reset.sndh */
TIMEDB_ENTRY( f01a3949,  0,   6144,          YM), /* 02796 - Starbuck - Alf.sndh */
TIMEDB_ENTRY( f028a321,  0,   2598,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  1,    174,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  2,    375,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  3,    147,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  4,     39,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  5,    153,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  6,     54,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  7,    177,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  8,    123,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321,  9,     18,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 10,    135,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 11,     69,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 12,     96,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 13,     63,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 14,     48,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 15,     36,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f028a321, 16,     39,          YM), /* 04736 - Al Lowe - Kings Quest - Quest for the Crown.sndh */
TIMEDB_ENTRY( f0309cdb,  0,  37317,       YM+TA), /* 04018 - Roggie - Legacy.sndh */
TIMEDB_ENTRY( f04c02ac,  0,    768,          YM), /* 01910 - Marcer - Not Ready.sndh */
TIMEDB_ENTRY( f0529ea6,  0,   6912,          YM), /* 02041 - Mad Max - The Last V8.sndh */
TIMEDB_ENTRY( f05369a4,  0,  33205, YM+TA+TB+TD), /* 03379 - STU - SUCH A SQUARE.sndh */
TIMEDB_ENTRY( f0665a0f,  0,   6133,          YM), /* 04329 - MCH and Mermaid - Elefantastic.sndh */
TIMEDB_ENTRY( f085566c,  0,   3835,          YM), /* 00027 - 505 - 19th.sndh */
TIMEDB_ENTRY( f093b98b,  0,  14586, YM+TA+TB+TD), /* 01011 - DMA-SC - Shining Dance.sndh */
TIMEDB_ENTRY( f0a7c121,  0,   9986,          YM), /* 00423 - Big Alec - PYM - Colours Go Bang.sndh */
TIMEDB_ENTRY( f0b2c86e,  0,  14591, YM+TA+TB+TD), /* 02488 - gwEm (Spaz^TLB conversion) - Crikey Wot a Tune!.sndh */
TIMEDB_ENTRY( f0ba9bde,  0,   7674,    YM+TB+TD), /* 00995 - Dma-Sc-Sector1^YM-Rockerz - Fantasia Tune 2.sndh */
TIMEDB_ENTRY( f0be2ab0,  0,  71891,       YM+TA), /* 03966 - Unknown - Sanxion Loader.sndh */
TIMEDB_ENTRY( f0e33ae4,  0,   7424,          YM), /* 03190 - n-a - Improvisation.sndh */
TIMEDB_ENTRY( f10cf937,  0,   1619,          YM), /* 03168 - Unknown - The Spider Dance - Tarantelle.sndh */
TIMEDB_ENTRY( f1211a3c,  0,    225,          YM), /* 04454 - Mike Davies - Human Killing Machine.sndh */
TIMEDB_ENTRY( f1211a3c,  1,    289,          YM), /* 04454 - Mike Davies - Human Killing Machine.sndh */
TIMEDB_ENTRY( f1211a3c,  2,  10080,          YM), /* 04454 - Mike Davies - Human Killing Machine.sndh */
TIMEDB_ENTRY( f1211a3c,  3,   6912,          YM), /* 04454 - Mike Davies - Human Killing Machine.sndh */
TIMEDB_ENTRY( f124c30b,  0,   5760,          YM), /* 02397 - Nexus 6 - Speherical Harmonies.sndh */
TIMEDB_ENTRY( f12ff40a,  0,   5427,          YM), /* 00092 - David Whittaker - Dan Dare III  - The Escape.sndh */
TIMEDB_ENTRY( f12ff40a,  1,   1152,          YM), /* 00092 - David Whittaker - Dan Dare III  - The Escape.sndh */
TIMEDB_ENTRY( f13a9be3,  0,   7298,          YM), /* 01053 - Dr Beat - Imagination - Snake.sndh */
TIMEDB_ENTRY( f141ce45,  0,  30720, YM+TA+TB+TD), /* 04571 - mOdmate-CheckPoint (orig. Mad Max-TEX) - Reassessing Teramis (Pokey Flavour).sndh */
TIMEDB_ENTRY( f1441012,  0,   7999,          YM), /* 03678 - Dma-Sc - Antiques - Ending.sndh */
TIMEDB_ENTRY( f14a3936,  0,   4468,          YM), /* 00160 - David Whittaker - Zombi.sndh */
TIMEDB_ENTRY( f14a3936,  1,  11614,          YM), /* 00160 - David Whittaker - Zombi.sndh */
TIMEDB_ENTRY( f14a3936,  2,     96,          YM), /* 00160 - David Whittaker - Zombi.sndh */
TIMEDB_ENTRY( f154403e,  0,  13416,       YM+TA), /* 04014 - Mad Max - Thalion Forever.sndh */
TIMEDB_ENTRY( f1893d7b,  0,  16896,          YM), /* 01090 - Techno - Fantasy - Hidden Screen.sndh */
TIMEDB_ENTRY( f1a269e0,  0,  75652,       YM+TA), /* 03959 - Scott Lemmon - RSMTST.sndh */
TIMEDB_ENTRY( f1a27de5,  0,  33990,       YM+TA), /* 02809 - Starbuck^Creative Mind - Ready to fight (Lightwarriorz 2).sndh */
TIMEDB_ENTRY( f1a2c561,  0,   6396,    YM+TA+TD), /* 00037 - 505 - Brutal Act.sndh */
TIMEDB_ENTRY( f1a5e780,  0,    386,          YM), /* 04498 - Reiner Glock - Biker.sndh */
TIMEDB_ENTRY( f1a5e780,  1,   1631,          YM), /* 04498 - Reiner Glock - Biker.sndh */
TIMEDB_ENTRY( f1ac7701,  0,   6191,          YM), /* 03689 - Dma-Sc - Saboteur 3 - In Game 2.sndh */
TIMEDB_ENTRY( f1c60210,  0,  10367, YM+TA+TB+TD), /* 03278 - Frequent-Ephidrena - Gatelys.sndh */
TIMEDB_ENTRY( f21858c3,  0,  41568,          YM), /* 01436 - Stax - Dull.sndh */
TIMEDB_ENTRY( f21e9f18,  0,  10304,          YM), /* 01560 - Dave Rogers - Zynaps.sndh */
TIMEDB_ENTRY( f21e9f18,  1,    640,          YM), /* 01560 - Dave Rogers - Zynaps.sndh */
TIMEDB_ENTRY( f2335193,  0,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  1,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  2,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  3,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  4,     34,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  5,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  6,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  7,     34,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  8,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193,  9,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 10,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 11,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 12,     31,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 13,     34,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 15,    563,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 16,    495,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 17,    563,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 18,    483,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 19,    772,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 20,    740,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 21,    502,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 22,    755,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 23,    467,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 24,   1262,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 25,    701,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2335193, 26,   1536,          YM), /* 04701 - Mikael Lundberg - Mushy Music.sndh */
TIMEDB_ENTRY( f2375bd6,  0,   1536,          YM), /* 01409 - Skyline - Snork Intro - Ooh Crikey.sndh */
TIMEDB_ENTRY( f2396e36,  0,   9022,       YM+TA), /* 01517 - Iso - ...And..Action!!.sndh */
TIMEDB_ENTRY( f23dcad2,  0,  13440,          YM), /* 00482 - Ultimatum - Count Zero 2.sndh */
TIMEDB_ENTRY( f23f63fc,  0,  23034,      YM+STE), /* 01676 - Jess - My First Resonant Burp (STe).sndh */
TIMEDB_ENTRY( f245e08a,  0,   3920,          YM), /* 03297 - gwEm - 1K_ller.sndh */
TIMEDB_ENTRY( f24e160d,  0,  36630,    YM+TA+TD), /* 02306 - Tao - Cream - Competition (EIL99).sndh */
TIMEDB_ENTRY( f2585808,  0,  13123, YM+TA+TB+TD), /* 03594 - Dma-Sc - YMetal.sndh */
TIMEDB_ENTRY( f2721c4c,  0,   3332,         STE), /* 03449 - 505 - Paradox - Direct Colour Zoomer.sndh */
TIMEDB_ENTRY( f278424e,  0,   1093,          YM), /* 04146 - Unknown - Baall2.sndh */
TIMEDB_ENTRY( f278424e,  1,     76,          YM), /* 04146 - Unknown - Baall2.sndh */
TIMEDB_ENTRY( f2798b85,  0,   8552,          YM), /* 04091 - Gary Wheaton - Scarborough Fair.sndh */
TIMEDB_ENTRY( f2809078,  0,  12909, YM+TA+TB+TD), /* 01625 - No More - Risk.sndh */
TIMEDB_ENTRY( f286a7c8,  0,  34054,       YM+TA), /* 04666 - Kijop - Axel F.sndh */
TIMEDB_ENTRY( f288c2be,  0,   7939,          YM), /* 00270 - Archimede - Cosmic Pirate Empire Intro.sndh */
TIMEDB_ENTRY( f28910f2,  0,   3168,    YM+TA+TD), /* 02316 - Tao - Cream (orig. Klaus Ude) - Lin Wu's Challenge 3.sndh */
TIMEDB_ENTRY( f28bd288,  0,   7561,          YM), /* 00171 - Martin Walker - Reach for The Skies.sndh */
TIMEDB_ENTRY( f28bd288,  1,   2640,          YM), /* 00171 - Martin Walker - Reach for The Skies.sndh */
TIMEDB_ENTRY( f28bd288,  2,   3116,          YM), /* 00171 - Martin Walker - Reach for The Skies.sndh */
TIMEDB_ENTRY( f28bd288,  3,   1854,          YM), /* 00171 - Martin Walker - Reach for The Skies.sndh */
TIMEDB_ENTRY( f292e42e,  0,   3269,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  1,    773,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  2,   2725,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  3,    389,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  4,   5765,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  5,   6077,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  6,   6445,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  7,   6149,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  8,   5680,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e,  9,   5669,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e, 10,   6053,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e, 11,   6245,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e, 12,   1365,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e, 13,    383,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f292e42e, 14,   2965,          YM), /* 01261 - Tim Follin - Sly Spy.sndh */
TIMEDB_ENTRY( f29318d3,  0,  59760,       YM+TA), /* 03228 - 505 - Pacemaker.sndh */
TIMEDB_ENTRY( f299bba7,  0,   1535,       YM+TA), /* 04777 - Excellence in Art - Who Made You Judge Judy And Executioner?.sndh */
TIMEDB_ENTRY( f29d28b0,  0,  29953,       YM+TA), /* 04821 - Twitch - Only You.sndh */
TIMEDB_ENTRY( f2af4cd0,  0,   9402,          YM), /* 01551 - Rhino - PMP Competition Entry.sndh */
TIMEDB_ENTRY( f2afe226,  0,   4984, YM+TA+TB+TD), /* 01849 - Lotek Style - Giana Sister Title Remix.sndh */
TIMEDB_ENTRY( f2c7cdc6,  0,  13006,       YM+TA), /* 02770 - !Cube - Stax Menu 78.sndh */
TIMEDB_ENTRY( f2e06124,  0,   9792,    YM+TA+TD), /* 02787 - Mad Max - 2006.sndh */
TIMEDB_ENTRY( f2ed13f3,  0,   5376, YM+TA+TB+TD), /* 03668 - Baggio - Compofiller 3000.sndh */
TIMEDB_ENTRY( f30f12fb,  0,  16384,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  1,  16596,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  2,   8192,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  3,  17712,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  4,  16128,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  5,  17520,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  6,  10368,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f30f12fb,  7,  10880,          YM), /* 04464 - Charles Callet - L'Affaire.sndh */
TIMEDB_ENTRY( f3193839,  0,    508,    YM+TA+TD), /* 02425 - 505-cpt^dox - loop01.sndh */
TIMEDB_ENTRY( f31b12f8,  0,   5534,          YM), /* 04076 - Gary Wheaton - Like a Hurricane.sndh */
TIMEDB_ENTRY( f3240c01,  0,   3552,          YM), /* 01203 - Excellence In Art - Summer.sndh */
TIMEDB_ENTRY( f32dd5b3,  0,  60930,       YM+TA), /* 04811 - Unknown - Spr.sndh */
TIMEDB_ENTRY( f35b9845,  0,   1686,          YM), /* 03184 - n-a - William Tell Overture.sndh */
TIMEDB_ENTRY( f363560d,  0,  28160, YM+TA+TB+TD), /* 02601 - STU - LAST MINUTE.sndh */
TIMEDB_ENTRY( f3637837,  0,   4475, YM+TA+TB+TD), /* 04322 - 505 - Three.sndh */
TIMEDB_ENTRY( f365698b,  0,   9600, YM+TA+TB+TD), /* 03026 - !Cube - Agent Hippel To The Rescue.sndh */
TIMEDB_ENTRY( f36b02f9,  0,   5898,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  1,    210,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  2,   1780,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  3,   7835,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  4,    334,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  5,   2730,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  6,    687,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f36b02f9,  7,   1820,          YM), /* 00578 - Jas C Brooke - Time Soldier.sndh */
TIMEDB_ENTRY( f371e8a6,  0,   2069,          YM), /* 03068 - Chris French - Jingle Bells.sndh */
TIMEDB_ENTRY( f37a6157,  0,   6912,          YM), /* 02002 - Mad Max - Demo Music 8.sndh */
TIMEDB_ENTRY( f37a73b3,  0,   5327,    YM+TA+TB), /* 00215 - 505 - Zirkuss.sndh */
TIMEDB_ENTRY( f3992b60,  0,   8824,       YM+TA), /* 02502 - lotekstyle - Spieler 1, Los!.sndh */
TIMEDB_ENTRY( f3acc758,  0,   2400,    YM+TA+TD), /* 00411 - Big Alec - Magick.sndh */
TIMEDB_ENTRY( f3b3dc75,  0,   3120,          YM), /* 00152 - David Whittaker - Warlock The Avenger.sndh */
TIMEDB_ENTRY( f3b4254a,  0,   5771,       YM+TD), /* 01614 - No More - Groovy.sndh */
TIMEDB_ENTRY( f3be48d8,  0,  12192,          YM), /* 02027 - Mad Max - Deep Strike.sndh */
TIMEDB_ENTRY( f3c6e8aa,  0,  12767, YM+TA+TB+TD), /* 03019 - Dubmood - Loonies Biohazard.sndh */
TIMEDB_ENTRY( f3fed517,  0,   4575,    YM+TB+TD), /* 03293 - Dma-Sc - ATIPYK Developments - Manical Drop - 1 Player music #2.sndh */
TIMEDB_ENTRY( f413fc38,  0,   1921,          YM), /* 04155 - Unknown - Orfana.sndh */
TIMEDB_ENTRY( f41645d7,  0,   1150,          YM), /* 01528 - Iso - Houseman.sndh */
TIMEDB_ENTRY( f4204a3f,  0,  11951,       YM+TA), /* 00812 - Crawdaddy - Suffer In Hell.sndh */
TIMEDB_ENTRY( f42539ef,  0,  71402,       YM+TA), /* 03809 - Martin Walker - Gettysburg.sndh */
TIMEDB_ENTRY( f4570dd1,  0,   5600,    YM+TA+TD), /* 02571 - !Cube-Aggression - Water.sndh */
TIMEDB_ENTRY( f45e0431,  0,  65185,       YM+TA), /* 03950 - Scott Lemmon - Fish.sndh */
TIMEDB_ENTRY( f4602101,  0,  18240,       YM+TA), /* 02182 - Mad Max - Bad Cat.sndh */
TIMEDB_ENTRY( f46d5b97,  0,   5396,       YM+TA), /* 04008 - Per Almered-XiA-Checkpoint - Idas Sommarvisa (G Riedel).sndh */
TIMEDB_ENTRY( f471ff46,  0,   2305,          YM), /* 03889 - Unknown - Menuet.sndh */
TIMEDB_ENTRY( f476f830,  0,   3264,          YM), /* 04835 - t1000 - Bits 29.sndh */
TIMEDB_ENTRY( f47ce127,  0,  12640,          YM), /* 00760 - Count Zero - Hidden Easter Egg.sndh */
TIMEDB_ENTRY( f490c357,  0,  10620,          YM), /* 00349 - Tony Williams - Sharkeys Moll.sndh */
TIMEDB_ENTRY( f490c357,  1,    193,          YM), /* 00349 - Tony Williams - Sharkeys Moll.sndh */
TIMEDB_ENTRY( f490c357,  2,    793,          YM), /* 00349 - Tony Williams - Sharkeys Moll.sndh */
TIMEDB_ENTRY( f493813b,  0,    518,          YM), /* 01494 - Havoc - This Sux.sndh */
TIMEDB_ENTRY( f49b7214,  0,   8448,       YM+TA), /* 04418 - EVE - Serious.sndh */
TIMEDB_ENTRY( f4a2cb20,  0,   2048,          YM), /* 00135 - David Whittaker - Snoopy.sndh */
TIMEDB_ENTRY( f4a2cb20,  1,   1877,          YM), /* 00135 - David Whittaker - Snoopy.sndh */
TIMEDB_ENTRY( f4a2cb20,  2,    145,          YM), /* 00135 - David Whittaker - Snoopy.sndh */
TIMEDB_ENTRY( f4a2cb20,  3,    145,          YM), /* 00135 - David Whittaker - Snoopy.sndh */
TIMEDB_ENTRY( f4b5c9f5,  0,   4096,          YM), /* 02932 - Chris Denman - Crime City.sndh */
TIMEDB_ENTRY( f4b5c9f5,  1,   2560,          YM), /* 02932 - Chris Denman - Crime City.sndh */
TIMEDB_ENTRY( f4b5c9f5,  2,   1024,          YM), /* 02932 - Chris Denman - Crime City.sndh */
TIMEDB_ENTRY( f4b5c9f5,  3,   1152,          YM), /* 02932 - Chris Denman - Crime City.sndh */
TIMEDB_ENTRY( f4b5c9f5,  4,   1024,          YM), /* 02932 - Chris Denman - Crime City.sndh */
TIMEDB_ENTRY( f4be8784,  0,  12840,          YM), /* 01642 - Jedi - Beeforol.sndh */
TIMEDB_ENTRY( f4d4c835,  0,  44456,          YM), /* 00005 - Gwem - Raid The Tower (YM Rockerz).sndh */
TIMEDB_ENTRY( f4f9c1ba,  0,   1919,       YM+TD), /* 03626 - Gabriel Yared-XiA - Betty.sndh */
TIMEDB_ENTRY( f4fc3f7c,  0,   9340,          YM), /* 01233 - Epic - Magic Lame.sndh */
TIMEDB_ENTRY( f4fe548e,  0,  15401,          YM), /* 03079 - Ron Robinson - Bridge Over Troubled Water.sndh */
TIMEDB_ENTRY( f4ff62c1,  0,   3072,          YM), /* 01728 - John Michael Phillips - Nebulus.sndh */
TIMEDB_ENTRY( f51be010,  0,   1275,    YM+TA+TB), /* 03502 - 505 - Suretrip 2 - Stars.sndh */
TIMEDB_ENTRY( f51f297f,  0,  33518,       YM+TA), /* 04801 - Unknown - Flash.sndh */
TIMEDB_ENTRY( f52bd8d5,  0,  12800,       YM+TA), /* 01546 - Ralph - Ralph Digi 3.sndh */
TIMEDB_ENTRY( f533be70,  0,   4928,          YM), /* 00296 - Gary Wheaton - Aru.sndh */
TIMEDB_ENTRY( f53f462b,  0,   4896,       YM+TA), /* 04587 - Technoblip - Groovin' Nostrils.sndh */
TIMEDB_ENTRY( f5482637,  0,    909,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  1,   1914,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  2,    322,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  3,    181,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  4,    306,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  5,   2885,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  6,    781,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  7,    538,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  8,    221,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637,  9,    661,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637, 10,    592,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637, 11,     99,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637, 12,   2682,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f5482637, 14,   1108,          YM), /* 04738 - Margaret Lowe - Kings Quest 3 - The Heir is Human.sndh */
TIMEDB_ENTRY( f548f91d,  0,   1536,          YM), /* 01102 - Techno - Techno 7.sndh */
TIMEDB_ENTRY( f54dd8a5,  0,   1536,          YM), /* 01413 - Skyline - Snork Intro - Syntax Terror.sndh */
TIMEDB_ENTRY( f550d110,  0,   8471, YM+TA+TB+TD), /* 02778 - !Cube - Back To Basics.sndh */
TIMEDB_ENTRY( f55d8144,  0,   4968,          YM), /* 03175 - Unknown - Talk To Me - by Stevie Nicks.sndh */
TIMEDB_ENTRY( f55e7bae,  0,   4473,       YM+TA), /* 01689 - Jess - Guile.sndh */
TIMEDB_ENTRY( f58379d5,  0,  20928,          YM), /* 01796 - Barry Leitch - Hero Quest.sndh */
TIMEDB_ENTRY( f58379d5,  1,  10864,          YM), /* 01796 - Barry Leitch - Hero Quest.sndh */
TIMEDB_ENTRY( f58379d5,  2,   8448,          YM), /* 01796 - Barry Leitch - Hero Quest.sndh */
TIMEDB_ENTRY( f58c1b47,  0,   7680,    YM+TA+TD), /* 02428 - 505 - Triplex 2.sndh */
TIMEDB_ENTRY( f5abd6a1,  0,  13055, YM+TA+TB+TD), /* 02777 - timbral-dhs-ymr - 2006 - outrun highway E6.sndh */
TIMEDB_ENTRY( f5b3a422,  0,   8960,          YM), /* 03467 - Marcer - Treasure.sndh */
TIMEDB_ENTRY( f5c0d269,  0,   7852,    YM+TA+TB), /* 00209 - 505 - UCM 11 Intro Endpart.sndh */
TIMEDB_ENTRY( f5de8f29,  0,   4992,          YM), /* 04857 - Big Alec - Pro BMX Simulator (cover).sndh */
TIMEDB_ENTRY( f5e39c6c,  0,   3840,          YM), /* 01058 - Jonathan Dunn - Elf - Menu.sndh */
TIMEDB_ENTRY( f5eb1ba7,  0,   3450,       YM+TA), /* 02641 - Miggy - Poker(Unfinished).sndh */
TIMEDB_ENTRY( f5f03f5d,  0,   9600,          YM), /* 01580 - Nigel Pritchard - Netherworld.sndh */
TIMEDB_ENTRY( f606699b,  0,  20020, YM+TA+TB+TD), /* 03600 - Marcer - RADIO Summerhack Version.sndh */
TIMEDB_ENTRY( f61d2f66,  0,   3583,          YM), /* 03686 - Dma-Sc - Saboteur 3 - Gains.sndh */
TIMEDB_ENTRY( f6237591,  0,   6912,       YM+TA), /* 00782 - Crawdaddy - Chinese story.sndh */
TIMEDB_ENTRY( f62e8f62,  0,   3584,          YM), /* 00791 - Crawdaddy - Hidden.sndh */
TIMEDB_ENTRY( f647f698,  0,   8960,    YM+TB+TD), /* 03462 - Marcer - Pandora.sndh */
TIMEDB_ENTRY( f64f8919,  0,  39853,          YM), /* 04383 - Roderic Mege - Unreal.sndh */
TIMEDB_ENTRY( f65568b3,  0,    285,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  1,    160,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  2,    498,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  3,    267,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  4,    911,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  5,   2045,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65568b3,  6,   1017,          YM), /* 00859 - Ben Daglish - Legends of Valour.sndh */
TIMEDB_ENTRY( f65c50e0,  0,   6583, YM+TA+TB+TD), /* 02992 - 505-paradox - on behalf of the queen (STE only).sndh */
TIMEDB_ENTRY( f668e3f0,  0,  10471,    YM+TA+TD), /* 03647 - 505 - Base Case.sndh */
TIMEDB_ENTRY( f699428a,  0,   5412,    YM+TA+TD), /* 04316 - 505-checkpoint - synthpop.sndh */
TIMEDB_ENTRY( f6b1345a,  0,   3072,          YM), /* 04519 - Antichrist - Robocop.sndh */
TIMEDB_ENTRY( f6b1d5ee,  0,   6086,          YM), /* 03098 - n-a - Good Time Charlie.sndh */
TIMEDB_ENTRY( f6d73f8f,  0,   2562,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  1,    155,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  2,    116,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  3,   2189,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  4,   1120,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  5,   2362,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  6,    594,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  7,   1920,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  8,   2571,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6d73f8f,  9,    259,          YM), /* 04535 - Jan Holmgren - Frog Race.sndh */
TIMEDB_ENTRY( f6dbcd1a,  0,   3072,          YM), /* 00734 - Sean Connolly - Goat Beard End.sndh */
TIMEDB_ENTRY( f6e0e5fa,  0,   7161,       YM+TA), /* 02769 - !Cube - Eye on Scene Maggie Dentro.sndh */
TIMEDB_ENTRY( f6e2634e,  0,   9978, YM+TA+TB+TD), /* 02423 - 505 - Checkpoint - hybrid.sndh */
TIMEDB_ENTRY( f6e83698,  0,  36992, YM+TA+TB+TD), /* 03863 - xyce - Cpp is Forbidden - Part 1 (sid).sndh */
TIMEDB_ENTRY( f71a80a8,  0,  21760,       YM+TD), /* 02608 - STU - NAUGHTY BRAT.sndh */
TIMEDB_ENTRY( f72491f6,  0,   2700,          YM), /* 03146 - n-a - Off To School Boogie.sndh */
TIMEDB_ENTRY( f72e19e6,  0,   2298,          YM), /* 04273 - 505 - lame01.sndh */
TIMEDB_ENTRY( f7376718,  0,  16984,       YM+TA), /* 00725 - Tebirod - No Cooper End Screen.sndh */
TIMEDB_ENTRY( f7441a76,  0,    743,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  1,   1210,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  2,    640,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  3,    192,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  4,    448,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  5,    576,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  6,    768,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  7,    512,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  8,   1024,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76,  9,    704,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 10,   1645,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 11,    643,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 12,    576,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 13,    832,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 14,    256,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 15,   1554,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 16,    672,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 17,    377,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 18,    480,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f7441a76, 19,    896,          YM), /* 04491 - Unknown - Anti matter.sndh */
TIMEDB_ENTRY( f74f2e17,  0,   6144,          YM), /* 04445 - Iso - Smooth Talking.sndh */
TIMEDB_ENTRY( f791df96,  0,  31040,          YM), /* 02690 - Philip Linde - RAS ATARI.sndh */
TIMEDB_ENTRY( f7ce16da,  0,  33024,    YM+TB+TD), /* 04850 - xFalcon - LASER CHIP TALK.sndh */
TIMEDB_ENTRY( f7d382af,  0,  12288,          YM), /* 01082 - Excellence In Art - Elftec 2.sndh */
TIMEDB_ENTRY( f7ddc72f,  0,   7677,          YM), /* 03549 - Jovis - Zwaqwam.sndh */
TIMEDB_ENTRY( f7e262be,  0,  11901, YM+TA+TB+TD), /* 00050 - 505 - Fukk2.sndh */
TIMEDB_ENTRY( f7e41fdd,  0,   4212,          YM), /* 01394 - Paul Shields - Airball.sndh */
TIMEDB_ENTRY( f7e41fdd,  1,  20767,          YM), /* 01394 - Paul Shields - Airball.sndh */
TIMEDB_ENTRY( f7e41fdd,  2,   5681,          YM), /* 01394 - Paul Shields - Airball.sndh */
TIMEDB_ENTRY( f7e41fdd,  3,   1152,          YM), /* 01394 - Paul Shields - Airball.sndh */
TIMEDB_ENTRY( f7e41fdd,  4,   2204,          YM), /* 01394 - Paul Shields - Airball.sndh */
TIMEDB_ENTRY( f7f9c7da,  0,  52263,       YM+TA), /* 03777 - Stewart Gardner - Stampede 6.sndh */
TIMEDB_ENTRY( f7fe995e,  0,   5373, YM+TA+TB+TD), /* 04258 - 505 - Checkpoint (2000) - Fettiger (Unrel.).sndh */
TIMEDB_ENTRY( f8105359,  0,  40512, YM+TA+TB+TD), /* 04349 - YERZMYEY-AY-RIDERS - CELESTIAL SPHERE.sndh */
TIMEDB_ENTRY( f81187b5,  0,   3916,          YM), /* 03092 - Unknown - Every Breath You Take - Sting.sndh */
TIMEDB_ENTRY( f8255e0a,  0,  19199,          YM), /* 04474 - Charles Callet - Murders in Venice.sndh */
TIMEDB_ENTRY( f839d7b9,  0,   5631, YM+TA+TB+TD), /* 02987 - 505-paradox - for the moon.sndh */
TIMEDB_ENTRY( f83d8b09,  0,   2300,    YM+TB+TD), /* 04261 - 505-checkpoint - Granit.sndh */
TIMEDB_ENTRY( f840a4d5,  0,  40227,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  1,  16899,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  2,  41475,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  3,  14339,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  4,  51452,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  5,  71273,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  6,  28163,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  7,   1435,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  8,   1191,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5,  9,   1503,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5, 10,    700,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5, 11,   1574,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f840a4d5, 12,   1503,          YM), /* 01795 - Harlequin - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  0,  27394,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  1,  49179,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  2,  10755,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  3,  28061,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  4,  28675,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  5,  27651,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  6,  63747,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  7,  12291,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  8,  16899,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4,  9,  56835,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 10,  11651,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 11,  23043,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 12,  68870,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 13,  15363,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 14,  50691,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 15,  62595,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 16,  11308,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 17,  26883,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 18,  62115,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 19,  12291,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f85661a4, 20,  21507,          YM), /* 01814 - Zone Warrior - Barry Leitch.sndh */
TIMEDB_ENTRY( f888195c,  0,  24832, YM+TA+TB+TD), /* 02086 - Mad Max - Lightforce - Sid.sndh */
TIMEDB_ENTRY( f889233b,  0,   1289,          YM), /* 03095 - n-a - .sndh */
TIMEDB_ENTRY( f88ceda9,  0,   4032,          YM), /* 00766 - Count Zero - Decade Demo - Loader.sndh */
TIMEDB_ENTRY( f894547e,  0,  94600,       YM+TA), /* 04009 - Manfred Becker - Robert in the Fire Factory.sndh */
TIMEDB_ENTRY( f8a9b5d0,  0,   6527, YM+TA+TB+TD), /* 03365 - gwEm - The Ancient Art of Chiptune.sndh */
TIMEDB_ENTRY( f8c6a251,  0,  17952,          YM), /* 02008 - Mad Max - Shades - Giana Sisters Remix.sndh */
TIMEDB_ENTRY( f8c7f8a6,  0,   7953, YM+TA+TB+TD), /* 04873 - 505 - Courage (compo version).sndh */
TIMEDB_ENTRY( f8c9634f,  0,   8448,          YM), /* 00492 - Ultimatum - Kidney Bean 8.sndh */
TIMEDB_ENTRY( f8d74c10,  0,  11011,          YM), /* 00271 - Archimede - Transbeauce Equinox.sndh */
TIMEDB_ENTRY( f8dfa131,  0,   9594, YM+TA+TB+TD), /* 02832 - Msg - Superfly tune 9.sndh */
TIMEDB_ENTRY( f8e3bf73,  0,   3840,          YM), /* 00481 - Ultimatum - Curio.sndh */
TIMEDB_ENTRY( f8e40be9,  0,  11904,       YM+TA), /* 04117 - Tao - Call Me - STMYGM 2 version.sndh */
TIMEDB_ENTRY( f90150aa,  0,  41496, YM+TA+TB+TD), /* 03397 - cerror - CRANBERRY PT1.sndh */
TIMEDB_ENTRY( f9026221,  0,   8058,          YM), /* 00885 - Adrian Dalecki - Leonardo 1.sndh */
TIMEDB_ENTRY( f9026221,  1,   1146,          YM), /* 00885 - Adrian Dalecki - Leonardo 1.sndh */
TIMEDB_ENTRY( f90bccb5,  0,  17045, YM+STE+TA+TB+TD), /* 03604 - Per Almered (XiA) 2010 - Talk Talk 2 - The Church of Excellence in Art main tune.sndh */
TIMEDB_ENTRY( f920fd49,  0,   4697,          YM), /* 03100 - n-a - Children's Corner # 2.sndh */
TIMEDB_ENTRY( f92ebe35,  0,    768,       YM+TA), /* 00944 - Dark Angel - Rantan Plan.sndh */
TIMEDB_ENTRY( f936b5c6,  0,  14669,          YM), /* 00917 - Dark Angel - Blood Theme.sndh */
TIMEDB_ENTRY( f946b406,  0,   2699,       YM+TA), /* 01703 - Jess - Once Again.sndh */
TIMEDB_ENTRY( f96b72d0,  0,   4607, YM+TA+TB+TD), /* 02836 - maktone-fairlight - hex.sndh */
TIMEDB_ENTRY( f989299b,  0,   2304,          YM), /* 01763 - Lap - Lap 3.sndh */
TIMEDB_ENTRY( f9975afa,  0,   2682,          YM), /* 01347 - Scavenger - Miami Spice.sndh */
TIMEDB_ENTRY( f9a3e049,  0,   5764,          YM), /* 01771 - Lap - Lap 11.sndh */
TIMEDB_ENTRY( f9a7e8fe,  0,  34048, YM+TA+TB+TD), /* 02628 - schmx-awp - Eat This.sndh */
TIMEDB_ENTRY( f9b67dc6,  0,   8696,    YM+TA+TD), /* 01861 - Lotek Style - Ode to DMA-SC.sndh */
TIMEDB_ENTRY( f9bd3c1e,  0,  26548,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  1,  24651,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  2,  29582,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  3,  41265,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  4,  29392,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  5,  29866,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9bd3c1e,  6,  35650,          YM), /* 03363 - Dave Kelly - Last Ninja 2.sndh */
TIMEDB_ENTRY( f9c35150,  0,   5754, YM+TA+TB+TD), /* 04291 - 505-cpt^pdx - nervous.sndh */
TIMEDB_ENTRY( f9cbd585,  0,  11904,       YM+TA), /* 04123 - Tao - Seagulls - STMYGM 2 version.sndh */
TIMEDB_ENTRY( f9cc8d5d,  0,  36885,          YM), /* 03476 - gwEm - Lazy Jones remix.sndh */
TIMEDB_ENTRY( f9d34064,  0,  51395,       YM+TA), /* 04041 - Roggie - Florence loves Dougal.sndh */
TIMEDB_ENTRY( f9d954dc,  0,  10745, YM+TA+TB+TD), /* 01968 - MSG - Godpey - Time.sndh */
TIMEDB_ENTRY( f9dce660,  0,   1463,          YM), /* 04484 - Nathan. Taylor - 180.sndh */
TIMEDB_ENTRY( f9dce660,  1,    576,          YM), /* 04484 - Nathan. Taylor - 180.sndh */
TIMEDB_ENTRY( f9dce660,  2,   3584,          YM), /* 04484 - Nathan. Taylor - 180.sndh */
TIMEDB_ENTRY( f9dce660,  3,   1861,          YM), /* 04484 - Nathan. Taylor - 180.sndh */
TIMEDB_ENTRY( f9e43f4a,  0,   9888,          YM), /* 02383 - Nexus 6 - Lost in Milky Way.sndh */
TIMEDB_ENTRY( f9e4b826,  0,   1513,          YM), /* 00139 - David Whittaker - Sporting Triangles.sndh */
TIMEDB_ENTRY( f9e4b826,  1,     78,          YM), /* 00139 - David Whittaker - Sporting Triangles.sndh */
TIMEDB_ENTRY( f9e4b826,  2,   1514,          YM), /* 00139 - David Whittaker - Sporting Triangles.sndh */
TIMEDB_ENTRY( f9e4b826,  3,    179,          YM), /* 00139 - David Whittaker - Sporting Triangles.sndh */
TIMEDB_ENTRY( f9e4b826,  4,    179,          YM), /* 00139 - David Whittaker - Sporting Triangles.sndh */
TIMEDB_ENTRY( fa335012,  0,   2664,          YM), /* 03927 - Unknown - Sound 1.sndh */
TIMEDB_ENTRY( fa34dc4a,  0,  14339,          YM), /* 01809 - Barry Leitch - Switchblade 2.sndh */
TIMEDB_ENTRY( fa34dc4a,  1,   9219,          YM), /* 01809 - Barry Leitch - Switchblade 2.sndh */
TIMEDB_ENTRY( fa480c49,  0,  56581,       YM+TA), /* 04129 - Mad Max - In 80 Days Around The World.sndh */
TIMEDB_ENTRY( fa64f502,  0,   6048,          YM), /* 02043 - Mad Max - Locomotion.sndh */
TIMEDB_ENTRY( fa71d171,  0,    775,          YM), /* 03913 - Unknown - E1.sndh */
TIMEDB_ENTRY( fa7dbc55,  0,   4416,          YM), /* 01992 - Mad Max - Airball Remix.sndh */
TIMEDB_ENTRY( fa8edb08,  0,   2496, YM+TA+TB+TD), /* 03865 - Mr Coke - Chasey End.sndh */
TIMEDB_ENTRY( fa9a5fac,  0,   8064,          YM), /* 00257 - Antichrist - New History.sndh */
TIMEDB_ENTRY( faad5275,  0,  99326,       YM+TA), /* 03706 - Sparky - Just 4 Fun.sndh */
TIMEDB_ENTRY( fabf8b04,  0,    195,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fabf8b04,  1,    597,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fabf8b04,  2,    322,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fabf8b04,  3,   4239,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fabf8b04,  4,   1963,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fabf8b04,  5,    104,          YM), /* 00564 - Jas C Brooke - Buggy Boy.sndh */
TIMEDB_ENTRY( fae15a74,  0,   3840,          YM), /* 00900 - Danny Oneway - Drumz.sndh */
TIMEDB_ENTRY( fae747b9,  0,   8763,          YM), /* 00149 - David Whittaker - Total Recall.sndh */
TIMEDB_ENTRY( fae747b9,  1,   8304,          YM), /* 00149 - David Whittaker - Total Recall.sndh */
TIMEDB_ENTRY( fae747b9,  2,   1905,          YM), /* 00149 - David Whittaker - Total Recall.sndh */
TIMEDB_ENTRY( fae747b9,  3,   2048,          YM), /* 00149 - David Whittaker - Total Recall.sndh */
TIMEDB_ENTRY( fae747b9,  4,   1917,          YM), /* 00149 - David Whittaker - Total Recall.sndh */
TIMEDB_ENTRY( faf39c3b,  0,   2112,          YM), /* 04681 - B Amundlien - Just Another Silly Sports sim.sndh */
TIMEDB_ENTRY( faf39c3b,  1,   2352,          YM), /* 04681 - B Amundlien - Just Another Silly Sports sim.sndh */
TIMEDB_ENTRY( faf39c3b,  2,   8896,          YM), /* 04681 - B Amundlien - Just Another Silly Sports sim.sndh */
TIMEDB_ENTRY( faf39c3b,  3,    304,          YM), /* 04681 - B Amundlien - Just Another Silly Sports sim.sndh */
TIMEDB_ENTRY( fb020342,  0,  12733,    YM+TA+TD), /* 02894 - Nemo - DHS - Bobbin Guru Deva.sndh */
TIMEDB_ENTRY( fb1d5028,  0,   1920,          YM), /* 04648 - ActoDi - Cute Animals.sndh */
TIMEDB_ENTRY( fb23df7b,  0,  10326,       YM+TA), /* 01621 - No More - Infinity.sndh */
TIMEDB_ENTRY( fb3140a4,  0,   2764,          YM), /* 00151 - David Whittaker - Voodoo Nightmare.sndh */
TIMEDB_ENTRY( fb3140a4,  1,   7801,          YM), /* 00151 - David Whittaker - Voodoo Nightmare.sndh */
TIMEDB_ENTRY( fb3140a4,  2,    124,          YM), /* 00151 - David Whittaker - Voodoo Nightmare.sndh */
TIMEDB_ENTRY( fb3140a4,  3,    135,          YM), /* 00151 - David Whittaker - Voodoo Nightmare.sndh */
TIMEDB_ENTRY( fb5bb413,  0,   1727,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  1,    136,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  2,   2303,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  3,   2303,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  4,   2303,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  5,    895,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  6,    120,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  7,   3966,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  8,   3359,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413,  9,   1140,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5bb413, 10,    511,          YM), /* 00350 - Tony Williams - Shinobi.sndh */
TIMEDB_ENTRY( fb5c1c6e,  0,   6144,          YM), /* 01982 - MSG - MSG 14.sndh */
TIMEDB_ENTRY( fb5d6332,  0,  10756,          YM), /* 02229 - Tao - Fractal Blast (Nostalgic-O-Demo).sndh */
TIMEDB_ENTRY( fb67d77b,  0,   5384,          YM), /* 01034 - Donald Fakk - Bluespace.sndh */
TIMEDB_ENTRY( fb7d564d,  0,   2080,          YM), /* 03060 - Ed Bogas - Star-Spangled Banner - F.S.Key.sndh */
TIMEDB_ENTRY( fb7ee1cc,  0,  57367,       YM+TA), /* 03804 - Sparky - Pompey Pirates Menu 106.sndh */
TIMEDB_ENTRY( fb8e2f01,  0,   8642,          YM), /* 00402 - Big Alec - What Time Is It ? Bouncy.sndh */
TIMEDB_ENTRY( fb96b36a,  0,   6833,       YM+TA), /* 03648 - 505-paradox - Bagatelle.sndh */
TIMEDB_ENTRY( fb9a23fc,  0,   6719, YM+TA+TB+TD), /* 03666 - Starbuck - Der Traum (YM Remix).sndh */
TIMEDB_ENTRY( fbdde9a7,  0,   9472,       YM+TA), /* 00780 - Crawdaddy - Bounty hunter.sndh */
TIMEDB_ENTRY( fbe2f4a0,  0,  15231,          YM), /* 00513 - Tangens - Synthetic Visions - Scroller.sndh */
TIMEDB_ENTRY( fbfd09ab,  0,  18829, YM+TA+TB+TD), /* 03609 - Dma-Sc - The 2149 Gang.sndh */
TIMEDB_ENTRY( fbfd4508,  0,   2727,       YM+TB), /* 02650 - Dma-Sc - POOZ - Highscores.sndh */
TIMEDB_ENTRY( fc0e9214,  0,   7681,       YM+TA), /* 04337 - Big Alec - Punish Your Machine - Main Menu.sndh */
TIMEDB_ENTRY( fc33afb1,  0,   6675,       YM+TA), /* 04550 - gwEm - Sillyventure 2014 Invite.sndh */
TIMEDB_ENTRY( fc39e0b9,  0,   9248, YM+TA+TB+TD), /* 01842 - Lotek Style - Diva.sndh */
TIMEDB_ENTRY( fc4a54b6,  0,   8640,       YM+TA), /* 01635 - Nexus - Dreamzone - Lame Screen.sndh */
TIMEDB_ENTRY( fc693ef5,  0,   3842,          YM), /* 00367 - Steve Barrett - Super Grand Prix.sndh */
TIMEDB_ENTRY( fc693ef5,  1,   1538,          YM), /* 00367 - Steve Barrett - Super Grand Prix.sndh */
TIMEDB_ENTRY( fc693ef5,  2,    194,          YM), /* 00367 - Steve Barrett - Super Grand Prix.sndh */
TIMEDB_ENTRY( fc693ef5,  3,    194,          YM), /* 00367 - Steve Barrett - Super Grand Prix.sndh */
TIMEDB_ENTRY( fc72b34f,  0,   4218,       YM+TA), /* 01356 - Scavenger - Pandora's Box.sndh */
TIMEDB_ENTRY( fc7861f8,  0,  10016,          YM), /* 03572 - YERZMYEY-AY-RIDERS - Yerzmix Part 4 - Silly Venture 2010.sndh */
TIMEDB_ENTRY( fc83d25a,  0,  36180, YM+TA+TB+TD), /* 02529 - Dark Angel - The !Song.sndh */
TIMEDB_ENTRY( fc8405c1,  0,  10496,          YM), /* 02060 - Mad Max - Warhawk.sndh */
TIMEDB_ENTRY( fc9f2ee6,  0,   4224,       YM+TA), /* 02870 - Crawdaddy - Whatta Hell.sndh */
TIMEDB_ENTRY( fcb4718d,  0,   1023,          YM), /* 04078 - Gary Wheaton - I'm knitting with only one needle.sndh */
TIMEDB_ENTRY( fcb6ec4f,  0,   4067,          YM), /* 00844 - Ben Daglish - Captain America.sndh */
TIMEDB_ENTRY( fcc0a5f6,  0,   4623,    YM+TA+TD), /* 04319 - 505-Checkpoint (2013) - Taotic.sndh */
TIMEDB_ENTRY( fccaffd2,  0,   6141,    YM+TB+TD), /* 04107 - 505 - Checkpoint - Blink.sndh */
TIMEDB_ENTRY( fcd8371f,  0,  17024,          YM), /* 03403 - Dave Lowe - Garfield 2 - Winters Tale.sndh */
TIMEDB_ENTRY( fcd8371f,  1,   4006,          YM), /* 03403 - Dave Lowe - Garfield 2 - Winters Tale.sndh */
TIMEDB_ENTRY( fcd8371f,  2,   4352,          YM), /* 03403 - Dave Lowe - Garfield 2 - Winters Tale.sndh */
TIMEDB_ENTRY( fcd906c1,  0,  22529,          YM), /* 02953 - Andy Severn - Heavy Metal Heroes.sndh */
TIMEDB_ENTRY( fcd906c1,  1,   8193,          YM), /* 02953 - Andy Severn - Heavy Metal Heroes.sndh */
TIMEDB_ENTRY( fcd906c1,  2,    162,          YM), /* 02953 - Andy Severn - Heavy Metal Heroes.sndh */
TIMEDB_ENTRY( fcd906c1,  3,    182,          YM), /* 02953 - Andy Severn - Heavy Metal Heroes.sndh */
TIMEDB_ENTRY( fcd941cf,  0,   1789,          YM), /* 01901 - Luebke - Jingle Bells.sndh */
TIMEDB_ENTRY( fcfcd5e5,  0,   9243,          YM), /* 01881 - Dave Lowe - Elite 2 - Frontier.sndh */
TIMEDB_ENTRY( fcfcd5e5,  1,   3836,          YM), /* 01881 - Dave Lowe - Elite 2 - Frontier.sndh */
TIMEDB_ENTRY( fcfcd5e5,  2,   4851,          YM), /* 01881 - Dave Lowe - Elite 2 - Frontier.sndh */
TIMEDB_ENTRY( fcfcd5e5,  3,   2587,          YM), /* 01881 - Dave Lowe - Elite 2 - Frontier.sndh */
TIMEDB_ENTRY( fcffaeb8,  0,  35105,       YM+TA), /* 03952 - Scott Lemmon - House of the Rising Sun.sndh */
TIMEDB_ENTRY( fd153c8e,  0,  19532,          YM), /* 02175 - Mad Max - 5th Gear.sndh */
TIMEDB_ENTRY( fd153c8e,  1,   3840,          YM), /* 02175 - Mad Max - 5th Gear.sndh */
TIMEDB_ENTRY( fd153c8e,  2,    769,          YM), /* 02175 - Mad Max - 5th Gear.sndh */
TIMEDB_ENTRY( fd153c8e,  3,   1152,          YM), /* 02175 - Mad Max - 5th Gear.sndh */
TIMEDB_ENTRY( fd153c8e,  4,    553,          YM), /* 02175 - Mad Max - 5th Gear.sndh */
TIMEDB_ENTRY( fd1e4db4,  0, 346369,       YM+TA), /* 04034 - Roggie - Roggie 16.sndh */
TIMEDB_ENTRY( fd40717c,  0,   1530,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd40717c,  1,   1336,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd40717c,  2,   6466,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd40717c,  3,   4979,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd40717c,  4,   2517,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd40717c,  5,    570,          YM), /* 00017 - 4-Mat - Space Gun.sndh */
TIMEDB_ENTRY( fd4e7906,  0,   8960,      YM+STE), /* 04514 - Excellence in Art - Summer Deligths.sndh */
TIMEDB_ENTRY( fd603263,  0,   3072,          YM), /* 02056 - Mad Max - Thanatos.sndh */
TIMEDB_ENTRY( fd8aad47,  0, 111966,       YM+TA), /* 03707 - Sparky - MC DESer - For DES and Juggler.sndh */
TIMEDB_ENTRY( fd8b4b10,  0,   7120,          YM), /* 02412 - Nexus 6 - Third.sndh */
TIMEDB_ENTRY( fdaed5e3,  0,   5628,          YM), /* 01227 - Epic - That's About To Explode.sndh */
TIMEDB_ENTRY( fdc9e72d,  0,    754,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  1,    409,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  2,    211,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  3,   1006,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  4,   1981,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  5,   1344,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  6,   3456,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  7,    768,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  8,   1944,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d,  9,   1238,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d, 10,   1620,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdc9e72d, 11,    816,          YM), /* 00172 - Martin Walker - R-Type 2.sndh */
TIMEDB_ENTRY( fdd0c43c,  0,    504,          YM), /* 03682 - Dma-Sc - Antiques - Loader 3.sndh */
TIMEDB_ENTRY( fdf9f5ee,  0,   8442,          YM), /* 02878 - Beast - Dodi Dodi.sndh */
TIMEDB_ENTRY( fe1de724,  0,   2048,          YM), /* 02930 - Chris Denman - Ancient Games.sndh */
TIMEDB_ENTRY( fe318f68,  0,   1907,          YM), /* 03073 - Peter Inzerillo - Carol of the Bells.sndh */
TIMEDB_ENTRY( fe4ad6a9,  0,  12288,    YM+TB+TD), /* 02847 - Helene - 28 Sept 06.sndh */
TIMEDB_ENTRY( fe7fb862,  0,   4000,    YM+TB+TD), /* 03338 - Marcer - n-a.sndh */
TIMEDB_ENTRY( fe81df89,  0,   1666,    YM+TA+TD), /* 02254 - Tao - Cream - Entrance (Steps).sndh */
TIMEDB_ENTRY( fe83dc25,  0,   2916,    YM+TA+TB), /* 00053 - 505 - Greetings.sndh */
TIMEDB_ENTRY( fe8462a0,  0,  11783,          YM), /* 02277 - Tao - Cream - Chicago (Just Buggin Greetings).sndh */
TIMEDB_ENTRY( fe8d78c4,  0,  12800,    YM+TB+TD), /* 02913 - helln - WRONGGGGG.sndh */
TIMEDB_ENTRY( fe92af3f,  0,   9677,    YM+TA+TD), /* 02749 - kenny d - folk off 'round the fields.sndh */
TIMEDB_ENTRY( fe97d66b,  0,  15354, YM+TA+TB+TD), /* 01853 - Lotek Style - Hopes - UCM 13 part 2.sndh */
TIMEDB_ENTRY( fe9d78ee,  0,   5120,          YM), /* 00251 - Antichrist - Christelle.sndh */
TIMEDB_ENTRY( febe29b7,  0,  13920,          YM), /* 02013 - Mad Max - Poker Music.sndh */
TIMEDB_ENTRY( fed9b8b7,  0,   3680,          YM), /* 04749 - Dave Lowe - Betrayal.sndh */
TIMEDB_ENTRY( fed9b8b7,  1,   2851,          YM), /* 04749 - Dave Lowe - Betrayal.sndh */
TIMEDB_ENTRY( feda6e4e,  0,    858,          YM), /* 04141 - Unknown - Karlchen.sndh */
TIMEDB_ENTRY( feda6e4e,  1,   2457,          YM), /* 04141 - Unknown - Karlchen.sndh */
TIMEDB_ENTRY( fef86d0d,  0,   6975, YM+TA+TB+TD), /* 03614 - 505-YM-Rockerz - Liberation (STE).sndh */
TIMEDB_ENTRY( ff018202,  0,  16896,          YM), /* 02579 - STU - BOMB SQUAD.sndh */
TIMEDB_ENTRY( ff0dc15d,  0,   2260,    YM+TA+TD), /* 02232 - Tao - Cream - Ikari Union Edit (Nostalgic-O-Demo, not used).sndh */
TIMEDB_ENTRY( ff106010,  0,   9317, YM+TA+TB+TD), /* 03588 - Divag - Crash to live.sndh */
TIMEDB_ENTRY( ff1111f4,  0,   6143,          YM), /* 03675 - Dma-Sc - Antiques - IFF Slideshow.sndh */
TIMEDB_ENTRY( ff34688e,  0,  15360,    YM+TB+TD), /* 02907 - helene - NONONYESYESYES.sndh */
TIMEDB_ENTRY( ff468dd0,  0,   4004,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  1,   6151,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  2,   4613,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  3,   4613,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  4,   3506,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  5,    257,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  6,    259,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  7,    148,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff468dd0,  8,   1921,          YM), /* 01737 - Richard Joseph - James Pond 2 Codename  Robocod.sndh */
TIMEDB_ENTRY( ff77da40,  0,  10369, YM+TA+TB+TD), /* 03670 - 505-paradox - Jacktro.sndh */
TIMEDB_ENTRY( ff7fae4d,  0,   6912,    YM+TA+TB), /* 01334 - Scavenger - Hate.sndh */
TIMEDB_ENTRY( ff7fcdf3,  0,  21690,          YM), /* 01682 - Jess - I Found The Dead Zone.sndh */
TIMEDB_ENTRY( ffa6a088,  0,   8447, YM+TA+TB+TD), /* 03266 - Dubmood - Intro Tune (4-Mat).sndh */
TIMEDB_ENTRY( ffacfd86,  0,   7104,      YM+STE), /* 04212 - Griffon - STE Folies - O Demo.sndh */
TIMEDB_ENTRY( ffb965d6,  0,  10368,          YM), /* 01640 - JB 007 - Librarie Intro.sndh */
TIMEDB_ENTRY( ffbaf782,  0,   4669,          YM), /* 00298 - Gary Wheaton - Let it Be.sndh */
TIMEDB_ENTRY( ffbfbad9,  0,   3072,          YM), /* 01148 - Tyan - Last Ninja Remake.sndh */
TIMEDB_ENTRY( ffc3f74c,  0,   3832,          YM), /* 00592 - Damon Chaplin - Last Trooper.sndh */
TIMEDB_ENTRY( ffc6a7e7,  0,   1320,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc6a7e7,  1,   1480,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc6a7e7,  2,    634,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc6a7e7,  3,    255,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc6a7e7,  4,   1492,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc6a7e7,  5,   1828,          YM), /* 04719 - Dean Sharples (Deano) - The Heavy Bunch.sndh */
TIMEDB_ENTRY( ffc8af8b,  0,  17916,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffc8af8b,  1,   5760,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffc8af8b,  2,   9648,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffc8af8b,  3,   7984,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffc8af8b,  4,  17408,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffc8af8b,  5,  18408,          YM), /* 04467 - Charles Callet - Passengers in the Wind.sndh */
TIMEDB_ENTRY( ffe1c341,  0,   4063,          YM), /* 03187 - n-a - Yoda's Theme.sndh */
TIMEDB_ENTRY( fff98e01,  0,   9983,    YM+TA+TB), /* 01361 - Scavenger - Readers Digest.sndh */
TIMEDB_ENTRY( fffaeaae,  0,  10362,    YM+TB+TD), /* 02477 - No More - leif Rullar conversion.sndh */
