/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#define GLADE_HOOKUP_OBJECT(component,widget,name) \
  g_object_set_data_full (G_OBJECT (component), name, \
    g_object_ref(G_OBJECT(widget)), (GDestroyNotify) g_object_unref)

#define GLADE_HOOKUP_OBJECT_NO_REF(component,widget,name) \
  g_object_set_data (G_OBJECT (component), name, widget)

GtkWidget*
create_mainwin (void)
{
  GtkWidget *mainwin;
  GtkWidget *vbox1;
  GtkWidget *menubar;
  GtkWidget *File;
  GtkWidget *File_menu;
  GtkWidget *open;
  GtkWidget *image724;
  GtkWidget *separator2;
  GtkWidget *add_files;
  GtkWidget *image725;
  GtkWidget *add_folders;
  GtkWidget *image726;
  GtkWidget *add_location1;
  GtkWidget *separatormenuitem1;
  GtkWidget *new_playlist1;
  GtkWidget *playlist_load;
  GtkWidget *playlist_save_as;
  GtkWidget *separator8;
  GtkWidget *quit;
  GtkWidget *image727;
  GtkWidget *Edit;
  GtkWidget *Edit_menu;
  GtkWidget *undo;
  GtkWidget *redo;
  GtkWidget *clear1;
  GtkWidget *image728;
  GtkWidget *select_all1;
  GtkWidget *deselect_all1;
  GtkWidget *invert_selection1;
  GtkWidget *Selection;
  GtkWidget *Selection_menu;
  GtkWidget *remove1;
  GtkWidget *image729;
  GtkWidget *crop1;
  GtkWidget *find1;
  GtkWidget *sort_by1;
  GtkWidget *sort_by1_menu;
  GtkWidget *title1;
  GtkWidget *track_nr1;
  GtkWidget *album1;
  GtkWidget *artist1;
  GtkWidget *date1;
  GtkWidget *random1;
  GtkWidget *custom2;
  GtkWidget *separator5;
  GtkWidget *preferences;
  GtkWidget *View;
  GtkWidget *View_menu;
  GtkWidget *view_status_bar;
  GtkWidget *view_eq;
  GtkWidget *design_mode1;
  GtkWidget *view_log;
  GtkWidget *Playback;
  GtkWidget *Playback_menu;
  GtkWidget *Order;
  GtkWidget *Order_menu;
  GSList *order_linear_group = NULL;
  GtkWidget *order_linear;
  GtkWidget *order_shuffle;
  GtkWidget *order_shuffle_albums;
  GtkWidget *order_random;
  GtkWidget *Looping;
  GtkWidget *Looping_menu;
  GSList *loop_all_group = NULL;
  GtkWidget *loop_all;
  GtkWidget *loop_single;
  GtkWidget *loop_disable;
  GtkWidget *scroll_follows_playback;
  GtkWidget *cursor_follows_playback;
  GtkWidget *stop_after_current;
  GtkWidget *stop_after_album;
  GtkWidget *stop_after_queue;
  GtkWidget *separator11;
  GtkWidget *jump_to_current_track1;
  GtkWidget *Help;
  GtkWidget *Help_menu;
  GtkWidget *help1;
  GtkWidget *image730;
  GtkWidget *changelog1;
  GtkWidget *separator10;
  GtkWidget *gpl1;
  GtkWidget *lgpl1;
  GtkWidget *separator9;
  GtkWidget *about1;
  GtkWidget *image731;
  GtkWidget *translators1;
  GtkWidget *image732;
  GtkWidget *plugins_bottom_vbox;
  GtkWidget *statusbar;
  GtkAccelGroup *accel_group;

  accel_group = gtk_accel_group_new ();

  mainwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_events (mainwin, GDK_POINTER_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK);
  gtk_window_set_title (GTK_WINDOW (mainwin), "DeaDBeeF");
  gtk_window_set_default_size (GTK_WINDOW (mainwin), 500, 300);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (mainwin), vbox1);

  menubar = gtk_menu_bar_new ();
  gtk_widget_show (menubar);
  gtk_box_pack_start (GTK_BOX (vbox1), menubar, FALSE, FALSE, 0);

  File = gtk_menu_item_new_with_mnemonic (_("_File"));
  gtk_widget_show (File);
  gtk_container_add (GTK_CONTAINER (menubar), File);

  File_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (File), File_menu);

  open = gtk_image_menu_item_new_with_mnemonic (_("_Open file(s)"));
  gtk_widget_show (open);
  gtk_container_add (GTK_CONTAINER (File_menu), open);

  image724 = gtk_image_new_from_stock ("gtk-open", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image724);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (open), image724);

  separator2 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator2);
  gtk_container_add (GTK_CONTAINER (File_menu), separator2);
  gtk_widget_set_sensitive (separator2, FALSE);

  add_files = gtk_image_menu_item_new_with_mnemonic (_("Add file(s)"));
  gtk_widget_show (add_files);
  gtk_container_add (GTK_CONTAINER (File_menu), add_files);

  image725 = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image725);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (add_files), image725);

  add_folders = gtk_image_menu_item_new_with_mnemonic (_("Add folder(s)"));
  gtk_widget_show (add_folders);
  gtk_container_add (GTK_CONTAINER (File_menu), add_folders);

  image726 = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image726);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (add_folders), image726);

  add_location1 = gtk_menu_item_new_with_mnemonic (_("Add location"));
  gtk_widget_show (add_location1);
  gtk_container_add (GTK_CONTAINER (File_menu), add_location1);

  separatormenuitem1 = gtk_separator_menu_item_new ();
  gtk_widget_show (separatormenuitem1);
  gtk_container_add (GTK_CONTAINER (File_menu), separatormenuitem1);
  gtk_widget_set_sensitive (separatormenuitem1, FALSE);

  new_playlist1 = gtk_menu_item_new_with_mnemonic (_("New playlist"));
  gtk_widget_show (new_playlist1);
  gtk_container_add (GTK_CONTAINER (File_menu), new_playlist1);

  playlist_load = gtk_menu_item_new_with_mnemonic (_("Load playlist"));
  gtk_widget_show (playlist_load);
  gtk_container_add (GTK_CONTAINER (File_menu), playlist_load);

  playlist_save_as = gtk_menu_item_new_with_mnemonic (_("Save playlist"));
  gtk_widget_show (playlist_save_as);
  gtk_container_add (GTK_CONTAINER (File_menu), playlist_save_as);

  separator8 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator8);
  gtk_container_add (GTK_CONTAINER (File_menu), separator8);
  gtk_widget_set_sensitive (separator8, FALSE);

  quit = gtk_image_menu_item_new_with_mnemonic (_("_Quit"));
  gtk_widget_show (quit);
  gtk_container_add (GTK_CONTAINER (File_menu), quit);

  image727 = gtk_image_new_from_stock ("gtk-quit", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image727);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (quit), image727);

  Edit = gtk_menu_item_new_with_mnemonic (_("_Edit"));
  gtk_widget_show (Edit);
  gtk_container_add (GTK_CONTAINER (menubar), Edit);

  Edit_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Edit), Edit_menu);

  undo = gtk_image_menu_item_new_from_stock ("gtk-undo", accel_group);
  gtk_widget_show (undo);
  gtk_container_add (GTK_CONTAINER (Edit_menu), undo);

  redo = gtk_image_menu_item_new_from_stock ("gtk-redo", accel_group);
  gtk_widget_show (redo);
  gtk_container_add (GTK_CONTAINER (Edit_menu), redo);

  clear1 = gtk_image_menu_item_new_with_mnemonic (_("_Clear"));
  gtk_widget_show (clear1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), clear1);

  image728 = gtk_image_new_from_stock ("gtk-clear", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image728);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (clear1), image728);

  select_all1 = gtk_menu_item_new_with_mnemonic (_("Select all"));
  gtk_widget_show (select_all1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), select_all1);

  deselect_all1 = gtk_menu_item_new_with_mnemonic (_("Deselect all"));
  gtk_widget_show (deselect_all1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), deselect_all1);

  invert_selection1 = gtk_menu_item_new_with_mnemonic (_("Invert selection"));
  gtk_widget_show (invert_selection1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), invert_selection1);

  Selection = gtk_menu_item_new_with_mnemonic (_("Selection"));
  gtk_widget_show (Selection);
  gtk_container_add (GTK_CONTAINER (Edit_menu), Selection);

  Selection_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Selection), Selection_menu);

  remove1 = gtk_image_menu_item_new_with_mnemonic (_("Remove"));
  gtk_widget_show (remove1);
  gtk_container_add (GTK_CONTAINER (Selection_menu), remove1);

  image729 = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image729);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (remove1), image729);

  crop1 = gtk_menu_item_new_with_mnemonic (_("Crop"));
  gtk_widget_show (crop1);
  gtk_container_add (GTK_CONTAINER (Selection_menu), crop1);

  find1 = gtk_menu_item_new_with_mnemonic (_("_Find"));
  gtk_widget_show (find1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), find1);

  sort_by1 = gtk_menu_item_new_with_mnemonic (_("Sort by"));
  gtk_widget_show (sort_by1);
  gtk_container_add (GTK_CONTAINER (Edit_menu), sort_by1);

  sort_by1_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (sort_by1), sort_by1_menu);

  title1 = gtk_menu_item_new_with_mnemonic (_("Title"));
  gtk_widget_show (title1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), title1);

  track_nr1 = gtk_menu_item_new_with_mnemonic (_("Track number"));
  gtk_widget_show (track_nr1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), track_nr1);

  album1 = gtk_menu_item_new_with_mnemonic (_("Album"));
  gtk_widget_show (album1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), album1);

  artist1 = gtk_menu_item_new_with_mnemonic (_("Artist"));
  gtk_widget_show (artist1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), artist1);

  date1 = gtk_menu_item_new_with_mnemonic (_("Date"));
  gtk_widget_show (date1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), date1);

  random1 = gtk_menu_item_new_with_mnemonic (_("Random"));
  gtk_widget_show (random1);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), random1);

  custom2 = gtk_menu_item_new_with_mnemonic (_("Custom"));
  gtk_widget_show (custom2);
  gtk_container_add (GTK_CONTAINER (sort_by1_menu), custom2);

  separator5 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator5);
  gtk_container_add (GTK_CONTAINER (Edit_menu), separator5);
  gtk_widget_set_sensitive (separator5, FALSE);

  preferences = gtk_menu_item_new_with_mnemonic (_("Preferences"));
  gtk_widget_show (preferences);
  gtk_container_add (GTK_CONTAINER (Edit_menu), preferences);

  View = gtk_menu_item_new_with_mnemonic (_("_View"));
  gtk_widget_show (View);
  gtk_container_add (GTK_CONTAINER (menubar), View);

  View_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (View), View_menu);

  view_status_bar = gtk_check_menu_item_new_with_mnemonic (_("Status bar"));
  gtk_widget_show (view_status_bar);
  gtk_container_add (GTK_CONTAINER (View_menu), view_status_bar);

  view_eq = gtk_check_menu_item_new_with_mnemonic (_("Equalizer"));
  gtk_widget_show (view_eq);
  gtk_container_add (GTK_CONTAINER (View_menu), view_eq);

  design_mode1 = gtk_check_menu_item_new_with_mnemonic (_("Design mode"));
  gtk_widget_show (design_mode1);
  gtk_container_add (GTK_CONTAINER (View_menu), design_mode1);

  view_log = gtk_check_menu_item_new_with_mnemonic (_("Log"));
  gtk_widget_show (view_log);
  gtk_container_add (GTK_CONTAINER (View_menu), view_log);

  Playback = gtk_menu_item_new_with_mnemonic (_("_Playback"));
  gtk_widget_show (Playback);
  gtk_container_add (GTK_CONTAINER (menubar), Playback);

  Playback_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Playback), Playback_menu);

  Order = gtk_menu_item_new_with_mnemonic (_("Shuffle"));
  gtk_widget_show (Order);
  gtk_container_add (GTK_CONTAINER (Playback_menu), Order);

  Order_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Order), Order_menu);

  order_linear = gtk_radio_menu_item_new_with_mnemonic (order_linear_group, _("Off"));
  order_linear_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (order_linear));
  gtk_widget_show (order_linear);
  gtk_container_add (GTK_CONTAINER (Order_menu), order_linear);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (order_linear), TRUE);

  order_shuffle = gtk_radio_menu_item_new_with_mnemonic (order_linear_group, _("Tracks"));
  order_linear_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (order_shuffle));
  gtk_widget_show (order_shuffle);
  gtk_container_add (GTK_CONTAINER (Order_menu), order_shuffle);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (order_shuffle), TRUE);

  order_shuffle_albums = gtk_radio_menu_item_new_with_mnemonic (order_linear_group, _("Albums"));
  order_linear_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (order_shuffle_albums));
  gtk_widget_show (order_shuffle_albums);
  gtk_container_add (GTK_CONTAINER (Order_menu), order_shuffle_albums);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (order_shuffle_albums), TRUE);

  order_random = gtk_radio_menu_item_new_with_mnemonic (order_linear_group, _("Random Tracks"));
  order_linear_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (order_random));
  gtk_widget_show (order_random);
  gtk_container_add (GTK_CONTAINER (Order_menu), order_random);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (order_random), TRUE);

  Looping = gtk_menu_item_new_with_mnemonic (_("Repeat"));
  gtk_widget_show (Looping);
  gtk_container_add (GTK_CONTAINER (Playback_menu), Looping);

  Looping_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Looping), Looping_menu);

  loop_all = gtk_radio_menu_item_new_with_mnemonic (loop_all_group, _("All Tracks"));
  loop_all_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (loop_all));
  gtk_widget_show (loop_all);
  gtk_container_add (GTK_CONTAINER (Looping_menu), loop_all);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (loop_all), TRUE);

  loop_single = gtk_radio_menu_item_new_with_mnemonic (loop_all_group, _("One Track"));
  loop_all_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (loop_single));
  gtk_widget_show (loop_single);
  gtk_container_add (GTK_CONTAINER (Looping_menu), loop_single);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (loop_single), TRUE);

  loop_disable = gtk_radio_menu_item_new_with_mnemonic (loop_all_group, _("Off"));
  loop_all_group = gtk_radio_menu_item_get_group (GTK_RADIO_MENU_ITEM (loop_disable));
  gtk_widget_show (loop_disable);
  gtk_container_add (GTK_CONTAINER (Looping_menu), loop_disable);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (loop_disable), TRUE);

  scroll_follows_playback = gtk_check_menu_item_new_with_mnemonic (_("Scroll follows playback"));
  gtk_widget_show (scroll_follows_playback);
  gtk_container_add (GTK_CONTAINER (Playback_menu), scroll_follows_playback);
  gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (scroll_follows_playback), TRUE);

  cursor_follows_playback = gtk_check_menu_item_new_with_mnemonic (_("Cursor follows playback"));
  gtk_widget_show (cursor_follows_playback);
  gtk_container_add (GTK_CONTAINER (Playback_menu), cursor_follows_playback);

  stop_after_current = gtk_check_menu_item_new_with_mnemonic (_("Stop after current track"));
  gtk_widget_show (stop_after_current);
  gtk_container_add (GTK_CONTAINER (Playback_menu), stop_after_current);

  stop_after_album = gtk_check_menu_item_new_with_mnemonic (_("Stop after current album"));
  gtk_widget_show (stop_after_album);
  gtk_container_add (GTK_CONTAINER (Playback_menu), stop_after_album);

  stop_after_queue = gtk_check_menu_item_new_with_mnemonic (_("Stop after playback queue"));
  gtk_widget_show (stop_after_queue);
  gtk_container_add (GTK_CONTAINER (Playback_menu), stop_after_queue);

  separator11 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator11);
  gtk_container_add (GTK_CONTAINER (Playback_menu), separator11);
  gtk_widget_set_sensitive (separator11, FALSE);

  jump_to_current_track1 = gtk_menu_item_new_with_mnemonic (_("Jump to current track"));
  gtk_widget_show (jump_to_current_track1);
  gtk_container_add (GTK_CONTAINER (Playback_menu), jump_to_current_track1);

  Help = gtk_menu_item_new_with_mnemonic (_("_Help"));
  gtk_widget_show (Help);
  gtk_container_add (GTK_CONTAINER (menubar), Help);

  Help_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (Help), Help_menu);

  help1 = gtk_image_menu_item_new_with_mnemonic (_("_Help"));
  gtk_widget_show (help1);
  gtk_container_add (GTK_CONTAINER (Help_menu), help1);

  image730 = gtk_image_new_from_stock ("gtk-help", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image730);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (help1), image730);

  changelog1 = gtk_menu_item_new_with_mnemonic (_("_ChangeLog"));
  gtk_widget_show (changelog1);
  gtk_container_add (GTK_CONTAINER (Help_menu), changelog1);

  separator10 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator10);
  gtk_container_add (GTK_CONTAINER (Help_menu), separator10);
  gtk_widget_set_sensitive (separator10, FALSE);

  gpl1 = gtk_menu_item_new_with_mnemonic (_("_GPLv2"));
  gtk_widget_show (gpl1);
  gtk_container_add (GTK_CONTAINER (Help_menu), gpl1);

  lgpl1 = gtk_menu_item_new_with_mnemonic (_("_LGPLv2.1"));
  gtk_widget_show (lgpl1);
  gtk_container_add (GTK_CONTAINER (Help_menu), lgpl1);

  separator9 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator9);
  gtk_container_add (GTK_CONTAINER (Help_menu), separator9);
  gtk_widget_set_sensitive (separator9, FALSE);

  about1 = gtk_image_menu_item_new_with_mnemonic (_("_About"));
  gtk_widget_show (about1);
  gtk_container_add (GTK_CONTAINER (Help_menu), about1);

  image731 = gtk_image_new_from_stock ("gtk-about", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image731);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (about1), image731);

  translators1 = gtk_image_menu_item_new_with_mnemonic (_("_Translators"));
  gtk_widget_show (translators1);
  gtk_container_add (GTK_CONTAINER (Help_menu), translators1);

  image732 = gtk_image_new_from_stock ("gtk-about", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image732);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (translators1), image732);

  plugins_bottom_vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (plugins_bottom_vbox);
  gtk_box_pack_start (GTK_BOX (vbox1), plugins_bottom_vbox, TRUE, TRUE, 0);

  statusbar = gtk_statusbar_new ();
  gtk_widget_show (statusbar);
  gtk_box_pack_start (GTK_BOX (vbox1), statusbar, FALSE, FALSE, 0);

  g_signal_connect_after ((gpointer) mainwin, "key_press_event",
                          G_CALLBACK (on_mainwin_key_press_event),
                          NULL);
  g_signal_connect ((gpointer) mainwin, "delete_event",
                    G_CALLBACK (on_mainwin_delete_event),
                    NULL);
  g_signal_connect ((gpointer) mainwin, "configure_event",
                    G_CALLBACK (on_mainwin_configure_event),
                    NULL);
  g_signal_connect ((gpointer) mainwin, "window_state_event",
                    G_CALLBACK (on_mainwin_window_state_event),
                    NULL);
  g_signal_connect ((gpointer) mainwin, "realize",
                    G_CALLBACK (on_mainwin_realize),
                    NULL);
  g_signal_connect ((gpointer) mainwin, "button_press_event",
                    G_CALLBACK (on_mainwin_button_press_event),
                    NULL);
  g_signal_connect ((gpointer) open, "activate",
                    G_CALLBACK (on_open_activate),
                    NULL);
  g_signal_connect ((gpointer) add_files, "activate",
                    G_CALLBACK (on_add_files_activate),
                    NULL);
  g_signal_connect ((gpointer) add_folders, "activate",
                    G_CALLBACK (on_add_folders_activate),
                    NULL);
  g_signal_connect ((gpointer) add_location1, "activate",
                    G_CALLBACK (on_add_location_activate),
                    NULL);
  g_signal_connect ((gpointer) new_playlist1, "activate",
                    G_CALLBACK (on_new_playlist1_activate),
                    NULL);
  g_signal_connect ((gpointer) playlist_load, "activate",
                    G_CALLBACK (on_playlist_load_activate),
                    NULL);
  g_signal_connect ((gpointer) playlist_save_as, "activate",
                    G_CALLBACK (on_playlist_save_as_activate),
                    NULL);
  g_signal_connect ((gpointer) quit, "activate",
                    G_CALLBACK (on_quit_activate),
                    NULL);
  g_signal_connect ((gpointer) undo, "activate",
                    G_CALLBACK (on_mainwin_undo_activate),
                    NULL);
  g_signal_connect ((gpointer) redo, "activate",
                    G_CALLBACK (on_mainwin_redo_activate),
                    NULL);
  g_signal_connect ((gpointer) clear1, "activate",
                    G_CALLBACK (on_clear1_activate),
                    NULL);
  g_signal_connect ((gpointer) select_all1, "activate",
                    G_CALLBACK (on_select_all1_activate),
                    NULL);
  g_signal_connect ((gpointer) deselect_all1, "activate",
                    G_CALLBACK (on_deselect_all1_activate),
                    NULL);
  g_signal_connect ((gpointer) invert_selection1, "activate",
                    G_CALLBACK (on_invert_selection1_activate),
                    NULL);
  g_signal_connect ((gpointer) remove1, "activate",
                    G_CALLBACK (on_remove1_activate),
                    NULL);
  g_signal_connect ((gpointer) crop1, "activate",
                    G_CALLBACK (on_crop1_activate),
                    NULL);
  g_signal_connect ((gpointer) find1, "activate",
                    G_CALLBACK (on_find_activate),
                    NULL);
  g_signal_connect ((gpointer) title1, "activate",
                    G_CALLBACK (on_sort_by_title_activate),
                    NULL);
  g_signal_connect ((gpointer) track_nr1, "activate",
                    G_CALLBACK (on_sort_by_track_nr_activate),
                    NULL);
  g_signal_connect ((gpointer) album1, "activate",
                    G_CALLBACK (on_sort_by_album_activate),
                    NULL);
  g_signal_connect ((gpointer) artist1, "activate",
                    G_CALLBACK (on_sort_by_artist_activate),
                    NULL);
  g_signal_connect ((gpointer) date1, "activate",
                    G_CALLBACK (on_sort_by_date_activate),
                    NULL);
  g_signal_connect ((gpointer) random1, "activate",
                    G_CALLBACK (on_sort_by_random_activate),
                    NULL);
  g_signal_connect ((gpointer) custom2, "activate",
                    G_CALLBACK (on_sort_by_custom_activate),
                    NULL);
  g_signal_connect ((gpointer) preferences, "activate",
                    G_CALLBACK (on_preferences_activate),
                    NULL);
  g_signal_connect ((gpointer) view_status_bar, "activate",
                    G_CALLBACK (on_toggle_status_bar_activate),
                    NULL);
  g_signal_connect ((gpointer) view_eq, "activate",
                    G_CALLBACK (on_toggle_eq),
                    NULL);
  g_signal_connect ((gpointer) design_mode1, "activate",
                    G_CALLBACK (on_design_mode1_activate),
                    NULL);
  g_signal_connect ((gpointer) view_log, "activate",
                    G_CALLBACK (on_view_log_activate),
                    NULL);
  g_signal_connect ((gpointer) order_linear, "activate",
                    G_CALLBACK (on_order_linear_activate),
                    NULL);
  g_signal_connect ((gpointer) order_shuffle, "activate",
                    G_CALLBACK (on_order_shuffle_activate),
                    NULL);
  g_signal_connect ((gpointer) order_shuffle_albums, "activate",
                    G_CALLBACK (on_order_shuffle_albums_activate),
                    NULL);
  g_signal_connect ((gpointer) order_random, "activate",
                    G_CALLBACK (on_order_random_activate),
                    NULL);
  g_signal_connect ((gpointer) loop_all, "activate",
                    G_CALLBACK (on_loop_all_activate),
                    NULL);
  g_signal_connect ((gpointer) loop_single, "activate",
                    G_CALLBACK (on_loop_single_activate),
                    NULL);
  g_signal_connect ((gpointer) loop_disable, "activate",
                    G_CALLBACK (on_loop_disable_activate),
                    NULL);
  g_signal_connect ((gpointer) scroll_follows_playback, "activate",
                    G_CALLBACK (on_scroll_follows_playback_activate),
                    NULL);
  g_signal_connect ((gpointer) cursor_follows_playback, "activate",
                    G_CALLBACK (on_cursor_follows_playback_activate),
                    NULL);
  g_signal_connect ((gpointer) stop_after_current, "activate",
                    G_CALLBACK (on_stop_after_current_activate),
                    NULL);
  g_signal_connect ((gpointer) stop_after_album, "activate",
                    G_CALLBACK (on_stop_after_album_activate),
                    NULL);
  g_signal_connect ((gpointer) stop_after_queue, "activate",
                    G_CALLBACK (on_stop_after_queue_activate),
                    NULL);
  g_signal_connect ((gpointer) jump_to_current_track1, "activate",
                    G_CALLBACK (on_jump_to_current_track1_activate),
                    NULL);
  g_signal_connect ((gpointer) help1, "activate",
                    G_CALLBACK (on_help1_activate),
                    NULL);
  g_signal_connect ((gpointer) changelog1, "activate",
                    G_CALLBACK (on_changelog1_activate),
                    NULL);
  g_signal_connect ((gpointer) gpl1, "activate",
                    G_CALLBACK (on_gpl1_activate),
                    NULL);
  g_signal_connect ((gpointer) lgpl1, "activate",
                    G_CALLBACK (on_lgpl1_activate),
                    NULL);
  g_signal_connect ((gpointer) about1, "activate",
                    G_CALLBACK (on_about1_activate),
                    NULL);
  g_signal_connect ((gpointer) translators1, "activate",
                    G_CALLBACK (on_translators1_activate),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (mainwin, mainwin, "mainwin");
  GLADE_HOOKUP_OBJECT (mainwin, vbox1, "vbox1");
  GLADE_HOOKUP_OBJECT (mainwin, menubar, "menubar");
  GLADE_HOOKUP_OBJECT (mainwin, File, "File");
  GLADE_HOOKUP_OBJECT (mainwin, File_menu, "File_menu");
  GLADE_HOOKUP_OBJECT (mainwin, open, "open");
  GLADE_HOOKUP_OBJECT (mainwin, image724, "image724");
  GLADE_HOOKUP_OBJECT (mainwin, separator2, "separator2");
  GLADE_HOOKUP_OBJECT (mainwin, add_files, "add_files");
  GLADE_HOOKUP_OBJECT (mainwin, image725, "image725");
  GLADE_HOOKUP_OBJECT (mainwin, add_folders, "add_folders");
  GLADE_HOOKUP_OBJECT (mainwin, image726, "image726");
  GLADE_HOOKUP_OBJECT (mainwin, add_location1, "add_location1");
  GLADE_HOOKUP_OBJECT (mainwin, separatormenuitem1, "separatormenuitem1");
  GLADE_HOOKUP_OBJECT (mainwin, new_playlist1, "new_playlist1");
  GLADE_HOOKUP_OBJECT (mainwin, playlist_load, "playlist_load");
  GLADE_HOOKUP_OBJECT (mainwin, playlist_save_as, "playlist_save_as");
  GLADE_HOOKUP_OBJECT (mainwin, separator8, "separator8");
  GLADE_HOOKUP_OBJECT (mainwin, quit, "quit");
  GLADE_HOOKUP_OBJECT (mainwin, image727, "image727");
  GLADE_HOOKUP_OBJECT (mainwin, Edit, "Edit");
  GLADE_HOOKUP_OBJECT (mainwin, Edit_menu, "Edit_menu");
  GLADE_HOOKUP_OBJECT (mainwin, undo, "undo");
  GLADE_HOOKUP_OBJECT (mainwin, redo, "redo");
  GLADE_HOOKUP_OBJECT (mainwin, clear1, "clear1");
  GLADE_HOOKUP_OBJECT (mainwin, image728, "image728");
  GLADE_HOOKUP_OBJECT (mainwin, select_all1, "select_all1");
  GLADE_HOOKUP_OBJECT (mainwin, deselect_all1, "deselect_all1");
  GLADE_HOOKUP_OBJECT (mainwin, invert_selection1, "invert_selection1");
  GLADE_HOOKUP_OBJECT (mainwin, Selection, "Selection");
  GLADE_HOOKUP_OBJECT (mainwin, Selection_menu, "Selection_menu");
  GLADE_HOOKUP_OBJECT (mainwin, remove1, "remove1");
  GLADE_HOOKUP_OBJECT (mainwin, image729, "image729");
  GLADE_HOOKUP_OBJECT (mainwin, crop1, "crop1");
  GLADE_HOOKUP_OBJECT (mainwin, find1, "find1");
  GLADE_HOOKUP_OBJECT (mainwin, sort_by1, "sort_by1");
  GLADE_HOOKUP_OBJECT (mainwin, sort_by1_menu, "sort_by1_menu");
  GLADE_HOOKUP_OBJECT (mainwin, title1, "title1");
  GLADE_HOOKUP_OBJECT (mainwin, track_nr1, "track_nr1");
  GLADE_HOOKUP_OBJECT (mainwin, album1, "album1");
  GLADE_HOOKUP_OBJECT (mainwin, artist1, "artist1");
  GLADE_HOOKUP_OBJECT (mainwin, date1, "date1");
  GLADE_HOOKUP_OBJECT (mainwin, random1, "random1");
  GLADE_HOOKUP_OBJECT (mainwin, custom2, "custom2");
  GLADE_HOOKUP_OBJECT (mainwin, separator5, "separator5");
  GLADE_HOOKUP_OBJECT (mainwin, preferences, "preferences");
  GLADE_HOOKUP_OBJECT (mainwin, View, "View");
  GLADE_HOOKUP_OBJECT (mainwin, View_menu, "View_menu");
  GLADE_HOOKUP_OBJECT (mainwin, view_status_bar, "view_status_bar");
  GLADE_HOOKUP_OBJECT (mainwin, view_eq, "view_eq");
  GLADE_HOOKUP_OBJECT (mainwin, design_mode1, "design_mode1");
  GLADE_HOOKUP_OBJECT (mainwin, view_log, "view_log");
  GLADE_HOOKUP_OBJECT (mainwin, Playback, "Playback");
  GLADE_HOOKUP_OBJECT (mainwin, Playback_menu, "Playback_menu");
  GLADE_HOOKUP_OBJECT (mainwin, Order, "Order");
  GLADE_HOOKUP_OBJECT (mainwin, Order_menu, "Order_menu");
  GLADE_HOOKUP_OBJECT (mainwin, order_linear, "order_linear");
  GLADE_HOOKUP_OBJECT (mainwin, order_shuffle, "order_shuffle");
  GLADE_HOOKUP_OBJECT (mainwin, order_shuffle_albums, "order_shuffle_albums");
  GLADE_HOOKUP_OBJECT (mainwin, order_random, "order_random");
  GLADE_HOOKUP_OBJECT (mainwin, Looping, "Looping");
  GLADE_HOOKUP_OBJECT (mainwin, Looping_menu, "Looping_menu");
  GLADE_HOOKUP_OBJECT (mainwin, loop_all, "loop_all");
  GLADE_HOOKUP_OBJECT (mainwin, loop_single, "loop_single");
  GLADE_HOOKUP_OBJECT (mainwin, loop_disable, "loop_disable");
  GLADE_HOOKUP_OBJECT (mainwin, scroll_follows_playback, "scroll_follows_playback");
  GLADE_HOOKUP_OBJECT (mainwin, cursor_follows_playback, "cursor_follows_playback");
  GLADE_HOOKUP_OBJECT (mainwin, stop_after_current, "stop_after_current");
  GLADE_HOOKUP_OBJECT (mainwin, stop_after_album, "stop_after_album");
  GLADE_HOOKUP_OBJECT (mainwin, stop_after_queue, "stop_after_queue");
  GLADE_HOOKUP_OBJECT (mainwin, separator11, "separator11");
  GLADE_HOOKUP_OBJECT (mainwin, jump_to_current_track1, "jump_to_current_track1");
  GLADE_HOOKUP_OBJECT (mainwin, Help, "Help");
  GLADE_HOOKUP_OBJECT (mainwin, Help_menu, "Help_menu");
  GLADE_HOOKUP_OBJECT (mainwin, help1, "help1");
  GLADE_HOOKUP_OBJECT (mainwin, image730, "image730");
  GLADE_HOOKUP_OBJECT (mainwin, changelog1, "changelog1");
  GLADE_HOOKUP_OBJECT (mainwin, separator10, "separator10");
  GLADE_HOOKUP_OBJECT (mainwin, gpl1, "gpl1");
  GLADE_HOOKUP_OBJECT (mainwin, lgpl1, "lgpl1");
  GLADE_HOOKUP_OBJECT (mainwin, separator9, "separator9");
  GLADE_HOOKUP_OBJECT (mainwin, about1, "about1");
  GLADE_HOOKUP_OBJECT (mainwin, image731, "image731");
  GLADE_HOOKUP_OBJECT (mainwin, translators1, "translators1");
  GLADE_HOOKUP_OBJECT (mainwin, image732, "image732");
  GLADE_HOOKUP_OBJECT (mainwin, plugins_bottom_vbox, "plugins_bottom_vbox");
  GLADE_HOOKUP_OBJECT (mainwin, statusbar, "statusbar");

  gtk_window_add_accel_group (GTK_WINDOW (mainwin), accel_group);

  return mainwin;
}

GtkWidget*
create_searchwin (void)
{
  GtkWidget *searchwin;
  GtkWidget *vbox4;
  GtkWidget *searchentry;
  GtkWidget *searchlist;

  searchwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_events (searchwin, GDK_KEY_PRESS_MASK);
  gtk_window_set_title (GTK_WINDOW (searchwin), _("Search"));
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (searchwin), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (searchwin), TRUE);

  vbox4 = gtk_vbox_new (FALSE, 4);
  gtk_widget_show (vbox4);
  gtk_container_add (GTK_CONTAINER (searchwin), vbox4);

  searchentry = gtk_entry_new ();
  gtk_widget_show (searchentry);
  gtk_box_pack_start (GTK_BOX (vbox4), searchentry, FALSE, FALSE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (searchentry), 8226);
  gtk_entry_set_activates_default (GTK_ENTRY (searchentry), TRUE);

  searchlist = create_ddb_listview_widget ("searchlist", "", "", 0, 0);
  gtk_widget_show (searchlist);
  gtk_box_pack_start (GTK_BOX (vbox4), searchlist, TRUE, TRUE, 0);
  gtk_widget_set_can_focus(searchlist, FALSE);
  gtk_widget_set_can_default(searchlist, FALSE);

  g_signal_connect ((gpointer) searchwin, "delete_event",
                    G_CALLBACK (gtk_widget_hide_on_delete),
                    NULL);
  g_signal_connect ((gpointer) searchwin, "configure_event",
                    G_CALLBACK (on_searchwin_configure_event),
                    NULL);
  g_signal_connect ((gpointer) searchwin, "window_state_event",
                    G_CALLBACK (on_searchwin_window_state_event),
                    NULL);
  g_signal_connect ((gpointer) searchwin, "key_press_event",
                    G_CALLBACK (on_searchwin_key_press_event),
                    NULL);
  g_signal_connect ((gpointer) searchentry, "changed",
                    G_CALLBACK (on_searchentry_changed),
                    NULL);
  g_signal_connect ((gpointer) searchentry, "activate",
                    G_CALLBACK (on_searchentry_activate),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (searchwin, searchwin, "searchwin");
  GLADE_HOOKUP_OBJECT (searchwin, vbox4, "vbox4");
  GLADE_HOOKUP_OBJECT (searchwin, searchentry, "searchentry");
  GLADE_HOOKUP_OBJECT (searchwin, searchlist, "searchlist");

  return searchwin;
}

GtkWidget*
create_traymenu (void)
{
  GtkWidget *traymenu;
  GtkWidget *stop2;
  GtkWidget *image59;
  GtkWidget *play2;
  GtkWidget *image60;
  GtkWidget *pause2;
  GtkWidget *image61;
  GtkWidget *previous2;
  GtkWidget *image62;
  GtkWidget *next2;
  GtkWidget *image63;
  GtkWidget *play_random1;
  GtkWidget *separator4;
  GtkWidget *about3;
  GtkWidget *image64;
  GtkWidget *separator3;
  GtkWidget *quit;
  GtkWidget *image65;

  traymenu = gtk_menu_new ();

  stop2 = gtk_image_menu_item_new_with_mnemonic (_("Stop"));
  gtk_widget_show (stop2);
  gtk_container_add (GTK_CONTAINER (traymenu), stop2);

  image59 = gtk_image_new_from_stock ("gtk-media-stop", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image59);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (stop2), image59);

  play2 = gtk_image_menu_item_new_with_mnemonic (_("Play"));
  gtk_widget_show (play2);
  gtk_container_add (GTK_CONTAINER (traymenu), play2);

  image60 = gtk_image_new_from_stock ("gtk-media-play", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image60);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (play2), image60);

  pause2 = gtk_image_menu_item_new_with_mnemonic (_("Pause"));
  gtk_widget_show (pause2);
  gtk_container_add (GTK_CONTAINER (traymenu), pause2);

  image61 = gtk_image_new_from_stock ("gtk-media-pause", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image61);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (pause2), image61);

  previous2 = gtk_image_menu_item_new_with_mnemonic (_("Previous"));
  gtk_widget_show (previous2);
  gtk_container_add (GTK_CONTAINER (traymenu), previous2);

  image62 = gtk_image_new_from_stock ("gtk-media-previous", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image62);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (previous2), image62);

  next2 = gtk_image_menu_item_new_with_mnemonic (_("Next"));
  gtk_widget_show (next2);
  gtk_container_add (GTK_CONTAINER (traymenu), next2);

  image63 = gtk_image_new_from_stock ("gtk-media-next", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image63);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (next2), image63);

  play_random1 = gtk_menu_item_new_with_mnemonic (_("Play Random"));
  gtk_widget_show (play_random1);
  gtk_container_add (GTK_CONTAINER (traymenu), play_random1);

  separator4 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator4);
  gtk_container_add (GTK_CONTAINER (traymenu), separator4);
  gtk_widget_set_sensitive (separator4, FALSE);

  about3 = gtk_image_menu_item_new_with_mnemonic (_("About"));
  gtk_widget_show (about3);
  gtk_container_add (GTK_CONTAINER (traymenu), about3);

  image64 = gtk_image_new_from_stock ("gtk-about", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image64);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (about3), image64);

  separator3 = gtk_separator_menu_item_new ();
  gtk_widget_show (separator3);
  gtk_container_add (GTK_CONTAINER (traymenu), separator3);
  gtk_widget_set_sensitive (separator3, FALSE);

  quit = gtk_image_menu_item_new_with_mnemonic (_("Quit"));
  gtk_widget_show (quit);
  gtk_container_add (GTK_CONTAINER (traymenu), quit);

  image65 = gtk_image_new_from_stock ("gtk-quit", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image65);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (quit), image65);

  g_signal_connect ((gpointer) stop2, "activate",
                    G_CALLBACK (on_stopbtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) play2, "activate",
                    G_CALLBACK (on_playbtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) pause2, "activate",
                    G_CALLBACK (on_pausebtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) previous2, "activate",
                    G_CALLBACK (on_prevbtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) next2, "activate",
                    G_CALLBACK (on_nextbtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) play_random1, "activate",
                    G_CALLBACK (on_playrand_clicked),
                    NULL);
  g_signal_connect ((gpointer) about3, "activate",
                    G_CALLBACK (on_about1_activate),
                    NULL);
  g_signal_connect ((gpointer) quit, "activate",
                    G_CALLBACK (on_quit_activate),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (traymenu, traymenu, "traymenu");
  GLADE_HOOKUP_OBJECT (traymenu, stop2, "stop2");
  GLADE_HOOKUP_OBJECT (traymenu, image59, "image59");
  GLADE_HOOKUP_OBJECT (traymenu, play2, "play2");
  GLADE_HOOKUP_OBJECT (traymenu, image60, "image60");
  GLADE_HOOKUP_OBJECT (traymenu, pause2, "pause2");
  GLADE_HOOKUP_OBJECT (traymenu, image61, "image61");
  GLADE_HOOKUP_OBJECT (traymenu, previous2, "previous2");
  GLADE_HOOKUP_OBJECT (traymenu, image62, "image62");
  GLADE_HOOKUP_OBJECT (traymenu, next2, "next2");
  GLADE_HOOKUP_OBJECT (traymenu, image63, "image63");
  GLADE_HOOKUP_OBJECT (traymenu, play_random1, "play_random1");
  GLADE_HOOKUP_OBJECT (traymenu, separator4, "separator4");
  GLADE_HOOKUP_OBJECT (traymenu, about3, "about3");
  GLADE_HOOKUP_OBJECT (traymenu, image64, "image64");
  GLADE_HOOKUP_OBJECT (traymenu, separator3, "separator3");
  GLADE_HOOKUP_OBJECT (traymenu, quit, "quit");
  GLADE_HOOKUP_OBJECT (traymenu, image65, "image65");

  return traymenu;
}

GtkWidget*
create_progressdlg (void)
{
  GtkWidget *progressdlg;
  GtkWidget *vbox6;
  GtkWidget *progresstitle;
  GtkWidget *hbox7;
  GtkWidget *label22;
  GtkWidget *cancelbtn;

  progressdlg = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width (GTK_CONTAINER (progressdlg), 12);
  gtk_window_set_title (GTK_WINDOW (progressdlg), "progressdlg");
  gtk_window_set_position (GTK_WINDOW (progressdlg), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_modal (GTK_WINDOW (progressdlg), TRUE);
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (progressdlg), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (progressdlg), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (progressdlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  vbox6 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox6);
  gtk_container_add (GTK_CONTAINER (progressdlg), vbox6);

  progresstitle = gtk_entry_new ();
  gtk_widget_show (progresstitle);
  gtk_box_pack_start (GTK_BOX (vbox6), progresstitle, TRUE, FALSE, 0);
  gtk_widget_set_size_request (progresstitle, 500, -1);
  gtk_widget_set_can_focus(progresstitle, FALSE);
  gtk_editable_set_editable (GTK_EDITABLE (progresstitle), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (progresstitle), 8226);

  hbox7 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox7);
  gtk_box_pack_start (GTK_BOX (vbox6), hbox7, FALSE, TRUE, 2);

  label22 = gtk_label_new ("");
  gtk_widget_show (label22);
  gtk_box_pack_start (GTK_BOX (hbox7), label22, TRUE, FALSE, 0);

  cancelbtn = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbtn);
  gtk_box_pack_start (GTK_BOX (hbox7), cancelbtn, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (progressdlg, progressdlg, "progressdlg");
  GLADE_HOOKUP_OBJECT (progressdlg, vbox6, "vbox6");
  GLADE_HOOKUP_OBJECT (progressdlg, progresstitle, "progresstitle");
  GLADE_HOOKUP_OBJECT (progressdlg, hbox7, "hbox7");
  GLADE_HOOKUP_OBJECT (progressdlg, label22, "label22");
  GLADE_HOOKUP_OBJECT (progressdlg, cancelbtn, "cancelbtn");

  return progressdlg;
}

GtkWidget*
create_helpwindow (void)
{
  GtkWidget *helpwindow;
  GtkWidget *scrolledwindow1;
  GtkWidget *helptext;

  helpwindow = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (helpwindow, 600, 400);
  gtk_widget_set_events (helpwindow, GDK_KEY_PRESS_MASK);
  gtk_window_set_title (GTK_WINDOW (helpwindow), _("Help"));
  gtk_window_set_modal (GTK_WINDOW (helpwindow), TRUE);
  gtk_window_set_destroy_with_parent (GTK_WINDOW (helpwindow), TRUE);
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (helpwindow), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (helpwindow), TRUE);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow1);
  gtk_container_add (GTK_CONTAINER (helpwindow), scrolledwindow1);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  helptext = gtk_text_view_new ();
  gtk_widget_show (helptext);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), helptext);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (helptext), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (helptext), GTK_WRAP_WORD);
  gtk_text_view_set_left_margin (GTK_TEXT_VIEW (helptext), 10);
  gtk_text_view_set_right_margin (GTK_TEXT_VIEW (helptext), 10);

  g_signal_connect ((gpointer) helpwindow, "key_press_event",
                    G_CALLBACK (on_helpwindow_key_press_event),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (helpwindow, helpwindow, "helpwindow");
  GLADE_HOOKUP_OBJECT (helpwindow, scrolledwindow1, "scrolledwindow1");
  GLADE_HOOKUP_OBJECT (helpwindow, helptext, "helptext");

  return helpwindow;
}

GtkWidget*
create_trackproperties (void)
{
  GtkWidget *trackproperties;
  GtkWidget *trkproperties_notebook;
  GtkWidget *vbox16;
  GtkWidget *hbox109;
  GtkWidget *label136;
  GtkWidget *filename;
  GtkWidget *scrolledwindow5;
  GtkWidget *metalist;
  GtkWidget *hbox98;
  GtkWidget *settings;
  GtkWidget *alignment24;
  GtkWidget *hbox99;
  GtkWidget *image522;
  GtkWidget *label123;
  GtkWidget *hbuttonbox1;
  GtkWidget *write_tags;
  GtkWidget *alignment11;
  GtkWidget *hbox52;
  GtkWidget *image390;
  GtkWidget *label88;
  GtkWidget *closebtn;
  GtkWidget *alignment12;
  GtkWidget *hbox53;
  GtkWidget *image391;
  GtkWidget *label89;
  GtkWidget *label64;
  GtkWidget *vbox24;
  GtkWidget *scrolledwindow6;
  GtkWidget *properties;
  GtkWidget *hbuttonbox2;
  GtkWidget *button2;
  GtkWidget *alignment13;
  GtkWidget *hbox54;
  GtkWidget *image392;
  GtkWidget *label90;
  GtkWidget *label80;

  trackproperties = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (trackproperties, 400, 400);
  gtk_window_set_title (GTK_WINDOW (trackproperties), _("Track Properties"));
  gtk_window_set_position (GTK_WINDOW (trackproperties), GTK_WIN_POS_MOUSE);
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (trackproperties), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (trackproperties), TRUE);

  trkproperties_notebook = gtk_notebook_new ();
  gtk_widget_show (trkproperties_notebook);
  gtk_container_add (GTK_CONTAINER (trackproperties), trkproperties_notebook);

  vbox16 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox16);
  gtk_container_add (GTK_CONTAINER (trkproperties_notebook), vbox16);
  gtk_container_set_border_width (GTK_CONTAINER (vbox16), 12);

  hbox109 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox109);
  gtk_box_pack_start (GTK_BOX (vbox16), hbox109, FALSE, FALSE, 0);

  label136 = gtk_label_new (_("Location:"));
  gtk_widget_show (label136);
  gtk_box_pack_start (GTK_BOX (hbox109), label136, FALSE, FALSE, 0);

  filename = gtk_entry_new ();
  gtk_widget_show (filename);
  gtk_box_pack_start (GTK_BOX (hbox109), filename, TRUE, TRUE, 0);
  gtk_editable_set_editable (GTK_EDITABLE (filename), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (filename), 8226);

  scrolledwindow5 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow5);
  gtk_box_pack_start (GTK_BOX (vbox16), scrolledwindow5, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow5), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow5), GTK_SHADOW_IN);

  metalist = gtk_tree_view_new ();
  gtk_widget_show (metalist);
  gtk_container_add (GTK_CONTAINER (scrolledwindow5), metalist);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (metalist), TRUE);

  hbox98 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox98);
  gtk_box_pack_start (GTK_BOX (vbox16), hbox98, FALSE, FALSE, 0);

  settings = gtk_button_new ();
  gtk_widget_show (settings);
  gtk_box_pack_start (GTK_BOX (hbox98), settings, FALSE, FALSE, 0);

  alignment24 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment24);
  gtk_container_add (GTK_CONTAINER (settings), alignment24);

  hbox99 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox99);
  gtk_container_add (GTK_CONTAINER (alignment24), hbox99);

  image522 = gtk_image_new_from_stock ("gtk-preferences", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image522);
  gtk_box_pack_start (GTK_BOX (hbox99), image522, FALSE, FALSE, 0);

  label123 = gtk_label_new_with_mnemonic (_("Settings"));
  gtk_widget_show (label123);
  gtk_box_pack_start (GTK_BOX (hbox99), label123, FALSE, FALSE, 0);

  hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (hbox98), hbuttonbox1, TRUE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox1), GTK_BUTTONBOX_END);

  write_tags = gtk_button_new ();
  gtk_widget_show (write_tags);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), write_tags);
  gtk_widget_set_can_default(write_tags, TRUE);

  alignment11 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment11);
  gtk_container_add (GTK_CONTAINER (write_tags), alignment11);

  hbox52 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox52);
  gtk_container_add (GTK_CONTAINER (alignment11), hbox52);

  image390 = gtk_image_new_from_stock ("gtk-apply", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image390);
  gtk_box_pack_start (GTK_BOX (hbox52), image390, FALSE, FALSE, 0);

  label88 = gtk_label_new_with_mnemonic (_("_Apply"));
  gtk_widget_show (label88);
  gtk_box_pack_start (GTK_BOX (hbox52), label88, FALSE, FALSE, 0);

  closebtn = gtk_button_new ();
  gtk_widget_show (closebtn);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), closebtn);
  gtk_widget_set_can_default(closebtn, TRUE);

  alignment12 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment12);
  gtk_container_add (GTK_CONTAINER (closebtn), alignment12);

  hbox53 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox53);
  gtk_container_add (GTK_CONTAINER (alignment12), hbox53);

  image391 = gtk_image_new_from_stock ("gtk-close", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image391);
  gtk_box_pack_start (GTK_BOX (hbox53), image391, FALSE, FALSE, 0);

  label89 = gtk_label_new_with_mnemonic (_("_Close"));
  gtk_widget_show (label89);
  gtk_box_pack_start (GTK_BOX (hbox53), label89, FALSE, FALSE, 0);

  label64 = gtk_label_new (_("Metadata"));
  gtk_widget_show (label64);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (trkproperties_notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (trkproperties_notebook), 0), label64);

  vbox24 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox24);
  gtk_container_add (GTK_CONTAINER (trkproperties_notebook), vbox24);
  gtk_container_set_border_width (GTK_CONTAINER (vbox24), 12);

  scrolledwindow6 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow6);
  gtk_box_pack_start (GTK_BOX (vbox24), scrolledwindow6, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow6), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow6), GTK_SHADOW_IN);

  properties = gtk_tree_view_new ();
  gtk_widget_show (properties);
  gtk_container_add (GTK_CONTAINER (scrolledwindow6), properties);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (properties), TRUE);

  hbuttonbox2 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox2);
  gtk_box_pack_start (GTK_BOX (vbox24), hbuttonbox2, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox2), GTK_BUTTONBOX_END);

  button2 = gtk_button_new ();
  gtk_widget_show (button2);
  gtk_container_add (GTK_CONTAINER (hbuttonbox2), button2);
  gtk_widget_set_can_default(button2, TRUE);

  alignment13 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment13);
  gtk_container_add (GTK_CONTAINER (button2), alignment13);

  hbox54 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox54);
  gtk_container_add (GTK_CONTAINER (alignment13), hbox54);

  image392 = gtk_image_new_from_stock ("gtk-close", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image392);
  gtk_box_pack_start (GTK_BOX (hbox54), image392, FALSE, FALSE, 0);

  label90 = gtk_label_new_with_mnemonic (_("_Close"));
  gtk_widget_show (label90);
  gtk_box_pack_start (GTK_BOX (hbox54), label90, FALSE, FALSE, 0);

  label80 = gtk_label_new (_("Properties"));
  gtk_widget_show (label80);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (trkproperties_notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (trkproperties_notebook), 1), label80);

  g_signal_connect ((gpointer) trackproperties, "key_press_event",
                    G_CALLBACK (on_trackproperties_key_press_event),
                    NULL);
  g_signal_connect ((gpointer) trackproperties, "delete_event",
                    G_CALLBACK (on_trackproperties_delete_event),
                    NULL);
  g_signal_connect ((gpointer) trackproperties, "configure_event",
                    G_CALLBACK (on_trackproperties_configure_event),
                    NULL);
  g_signal_connect ((gpointer) trackproperties, "window_state_event",
                    G_CALLBACK (on_trackproperties_window_state_event),
                    NULL);
  g_signal_connect ((gpointer) metalist, "button_press_event",
                    G_CALLBACK (on_metalist_button_press_event),
                    NULL);
  g_signal_connect ((gpointer) settings, "clicked",
                    G_CALLBACK (on_tagwriter_settings_clicked),
                    NULL);
  g_signal_connect ((gpointer) write_tags, "clicked",
                    G_CALLBACK (on_write_tags_clicked),
                    NULL);
  g_signal_connect ((gpointer) closebtn, "clicked",
                    G_CALLBACK (on_closebtn_clicked),
                    NULL);
  g_signal_connect ((gpointer) button2, "clicked",
                    G_CALLBACK (on_closebtn_clicked),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (trackproperties, trackproperties, "trackproperties");
  GLADE_HOOKUP_OBJECT (trackproperties, trkproperties_notebook, "trkproperties_notebook");
  GLADE_HOOKUP_OBJECT (trackproperties, vbox16, "vbox16");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox109, "hbox109");
  GLADE_HOOKUP_OBJECT (trackproperties, label136, "label136");
  GLADE_HOOKUP_OBJECT (trackproperties, filename, "filename");
  GLADE_HOOKUP_OBJECT (trackproperties, scrolledwindow5, "scrolledwindow5");
  GLADE_HOOKUP_OBJECT (trackproperties, metalist, "metalist");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox98, "hbox98");
  GLADE_HOOKUP_OBJECT (trackproperties, settings, "settings");
  GLADE_HOOKUP_OBJECT (trackproperties, alignment24, "alignment24");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox99, "hbox99");
  GLADE_HOOKUP_OBJECT (trackproperties, image522, "image522");
  GLADE_HOOKUP_OBJECT (trackproperties, label123, "label123");
  GLADE_HOOKUP_OBJECT (trackproperties, hbuttonbox1, "hbuttonbox1");
  GLADE_HOOKUP_OBJECT (trackproperties, write_tags, "write_tags");
  GLADE_HOOKUP_OBJECT (trackproperties, alignment11, "alignment11");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox52, "hbox52");
  GLADE_HOOKUP_OBJECT (trackproperties, image390, "image390");
  GLADE_HOOKUP_OBJECT (trackproperties, label88, "label88");
  GLADE_HOOKUP_OBJECT (trackproperties, closebtn, "closebtn");
  GLADE_HOOKUP_OBJECT (trackproperties, alignment12, "alignment12");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox53, "hbox53");
  GLADE_HOOKUP_OBJECT (trackproperties, image391, "image391");
  GLADE_HOOKUP_OBJECT (trackproperties, label89, "label89");
  GLADE_HOOKUP_OBJECT (trackproperties, label64, "label64");
  GLADE_HOOKUP_OBJECT (trackproperties, vbox24, "vbox24");
  GLADE_HOOKUP_OBJECT (trackproperties, scrolledwindow6, "scrolledwindow6");
  GLADE_HOOKUP_OBJECT (trackproperties, properties, "properties");
  GLADE_HOOKUP_OBJECT (trackproperties, hbuttonbox2, "hbuttonbox2");
  GLADE_HOOKUP_OBJECT (trackproperties, button2, "button2");
  GLADE_HOOKUP_OBJECT (trackproperties, alignment13, "alignment13");
  GLADE_HOOKUP_OBJECT (trackproperties, hbox54, "hbox54");
  GLADE_HOOKUP_OBJECT (trackproperties, image392, "image392");
  GLADE_HOOKUP_OBJECT (trackproperties, label90, "label90");
  GLADE_HOOKUP_OBJECT (trackproperties, label80, "label80");

  return trackproperties;
}

GtkWidget*
create_editcolumndlg (void)
{
  GtkWidget *editcolumndlg;
  GtkWidget *dialog_vbox1;
  GtkWidget *vbox14;
  GtkWidget *hbox29;
  GtkWidget *label26;
  GtkWidget *title;
  GtkWidget *hbox30;
  GtkWidget *label37;
  GtkWidget *id;
  GtkWidget *hbox31;
  GtkWidget *fmtlabel;
  GtkWidget *hbox74;
  GtkWidget *format;
  GtkWidget *title_formatting_help_link;
  GtkWidget *hbox141;
  GtkWidget *label164;
  GtkWidget *sort_format;
  GtkWidget *label165;
  GtkWidget *hbox32;
  GtkWidget *label38;
  GtkWidget *align;
  GtkWidget *hbox138;
  GtkWidget *color_override;
  GtkWidget *color;
  GtkWidget *dialog_action_area1;
  GtkWidget *cancelbutton1;
  GtkWidget *alignment9;
  GtkWidget *hbox50;
  GtkWidget *image388;
  GtkWidget *label86;
  GtkWidget *okbutton1;
  GtkWidget *alignment8;
  GtkWidget *hbox49;
  GtkWidget *image387;
  GtkWidget *label85;

  editcolumndlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (editcolumndlg), "editcolumndlg");
  gtk_window_set_position (GTK_WINDOW (editcolumndlg), GTK_WIN_POS_MOUSE);
  gtk_window_set_modal (GTK_WINDOW (editcolumndlg), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (editcolumndlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox1 = gtk_dialog_get_content_area (GTK_DIALOG (editcolumndlg));
  gtk_widget_show (dialog_vbox1);

  vbox14 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox14);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), vbox14, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox14), 12);

  hbox29 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox29);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox29, FALSE, FALSE, 0);

  label26 = gtk_label_new (_("Title:"));
  gtk_widget_show (label26);
  gtk_box_pack_start (GTK_BOX (hbox29), label26, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label26), 0, 0.5);

  title = gtk_entry_new ();
  gtk_widget_show (title);
  gtk_box_pack_start (GTK_BOX (hbox29), title, TRUE, TRUE, 0);
  gtk_entry_set_text (GTK_ENTRY (title), _("Enter new column title here"));
  gtk_entry_set_invisible_char (GTK_ENTRY (title), 9679);
  gtk_entry_set_activates_default (GTK_ENTRY (title), TRUE);

  hbox30 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox30);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox30, FALSE, FALSE, 0);

  label37 = gtk_label_new (_("Type:"));
  gtk_widget_show (label37);
  gtk_box_pack_start (GTK_BOX (hbox30), label37, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label37), 0, 0.5);

  id = gtk_combo_box_text_new ();
  gtk_widget_show (id);
  gtk_box_pack_start (GTK_BOX (hbox30), id, TRUE, TRUE, 0);

  hbox31 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox31);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox31, FALSE, FALSE, 0);

  fmtlabel = gtk_label_new (_("Format:"));
  gtk_widget_show (fmtlabel);
  gtk_box_pack_start (GTK_BOX (hbox31), fmtlabel, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (fmtlabel), 0, 0.5);

  hbox74 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox74);
  gtk_box_pack_start (GTK_BOX (hbox31), hbox74, TRUE, TRUE, 0);

  format = gtk_entry_new ();
  gtk_widget_show (format);
  gtk_box_pack_start (GTK_BOX (hbox74), format, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (format), 9679);
  gtk_entry_set_activates_default (GTK_ENTRY (format), TRUE);

  title_formatting_help_link = title_formatting_help_link_create ("title_formatting_help_link", "", "", 0, 0);
  gtk_widget_show (title_formatting_help_link);
  gtk_box_pack_start (GTK_BOX (hbox74), title_formatting_help_link, FALSE, TRUE, 0);
  gtk_widget_set_can_focus(title_formatting_help_link, FALSE);
  gtk_widget_set_can_default(title_formatting_help_link, FALSE);

  hbox141 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox141);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox141, TRUE, TRUE, 0);

  label164 = gtk_label_new (_("Sort format:"));
  gtk_widget_show (label164);
  gtk_box_pack_start (GTK_BOX (hbox141), label164, FALSE, FALSE, 0);

  sort_format = gtk_entry_new ();
  gtk_widget_show (sort_format);
  gtk_box_pack_start (GTK_BOX (hbox141), sort_format, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (sort_format), 8226);

  label165 = gtk_label_new (_("(if non-empty)"));
  gtk_widget_show (label165);
  gtk_box_pack_start (GTK_BOX (hbox141), label165, FALSE, FALSE, 0);

  hbox32 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox32);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox32, FALSE, FALSE, 0);

  label38 = gtk_label_new (_("Alignment:"));
  gtk_widget_show (label38);
  gtk_box_pack_start (GTK_BOX (hbox32), label38, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label38), 0, 0.5);

  align = gtk_combo_box_text_new ();
  gtk_widget_show (align);
  gtk_box_pack_start (GTK_BOX (hbox32), align, TRUE, TRUE, 0);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (align), _("Left"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (align), _("Right"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (align), _("Center"));

  hbox138 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox138);
  gtk_box_pack_start (GTK_BOX (vbox14), hbox138, TRUE, TRUE, 0);

  color_override = gtk_check_button_new_with_mnemonic (_("Text color:"));
  gtk_widget_show (color_override);
  gtk_box_pack_start (GTK_BOX (hbox138), color_override, FALSE, FALSE, 0);

  color = gtk_color_button_new ();
  gtk_widget_show (color);
  gtk_box_pack_start (GTK_BOX (hbox138), color, TRUE, TRUE, 0);

  dialog_action_area1 = gtk_dialog_get_action_area (GTK_DIALOG (editcolumndlg));
  gtk_widget_show (dialog_action_area1);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area1), GTK_BUTTONBOX_END);

  cancelbutton1 = gtk_button_new ();
  gtk_widget_show (cancelbutton1);
  gtk_dialog_add_action_widget (GTK_DIALOG (editcolumndlg), cancelbutton1, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton1, TRUE);

  alignment9 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment9);
  gtk_container_add (GTK_CONTAINER (cancelbutton1), alignment9);

  hbox50 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox50);
  gtk_container_add (GTK_CONTAINER (alignment9), hbox50);

  image388 = gtk_image_new_from_stock ("gtk-cancel", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image388);
  gtk_box_pack_start (GTK_BOX (hbox50), image388, FALSE, FALSE, 0);

  label86 = gtk_label_new_with_mnemonic (_("_Cancel"));
  gtk_widget_show (label86);
  gtk_box_pack_start (GTK_BOX (hbox50), label86, FALSE, FALSE, 0);

  okbutton1 = gtk_button_new ();
  gtk_widget_show (okbutton1);
  gtk_dialog_add_action_widget (GTK_DIALOG (editcolumndlg), okbutton1, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton1, TRUE);

  alignment8 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment8);
  gtk_container_add (GTK_CONTAINER (okbutton1), alignment8);

  hbox49 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox49);
  gtk_container_add (GTK_CONTAINER (alignment8), hbox49);

  image387 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image387);
  gtk_box_pack_start (GTK_BOX (hbox49), image387, FALSE, FALSE, 0);

  label85 = gtk_label_new_with_mnemonic (_("_OK"));
  gtk_widget_show (label85);
  gtk_box_pack_start (GTK_BOX (hbox49), label85, FALSE, FALSE, 0);

  g_signal_connect ((gpointer) title, "changed",
                    G_CALLBACK (on_editcolumn_title_changed),
                    NULL);
  g_signal_connect ((gpointer) id, "changed",
                    G_CALLBACK (on_column_id_changed),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (editcolumndlg, editcolumndlg, "editcolumndlg");
  GLADE_HOOKUP_OBJECT_NO_REF (editcolumndlg, dialog_vbox1, "dialog_vbox1");
  GLADE_HOOKUP_OBJECT (editcolumndlg, vbox14, "vbox14");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox29, "hbox29");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label26, "label26");
  GLADE_HOOKUP_OBJECT (editcolumndlg, title, "title");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox30, "hbox30");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label37, "label37");
  GLADE_HOOKUP_OBJECT (editcolumndlg, id, "id");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox31, "hbox31");
  GLADE_HOOKUP_OBJECT (editcolumndlg, fmtlabel, "fmtlabel");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox74, "hbox74");
  GLADE_HOOKUP_OBJECT (editcolumndlg, format, "format");
  GLADE_HOOKUP_OBJECT (editcolumndlg, title_formatting_help_link, "title_formatting_help_link");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox141, "hbox141");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label164, "label164");
  GLADE_HOOKUP_OBJECT (editcolumndlg, sort_format, "sort_format");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label165, "label165");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox32, "hbox32");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label38, "label38");
  GLADE_HOOKUP_OBJECT (editcolumndlg, align, "align");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox138, "hbox138");
  GLADE_HOOKUP_OBJECT (editcolumndlg, color_override, "color_override");
  GLADE_HOOKUP_OBJECT (editcolumndlg, color, "color");
  GLADE_HOOKUP_OBJECT_NO_REF (editcolumndlg, dialog_action_area1, "dialog_action_area1");
  GLADE_HOOKUP_OBJECT (editcolumndlg, cancelbutton1, "cancelbutton1");
  GLADE_HOOKUP_OBJECT (editcolumndlg, alignment9, "alignment9");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox50, "hbox50");
  GLADE_HOOKUP_OBJECT (editcolumndlg, image388, "image388");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label86, "label86");
  GLADE_HOOKUP_OBJECT (editcolumndlg, okbutton1, "okbutton1");
  GLADE_HOOKUP_OBJECT (editcolumndlg, alignment8, "alignment8");
  GLADE_HOOKUP_OBJECT (editcolumndlg, hbox49, "hbox49");
  GLADE_HOOKUP_OBJECT (editcolumndlg, image387, "image387");
  GLADE_HOOKUP_OBJECT (editcolumndlg, label85, "label85");

  return editcolumndlg;
}

GtkWidget*
create_prefwin (void)
{
  GtkWidget *prefwin;
  GtkWidget *dialog_vbox2;
  GtkWidget *notebook;
  GtkWidget *vbox10;
  GtkWidget *hbox11;
  GtkWidget *label23;
  GtkWidget *pref_output_plugin;
  GtkWidget *hbox12;
  GtkWidget *label4;
  GtkWidget *pref_soundcard;
  GtkWidget *convert8to16;
  GtkWidget *convert16to24;
  GtkWidget *hbox151;
  GtkWidget *label_bit_override;
  GtkWidget *combo_bit_override;
  GtkWidget *frame15;
  GtkWidget *alignment31;
  GtkWidget *vbox52;
  GtkWidget *hbox145;
  GtkWidget *label_direct_sr;
  GtkWidget *comboboxentry_direct_sr;
  GtkWidget *frame16;
  GtkWidget *alignment32;
  GtkWidget *vbox53;
  GtkWidget *hbox147;
  GtkWidget *label_sr_mult_48;
  GtkWidget *comboboxentry_sr_mult_48;
  GtkWidget *hbox148;
  GtkWidget *label_sr_mult_44;
  GtkWidget *comboboxentry_sr_mult_44;
  GtkWidget *checkbutton_dependent_sr;
  GtkWidget *checkbutton_sr_override;
  GtkWidget *Sound;
  GtkWidget *vbox8;
  GtkWidget *frame14;
  GtkWidget *alignment30;
  GtkWidget *vbox48;
  GtkWidget *hbox10;
  GtkWidget *label8;
  GtkWidget *pref_replaygain_source_mode;
  GtkWidget *hbox139;
  GtkWidget *label163;
  GtkWidget *pref_replaygain_processing;
  GtkWidget *table1;
  GtkWidget *hbox100;
  GtkWidget *label125;
  GtkWidget *replaygain_preamp;
  GtkWidget *label126;
  GtkWidget *label124;
  GtkWidget *label124a;
  GtkWidget *hbox100a;
  GtkWidget *label125a;
  GtkWidget *global_preamp;
  GtkWidget *label126a;
  GtkWidget *label162;
  GtkWidget *hbox66;
  GtkWidget *cli_add_to_playlist;
  GtkWidget *cli_playlist_name;
  GtkWidget *resume_last_session;
  GtkWidget *resume_always_paused;
  GtkWidget *ignore_archives;
  GtkWidget *reset_autostop;
  GtkWidget *reset_autostopqueue;
  GtkWidget *reset_autostopalbum;
  GtkWidget *label39;
  GtkWidget *vbox29;
  GtkWidget *prefwin_dsp_vbox;
  GtkWidget *label110;
  GtkWidget *notebook5;
  GtkWidget *vbox9;
  GtkWidget *minimize_on_startup;
  GtkWidget *pref_close_send_to_tray;
  GtkWidget *hide_tray_icon;
  GtkWidget *move_to_trash;
  GtkWidget *enable_shift_jis_recoding;
  GtkWidget *enable_cp1251_recoding;
  GtkWidget *enable_cp936_recoding;
  GtkWidget *hbox102;
  GtkWidget *label129;
  GtkWidget *gui_fps;
  GtkWidget *hbox64;
  GtkWidget *label101;
  GtkWidget *titlebar_format_playing;
  GtkWidget *hbox65;
  GtkWidget *label102;
  GtkWidget *titlebar_format_stopped;
  GtkWidget *hbox101;
  GtkWidget *label128;
  GtkWidget *gui_plugin;
  GtkWidget *display_seltime;
  GtkWidget *label147;
  GtkWidget *vbox44;
  GtkWidget *mmb_delete_playlist;
  GtkWidget *hide_delete_from_disk;
  GtkWidget *skip_deleted_songs;
  GtkWidget *auto_name_playlist_from_folder;
  GtkWidget *auto_size_columns;
  GtkWidget *hbox141;
  GtkWidget *label164;
  GObject *listview_group_spacing_adj;
  GtkWidget *listview_group_spacing;
  GtkWidget *label165;
  GtkWidget *label148;
  GtkWidget *label2;
  GtkWidget *notebook4;
  GtkWidget *vbox21;
  GtkWidget *override_bar_colors;
  GtkWidget *bar_colors_group;
  GtkWidget *label43;
  GtkWidget *label47;
  GtkWidget *bar_foreground;
  GtkWidget *bar_background;
  GtkWidget *label73;
  GtkWidget *vbox22;
  GtkWidget *override_tabstrip_colors;
  GtkWidget *tabstrip_colors_group;
  GtkWidget *tabstrip_base;
  GtkWidget *label76;
  GtkWidget *tabstrip_dark;
  GtkWidget *label44;
  GtkWidget *tabstrip_light;
  GtkWidget *label46;
  GtkWidget *tabstrip_mid;
  GtkWidget *label45;
  GtkWidget *label127;
  GtkWidget *tabstrip_text;
  GtkWidget *tabstrip_selected_text;
  GtkWidget *hbox136;
  GtkWidget *tabstrip_playing_bold;
  GtkWidget *tabstrip_playing_italic;
  GtkWidget *hbox137;
  GtkWidget *tabstrip_selected_bold;
  GtkWidget *tabstrip_selected_italic;
  GtkWidget *tabstrip_text_font;
  GtkWidget *label152;
  GtkWidget *label153;
  GtkWidget *tabstrip_playing_text;
  GtkWidget *label74;
  GtkWidget *vbox23;
  GtkWidget *frame11;
  GtkWidget *listview_colors_group;
  GtkWidget *frame12;
  GtkWidget *alignment28;
  GtkWidget *table3;
  GtkWidget *label77;
  GtkWidget *listview_text;
  GtkWidget *label150;
  GtkWidget *listview_playing_text;
  GtkWidget *label61;
  GtkWidget *listview_selected_text;
  GtkWidget *label154;
  GtkWidget *listview_column_text;
  GtkWidget *listview_column_text_font;
  GtkWidget *label149;
  GtkWidget *listview_group_text;
  GtkWidget *listview_group_text_font;
  GtkWidget *listview_text_font;
  GtkWidget *label158;
  GtkWidget *vseparator1;
  GtkWidget *frame13;
  GtkWidget *alignment29;
  GtkWidget *table4;
  GtkWidget *label58;
  GtkWidget *label59;
  GtkWidget *listview_odd_row;
  GtkWidget *label78;
  GtkWidget *listview_selected_row;
  GtkWidget *label62;
  GtkWidget *listview_cursor;
  GtkWidget *listview_even_row;
  GtkWidget *label159;
  GtkWidget *override_listview_colors;
  GtkWidget *table2;
  GtkWidget *label155;
  GtkWidget *label156;
  GtkWidget *listview_playing_text_bold;
  GtkWidget *listview_selected_text_bold;
  GtkWidget *listview_playing_text_italic;
  GtkWidget *listview_selected_text_italic;
  GtkWidget *label75;
  GtkWidget *table5;
  GtkWidget *use_visualization_base_color;
  GtkWidget *visualization_custom_color_button;
  GtkWidget *use_visualization_background_color;
  GtkWidget *visualization_custom_background_color_button;
  GtkWidget *label175;
  GtkWidget *label100;
  GtkWidget *vbox55;
  GtkWidget *toggle_medialib_on;
  GtkWidget *hseparator7;
  GtkWidget *label174;
  GtkWidget *vbox56;
  GtkWidget *scrolledwindow17;
  GtkWidget *treeview_medialib_folders;
  GtkWidget *hbox150;
  GtkWidget *button_medialib_add_folder;
  GtkWidget *image714;
  GtkWidget *button_medialib_remove_folder;
  GtkWidget *image715;
  GtkWidget *label173;
  GtkWidget *vbox11;
  GtkWidget *pref_network_enableproxy;
  GtkWidget *hbox13;
  GtkWidget *label18;
  GtkWidget *pref_network_proxyaddress;
  GtkWidget *hbox14;
  GtkWidget *label19;
  GtkWidget *pref_network_proxyport;
  GtkWidget *hbox15;
  GtkWidget *label20;
  GtkWidget *pref_network_proxytype;
  GtkWidget *hbox61;
  GtkWidget *label97;
  GtkWidget *proxyuser;
  GtkWidget *hbox62;
  GtkWidget *label98;
  GtkWidget *proxypassword;
  GtkWidget *hseparator1;
  GtkWidget *hbox104;
  GtkWidget *label131;
  GtkWidget *useragent;
  GtkWidget *hseparator5;
  GtkWidget *edit_content_type_mapping;
  GtkWidget *label16;
  GtkWidget *vbox36;
  GtkWidget *hbox119;
  GtkWidget *label144;
  GtkWidget *hbuttonbox3;
  GtkWidget *hotkey_add;
  GtkWidget *hotkey_remove;
  GtkWidget *scrolledwindow9;
  GtkWidget *hotkeys_list;
  GtkWidget *hbox105;
  GtkWidget *label133;
  GtkWidget *hotkeys_actions;
  GtkWidget *hbox106;
  GtkWidget *label134;
  GtkWidget *hotkeys_set_key;
  GtkWidget *hseparator6;
  GtkWidget *hbox120;
  GtkWidget *label145;
  GtkWidget *hbuttonbox4;
  GtkWidget *hotkeys_apply;
  GtkWidget *hotkeys_revert;
  GtkWidget *hotkeys_defaults;
  GtkWidget *alignment26;
  GtkWidget *hbox121;
  GtkWidget *image638;
  GtkWidget *label146;
  GtkWidget *label132;
  GtkWidget *hpaned1;
  GtkWidget *scrolledwindow2;
  GtkWidget *pref_pluginlist;
  GtkWidget *alignment33;
  GtkWidget *vbox54;
  GtkWidget *hbox149;
  GtkWidget *fillerlabel1;
  GtkWidget *plugin_tabbtn_hbtnbox;
  GtkWidget *plugin_conf_tab_btn;
  GSList *plugin_conf_tab_btn_group = NULL;
  GtkWidget *plugin_info_tab_btn;
  GtkWidget *plugin_license_tab_btn;
  GtkWidget *fillerlabel2;
  GtkWidget *plugin_notebook;
  GtkWidget *plugin_conf_tab_vbox;
  GtkWidget *plug_conf_dlg_scrolledwindow;
  GtkWidget *plug_conf_dlg_viewport;
  GtkWidget *plugin_actions_btnbox;
  GtkWidget *plugin_conf_reset_btn;
  GtkWidget *label170;
  GtkWidget *vbox12;
  GtkWidget *hbox103;
  GtkWidget *label130;
  GtkWidget *plug_version;
  GtkWidget *scrolledwindow8;
  GtkWidget *plug_description;
  GtkWidget *hbox20;
  GtkWidget *weblink;
  GtkWidget *label171;
  GtkWidget *scrolledwindow16;
  GtkWidget *plug_license;
  GtkWidget *label172;
  GtkWidget *label3;
  GtkWidget *dialog_action_area2;
  GtkWidget *closebutton1;

  prefwin = gtk_dialog_new ();
  gtk_container_set_border_width (GTK_CONTAINER (prefwin), 6);
  gtk_window_set_title (GTK_WINDOW (prefwin), _("Preferences"));
  gtk_window_set_position (GTK_WINDOW (prefwin), GTK_WIN_POS_CENTER);
  gtk_window_set_type_hint (GTK_WINDOW (prefwin), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox2 = gtk_dialog_get_content_area (GTK_DIALOG (prefwin));
  gtk_widget_show (dialog_vbox2);

  notebook = gtk_notebook_new ();
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (dialog_vbox2), notebook, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (notebook), 6);
  gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook), TRUE);

  vbox10 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox10);
  gtk_container_add (GTK_CONTAINER (notebook), vbox10);
  gtk_container_set_border_width (GTK_CONTAINER (vbox10), 12);

  hbox11 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox11);
  gtk_box_pack_start (GTK_BOX (vbox10), hbox11, FALSE, FALSE, 0);

  label23 = gtk_label_new (_("Output plugin:"));
  gtk_widget_show (label23);
  gtk_box_pack_start (GTK_BOX (hbox11), label23, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label23), 0, 0.5);

  pref_output_plugin = gtk_combo_box_text_new ();
  gtk_widget_show (pref_output_plugin);
  gtk_box_pack_start (GTK_BOX (hbox11), pref_output_plugin, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (pref_output_plugin), 1);

  hbox12 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox12);
  gtk_box_pack_start (GTK_BOX (vbox10), hbox12, FALSE, FALSE, 0);

  label4 = gtk_label_new (_("Output device:"));
  gtk_widget_show (label4);
  gtk_box_pack_start (GTK_BOX (hbox12), label4, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label4), 0, 0.5);

  pref_soundcard = gtk_combo_box_text_new ();
  gtk_widget_show (pref_soundcard);
  gtk_box_pack_start (GTK_BOX (hbox12), pref_soundcard, TRUE, TRUE, 0);

  convert8to16 = gtk_check_button_new_with_mnemonic (_("Always convert 8 bit audio to 16 bit"));
  gtk_widget_show (convert8to16);
  gtk_box_pack_start (GTK_BOX (vbox10), convert8to16, FALSE, FALSE, 0);

  convert16to24 = gtk_check_button_new_with_mnemonic (_("Always convert 16 bit audio to 24 bit"));
  gtk_widget_show (convert16to24);
  gtk_box_pack_start (GTK_BOX (vbox10), convert16to24, FALSE, FALSE, 0);

  hbox151 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox151);
  gtk_box_pack_start (GTK_BOX (vbox10), hbox151, FALSE, TRUE, 0);

  label_bit_override = gtk_label_new (_("Override bit depth:"));
  gtk_widget_show (label_bit_override);
  gtk_box_pack_start (GTK_BOX (hbox151), label_bit_override, FALSE, FALSE, 0);

  combo_bit_override = gtk_combo_box_text_new ();
  gtk_widget_show (combo_bit_override);
  gtk_box_pack_start (GTK_BOX (hbox151), combo_bit_override, TRUE, TRUE, 0);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("Disabled"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("8 bit"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("16 bit"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("24 bit"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("32 bit"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo_bit_override), _("32 bit float"));

  frame15 = gtk_frame_new (NULL);
  gtk_widget_show (frame15);
  gtk_box_pack_start (GTK_BOX (vbox10), frame15, FALSE, TRUE, 0);

  alignment31 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment31);
  gtk_container_add (GTK_CONTAINER (frame15), alignment31);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment31), 8, 0, 12, 0);

  vbox52 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox52);
  gtk_container_add (GTK_CONTAINER (alignment31), vbox52);

  hbox145 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox145);
  gtk_box_pack_start (GTK_BOX (vbox52), hbox145, TRUE, TRUE, 0);

  label_direct_sr = gtk_label_new (_("Target samplerate:"));
  gtk_widget_show (label_direct_sr);
  gtk_box_pack_start (GTK_BOX (hbox145), label_direct_sr, FALSE, FALSE, 0);

  comboboxentry_direct_sr = gtk_combo_box_entry_new_text ();
  gtk_widget_show (comboboxentry_direct_sr);
  gtk_box_pack_start (GTK_BOX (hbox145), comboboxentry_direct_sr, FALSE, FALSE, 0);
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "44100");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "48000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "88200");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "96000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "176400");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_direct_sr), "192000");

  frame16 = gtk_frame_new (NULL);
  gtk_widget_show (frame16);
  gtk_box_pack_start (GTK_BOX (vbox52), frame16, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame16), 8);

  alignment32 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment32);
  gtk_container_add (GTK_CONTAINER (frame16), alignment32);
  gtk_container_set_border_width (GTK_CONTAINER (alignment32), 8);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment32), 0, 0, 12, 0);

  vbox53 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox53);
  gtk_container_add (GTK_CONTAINER (alignment32), vbox53);

  hbox147 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox147);
  gtk_box_pack_start (GTK_BOX (vbox53), hbox147, TRUE, TRUE, 0);

  label_sr_mult_48 = gtk_label_new (_("For multiples of 48KHz (96K, 192K, ...):"));
  gtk_widget_show (label_sr_mult_48);
  gtk_box_pack_start (GTK_BOX (hbox147), label_sr_mult_48, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label_sr_mult_48), 1, 0.5);

  comboboxentry_sr_mult_48 = gtk_combo_box_entry_new_text ();
  gtk_widget_show (comboboxentry_sr_mult_48);
  gtk_box_pack_start (GTK_BOX (hbox147), comboboxentry_sr_mult_48, FALSE, FALSE, 0);
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "44100");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "48000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "88200");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "96000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "176400");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_48), "192000");

  hbox148 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox148);
  gtk_box_pack_start (GTK_BOX (vbox53), hbox148, TRUE, TRUE, 0);

  label_sr_mult_44 = gtk_label_new (_("For multiples of 44.1KHz (88.2K, 176.4K, ...):"));
  gtk_widget_show (label_sr_mult_44);
  gtk_box_pack_start (GTK_BOX (hbox148), label_sr_mult_44, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label_sr_mult_44), 1, 0.5);

  comboboxentry_sr_mult_44 = gtk_combo_box_entry_new_text ();
  gtk_widget_show (comboboxentry_sr_mult_44);
  gtk_box_pack_start (GTK_BOX (hbox148), comboboxentry_sr_mult_44, FALSE, FALSE, 0);
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "44100");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "48000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "88200");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "96000");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "176400");
  gtk_combo_box_append_text (GTK_COMBO_BOX (comboboxentry_sr_mult_44), "192000");

  checkbutton_dependent_sr = gtk_check_button_new_with_mnemonic (_("Based on input samplerate"));
  gtk_widget_show (checkbutton_dependent_sr);
  gtk_frame_set_label_widget (GTK_FRAME (frame16), checkbutton_dependent_sr);

  checkbutton_sr_override = gtk_check_button_new_with_mnemonic (_("Override samplerate"));
  gtk_widget_show (checkbutton_sr_override);
  gtk_frame_set_label_widget (GTK_FRAME (frame15), checkbutton_sr_override);

  Sound = gtk_label_new (_("Sound"));
  gtk_widget_show (Sound);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 0), Sound);

  vbox8 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox8);
  gtk_container_add (GTK_CONTAINER (notebook), vbox8);
  gtk_container_set_border_width (GTK_CONTAINER (vbox8), 12);

  frame14 = gtk_frame_new (NULL);
  gtk_widget_show (frame14);
  gtk_box_pack_start (GTK_BOX (vbox8), frame14, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame14), GTK_SHADOW_NONE);

  alignment30 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment30);
  gtk_container_add (GTK_CONTAINER (frame14), alignment30);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment30), 0, 0, 12, 0);

  vbox48 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox48);
  gtk_container_add (GTK_CONTAINER (alignment30), vbox48);

  hbox10 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox10);
  gtk_box_pack_start (GTK_BOX (vbox48), hbox10, TRUE, TRUE, 0);

  label8 = gtk_label_new (_("Source mode:"));
  gtk_widget_show (label8);
  gtk_box_pack_start (GTK_BOX (hbox10), label8, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label8), 0, 0.5);

  pref_replaygain_source_mode = gtk_combo_box_text_new ();
  gtk_widget_show (pref_replaygain_source_mode);
  gtk_box_pack_start (GTK_BOX (hbox10), pref_replaygain_source_mode, TRUE, TRUE, 0);
  gtk_widget_set_size_request (pref_replaygain_source_mode, 337, -1);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_source_mode), _("By playback order"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_source_mode), _("Track"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_source_mode), _("Album"));

  hbox139 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox139);
  gtk_box_pack_start (GTK_BOX (vbox48), hbox139, TRUE, TRUE, 0);

  label163 = gtk_label_new (_("Processing:"));
  gtk_widget_show (label163);
  gtk_box_pack_start (GTK_BOX (hbox139), label163, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label163), 0, 0.5);

  pref_replaygain_processing = gtk_combo_box_text_new ();
  gtk_widget_show (pref_replaygain_processing);
  gtk_box_pack_start (GTK_BOX (hbox139), pref_replaygain_processing, TRUE, TRUE, 0);
  gtk_widget_set_size_request (pref_replaygain_processing, 337, -1);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_processing), _("None"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_processing), _("Apply gain"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_processing), _("Apply gain and prevent clipping according to peak"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_replaygain_processing), _("Only prevent clipping"));

  table1 = gtk_table_new (2, 2, FALSE);
  gtk_widget_show (table1);
  gtk_box_pack_start (GTK_BOX (vbox48), table1, TRUE, TRUE, 0);
  gtk_table_set_col_spacings (GTK_TABLE (table1), 8);

  hbox100 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox100);
  gtk_table_attach (GTK_TABLE (table1), hbox100, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

  label125 = gtk_label_new (_("-12 dB"));
  gtk_widget_show (label125);
  gtk_box_pack_start (GTK_BOX (hbox100), label125, FALSE, FALSE, 0);

  replaygain_preamp = gtk_hscale_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, -12, 12, 0, 0, 0)));
  gtk_widget_show (replaygain_preamp);
  gtk_box_pack_start (GTK_BOX (hbox100), replaygain_preamp, TRUE, TRUE, 0);
  gtk_scale_set_value_pos (GTK_SCALE (replaygain_preamp), GTK_POS_BOTTOM);
  gtk_scale_set_digits (GTK_SCALE (replaygain_preamp), 0);

  label126 = gtk_label_new (_("+12 dB"));
  gtk_widget_show (label126);
  gtk_box_pack_start (GTK_BOX (hbox100), label126, FALSE, FALSE, 0);

  label124 = gtk_label_new (_("Preamp with RG info:"));
  gtk_widget_show (label124);
  gtk_table_attach (GTK_TABLE (table1), label124, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label124), 0, 0.5);

  label124a = gtk_label_new (_("Preamp without RG info:"));
  gtk_widget_show (label124a);
  gtk_table_attach (GTK_TABLE (table1), label124a, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label124a), 0, 0.5);

  hbox100a = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox100a);
  gtk_table_attach (GTK_TABLE (table1), hbox100a, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

  label125a = gtk_label_new (_("-12 dB"));
  gtk_widget_show (label125a);
  gtk_box_pack_start (GTK_BOX (hbox100a), label125a, FALSE, FALSE, 0);

  global_preamp = gtk_hscale_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, -12, 12, 0, 0, 0)));
  gtk_widget_show (global_preamp);
  gtk_box_pack_start (GTK_BOX (hbox100a), global_preamp, TRUE, TRUE, 0);
  gtk_scale_set_value_pos (GTK_SCALE (global_preamp), GTK_POS_BOTTOM);
  gtk_scale_set_digits (GTK_SCALE (global_preamp), 0);

  label126a = gtk_label_new (_("+12 dB"));
  gtk_widget_show (label126a);
  gtk_box_pack_start (GTK_BOX (hbox100a), label126a, FALSE, FALSE, 0);

  label162 = gtk_label_new (_("<b>ReplayGain</b>"));
  gtk_widget_show (label162);
  gtk_frame_set_label_widget (GTK_FRAME (frame14), label162);
  gtk_label_set_use_markup (GTK_LABEL (label162), TRUE);

  hbox66 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox66);
  gtk_box_pack_start (GTK_BOX (vbox8), hbox66, FALSE, FALSE, 0);

  cli_add_to_playlist = gtk_check_button_new_with_mnemonic (_("Add files from command line (or file manager) to this playlist:"));
  gtk_widget_show (cli_add_to_playlist);
  gtk_box_pack_start (GTK_BOX (hbox66), cli_add_to_playlist, FALSE, FALSE, 0);

  cli_playlist_name = gtk_entry_new ();
  gtk_widget_show (cli_playlist_name);
  gtk_box_pack_start (GTK_BOX (hbox66), cli_playlist_name, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (cli_playlist_name), 9679);

  resume_last_session = gtk_check_button_new_with_mnemonic (_("Resume previous session on startup"));
  gtk_widget_show (resume_last_session);
  gtk_box_pack_start (GTK_BOX (vbox8), resume_last_session, FALSE, FALSE, 0);

  resume_always_paused = gtk_check_button_new_with_mnemonic (_("Always resume session in paused state"));
  gtk_widget_show (resume_always_paused);
  gtk_box_pack_start (GTK_BOX (vbox8), resume_always_paused, FALSE, FALSE, 0);

  ignore_archives = gtk_check_button_new_with_mnemonic (_("Don't add from archives when adding folders"));
  gtk_widget_show (ignore_archives);
  gtk_box_pack_start (GTK_BOX (vbox8), ignore_archives, FALSE, FALSE, 0);

  reset_autostop = gtk_check_button_new_with_mnemonic (_("\"Stop after current track\" option will switch off after triggering"));
  gtk_widget_show (reset_autostop);
  gtk_box_pack_start (GTK_BOX (vbox8), reset_autostop, FALSE, FALSE, 0);

  reset_autostopqueue = gtk_check_button_new_with_mnemonic (_("\"Stop after queue\" option will switch off after triggering"));
  gtk_widget_show (reset_autostopqueue);
  gtk_box_pack_start (GTK_BOX (vbox8), reset_autostopqueue, FALSE, FALSE, 0);

  reset_autostopalbum = gtk_check_button_new_with_mnemonic (_("\"Stop after current album\" option will switch off after triggering"));
  gtk_widget_show (reset_autostopalbum);
  gtk_box_pack_start (GTK_BOX (vbox8), reset_autostopalbum, FALSE, FALSE, 0);

  label39 = gtk_label_new (_("Playback"));
  gtk_widget_show (label39);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1), label39);

  vbox29 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox29);
  gtk_container_add (GTK_CONTAINER (notebook), vbox29);
  gtk_container_set_border_width (GTK_CONTAINER (vbox29), 12);

  prefwin_dsp_vbox = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (prefwin_dsp_vbox);
  gtk_box_pack_start (GTK_BOX (vbox29), prefwin_dsp_vbox, TRUE, TRUE, 0);

  label110 = gtk_label_new (_("DSP"));
  gtk_widget_show (label110);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 2), label110);

  notebook5 = gtk_notebook_new ();
  gtk_widget_show (notebook5);
  gtk_container_add (GTK_CONTAINER (notebook), notebook5);
  gtk_container_set_border_width (GTK_CONTAINER (notebook5), 12);

  vbox9 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox9);
  gtk_container_add (GTK_CONTAINER (notebook5), vbox9);
  gtk_container_set_border_width (GTK_CONTAINER (vbox9), 12);

  minimize_on_startup = gtk_check_button_new_with_mnemonic (_("Minimize to tray on startup"));
  gtk_widget_show (minimize_on_startup);
  gtk_box_pack_start (GTK_BOX (vbox9), minimize_on_startup, FALSE, FALSE, 0);

  pref_close_send_to_tray = gtk_check_button_new_with_mnemonic (_("Close minimizes to tray"));
  gtk_widget_show (pref_close_send_to_tray);
  gtk_box_pack_start (GTK_BOX (vbox9), pref_close_send_to_tray, FALSE, FALSE, 0);

  hide_tray_icon = gtk_check_button_new_with_mnemonic (_("Hide system tray icon"));
  gtk_widget_show (hide_tray_icon);
  gtk_box_pack_start (GTK_BOX (vbox9), hide_tray_icon, FALSE, FALSE, 0);

  move_to_trash = gtk_check_button_new_with_mnemonic (_("Delete from disk should use Trash"));
  gtk_widget_show (move_to_trash);
  gtk_box_pack_start (GTK_BOX (vbox9), move_to_trash, FALSE, FALSE, 0);

  enable_shift_jis_recoding = gtk_check_button_new_with_mnemonic (_("Enable Japanese SHIFT-JIS detection and recoding"));
  gtk_widget_show (enable_shift_jis_recoding);
  gtk_box_pack_start (GTK_BOX (vbox9), enable_shift_jis_recoding, FALSE, FALSE, 0);

  enable_cp1251_recoding = gtk_check_button_new_with_mnemonic (_("Enable cyrillic CP1251 detection and recoding"));
  gtk_widget_show (enable_cp1251_recoding);
  gtk_box_pack_start (GTK_BOX (vbox9), enable_cp1251_recoding, FALSE, FALSE, 0);

  enable_cp936_recoding = gtk_check_button_new_with_mnemonic (_("Enable Chinese CP936 detection and recoding"));
  gtk_widget_show (enable_cp936_recoding);
  gtk_box_pack_start (GTK_BOX (vbox9), enable_cp936_recoding, FALSE, FALSE, 0);

  hbox102 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox102);
  gtk_box_pack_start (GTK_BOX (vbox9), hbox102, FALSE, FALSE, 0);

  label129 = gtk_label_new (_("Interface refresh rate (times per second):"));
  gtk_widget_show (label129);
  gtk_box_pack_start (GTK_BOX (hbox102), label129, FALSE, FALSE, 0);

  gui_fps = gtk_hscale_new (GTK_ADJUSTMENT (gtk_adjustment_new (10, 1, 30, 0, 0, 0)));
  gtk_widget_show (gui_fps);
  gtk_box_pack_start (GTK_BOX (hbox102), gui_fps, TRUE, TRUE, 0);
  gtk_scale_set_value_pos (GTK_SCALE (gui_fps), GTK_POS_RIGHT);
  gtk_scale_set_digits (GTK_SCALE (gui_fps), 0);

  hbox64 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox64);
  gtk_box_pack_start (GTK_BOX (vbox9), hbox64, FALSE, FALSE, 0);

  label101 = gtk_label_new (_("Titlebar text while playing:"));
  gtk_widget_show (label101);
  gtk_box_pack_start (GTK_BOX (hbox64), label101, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label101), 0, 0.5);

  titlebar_format_playing = gtk_entry_new ();
  gtk_widget_show (titlebar_format_playing);
  gtk_box_pack_start (GTK_BOX (hbox64), titlebar_format_playing, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (titlebar_format_playing), 8226);

  hbox65 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox65);
  gtk_box_pack_start (GTK_BOX (vbox9), hbox65, FALSE, FALSE, 0);

  label102 = gtk_label_new (_("Titlebar text while stopped:"));
  gtk_widget_show (label102);
  gtk_box_pack_start (GTK_BOX (hbox65), label102, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label102), 0, 0.5);

  titlebar_format_stopped = gtk_entry_new ();
  gtk_widget_show (titlebar_format_stopped);
  gtk_box_pack_start (GTK_BOX (hbox65), titlebar_format_stopped, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (titlebar_format_stopped), 8226);

  hbox101 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox101);
  gtk_box_pack_start (GTK_BOX (vbox9), hbox101, FALSE, FALSE, 0);

  label128 = gtk_label_new (_("GUI Plugin (changing requires restart):"));
  gtk_widget_show (label128);
  gtk_box_pack_start (GTK_BOX (hbox101), label128, FALSE, FALSE, 0);

  gui_plugin = gtk_combo_box_text_new ();
  gtk_widget_show (gui_plugin);
  gtk_box_pack_start (GTK_BOX (hbox101), gui_plugin, TRUE, TRUE, 0);

  display_seltime = gtk_check_button_new_with_mnemonic (_("Display selection playback time in status bar"));
  gtk_widget_show (display_seltime);
  gtk_box_pack_start (GTK_BOX (vbox9), display_seltime, FALSE, FALSE, 0);

  label147 = gtk_label_new (_("Player"));
  gtk_widget_show (label147);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook5), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook5), 0), label147);

  vbox44 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox44);
  gtk_container_add (GTK_CONTAINER (notebook5), vbox44);
  gtk_container_set_border_width (GTK_CONTAINER (vbox44), 12);

  mmb_delete_playlist = gtk_check_button_new_with_mnemonic (_("Close playlists with middle mouse button"));
  gtk_widget_show (mmb_delete_playlist);
  gtk_box_pack_start (GTK_BOX (vbox44), mmb_delete_playlist, FALSE, FALSE, 0);

  hide_delete_from_disk = gtk_check_button_new_with_mnemonic (_("Hide \"Delete from Disk\" context menu item"));
  gtk_widget_show (hide_delete_from_disk);
  gtk_box_pack_start (GTK_BOX (vbox44), hide_delete_from_disk, FALSE, FALSE, 0);

  skip_deleted_songs = gtk_check_button_new_with_mnemonic (_("Switch to the next song when the currently played one is removed from disk"));
  gtk_widget_show (skip_deleted_songs);
  gtk_box_pack_start (GTK_BOX (vbox44), skip_deleted_songs, FALSE, FALSE, 0);

  auto_name_playlist_from_folder = gtk_check_button_new_with_mnemonic (_("Auto-name playlists when adding a single folder"));
  gtk_widget_show (auto_name_playlist_from_folder);
  gtk_box_pack_start (GTK_BOX (vbox44), auto_name_playlist_from_folder, FALSE, FALSE, 0);

  auto_size_columns = gtk_check_button_new_with_mnemonic (_("Auto-resize columns to fit the window"));
  gtk_widget_show (auto_size_columns);
  gtk_box_pack_start (GTK_BOX (vbox44), auto_size_columns, FALSE, FALSE, 0);

  hbox141 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox141);
  gtk_box_pack_start (GTK_BOX (vbox44), hbox141, FALSE, FALSE, 0);

  label164 = gtk_label_new (_("Group spacing:"));
  gtk_widget_show (label164);
  gtk_box_pack_start (GTK_BOX (hbox141), label164, FALSE, FALSE, 0);

  listview_group_spacing_adj = G_OBJECT(gtk_adjustment_new (0, 0, 1000, 1, 10, 0));
  listview_group_spacing = gtk_spin_button_new (GTK_ADJUSTMENT (listview_group_spacing_adj), 1, 0);
  gtk_widget_show (listview_group_spacing);
  gtk_box_pack_start (GTK_BOX (hbox141), listview_group_spacing, FALSE, TRUE, 0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (listview_group_spacing), TRUE);

  label165 = gtk_label_new (_("pixels"));
  gtk_widget_show (label165);
  gtk_box_pack_start (GTK_BOX (hbox141), label165, FALSE, FALSE, 0);

  label148 = gtk_label_new (_("Playlist"));
  gtk_widget_show (label148);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook5), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook5), 1), label148);

  label2 = gtk_label_new (_("GUI/Misc"));
  gtk_widget_show (label2);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3), label2);

  notebook4 = gtk_notebook_new ();
  gtk_widget_show (notebook4);
  gtk_container_add (GTK_CONTAINER (notebook), notebook4);
  gtk_container_set_border_width (GTK_CONTAINER (notebook4), 12);
  gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook4), TRUE);

  vbox21 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox21);
  gtk_container_add (GTK_CONTAINER (notebook4), vbox21);
  gtk_container_set_border_width (GTK_CONTAINER (vbox21), 12);

  override_bar_colors = gtk_check_button_new_with_mnemonic (_("Override"));
  gtk_widget_show (override_bar_colors);
  gtk_box_pack_start (GTK_BOX (vbox21), override_bar_colors, FALSE, FALSE, 0);

  bar_colors_group = gtk_table_new (2, 2, TRUE);
  gtk_widget_show (bar_colors_group);
  gtk_box_pack_start (GTK_BOX (vbox21), bar_colors_group, TRUE, TRUE, 0);
  gtk_table_set_col_spacings (GTK_TABLE (bar_colors_group), 8);

  label43 = gtk_label_new (_("Bars"));
  gtk_widget_show (label43);
  gtk_table_attach (GTK_TABLE (bar_colors_group), label43, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label43), 0, 0.5);

  label47 = gtk_label_new (_("Background (EQ)"));
  gtk_widget_show (label47);
  gtk_table_attach (GTK_TABLE (bar_colors_group), label47, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label47), 0, 0.5);

  bar_foreground = gtk_color_button_new ();
  gtk_widget_show (bar_foreground);
  gtk_table_attach (GTK_TABLE (bar_colors_group), bar_foreground, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND),
                    (GtkAttachOptions) (0), 0, 0);

  bar_background = gtk_color_button_new ();
  gtk_widget_show (bar_background);
  gtk_table_attach (GTK_TABLE (bar_colors_group), bar_background, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND),
                    (GtkAttachOptions) (0), 0, 0);

  label73 = gtk_label_new (_("Seekbar/Volumebar/EQ"));
  gtk_widget_show (label73);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook4), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook4), 0), label73);

  vbox22 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox22);
  gtk_container_add (GTK_CONTAINER (notebook4), vbox22);
  gtk_container_set_border_width (GTK_CONTAINER (vbox22), 12);

  override_tabstrip_colors = gtk_check_button_new_with_mnemonic (_("Override"));
  gtk_widget_show (override_tabstrip_colors);
  gtk_box_pack_start (GTK_BOX (vbox22), override_tabstrip_colors, FALSE, FALSE, 0);

  tabstrip_colors_group = gtk_table_new (7, 3, FALSE);
  gtk_widget_show (tabstrip_colors_group);
  gtk_box_pack_start (GTK_BOX (vbox22), tabstrip_colors_group, TRUE, TRUE, 0);
  gtk_table_set_row_spacings (GTK_TABLE (tabstrip_colors_group), 4);
  gtk_table_set_col_spacings (GTK_TABLE (tabstrip_colors_group), 8);

  tabstrip_base = gtk_color_button_new ();
  gtk_widget_show (tabstrip_base);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_base, 1, 2, 6, 7,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label76 = gtk_label_new (_("Base:"));
  gtk_widget_show (label76);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label76, 0, 1, 6, 7,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label76), 0, 0.5);

  tabstrip_dark = gtk_color_button_new ();
  gtk_widget_show (tabstrip_dark);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_dark, 1, 2, 5, 6,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label44 = gtk_label_new (_("Dark:"));
  gtk_widget_show (label44);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label44, 0, 1, 5, 6,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label44), 0, 0.5);

  tabstrip_light = gtk_color_button_new ();
  gtk_widget_show (tabstrip_light);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_light, 1, 2, 4, 5,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label46 = gtk_label_new (_("Light:"));
  gtk_widget_show (label46);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label46, 0, 1, 4, 5,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label46), 0, 0.5);

  tabstrip_mid = gtk_color_button_new ();
  gtk_widget_show (tabstrip_mid);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_mid, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label45 = gtk_label_new (_("Middle:"));
  gtk_widget_show (label45);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label45, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label45), 0, 0.5);

  label127 = gtk_label_new (_("Text:"));
  gtk_widget_show (label127);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label127, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label127), 0, 0.5);

  tabstrip_text = gtk_color_button_new ();
  gtk_widget_show (tabstrip_text);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_text, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  tabstrip_selected_text = gtk_color_button_new ();
  gtk_widget_show (tabstrip_selected_text);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_selected_text, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  hbox136 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox136);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), hbox136, 2, 3, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

  tabstrip_playing_bold = gtk_check_button_new_with_mnemonic (_("Bold"));
  gtk_widget_show (tabstrip_playing_bold);
  gtk_box_pack_start (GTK_BOX (hbox136), tabstrip_playing_bold, FALSE, FALSE, 0);

  tabstrip_playing_italic = gtk_check_button_new_with_mnemonic (_("Italic"));
  gtk_widget_show (tabstrip_playing_italic);
  gtk_box_pack_start (GTK_BOX (hbox136), tabstrip_playing_italic, FALSE, FALSE, 0);

  hbox137 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox137);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), hbox137, 2, 3, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

  tabstrip_selected_bold = gtk_check_button_new_with_mnemonic (_("Bold"));
  gtk_widget_show (tabstrip_selected_bold);
  gtk_box_pack_start (GTK_BOX (hbox137), tabstrip_selected_bold, FALSE, FALSE, 0);

  tabstrip_selected_italic = gtk_check_button_new_with_mnemonic (_("Italic"));
  gtk_widget_show (tabstrip_selected_italic);
  gtk_box_pack_start (GTK_BOX (hbox137), tabstrip_selected_italic, FALSE, FALSE, 0);

  tabstrip_text_font = gtk_font_button_new ();
  gtk_widget_show (tabstrip_text_font);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_text_font, 2, 3, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0);

  label152 = gtk_label_new (_("Playing text:"));
  gtk_widget_show (label152);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label152, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label152), 0, 0.5);

  label153 = gtk_label_new (_("Selected text:"));
  gtk_widget_show (label153);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), label153, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label153), 0, 0.5);

  tabstrip_playing_text = gtk_color_button_new ();
  gtk_widget_show (tabstrip_playing_text);
  gtk_table_attach (GTK_TABLE (tabstrip_colors_group), tabstrip_playing_text, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label74 = gtk_label_new (_("Tab strip"));
  gtk_widget_show (label74);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook4), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook4), 1), label74);

  vbox23 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox23);
  gtk_container_add (GTK_CONTAINER (notebook4), vbox23);
  gtk_container_set_border_width (GTK_CONTAINER (vbox23), 12);

  frame11 = gtk_frame_new (NULL);
  gtk_widget_show (frame11);
  gtk_box_pack_start (GTK_BOX (vbox23), frame11, TRUE, TRUE, 0);

  listview_colors_group = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (listview_colors_group);
  gtk_container_add (GTK_CONTAINER (frame11), listview_colors_group);
  gtk_container_set_border_width (GTK_CONTAINER (listview_colors_group), 8);

  frame12 = gtk_frame_new (NULL);
  gtk_widget_show (frame12);
  gtk_box_pack_start (GTK_BOX (listview_colors_group), frame12, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame12), GTK_SHADOW_NONE);

  alignment28 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment28);
  gtk_container_add (GTK_CONTAINER (frame12), alignment28);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment28), 12, 12, 12, 12);

  table3 = gtk_table_new (5, 3, FALSE);
  gtk_widget_show (table3);
  gtk_container_add (GTK_CONTAINER (alignment28), table3);
  gtk_table_set_row_spacings (GTK_TABLE (table3), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table3), 8);

  label77 = gtk_label_new (_("Normal track(s):"));
  gtk_widget_show (label77);
  gtk_table_attach (GTK_TABLE (table3), label77, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label77), 0, 0.5);

  listview_text = gtk_color_button_new ();
  gtk_widget_show (listview_text);
  gtk_table_attach (GTK_TABLE (table3), listview_text, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label150 = gtk_label_new (_("Playing track:"));
  gtk_widget_show (label150);
  gtk_table_attach (GTK_TABLE (table3), label150, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label150), 0, 0.5);

  listview_playing_text = gtk_color_button_new ();
  gtk_widget_show (listview_playing_text);
  gtk_table_attach (GTK_TABLE (table3), listview_playing_text, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label61 = gtk_label_new (_("Selected track(s):"));
  gtk_widget_show (label61);
  gtk_table_attach (GTK_TABLE (table3), label61, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label61), 0, 0.5);

  listview_selected_text = gtk_color_button_new ();
  gtk_widget_show (listview_selected_text);
  gtk_table_attach (GTK_TABLE (table3), listview_selected_text, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label154 = gtk_label_new (_("Columns:"));
  gtk_widget_show (label154);
  gtk_table_attach (GTK_TABLE (table3), label154, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label154), 0, 0.5);

  listview_column_text = gtk_color_button_new ();
  gtk_widget_show (listview_column_text);
  gtk_table_attach (GTK_TABLE (table3), listview_column_text, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_column_text_font = gtk_font_button_new ();
  gtk_widget_show (listview_column_text_font);
  gtk_table_attach (GTK_TABLE (table3), listview_column_text_font, 2, 3, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label149 = gtk_label_new (_("Group header:"));
  gtk_widget_show (label149);
  gtk_table_attach (GTK_TABLE (table3), label149, 0, 1, 4, 5,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label149), 0, 0.5);

  listview_group_text = gtk_color_button_new ();
  gtk_widget_show (listview_group_text);
  gtk_table_attach (GTK_TABLE (table3), listview_group_text, 1, 2, 4, 5,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_group_text_font = gtk_font_button_new ();
  gtk_widget_show (listview_group_text_font);
  gtk_table_attach (GTK_TABLE (table3), listview_group_text_font, 2, 3, 4, 5,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_text_font = gtk_font_button_new ();
  gtk_widget_show (listview_text_font);
  gtk_table_attach (GTK_TABLE (table3), listview_text_font, 2, 3, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label158 = gtk_label_new (_("<b>Foreground:</b>"));
  gtk_widget_show (label158);
  gtk_frame_set_label_widget (GTK_FRAME (frame12), label158);
  gtk_label_set_use_markup (GTK_LABEL (label158), TRUE);

  vseparator1 = gtk_vseparator_new ();
  gtk_widget_show (vseparator1);
  gtk_box_pack_start (GTK_BOX (listview_colors_group), vseparator1, FALSE, TRUE, 5);

  frame13 = gtk_frame_new (NULL);
  gtk_widget_show (frame13);
  gtk_box_pack_start (GTK_BOX (listview_colors_group), frame13, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame13), GTK_SHADOW_NONE);

  alignment29 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment29);
  gtk_container_add (GTK_CONTAINER (frame13), alignment29);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment29), 11, 12, 12, 12);

  table4 = gtk_table_new (4, 2, FALSE);
  gtk_widget_show (table4);
  gtk_container_add (GTK_CONTAINER (alignment29), table4);
  gtk_table_set_row_spacings (GTK_TABLE (table4), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table4), 8);

  label58 = gtk_label_new (_("Even:"));
  gtk_widget_show (label58);
  gtk_table_attach (GTK_TABLE (table4), label58, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label58), 0, 0.5);

  label59 = gtk_label_new (_("Odd:"));
  gtk_widget_show (label59);
  gtk_table_attach (GTK_TABLE (table4), label59, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label59), 0, 0.5);

  listview_odd_row = gtk_color_button_new ();
  gtk_widget_show (listview_odd_row);
  gtk_table_attach (GTK_TABLE (table4), listview_odd_row, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label78 = gtk_label_new (_("Selected:"));
  gtk_widget_show (label78);
  gtk_table_attach (GTK_TABLE (table4), label78, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label78), 0, 0.5);

  listview_selected_row = gtk_color_button_new ();
  gtk_widget_show (listview_selected_row);
  gtk_table_attach (GTK_TABLE (table4), listview_selected_row, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label62 = gtk_label_new (_("Cursor:"));
  gtk_widget_show (label62);
  gtk_table_attach (GTK_TABLE (table4), label62, 0, 1, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label62), 0, 0.5);

  listview_cursor = gtk_color_button_new ();
  gtk_widget_show (listview_cursor);
  gtk_table_attach (GTK_TABLE (table4), listview_cursor, 1, 2, 3, 4,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_even_row = gtk_color_button_new ();
  gtk_widget_show (listview_even_row);
  gtk_table_attach (GTK_TABLE (table4), listview_even_row, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label159 = gtk_label_new (_("<b>Background:</b>"));
  gtk_widget_show (label159);
  gtk_frame_set_label_widget (GTK_FRAME (frame13), label159);
  gtk_label_set_use_markup (GTK_LABEL (label159), TRUE);

  override_listview_colors = gtk_check_button_new_with_mnemonic (_("Override (loses GTK treeview theming, but speeds up rendering)"));
  gtk_widget_show (override_listview_colors);
  gtk_frame_set_label_widget (GTK_FRAME (frame11), override_listview_colors);

  table2 = gtk_table_new (2, 3, FALSE);
  gtk_widget_show (table2);
  gtk_box_pack_start (GTK_BOX (vbox23), table2, TRUE, TRUE, 0);
  gtk_table_set_row_spacings (GTK_TABLE (table2), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table2), 8);

  label155 = gtk_label_new (_("Playing track:"));
  gtk_widget_show (label155);
  gtk_table_attach (GTK_TABLE (table2), label155, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label155), 0, 0.5);

  label156 = gtk_label_new (_("Selected track(s):"));
  gtk_widget_show (label156);
  gtk_table_attach (GTK_TABLE (table2), label156, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label156), 0, 0.5);

  listview_playing_text_bold = gtk_check_button_new_with_mnemonic (_("Bold"));
  gtk_widget_show (listview_playing_text_bold);
  gtk_table_attach (GTK_TABLE (table2), listview_playing_text_bold, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_selected_text_bold = gtk_check_button_new_with_mnemonic (_("Bold"));
  gtk_widget_show (listview_selected_text_bold);
  gtk_table_attach (GTK_TABLE (table2), listview_selected_text_bold, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_playing_text_italic = gtk_check_button_new_with_mnemonic (_("Italic"));
  gtk_widget_show (listview_playing_text_italic);
  gtk_table_attach (GTK_TABLE (table2), listview_playing_text_italic, 2, 3, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  listview_selected_text_italic = gtk_check_button_new_with_mnemonic (_("Italic"));
  gtk_widget_show (listview_selected_text_italic);
  gtk_table_attach (GTK_TABLE (table2), listview_selected_text_italic, 2, 3, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label75 = gtk_label_new (_("Playlist"));
  gtk_widget_show (label75);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook4), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook4), 2), label75);

  table5 = gtk_table_new (2, 2, FALSE);
  gtk_widget_show (table5);
  gtk_container_add (GTK_CONTAINER (notebook4), table5);
  gtk_container_set_border_width (GTK_CONTAINER (table5), 12);
  gtk_table_set_row_spacings (GTK_TABLE (table5), 8);
  gtk_table_set_col_spacings (GTK_TABLE (table5), 8);

  use_visualization_base_color = gtk_check_button_new_with_mnemonic (_("Custom visualization base color"));
  gtk_widget_show (use_visualization_base_color);
  gtk_table_attach (GTK_TABLE (table5), use_visualization_base_color, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  visualization_custom_color_button = gtk_color_button_new ();
  gtk_widget_show (visualization_custom_color_button);
  gtk_table_attach (GTK_TABLE (table5), visualization_custom_color_button, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  use_visualization_background_color = gtk_check_button_new_with_mnemonic (_("Custom visualization background color"));
  gtk_widget_show (use_visualization_background_color);
  gtk_table_attach (GTK_TABLE (table5), use_visualization_background_color, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  visualization_custom_background_color_button = gtk_color_button_new ();
  gtk_widget_show (visualization_custom_background_color_button);
  gtk_table_attach (GTK_TABLE (table5), visualization_custom_background_color_button, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label175 = gtk_label_new (_("Visualizations"));
  gtk_widget_show (label175);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook4), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook4), 3), label175);

  label100 = gtk_label_new (_("Appearance"));
  gtk_widget_show (label100);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 4), label100);

  vbox55 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox55);
  gtk_container_add (GTK_CONTAINER (notebook), vbox55);
  gtk_container_set_border_width (GTK_CONTAINER (vbox55), 12);

  toggle_medialib_on = gtk_check_button_new_with_mnemonic (_("Enable media library"));
  gtk_widget_show (toggle_medialib_on);
  gtk_box_pack_start (GTK_BOX (vbox55), toggle_medialib_on, FALSE, FALSE, 0);

  hseparator7 = gtk_hseparator_new ();
  gtk_widget_show (hseparator7);
  gtk_box_pack_start (GTK_BOX (vbox55), hseparator7, FALSE, FALSE, 0);

  label174 = gtk_label_new (_("Add / remove music folders"));
  gtk_widget_show (label174);
  gtk_box_pack_start (GTK_BOX (vbox55), label174, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label174), 0, 0.5);

  vbox56 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox56);
  gtk_box_pack_start (GTK_BOX (vbox55), vbox56, TRUE, TRUE, 0);

  scrolledwindow17 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow17);
  gtk_box_pack_start (GTK_BOX (vbox56), scrolledwindow17, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow17), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow17), GTK_SHADOW_IN);

  treeview_medialib_folders = gtk_tree_view_new ();
  gtk_widget_show (treeview_medialib_folders);
  gtk_container_add (GTK_CONTAINER (scrolledwindow17), treeview_medialib_folders);
  gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (treeview_medialib_folders), FALSE);

  hbox150 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox150);
  gtk_box_pack_start (GTK_BOX (vbox56), hbox150, FALSE, TRUE, 0);

  button_medialib_add_folder = gtk_button_new ();
  gtk_widget_show (button_medialib_add_folder);
  gtk_box_pack_start (GTK_BOX (hbox150), button_medialib_add_folder, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_medialib_add_folder, 38, -1);

  image714 = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image714);
  gtk_container_add (GTK_CONTAINER (button_medialib_add_folder), image714);

  button_medialib_remove_folder = gtk_button_new ();
  gtk_widget_show (button_medialib_remove_folder);
  gtk_box_pack_start (GTK_BOX (hbox150), button_medialib_remove_folder, FALSE, FALSE, 0);
  gtk_widget_set_size_request (button_medialib_remove_folder, 38, -1);

  image715 = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image715);
  gtk_container_add (GTK_CONTAINER (button_medialib_remove_folder), image715);

  label173 = gtk_label_new (_("Media Library"));
  gtk_widget_show (label173);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 5), label173);

  vbox11 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox11);
  gtk_container_add (GTK_CONTAINER (notebook), vbox11);
  gtk_container_set_border_width (GTK_CONTAINER (vbox11), 12);

  pref_network_enableproxy = gtk_check_button_new_with_mnemonic (_("Enable Proxy Server"));
  gtk_widget_show (pref_network_enableproxy);
  gtk_box_pack_start (GTK_BOX (vbox11), pref_network_enableproxy, FALSE, FALSE, 0);

  hbox13 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox13);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox13, FALSE, FALSE, 0);

  label18 = gtk_label_new (_("Proxy Server Address:"));
  gtk_widget_show (label18);
  gtk_box_pack_start (GTK_BOX (hbox13), label18, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label18), 0, 0.5);

  pref_network_proxyaddress = gtk_entry_new ();
  gtk_widget_show (pref_network_proxyaddress);
  gtk_box_pack_start (GTK_BOX (hbox13), pref_network_proxyaddress, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (pref_network_proxyaddress), 8226);

  hbox14 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox14);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox14, FALSE, FALSE, 0);

  label19 = gtk_label_new (_("Proxy Server Port:"));
  gtk_widget_show (label19);
  gtk_box_pack_start (GTK_BOX (hbox14), label19, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label19), 0, 0.5);

  pref_network_proxyport = gtk_entry_new ();
  gtk_widget_show (pref_network_proxyport);
  gtk_box_pack_start (GTK_BOX (hbox14), pref_network_proxyport, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (pref_network_proxyport), 8226);

  hbox15 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox15);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox15, FALSE, FALSE, 0);

  label20 = gtk_label_new (_("Proxy Type:"));
  gtk_widget_show (label20);
  gtk_box_pack_start (GTK_BOX (hbox15), label20, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label20), 0, 0.5);

  pref_network_proxytype = gtk_combo_box_text_new ();
  gtk_widget_show (pref_network_proxytype);
  gtk_box_pack_start (GTK_BOX (hbox15), pref_network_proxytype, TRUE, TRUE, 0);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "HTTP");
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "HTTP_1_0");
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "SOCKS4");
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "SOCKS5");
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "SOCKS4A");
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (pref_network_proxytype), "SOCKS5_HOSTNAME");

  hbox61 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox61);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox61, FALSE, FALSE, 0);

  label97 = gtk_label_new (_("Proxy Username:"));
  gtk_widget_show (label97);
  gtk_box_pack_start (GTK_BOX (hbox61), label97, FALSE, FALSE, 0);

  proxyuser = gtk_entry_new ();
  gtk_widget_show (proxyuser);
  gtk_box_pack_start (GTK_BOX (hbox61), proxyuser, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (proxyuser), 9679);

  hbox62 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox62);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox62, FALSE, FALSE, 0);

  label98 = gtk_label_new (_("Proxy Password:"));
  gtk_widget_show (label98);
  gtk_box_pack_start (GTK_BOX (hbox62), label98, FALSE, FALSE, 0);

  proxypassword = gtk_entry_new ();
  gtk_widget_show (proxypassword);
  gtk_box_pack_start (GTK_BOX (hbox62), proxypassword, TRUE, TRUE, 0);
  gtk_entry_set_visibility (GTK_ENTRY (proxypassword), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (proxypassword), 9679);

  hseparator1 = gtk_hseparator_new ();
  gtk_widget_show (hseparator1);
  gtk_box_pack_start (GTK_BOX (vbox11), hseparator1, FALSE, TRUE, 0);

  hbox104 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox104);
  gtk_box_pack_start (GTK_BOX (vbox11), hbox104, FALSE, TRUE, 0);

  label131 = gtk_label_new (_("HTTP User Agent:"));
  gtk_widget_show (label131);
  gtk_box_pack_start (GTK_BOX (hbox104), label131, FALSE, FALSE, 0);

  useragent = gtk_entry_new ();
  gtk_widget_show (useragent);
  gtk_box_pack_start (GTK_BOX (hbox104), useragent, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (useragent), 8226);

  hseparator5 = gtk_hseparator_new ();
  gtk_widget_show (hseparator5);
  gtk_box_pack_start (GTK_BOX (vbox11), hseparator5, FALSE, FALSE, 0);

  edit_content_type_mapping = gtk_button_new_with_mnemonic (_("Edit Content-Type Mapping"));
  gtk_widget_show (edit_content_type_mapping);
  gtk_box_pack_start (GTK_BOX (vbox11), edit_content_type_mapping, FALSE, FALSE, 0);

  label16 = gtk_label_new (_("Network"));
  gtk_widget_show (label16);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 6), label16);

  vbox36 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox36);
  gtk_container_add (GTK_CONTAINER (notebook), vbox36);
  gtk_container_set_border_width (GTK_CONTAINER (vbox36), 12);

  hbox119 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox119);
  gtk_box_pack_start (GTK_BOX (vbox36), hbox119, FALSE, FALSE, 0);

  label144 = gtk_label_new (_("Assigned Hotkeys:"));
  gtk_widget_show (label144);
  gtk_box_pack_start (GTK_BOX (hbox119), label144, FALSE, FALSE, 0);

  hbuttonbox3 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox3);
  gtk_box_pack_start (GTK_BOX (hbox119), hbuttonbox3, TRUE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox3), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox3), 8);

  hotkey_add = gtk_button_new_from_stock ("gtk-add");
  gtk_widget_show (hotkey_add);
  gtk_container_add (GTK_CONTAINER (hbuttonbox3), hotkey_add);
  gtk_widget_set_can_default(hotkey_add, TRUE);

  hotkey_remove = gtk_button_new_from_stock ("gtk-remove");
  gtk_widget_show (hotkey_remove);
  gtk_container_add (GTK_CONTAINER (hbuttonbox3), hotkey_remove);
  gtk_widget_set_can_default(hotkey_remove, TRUE);

  scrolledwindow9 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow9);
  gtk_box_pack_start (GTK_BOX (vbox36), scrolledwindow9, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow9), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow9), GTK_SHADOW_IN);

  hotkeys_list = gtk_tree_view_new ();
  gtk_widget_show (hotkeys_list);
  gtk_container_add (GTK_CONTAINER (scrolledwindow9), hotkeys_list);
  gtk_tree_view_set_enable_search (GTK_TREE_VIEW (hotkeys_list), FALSE);

  hbox105 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox105);
  gtk_box_pack_start (GTK_BOX (vbox36), hbox105, FALSE, FALSE, 0);

  label133 = gtk_label_new (_("Action:"));
  gtk_widget_show (label133);
  gtk_box_pack_start (GTK_BOX (hbox105), label133, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label133), TRUE);

  hotkeys_actions = gtk_button_new_with_mnemonic (_("<Not set>"));
  gtk_widget_show (hotkeys_actions);
  gtk_box_pack_start (GTK_BOX (hbox105), hotkeys_actions, FALSE, FALSE, 0);

  hbox106 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox106);
  gtk_box_pack_start (GTK_BOX (vbox36), hbox106, FALSE, FALSE, 0);

  label134 = gtk_label_new (_("Key combination:"));
  gtk_widget_show (label134);
  gtk_box_pack_start (GTK_BOX (hbox106), label134, FALSE, FALSE, 0);
  gtk_label_set_use_markup (GTK_LABEL (label134), TRUE);

  hotkeys_set_key = gtk_button_new_with_mnemonic (_("<Not set>"));
  gtk_widget_show (hotkeys_set_key);
  gtk_box_pack_start (GTK_BOX (hbox106), hotkeys_set_key, FALSE, FALSE, 0);

  hseparator6 = gtk_hseparator_new ();
  gtk_widget_show (hseparator6);
  gtk_box_pack_start (GTK_BOX (vbox36), hseparator6, FALSE, FALSE, 0);

  hbox120 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox120);
  gtk_box_pack_start (GTK_BOX (vbox36), hbox120, FALSE, FALSE, 0);

  label145 = gtk_label_new (_("Use the Apply button to save your changes,\nor the Revert button to undo your changes.\nThe changes will NOT be saved\nif you don't press Apply."));
  gtk_widget_show (label145);
  gtk_box_pack_start (GTK_BOX (hbox120), label145, FALSE, FALSE, 0);

  hbuttonbox4 = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox4);
  gtk_box_pack_start (GTK_BOX (hbox120), hbuttonbox4, TRUE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox4), GTK_BUTTONBOX_END);
  gtk_box_set_spacing (GTK_BOX (hbuttonbox4), 8);

  hotkeys_apply = gtk_button_new_from_stock ("gtk-apply");
  gtk_widget_show (hotkeys_apply);
  gtk_container_add (GTK_CONTAINER (hbuttonbox4), hotkeys_apply);
  gtk_widget_set_can_default(hotkeys_apply, TRUE);

  hotkeys_revert = gtk_button_new_from_stock ("gtk-revert-to-saved");
  gtk_widget_show (hotkeys_revert);
  gtk_container_add (GTK_CONTAINER (hbuttonbox4), hotkeys_revert);
  gtk_widget_set_can_default(hotkeys_revert, TRUE);

  hotkeys_defaults = gtk_button_new ();
  gtk_widget_show (hotkeys_defaults);
  gtk_container_add (GTK_CONTAINER (hbuttonbox4), hotkeys_defaults);
  gtk_widget_set_can_default(hotkeys_defaults, TRUE);

  alignment26 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment26);
  gtk_container_add (GTK_CONTAINER (hotkeys_defaults), alignment26);

  hbox121 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox121);
  gtk_container_add (GTK_CONTAINER (alignment26), hbox121);

  image638 = gtk_image_new_from_stock ("gtk-revert-to-saved", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image638);
  gtk_box_pack_start (GTK_BOX (hbox121), image638, FALSE, FALSE, 0);

  label146 = gtk_label_new_with_mnemonic (_("_Defaults"));
  gtk_widget_show (label146);
  gtk_box_pack_start (GTK_BOX (hbox121), label146, FALSE, FALSE, 0);

  label132 = gtk_label_new (_("Hotkeys"));
  gtk_widget_show (label132);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 7), label132);

  hpaned1 = gtk_hpaned_new ();
  gtk_widget_show (hpaned1);
  gtk_container_add (GTK_CONTAINER (notebook), hpaned1);
  gtk_container_set_border_width (GTK_CONTAINER (hpaned1), 12);

  scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow2);
  gtk_paned_pack1 (GTK_PANED (hpaned1), scrolledwindow2, FALSE, FALSE);
  gtk_widget_set_size_request (scrolledwindow2, 280, -1);
  gtk_container_set_border_width (GTK_CONTAINER (scrolledwindow2), 3);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow2), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow2), GTK_SHADOW_IN);

  pref_pluginlist = gtk_tree_view_new ();
  gtk_widget_show (pref_pluginlist);
  gtk_container_add (GTK_CONTAINER (scrolledwindow2), pref_pluginlist);
  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (pref_pluginlist), TRUE);

  alignment33 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment33);
  gtk_paned_pack2 (GTK_PANED (hpaned1), alignment33, TRUE, FALSE);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment33), 0, 0, 16, 0);

  vbox54 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox54);
  gtk_container_add (GTK_CONTAINER (alignment33), vbox54);

  hbox149 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox149);
  gtk_box_pack_start (GTK_BOX (vbox54), hbox149, FALSE, FALSE, 0);

  fillerlabel1 = gtk_label_new ("");
  gtk_widget_show (fillerlabel1);
  gtk_box_pack_start (GTK_BOX (hbox149), fillerlabel1, TRUE, FALSE, 0);

  plugin_tabbtn_hbtnbox = gtk_hbutton_box_new ();
  gtk_widget_show (plugin_tabbtn_hbtnbox);
  gtk_box_pack_start (GTK_BOX (hbox149), plugin_tabbtn_hbtnbox, FALSE, FALSE, 0);

  plugin_conf_tab_btn = gtk_radio_button_new_with_mnemonic (NULL, _("Configuration"));
  gtk_widget_show (plugin_conf_tab_btn);
  gtk_container_add (GTK_CONTAINER (plugin_tabbtn_hbtnbox), plugin_conf_tab_btn);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (plugin_conf_tab_btn), plugin_conf_tab_btn_group);
  plugin_conf_tab_btn_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (plugin_conf_tab_btn));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (plugin_conf_tab_btn), TRUE);
  gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (plugin_conf_tab_btn), FALSE);

  plugin_info_tab_btn = gtk_radio_button_new_with_mnemonic (NULL, _("Info"));
  gtk_widget_show (plugin_info_tab_btn);
  gtk_container_add (GTK_CONTAINER (plugin_tabbtn_hbtnbox), plugin_info_tab_btn);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (plugin_info_tab_btn), plugin_conf_tab_btn_group);
  plugin_conf_tab_btn_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (plugin_info_tab_btn));
  gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (plugin_info_tab_btn), FALSE);

  plugin_license_tab_btn = gtk_radio_button_new_with_mnemonic (NULL, _("License"));
  gtk_widget_show (plugin_license_tab_btn);
  gtk_container_add (GTK_CONTAINER (plugin_tabbtn_hbtnbox), plugin_license_tab_btn);
  gtk_radio_button_set_group (GTK_RADIO_BUTTON (plugin_license_tab_btn), plugin_conf_tab_btn_group);
  plugin_conf_tab_btn_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (plugin_license_tab_btn));
  gtk_toggle_button_set_mode (GTK_TOGGLE_BUTTON (plugin_license_tab_btn), FALSE);

  fillerlabel2 = gtk_label_new ("");
  gtk_widget_show (fillerlabel2);
  gtk_box_pack_start (GTK_BOX (hbox149), fillerlabel2, TRUE, FALSE, 0);

  plugin_notebook = gtk_notebook_new ();
  gtk_widget_show (plugin_notebook);
  gtk_box_pack_start (GTK_BOX (vbox54), plugin_notebook, TRUE, TRUE, 0);
  gtk_notebook_set_show_border (GTK_NOTEBOOK (plugin_notebook), FALSE);

  plugin_conf_tab_vbox = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (plugin_conf_tab_vbox);
  gtk_container_add (GTK_CONTAINER (plugin_notebook), plugin_conf_tab_vbox);

  plug_conf_dlg_scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (plug_conf_dlg_scrolledwindow);
  gtk_box_pack_start (GTK_BOX (plugin_conf_tab_vbox), plug_conf_dlg_scrolledwindow, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (plug_conf_dlg_scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  plug_conf_dlg_viewport = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (plug_conf_dlg_viewport);
  gtk_container_add (GTK_CONTAINER (plug_conf_dlg_scrolledwindow), plug_conf_dlg_viewport);
  gtk_viewport_set_shadow_type (GTK_VIEWPORT (plug_conf_dlg_viewport), GTK_SHADOW_NONE);

  plugin_actions_btnbox = gtk_hbutton_box_new ();
  gtk_box_pack_start (GTK_BOX (plugin_conf_tab_vbox), plugin_actions_btnbox, FALSE, TRUE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (plugin_actions_btnbox), GTK_BUTTONBOX_END);

  plugin_conf_reset_btn = gtk_button_new_with_mnemonic (_("Reset to defaults"));
  gtk_widget_show (plugin_conf_reset_btn);
  gtk_container_add (GTK_CONTAINER (plugin_actions_btnbox), plugin_conf_reset_btn);
  gtk_widget_set_can_default(plugin_conf_reset_btn, TRUE);

  label170 = gtk_label_new (_("Configuration"));
  gtk_widget_show (label170);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (plugin_notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (plugin_notebook), 0), label170);

  vbox12 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox12);
  gtk_container_add (GTK_CONTAINER (plugin_notebook), vbox12);

  hbox103 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox103);
  gtk_box_pack_start (GTK_BOX (vbox12), hbox103, FALSE, TRUE, 0);

  label130 = gtk_label_new (_("Version: "));
  gtk_widget_show (label130);
  gtk_box_pack_start (GTK_BOX (hbox103), label130, FALSE, FALSE, 0);

  plug_version = gtk_entry_new ();
  gtk_widget_show (plug_version);
  gtk_box_pack_start (GTK_BOX (hbox103), plug_version, TRUE, TRUE, 0);
  gtk_editable_set_editable (GTK_EDITABLE (plug_version), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (plug_version), 8226);

  scrolledwindow8 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow8);
  gtk_box_pack_start (GTK_BOX (vbox12), scrolledwindow8, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow8), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow8), GTK_SHADOW_IN);

  plug_description = gtk_text_view_new ();
  gtk_widget_show (plug_description);
  gtk_container_add (GTK_CONTAINER (scrolledwindow8), plug_description);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (plug_description), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (plug_description), FALSE);

  hbox20 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox20);
  gtk_box_pack_start (GTK_BOX (vbox12), hbox20, FALSE, FALSE, 0);

  weblink = create_plugin_weblink ("weblink", "", "", 0, 0);
  gtk_widget_show (weblink);
  gtk_box_pack_start (GTK_BOX (hbox20), weblink, TRUE, FALSE, 0);
  gtk_widget_set_can_focus(weblink, FALSE);
  gtk_widget_set_can_default(weblink, FALSE);

  label171 = gtk_label_new (_("Info"));
  gtk_widget_show (label171);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (plugin_notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (plugin_notebook), 1), label171);

  scrolledwindow16 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow16);
  gtk_container_add (GTK_CONTAINER (plugin_notebook), scrolledwindow16);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow16), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow16), GTK_SHADOW_IN);

  plug_license = gtk_text_view_new ();
  gtk_widget_show (plug_license);
  gtk_container_add (GTK_CONTAINER (scrolledwindow16), plug_license);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (plug_license), FALSE);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (plug_license), FALSE);

  label172 = gtk_label_new (_("License"));
  gtk_widget_show (label172);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (plugin_notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (plugin_notebook), 2), label172);

  label3 = gtk_label_new (_("Plugins"));
  gtk_widget_show (label3);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 8), label3);
  gtk_misc_set_alignment (GTK_MISC (label3), 0.48, 0.5);

  dialog_action_area2 = gtk_dialog_get_action_area (GTK_DIALOG (prefwin));
  gtk_widget_show (dialog_action_area2);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area2), GTK_BUTTONBOX_END);

  closebutton1 = gtk_button_new_from_stock ("gtk-close");
  gtk_widget_show (closebutton1);
  gtk_dialog_add_action_widget (GTK_DIALOG (prefwin), closebutton1, GTK_RESPONSE_CLOSE);
  gtk_widget_set_can_default(closebutton1, TRUE);

  g_signal_connect ((gpointer) prefwin, "configure_event",
                    G_CALLBACK (on_prefwin_configure_event),
                    NULL);
  g_signal_connect ((gpointer) prefwin, "window_state_event",
                    G_CALLBACK (on_prefwin_window_state_event),
                    NULL);
  g_signal_connect ((gpointer) prefwin, "realize",
                    G_CALLBACK (on_prefwin_realize),
                    NULL);
  g_signal_connect ((gpointer) prefwin, "key_press_event",
                    G_CALLBACK (on_prefwin_key_press_event),
                    NULL);
  g_signal_connect ((gpointer) convert8to16, "toggled",
                    G_CALLBACK (on_convert8to16_toggled),
                    NULL);
  g_signal_connect ((gpointer) convert16to24, "toggled",
                    G_CALLBACK (on_convert16to24_toggled),
                    NULL);
  g_signal_connect ((gpointer) combo_bit_override, "changed",
                    G_CALLBACK (on_combo_bit_override_changed),
                    NULL);
  g_signal_connect ((gpointer) comboboxentry_direct_sr, "changed",
                    G_CALLBACK (on_comboboxentry_direct_sr_changed),
                    NULL);
  g_signal_connect ((gpointer) comboboxentry_sr_mult_48, "changed",
                    G_CALLBACK (on_comboboxentry_sr_mult_48_changed),
                    NULL);
  g_signal_connect ((gpointer) comboboxentry_sr_mult_44, "changed",
                    G_CALLBACK (on_comboboxentry_sr_mult_44_changed),
                    NULL);
  g_signal_connect ((gpointer) checkbutton_dependent_sr, "toggled",
                    G_CALLBACK (on_checkbutton_dependent_sr_toggled),
                    NULL);
  g_signal_connect ((gpointer) checkbutton_sr_override, "toggled",
                    G_CALLBACK (on_checkbutton_sr_override_toggled),
                    NULL);
  g_signal_connect ((gpointer) pref_replaygain_source_mode, "changed",
                    G_CALLBACK (on_pref_replaygain_source_mode_changed),
                    NULL);
  g_signal_connect ((gpointer) pref_replaygain_processing, "changed",
                    G_CALLBACK (on_pref_replaygain_processing_changed),
                    NULL);
  g_signal_connect ((gpointer) replaygain_preamp, "value_changed",
                    G_CALLBACK (on_replaygain_preamp_value_changed),
                    NULL);
  g_signal_connect ((gpointer) global_preamp, "value_changed",
                    G_CALLBACK (on_global_preamp_value_changed),
                    NULL);
  g_signal_connect ((gpointer) cli_add_to_playlist, "toggled",
                    G_CALLBACK (on_cli_add_to_playlist_toggled),
                    NULL);
  g_signal_connect ((gpointer) cli_playlist_name, "changed",
                    G_CALLBACK (on_cli_playlist_name_changed),
                    NULL);
  g_signal_connect ((gpointer) resume_last_session, "toggled",
                    G_CALLBACK (on_resume_last_session_toggled),
                    NULL);
  g_signal_connect ((gpointer) resume_always_paused, "toggled",
                    G_CALLBACK (on_resume_always_paused_toggled),
                    NULL);
  g_signal_connect ((gpointer) ignore_archives, "toggled",
                    G_CALLBACK (on_ignore_archives_toggled),
                    NULL);
  g_signal_connect ((gpointer) reset_autostop, "toggled",
                    G_CALLBACK (on_reset_autostop_toggled),
                    NULL);
  g_signal_connect ((gpointer) reset_autostopqueue, "toggled",
                    G_CALLBACK (on_reset_autostopqueue_toggled),
                    NULL);
  g_signal_connect ((gpointer) reset_autostopalbum, "toggled",
                    G_CALLBACK (on_reset_autostopalbum_toggled),
                    NULL);
  g_signal_connect ((gpointer) minimize_on_startup, "clicked",
                    G_CALLBACK (on_minimize_on_startup_clicked),
                    NULL);
  g_signal_connect ((gpointer) pref_close_send_to_tray, "clicked",
                    G_CALLBACK (on_pref_close_send_to_tray_clicked),
                    NULL);
  g_signal_connect ((gpointer) hide_tray_icon, "toggled",
                    G_CALLBACK (on_hide_tray_icon_toggled),
                    NULL);
  g_signal_connect ((gpointer) move_to_trash, "clicked",
                    G_CALLBACK (on_move_to_trash_clicked),
                    NULL);
  g_signal_connect ((gpointer) enable_shift_jis_recoding, "toggled",
                    G_CALLBACK (on_enable_shift_jis_recoding_toggled),
                    NULL);
  g_signal_connect ((gpointer) enable_cp1251_recoding, "toggled",
                    G_CALLBACK (on_enable_cp1251_recoding_toggled),
                    NULL);
  g_signal_connect ((gpointer) enable_cp936_recoding, "toggled",
                    G_CALLBACK (on_enable_cp936_recoding_toggled),
                    NULL);
  g_signal_connect ((gpointer) gui_fps, "value_changed",
                    G_CALLBACK (on_gui_fps_value_changed),
                    NULL);
  g_signal_connect ((gpointer) titlebar_format_playing, "changed",
                    G_CALLBACK (on_titlebar_format_playing_changed),
                    NULL);
  g_signal_connect ((gpointer) titlebar_format_stopped, "changed",
                    G_CALLBACK (on_titlebar_format_stopped_changed),
                    NULL);
  g_signal_connect ((gpointer) gui_plugin, "changed",
                    G_CALLBACK (on_gui_plugin_changed),
                    NULL);
  g_signal_connect ((gpointer) display_seltime, "toggled",
                    G_CALLBACK (on_display_seltime_toggled),
                    NULL);
  g_signal_connect ((gpointer) mmb_delete_playlist, "toggled",
                    G_CALLBACK (on_mmb_delete_playlist_toggled),
                    NULL);
  g_signal_connect ((gpointer) hide_delete_from_disk, "toggled",
                    G_CALLBACK (on_hide_delete_from_disk_toggled),
                    NULL);
  g_signal_connect ((gpointer) skip_deleted_songs, "toggled",
                    G_CALLBACK (on_skip_deleted_songs_toggled),
                    NULL);
  g_signal_connect ((gpointer) auto_name_playlist_from_folder, "toggled",
                    G_CALLBACK (on_auto_name_playlist_from_folder_toggled),
                    NULL);
  g_signal_connect ((gpointer) auto_size_columns, "toggled",
                    G_CALLBACK (on_auto_size_columns_toggled),
                    NULL);
  g_signal_connect ((gpointer) listview_group_spacing, "value_changed",
                    G_CALLBACK (on_listview_group_spacing_value_changed),
                    NULL);
  g_signal_connect ((gpointer) override_bar_colors, "toggled",
                    G_CALLBACK (on_override_bar_colors_toggled),
                    NULL);
  g_signal_connect ((gpointer) bar_foreground, "color_set",
                    G_CALLBACK (on_bar_foreground_color_set),
                    NULL);
  g_signal_connect ((gpointer) bar_background, "color_set",
                    G_CALLBACK (on_bar_background_color_set),
                    NULL);
  g_signal_connect ((gpointer) override_tabstrip_colors, "toggled",
                    G_CALLBACK (on_override_tabstrip_colors_toggled),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_base, "color_set",
                    G_CALLBACK (on_tabstrip_base_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_dark, "color_set",
                    G_CALLBACK (on_tabstrip_dark_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_light, "color_set",
                    G_CALLBACK (on_tabstrip_light_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_mid, "color_set",
                    G_CALLBACK (on_tabstrip_mid_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_text, "color_set",
                    G_CALLBACK (on_tabstrip_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_selected_text, "color_set",
                    G_CALLBACK (on_tabstrip_selected_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_playing_bold, "toggled",
                    G_CALLBACK (on_tabstrip_playing_bold_toggled),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_playing_italic, "toggled",
                    G_CALLBACK (on_tabstrip_playing_italic_toggled),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_selected_bold, "toggled",
                    G_CALLBACK (on_tabstrip_selected_bold_toggled),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_selected_italic, "toggled",
                    G_CALLBACK (on_tabstrip_selected_italic_toggled),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_text_font, "font_set",
                    G_CALLBACK (on_tabstrip_text_font_set),
                    NULL);
  g_signal_connect ((gpointer) tabstrip_playing_text, "color_set",
                    G_CALLBACK (on_tabstrip_playing_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_text, "color_set",
                    G_CALLBACK (on_listview_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_playing_text, "color_set",
                    G_CALLBACK (on_listview_playing_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_selected_text, "color_set",
                    G_CALLBACK (on_listview_selected_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_column_text, "color_set",
                    G_CALLBACK (on_listview_column_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_column_text_font, "font_set",
                    G_CALLBACK (on_listview_column_text_font_set),
                    NULL);
  g_signal_connect ((gpointer) listview_group_text, "color_set",
                    G_CALLBACK (on_listview_group_text_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_group_text_font, "font_set",
                    G_CALLBACK (on_listview_group_text_font_set),
                    NULL);
  g_signal_connect ((gpointer) listview_text_font, "font_set",
                    G_CALLBACK (on_listview_text_font_set),
                    NULL);
  g_signal_connect ((gpointer) listview_odd_row, "color_set",
                    G_CALLBACK (on_listview_odd_row_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_selected_row, "color_set",
                    G_CALLBACK (on_listview_selected_row_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_cursor, "color_set",
                    G_CALLBACK (on_listview_cursor_color_set),
                    NULL);
  g_signal_connect ((gpointer) listview_even_row, "color_set",
                    G_CALLBACK (on_listview_even_row_color_set),
                    NULL);
  g_signal_connect ((gpointer) override_listview_colors, "toggled",
                    G_CALLBACK (on_override_listview_colors_toggled),
                    NULL);
  g_signal_connect ((gpointer) listview_playing_text_bold, "toggled",
                    G_CALLBACK (on_listview_playing_text_bold_toggled),
                    NULL);
  g_signal_connect ((gpointer) listview_selected_text_bold, "toggled",
                    G_CALLBACK (on_listview_selected_text_bold_toggled),
                    NULL);
  g_signal_connect ((gpointer) listview_playing_text_italic, "toggled",
                    G_CALLBACK (on_listview_playing_text_italic_toggled),
                    NULL);
  g_signal_connect ((gpointer) listview_selected_text_italic, "toggled",
                    G_CALLBACK (on_listview_selected_text_italic_toggled),
                    NULL);
  g_signal_connect ((gpointer) use_visualization_base_color, "toggled",
                    G_CALLBACK (on_use_visualization_base_color_toggled),
                    NULL);
  g_signal_connect ((gpointer) visualization_custom_color_button, "color_set",
                    G_CALLBACK (on_visualization_custom_color_button_color_set),
                    NULL);
  g_signal_connect ((gpointer) use_visualization_background_color, "toggled",
                    G_CALLBACK (on_use_visualization_background_color_toggled),
                    NULL);
  g_signal_connect ((gpointer) visualization_custom_background_color_button, "color_set",
                    G_CALLBACK (on_visualization_custom_background_color_button_color_set),
                    NULL);
  g_signal_connect ((gpointer) pref_network_enableproxy, "clicked",
                    G_CALLBACK (on_pref_network_enableproxy_clicked),
                    NULL);
  g_signal_connect ((gpointer) pref_network_proxyaddress, "changed",
                    G_CALLBACK (on_pref_network_proxyaddress_changed),
                    NULL);
  g_signal_connect ((gpointer) pref_network_proxyport, "changed",
                    G_CALLBACK (on_pref_network_proxyport_changed),
                    NULL);
  g_signal_connect ((gpointer) pref_network_proxytype, "changed",
                    G_CALLBACK (on_pref_network_proxytype_changed),
                    NULL);
  g_signal_connect ((gpointer) proxyuser, "changed",
                    G_CALLBACK (on_proxyuser_changed),
                    NULL);
  g_signal_connect ((gpointer) proxypassword, "changed",
                    G_CALLBACK (on_proxypassword_changed),
                    NULL);
  g_signal_connect ((gpointer) useragent, "changed",
                    G_CALLBACK (on_useragent_changed),
                    NULL);
  g_signal_connect ((gpointer) edit_content_type_mapping, "clicked",
                    G_CALLBACK (on_edit_content_type_mapping_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkey_add, "clicked",
                    G_CALLBACK (on_hotkey_add_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkey_remove, "clicked",
                    G_CALLBACK (on_hotkey_remove_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_list, "cursor_changed",
                    G_CALLBACK (on_hotkeys_list_cursor_changed),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_actions, "clicked",
                    G_CALLBACK (on_hotkeys_actions_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_set_key, "clicked",
                    G_CALLBACK (on_hotkeys_set_key_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_set_key, "key_press_event",
                    G_CALLBACK (on_hotkeys_set_key_key_press_event),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_apply, "clicked",
                    G_CALLBACK (on_hotkeys_apply_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_revert, "clicked",
                    G_CALLBACK (on_hotkeys_revert_clicked),
                    NULL);
  g_signal_connect ((gpointer) hotkeys_defaults, "clicked",
                    G_CALLBACK (on_hotkeys_defaults_clicked),
                    NULL);
  g_signal_connect ((gpointer) pref_pluginlist, "cursor_changed",
                    G_CALLBACK (on_pref_pluginlist_cursor_changed),
                    NULL);
  g_signal_connect ((gpointer) pref_pluginlist, "row_activated",
                    G_CALLBACK (on_pref_pluginlist_row_activated),
                    NULL);
  g_signal_connect ((gpointer) pref_pluginlist, "button_press_event",
                    G_CALLBACK (on_pref_pluginlist_button_press_event),
                    NULL);
  g_signal_connect ((gpointer) plugin_conf_tab_btn, "clicked",
                    G_CALLBACK (on_plugin_conf_tab_btn_clicked),
                    NULL);
  g_signal_connect ((gpointer) plugin_info_tab_btn, "clicked",
                    G_CALLBACK (on_plugin_info_tab_btn_clicked),
                    NULL);
  g_signal_connect ((gpointer) plugin_license_tab_btn, "clicked",
                    G_CALLBACK (on_plugin_license_tab_btn_clicked),
                    NULL);
  g_signal_connect ((gpointer) plugin_notebook, "switch_page",
                    G_CALLBACK (on_plugin_notebook_switch_page),
                    NULL);
  g_signal_connect ((gpointer) plugin_conf_reset_btn, "clicked",
                    G_CALLBACK (on_plugin_conf_reset_btn_clicked),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (prefwin, prefwin, "prefwin");
  GLADE_HOOKUP_OBJECT_NO_REF (prefwin, dialog_vbox2, "dialog_vbox2");
  GLADE_HOOKUP_OBJECT (prefwin, notebook, "notebook");
  GLADE_HOOKUP_OBJECT (prefwin, vbox10, "vbox10");
  GLADE_HOOKUP_OBJECT (prefwin, hbox11, "hbox11");
  GLADE_HOOKUP_OBJECT (prefwin, label23, "label23");
  GLADE_HOOKUP_OBJECT (prefwin, pref_output_plugin, "pref_output_plugin");
  GLADE_HOOKUP_OBJECT (prefwin, hbox12, "hbox12");
  GLADE_HOOKUP_OBJECT (prefwin, label4, "label4");
  GLADE_HOOKUP_OBJECT (prefwin, pref_soundcard, "pref_soundcard");
  GLADE_HOOKUP_OBJECT (prefwin, convert8to16, "convert8to16");
  GLADE_HOOKUP_OBJECT (prefwin, convert16to24, "convert16to24");
  GLADE_HOOKUP_OBJECT (prefwin, hbox151, "hbox151");
  GLADE_HOOKUP_OBJECT (prefwin, label_bit_override, "label_bit_override");
  GLADE_HOOKUP_OBJECT (prefwin, combo_bit_override, "combo_bit_override");
  GLADE_HOOKUP_OBJECT (prefwin, frame15, "frame15");
  GLADE_HOOKUP_OBJECT (prefwin, alignment31, "alignment31");
  GLADE_HOOKUP_OBJECT (prefwin, vbox52, "vbox52");
  GLADE_HOOKUP_OBJECT (prefwin, hbox145, "hbox145");
  GLADE_HOOKUP_OBJECT (prefwin, label_direct_sr, "label_direct_sr");
  GLADE_HOOKUP_OBJECT (prefwin, comboboxentry_direct_sr, "comboboxentry_direct_sr");
  GLADE_HOOKUP_OBJECT (prefwin, frame16, "frame16");
  GLADE_HOOKUP_OBJECT (prefwin, alignment32, "alignment32");
  GLADE_HOOKUP_OBJECT (prefwin, vbox53, "vbox53");
  GLADE_HOOKUP_OBJECT (prefwin, hbox147, "hbox147");
  GLADE_HOOKUP_OBJECT (prefwin, label_sr_mult_48, "label_sr_mult_48");
  GLADE_HOOKUP_OBJECT (prefwin, comboboxentry_sr_mult_48, "comboboxentry_sr_mult_48");
  GLADE_HOOKUP_OBJECT (prefwin, hbox148, "hbox148");
  GLADE_HOOKUP_OBJECT (prefwin, label_sr_mult_44, "label_sr_mult_44");
  GLADE_HOOKUP_OBJECT (prefwin, comboboxentry_sr_mult_44, "comboboxentry_sr_mult_44");
  GLADE_HOOKUP_OBJECT (prefwin, checkbutton_dependent_sr, "checkbutton_dependent_sr");
  GLADE_HOOKUP_OBJECT (prefwin, checkbutton_sr_override, "checkbutton_sr_override");
  GLADE_HOOKUP_OBJECT (prefwin, Sound, "Sound");
  GLADE_HOOKUP_OBJECT (prefwin, vbox8, "vbox8");
  GLADE_HOOKUP_OBJECT (prefwin, frame14, "frame14");
  GLADE_HOOKUP_OBJECT (prefwin, alignment30, "alignment30");
  GLADE_HOOKUP_OBJECT (prefwin, vbox48, "vbox48");
  GLADE_HOOKUP_OBJECT (prefwin, hbox10, "hbox10");
  GLADE_HOOKUP_OBJECT (prefwin, label8, "label8");
  GLADE_HOOKUP_OBJECT (prefwin, pref_replaygain_source_mode, "pref_replaygain_source_mode");
  GLADE_HOOKUP_OBJECT (prefwin, hbox139, "hbox139");
  GLADE_HOOKUP_OBJECT (prefwin, label163, "label163");
  GLADE_HOOKUP_OBJECT (prefwin, pref_replaygain_processing, "pref_replaygain_processing");
  GLADE_HOOKUP_OBJECT (prefwin, table1, "table1");
  GLADE_HOOKUP_OBJECT (prefwin, hbox100, "hbox100");
  GLADE_HOOKUP_OBJECT (prefwin, label125, "label125");
  GLADE_HOOKUP_OBJECT (prefwin, replaygain_preamp, "replaygain_preamp");
  GLADE_HOOKUP_OBJECT (prefwin, label126, "label126");
  GLADE_HOOKUP_OBJECT (prefwin, label124, "label124");
  GLADE_HOOKUP_OBJECT (prefwin, label124a, "label124a");
  GLADE_HOOKUP_OBJECT (prefwin, hbox100a, "hbox100a");
  GLADE_HOOKUP_OBJECT (prefwin, label125a, "label125a");
  GLADE_HOOKUP_OBJECT (prefwin, global_preamp, "global_preamp");
  GLADE_HOOKUP_OBJECT (prefwin, label126a, "label126a");
  GLADE_HOOKUP_OBJECT (prefwin, label162, "label162");
  GLADE_HOOKUP_OBJECT (prefwin, hbox66, "hbox66");
  GLADE_HOOKUP_OBJECT (prefwin, cli_add_to_playlist, "cli_add_to_playlist");
  GLADE_HOOKUP_OBJECT (prefwin, cli_playlist_name, "cli_playlist_name");
  GLADE_HOOKUP_OBJECT (prefwin, resume_last_session, "resume_last_session");
  GLADE_HOOKUP_OBJECT (prefwin, resume_always_paused, "resume_always_paused");
  GLADE_HOOKUP_OBJECT (prefwin, ignore_archives, "ignore_archives");
  GLADE_HOOKUP_OBJECT (prefwin, reset_autostop, "reset_autostop");
  GLADE_HOOKUP_OBJECT (prefwin, reset_autostopqueue, "reset_autostopqueue");
  GLADE_HOOKUP_OBJECT (prefwin, reset_autostopalbum, "reset_autostopalbum");
  GLADE_HOOKUP_OBJECT (prefwin, label39, "label39");
  GLADE_HOOKUP_OBJECT (prefwin, vbox29, "vbox29");
  GLADE_HOOKUP_OBJECT (prefwin, prefwin_dsp_vbox, "prefwin_dsp_vbox");
  GLADE_HOOKUP_OBJECT (prefwin, label110, "label110");
  GLADE_HOOKUP_OBJECT (prefwin, notebook5, "notebook5");
  GLADE_HOOKUP_OBJECT (prefwin, vbox9, "vbox9");
  GLADE_HOOKUP_OBJECT (prefwin, minimize_on_startup, "minimize_on_startup");
  GLADE_HOOKUP_OBJECT (prefwin, pref_close_send_to_tray, "pref_close_send_to_tray");
  GLADE_HOOKUP_OBJECT (prefwin, hide_tray_icon, "hide_tray_icon");
  GLADE_HOOKUP_OBJECT (prefwin, move_to_trash, "move_to_trash");
  GLADE_HOOKUP_OBJECT (prefwin, enable_shift_jis_recoding, "enable_shift_jis_recoding");
  GLADE_HOOKUP_OBJECT (prefwin, enable_cp1251_recoding, "enable_cp1251_recoding");
  GLADE_HOOKUP_OBJECT (prefwin, enable_cp936_recoding, "enable_cp936_recoding");
  GLADE_HOOKUP_OBJECT (prefwin, hbox102, "hbox102");
  GLADE_HOOKUP_OBJECT (prefwin, label129, "label129");
  GLADE_HOOKUP_OBJECT (prefwin, gui_fps, "gui_fps");
  GLADE_HOOKUP_OBJECT (prefwin, hbox64, "hbox64");
  GLADE_HOOKUP_OBJECT (prefwin, label101, "label101");
  GLADE_HOOKUP_OBJECT (prefwin, titlebar_format_playing, "titlebar_format_playing");
  GLADE_HOOKUP_OBJECT (prefwin, hbox65, "hbox65");
  GLADE_HOOKUP_OBJECT (prefwin, label102, "label102");
  GLADE_HOOKUP_OBJECT (prefwin, titlebar_format_stopped, "titlebar_format_stopped");
  GLADE_HOOKUP_OBJECT (prefwin, hbox101, "hbox101");
  GLADE_HOOKUP_OBJECT (prefwin, label128, "label128");
  GLADE_HOOKUP_OBJECT (prefwin, gui_plugin, "gui_plugin");
  GLADE_HOOKUP_OBJECT (prefwin, display_seltime, "display_seltime");
  GLADE_HOOKUP_OBJECT (prefwin, label147, "label147");
  GLADE_HOOKUP_OBJECT (prefwin, vbox44, "vbox44");
  GLADE_HOOKUP_OBJECT (prefwin, mmb_delete_playlist, "mmb_delete_playlist");
  GLADE_HOOKUP_OBJECT (prefwin, hide_delete_from_disk, "hide_delete_from_disk");
  GLADE_HOOKUP_OBJECT (prefwin, skip_deleted_songs, "skip_deleted_songs");
  GLADE_HOOKUP_OBJECT (prefwin, auto_name_playlist_from_folder, "auto_name_playlist_from_folder");
  GLADE_HOOKUP_OBJECT (prefwin, auto_size_columns, "auto_size_columns");
  GLADE_HOOKUP_OBJECT (prefwin, hbox141, "hbox141");
  GLADE_HOOKUP_OBJECT (prefwin, label164, "label164");
  GLADE_HOOKUP_OBJECT (prefwin, listview_group_spacing, "listview_group_spacing");
  GLADE_HOOKUP_OBJECT (prefwin, label165, "label165");
  GLADE_HOOKUP_OBJECT (prefwin, label148, "label148");
  GLADE_HOOKUP_OBJECT (prefwin, label2, "label2");
  GLADE_HOOKUP_OBJECT (prefwin, notebook4, "notebook4");
  GLADE_HOOKUP_OBJECT (prefwin, vbox21, "vbox21");
  GLADE_HOOKUP_OBJECT (prefwin, override_bar_colors, "override_bar_colors");
  GLADE_HOOKUP_OBJECT (prefwin, bar_colors_group, "bar_colors_group");
  GLADE_HOOKUP_OBJECT (prefwin, label43, "label43");
  GLADE_HOOKUP_OBJECT (prefwin, label47, "label47");
  GLADE_HOOKUP_OBJECT (prefwin, bar_foreground, "bar_foreground");
  GLADE_HOOKUP_OBJECT (prefwin, bar_background, "bar_background");
  GLADE_HOOKUP_OBJECT (prefwin, label73, "label73");
  GLADE_HOOKUP_OBJECT (prefwin, vbox22, "vbox22");
  GLADE_HOOKUP_OBJECT (prefwin, override_tabstrip_colors, "override_tabstrip_colors");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_colors_group, "tabstrip_colors_group");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_base, "tabstrip_base");
  GLADE_HOOKUP_OBJECT (prefwin, label76, "label76");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_dark, "tabstrip_dark");
  GLADE_HOOKUP_OBJECT (prefwin, label44, "label44");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_light, "tabstrip_light");
  GLADE_HOOKUP_OBJECT (prefwin, label46, "label46");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_mid, "tabstrip_mid");
  GLADE_HOOKUP_OBJECT (prefwin, label45, "label45");
  GLADE_HOOKUP_OBJECT (prefwin, label127, "label127");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_text, "tabstrip_text");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_selected_text, "tabstrip_selected_text");
  GLADE_HOOKUP_OBJECT (prefwin, hbox136, "hbox136");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_playing_bold, "tabstrip_playing_bold");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_playing_italic, "tabstrip_playing_italic");
  GLADE_HOOKUP_OBJECT (prefwin, hbox137, "hbox137");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_selected_bold, "tabstrip_selected_bold");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_selected_italic, "tabstrip_selected_italic");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_text_font, "tabstrip_text_font");
  GLADE_HOOKUP_OBJECT (prefwin, label152, "label152");
  GLADE_HOOKUP_OBJECT (prefwin, label153, "label153");
  GLADE_HOOKUP_OBJECT (prefwin, tabstrip_playing_text, "tabstrip_playing_text");
  GLADE_HOOKUP_OBJECT (prefwin, label74, "label74");
  GLADE_HOOKUP_OBJECT (prefwin, vbox23, "vbox23");
  GLADE_HOOKUP_OBJECT (prefwin, frame11, "frame11");
  GLADE_HOOKUP_OBJECT (prefwin, listview_colors_group, "listview_colors_group");
  GLADE_HOOKUP_OBJECT (prefwin, frame12, "frame12");
  GLADE_HOOKUP_OBJECT (prefwin, alignment28, "alignment28");
  GLADE_HOOKUP_OBJECT (prefwin, table3, "table3");
  GLADE_HOOKUP_OBJECT (prefwin, label77, "label77");
  GLADE_HOOKUP_OBJECT (prefwin, listview_text, "listview_text");
  GLADE_HOOKUP_OBJECT (prefwin, label150, "label150");
  GLADE_HOOKUP_OBJECT (prefwin, listview_playing_text, "listview_playing_text");
  GLADE_HOOKUP_OBJECT (prefwin, label61, "label61");
  GLADE_HOOKUP_OBJECT (prefwin, listview_selected_text, "listview_selected_text");
  GLADE_HOOKUP_OBJECT (prefwin, label154, "label154");
  GLADE_HOOKUP_OBJECT (prefwin, listview_column_text, "listview_column_text");
  GLADE_HOOKUP_OBJECT (prefwin, listview_column_text_font, "listview_column_text_font");
  GLADE_HOOKUP_OBJECT (prefwin, label149, "label149");
  GLADE_HOOKUP_OBJECT (prefwin, listview_group_text, "listview_group_text");
  GLADE_HOOKUP_OBJECT (prefwin, listview_group_text_font, "listview_group_text_font");
  GLADE_HOOKUP_OBJECT (prefwin, listview_text_font, "listview_text_font");
  GLADE_HOOKUP_OBJECT (prefwin, label158, "label158");
  GLADE_HOOKUP_OBJECT (prefwin, vseparator1, "vseparator1");
  GLADE_HOOKUP_OBJECT (prefwin, frame13, "frame13");
  GLADE_HOOKUP_OBJECT (prefwin, alignment29, "alignment29");
  GLADE_HOOKUP_OBJECT (prefwin, table4, "table4");
  GLADE_HOOKUP_OBJECT (prefwin, label58, "label58");
  GLADE_HOOKUP_OBJECT (prefwin, label59, "label59");
  GLADE_HOOKUP_OBJECT (prefwin, listview_odd_row, "listview_odd_row");
  GLADE_HOOKUP_OBJECT (prefwin, label78, "label78");
  GLADE_HOOKUP_OBJECT (prefwin, listview_selected_row, "listview_selected_row");
  GLADE_HOOKUP_OBJECT (prefwin, label62, "label62");
  GLADE_HOOKUP_OBJECT (prefwin, listview_cursor, "listview_cursor");
  GLADE_HOOKUP_OBJECT (prefwin, listview_even_row, "listview_even_row");
  GLADE_HOOKUP_OBJECT (prefwin, label159, "label159");
  GLADE_HOOKUP_OBJECT (prefwin, override_listview_colors, "override_listview_colors");
  GLADE_HOOKUP_OBJECT (prefwin, table2, "table2");
  GLADE_HOOKUP_OBJECT (prefwin, label155, "label155");
  GLADE_HOOKUP_OBJECT (prefwin, label156, "label156");
  GLADE_HOOKUP_OBJECT (prefwin, listview_playing_text_bold, "listview_playing_text_bold");
  GLADE_HOOKUP_OBJECT (prefwin, listview_selected_text_bold, "listview_selected_text_bold");
  GLADE_HOOKUP_OBJECT (prefwin, listview_playing_text_italic, "listview_playing_text_italic");
  GLADE_HOOKUP_OBJECT (prefwin, listview_selected_text_italic, "listview_selected_text_italic");
  GLADE_HOOKUP_OBJECT (prefwin, label75, "label75");
  GLADE_HOOKUP_OBJECT (prefwin, table5, "table5");
  GLADE_HOOKUP_OBJECT (prefwin, use_visualization_base_color, "use_visualization_base_color");
  GLADE_HOOKUP_OBJECT (prefwin, visualization_custom_color_button, "visualization_custom_color_button");
  GLADE_HOOKUP_OBJECT (prefwin, use_visualization_background_color, "use_visualization_background_color");
  GLADE_HOOKUP_OBJECT (prefwin, visualization_custom_background_color_button, "visualization_custom_background_color_button");
  GLADE_HOOKUP_OBJECT (prefwin, label175, "label175");
  GLADE_HOOKUP_OBJECT (prefwin, label100, "label100");
  GLADE_HOOKUP_OBJECT (prefwin, vbox55, "vbox55");
  GLADE_HOOKUP_OBJECT (prefwin, toggle_medialib_on, "toggle_medialib_on");
  GLADE_HOOKUP_OBJECT (prefwin, hseparator7, "hseparator7");
  GLADE_HOOKUP_OBJECT (prefwin, label174, "label174");
  GLADE_HOOKUP_OBJECT (prefwin, vbox56, "vbox56");
  GLADE_HOOKUP_OBJECT (prefwin, scrolledwindow17, "scrolledwindow17");
  GLADE_HOOKUP_OBJECT (prefwin, treeview_medialib_folders, "treeview_medialib_folders");
  GLADE_HOOKUP_OBJECT (prefwin, hbox150, "hbox150");
  GLADE_HOOKUP_OBJECT (prefwin, button_medialib_add_folder, "button_medialib_add_folder");
  GLADE_HOOKUP_OBJECT (prefwin, image714, "image714");
  GLADE_HOOKUP_OBJECT (prefwin, button_medialib_remove_folder, "button_medialib_remove_folder");
  GLADE_HOOKUP_OBJECT (prefwin, image715, "image715");
  GLADE_HOOKUP_OBJECT (prefwin, label173, "label173");
  GLADE_HOOKUP_OBJECT (prefwin, vbox11, "vbox11");
  GLADE_HOOKUP_OBJECT (prefwin, pref_network_enableproxy, "pref_network_enableproxy");
  GLADE_HOOKUP_OBJECT (prefwin, hbox13, "hbox13");
  GLADE_HOOKUP_OBJECT (prefwin, label18, "label18");
  GLADE_HOOKUP_OBJECT (prefwin, pref_network_proxyaddress, "pref_network_proxyaddress");
  GLADE_HOOKUP_OBJECT (prefwin, hbox14, "hbox14");
  GLADE_HOOKUP_OBJECT (prefwin, label19, "label19");
  GLADE_HOOKUP_OBJECT (prefwin, pref_network_proxyport, "pref_network_proxyport");
  GLADE_HOOKUP_OBJECT (prefwin, hbox15, "hbox15");
  GLADE_HOOKUP_OBJECT (prefwin, label20, "label20");
  GLADE_HOOKUP_OBJECT (prefwin, pref_network_proxytype, "pref_network_proxytype");
  GLADE_HOOKUP_OBJECT (prefwin, hbox61, "hbox61");
  GLADE_HOOKUP_OBJECT (prefwin, label97, "label97");
  GLADE_HOOKUP_OBJECT (prefwin, proxyuser, "proxyuser");
  GLADE_HOOKUP_OBJECT (prefwin, hbox62, "hbox62");
  GLADE_HOOKUP_OBJECT (prefwin, label98, "label98");
  GLADE_HOOKUP_OBJECT (prefwin, proxypassword, "proxypassword");
  GLADE_HOOKUP_OBJECT (prefwin, hseparator1, "hseparator1");
  GLADE_HOOKUP_OBJECT (prefwin, hbox104, "hbox104");
  GLADE_HOOKUP_OBJECT (prefwin, label131, "label131");
  GLADE_HOOKUP_OBJECT (prefwin, useragent, "useragent");
  GLADE_HOOKUP_OBJECT (prefwin, hseparator5, "hseparator5");
  GLADE_HOOKUP_OBJECT (prefwin, edit_content_type_mapping, "edit_content_type_mapping");
  GLADE_HOOKUP_OBJECT (prefwin, label16, "label16");
  GLADE_HOOKUP_OBJECT (prefwin, vbox36, "vbox36");
  GLADE_HOOKUP_OBJECT (prefwin, hbox119, "hbox119");
  GLADE_HOOKUP_OBJECT (prefwin, label144, "label144");
  GLADE_HOOKUP_OBJECT (prefwin, hbuttonbox3, "hbuttonbox3");
  GLADE_HOOKUP_OBJECT (prefwin, hotkey_add, "hotkey_add");
  GLADE_HOOKUP_OBJECT (prefwin, hotkey_remove, "hotkey_remove");
  GLADE_HOOKUP_OBJECT (prefwin, scrolledwindow9, "scrolledwindow9");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_list, "hotkeys_list");
  GLADE_HOOKUP_OBJECT (prefwin, hbox105, "hbox105");
  GLADE_HOOKUP_OBJECT (prefwin, label133, "label133");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_actions, "hotkeys_actions");
  GLADE_HOOKUP_OBJECT (prefwin, hbox106, "hbox106");
  GLADE_HOOKUP_OBJECT (prefwin, label134, "label134");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_set_key, "hotkeys_set_key");
  GLADE_HOOKUP_OBJECT (prefwin, hseparator6, "hseparator6");
  GLADE_HOOKUP_OBJECT (prefwin, hbox120, "hbox120");
  GLADE_HOOKUP_OBJECT (prefwin, label145, "label145");
  GLADE_HOOKUP_OBJECT (prefwin, hbuttonbox4, "hbuttonbox4");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_apply, "hotkeys_apply");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_revert, "hotkeys_revert");
  GLADE_HOOKUP_OBJECT (prefwin, hotkeys_defaults, "hotkeys_defaults");
  GLADE_HOOKUP_OBJECT (prefwin, alignment26, "alignment26");
  GLADE_HOOKUP_OBJECT (prefwin, hbox121, "hbox121");
  GLADE_HOOKUP_OBJECT (prefwin, image638, "image638");
  GLADE_HOOKUP_OBJECT (prefwin, label146, "label146");
  GLADE_HOOKUP_OBJECT (prefwin, label132, "label132");
  GLADE_HOOKUP_OBJECT (prefwin, hpaned1, "hpaned1");
  GLADE_HOOKUP_OBJECT (prefwin, scrolledwindow2, "scrolledwindow2");
  GLADE_HOOKUP_OBJECT (prefwin, pref_pluginlist, "pref_pluginlist");
  GLADE_HOOKUP_OBJECT (prefwin, alignment33, "alignment33");
  GLADE_HOOKUP_OBJECT (prefwin, vbox54, "vbox54");
  GLADE_HOOKUP_OBJECT (prefwin, hbox149, "hbox149");
  GLADE_HOOKUP_OBJECT (prefwin, fillerlabel1, "fillerlabel1");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_tabbtn_hbtnbox, "plugin_tabbtn_hbtnbox");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_conf_tab_btn, "plugin_conf_tab_btn");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_info_tab_btn, "plugin_info_tab_btn");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_license_tab_btn, "plugin_license_tab_btn");
  GLADE_HOOKUP_OBJECT (prefwin, fillerlabel2, "fillerlabel2");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_notebook, "plugin_notebook");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_conf_tab_vbox, "plugin_conf_tab_vbox");
  GLADE_HOOKUP_OBJECT (prefwin, plug_conf_dlg_scrolledwindow, "plug_conf_dlg_scrolledwindow");
  GLADE_HOOKUP_OBJECT (prefwin, plug_conf_dlg_viewport, "plug_conf_dlg_viewport");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_actions_btnbox, "plugin_actions_btnbox");
  GLADE_HOOKUP_OBJECT (prefwin, plugin_conf_reset_btn, "plugin_conf_reset_btn");
  GLADE_HOOKUP_OBJECT (prefwin, label170, "label170");
  GLADE_HOOKUP_OBJECT (prefwin, vbox12, "vbox12");
  GLADE_HOOKUP_OBJECT (prefwin, hbox103, "hbox103");
  GLADE_HOOKUP_OBJECT (prefwin, label130, "label130");
  GLADE_HOOKUP_OBJECT (prefwin, plug_version, "plug_version");
  GLADE_HOOKUP_OBJECT (prefwin, scrolledwindow8, "scrolledwindow8");
  GLADE_HOOKUP_OBJECT (prefwin, plug_description, "plug_description");
  GLADE_HOOKUP_OBJECT (prefwin, hbox20, "hbox20");
  GLADE_HOOKUP_OBJECT (prefwin, weblink, "weblink");
  GLADE_HOOKUP_OBJECT (prefwin, label171, "label171");
  GLADE_HOOKUP_OBJECT (prefwin, scrolledwindow16, "scrolledwindow16");
  GLADE_HOOKUP_OBJECT (prefwin, plug_license, "plug_license");
  GLADE_HOOKUP_OBJECT (prefwin, label172, "label172");
  GLADE_HOOKUP_OBJECT (prefwin, label3, "label3");
  GLADE_HOOKUP_OBJECT_NO_REF (prefwin, dialog_action_area2, "dialog_action_area2");
  GLADE_HOOKUP_OBJECT (prefwin, closebutton1, "closebutton1");

  return prefwin;
}

GtkWidget*
create_entrydialog (void)
{
  GtkWidget *entrydialog;
  GtkWidget *dialog_vbox3;
  GtkWidget *vbox15;
  GtkWidget *hbox33;
  GtkWidget *title_label;
  GtkWidget *title;
  GtkWidget *dialog_action_area3;
  GtkWidget *cancelbutton2;
  GtkWidget *alignment17;
  GtkWidget *hbox58;
  GtkWidget *image396;
  GtkWidget *label94;
  GtkWidget *okbutton2;
  GtkWidget *alignment16;
  GtkWidget *hbox57;
  GtkWidget *image395;
  GtkWidget *label93;

  entrydialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (entrydialog), "EntryDialog");
  gtk_window_set_position (GTK_WINDOW (entrydialog), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_destroy_with_parent (GTK_WINDOW (entrydialog), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (entrydialog), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox3 = gtk_dialog_get_content_area (GTK_DIALOG (entrydialog));
  gtk_widget_show (dialog_vbox3);

  vbox15 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox15);
  gtk_box_pack_start (GTK_BOX (dialog_vbox3), vbox15, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox15), 12);

  hbox33 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox33);
  gtk_box_pack_start (GTK_BOX (vbox15), hbox33, TRUE, TRUE, 0);

  title_label = gtk_label_new (_("Title:"));
  gtk_widget_show (title_label);
  gtk_box_pack_start (GTK_BOX (hbox33), title_label, FALSE, FALSE, 0);

  title = gtk_entry_new ();
  gtk_widget_show (title);
  gtk_box_pack_start (GTK_BOX (hbox33), title, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (title), 8226);
  gtk_entry_set_activates_default (GTK_ENTRY (title), TRUE);

  dialog_action_area3 = gtk_dialog_get_action_area (GTK_DIALOG (entrydialog));
  gtk_widget_show (dialog_action_area3);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area3), GTK_BUTTONBOX_END);

  cancelbutton2 = gtk_button_new ();
  gtk_widget_show (cancelbutton2);
  gtk_dialog_add_action_widget (GTK_DIALOG (entrydialog), cancelbutton2, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton2, TRUE);

  alignment17 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment17);
  gtk_container_add (GTK_CONTAINER (cancelbutton2), alignment17);

  hbox58 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox58);
  gtk_container_add (GTK_CONTAINER (alignment17), hbox58);

  image396 = gtk_image_new_from_stock ("gtk-cancel", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image396);
  gtk_box_pack_start (GTK_BOX (hbox58), image396, FALSE, FALSE, 0);

  label94 = gtk_label_new_with_mnemonic (_("_Cancel"));
  gtk_widget_show (label94);
  gtk_box_pack_start (GTK_BOX (hbox58), label94, FALSE, FALSE, 0);

  okbutton2 = gtk_button_new ();
  gtk_widget_show (okbutton2);
  gtk_dialog_add_action_widget (GTK_DIALOG (entrydialog), okbutton2, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton2, TRUE);

  alignment16 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment16);
  gtk_container_add (GTK_CONTAINER (okbutton2), alignment16);

  hbox57 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox57);
  gtk_container_add (GTK_CONTAINER (alignment16), hbox57);

  image395 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image395);
  gtk_box_pack_start (GTK_BOX (hbox57), image395, FALSE, FALSE, 0);

  label93 = gtk_label_new_with_mnemonic (_("_OK"));
  gtk_widget_show (label93);
  gtk_box_pack_start (GTK_BOX (hbox57), label93, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (entrydialog, entrydialog, "entrydialog");
  GLADE_HOOKUP_OBJECT_NO_REF (entrydialog, dialog_vbox3, "dialog_vbox3");
  GLADE_HOOKUP_OBJECT (entrydialog, vbox15, "vbox15");
  GLADE_HOOKUP_OBJECT (entrydialog, hbox33, "hbox33");
  GLADE_HOOKUP_OBJECT (entrydialog, title_label, "title_label");
  GLADE_HOOKUP_OBJECT (entrydialog, title, "title");
  GLADE_HOOKUP_OBJECT_NO_REF (entrydialog, dialog_action_area3, "dialog_action_area3");
  GLADE_HOOKUP_OBJECT (entrydialog, cancelbutton2, "cancelbutton2");
  GLADE_HOOKUP_OBJECT (entrydialog, alignment17, "alignment17");
  GLADE_HOOKUP_OBJECT (entrydialog, hbox58, "hbox58");
  GLADE_HOOKUP_OBJECT (entrydialog, image396, "image396");
  GLADE_HOOKUP_OBJECT (entrydialog, label94, "label94");
  GLADE_HOOKUP_OBJECT (entrydialog, okbutton2, "okbutton2");
  GLADE_HOOKUP_OBJECT (entrydialog, alignment16, "alignment16");
  GLADE_HOOKUP_OBJECT (entrydialog, hbox57, "hbox57");
  GLADE_HOOKUP_OBJECT (entrydialog, image395, "image395");
  GLADE_HOOKUP_OBJECT (entrydialog, label93, "label93");

  return entrydialog;
}

GtkWidget*
create_addlocationdlg (void)
{
  GtkWidget *addlocationdlg;
  GtkWidget *dialog_vbox4;
  GtkWidget *vbox45;
  GtkWidget *hbox34;
  GtkWidget *label41;
  GtkWidget *addlocation_entry;
  GtkWidget *hbox122;
  GtkWidget *set_custom_title;
  GtkWidget *custom_title;
  GtkWidget *addlocation_actionarea;
  GtkWidget *cancelbutton3;
  GtkWidget *alignment18;
  GtkWidget *hbox59;
  GtkWidget *image397;
  GtkWidget *label95;
  GtkWidget *okbutton3;
  GtkWidget *alignment19;
  GtkWidget *hbox60;
  GtkWidget *image398;
  GtkWidget *label96;

  addlocationdlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (addlocationdlg), _("Add location"));
  gtk_window_set_type_hint (GTK_WINDOW (addlocationdlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox4 = gtk_dialog_get_content_area (GTK_DIALOG (addlocationdlg));
  gtk_widget_show (dialog_vbox4);

  vbox45 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox45);
  gtk_box_pack_start (GTK_BOX (dialog_vbox4), vbox45, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox45), 12);

  hbox34 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox34);
  gtk_box_pack_start (GTK_BOX (vbox45), hbox34, TRUE, TRUE, 0);

  label41 = gtk_label_new (_("URL:"));
  gtk_widget_show (label41);
  gtk_box_pack_start (GTK_BOX (hbox34), label41, FALSE, FALSE, 0);

  addlocation_entry = gtk_entry_new ();
  gtk_widget_show (addlocation_entry);
  gtk_box_pack_start (GTK_BOX (hbox34), addlocation_entry, TRUE, TRUE, 0);
  gtk_widget_set_size_request (addlocation_entry, 297, -1);
  gtk_entry_set_invisible_char (GTK_ENTRY (addlocation_entry), 8226);
  gtk_entry_set_activates_default (GTK_ENTRY (addlocation_entry), TRUE);

  hbox122 = gtk_hbox_new (FALSE, 8);
  gtk_box_pack_start (GTK_BOX (vbox45), hbox122, TRUE, TRUE, 0);

  set_custom_title = gtk_check_button_new_with_mnemonic (_("Set Custom Title"));
  gtk_widget_show (set_custom_title);
  gtk_box_pack_start (GTK_BOX (hbox122), set_custom_title, FALSE, FALSE, 0);

  custom_title = gtk_entry_new ();
  gtk_widget_show (custom_title);
  gtk_box_pack_start (GTK_BOX (hbox122), custom_title, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (custom_title), 8226);

  addlocation_actionarea = gtk_dialog_get_action_area (GTK_DIALOG (addlocationdlg));
  gtk_widget_show (addlocation_actionarea);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (addlocation_actionarea), GTK_BUTTONBOX_END);

  cancelbutton3 = gtk_button_new ();
  gtk_widget_show (cancelbutton3);
  gtk_dialog_add_action_widget (GTK_DIALOG (addlocationdlg), cancelbutton3, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton3, TRUE);

  alignment18 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment18);
  gtk_container_add (GTK_CONTAINER (cancelbutton3), alignment18);

  hbox59 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox59);
  gtk_container_add (GTK_CONTAINER (alignment18), hbox59);

  image397 = gtk_image_new_from_stock ("gtk-cancel", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image397);
  gtk_box_pack_start (GTK_BOX (hbox59), image397, FALSE, FALSE, 0);

  label95 = gtk_label_new_with_mnemonic (_("_Cancel"));
  gtk_widget_show (label95);
  gtk_box_pack_start (GTK_BOX (hbox59), label95, FALSE, FALSE, 0);

  okbutton3 = gtk_button_new ();
  gtk_widget_show (okbutton3);
  gtk_dialog_add_action_widget (GTK_DIALOG (addlocationdlg), okbutton3, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton3, TRUE);

  alignment19 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment19);
  gtk_container_add (GTK_CONTAINER (okbutton3), alignment19);

  hbox60 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox60);
  gtk_container_add (GTK_CONTAINER (alignment19), hbox60);

  image398 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image398);
  gtk_box_pack_start (GTK_BOX (hbox60), image398, FALSE, FALSE, 0);

  label96 = gtk_label_new_with_mnemonic (_("_OK"));
  gtk_widget_show (label96);
  gtk_box_pack_start (GTK_BOX (hbox60), label96, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (addlocationdlg, addlocationdlg, "addlocationdlg");
  GLADE_HOOKUP_OBJECT_NO_REF (addlocationdlg, dialog_vbox4, "dialog_vbox4");
  GLADE_HOOKUP_OBJECT (addlocationdlg, vbox45, "vbox45");
  GLADE_HOOKUP_OBJECT (addlocationdlg, hbox34, "hbox34");
  GLADE_HOOKUP_OBJECT (addlocationdlg, label41, "label41");
  GLADE_HOOKUP_OBJECT (addlocationdlg, addlocation_entry, "addlocation_entry");
  GLADE_HOOKUP_OBJECT (addlocationdlg, hbox122, "hbox122");
  GLADE_HOOKUP_OBJECT (addlocationdlg, set_custom_title, "set_custom_title");
  GLADE_HOOKUP_OBJECT (addlocationdlg, custom_title, "custom_title");
  GLADE_HOOKUP_OBJECT_NO_REF (addlocationdlg, addlocation_actionarea, "addlocation_actionarea");
  GLADE_HOOKUP_OBJECT (addlocationdlg, cancelbutton3, "cancelbutton3");
  GLADE_HOOKUP_OBJECT (addlocationdlg, alignment18, "alignment18");
  GLADE_HOOKUP_OBJECT (addlocationdlg, hbox59, "hbox59");
  GLADE_HOOKUP_OBJECT (addlocationdlg, image397, "image397");
  GLADE_HOOKUP_OBJECT (addlocationdlg, label95, "label95");
  GLADE_HOOKUP_OBJECT (addlocationdlg, okbutton3, "okbutton3");
  GLADE_HOOKUP_OBJECT (addlocationdlg, alignment19, "alignment19");
  GLADE_HOOKUP_OBJECT (addlocationdlg, hbox60, "hbox60");
  GLADE_HOOKUP_OBJECT (addlocationdlg, image398, "image398");
  GLADE_HOOKUP_OBJECT (addlocationdlg, label96, "label96");

  return addlocationdlg;
}

GtkWidget*
create_groupbydlg (void)
{
  GtkWidget *groupbydlg;
  GtkWidget *dialog_vbox5;
  GtkWidget *vbox25;
  GtkWidget *hbox46;
  GtkWidget *label81;
  GtkWidget *hbox75;
  GtkWidget *format;
  GtkWidget *custom1;
  GtkWidget *dialog_action_area4;
  GtkWidget *cancelbutton4;
  GtkWidget *alignment7;
  GtkWidget *hbox48;
  GtkWidget *image386;
  GtkWidget *label84;
  GtkWidget *okbutton4;
  GtkWidget *alignment6;
  GtkWidget *hbox47;
  GtkWidget *image385;
  GtkWidget *label83;

  groupbydlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (groupbydlg), _("Group By ..."));
  gtk_window_set_type_hint (GTK_WINDOW (groupbydlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox5 = gtk_dialog_get_content_area (GTK_DIALOG (groupbydlg));
  gtk_widget_show (dialog_vbox5);

  vbox25 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox25);
  gtk_box_pack_start (GTK_BOX (dialog_vbox5), vbox25, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox25), 12);

  hbox46 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox46);
  gtk_box_pack_start (GTK_BOX (vbox25), hbox46, FALSE, FALSE, 0);

  label81 = gtk_label_new (_("Format:"));
  gtk_widget_show (label81);
  gtk_box_pack_start (GTK_BOX (hbox46), label81, FALSE, FALSE, 0);

  hbox75 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox75);
  gtk_box_pack_start (GTK_BOX (hbox46), hbox75, TRUE, TRUE, 0);

  format = gtk_entry_new ();
  gtk_widget_show (format);
  gtk_box_pack_start (GTK_BOX (hbox75), format, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (format), 9679);
  gtk_entry_set_activates_default (GTK_ENTRY (format), TRUE);

  custom1 = title_formatting_help_link_create ("custom1", "", "", 0, 0);
  gtk_widget_show (custom1);
  gtk_box_pack_start (GTK_BOX (hbox75), custom1, FALSE, TRUE, 0);
  gtk_widget_set_can_focus(custom1, FALSE);
  gtk_widget_set_can_default(custom1, FALSE);

  dialog_action_area4 = gtk_dialog_get_action_area (GTK_DIALOG (groupbydlg));
  gtk_widget_show (dialog_action_area4);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area4), GTK_BUTTONBOX_END);

  cancelbutton4 = gtk_button_new ();
  gtk_widget_show (cancelbutton4);
  gtk_dialog_add_action_widget (GTK_DIALOG (groupbydlg), cancelbutton4, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton4, TRUE);

  alignment7 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment7);
  gtk_container_add (GTK_CONTAINER (cancelbutton4), alignment7);

  hbox48 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox48);
  gtk_container_add (GTK_CONTAINER (alignment7), hbox48);

  image386 = gtk_image_new_from_stock ("gtk-cancel", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image386);
  gtk_box_pack_start (GTK_BOX (hbox48), image386, FALSE, FALSE, 0);

  label84 = gtk_label_new_with_mnemonic (_("_Cancel"));
  gtk_widget_show (label84);
  gtk_box_pack_start (GTK_BOX (hbox48), label84, FALSE, FALSE, 0);

  okbutton4 = gtk_button_new ();
  gtk_widget_show (okbutton4);
  gtk_dialog_add_action_widget (GTK_DIALOG (groupbydlg), okbutton4, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton4, TRUE);

  alignment6 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment6);
  gtk_container_add (GTK_CONTAINER (okbutton4), alignment6);

  hbox47 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox47);
  gtk_container_add (GTK_CONTAINER (alignment6), hbox47);

  image385 = gtk_image_new_from_stock ("gtk-ok", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image385);
  gtk_box_pack_start (GTK_BOX (hbox47), image385, FALSE, FALSE, 0);

  label83 = gtk_label_new_with_mnemonic (_("_OK"));
  gtk_widget_show (label83);
  gtk_box_pack_start (GTK_BOX (hbox47), label83, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (groupbydlg, groupbydlg, "groupbydlg");
  GLADE_HOOKUP_OBJECT_NO_REF (groupbydlg, dialog_vbox5, "dialog_vbox5");
  GLADE_HOOKUP_OBJECT (groupbydlg, vbox25, "vbox25");
  GLADE_HOOKUP_OBJECT (groupbydlg, hbox46, "hbox46");
  GLADE_HOOKUP_OBJECT (groupbydlg, label81, "label81");
  GLADE_HOOKUP_OBJECT (groupbydlg, hbox75, "hbox75");
  GLADE_HOOKUP_OBJECT (groupbydlg, format, "format");
  GLADE_HOOKUP_OBJECT (groupbydlg, custom1, "custom1");
  GLADE_HOOKUP_OBJECT_NO_REF (groupbydlg, dialog_action_area4, "dialog_action_area4");
  GLADE_HOOKUP_OBJECT (groupbydlg, cancelbutton4, "cancelbutton4");
  GLADE_HOOKUP_OBJECT (groupbydlg, alignment7, "alignment7");
  GLADE_HOOKUP_OBJECT (groupbydlg, hbox48, "hbox48");
  GLADE_HOOKUP_OBJECT (groupbydlg, image386, "image386");
  GLADE_HOOKUP_OBJECT (groupbydlg, label84, "label84");
  GLADE_HOOKUP_OBJECT (groupbydlg, okbutton4, "okbutton4");
  GLADE_HOOKUP_OBJECT (groupbydlg, alignment6, "alignment6");
  GLADE_HOOKUP_OBJECT (groupbydlg, hbox47, "hbox47");
  GLADE_HOOKUP_OBJECT (groupbydlg, image385, "image385");
  GLADE_HOOKUP_OBJECT (groupbydlg, label83, "label83");

  return groupbydlg;
}

GtkWidget*
create_sortbydlg (void)
{
  GtkWidget *sortbydlg;
  GtkWidget *dialog_vbox8;
  GtkWidget *sortvbox;
  GtkWidget *sortframe;
  GtkWidget *sortfmtalignment;
  GtkWidget *sortfmt;
  GtkWidget *label188;
  GtkWidget *sortbtnbox;
  GtkWidget *sorthelpalignment;
  GtkWidget *sorthelplink;
  GtkWidget *sortorder;
  GtkWidget *sortcancel;
  GtkWidget *sortok;
  GtkWidget *dialog_action_area7;

  sortbydlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (sortbydlg), _("Sort by ..."));
  gtk_window_set_modal (GTK_WINDOW (sortbydlg), TRUE);
  gtk_window_set_default_size (GTK_WINDOW (sortbydlg), 540, 180);
  gtk_window_set_type_hint (GTK_WINDOW (sortbydlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox8 = gtk_dialog_get_content_area (GTK_DIALOG (sortbydlg));
  gtk_widget_show (dialog_vbox8);

  sortvbox = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (sortvbox);
  gtk_box_pack_start (GTK_BOX (dialog_vbox8), sortvbox, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (sortvbox), 12);

  sortframe = gtk_frame_new (NULL);
  gtk_widget_show (sortframe);
  gtk_box_pack_start (GTK_BOX (sortvbox), sortframe, TRUE, TRUE, 0);
  gtk_frame_set_label_align (GTK_FRAME (sortframe), 0.02, 0.5);
  gtk_frame_set_shadow_type (GTK_FRAME (sortframe), GTK_SHADOW_IN);

  sortfmtalignment = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (sortfmtalignment);
  gtk_container_add (GTK_CONTAINER (sortframe), sortfmtalignment);
  gtk_alignment_set_padding (GTK_ALIGNMENT (sortfmtalignment), 8, 8, 8, 12);

  sortfmt = gtk_text_view_new ();
  gtk_widget_show (sortfmt);
  gtk_container_add (GTK_CONTAINER (sortfmtalignment), sortfmt);
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW (sortfmt), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (sortfmt), GTK_WRAP_WORD);

  label188 = gtk_label_new (_("<b>Format</b>"));
  gtk_widget_show (label188);
  gtk_frame_set_label_widget (GTK_FRAME (sortframe), label188);
  gtk_label_set_use_markup (GTK_LABEL (label188), TRUE);

  sortbtnbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (sortbtnbox);
  gtk_box_pack_start (GTK_BOX (sortvbox), sortbtnbox, FALSE, FALSE, 0);

  sorthelpalignment = gtk_alignment_new (0, 0.5, 0, 1);
  gtk_widget_show (sorthelpalignment);
  gtk_box_pack_start (GTK_BOX (sortbtnbox), sorthelpalignment, TRUE, TRUE, 8);

  sorthelplink = title_formatting_help_link_create ("sorthelplink", "", "", 0, 0);
  gtk_widget_show (sorthelplink);
  gtk_container_add (GTK_CONTAINER (sorthelpalignment), sorthelplink);
  gtk_widget_set_can_focus(sorthelplink, FALSE);
  gtk_widget_set_can_default(sorthelplink, FALSE);

  sortorder = gtk_combo_box_text_new ();
  gtk_widget_show (sortorder);
  gtk_box_pack_start (GTK_BOX (sortbtnbox), sortorder, FALSE, FALSE, 8);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (sortorder), _("Ascending"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (sortorder), _("Descending"));

  sortcancel = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (sortcancel);
  gtk_box_pack_start (GTK_BOX (sortbtnbox), sortcancel, FALSE, FALSE, 8);
  gtk_widget_set_can_default(sortcancel, TRUE);

  sortok = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (sortok);
  gtk_box_pack_start (GTK_BOX (sortbtnbox), sortok, FALSE, FALSE, 8);
  gtk_widget_set_can_default(sortok, TRUE);

  dialog_action_area7 = gtk_dialog_get_action_area (GTK_DIALOG (sortbydlg));
  gtk_widget_show (dialog_action_area7);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area7), GTK_BUTTONBOX_END);

  g_signal_connect ((gpointer) sortfmt, "show",
                    G_CALLBACK (on_sortfmt_show),
                    NULL);
  g_signal_connect ((gpointer) sortcancel, "clicked",
                    G_CALLBACK (on_sortcancel_clicked),
                    NULL);
  g_signal_connect ((gpointer) sortok, "clicked",
                    G_CALLBACK (on_sortok_clicked),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (sortbydlg, sortbydlg, "sortbydlg");
  GLADE_HOOKUP_OBJECT_NO_REF (sortbydlg, dialog_vbox8, "dialog_vbox8");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortvbox, "sortvbox");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortframe, "sortframe");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortfmtalignment, "sortfmtalignment");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortfmt, "sortfmt");
  GLADE_HOOKUP_OBJECT (sortbydlg, label188, "label188");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortbtnbox, "sortbtnbox");
  GLADE_HOOKUP_OBJECT (sortbydlg, sorthelpalignment, "sorthelpalignment");
  GLADE_HOOKUP_OBJECT (sortbydlg, sorthelplink, "sorthelplink");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortorder, "sortorder");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortcancel, "sortcancel");
  GLADE_HOOKUP_OBJECT (sortbydlg, sortok, "sortok");
  GLADE_HOOKUP_OBJECT_NO_REF (sortbydlg, dialog_action_area7, "dialog_action_area7");

  return sortbydlg;
}

GtkWidget*
create_select_dsp_plugin (void)
{
  GtkWidget *select_dsp_plugin;
  GtkWidget *dialog_vbox10;
  GtkWidget *vbox31;
  GtkWidget *hbox85;
  GtkWidget *label113;
  GtkWidget *plugin;
  GtkWidget *dialog_action_area9;
  GtkWidget *cancelbutton7;
  GtkWidget *okbutton7;

  select_dsp_plugin = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (select_dsp_plugin), _("Select DSP Plugin"));
  gtk_window_set_modal (GTK_WINDOW (select_dsp_plugin), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (select_dsp_plugin), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox10 = gtk_dialog_get_content_area (GTK_DIALOG (select_dsp_plugin));
  gtk_widget_show (dialog_vbox10);

  vbox31 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox31);
  gtk_box_pack_start (GTK_BOX (dialog_vbox10), vbox31, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox31), 12);

  hbox85 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox85);
  gtk_box_pack_start (GTK_BOX (vbox31), hbox85, FALSE, FALSE, 0);

  label113 = gtk_label_new (_("Plugin"));
  gtk_widget_show (label113);
  gtk_box_pack_start (GTK_BOX (hbox85), label113, FALSE, FALSE, 0);

  plugin = gtk_combo_box_text_new ();
  gtk_widget_show (plugin);
  gtk_box_pack_start (GTK_BOX (hbox85), plugin, TRUE, TRUE, 0);

  dialog_action_area9 = gtk_dialog_get_action_area (GTK_DIALOG (select_dsp_plugin));
  gtk_widget_show (dialog_action_area9);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area9), GTK_BUTTONBOX_END);

  cancelbutton7 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton7);
  gtk_dialog_add_action_widget (GTK_DIALOG (select_dsp_plugin), cancelbutton7, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton7, TRUE);

  okbutton7 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton7);
  gtk_dialog_add_action_widget (GTK_DIALOG (select_dsp_plugin), okbutton7, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton7, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (select_dsp_plugin, select_dsp_plugin, "select_dsp_plugin");
  GLADE_HOOKUP_OBJECT_NO_REF (select_dsp_plugin, dialog_vbox10, "dialog_vbox10");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, vbox31, "vbox31");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, hbox85, "hbox85");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, label113, "label113");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, plugin, "plugin");
  GLADE_HOOKUP_OBJECT_NO_REF (select_dsp_plugin, dialog_action_area9, "dialog_action_area9");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, cancelbutton7, "cancelbutton7");
  GLADE_HOOKUP_OBJECT (select_dsp_plugin, okbutton7, "okbutton7");

  return select_dsp_plugin;
}

GtkWidget*
create_tagwritersettings (void)
{
  GtkWidget *tagwritersettings;
  GtkWidget *dialog_vbox11;
  GtkWidget *vbox32;
  GtkWidget *frame8;
  GtkWidget *alignment21;
  GtkWidget *vbox33;
  GtkWidget *hbox89;
  GtkWidget *write_id3v2;
  GtkWidget *write_id3v1;
  GtkWidget *write_apev2;
  GtkWidget *hbox90;
  GtkWidget *strip_id3v2;
  GtkWidget *strip_id3v1;
  GtkWidget *strip_apev2;
  GtkWidget *hbox91;
  GtkWidget *label118;
  GtkWidget *id3v2_version;
  GtkWidget *hbox92;
  GtkWidget *label119;
  GtkWidget *id3v1_encoding;
  GtkWidget *label120;
  GtkWidget *hbox93;
  GtkWidget *frame9;
  GtkWidget *alignment22;
  GtkWidget *vbox34;
  GtkWidget *hbox94;
  GtkWidget *ape_write_id3v2;
  GtkWidget *ape_write_apev2;
  GtkWidget *hbox95;
  GtkWidget *ape_strip_id3v2;
  GtkWidget *ape_strip_apev2;
  GtkWidget *label121;
  GtkWidget *frame10;
  GtkWidget *alignment23;
  GtkWidget *vbox35;
  GtkWidget *hbox96;
  GtkWidget *wv_write_apev2;
  GtkWidget *wv_write_id3v1;
  GtkWidget *hbox97;
  GtkWidget *wv_strip_apev2;
  GtkWidget *wv_strip_id3v1;
  GtkWidget *label122;
  GtkWidget *dialog_action_area10;
  GtkWidget *closebutton2;

  tagwritersettings = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (tagwritersettings), _("Tag Writer Settings"));
  gtk_window_set_modal (GTK_WINDOW (tagwritersettings), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (tagwritersettings), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox11 = gtk_dialog_get_content_area (GTK_DIALOG (tagwritersettings));
  gtk_widget_show (dialog_vbox11);

  vbox32 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox32);
  gtk_box_pack_start (GTK_BOX (dialog_vbox11), vbox32, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox32), 12);

  frame8 = gtk_frame_new (NULL);
  gtk_widget_show (frame8);
  gtk_box_pack_start (GTK_BOX (vbox32), frame8, FALSE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame8), GTK_SHADOW_NONE);

  alignment21 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment21);
  gtk_container_add (GTK_CONTAINER (frame8), alignment21);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment21), 0, 0, 12, 0);

  vbox33 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox33);
  gtk_container_add (GTK_CONTAINER (alignment21), vbox33);
  gtk_container_set_border_width (GTK_CONTAINER (vbox33), 12);

  hbox89 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox89);
  gtk_box_pack_start (GTK_BOX (vbox33), hbox89, FALSE, FALSE, 0);

  write_id3v2 = gtk_check_button_new_with_mnemonic (_("Write ID3v2"));
  gtk_widget_show (write_id3v2);
  gtk_box_pack_start (GTK_BOX (hbox89), write_id3v2, FALSE, FALSE, 0);

  write_id3v1 = gtk_check_button_new_with_mnemonic (_("Write ID3v1"));
  gtk_widget_show (write_id3v1);
  gtk_box_pack_start (GTK_BOX (hbox89), write_id3v1, FALSE, FALSE, 0);

  write_apev2 = gtk_check_button_new_with_mnemonic (_("Write APEv2"));
  gtk_widget_show (write_apev2);
  gtk_box_pack_start (GTK_BOX (hbox89), write_apev2, FALSE, FALSE, 0);

  hbox90 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox90);
  gtk_box_pack_start (GTK_BOX (vbox33), hbox90, FALSE, FALSE, 0);

  strip_id3v2 = gtk_check_button_new_with_mnemonic (_("Strip ID3v2"));
  gtk_widget_show (strip_id3v2);
  gtk_box_pack_start (GTK_BOX (hbox90), strip_id3v2, FALSE, FALSE, 0);

  strip_id3v1 = gtk_check_button_new_with_mnemonic (_("Strip ID3v1"));
  gtk_widget_show (strip_id3v1);
  gtk_box_pack_start (GTK_BOX (hbox90), strip_id3v1, FALSE, FALSE, 0);

  strip_apev2 = gtk_check_button_new_with_mnemonic (_("Strip APEv2"));
  gtk_widget_show (strip_apev2);
  gtk_box_pack_start (GTK_BOX (hbox90), strip_apev2, FALSE, FALSE, 0);

  hbox91 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox91);
  gtk_box_pack_start (GTK_BOX (vbox33), hbox91, TRUE, TRUE, 0);

  label118 = gtk_label_new (_("ID3v2 version"));
  gtk_widget_show (label118);
  gtk_box_pack_start (GTK_BOX (hbox91), label118, FALSE, FALSE, 0);

  id3v2_version = gtk_combo_box_text_new ();
  gtk_widget_show (id3v2_version);
  gtk_box_pack_start (GTK_BOX (hbox91), id3v2_version, TRUE, TRUE, 0);
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (id3v2_version), _("2.3 (Recommended)"));
  gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (id3v2_version), _("2.4"));

  hbox92 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox92);
  gtk_box_pack_start (GTK_BOX (vbox33), hbox92, TRUE, TRUE, 0);

  label119 = gtk_label_new (_("ID3v1 character encoding (default is iso8859-1)"));
  gtk_widget_show (label119);
  gtk_box_pack_start (GTK_BOX (hbox92), label119, FALSE, FALSE, 0);

  id3v1_encoding = gtk_entry_new ();
  gtk_widget_show (id3v1_encoding);
  gtk_box_pack_start (GTK_BOX (hbox92), id3v1_encoding, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (id3v1_encoding), 9679);

  label120 = gtk_label_new ("<b>MP3</b>");
  gtk_widget_show (label120);
  gtk_frame_set_label_widget (GTK_FRAME (frame8), label120);
  gtk_label_set_use_markup (GTK_LABEL (label120), TRUE);

  hbox93 = gtk_hbox_new (TRUE, 0);
  gtk_widget_show (hbox93);
  gtk_box_pack_start (GTK_BOX (vbox32), hbox93, FALSE, TRUE, 0);

  frame9 = gtk_frame_new (NULL);
  gtk_widget_show (frame9);
  gtk_box_pack_start (GTK_BOX (hbox93), frame9, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame9), GTK_SHADOW_NONE);

  alignment22 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment22);
  gtk_container_add (GTK_CONTAINER (frame9), alignment22);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment22), 0, 0, 12, 0);

  vbox34 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox34);
  gtk_container_add (GTK_CONTAINER (alignment22), vbox34);
  gtk_container_set_border_width (GTK_CONTAINER (vbox34), 12);

  hbox94 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox94);
  gtk_box_pack_start (GTK_BOX (vbox34), hbox94, TRUE, TRUE, 0);

  ape_write_id3v2 = gtk_check_button_new_with_mnemonic (_("Write ID3v2.4"));
  gtk_widget_show (ape_write_id3v2);
  gtk_box_pack_start (GTK_BOX (hbox94), ape_write_id3v2, FALSE, FALSE, 0);

  ape_write_apev2 = gtk_check_button_new_with_mnemonic (_("Write APEv2"));
  gtk_widget_show (ape_write_apev2);
  gtk_box_pack_start (GTK_BOX (hbox94), ape_write_apev2, FALSE, FALSE, 0);

  hbox95 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox95);
  gtk_box_pack_start (GTK_BOX (vbox34), hbox95, TRUE, TRUE, 0);

  ape_strip_id3v2 = gtk_check_button_new_with_mnemonic (_("Strip ID3v2"));
  gtk_widget_show (ape_strip_id3v2);
  gtk_box_pack_start (GTK_BOX (hbox95), ape_strip_id3v2, FALSE, FALSE, 0);

  ape_strip_apev2 = gtk_check_button_new_with_mnemonic (_("Strip APEv2"));
  gtk_widget_show (ape_strip_apev2);
  gtk_box_pack_start (GTK_BOX (hbox95), ape_strip_apev2, FALSE, FALSE, 0);

  label121 = gtk_label_new ("<b>APE</b>");
  gtk_widget_show (label121);
  gtk_frame_set_label_widget (GTK_FRAME (frame9), label121);
  gtk_label_set_use_markup (GTK_LABEL (label121), TRUE);

  frame10 = gtk_frame_new (NULL);
  gtk_widget_show (frame10);
  gtk_box_pack_start (GTK_BOX (hbox93), frame10, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame10), GTK_SHADOW_NONE);

  alignment23 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_show (alignment23);
  gtk_container_add (GTK_CONTAINER (frame10), alignment23);
  gtk_alignment_set_padding (GTK_ALIGNMENT (alignment23), 0, 0, 12, 0);

  vbox35 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox35);
  gtk_container_add (GTK_CONTAINER (alignment23), vbox35);
  gtk_container_set_border_width (GTK_CONTAINER (vbox35), 12);

  hbox96 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox96);
  gtk_box_pack_start (GTK_BOX (vbox35), hbox96, FALSE, FALSE, 0);

  wv_write_apev2 = gtk_check_button_new_with_mnemonic (_("Write APEv2"));
  gtk_widget_show (wv_write_apev2);
  gtk_box_pack_start (GTK_BOX (hbox96), wv_write_apev2, FALSE, FALSE, 0);

  wv_write_id3v1 = gtk_check_button_new_with_mnemonic (_("Write ID3v1"));
  gtk_widget_show (wv_write_id3v1);
  gtk_box_pack_start (GTK_BOX (hbox96), wv_write_id3v1, FALSE, FALSE, 0);

  hbox97 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox97);
  gtk_box_pack_start (GTK_BOX (vbox35), hbox97, FALSE, FALSE, 0);

  wv_strip_apev2 = gtk_check_button_new_with_mnemonic (_("Strip APEv2"));
  gtk_widget_show (wv_strip_apev2);
  gtk_box_pack_start (GTK_BOX (hbox97), wv_strip_apev2, FALSE, FALSE, 0);

  wv_strip_id3v1 = gtk_check_button_new_with_mnemonic (_("Strip ID3v1"));
  gtk_widget_show (wv_strip_id3v1);
  gtk_box_pack_start (GTK_BOX (hbox97), wv_strip_id3v1, FALSE, FALSE, 0);

  label122 = gtk_label_new ("<b>WavPack</b>");
  gtk_widget_show (label122);
  gtk_frame_set_label_widget (GTK_FRAME (frame10), label122);
  gtk_label_set_use_markup (GTK_LABEL (label122), TRUE);

  dialog_action_area10 = gtk_dialog_get_action_area (GTK_DIALOG (tagwritersettings));
  gtk_widget_show (dialog_action_area10);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area10), GTK_BUTTONBOX_END);

  closebutton2 = gtk_button_new_from_stock ("gtk-close");
  gtk_widget_show (closebutton2);
  gtk_dialog_add_action_widget (GTK_DIALOG (tagwritersettings), closebutton2, GTK_RESPONSE_CLOSE);
  gtk_widget_set_can_default(closebutton2, TRUE);

  g_signal_connect ((gpointer) write_id3v2, "toggled",
                    G_CALLBACK (on_write_id3v2_toggled),
                    NULL);
  g_signal_connect ((gpointer) write_id3v1, "toggled",
                    G_CALLBACK (on_write_id3v1_toggled),
                    NULL);
  g_signal_connect ((gpointer) write_apev2, "toggled",
                    G_CALLBACK (on_write_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) strip_id3v2, "toggled",
                    G_CALLBACK (on_strip_id3v2_toggled),
                    NULL);
  g_signal_connect ((gpointer) strip_id3v1, "toggled",
                    G_CALLBACK (on_strip_id3v1_toggled),
                    NULL);
  g_signal_connect ((gpointer) strip_apev2, "toggled",
                    G_CALLBACK (on_strip_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) id3v2_version, "changed",
                    G_CALLBACK (on_id3v2_version_changed),
                    NULL);
  g_signal_connect ((gpointer) id3v1_encoding, "changed",
                    G_CALLBACK (on_id3v1_encoding_changed),
                    NULL);
  g_signal_connect ((gpointer) ape_write_id3v2, "toggled",
                    G_CALLBACK (on_ape_write_id3v2_toggled),
                    NULL);
  g_signal_connect ((gpointer) ape_write_apev2, "toggled",
                    G_CALLBACK (on_ape_write_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) ape_strip_id3v2, "toggled",
                    G_CALLBACK (on_ape_strip_id3v2_toggled),
                    NULL);
  g_signal_connect ((gpointer) ape_strip_apev2, "toggled",
                    G_CALLBACK (on_ape_strip_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) wv_write_apev2, "toggled",
                    G_CALLBACK (on_wv_write_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) wv_write_id3v1, "toggled",
                    G_CALLBACK (on_wv_write_id3v1_toggled),
                    NULL);
  g_signal_connect ((gpointer) wv_strip_apev2, "toggled",
                    G_CALLBACK (on_wv_strip_apev2_toggled),
                    NULL);
  g_signal_connect ((gpointer) wv_strip_id3v1, "toggled",
                    G_CALLBACK (on_wv_strip_id3v1_toggled),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (tagwritersettings, tagwritersettings, "tagwritersettings");
  GLADE_HOOKUP_OBJECT_NO_REF (tagwritersettings, dialog_vbox11, "dialog_vbox11");
  GLADE_HOOKUP_OBJECT (tagwritersettings, vbox32, "vbox32");
  GLADE_HOOKUP_OBJECT (tagwritersettings, frame8, "frame8");
  GLADE_HOOKUP_OBJECT (tagwritersettings, alignment21, "alignment21");
  GLADE_HOOKUP_OBJECT (tagwritersettings, vbox33, "vbox33");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox89, "hbox89");
  GLADE_HOOKUP_OBJECT (tagwritersettings, write_id3v2, "write_id3v2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, write_id3v1, "write_id3v1");
  GLADE_HOOKUP_OBJECT (tagwritersettings, write_apev2, "write_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox90, "hbox90");
  GLADE_HOOKUP_OBJECT (tagwritersettings, strip_id3v2, "strip_id3v2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, strip_id3v1, "strip_id3v1");
  GLADE_HOOKUP_OBJECT (tagwritersettings, strip_apev2, "strip_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox91, "hbox91");
  GLADE_HOOKUP_OBJECT (tagwritersettings, label118, "label118");
  GLADE_HOOKUP_OBJECT (tagwritersettings, id3v2_version, "id3v2_version");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox92, "hbox92");
  GLADE_HOOKUP_OBJECT (tagwritersettings, label119, "label119");
  GLADE_HOOKUP_OBJECT (tagwritersettings, id3v1_encoding, "id3v1_encoding");
  GLADE_HOOKUP_OBJECT (tagwritersettings, label120, "label120");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox93, "hbox93");
  GLADE_HOOKUP_OBJECT (tagwritersettings, frame9, "frame9");
  GLADE_HOOKUP_OBJECT (tagwritersettings, alignment22, "alignment22");
  GLADE_HOOKUP_OBJECT (tagwritersettings, vbox34, "vbox34");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox94, "hbox94");
  GLADE_HOOKUP_OBJECT (tagwritersettings, ape_write_id3v2, "ape_write_id3v2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, ape_write_apev2, "ape_write_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox95, "hbox95");
  GLADE_HOOKUP_OBJECT (tagwritersettings, ape_strip_id3v2, "ape_strip_id3v2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, ape_strip_apev2, "ape_strip_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, label121, "label121");
  GLADE_HOOKUP_OBJECT (tagwritersettings, frame10, "frame10");
  GLADE_HOOKUP_OBJECT (tagwritersettings, alignment23, "alignment23");
  GLADE_HOOKUP_OBJECT (tagwritersettings, vbox35, "vbox35");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox96, "hbox96");
  GLADE_HOOKUP_OBJECT (tagwritersettings, wv_write_apev2, "wv_write_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, wv_write_id3v1, "wv_write_id3v1");
  GLADE_HOOKUP_OBJECT (tagwritersettings, hbox97, "hbox97");
  GLADE_HOOKUP_OBJECT (tagwritersettings, wv_strip_apev2, "wv_strip_apev2");
  GLADE_HOOKUP_OBJECT (tagwritersettings, wv_strip_id3v1, "wv_strip_id3v1");
  GLADE_HOOKUP_OBJECT (tagwritersettings, label122, "label122");
  GLADE_HOOKUP_OBJECT_NO_REF (tagwritersettings, dialog_action_area10, "dialog_action_area10");
  GLADE_HOOKUP_OBJECT (tagwritersettings, closebutton2, "closebutton2");

  return tagwritersettings;
}

GtkWidget*
create_ctmappingdlg (void)
{
  GtkWidget *ctmappingdlg;
  GtkWidget *dialog_vbox12;
  GtkWidget *vbox41;
  GtkWidget *textview1;
  GtkWidget *hbox110;
  GtkWidget *scrolledwindow11;
  GtkWidget *ctmappinglist;
  GtkWidget *vbox39;
  GtkWidget *ctmapping_add;
  GtkWidget *ctmapping_remove;
  GtkWidget *ctmapping_edit;
  GtkWidget *ctmapping_reset;
  GtkWidget *alignment25;
  GtkWidget *hbox113;
  GtkWidget *image628;
  GtkWidget *label139;
  GtkWidget *dialog_action_area11;
  GtkWidget *ctmapping_apply;
  GtkWidget *ctmapping_cancel;
  GtkWidget *ctmapping_ok;

  ctmappingdlg = gtk_dialog_new ();
  gtk_widget_set_size_request (ctmappingdlg, 492, 328);
  gtk_window_set_title (GTK_WINDOW (ctmappingdlg), _("Content-Type Mapping"));
  gtk_window_set_position (GTK_WINDOW (ctmappingdlg), GTK_WIN_POS_MOUSE);
  gtk_window_set_modal (GTK_WINDOW (ctmappingdlg), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (ctmappingdlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox12 = gtk_dialog_get_content_area (GTK_DIALOG (ctmappingdlg));
  gtk_widget_show (dialog_vbox12);

  vbox41 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox41);
  gtk_box_pack_start (GTK_BOX (dialog_vbox12), vbox41, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox41), 12);

  textview1 = gtk_text_view_new ();
  gtk_widget_show (textview1);
  gtk_box_pack_start (GTK_BOX (vbox41), textview1, FALSE, TRUE, 0);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview1), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview1), GTK_WRAP_WORD);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview1), FALSE);
  gtk_text_view_set_pixels_above_lines (GTK_TEXT_VIEW (textview1), 8);
  gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW (textview1), 8);
  gtk_text_view_set_left_margin (GTK_TEXT_VIEW (textview1), 8);
  gtk_text_view_set_right_margin (GTK_TEXT_VIEW (textview1), 8);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview1)), _("This table defines the binding between network stream content types and DeaDBeeF decoder plugins. For example, mp3 files can have content type \"audio/x-mpeg\", and need to be decoded by DeaDBeeF's own \"stdmpg\" plugin, or \"ffmpeg\" plugin."), -1);

  hbox110 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox110);
  gtk_box_pack_start (GTK_BOX (vbox41), hbox110, TRUE, TRUE, 0);

  scrolledwindow11 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow11);
  gtk_box_pack_start (GTK_BOX (hbox110), scrolledwindow11, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow11), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow11), GTK_SHADOW_IN);

  ctmappinglist = gtk_tree_view_new ();
  gtk_widget_show (ctmappinglist);
  gtk_container_add (GTK_CONTAINER (scrolledwindow11), ctmappinglist);

  vbox39 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox39);
  gtk_box_pack_start (GTK_BOX (hbox110), vbox39, FALSE, TRUE, 0);

  ctmapping_add = gtk_button_new_from_stock ("gtk-add");
  gtk_widget_show (ctmapping_add);
  gtk_box_pack_start (GTK_BOX (vbox39), ctmapping_add, FALSE, FALSE, 0);

  ctmapping_remove = gtk_button_new_from_stock ("gtk-remove");
  gtk_widget_show (ctmapping_remove);
  gtk_box_pack_start (GTK_BOX (vbox39), ctmapping_remove, FALSE, FALSE, 0);

  ctmapping_edit = gtk_button_new_from_stock ("gtk-edit");
  gtk_widget_show (ctmapping_edit);
  gtk_box_pack_start (GTK_BOX (vbox39), ctmapping_edit, FALSE, FALSE, 0);

  ctmapping_reset = gtk_button_new ();
  gtk_widget_show (ctmapping_reset);
  gtk_box_pack_start (GTK_BOX (vbox39), ctmapping_reset, FALSE, FALSE, 0);

  alignment25 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment25);
  gtk_container_add (GTK_CONTAINER (ctmapping_reset), alignment25);

  hbox113 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox113);
  gtk_container_add (GTK_CONTAINER (alignment25), hbox113);

  image628 = gtk_image_new_from_stock ("gtk-revert-to-saved", GTK_ICON_SIZE_BUTTON);
  gtk_widget_show (image628);
  gtk_box_pack_start (GTK_BOX (hbox113), image628, FALSE, FALSE, 0);

  label139 = gtk_label_new_with_mnemonic (_("Reset to defaults"));
  gtk_widget_show (label139);
  gtk_box_pack_start (GTK_BOX (hbox113), label139, FALSE, FALSE, 0);

  dialog_action_area11 = gtk_dialog_get_action_area (GTK_DIALOG (ctmappingdlg));
  gtk_widget_show (dialog_action_area11);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area11), GTK_BUTTONBOX_END);

  ctmapping_apply = gtk_button_new_from_stock ("gtk-apply");
  gtk_widget_show (ctmapping_apply);
  gtk_dialog_add_action_widget (GTK_DIALOG (ctmappingdlg), ctmapping_apply, GTK_RESPONSE_APPLY);
  gtk_widget_set_can_default(ctmapping_apply, TRUE);

  ctmapping_cancel = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (ctmapping_cancel);
  gtk_dialog_add_action_widget (GTK_DIALOG (ctmappingdlg), ctmapping_cancel, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(ctmapping_cancel, TRUE);

  ctmapping_ok = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (ctmapping_ok);
  gtk_dialog_add_action_widget (GTK_DIALOG (ctmappingdlg), ctmapping_ok, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(ctmapping_ok, TRUE);

  g_signal_connect ((gpointer) ctmapping_add, "clicked",
                    G_CALLBACK (on_ctmapping_add_clicked),
                    NULL);
  g_signal_connect ((gpointer) ctmapping_remove, "clicked",
                    G_CALLBACK (on_ctmapping_remove_clicked),
                    NULL);
  g_signal_connect ((gpointer) ctmapping_edit, "clicked",
                    G_CALLBACK (on_ctmapping_edit_clicked),
                    NULL);
  g_signal_connect ((gpointer) ctmapping_reset, "clicked",
                    G_CALLBACK (on_ctmapping_reset_clicked),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingdlg, ctmappingdlg, "ctmappingdlg");
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingdlg, dialog_vbox12, "dialog_vbox12");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, vbox41, "vbox41");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, textview1, "textview1");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, hbox110, "hbox110");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, scrolledwindow11, "scrolledwindow11");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmappinglist, "ctmappinglist");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, vbox39, "vbox39");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_add, "ctmapping_add");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_remove, "ctmapping_remove");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_edit, "ctmapping_edit");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_reset, "ctmapping_reset");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, alignment25, "alignment25");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, hbox113, "hbox113");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, image628, "image628");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, label139, "label139");
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingdlg, dialog_action_area11, "dialog_action_area11");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_apply, "ctmapping_apply");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_cancel, "ctmapping_cancel");
  GLADE_HOOKUP_OBJECT (ctmappingdlg, ctmapping_ok, "ctmapping_ok");

  return ctmappingdlg;
}

GtkWidget*
create_ctmappingeditdlg (void)
{
  GtkWidget *ctmappingeditdlg;
  GtkWidget *dialog_vbox13;
  GtkWidget *vbox40;
  GtkWidget *hbox111;
  GtkWidget *label137;
  GtkWidget *content_type;
  GtkWidget *hbox112;
  GtkWidget *label138;
  GtkWidget *plugins;
  GtkWidget *dialog_action_area12;
  GtkWidget *cancelbutton8;
  GtkWidget *okbutton8;

  ctmappingeditdlg = gtk_dialog_new ();
  gtk_widget_set_size_request (ctmappingeditdlg, 300, 140);
  gtk_window_set_title (GTK_WINDOW (ctmappingeditdlg), _("Edit Content-Type Mapping"));
  gtk_window_set_modal (GTK_WINDOW (ctmappingeditdlg), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (ctmappingeditdlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox13 = gtk_dialog_get_content_area (GTK_DIALOG (ctmappingeditdlg));
  gtk_widget_show (dialog_vbox13);

  vbox40 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox40);
  gtk_box_pack_start (GTK_BOX (dialog_vbox13), vbox40, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox40), 12);

  hbox111 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox111);
  gtk_box_pack_start (GTK_BOX (vbox40), hbox111, FALSE, TRUE, 0);

  label137 = gtk_label_new (_("Content-Type:"));
  gtk_widget_show (label137);
  gtk_box_pack_start (GTK_BOX (hbox111), label137, FALSE, FALSE, 0);

  content_type = gtk_entry_new ();
  gtk_widget_show (content_type);
  gtk_box_pack_start (GTK_BOX (hbox111), content_type, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (content_type), 8226);

  hbox112 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox112);
  gtk_box_pack_start (GTK_BOX (vbox40), hbox112, FALSE, TRUE, 0);

  label138 = gtk_label_new (_("Plugins:"));
  gtk_widget_show (label138);
  gtk_box_pack_start (GTK_BOX (hbox112), label138, FALSE, FALSE, 0);

  plugins = gtk_entry_new ();
  gtk_widget_show (plugins);
  gtk_box_pack_start (GTK_BOX (hbox112), plugins, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (plugins), 8226);

  dialog_action_area12 = gtk_dialog_get_action_area (GTK_DIALOG (ctmappingeditdlg));
  gtk_widget_show (dialog_action_area12);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area12), GTK_BUTTONBOX_END);

  cancelbutton8 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton8);
  gtk_dialog_add_action_widget (GTK_DIALOG (ctmappingeditdlg), cancelbutton8, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton8, TRUE);

  okbutton8 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton8);
  gtk_dialog_add_action_widget (GTK_DIALOG (ctmappingeditdlg), okbutton8, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton8, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingeditdlg, ctmappingeditdlg, "ctmappingeditdlg");
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingeditdlg, dialog_vbox13, "dialog_vbox13");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, vbox40, "vbox40");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, hbox111, "hbox111");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, label137, "label137");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, content_type, "content_type");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, hbox112, "hbox112");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, label138, "label138");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, plugins, "plugins");
  GLADE_HOOKUP_OBJECT_NO_REF (ctmappingeditdlg, dialog_action_area12, "dialog_action_area12");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, cancelbutton8, "cancelbutton8");
  GLADE_HOOKUP_OBJECT (ctmappingeditdlg, okbutton8, "okbutton8");

  return ctmappingeditdlg;
}

GtkWidget*
create_button_properties (void)
{
  GtkWidget *button_properties;
  GtkWidget *dialog_vbox14;
  GtkWidget *vbox42;
  GtkWidget *hbox114;
  GtkWidget *use_color;
  GtkWidget *color;
  GtkWidget *hbox118;
  GtkWidget *use_textcolor;
  GtkWidget *textcolor;
  GtkWidget *hbox115;
  GtkWidget *label141;
  GtkWidget *icon;
  GtkWidget *hbox116;
  GtkWidget *label142;
  GtkWidget *label;
  GtkWidget *hbox117;
  GtkWidget *label143;
  GtkWidget *action;
  GtkWidget *dialog_action_area13;
  GtkWidget *applybutton1;
  GtkWidget *cancelbutton9;
  GtkWidget *okbutton9;

  button_properties = gtk_dialog_new ();
  gtk_widget_set_size_request (button_properties, -1, 250);
  gtk_window_set_title (GTK_WINDOW (button_properties), _("Button properties"));
  gtk_window_set_type_hint (GTK_WINDOW (button_properties), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox14 = gtk_dialog_get_content_area (GTK_DIALOG (button_properties));
  gtk_widget_show (dialog_vbox14);

  vbox42 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox42);
  gtk_box_pack_start (GTK_BOX (dialog_vbox14), vbox42, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox42), 12);

  hbox114 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox114);
  gtk_box_pack_start (GTK_BOX (vbox42), hbox114, FALSE, FALSE, 0);

  use_color = gtk_check_button_new_with_mnemonic (_("Color:"));
  gtk_widget_show (use_color);
  gtk_box_pack_start (GTK_BOX (hbox114), use_color, FALSE, FALSE, 0);

  color = gtk_color_button_new ();
  gtk_widget_show (color);
  gtk_box_pack_start (GTK_BOX (hbox114), color, FALSE, FALSE, 0);

  hbox118 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox118);
  gtk_box_pack_start (GTK_BOX (vbox42), hbox118, FALSE, FALSE, 0);

  use_textcolor = gtk_check_button_new_with_mnemonic (_("Text color:"));
  gtk_widget_show (use_textcolor);
  gtk_box_pack_start (GTK_BOX (hbox118), use_textcolor, FALSE, FALSE, 0);

  textcolor = gtk_color_button_new ();
  gtk_widget_show (textcolor);
  gtk_box_pack_start (GTK_BOX (hbox118), textcolor, FALSE, FALSE, 0);

  hbox115 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox115);
  gtk_box_pack_start (GTK_BOX (vbox42), hbox115, FALSE, TRUE, 0);

  label141 = gtk_label_new (_("Icon:"));
  gtk_widget_show (label141);
  gtk_box_pack_start (GTK_BOX (hbox115), label141, FALSE, FALSE, 0);

  icon = gtk_combo_box_text_new ();
  gtk_widget_show (icon);
  gtk_box_pack_start (GTK_BOX (hbox115), icon, TRUE, TRUE, 0);

  hbox116 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox116);
  gtk_box_pack_start (GTK_BOX (vbox42), hbox116, FALSE, TRUE, 0);

  label142 = gtk_label_new (_("Label:"));
  gtk_widget_show (label142);
  gtk_box_pack_start (GTK_BOX (hbox116), label142, FALSE, FALSE, 0);

  label = gtk_entry_new ();
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox116), label, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (label), 8226);

  hbox117 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox117);
  gtk_box_pack_start (GTK_BOX (vbox42), hbox117, FALSE, TRUE, 0);

  label143 = gtk_label_new (_("Action:"));
  gtk_widget_show (label143);
  gtk_box_pack_start (GTK_BOX (hbox117), label143, FALSE, FALSE, 0);

  action = gtk_button_new_with_mnemonic (_("<Not set>"));
  gtk_widget_show (action);
  gtk_box_pack_start (GTK_BOX (hbox117), action, TRUE, TRUE, 0);

  dialog_action_area13 = gtk_dialog_get_action_area (GTK_DIALOG (button_properties));
  gtk_widget_show (dialog_action_area13);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area13), GTK_BUTTONBOX_END);

  applybutton1 = gtk_button_new_from_stock ("gtk-apply");
  gtk_widget_show (applybutton1);
  gtk_dialog_add_action_widget (GTK_DIALOG (button_properties), applybutton1, GTK_RESPONSE_APPLY);
  gtk_widget_set_can_default(applybutton1, TRUE);

  cancelbutton9 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton9);
  gtk_dialog_add_action_widget (GTK_DIALOG (button_properties), cancelbutton9, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton9, TRUE);

  okbutton9 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton9);
  gtk_dialog_add_action_widget (GTK_DIALOG (button_properties), okbutton9, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton9, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (button_properties, button_properties, "button_properties");
  GLADE_HOOKUP_OBJECT_NO_REF (button_properties, dialog_vbox14, "dialog_vbox14");
  GLADE_HOOKUP_OBJECT (button_properties, vbox42, "vbox42");
  GLADE_HOOKUP_OBJECT (button_properties, hbox114, "hbox114");
  GLADE_HOOKUP_OBJECT (button_properties, use_color, "use_color");
  GLADE_HOOKUP_OBJECT (button_properties, color, "color");
  GLADE_HOOKUP_OBJECT (button_properties, hbox118, "hbox118");
  GLADE_HOOKUP_OBJECT (button_properties, use_textcolor, "use_textcolor");
  GLADE_HOOKUP_OBJECT (button_properties, textcolor, "textcolor");
  GLADE_HOOKUP_OBJECT (button_properties, hbox115, "hbox115");
  GLADE_HOOKUP_OBJECT (button_properties, label141, "label141");
  GLADE_HOOKUP_OBJECT (button_properties, icon, "icon");
  GLADE_HOOKUP_OBJECT (button_properties, hbox116, "hbox116");
  GLADE_HOOKUP_OBJECT (button_properties, label142, "label142");
  GLADE_HOOKUP_OBJECT (button_properties, label, "label");
  GLADE_HOOKUP_OBJECT (button_properties, hbox117, "hbox117");
  GLADE_HOOKUP_OBJECT (button_properties, label143, "label143");
  GLADE_HOOKUP_OBJECT (button_properties, action, "action");
  GLADE_HOOKUP_OBJECT_NO_REF (button_properties, dialog_action_area13, "dialog_action_area13");
  GLADE_HOOKUP_OBJECT (button_properties, applybutton1, "applybutton1");
  GLADE_HOOKUP_OBJECT (button_properties, cancelbutton9, "cancelbutton9");
  GLADE_HOOKUP_OBJECT (button_properties, okbutton9, "okbutton9");

  return button_properties;
}

GtkWidget*
create_select_action (void)
{
  GtkWidget *select_action;
  GtkWidget *dialog_vbox15;
  GtkWidget *vbox43;
  GtkWidget *scrolledwindow12;
  GtkWidget *actions;
  GtkWidget *dialog_action_area14;
  GtkWidget *cancelbutton10;
  GtkWidget *okbutton10;

  select_action = gtk_dialog_new ();
  gtk_widget_set_size_request (select_action, 494, 349);
  gtk_window_set_title (GTK_WINDOW (select_action), _("Select action"));
  gtk_window_set_type_hint (GTK_WINDOW (select_action), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox15 = gtk_dialog_get_content_area (GTK_DIALOG (select_action));
  gtk_widget_show (dialog_vbox15);

  vbox43 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox43);
  gtk_box_pack_start (GTK_BOX (dialog_vbox15), vbox43, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox43), 12);

  scrolledwindow12 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow12);
  gtk_box_pack_start (GTK_BOX (vbox43), scrolledwindow12, TRUE, TRUE, 0);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow12), GTK_SHADOW_IN);

  actions = gtk_tree_view_new ();
  gtk_widget_show (actions);
  gtk_container_add (GTK_CONTAINER (scrolledwindow12), actions);

  dialog_action_area14 = gtk_dialog_get_action_area (GTK_DIALOG (select_action));
  gtk_widget_show (dialog_action_area14);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area14), GTK_BUTTONBOX_END);

  cancelbutton10 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton10);
  gtk_dialog_add_action_widget (GTK_DIALOG (select_action), cancelbutton10, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton10, TRUE);

  okbutton10 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton10);
  gtk_dialog_add_action_widget (GTK_DIALOG (select_action), okbutton10, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton10, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (select_action, select_action, "select_action");
  GLADE_HOOKUP_OBJECT_NO_REF (select_action, dialog_vbox15, "dialog_vbox15");
  GLADE_HOOKUP_OBJECT (select_action, vbox43, "vbox43");
  GLADE_HOOKUP_OBJECT (select_action, scrolledwindow12, "scrolledwindow12");
  GLADE_HOOKUP_OBJECT (select_action, actions, "actions");
  GLADE_HOOKUP_OBJECT_NO_REF (select_action, dialog_action_area14, "dialog_action_area14");
  GLADE_HOOKUP_OBJECT (select_action, cancelbutton10, "cancelbutton10");
  GLADE_HOOKUP_OBJECT (select_action, okbutton10, "okbutton10");

  return select_action;
}

GtkWidget*
create_setcustomtitledlg (void)
{
  GtkWidget *setcustomtitledlg;
  GtkWidget *dialog_vbox16;
  GtkWidget *vbox46;
  GtkWidget *textview2;
  GtkWidget *hbox123;
  GtkWidget *set_custom_title;
  GtkWidget *custom_title;
  GtkWidget *dialog_action_area15;
  GtkWidget *cancelbutton11;
  GtkWidget *okbutton11;

  setcustomtitledlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (setcustomtitledlg), _("Set Custom Title"));
  gtk_window_set_modal (GTK_WINDOW (setcustomtitledlg), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (setcustomtitledlg), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox16 = gtk_dialog_get_content_area (GTK_DIALOG (setcustomtitledlg));
  gtk_widget_show (dialog_vbox16);

  vbox46 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox46);
  gtk_box_pack_start (GTK_BOX (dialog_vbox16), vbox46, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox46), 12);

  textview2 = gtk_text_view_new ();
  gtk_widget_show (textview2);
  gtk_box_pack_start (GTK_BOX (vbox46), textview2, FALSE, TRUE, 0);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview2), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview2), GTK_WRAP_WORD);
  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (textview2), FALSE);
  gtk_text_view_set_pixels_above_lines (GTK_TEXT_VIEW (textview2), 8);
  gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW (textview2), 8);
  gtk_text_view_set_left_margin (GTK_TEXT_VIEW (textview2), 8);
  gtk_text_view_set_right_margin (GTK_TEXT_VIEW (textview2), 8);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview2)), _("This dialog allows setting custom title for any track. This is most useful for radio stations. An option to set the custom title is also present in the \"Add Location\" dialog. The title itself is visible in columns displaying the \"Artist\" metadata field. It should look like \"[custom] artist\" if the Artist field is present, or just \"custom\" otherwise."), -1);

  hbox123 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox123);
  gtk_box_pack_start (GTK_BOX (vbox46), hbox123, FALSE, TRUE, 0);

  set_custom_title = gtk_check_button_new_with_mnemonic (_("Set Custom Title"));
  gtk_widget_show (set_custom_title);
  gtk_box_pack_start (GTK_BOX (hbox123), set_custom_title, FALSE, FALSE, 0);

  custom_title = gtk_entry_new ();
  gtk_widget_show (custom_title);
  gtk_box_pack_start (GTK_BOX (hbox123), custom_title, TRUE, TRUE, 0);
  gtk_entry_set_invisible_char (GTK_ENTRY (custom_title), 8226);

  dialog_action_area15 = gtk_dialog_get_action_area (GTK_DIALOG (setcustomtitledlg));
  gtk_widget_show (dialog_action_area15);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area15), GTK_BUTTONBOX_END);

  cancelbutton11 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton11);
  gtk_dialog_add_action_widget (GTK_DIALOG (setcustomtitledlg), cancelbutton11, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton11, TRUE);

  okbutton11 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton11);
  gtk_dialog_add_action_widget (GTK_DIALOG (setcustomtitledlg), okbutton11, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton11, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (setcustomtitledlg, setcustomtitledlg, "setcustomtitledlg");
  GLADE_HOOKUP_OBJECT_NO_REF (setcustomtitledlg, dialog_vbox16, "dialog_vbox16");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, vbox46, "vbox46");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, textview2, "textview2");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, hbox123, "hbox123");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, set_custom_title, "set_custom_title");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, custom_title, "custom_title");
  GLADE_HOOKUP_OBJECT_NO_REF (setcustomtitledlg, dialog_action_area15, "dialog_action_area15");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, cancelbutton11, "cancelbutton11");
  GLADE_HOOKUP_OBJECT (setcustomtitledlg, okbutton11, "okbutton11");

  return setcustomtitledlg;
}

GtkWidget*
create_trkproperties_popup_menu (void)
{
  GtkWidget *trkproperties_popup_menu;
  GtkWidget *trkproperties_edit;
  GtkWidget *trkproperties_edit_in_place;
  GtkWidget *trkproperties_remove;
  GtkWidget *trkproperties_crop;
  GtkWidget *trkproperties_add_new_field;

  trkproperties_popup_menu = gtk_menu_new ();

  trkproperties_edit = gtk_menu_item_new_with_mnemonic (_("Edit"));
  gtk_widget_show (trkproperties_edit);
  gtk_container_add (GTK_CONTAINER (trkproperties_popup_menu), trkproperties_edit);

  trkproperties_edit_in_place = gtk_menu_item_new_with_mnemonic (_("Edit (in place)"));
  gtk_widget_show (trkproperties_edit_in_place);
  gtk_container_add (GTK_CONTAINER (trkproperties_popup_menu), trkproperties_edit_in_place);

  trkproperties_remove = gtk_menu_item_new_with_mnemonic (_("Remove"));
  gtk_widget_show (trkproperties_remove);
  gtk_container_add (GTK_CONTAINER (trkproperties_popup_menu), trkproperties_remove);

  trkproperties_crop = gtk_menu_item_new_with_mnemonic (_("Crop"));
  gtk_widget_show (trkproperties_crop);
  gtk_container_add (GTK_CONTAINER (trkproperties_popup_menu), trkproperties_crop);

  trkproperties_add_new_field = gtk_menu_item_new_with_mnemonic (_("Add new field..."));
  gtk_widget_show (trkproperties_add_new_field);
  gtk_container_add (GTK_CONTAINER (trkproperties_popup_menu), trkproperties_add_new_field);

  g_signal_connect ((gpointer) trkproperties_edit, "activate",
                    G_CALLBACK (on_trkproperties_edit_activate),
                    NULL);
  g_signal_connect ((gpointer) trkproperties_edit_in_place, "activate",
                    G_CALLBACK (on_trkproperties_edit_in_place_activate),
                    NULL);
  g_signal_connect ((gpointer) trkproperties_remove, "activate",
                    G_CALLBACK (on_trkproperties_remove_activate),
                    NULL);
  g_signal_connect ((gpointer) trkproperties_crop, "activate",
                    G_CALLBACK (on_trkproperties_crop_activate),
                    NULL);
  g_signal_connect ((gpointer) trkproperties_add_new_field, "activate",
                    G_CALLBACK (on_trkproperties_add_new_field_activate),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (trkproperties_popup_menu, trkproperties_popup_menu, "trkproperties_popup_menu");
  GLADE_HOOKUP_OBJECT (trkproperties_popup_menu, trkproperties_edit, "trkproperties_edit");
  GLADE_HOOKUP_OBJECT (trkproperties_popup_menu, trkproperties_edit_in_place, "trkproperties_edit_in_place");
  GLADE_HOOKUP_OBJECT (trkproperties_popup_menu, trkproperties_remove, "trkproperties_remove");
  GLADE_HOOKUP_OBJECT (trkproperties_popup_menu, trkproperties_crop, "trkproperties_crop");
  GLADE_HOOKUP_OBJECT (trkproperties_popup_menu, trkproperties_add_new_field, "trkproperties_add_new_field");

  return trkproperties_popup_menu;
}

GtkWidget*
create_edit_tag_value_dlg (void)
{
  GtkWidget *edit_tag_value_dlg;
  GtkWidget *dialog_vbox17;
  GtkWidget *vbox47;
  GtkWidget *label160;
  GtkWidget *field_name;
  GtkWidget *label161;
  GtkWidget *scrolledwindow13;
  GtkWidget *field_value;
  GtkWidget *label187;
  GtkWidget *dialog_action_area16;
  GtkWidget *cancelbutton12;
  GtkWidget *okbutton12;

  edit_tag_value_dlg = gtk_dialog_new ();
  gtk_widget_set_size_request (edit_tag_value_dlg, 300, 250);
  gtk_window_set_title (GTK_WINDOW (edit_tag_value_dlg), _("Edit Value"));
  gtk_window_set_type_hint (GTK_WINDOW (edit_tag_value_dlg), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_dialog_set_has_separator (GTK_DIALOG (edit_tag_value_dlg), FALSE);

  dialog_vbox17 = gtk_dialog_get_content_area (GTK_DIALOG (edit_tag_value_dlg));
  gtk_widget_show (dialog_vbox17);

  vbox47 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox47);
  gtk_box_pack_start (GTK_BOX (dialog_vbox17), vbox47, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox47), 12);
  gtk_widget_set_can_focus(vbox47, TRUE);

  label160 = gtk_label_new (_("Field name:"));
  gtk_widget_show (label160);
  gtk_box_pack_start (GTK_BOX (vbox47), label160, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label160), 0, 0.5);

  field_name = gtk_entry_new ();
  gtk_widget_show (field_name);
  gtk_box_pack_start (GTK_BOX (vbox47), field_name, FALSE, FALSE, 0);
  gtk_widget_set_can_focus(field_name, FALSE);
  gtk_editable_set_editable (GTK_EDITABLE (field_name), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (field_name), 8226);

  label161 = gtk_label_new (_("Field value:"));
  gtk_widget_show (label161);
  gtk_box_pack_start (GTK_BOX (vbox47), label161, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label161), 0, 0.5);

  scrolledwindow13 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow13);
  gtk_box_pack_start (GTK_BOX (vbox47), scrolledwindow13, TRUE, TRUE, 0);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow13), GTK_SHADOW_IN);

  field_value = gtk_text_view_new ();
  gtk_widget_show (field_value);
  gtk_container_add (GTK_CONTAINER (scrolledwindow13), field_value);

  label187 = gtk_label_new (_("Use ; character to separate multiple values."));
  gtk_widget_show (label187);
  gtk_box_pack_start (GTK_BOX (vbox47), label187, FALSE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (label187), 0, 0.5);

  dialog_action_area16 = gtk_dialog_get_action_area (GTK_DIALOG (edit_tag_value_dlg));
  gtk_widget_show (dialog_action_area16);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area16), GTK_BUTTONBOX_END);

  cancelbutton12 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton12);
  gtk_dialog_add_action_widget (GTK_DIALOG (edit_tag_value_dlg), cancelbutton12, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton12, TRUE);

  okbutton12 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton12);
  gtk_dialog_add_action_widget (GTK_DIALOG (edit_tag_value_dlg), okbutton12, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton12, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (edit_tag_value_dlg, edit_tag_value_dlg, "edit_tag_value_dlg");
  GLADE_HOOKUP_OBJECT_NO_REF (edit_tag_value_dlg, dialog_vbox17, "dialog_vbox17");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, vbox47, "vbox47");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, label160, "label160");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, field_name, "field_name");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, label161, "label161");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, scrolledwindow13, "scrolledwindow13");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, field_value, "field_value");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, label187, "label187");
  GLADE_HOOKUP_OBJECT_NO_REF (edit_tag_value_dlg, dialog_action_area16, "dialog_action_area16");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, cancelbutton12, "cancelbutton12");
  GLADE_HOOKUP_OBJECT (edit_tag_value_dlg, okbutton12, "okbutton12");

  gtk_widget_grab_focus (field_value);
  return edit_tag_value_dlg;
}

GtkWidget*
create_log_window (void)
{
  GtkWidget *log_window;
  GtkWidget *vbox49;
  GtkWidget *scrolledwindow14;
  GtkWidget *logwindow_textview;
  GtkWidget *hbox140;
  GtkWidget *autoopen_button;
  GtkWidget *log_clear;

  log_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_events (log_window, GDK_KEY_PRESS_MASK);
  gtk_window_set_title (GTK_WINDOW (log_window), _("DeaDBeeF Log"));
  gtk_window_set_default_size (GTK_WINDOW (log_window), 400, 200);
  gtk_window_set_destroy_with_parent (GTK_WINDOW (log_window), TRUE);
  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (log_window), TRUE);
  gtk_window_set_skip_pager_hint (GTK_WINDOW (log_window), TRUE);

  vbox49 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox49);
  gtk_container_add (GTK_CONTAINER (log_window), vbox49);

  scrolledwindow14 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow14);
  gtk_box_pack_start (GTK_BOX (vbox49), scrolledwindow14, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow14), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow14), GTK_SHADOW_ETCHED_IN);

  logwindow_textview = gtk_text_view_new ();
  gtk_widget_show (logwindow_textview);
  gtk_container_add (GTK_CONTAINER (scrolledwindow14), logwindow_textview);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (logwindow_textview), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (logwindow_textview), GTK_WRAP_WORD);
  gtk_text_view_set_left_margin (GTK_TEXT_VIEW (logwindow_textview), 10);
  gtk_text_view_set_right_margin (GTK_TEXT_VIEW (logwindow_textview), 10);

  hbox140 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox140);
  gtk_box_pack_start (GTK_BOX (vbox49), hbox140, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox140), 5);

  autoopen_button = gtk_check_button_new_with_mnemonic (_("Auto-open this window on errors"));
  gtk_widget_show (autoopen_button);
  gtk_box_pack_start (GTK_BOX (hbox140), autoopen_button, FALSE, FALSE, 0);

  log_clear = gtk_button_new_with_mnemonic (_("Clear"));
  gtk_widget_show (log_clear);
  gtk_box_pack_end (GTK_BOX (hbox140), log_clear, FALSE, FALSE, 0);

  g_signal_connect ((gpointer) log_window, "key_press_event",
                    G_CALLBACK (on_log_window_key_press_event),
                    NULL);
  g_signal_connect ((gpointer) autoopen_button, "toggled",
                    G_CALLBACK (on_autoopen_button_toggled),
                    NULL);
  g_signal_connect ((gpointer) log_clear, "clicked",
                    G_CALLBACK (on_log_clear_clicked),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (log_window, log_window, "log_window");
  GLADE_HOOKUP_OBJECT (log_window, vbox49, "vbox49");
  GLADE_HOOKUP_OBJECT (log_window, scrolledwindow14, "scrolledwindow14");
  GLADE_HOOKUP_OBJECT (log_window, logwindow_textview, "logwindow_textview");
  GLADE_HOOKUP_OBJECT (log_window, hbox140, "hbox140");
  GLADE_HOOKUP_OBJECT (log_window, autoopen_button, "autoopen_button");
  GLADE_HOOKUP_OBJECT (log_window, log_clear, "log_clear");

  return log_window;
}

GtkWidget*
create_rg_scan_progress (void)
{
  GtkWidget *rg_scan_progress;
  GtkWidget *vbox50;
  GtkWidget *hbox142;
  GtkWidget *label166;
  GtkWidget *rg_scan_progress_file;
  GtkWidget *rg_scan_progress_bar;
  GtkWidget *hbox143;
  GtkWidget *rg_scan_progress_status;
  GtkWidget *rg_scan_progress_cancel;

  rg_scan_progress = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (rg_scan_progress, 650, -1);
  gtk_window_set_title (GTK_WINDOW (rg_scan_progress), _("ReplayGain Scan Progress"));

  vbox50 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox50);
  gtk_container_add (GTK_CONTAINER (rg_scan_progress), vbox50);
  gtk_container_set_border_width (GTK_CONTAINER (vbox50), 12);

  hbox142 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox142);
  gtk_box_pack_start (GTK_BOX (vbox50), hbox142, FALSE, TRUE, 0);

  label166 = gtk_label_new (_("Current File:"));
  gtk_widget_show (label166);
  gtk_box_pack_start (GTK_BOX (hbox142), label166, FALSE, FALSE, 0);

  rg_scan_progress_file = gtk_entry_new ();
  gtk_widget_show (rg_scan_progress_file);
  gtk_box_pack_start (GTK_BOX (hbox142), rg_scan_progress_file, TRUE, TRUE, 0);
  gtk_editable_set_editable (GTK_EDITABLE (rg_scan_progress_file), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (rg_scan_progress_file), 8226);

  rg_scan_progress_bar = gtk_progress_bar_new ();
  gtk_widget_show (rg_scan_progress_bar);
  gtk_box_pack_start (GTK_BOX (vbox50), rg_scan_progress_bar, FALSE, FALSE, 0);

  hbox143 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox143);
  gtk_box_pack_start (GTK_BOX (vbox50), hbox143, FALSE, TRUE, 0);

  rg_scan_progress_status = gtk_label_new (_("Status"));
  gtk_widget_show (rg_scan_progress_status);
  gtk_box_pack_start (GTK_BOX (hbox143), rg_scan_progress_status, TRUE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (rg_scan_progress_status), 0, 0.5);
  gtk_label_set_ellipsize (GTK_LABEL (rg_scan_progress_status), PANGO_ELLIPSIZE_END);
  gtk_label_set_single_line_mode (GTK_LABEL (rg_scan_progress_status), TRUE);

  rg_scan_progress_cancel = gtk_button_new_with_mnemonic (_("Cancel"));
  gtk_widget_show (rg_scan_progress_cancel);
  gtk_box_pack_start (GTK_BOX (hbox143), rg_scan_progress_cancel, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (rg_scan_progress, rg_scan_progress, "rg_scan_progress");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, vbox50, "vbox50");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, hbox142, "hbox142");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, label166, "label166");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, rg_scan_progress_file, "rg_scan_progress_file");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, rg_scan_progress_bar, "rg_scan_progress_bar");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, hbox143, "hbox143");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, rg_scan_progress_status, "rg_scan_progress_status");
  GLADE_HOOKUP_OBJECT (rg_scan_progress, rg_scan_progress_cancel, "rg_scan_progress_cancel");

  return rg_scan_progress;
}

GtkWidget*
create_rg_scan_results (void)
{
  GtkWidget *rg_scan_results;
  GtkWidget *vbox51;
  GtkWidget *scrolledwindow15;
  GtkWidget *rg_scan_results_list;
  GtkWidget *hbox144;
  GtkWidget *rg_scan_results_status;
  GtkWidget *rg_scan_results_update;
  GtkWidget *rg_scan_results_cancel;

  rg_scan_results = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (rg_scan_results, 1000, 350);
  gtk_window_set_title (GTK_WINDOW (rg_scan_results), _("ReplayGain Scan Results"));

  vbox51 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox51);
  gtk_container_add (GTK_CONTAINER (rg_scan_results), vbox51);
  gtk_container_set_border_width (GTK_CONTAINER (vbox51), 12);

  scrolledwindow15 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow15);
  gtk_box_pack_start (GTK_BOX (vbox51), scrolledwindow15, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow15), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow15), GTK_SHADOW_IN);

  rg_scan_results_list = gtk_tree_view_new ();
  gtk_widget_show (rg_scan_results_list);
  gtk_container_add (GTK_CONTAINER (scrolledwindow15), rg_scan_results_list);

  hbox144 = gtk_hbox_new (FALSE, 8);
  gtk_widget_show (hbox144);
  gtk_box_pack_start (GTK_BOX (vbox51), hbox144, FALSE, TRUE, 0);

  rg_scan_results_status = gtk_label_new (_("Status"));
  gtk_widget_show (rg_scan_results_status);
  gtk_box_pack_start (GTK_BOX (hbox144), rg_scan_results_status, TRUE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (rg_scan_results_status), 0, 0.5);
  gtk_label_set_ellipsize (GTK_LABEL (rg_scan_results_status), PANGO_ELLIPSIZE_END);
  gtk_label_set_single_line_mode (GTK_LABEL (rg_scan_results_status), TRUE);

  rg_scan_results_update = gtk_button_new_with_mnemonic (_("Update File Tags"));
  gtk_widget_show (rg_scan_results_update);
  gtk_box_pack_start (GTK_BOX (hbox144), rg_scan_results_update, FALSE, FALSE, 0);

  rg_scan_results_cancel = gtk_button_new_with_mnemonic (_("Cancel"));
  gtk_widget_show (rg_scan_results_cancel);
  gtk_box_pack_start (GTK_BOX (hbox144), rg_scan_results_cancel, FALSE, FALSE, 0);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (rg_scan_results, rg_scan_results, "rg_scan_results");
  GLADE_HOOKUP_OBJECT (rg_scan_results, vbox51, "vbox51");
  GLADE_HOOKUP_OBJECT (rg_scan_results, scrolledwindow15, "scrolledwindow15");
  GLADE_HOOKUP_OBJECT (rg_scan_results, rg_scan_results_list, "rg_scan_results_list");
  GLADE_HOOKUP_OBJECT (rg_scan_results, hbox144, "hbox144");
  GLADE_HOOKUP_OBJECT (rg_scan_results, rg_scan_results_status, "rg_scan_results_status");
  GLADE_HOOKUP_OBJECT (rg_scan_results, rg_scan_results_update, "rg_scan_results_update");
  GLADE_HOOKUP_OBJECT (rg_scan_results, rg_scan_results_cancel, "rg_scan_results_cancel");

  return rg_scan_results;
}

GtkWidget*
create_plugin_list_popup_menu (void)
{
  GtkWidget *plugin_list_popup_menu;
  GtkWidget *copy_plugin_report_menuitem;
  GtkWidget *image688;
  GtkWidget *only_show_plugins_with_configuration1;

  plugin_list_popup_menu = gtk_menu_new ();

  copy_plugin_report_menuitem = gtk_image_menu_item_new_with_mnemonic (_("Copy report"));
  gtk_widget_show (copy_plugin_report_menuitem);
  gtk_container_add (GTK_CONTAINER (plugin_list_popup_menu), copy_plugin_report_menuitem);

  image688 = gtk_image_new_from_stock ("gtk-copy", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image688);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (copy_plugin_report_menuitem), image688);

  only_show_plugins_with_configuration1 = gtk_check_menu_item_new_with_mnemonic (_("Only show plugins with configuration"));
  gtk_widget_show (only_show_plugins_with_configuration1);
  gtk_container_add (GTK_CONTAINER (plugin_list_popup_menu), only_show_plugins_with_configuration1);

  g_signal_connect ((gpointer) copy_plugin_report_menuitem, "activate",
                    G_CALLBACK (on_copy_plugin_report_menuitem_activate),
                    NULL);
  g_signal_connect ((gpointer) only_show_plugins_with_configuration1, "activate",
                    G_CALLBACK (on_only_show_plugins_with_configuration1_activate),
                    NULL);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (plugin_list_popup_menu, plugin_list_popup_menu, "plugin_list_popup_menu");
  GLADE_HOOKUP_OBJECT (plugin_list_popup_menu, copy_plugin_report_menuitem, "copy_plugin_report_menuitem");
  GLADE_HOOKUP_OBJECT (plugin_list_popup_menu, image688, "image688");
  GLADE_HOOKUP_OBJECT (plugin_list_popup_menu, only_show_plugins_with_configuration1, "only_show_plugins_with_configuration1");

  return plugin_list_popup_menu;
}

GtkWidget*
create_edit_multiple_tracks_dialog (void)
{
  GtkWidget *edit_multiple_tracks_dialog;
  GtkWidget *dialog_vbox18;
  GtkWidget *vbox60;
  GtkWidget *label183;
  GtkWidget *entry_field_name;
  GtkWidget *notebook;
  GtkWidget *scrolledwindow22;
  GtkWidget *textview_single_value;
  GtkWidget *label184;
  GtkWidget *scrolledwindow23;
  GtkWidget *treeview_individual_values;
  GtkWidget *label185;
  GtkWidget *label186;
  GtkWidget *dialog_action_area17;
  GtkWidget *cancelbutton13;
  GtkWidget *okbutton13;

  edit_multiple_tracks_dialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (edit_multiple_tracks_dialog), _("Edit Multiple Tracks"));
  gtk_window_set_type_hint (GTK_WINDOW (edit_multiple_tracks_dialog), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_dialog_set_has_separator (GTK_DIALOG (edit_multiple_tracks_dialog), FALSE);

  dialog_vbox18 = gtk_dialog_get_content_area (GTK_DIALOG (edit_multiple_tracks_dialog));
  gtk_widget_show (dialog_vbox18);

  vbox60 = gtk_vbox_new (FALSE, 8);
  gtk_widget_show (vbox60);
  gtk_box_pack_start (GTK_BOX (dialog_vbox18), vbox60, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox60), 12);

  label183 = gtk_label_new (_("Field Name:"));
  gtk_widget_show (label183);
  gtk_box_pack_start (GTK_BOX (vbox60), label183, FALSE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (label183), 0, 0.5);

  entry_field_name = gtk_entry_new ();
  gtk_widget_show (entry_field_name);
  gtk_box_pack_start (GTK_BOX (vbox60), entry_field_name, FALSE, FALSE, 0);
  gtk_widget_set_can_focus(entry_field_name, FALSE);
  gtk_editable_set_editable (GTK_EDITABLE (entry_field_name), FALSE);
  gtk_entry_set_invisible_char (GTK_ENTRY (entry_field_name), 9679);

  notebook = gtk_notebook_new ();
  gtk_widget_show (notebook);
  gtk_box_pack_start (GTK_BOX (vbox60), notebook, TRUE, TRUE, 0);

  scrolledwindow22 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow22);
  gtk_container_add (GTK_CONTAINER (notebook), scrolledwindow22);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow22), GTK_SHADOW_IN);

  textview_single_value = gtk_text_view_new ();
  gtk_widget_show (textview_single_value);
  gtk_container_add (GTK_CONTAINER (scrolledwindow22), textview_single_value);

  label184 = gtk_label_new (_("Single Value"));
  gtk_widget_show (label184);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 0), label184);

  scrolledwindow23 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow23);
  gtk_container_add (GTK_CONTAINER (notebook), scrolledwindow23);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow23), GTK_SHADOW_IN);

  treeview_individual_values = gtk_tree_view_new ();
  gtk_widget_show (treeview_individual_values);
  gtk_container_add (GTK_CONTAINER (scrolledwindow23), treeview_individual_values);

  label185 = gtk_label_new (_("Individual Values"));
  gtk_widget_show (label185);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1), label185);

  label186 = gtk_label_new (_("Use ; character to separate multiple values."));
  gtk_widget_show (label186);
  gtk_box_pack_start (GTK_BOX (vbox60), label186, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (label186), 0, 0.5);

  dialog_action_area17 = gtk_dialog_get_action_area (GTK_DIALOG (edit_multiple_tracks_dialog));
  gtk_widget_show (dialog_action_area17);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area17), GTK_BUTTONBOX_END);

  cancelbutton13 = gtk_button_new_from_stock ("gtk-cancel");
  gtk_widget_show (cancelbutton13);
  gtk_dialog_add_action_widget (GTK_DIALOG (edit_multiple_tracks_dialog), cancelbutton13, GTK_RESPONSE_CANCEL);
  gtk_widget_set_can_default(cancelbutton13, TRUE);

  okbutton13 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton13);
  gtk_dialog_add_action_widget (GTK_DIALOG (edit_multiple_tracks_dialog), okbutton13, GTK_RESPONSE_OK);
  gtk_widget_set_can_default(okbutton13, TRUE);

  /* Store pointers to all widgets, for use by lookup_widget(). */
  GLADE_HOOKUP_OBJECT_NO_REF (edit_multiple_tracks_dialog, edit_multiple_tracks_dialog, "edit_multiple_tracks_dialog");
  GLADE_HOOKUP_OBJECT_NO_REF (edit_multiple_tracks_dialog, dialog_vbox18, "dialog_vbox18");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, vbox60, "vbox60");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, label183, "label183");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, entry_field_name, "entry_field_name");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, notebook, "notebook");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, scrolledwindow22, "scrolledwindow22");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, textview_single_value, "textview_single_value");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, label184, "label184");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, scrolledwindow23, "scrolledwindow23");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, treeview_individual_values, "treeview_individual_values");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, label185, "label185");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, label186, "label186");
  GLADE_HOOKUP_OBJECT_NO_REF (edit_multiple_tracks_dialog, dialog_action_area17, "dialog_action_area17");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, cancelbutton13, "cancelbutton13");
  GLADE_HOOKUP_OBJECT (edit_multiple_tracks_dialog, okbutton13, "okbutton13");

  gtk_widget_grab_focus (textview_single_value);
  return edit_multiple_tracks_dialog;
}

